/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#ifndef SM_COMMAND_UTILITY_H_
#define SM_COMMAND_UTILITY_H_

#include <nn/gx.h>
#include "../include/SmBase.h"

class SmStaticCommandCache;

//---------------------------------------------------------------------------
// Class for creating and re-using commands.
class SmCommandReuser
{
    friend class SmStaticCommandCache;

public:
    SmCommandReuser()
    : m_StoreCommandListId(0),
      m_CommandListId(0),
      m_BufferOffset(0),
      m_BufferSize(0),
      m_RequestId(0),
      m_RequestSize(0)
    {}

    // Start creating cache.
    void StartCaching(GLuint cachingCommandListId, bool clearCmdCache = true)
    {
        // Store currently bound command list.
        nngxGetCmdlistParameteri( NN_GX_CMDLIST_BINDING, &m_StoreCommandListId );

        // Switch to cache command list.
        m_CommandListId = cachingCommandListId;
        nngxBindCmdlist(cachingCommandListId);

         if ( clearCmdCache )
        {
            nngxClearCmdlist();
        }
        nngxStartCmdlistSave();
        NW_GL_ASSERT();
    }

    // Stop creating cache.
    void EndCaching()
    {
//        nngxSplitDrawCmdlist();
        NW_GL_ASSERT();

        nngxStopCmdlistSave(
            &this->m_BufferOffset,
            &this->m_BufferSize,
            &this->m_RequestId,
            &this->m_RequestSize);
        NW_GL_ASSERT();

        nngxBindCmdlist( m_StoreCommandListId );
    }

    // Use cached commands.
    void UseCommand( bool isCopyBuffer )
    {
        GLboolean copyBuffer = isCopyBuffer ? GL_TRUE : GL_FALSE;
        nngxUseSavedCmdlist(
            this->m_CommandListId,
            this->m_BufferOffset,
            this->m_BufferSize,
            this->m_RequestId,
            this->m_RequestSize,
            0,
            copyBuffer);
        NW_GL_ASSERT();
    }

    // Is it cached?
    bool IsCache()
    {
         if ( m_BufferSize > 0 )
        {
            return true;
        }
        else
        {
            return false;
        }
    }

    // Get the size of cached data.
    GLsizei GetCacheSize()
    {
        return m_BufferSize;
    }

private:


    GLint       m_StoreCommandListId;
    GLuint      m_CommandListId;
    GLuint      m_BufferOffset;
    GLsizei     m_BufferSize;
    GLuint      m_RequestId;
    GLsizei     m_RequestSize;
};


//---------------------------------------------------------------------------
// Class for creating a static command cache and re-using commands by cutting regions out of it.
//
class SmStaticCommandCache : public SmBase
{
public:
    struct CopiedCmdCache
    {
        GLvoid* cmdBuff;
        GLsizei cmdBuffSize;

        bool IsCopied()
        {
             if ( cmdBuffSize )
            {
                return true;
            }

            return false;
        }

        void Init()
        {
            cmdBuff = NULL;
            cmdBuffSize = 0;
        }
    };

    // Constructor
    SmStaticCommandCache();

    // Destructor
    ~SmStaticCommandCache();

    static void StartStaticCmdCache( SmCommandReuser& cmdReuser );
    static void EndStaticCmdCache( SmCommandReuser& cmdReuser );

    static void CreateCopiedCmdCache( CopiedCmdCache& cmdCacheCopied );
    static void DestroyCopiedCmdCache( CopiedCmdCache& cmdCacheCopied );
    static void UseCopiedCmdCache( CopiedCmdCache& cmdCacheCopied );

private:
    static GLuint               m_StaticCommandList;
};


#endif // SM_COMMAND_UTILITY_H_
