/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/
#ifndef SM_PERF_H_
#define SM_PERF_H_

//#include <nw/demo.h>

#include "../include/SmPrimitive.h"
#include "../include/SmCommandUtility.h"

#if NN_CURRENT_VERSION_NUMBER >= NN_VERSION_NUMBER(4,0,0,0)
#pragma diag_suppress 1301 // padding inserted in struct.
#pragma diag_suppress 2530 // padding added to end of struct.
#endif
//------------------------------------------------------------------------------
//
//
struct Cost
{
    nn::os::Tick    costStart;
    nn::os::Tick    costEnd;
    char            costName[64];

    uint            gpuStart;
    uint            gpuEnd;

    nw::ut::Color8  costColor;

    void init()
    {
        std::memset( costName, 0, sizeof(costName) );
        gpuStart  = 0;
        gpuEnd    = 0;
        costColor = 0;
    }
};

typedef nw::ut::FixedSizeArray<Cost, 32> CostArray;

//------------------------------------------------------------------------------
// Class that measures processing cost.
class SmPerf
{
public:
    static const int BUFFER_COUNT = 3;

    SmPerf()
    : m_CurrentBuffer( 0 ),
      m_CurrentCost( -1 )
    {
        m_CostArray[0].clear();
        m_CostArray[1].clear();
        m_CostArray[2].clear();
    }

    ~SmPerf()
    {
        m_CostArray[0].clear();
        m_CostArray[1].clear();
        m_CostArray[2].clear();
    }

    // Starts measurements.
    virtual void Begin( const char* costName, u32 costColor )
    {
        NW_UNUSED_VARIABLE(costName);
        NW_UNUSED_VARIABLE(costColor);
    }

    // Ends measures.
    virtual void End(){}

    // Resets
    // Resets the current frame results due to the frame end.
    virtual void Reset(nn::os::Tick){}


    // Gets the current time.
    nn::os::Tick GetTime()
    {
        return nn::os::Tick::GetSystemCurrent();
    }


    // Gets the current number of registered costs.
    uint GetCostCount()
    {
        return m_CostArray[IdxResult()].size();
    }


    // Gets the current registered cost names.
    const char* GetCostName( uint costNo )
    {
        return m_CostArray[IdxResult()][costNo].costName;
    }


    // Gets the current cost colors.
    nw::ut::Color8 GetCostColor( uint costNo )
    {
        return m_CostArray[IdxResult()][costNo].costColor;
    }


    // Gets the current registered costs (msec).

    nn::os::Tick GetCostStart( uint costNo ) const
    {
        if ( m_CostArray[IdxResult()].size() > costNo )
        {
            return m_CostArray[IdxResult()][costNo].costStart;
        }

        return nn::os::Tick();
    }

    nn::os::Tick GetCostEnd( uint costNo ) const
    {
        if ( m_CostArray[IdxResult()].size() > costNo )
        {
            return m_CostArray[IdxResult()][costNo].costEnd;
        }

        return nn::os::Tick();
    }

    nn::os::Tick GetFrameStart() const
    {
        return m_FrameStart[IdxResult()];
    }

protected:
    CostArray       m_CostArray[BUFFER_COUNT];
    nn::os::Tick    m_FrameStart[BUFFER_COUNT];
    uint            m_CurrentBuffer;
    s32             m_CurrentCost;

    int IdxPrepare() const
    {
        return (m_CurrentBuffer + 1) % BUFFER_COUNT;
    }
    int IdxCurrent() const
    {
        return m_CurrentBuffer;
    }
    int IdxResult() const
    {
        return (m_CurrentBuffer - 1 + BUFFER_COUNT) % BUFFER_COUNT;
    }

    void SwapBuffer()
    {
        m_CurrentBuffer = IdxPrepare();

        m_CostArray[IdxPrepare()].clear();
        m_CurrentCost = -1;
    }
};




//------------------------------------------------------------------------------
// Class that measures CPU processing cost.
class SmCpuPerf : public SmPerf
{
public:
    // Constructor
    SmCpuPerf();

    // Starts measurements.
    virtual void Begin( const char* costName, u32 costColor );

    // Ends measures.
    virtual void End();

    // Resets
    virtual void Reset(nn::os::Tick startTime);
};


//------------------------------------------------------------------------------
// Class that measures GPU processing cost.
class SmGpuPerf : public SmPerf
{
public:
    // Constructor
    SmGpuPerf();

    // Destructor
    ~SmGpuPerf();

    // Starts measurements.
    virtual void Begin( const char* costName, u32 costColor );

    // Ends measures.
    virtual void End();

    void SetCallback() const;

    // The callback function to be registered.
    void Invoke(GLint id);
    static void Callback(GLint id);

    // Resets
    virtual void Reset(nn::os::Tick startTime);

    // Calls when the command list is launched.
    void SetCmdlistRun();

private:
    f32     m_MillisecCost;
    uint    m_StartReqCount;
    uint    m_EndReqCount;
};


//------------------------------------------------------------------------------
// Class that renders costs as measured results.
class SmPerfDrawer : public SmBase
{
    // 1frame width
    #define PERF_ONE_FRAME_WIDTH    (160.f)

public:
    SmPerfDrawer();
    ~SmPerfDrawer();

    void Render( SmPerf* cpuPerf, SmPerf* gpuPerf );
    void RenderBar();

private:
    Sm2DDynamicPrimPC       m_PerfBar;
    Sm2DPrimPC*             m_pPerfBarSplit;
    SmCommandReuser         m_PerfBarSplitCmd;
};
#if NN_CURRENT_VERSION_NUMBER >= NN_VERSION_NUMBER(4,0,0,0)
#pragma diag_default 2530 // padding added to end of struct.
#pragma diag_default 1301 // padding inserted in struct.
#endif


#endif  // SM_PERF_H_
