/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/
#ifndef FRAMEWORK_H__
#define FRAMEWORK_H__

#include "../include/SmMessage.h"

#if NN_CURRENT_VERSION_NUMBER >= NN_VERSION_NUMBER(4,0,0,0)
#pragma diag_suppress 1301 // padding inserted in struct.
#pragma diag_suppress 2530 // padding added to end of struct.
#endif
//------------------------------------------------------------------------------
//
class FrameWorkStatus
{
public:
    static const int STEPFRAME_MIN = 1; // 60fps maximum.
    static const int STEPFRAME_MAX = 6; // 10fps minimum

    //
    FrameWorkStatus()
    {
        m_IsCommandDouble   = true;
        m_IsLayoutVisible   = true;
        m_IsCpuMeterVisible = true;
        m_IsPalying         = true;
        m_IsFirstLoop       = true;
        m_IsAddFrame        = false;

        m_Frame = 0.f;
        m_LastVSyncCounter = 0;
    }

    // Switches command buffer mode.
    void SwapCommandBufferMode()
    {
        if ( m_IsCommandDouble )
        {
            m_IsCommandDouble = false;
            NW_DEV_LOG( "[DEMO] Command Buffer Single\n\n" );
        }
        else
        {
            m_IsCommandDouble = true;
            NW_DEV_LOG( "[DEMO] Command Buffer Double\n\n" );
        }
    }

    // Switches layout rendering on or off.
    void SwapLayoutVisible()
    {
        if ( m_IsLayoutVisible )
        {
            m_IsLayoutVisible = false;
            NW_DEV_LOG( "[DEMO] Layout Draw Off\n\n" );
        }
        else
        {
            m_IsLayoutVisible = true;
            NW_DEV_LOG( "[DEMO] Layout Draw On\n\n" );
        }
    }

    // Switches OpenGL on or off.
    void SwapCpuMeterVisible()
    {
        if ( m_IsCpuMeterVisible )
        {
            m_IsCpuMeterVisible = false;
            NW_DEV_LOG( "[DEMO] Cpu Meter Draw Off\n\n" );
        }
        else
        {
            m_IsCpuMeterVisible = true;
            NW_DEV_LOG( "[DEMO] Cpu Meter Draw On\n\n" );
        }
    }

    // Switches between play and stop.
    void SwapPlay()
    {
        if ( m_IsPalying )
        {
            m_IsPalying = false;
            NW_DEV_LOG( "[DEMO] Demo Stop\n\n" );
        }
        else
        {
            m_IsPalying = true;
            NW_DEV_LOG( "[DEMO] Demo Play\n\n" );
        }
    }

    // Advances the frames.
    void AddFrame( f32 addFrame )
    {
        m_Frame += addFrame;
    }

    // Get the frame.
    f32  GetFrame()
    {
        return m_Frame;
    }

    // Sets the frame.
    void SetFrame( f32 frane )
    {
        m_Frame = frane;
    }

    // Sets the frame advance flag.
    void SetAddFrame()
    {
        m_IsAddFrame = true;
        NW_DEV_LOG( "[DEMO] Demo Add Frame\n\n" );
    }

    // Sets the frame advance flag.
    bool GetAddFrame()
    {
        bool temp = m_IsAddFrame;
        m_IsAddFrame = false;
        return temp;
    }

    bool IsCommandBufferDouble()    { return m_IsCommandDouble; }
    bool IsLayoutVisible()          { return m_IsLayoutVisible; }
    bool IsCpuMeterVisible()        { return m_IsCpuMeterVisible; }
    bool IsPlaying()                { return m_IsPalying; }

    void SetTimerStart()
    {
        m_StartTime = nn::os::Tick::GetSystemCurrent();
    }

    void SetTimerEnd()
    {
        m_EndTime = nn::os::Tick::GetSystemCurrent();
        m_AllTime = m_EndTime - m_StartTime;
    }

    nn::os::Tick GetStartTime() const
    {
        return m_StartTime;
    }

    f32 GetAllTime()
    {
        return m_AllTime.ToMilliSeconds()/1000.f;
    }

    void SetGpuWaitStart()
    {
        m_GpuWaitStartTime = nn::os::Tick::GetSystemCurrent();
    }

    void SetGpuWaitEnd()
    {
        m_GpuWaitEndTime = nn::os::Tick::GetSystemCurrent();
        m_GpuWaitTime = m_GpuWaitEndTime - m_GpuWaitStartTime;
    }

    f32 GetGpuWaitTime()
    {
        return m_GpuWaitTime.ToMilliSeconds()/1000.f;
    }

    // Gets the number of requests accumulated in the command buffer.
    void UpdateUsedReqCount()
    {
        nngxGetCmdlistParameteri(NN_GX_CMDLIST_USED_REQCOUNT, &m_UsedReqCount);
    }

    GLint GetUsedReqCount()
    {
        return m_UsedReqCount;
    }

    int UpdateVSyncCounter()
    {
        int prevCounter = m_LastVSyncCounter;
        m_LastVSyncCounter = nngxCheckVSync(NN_GX_DISPLAY_BOTH);

        if (m_LastVSyncCounter < prevCounter)
        {
            // The accurate value is not known because nngxCheckVSync returns to 0 once it reaches its upper limit.
            return -1;
        }
        else
        {
            return m_LastVSyncCounter - prevCounter;
        }
    }

private:
    bool                    m_IsCommandDouble;      // Command buffer double/single.
    bool                    m_IsLayoutVisible;      // Render/do not render layout.
    bool                    m_IsCpuMeterVisible;    // Render/do not render OpenGL.
    bool                    m_IsPalying;            // Playing or stopped.
    bool                    m_IsAddFrame;           // Frame advance.
    bool                    m_IsFirstLoop;
    f32                     m_Frame;

    nn::os::Tick            m_StartTime;
    nn::os::Tick            m_EndTime;
    nn::os::Tick            m_AllTime;

    nn::os::Tick            m_GpuWaitStartTime;
    nn::os::Tick            m_GpuWaitEndTime;
    nn::os::Tick            m_GpuWaitTime;

    GLint                   m_UsedReqCount;
    GLint                   m_LastVSyncCounter;
};
#if NN_CURRENT_VERSION_NUMBER >= NN_VERSION_NUMBER(4,0,0,0)
#pragma diag_default 2530 // padding added to end of struct.
#pragma diag_default 1301 // padding inserted in struct.
#endif


#endif  // FRAMEWORK_H__
