/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/
#include <nn/os.h>

#include <nw/types.h>
//#include <nw/demo.h>
#include <nw/dev.h>
#include <nw/gfx.h>
#include <nw/ut.h>

#include "../include/LytCtrl.h"


namespace
{


/*
  
 */
void
InitGX()
{
#ifdef NW_PLATFORM_CTR
    GLenum curCmdGenMode = 0;
    nngxGetCommandGenerationMode(&curCmdGenMode);

    if (curCmdGenMode != NN_GX_CMDGEN_MODE_UNCONDITIONAL)
    {
        // Immediately issues commands of the NN_GX_STATE_OTHER state.
        nngxSetCommandGenerationMode(NN_GX_CMDGEN_MODE_UNCONDITIONAL);
    }
#endif

    glClearColor(0.2f, 0.2f, 0.2f, 0.0f);
    glDisable(GL_CULL_FACE);
    glDisable(GL_DEPTH_TEST);
    glDisable(GL_SCISSOR_TEST);
    glDisable(GL_STENCIL_TEST);
    glDisable(GL_POLYGON_OFFSET_FILL);
    glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);
    glDepthMask(GL_TRUE);
#ifdef NW_PLATFORM_CTR
    glDisable(GL_EARLY_DEPTH_TEST_DMP);
#endif

#ifdef NW_PLATFORM_CTR
    // Returns the command issue mode back to what it was.
    if (curCmdGenMode != NN_GX_CMDGEN_MODE_UNCONDITIONAL)
    {
        nngxSetCommandGenerationMode(curCmdGenMode);
    }

    // Issues command based on validation for states that are not the NN_GX_STATE_OTHERS state.
    const GLbitfield stateMask =
        NN_GX_STATE_TRIOFFSET |
        NN_GX_STATE_SCISSOR |
        NN_GX_STATE_FBACCESS;

    nngxUpdateState(stateMask);
    NW_LYT_DRAWER_VALIDATE(stateMask);
#endif

    NW_GL_ASSERT();
}


/*
  
 */
LytCtrl::LytCtrl()
{
    m_ResourcesArray.clear();

    // Loads files like the shader file.
    {
        m_GraphicsResource.StartSetup();
        const wchar_t* resourcePath = 0;
        for (int i = 0;
             (resourcePath = m_GraphicsResource.GetResourcePath(i)) != NULL;
             ++i)
        {
            SmFile file(true /* autoFree */);
            if (!file.Read(resourcePath, m_Allocator))
            {
                NW_FATAL_ERROR("can not read lyt resource file.");
            }
            m_GraphicsResource.SetResource(i, file.Buffer(), file.Size(), false /* bFree */);
        }
        m_GraphicsResource.FinishSetup();

        m_DrawInfo.SetGraphicsResource( &m_GraphicsResource );
    }

    // Initializes Drawer.
    m_Drawer.Initialize( m_GraphicsResource );
}


/*
  
 */
LytCtrl::~LytCtrl()
{
    uint i = 0;

    // Destroys the SmLayout instance.
    for ( i=0; i<m_ResourcesArray.size(); i++ )
    {
        SM_SAFE_DELETE( m_SmLytInstanceArray[i] );
    }

    // Destroys the SmLytResource instance.
    for ( i=0; i<m_SmLytResourceArray.size(); i++ )
    {
        SM_SAFE_DELETE( m_SmLytResourceArray[i] );
    }

    // Releases the loaded layout resources.
    for ( i=0; i<m_ResourcesArray.size(); i++ )
    {
        SM_SAFE_DELETE( m_ResourcesArray[i].resAccessor );
        m_ResourcesArray[i].lytFile.Release();
    }
}


/*
  
 */
SmLayoutResource*
LytCtrl::CreateSmLayoutResource( const wchar_t* archivePath, LytResourceFile& lytResFile )
{
    LytResourceSet      resSet;
    SmLayoutResource*   smLytResource = NULL;
    bool                isPushed = false;
    NW_UNUSED_VARIABLE(isPushed);

    // Loads layout binary resources (archives).
    if (!resSet.lytFile.Read( archivePath, m_DeviceAllocator ) )
    {
        NW_FATAL_ERROR("can not open layout archive.\n");
    }

    // Specifies the binary resource root directory and generates the resource assessor.
    resSet.resAccessor = new nw::lyt::ArcResourceAccessor;
    if (!resSet.resAccessor->Attach(resSet.lytFile.Buffer(), "."))
    {
        NW_FATAL_ERROR("can not attach layout archive.\n");
    }

    // Generates the layout resource class.
    smLytResource = new SmLayoutResource();

    // Generates nw::lyt::Layout.
    const void* lytRes = resSet.resAccessor->GetResource( 0, lytResFile.clyt.c_str() );
    NW_NULL_ASSERT( lytRes );
     if ( !smLytResource->InitLayout( lytRes, resSet.resAccessor ) )
    {
        NW_FATAL_ERROR("can not setup layout archive.\n");
    }

    // Loads the animation resource.
    for ( uint i = 0; i < lytResFile.clan.size(); i++ )
    {
        const void* lpaRes = resSet.resAccessor->GetResource( 0, lytResFile.clan[i].c_str() );
         if ( lpaRes )
        {
             if ( !smLytResource->AddAnimation( lpaRes, resSet.resAccessor ) )
            {
                NW_FATAL_ERROR("can not setup layout animation archive.\n");
            }
        }
    }

    isPushed = m_SmLytResourceArray.push_back( smLytResource );
    NW_ASSERT(isPushed);

    isPushed = m_ResourcesArray.push_back( resSet );
    NW_ASSERT(isPushed);

    // Sets up the camera.
    setupCamera( smLytResource->GetLayout()->GetLayoutRect() );

    return smLytResource;
}


/*
  
 */
SmLayout*
LytCtrl::CreateSmLayout( SmLayoutResource* layoutResource )
{
    SmLayout* smLyt = new SmLayout( layoutResource );
    NW_NULL_ASSERT( smLyt );

    // Maintains the generated instance.
    bool isPushed = m_SmLytInstanceArray.push_back( smLyt );
    NW_ASSERT(isPushed);

    return smLyt;
}


/*
  
 */
void
LytCtrl::Render()
{
    if ( m_SmLytInstanceArray.empty() ) return;

    InitGX();

    // Makes enough entries for the layout of the maintained SmLaytou.
    for ( uint i = 0; i < m_SmLytInstanceArray.size(); i++ )
    {
        SmLayout* pLayout = m_SmLytInstanceArray[i];

        if ( pLayout->GetVisible() )
        {
            // Reflects the data in the layout.
            if ( pLayout->SetupLayout() )
            {
                SmLayoutResource* smLayoutRes = pLayout->GetLayoutResource();
                nw::lyt::Layout* layout = smLayoutRes->GetLayout();
                layout->Animate();
                layout->CalculateMtx( m_DrawInfo );
                m_Drawer.DrawBegin( m_DrawInfo );
                m_Drawer.Draw( layout, m_DrawInfo );
                m_Drawer.DrawEnd( m_DrawInfo );

                pLayout->ResetLayout();
            }
        }
    }
}


/*
  
 */
void
LytCtrl::setupCamera( nw::ut::Rect rect )
{
    f32 znear = 0.f;
    f32 zfar = 500.f;

    // Set the projection matrix for an orthogonal projection
    // (Since the layout data is horizontal, change its direction.)
    nw::math::MTX44 projMtx;
    nw::math::MTX44OrthoPivot(
        &projMtx,
        rect.left,   // Left
        rect.right,  // Right
        rect.bottom, // Bottom
        rect.top,    // Top
        znear,
        zfar,
        nw::math::PIVOT_UPSIDE_TO_TOP);
    m_DrawInfo.SetProjectionMtx(projMtx);

    // Set the modelview matrix
    // (Since the layout data is horizontal, the upper part of the screen is the -X direction of the layout.)
    nw::math::VEC3 pos(0, 0, 1);
    nw::math::VEC3 up(0, 1, 0);
    nw::math::VEC3 target(0, 0, 0);

    nw::math::MTX34 viewMtx;
    nw::math::MTX34LookAt(&viewMtx, &pos, &up, &target);
    m_DrawInfo.SetViewMtx(viewMtx);
}

} // Namespace


