/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#include "../include/SmModel.h"

namespace
{

//----------------------------------------
// Constructor
SmModel::SmModel()
: m_pParentNode( NULL ),
  m_pSceneNode( NULL ),
  m_Model( NULL ),
  m_SkeletalModel( NULL ),
  m_AnimationFrame( 0.f ),
  m_CurrentAnimEvNo( -1 ),
  m_CostrainId( -1 ),
  m_Duplicated( false )
{
    m_AnimEvaluatorArray.clear();
}


//----------------------------------------
//
SmModel::~SmModel()
{

}


//----------------------------------------
// Set the node.
void
SmModel::SetSceneNode( nw::gfx::SceneNode* pParentNode, nw::gfx::SceneNode* pSceneNode )
{
    NW_NULL_ASSERT(pParentNode);

    m_pParentNode = pParentNode;
    if ( pSceneNode )
    {
        m_pSceneNode = pSceneNode;
    }
}


//----------------------------------------
// Set the model.
void
SmModel::SetModel( nw::gfx::Model* pModel, nw::gfx::SkeletalModel* pSkelModel )
{
    m_Model         = pModel;
    m_SkeletalModel = pSkelModel;
}


//----------------------------------------
// Resolve updates, constraints, etc.
void
SmModel::Update()
{
    // Resolve parent-child relationships.
    updateConstrain();
}



//----------------------------------------
// position
void
SmModel::SetPosition( f32 x, f32 y, f32 z )
{
    nw::gfx::TransformNode* transformNode =
        nw::ut::DynamicCast<nw::gfx::TransformNode*>(m_pSceneNode);
    NW_NULL_ASSERT(transformNode);

    transformNode->Transform().SetTranslate( x, y, z );
}


//----------------------------------------
// Enlarge, reduce
void
SmModel::SetScale( f32 x, f32 y, f32 z )
{
    nw::gfx::TransformNode* transformNode =
        nw::ut::DynamicCast<nw::gfx::TransformNode*>(m_pSceneNode);
    NW_NULL_ASSERT(transformNode);

    transformNode->Transform().SetScale( x, y, z );
}


//----------------------------------------
// Degrees of rotation.
void
SmModel::SetRotate( f32 x, f32 y, f32 z )
{
    nw::gfx::TransformNode* transformNode =
        nw::ut::DynamicCast<nw::gfx::TransformNode*>(m_pSceneNode);
    NW_NULL_ASSERT(transformNode);

    transformNode->Transform().SetRotateXYZ( NW_MATH_DEG_TO_RAD(x),
                                             NW_MATH_DEG_TO_RAD(y),
                                             NW_MATH_DEG_TO_RAD(z) );
}


//----------------------------------------
// Matrix
void
SmModel::SetMatrix( const nw::math::MTX34* matrix )
{
    NW_NULL_ASSERT( matrix );

    nw::gfx::TransformNode* transformNode =
        nw::ut::DynamicCast<nw::gfx::TransformNode*>(m_pSceneNode);
    NW_NULL_ASSERT(transformNode);

    transformNode->Transform().SetTransformMatrix( *matrix );
}


void SmModel::SetWorldMatrix( const nw::math::MTX34* matrix )
{
    NW_NULL_ASSERT( matrix );

    nw::gfx::TransformNode* transformNode =
        nw::ut::DynamicCast<nw::gfx::TransformNode*>(m_pSceneNode);
    NW_NULL_ASSERT(transformNode);

    transformNode->WorldMatrix() = *matrix;
}




//----------------------------------------
// Control visibility.
void
SmModel::SetVisible( bool visible )
{
    if ( !m_SkeletalModel ) return;
    m_SkeletalModel->SetVisible( visible );

    // Determine whether the object is registered in the scene tree based on whether it has a parent.
    if ( !visible )
    {
        if ( m_pSceneNode->GetParent() )
        {
            m_pParentNode->DetachChild( m_pSceneNode );
        }
    }
    else
    {
        if ( !m_pSceneNode->GetParent() )
        {
            m_pParentNode->AttachChild( m_pSceneNode );
        }
    }
}

void
SmModel::SwapVisible()
{
    if ( !m_SkeletalModel ) return;

    // Determine whether the object is registered in the scene tree based on whether it has a parent.
    if ( m_SkeletalModel->IsVisible() )
    {
        m_SkeletalModel->SetVisible( false );

        if ( m_pSceneNode->GetParent() )
        {
            m_pParentNode->DetachChild( m_pSceneNode );
        }
    }
    else
    {
        m_SkeletalModel->SetVisible( true );

        if ( !m_pSceneNode->GetParent() )
        {
            m_pParentNode->AttachChild( m_pSceneNode );
        }
    }
}

bool
SmModel::IsVisible()
{
    if ( !m_SkeletalModel ) return false;
    return m_SkeletalModel->IsVisible();
}


//----------------------------------------
// Copy.
SmModel* SmModel::Duplicate()
{
    SmModel* model = new SmModel();

#if 1
    // There are no shared skeletons.
    nw::gfx::SkeletalModel* skeletalModel = nw::gfx::SkeletalModel::Builder()
        .Create( NULL, nw::gfx::ResSceneObject( m_SkeletalModel->GetResModel() ), m_Allocator );
#else
    // There are shared skeletons.
    nw::gfx::SkeletalModel* skeletalModel = nw::gfx::SkeletalModel::Builder()
        .SharedSkeleton( m_SkeletalModel->GetSkeleton() )
        .Create(NULL, nw::gfx::ResSceneObject(m_SkeletalModel->GetResModel()), m_Allocator);
#endif

    NW_NULL_ASSERT(skeletalModel);

    // Copy values from model being copied.
    model->m_SkeletalModel      = skeletalModel;
    model->m_Duplicated         = true;
    model->m_pSceneNode         = static_cast<nw::gfx::SceneNode*>(model->m_SkeletalModel);

    model->SetLayerId( this->GetLayerId() );

    // todo: To animate based on parameters,
    //      animation must be copied.
    model->m_AnimEvaluatorArray.CopyFrom( m_AnimEvaluatorArray );

    return model;
}


//----------------------------------------
// Add animations.
void
SmModel::AddAnimEvaluator( nw::gfx::TransformAnimEvaluator* animEvaluator )
{
    m_AnimEvaluatorArray.push_back( animEvaluator );
}


//----------------------------------------
// Get the number of animations.
uint
SmModel::GetAnimEvaluatorNum() const
{
    return m_AnimEvaluatorArray.size();
}


//----------------------------------------
// Configure animations.
bool
SmModel::SetAnimEvaluatorNo( uint animEvaluatorNo )
{
    if ( animEvaluatorNo >= GetAnimEvaluatorNum() ) return false;
    if ( !m_SkeletalModel ) return false;

    nw::gfx::AnimGroup* animGroup = m_SkeletalModel->GetSkeletalAnimGroup();
    if (animGroup == NULL){ NW_FATAL_ERROR("Don't have AnimGroup");; }

    // Bind the animation.
    m_AnimEvaluatorArray[animEvaluatorNo]->Release();
    bool bindResult = m_AnimEvaluatorArray[animEvaluatorNo]->Bind(animGroup);

    // Set animation for model.
    m_SkeletalModel->SetSkeletalAnimObject(m_AnimEvaluatorArray[animEvaluatorNo]);

    // Store number.
    m_CurrentAnimEvNo = animEvaluatorNo;

    return true;
}


//----------------------------------------
// Configure the animation frame.
void
SmModel::SetAnimationFrame( f32 setFrame )
{
    m_AnimationFrame = setFrame;
    if ( m_CurrentAnimEvNo == -1 ) return;

    // Do not update animation not visible.
    if ( !this->IsVisible() ) return;

    // Set frame count.
    m_AnimEvaluatorArray[m_CurrentAnimEvNo]->SetFrame(m_AnimationFrame);
}


//----------------------------------------
// Advance the animation frame.
// Return true if we've reached the last frame.
bool
SmModel::AddAnimationFrame( f32 addFrame, bool loop )
{
    bool ret = false;

    m_AnimationFrame += addFrame;
    if ( m_CurrentAnimEvNo == -1 ) return ret;

    // Fixed in the last frame.
    f32 endFrame = m_AnimEvaluatorArray[m_CurrentAnimEvNo]->AnimFrameController().GetEndFrame();
    if ( m_AnimationFrame > endFrame )
    {
        if ( loop )
        {
            m_AnimationFrame = 0.f;
        }
        else
        {
            m_AnimationFrame = endFrame;
        }

        ret = true;
    }

    // Do not update animation not visible.
    if ( !this->IsVisible() ) return ret;

    // Set frame count.
    m_AnimEvaluatorArray[m_CurrentAnimEvNo]->SetFrame(m_AnimationFrame);

    return ret;
}


//----------------------------------------
// Get the animation frame.
f32
SmModel::GetAnimationFrame() const
{
    return m_AnimationFrame;
}


//----------------------------------------
// Get the largest frame of the animation being configured.
f32
SmModel::GetAnimationMaxFrame() const
{
    if ( m_CurrentAnimEvNo == -1 ) return 0.f;

    return m_AnimEvaluatorArray[m_CurrentAnimEvNo]->AnimFrameController().GetEndFrame();
}


//----------------------------------------
// Configure material animation.
bool
SmModel::SetMaterialAnim( nw::gfx::AnimEvaluator* animEvaluator )
{
    nw::gfx::AnimGroup* animGroup = m_Model->GetMaterialAnimGroup();
    if ( !animGroup ) return false;

    bool bindResult = animEvaluator->Bind( animGroup );
    if ( !bindResult ) return false;

    // Set animation for model.
    m_Model->SetMaterialAnimObject( animEvaluator );

    return true;
}



//----------------------------------------
// Specify parent node ID.
void
SmModel::SetConstrainId( s32 boneId )
{
    m_CostrainId = boneId;

    nw::gfx::TransformNode* transformNode =
        nw::ut::DynamicCast<nw::gfx::TransformNode*>(m_pSceneNode);
    NW_NULL_ASSERT(transformNode);

    // Cancel with -1.
    if ( m_CostrainId == -1 )
    {
        // Raise FLAG_IS_WORLDMATRIX_CALCULATION_ENABLED flag.
        transformNode->Transform().EnableFlags(nw::gfx::CalculatedTransform::FLAG_IS_WORLDMATRIX_CALCULATION_ENABLED);

        transformNode->Transform().DisableFlags(nw::gfx::CalculatedTransform::FLAG_IS_WORLDMATRIX_CALCULATION_ENABLED);

    }
    else
    {
        // Lower FLAG_IS_WORLDMATRIX_CALCULATION_ENABLED flag.
        transformNode->Transform().DisableFlags(nw::gfx::CalculatedTransform::FLAG_IS_WORLDMATRIX_CALCULATION_ENABLED);

        transformNode->Transform().EnableFlags(nw::gfx::CalculatedTransform::FLAG_IS_WORLDMATRIX_CALCULATION_ENABLED);
    }
}


//----------------------------------------
// Resolve parent-child inheritance.
void
SmModel::updateConstrain()
{
     if ( m_CostrainId == -1 ) return;

    // Does parent have skeletal structure?
    nw::gfx::SkeletalModel* pParentSkeletalModel = nw::ut::DynamicCast<nw::gfx::SkeletalModel*>(m_pParentNode);

     if ( pParentSkeletalModel )
    {
        // Get Skeleton.
        nw::gfx::Skeleton* pSkeleton = pParentSkeletalModel->GetSkeleton();
        NW_NULL_ASSERT(pSkeleton);

        nw::gfx::Skeleton::TransformPose& transformPose = pSkeleton->WorldTransformPose();
//        nw::gfx::Skeleton::TransformPose& transformPose = pSkeleton->LocalTransformPose();
         if ( transformPose.GetBonesCount() < m_CostrainId ) return;

        // Get matrix for specified node and set to world matrix of self.
        nw::gfx::Skeleton::TransformPose::Transform* transForm = transformPose.GetTransform( m_CostrainId );
//        this->SetWorldMatrix( &transForm->TransformMatrix() );
        this->SetMatrix( &transForm->TransformMatrix() );
    }
    else
    {
        // Do nothing if there is no skeleton.
    }
}


//----------------------------------------
// Testing features.
// Show/hide meshes in model based on whether their value is 0 or 1.
void
SmModel::SetVisibleZone( f32 term )
{
    if ( !m_SkeletalModel ) return;

    nw::gfx::ResMeshArray meshArray = m_SkeletalModel->GetResMeshes();

    s32 visibleNum = meshArray.size() * term;

    for ( int i = 0; i < meshArray.size(); i++ )
    {
        if ( i < visibleNum )
        {
            meshArray[i].SetVisible( true );
        }
        else
        {
            meshArray[i].SetVisible( false );
        }
    }
}



} // Namespace


