/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/
#include <nn/os.h>

#include <nw/types.h>
#include <nw/ut.h>

#include "../include/SmPerf.h"

namespace
{


SmGpuPerf* s_GpuPerf = NULL;

//----------------------------------------
// Constructor
SmCpuPerf::SmCpuPerf()
: SmPerf(){}


//----------------------------------------
// Starts measurements.
void
SmCpuPerf::Begin( const char* costName, u32 costColor  )
{
    NW_NULL_ASSERT( costName );

    Cost addCost;
    addCost.init();

    nw::ut::strncpy( addCost.costName, 32, costName, strlen( costName ) );
    addCost.costColor = costColor;

    bool isPushed = m_CostArray[IdxPrepare()].push_back( addCost );
    NW_ASSERT(isPushed);

    m_CostArray[IdxPrepare()].back().costStart = this->GetTime();
}


//----------------------------------------
// Ends measures.
void
SmCpuPerf::End()
{
    NW_ASSERT( !m_CostArray[IdxPrepare()].empty() );
    m_CostArray[IdxPrepare()].back().costEnd = this->GetTime();
}

//----------------------------------------------------------
void SmCpuPerf::Reset( nn::os::Tick startTime )
{
    SwapBuffer();

    m_FrameStart[IdxPrepare()] = startTime;
}




//----------------------------------------
// Constructor
SmGpuPerf::SmGpuPerf()
: SmPerf()
{
    if ( s_GpuPerf )
    {
        NW_FATAL_ERROR("SmGpuPerf instance is exsit.\n");
    }

    // Registers itself.
    s_GpuPerf = this;
}


//----------------------------------------
// Destructor
SmGpuPerf::~SmGpuPerf()
{
    s_GpuPerf = NULL;
}

//----------------------------------------
// Configures callback processing for starting measurements.
void
SmGpuPerf::Begin( const char* costName, u32 costColor  )
{
    Cost addCost;
    addCost.init();

    nw::ut::strncpy( addCost.costName, 32, costName, strlen( costName ) );
    addCost.costColor = costColor;

    GLint requestCount = 0;
    nngxGetCmdlistParameteri(NN_GX_CMDLIST_USED_REQCOUNT, &requestCount);

    addCost.gpuStart    = requestCount;

    bool isPushed = m_CostArray[IdxPrepare()].push_back( addCost );
    NW_ASSERT(isPushed);

    // Has the start time of the first command request set from the outside.
    if ( requestCount > 0 )
    {
        // Sets callback.
        nngxEnableCmdlistCallback( requestCount );
        NW_GL_ASSERT();
    }
}


//----------------------------------------
// Configures callback processing for ending measurements.
void
SmGpuPerf::End()
{
    NW_ASSERT( !m_CostArray[IdxPrepare()].empty() );

    GLint requestCount = 0;
    nngxGetCmdlistParameteri(NN_GX_CMDLIST_USED_REQCOUNT, &requestCount);

    m_CostArray[IdxPrepare()].back().gpuEnd = requestCount;

    nngxEnableCmdlistCallback( requestCount );
    NW_GL_ASSERT();
}

//----------------------------------------
// Calls the callback.
void
SmGpuPerf::Invoke(GLint id)
{
    // Searches to find out if there are costs specified by "id."
    for ( int i = 0; i<m_CostArray[IdxCurrent()].size(); i++ )
    {
        if ( m_CostArray[IdxCurrent()][i].gpuStart == id )
        {
            m_CostArray[IdxCurrent()][i].costStart = this->GetTime();
        }
        if ( m_CostArray[IdxCurrent()][i].gpuEnd == id )
        {
            m_CostArray[IdxCurrent()][i].costEnd = this->GetTime();
        }
    }
}


//----------------------------------------
// The callback function to be registered.
void
SmGpuPerf::Callback(GLint id)
{
    if ( s_GpuPerf )
    {
        s_GpuPerf->Invoke(id);
    }
}


//----------------------------------------
// Resets
void
SmGpuPerf::Reset(nn::os::Tick startTime)
{
    m_FrameStart[IdxPrepare()] = startTime;

    SwapBuffer();
}
//----------------------------------------------------------
void SmGpuPerf::SetCallback() const
{
    // Register callback function
    nngxSetCmdlistCallback(SmGpuPerf::Callback);
}

//----------------------------------------------------------
void SmGpuPerf::SetCmdlistRun()
{
    for ( int i = 0; i<m_CostArray[IdxCurrent()].size(); i++ )
    {
        // The first command request start time is recorded here because it cannot be recorded at callback.
        if ( m_CostArray[IdxCurrent()][i].gpuStart == 0 )
        {
            m_CostArray[IdxCurrent()][i].costStart = this->GetTime();
        }
    }
}

//----------------------------------------
// Constructor
SmPerfDrawer::SmPerfDrawer()
: m_pPerfBarSplit( NULL )
{

}


//----------------------------------------
// Destructor
SmPerfDrawer::~SmPerfDrawer()
{

}




//----------------------------------------
// Renders the process bar.
void
SmPerfDrawer::Render( SmPerf* cpuPerf, SmPerf* gpuPerf )
{
    NW_NULL_ASSERT( cpuPerf );
    NW_NULL_ASSERT( gpuPerf );

    // Registers rect.
    #define _SET_BAR_RECT( prim, x, y, w, h, color ) \
        prim.SetVertex( x,     y,     color ); \
        prim.SetVertex( x,     y + h, color ); \
        prim.SetVertex( x + w, y + h, color ); \
        prim.SetVertex( x + w, y + h, color ); \
        prim.SetVertex( x + w, y,     color ); \
        prim.SetVertex( x,     y,     color ); \

    nn::os::Tick frameStart;

    // Process bar
    if ( m_PerfBar.Begin( cpuPerf->GetCostCount() * 6 + gpuPerf->GetCostCount() * 6, GL_TRIANGLES, 0.f, 0.f ) )
    {
        // CPU process bar
        frameStart = cpuPerf->GetFrameStart();
        for ( int i = 0; i<cpuPerf->GetCostCount(); i++ )
        {
            f32 start = (cpuPerf->GetCostStart(i) - frameStart).ToMicroSeconds() / 16000.f * PERF_ONE_FRAME_WIDTH;
            f32 end = (cpuPerf->GetCostEnd(i) - frameStart).ToMicroSeconds() / 16000.f * PERF_ONE_FRAME_WIDTH;

            _SET_BAR_RECT( m_PerfBar, start, 5.f, end - start, 4.f, cpuPerf->GetCostColor(i) );
        }

        // GPU process bar
        frameStart = gpuPerf->GetFrameStart();
        for ( int i = 0; i<gpuPerf->GetCostCount(); i++ )
        {
            f32 start = (gpuPerf->GetCostStart(i) - frameStart).ToMicroSeconds() / 16000.f * PERF_ONE_FRAME_WIDTH;
            f32 end = (gpuPerf->GetCostEnd(i) - frameStart).ToMicroSeconds() / 16000.f * PERF_ONE_FRAME_WIDTH;

            _SET_BAR_RECT( m_PerfBar, start, 20.f, end - start, 4.f, gpuPerf->GetCostColor(i) );
        }

        m_PerfBar.End();
    }

    nngxFlush3DCommand();
}


} // Namespace


