﻿/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#include <nw/demo/demo_DisplayBufferSwapper.h>
#include <nn/gx.h>


namespace nw
{
namespace demo
{

NW_UT_RUNTIME_TYPEINFO_ROOT_DEFINITION(DisplayBufferSwapper);

//----------------------------------------
DisplayBufferSwapper*
DisplayBufferSwapper::Builder::Create(
    os::IAllocator* allocator
)
{
    NW_ASSERT(0 < m_Description.bufferCount);

    void* memory = allocator->Alloc(sizeof(DisplayBufferSwapper));
    DisplayBufferSwapper* bufferSwapper =
        new(memory) DisplayBufferSwapper(allocator, m_Description);

    return bufferSwapper;
}

//----------------------------------------
DisplayBufferSwapper::DisplayBufferSwapper(
    os::IAllocator* allocator,
    const Description& description
)
: GfxObject(allocator),
  m_Description(description),
  m_CurrentDisplay(0),
  m_DisplayBuffers(0)
{
    NW_ASSERT(0 <= m_Description.screenKind < BOTH_SCREENS);
    NW_ASSERT(m_Description.bufferCount);
    NW_ASSERT(0 <= m_Description.memoryArea < nw::gfx::GRAPHICS_MEMORY_AREA_COUNT);
    NW_ASSERT(0 <= m_Description.displayOffsetX);
    NW_ASSERT(0 <= m_Description.displayOffsetY);

    GLenum display = 0;
    if (m_Description.screenKind == UPPER_SCREEN)
    {
        display = NN_GX_DISPLAY0;
    }
    else if (m_Description.screenKind == LOWER_SCREEN)
    {
        display = NN_GX_DISPLAY1;
    }
    else if (m_Description.screenKind == EXTENSION_SCREEN)
    {
        display = NN_GX_DISPLAY0_EXT;
    }
    else
    {
        NW_FATAL_ERROR();
    }

    nngxActiveDisplay(display);

    NW_ASSERT(m_Description.bufferCount != 0);

    void* memory = allocator->Alloc(sizeof(GLuint) * m_Description.bufferCount);
    m_DisplayBuffers = static_cast<GLuint*>(memory);
    nngxGenDisplaybuffers(m_Description.bufferCount, m_DisplayBuffers);

    for (s32 i = 0; i < m_Description.bufferCount; ++i)
    {
        nngxBindDisplaybuffer(m_DisplayBuffers[i]);

        nngxDisplaybufferStorage(
            m_Description.format,
            m_Description.height,
            m_Description.width,
            m_Description.memoryArea);
    }

    nngxDisplayEnv(m_Description.displayOffsetY, m_Description.displayOffsetX);
    nngxBindDisplaybuffer(0);
    m_CurrentDisplay = 0;
}

//----------------------------------------
void
DisplayBufferSwapper::MakeTransferBufferCommand(
    gfx::IRenderTarget* target,
    bool isMultiCommandList
){
    if (target == NULL)
    {
        return;
    }

    NW_ASSERT(m_Description.transferMode < TRANSFER_MODE_ANTIALIASE_COUNT);
    NW_ASSERT(0 <= m_Description.transferOffsetX && m_Description.transferOffsetX % 8 == 0);
    NW_ASSERT(0 <= m_Description.transferOffsetY && m_Description.transferOffsetY % 8 == 0);

    GLuint bufferObject = target->GetBufferObject();

    NW_NULL_ASSERT(bufferObject);

    glBindFramebuffer(GL_FRAMEBUFFER, bufferObject);

    GLenum mode = NULL;
    if (m_Description.transferMode == TRANSFER_MODE_ANTIALIASE_NOT_USED)
    {
        mode = NN_GX_ANTIALIASE_NOT_USED;
    }
    else if (m_Description.transferMode == TRANSFER_MODE_ANTIALIASE_2x1)
    {
        mode = NN_GX_ANTIALIASE_2x1;
    }
    else if (m_Description.transferMode == TRANSFER_MODE_ANTIALIASE_2x2)
    {
        mode = NN_GX_ANTIALIASE_2x2;
    }

    unsigned int targetCount =
        (m_CurrentDisplay + (isMultiCommandList ? 1 : 0)) % m_Description.bufferCount;

    nngxTransferRenderImage(
        m_DisplayBuffers[targetCount],
        mode,
        m_Description.isTransferFlipX,
        m_Description.transferOffsetY,
        m_Description.transferOffsetX);

    nngxSplitDrawCmdlist();
}

//----------------------------------------
void
DisplayBufferSwapper::ActivateBuffer()
{
    GLenum display = NULL;

    if(m_Description.screenKind == UPPER_SCREEN)
    {
        display = NN_GX_DISPLAY0;
    }
    else if(m_Description.screenKind == LOWER_SCREEN)
    {
        display = NN_GX_DISPLAY1;
    }
    else if(m_Description.screenKind == EXTENSION_SCREEN)
    {
        display = NN_GX_DISPLAY0_EXT;
    }
    else
    {
        NW_FATAL_ERROR();
    }

    nngxActiveDisplay(display);
    nngxBindDisplaybuffer(m_DisplayBuffers[m_CurrentDisplay]);

    ++m_CurrentDisplay;
    if (m_CurrentDisplay == m_Description.bufferCount)
    {
        m_CurrentDisplay = 0;
    }
}

} // namespace demo
} // namespace nw
