/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#include <nw/ut/ut_Preprocessor.h>
#include <nw/demo/demo_GraphicsMemoryAllocator.h>
#include <nn.h>
#include <nn/gx.h>

namespace nw
{
namespace demo
{

const int ALIGNMENT_SYSTEM_BUFFER     = 4;
const int ALIGNMENT_VERTEX            = 4;
const int ALIGNMENT_TEXTURE           = 128;
const int ALIGNMENT_RENDER_BUFFER     = 64;    // 24-bit format (RGB8, D24) buffer must be aligned to 96 bytes
const int ALIGNMENT_DISPLAY_BUFFER    = 16;
const int ALIGNMENT_3D_COMMAND_BUFFER = 16;

//--------------------------------------------------------------------------
void
GraphicsMemoryAllocator::Initialize(os::IAllocator* allocator)
{
    m_CurrentAddrVramA = nn::gx::GetVramStartAddr(nn::gx::MEM_VRAMA);
    m_CurrentAddrVramB = nn::gx::GetVramStartAddr(nn::gx::MEM_VRAMB);

    m_DeviceMemoryAllocator = allocator;

    m_Initialized = true;
}

//--------------------------------------------------------------------------
void
GraphicsMemoryAllocator::Finalize(void)
{
    m_CurrentAddrVramA = nn::gx::GetVramStartAddr(nn::gx::MEM_VRAMA);
    m_CurrentAddrVramB = nn::gx::GetVramStartAddr(nn::gx::MEM_VRAMB);
}

//--------------------------------------------------------------------------
void*
GraphicsMemoryAllocator::Allocate(GLenum area, GLenum aim, GLuint id, GLsizei size)
{
    NW_UNUSED_VARIABLE(id);

    NW_ASSERTMSG (m_Initialized, "[DEMO:GraphicsMemoryAllocator] not initialized.\n");

    if (size == 0) { return 0; }

    // At present it is unclear how much memory is reserved in the CTR when it comes to VRAM-A or VRAM-B

    // When allocating to VRAM, for simplicity treat VRAM like a stack.
    // Also, do not worry about fragmentation.
    // To deal with fragmentation, you may need to retain information about the memory allocated as "area," "aim" and "id" when allocation is successful so they can be subsequently deallocated.
    // 

    int addrAlign = 8;
    void* resultAddr = NULL;

    // Take into account the alignment restrictions on each data set's position
    switch (aim)
    {
    case NN_GX_MEM_SYSTEM:
        addrAlign = ALIGNMENT_SYSTEM_BUFFER;
        break;
    case NN_GX_MEM_TEXTURE:
        addrAlign = ALIGNMENT_TEXTURE;
        break;
    case NN_GX_MEM_VERTEXBUFFER:
        addrAlign = ALIGNMENT_VERTEX;
        break;
    case NN_GX_MEM_RENDERBUFFER:
        addrAlign = ALIGNMENT_RENDER_BUFFER;
        break;
    case NN_GX_MEM_DISPLAYBUFFER:
        addrAlign = ALIGNMENT_DISPLAY_BUFFER;
        break;
    case NN_GX_MEM_COMMANDBUFFER:
        addrAlign = ALIGNMENT_3D_COMMAND_BUFFER;
        break;
    default:
        NW_FATAL_ERROR("[DEMO:GraphicsMemoryAllocator] invalid parameter.(0x%X)\n", aim);
        break;
    }


    switch (area)
    {
    case NN_GX_MEM_FCRAM:
        // When using FCRAM, no need to manage this because it is allocated from the expanded heap that is supported by the SDK
        resultAddr = m_DeviceMemoryAllocator->Alloc(size, addrAlign);
        break;
    case NN_GX_MEM_VRAMA:
        {
            if (MathRoundup(m_CurrentAddrVramA, addrAlign) + size > nn::gx::GetVramEndAddr(nn::gx::MEM_VRAMA))
            {
                NW_FATAL_ERROR("[DEMO:GraphicsMemoryAllocator] lack of resource on VRAM-A.\n");
            }
            m_CurrentAddrVramA = MathRoundup(m_CurrentAddrVramA, addrAlign);
            resultAddr = reinterpret_cast<void*>(m_CurrentAddrVramA);
            m_CurrentAddrVramA += size;
        }
        break;
    case NN_GX_MEM_VRAMB:
        {
            if (MathRoundup(m_CurrentAddrVramB, addrAlign) + size > nn::gx::GetVramEndAddr(nn::gx::MEM_VRAMB))
            {
                NW_FATAL_ERROR("[DEMO:GraphicsMemoryAllocator] lack of resource on VRAM-B.\n");
            }
            m_CurrentAddrVramB = MathRoundup(m_CurrentAddrVramB, addrAlign);
            resultAddr = reinterpret_cast<void*>(m_CurrentAddrVramB);
            m_CurrentAddrVramB += size;
        }
        break;
    default:
        // This is not likely in normal use; FATAL_ERROR
        NW_FATAL_ERROR("[DEMO:GraphicsMemoryAllocator] invalid parameter.(0x%X)\n", area);
        break;
    }

    NW_NULL_ASSERT(resultAddr);

    return resultAddr;
}

//--------------------------------------------------------------------------
void
GraphicsMemoryAllocator::Deallocate(GLenum area, GLenum aim, GLuint id, void* addr)
{
    NW_ASSERTMSG (m_Initialized, "[DEMO:GraphicsMemoryAllocator] not initialized.\n");
    NW_UNUSED_VARIABLE(aim);
    NW_UNUSED_VARIABLE(id);

    switch (area)
    {
    case NN_GX_MEM_FCRAM:
        m_DeviceMemoryAllocator->Free(addr);
        break;

    default:
        // TODO: Implement a deallocator for the graphics memory allocator
        break;
    }
}


} // namespace demo
} // namespace nw
