/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/


//------------------------------------------------------------------
// Demo: ArchiveFont
// 
// Overview
//   This is a sample of building and destroying nw::font::ArchiveFont.
//   Two methods for building a font resource are indicated: the first loads it all into memory and then builds; the second builds as it loads successively.
// 
// 
// Controls
//   None.
// 
//------------------------------------------------------------------

#include <GLES2/gl2.h>
#include <GLES2/gl2ext.h>
#include <nn.h>
#include <nn/math.h>
#include <nn/fs.h>
#include <nn/applet.h>
#include <nw/assert.h>
#include <nw/font/font_WideTextWriter.h>
#include <nw/font/font_ArchiveFont.h>
#include <nw/font/font_RectDrawer.h>
#include <nw/demo.h>

#include <algorithm>

namespace
{

/*
    Load buffer size for sequential loading.

    The size of the buffer specified in nw::font::ArchiveFont::GetRequireBufferSize()
    must be nw::font::ArchiveFont::HEADER_SIZE or larger.
*/
const int DEMO_STREAMING_READ_BUFFER_SIZE   = 16 * 1024;

const char DEMO_LOAD_GROUPS[]    = "ascii";
// Load ASCII characters

//---------------------------------------------------------------------------
//
//
//
//
//
//
//---------------------------------------------------------------------------
void*
MemAlloc(
    size_t  size,
    u8      alignment = 4
)
{
    return nw::demo::SimpleApp::Allocate(size, alignment);
}

//---------------------------------------------------------------------------
//
//
//
//
//
//
//---------------------------------------------------------------------------
void*
DevMemAlloc(
    size_t  size,
    u8      alignment = 4
)
{
    return nw::demo::SimpleApp::AllocateDeviceMemory(size, alignment);
}

//---------------------------------------------------------------------------
//
//
//
//---------------------------------------------------------------------------
void
MemFree(void* memory)
{
    nw::demo::SimpleApp::Free(memory);
}

//---------------------------------------------------------------------------
//
//
//
//---------------------------------------------------------------------------
void*
InitShaders(nw::font::RectDrawer* pDrawer)
{
    const wchar_t* shaders = NW_DEMO_FILE_PATH( NW_FONT_RECTDRAWER_SHADERBINARY );
    nn::fs::FileReader shaderReader(shaders);

    const u32 fileSize = (u32)shaderReader.GetSize();

    void* shaderBinary = MemAlloc(fileSize);
    NN_NULL_ASSERT(shaderBinary);

    s32 read = shaderReader.Read(shaderBinary, fileSize);
    NN_ASSERT(read == fileSize);

    const u32 vtxBufCmdBufSize =
        nw::font::RectDrawer::GetVertexBufferCommandBufferSize(shaderBinary, fileSize);
    void *const vtxBufCmdBuf = DevMemAlloc(vtxBufCmdBufSize);
    NN_NULL_ASSERT(vtxBufCmdBuf);
    pDrawer->Initialize(vtxBufCmdBuf, shaderBinary, fileSize);

    MemFree(shaderBinary);

    return vtxBufCmdBuf;
}

//---------------------------------------------------------------------------
//
//
//
//
//---------------------------------------------------------------------------
void
InitDraw(
    int     width,
    int     height
)
{
    // Color buffer information
    // Switches the width and height to match the LCD direction.
    const nw::font::ColorBufferInfo colBufInfo = { height, width, PICA_DATA_DEPTH24_STENCIL8_EXT };

    const u32 screenSettingCommands[] =
    {
        // Viewport settings
        NW_FONT_CMD_SET_VIEWPORT( 0, 0, colBufInfo.width, colBufInfo.height ),

        // Disable scissoring
        NW_FONT_CMD_SET_DISABLE_SCISSOR( colBufInfo ),

        // Disable W buffer
        // Set the depth range
        // Disable polygon offset
        NW_FONT_CMD_SET_WBUFFER_DEPTHRANGE_POLYGONOFFSET(
            0.0f,           // wScale : W buffer is disabled at 0.0
            0.0f,           // depth range near
            1.0f,           // depth range far
            0,              // polygon offset units : polygon offset is disabled with 0.0
            colBufInfo),
    };

    nngxAdd3DCommand(screenSettingCommands, sizeof(screenSettingCommands), true);

    static const u32 s_InitCommands[] =
    {
        // Disable culling
        NW_FONT_CMD_SET_CULL_FACE( NW_FONT_CMD_CULL_FACE_DISABLE ),

        // Disable stencil test
        NW_FONT_CMD_SET_DISABLE_STENCIL_TEST(),

        // Disable depth test
        // Make all elements of the color buffer writable
        NW_FONT_CMD_SET_DEPTH_FUNC_COLOR_MASK(
            false,  // isDepthTestEnabled
            0,      // depthFunc
            true,   // depthMask
            true,   // red
            true,   // Green
            true,   // Blue
            true),  // Alpha

        // Disable early depth test
        NW_FONT_CMD_SET_ENABLE_EARLY_DEPTH_TEST( false ),

        // Framebuffer access control
        NW_FONT_CMD_SET_FBACCESS(
            true,   // colorRead
            true,   // colorWrite
            false,  // depthRead
            false,  // depthWrite
            false,  // stencilRead
            false), // stencilWrite
    };

    nngxAdd3DCommand(s_InitCommands, sizeof(s_InitCommands), true);
}

//---------------------------------------------------------------------------
//
//
//
//
//
//
//
//
//---------------------------------------------------------------------------
bool
InitFont(
    nw::font::ArchiveFont*  pFont,
    nn::fs::FileReader*     pFileReader,
    const char*             glyphGroups
)
{
    bool bSuccess;

    // Loads all font resource to the memory.
    const u32 fileSize = (u32)pFileReader->GetSize();
    if (fileSize <= 0)
    {
        return false;
    }


    const u32 readBufferSize = fileSize;
    void* readBuffer = MemAlloc(readBufferSize);
    if (readBuffer == NULL)
    {
        return false;
    }

    const int readSize = pFileReader->Read(readBuffer, readBufferSize);
    if (readSize != readBufferSize)
    {
        MemFree(readBuffer);
        return false;
    }


    // Gets the buffer size necessary for building the font, and allocates the buffer.
    // Specifies the glyph group to load for the second argument of GetRequireBufferSize.
    const u32 fontBufferSize = nw::font::ArchiveFont::GetRequireBufferSize(
                                                    readBuffer, glyphGroups);
    if (fontBufferSize == 0)
    {
        MemFree(readBuffer);
        return false;
    }

    void* const fontBuffer = DevMemAlloc(fontBufferSize, nw::font::GlyphDataAlignment);
    if (fontBuffer == NULL)
    {
        MemFree(readBuffer);
        return false;
    }

    // Builds fonts.
    // Specifies the glyph group to load for the fourth argument of Construct.
    // Same glyph group used when obtaining the fontBufferSize must be used.
    bSuccess = pFont->Construct(fontBuffer, fontBufferSize, readBuffer, glyphGroups);
    NW_ASSERT( bSuccess );


    //--- Fails if the font is already built or if the resource is invalid.
    if (! bSuccess)
    {
        MemFree(fontBuffer);
    }

    // readBuffer will not be used anymore, so it may be released.
    // fontBuffer must not be released if the build is successful.
    MemFree(readBuffer);

    // The size associated with building is reported.
    NN_LOG("InitFont:\n"
             "  fileSize=%7lld readBufferSize=%7d fontBufferSize=%7d\n",
             pFileReader->GetSize(), readBufferSize, fontBufferSize );

    return bSuccess;
}

//---------------------------------------------------------------------------
//
//
//
//
//
//
//
//---------------------------------------------------------------------------
bool
InitFontStreaming(
    nw::font::ArchiveFont*  pFont,
    nn::fs::FileReader*     pFileReader,
    const char*             glyphGroups
)
{
    s32 readSize;
    nw::font::ArchiveFont::ConstructContext context;
    nw::font::ArchiveFont::ConstructResult ret = nw::font::ArchiveFont::CONSTRUCT_ERROR;

    // Allocates buffer for sequential loading.
    const u32 readBufferSize = DEMO_STREAMING_READ_BUFFER_SIZE;

    // The size of the buffer specified by the nw::font::ArchiveFont::GetRequireBufferSize function must be nw::font::ArchiveFont::HEADER_SIZE or larger.
    // 
    NN_ASSERT(readBufferSize >= nw::font::ArchiveFont::HEADER_SIZE);

    void* const readBuffer = MemAlloc(readBufferSize);
    if (readBuffer == NULL)
    {
        NN_LOG("can't alloc %d bytes", readBufferSize);
        return false;
    }

    // To get the buffer size needed to build a font, read the start of the font resource.
    // 
    readSize = pFileReader->Read(readBuffer, nw::font::ArchiveFont::HEADER_SIZE);
    if (readSize != nw::font::ArchiveFont::HEADER_SIZE)
    {
        MemFree(readBuffer);
        NN_LOG("Fail to load %d bytes. Maybe too small bcfna.", nw::font::ArchiveFont::HEADER_SIZE);
        return false;
    }

    // Gets the buffer size necessary for building the font, and allocates the buffer.
    // Specifies the glyph group to load for the second argument of GetRequireBufferSize.
    const u32 fontBufferSize = nw::font::ArchiveFont::GetRequireBufferSize(
                                                    readBuffer, glyphGroups);
    if (fontBufferSize == 0)
    {
        MemFree(readBuffer);
        NN_LOG("ArchiveFont::GetRequireBufferSize failed");
        return false;
    }

    void* fontBuffer = DevMemAlloc(fontBufferSize, nw::font::GlyphDataAlignment);
    if (fontBuffer == NULL)
    {
        MemFree(readBuffer);
        NW_WARNING(false, "can't alloc %d bytes", fontBufferSize);
        return false;
    }


    // Starts font building.
    // Specifies the glyph group to load for the fourth argument of InitStreamingConstruct.
    // Same glyph group used when obtaining the fontBufferSize must be used.
    pFont->InitStreamingConstruct(&context, fontBuffer, fontBufferSize, glyphGroups);

    // Loads resource successively from the file stream, and passes the resource to the StreamingConstruct.
    while (readSize > 0)
    {
        ret = pFont->StreamingConstruct(&context, readBuffer, static_cast<u32>(readSize));
        if (ret == nw::font::ArchiveFont::CONSTRUCT_ERROR)
        {
            MemFree(fontBuffer);
            MemFree(readBuffer);
            NN_LOG("ArchiveFont::StreamingConstruct return error");
            return false;
        }

        readSize = pFileReader->Read(readBuffer, readBufferSize);
    }

    if (readSize < 0)
    {
        // Error
        MemFree(fontBuffer);
        MemFree(readBuffer);
        NN_LOG("readSize(=%d) < 0", readSize);
        return false;
    }

    // Resource load completion / font building should also be completed.
    NW_ASSERT(ret == nw::font::ArchiveFont::CONSTRUCT_FINISH);

    // readBuffer will not be used anymore, so it may be released.
    // fontBuffer must not be released if the build is successful.
    MemFree(readBuffer);

    // The size associated with building is reported.
    NN_LOG("InitFontStreaming:\n"
             "  fileSize=%7lld readBufferSize=%7d fontBufferSize=%7d\n",
             pFileReader->GetSize(), readBufferSize, fontBufferSize );

    return true;
}

//---------------------------------------------------------------------------
//
//
//
//---------------------------------------------------------------------------
void
CleanupFont(nw::font::ArchiveFont* pFont)
{
    // Destroys fonts.
    void* buffer = pFont->Destroy();

    // If the font is already built, a pointer to the buffer allocated at the build operation will be returned.
    if (buffer != NULL)
    {
        MemFree(buffer);
    }

    // After font data is destroyed, it cannot be used as a font until another Construct.
}

//---------------------------------------------------------------------------
//
//
//
//
//
//---------------------------------------------------------------------------
nw::font::DispStringBuffer*
AllocDispStringBuffer(int charMax)
{
    const u32 DrawBufferSize = nw::font::CharWriter::GetDispStringBufferSize(charMax);
    void *const bufMem = MemAlloc(DrawBufferSize);
    NN_NULL_ASSERT(bufMem);

    return nw::font::CharWriter::InitDispStringBuffer(bufMem, charMax);
}

//---------------------------------------------------------------------------
//
//
//
//
//
//---------------------------------------------------------------------------
void
SetupTextCamera(
    nw::font::RectDrawer*   pDrawer,
    int         			width,
    int         			height
)
{
    // Set the projection matrix for an orthogonal projection
    {
        // Sets the origin at the upper left, and the Y and Z axes in opposite directions.
        nn::math::MTX44 proj;
        f32 znear   =  0.0f;
        f32 zfar    = -1.0f;
        f32 t       =  0;
        f32 b       =  static_cast<f32>(height);
        f32 l       =  0;
        f32 r       =  static_cast<f32>(width);
        nn::math::MTX44OrthoPivot(&proj, l, r, b, t, znear, zfar, nn::math::PIVOT_UPSIDE_TO_TOP);
        pDrawer->SetProjectionMtx(proj);
    }

    // Set the model view matrix as an identity matrix
    {
        nn::math::MTX34 mv;
        nn::math::MTX34Identity(&mv);
        pDrawer->SetViewMtxForText(mv);
    }
}

//---------------------------------------------------------------------------
//
//
//
//
//
//
//
//
//---------------------------------------------------------------------------
void
DrawAscii(
    nw::font::RectDrawer*           pDrawer,
    nw::font::DispStringBuffer*     pDrawStringBuf,
    const nw::font::Font*           pFont,
    int         			        width,
    int         			        height,
    const wchar_t*                  title
)
{
    nw::font::WideTextWriter writer;
    writer.SetDispStringBuffer(pDrawStringBuf);
    writer.SetFont(pFont);

    writer.SetCursor(0, 0);

    writer.StartPrint();
    writer.Printf(L"DEMO: %ls\n", title);

    // Displays text sample.
    writer.Print(L"Glyph Sample:\n");
    writer.MoveCursorX(5);
    writer.Print(L" !\"#$%&'()*+,-./0123456789:;<=>?\n");
    writer.Print(L"@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_\n");
    writer.Print(L"`abcdefghijklmnopqrstuvwxyz{|}~\n");
    writer.EndPrint();

    pDrawer->BuildTextCommand(&writer);

    pDrawer->DrawBegin();

        SetupTextCamera(pDrawer, width, height);
        writer.UseCommandBuffer();

    pDrawer->DrawEnd();
}

}   // Namespace

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------
void
nnMain()
{
    nn::applet::Enable();

    // fs initialization
    nn::fs::Initialize();

    nw::demo::SimpleApp& demoApp = nw::demo::SimpleApp::GetInstance();

    demoApp.Initialize();

    nw::font::ArchiveFont fontNormal;
    nw::font::ArchiveFont fontStream;

    // Build font
    {
        nn::fs::FileReader fileReader(NW_DEMO_FILE_PATH(L"tahoma.bcfna"));

        fileReader.Seek(0, nn::fs::POSITION_BASE_BEGIN);
        const bool bSuccess = InitFont(&fontNormal, &fileReader, DEMO_LOAD_GROUPS);
        NW_ASSERTMSG(bSuccess, "Fail to load ArchiveFont by default.");

        fileReader.Seek(0, nn::fs::POSITION_BASE_BEGIN);
        const bool bSuccessStream = InitFontStreaming(&fontStream, &fileReader, DEMO_LOAD_GROUPS);
        NW_ASSERTMSG(bSuccessStream, "Fail to load ArchiveFont by streaming.");
    }

    // Build the render resource
    nw::font::RectDrawer drawer;

    void *const drawerBuf = InitShaders(&drawer);

    // Allocate buffer for render string
    nw::font::DispStringBuffer *const pDrawStringBuf = AllocDispStringBuffer(512);

    const nw::ut::FloatColor clearColor(0.3f, 0.3f, 0.3f, 1.0f);
    const f32 clearDepth = 0.0f;

    // Display font data
    volatile bool loop = true;
    while (loop)
    {
        demoApp.SetRenderingTarget(demoApp.DISPLAY0);
        {
            demoApp.GetFrameBufferObject().ClearBuffer(clearColor, clearDepth);

            InitDraw(demoApp.DISPLAY0_WIDTH, demoApp.DISPLAY0_HEIGHT);

            DrawAscii(
                &drawer,
                pDrawStringBuf,
                &fontNormal,
                demoApp.DISPLAY0_WIDTH,
                demoApp.DISPLAY0_HEIGHT,
                L"ArchiveFont - Construct");
        }

        demoApp.SetRenderingTarget(demoApp.DISPLAY1);
        {
            demoApp.GetFrameBufferObject().ClearBuffer(clearColor, clearDepth);

            InitDraw(demoApp.DISPLAY1_WIDTH, demoApp.DISPLAY1_HEIGHT);

            DrawAscii(
                &drawer,
                pDrawStringBuf,
                &fontStream,
                demoApp.DISPLAY1_WIDTH,
                demoApp.DISPLAY1_HEIGHT,
                L"ArchiveFont - StreamingConstruct");
        }

        demoApp.SwapBuffer(demoApp.DISPLAY_BOTH);
    }

    // Free the render string buffer
    MemFree(pDrawStringBuf);

    drawer.Finalize();

    MemFree(drawerBuf);

    // Destroy font
    CleanupFont(&fontStream);
    CleanupFont(&fontNormal);
}
