/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/


//------------------------------------------------------------------
// Demo: PackedFont
// 
// Overview
//   This is a sample of building and destroying nw::font::PackedFont.
//   Two methods for building a font resource are indicated: the first loads it all into memory and then builds; the second builds as it loads successively.
// 
// 
// Controls
//   None.
// 
//------------------------------------------------------------------


#include <GLES2/gl2.h>
#include <GLES2/gl2ext.h>
#include <nn.h>
#include <nn/math.h>
#include <nn/fs.h>
#include <nn/applet.h>
#include <nw/assert.h>
#include <nw/font/font_WideTextWriter.h>
#include <nw/font/font_PackedFont.h>
#include <nw/demo.h>

#include <algorithm>

namespace
{

/*
    Load buffer size for sequential loading.
    The size of the buffer specified by the nw::font::PackedFont::GetRequireBufferSize function must be nw::font::PackedFont::HEADER_SIZE or larger.
    

*/
const int DEMO_STREAMING_READ_BUFFER_SIZE   = 16 * 1024;

const char DEMO_LOAD_GROUPS[]    = "ascii";
    // Load ASCII characters
const f32 DEMO_CACHE_RATE        = 0.3f;
    // Allocate a cache that is 30% worth of the sheet loaded.

//---------------------------------------------------------------------------
//
//
//
//
//
//
//---------------------------------------------------------------------------
void*
MemAlloc(
    size_t  size,
    u8      alignment = 4
)
{
#if defined(NW_PLATFORM_CTR)
    return nw::demo::SimpleApp::Allocate(size, alignment);
#else
    return nw::demo::Alloc(size, alignment);
#endif
}

//---------------------------------------------------------------------------
//
//
//
//
//
//
//---------------------------------------------------------------------------
void*
DevMemAlloc(
    size_t  size,
    u8      alignment = 4
)
{
#if defined(NW_PLATFORM_CTR)
    return nw::demo::SimpleApp::AllocateDeviceMemory(size, alignment);
#else
    return nw::demo::Alloc(size, alignment);
#endif
}

//---------------------------------------------------------------------------
//
//
//
//---------------------------------------------------------------------------
void
MemFree(void* memory)
{
#if defined(NW_PLATFORM_CTR)
    nw::demo::SimpleApp::Free(memory);
#else
    nw::demo::Free(memory);
#endif
}

//---------------------------------------------------------------------------
//
//
//
//---------------------------------------------------------------------------
void
InitShaders(nw::font::TextWriterResource* pResource)
{
    const wchar_t* shaders = NW_DEMO_FILE_PATH( NW_FONT_SHADERBINARY );
    nn::fs::FileReader shaderReader(shaders);

    const u32 fileSize = (u32)shaderReader.GetSize();

    void* shaderBinary = MemAlloc(fileSize);

    s32 read = shaderReader.Read(shaderBinary, fileSize);
    NN_ASSERT(read == fileSize);

    pResource->InitResource(shaderBinary, fileSize);

    MemFree(shaderBinary);
}

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------
void
InitDraw()
{
    glClearColor(0.3f, 0.3f, 0.3f, 1.0f);

    glDisable(GL_CULL_FACE);            // Disable culling
    glDisable(GL_SCISSOR_TEST);         // Disable scissoring
    glDisable(GL_POLYGON_OFFSET_FILL);  // Disable polygon offset
#if defined(NW_PLATFORM_CTR)
    glDisable(GL_EARLY_DEPTH_TEST_DMP); // Disable early depth test
#endif
    glDisable(GL_DEPTH_TEST);           // Disable depth test
    glDisable(GL_STENCIL_TEST);         // Disable stencil test

    // Make all elements of the color buffer writable
    glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);
}

//---------------------------------------------------------------------------
//
//
//
//
//
//
//
//
//
//---------------------------------------------------------------------------
bool
InitFont(
    nw::font::PackedFont*   pFont,
    nn::fs::FileReader*     pFileReader,
    const char*             glyphGroups,
    f32                     rate
)
{
    bool bSuccess;

    // Loads all font resource to the memory.
    const u32 fileSize = (u32)pFileReader->GetSize();
    if( fileSize <= 0 )
    {
        NN_LOG("fail to get FileStream size.");
        return false;
    }

    const u32 readBufferSize = fileSize;
    void* readBuffer = MemAlloc(readBufferSize);
    if (readBuffer == NULL)
    {
        NN_LOG("fail to allocate memory for resource buffer(%d byte).", readBufferSize);
        return false;
    }

    const int readSize = pFileReader->Read(readBuffer, readBufferSize);
    if (readSize != readBufferSize)
    {
        MemFree(readBuffer);
        NN_LOG("fail to read resource file.");
        return false;
    }


    // Gets the buffer size necessary for building the font, and allocates the buffer.
    // Specifies the glyph group to load for the second argument of GetRequireBufferSize.
    // Specifies the ratio of the area to cache the expanded sheet for the third argument of GetRequireBufferSize.
    // Specifies the ratio in values 0.0 ~ 1.0.
    // Decreasing the ratio will decrease the amount of necessary memory, but when rendering many characters it will worsen the performance.
    // Increasing the ratio will have the opposite effect.
    const u32 fontBufferSize = nw::font::PackedFont::GetRequireBufferSize(
                                            readBuffer, glyphGroups, rate);
    if (fontBufferSize == 0)
    {
        MemFree(readBuffer);
        NN_LOG("fail to GetRequireBufferSize.");
        return false;
    }

    void* const fontBuffer = DevMemAlloc(fontBufferSize, nw::font::GlyphDataAlignment);
    if (fontBuffer == NULL)
    {
        MemFree(readBuffer);
        NW_WARNING(false, "fail to allocate memory for font buffer(%d byte).", fontBufferSize);
        return false;
    }

    // Builds fonts.
    // Specifies the glyph group to load for the fourth argument of Construct.
    // Same glyph group used when obtaining the fontBufferSize must be used.
    bSuccess = pFont->Construct(fontBuffer, fontBufferSize, readBuffer, glyphGroups);
    NW_ASSERT(bSuccess);


    //--- Fails if the font is already built or if the resource is invalid.
    if( ! bSuccess )
    {
        NN_LOG("fail to Construct.");
        MemFree(fontBuffer);
    }

    // readBuffer will not be used anymore, so it may be released.
    // fontBuffer must not be released if the build is successful.
    MemFree(readBuffer);

    // The size associated with building is reported.
    NN_LOG("InitFont:\n"
             "  fileSize=%7lld readBufferSize=%7d fontBufferSize=%7d\n",
             pFileReader->GetSize(), readBufferSize, fontBufferSize);

    return bSuccess;
}

//---------------------------------------------------------------------------
//
//
//
//
//
//
//
//
//---------------------------------------------------------------------------
bool
InitFontStreaming(
    nw::font::PackedFont*   pFont,
    nn::fs::FileReader*     pFileReader,
    const char*             glyphGroups,
    f32                     rate
)
{
    s32 readSize;
    nw::font::PackedFont::ConstructContext context;
    nw::font::PackedFont::ConstructResult ret = nw::font::PackedFont::CONSTRUCT_ERROR;

    // Allocates buffer for sequential loading.
    const u32 readBufferSize = DEMO_STREAMING_READ_BUFFER_SIZE;

    // The size of the buffer specified by the nw::font::PackedFont::GetRequireBufferSize function must be nw::font::PackedFont::HEADER_SIZE or larger.
    // 
    NN_ASSERT(readBufferSize >= nw::font::PackedFont::HEADER_SIZE);

    void* const readBuffer = MemAlloc(readBufferSize);
    if (readBuffer == NULL)
    {
        NW_WARNING(false, "fail to allocate read buffer(%d byte).", readBufferSize);
        return false;
    }

    // To get the buffer size needed to build a font, read the start of the font resource.
    // 
    readSize = pFileReader->Read(readBuffer, nw::font::PackedFont::HEADER_SIZE);
    if (readSize != nw::font::PackedFont::HEADER_SIZE)
    {
        MemFree(readBuffer);
        NN_LOG("Fail to load %d bytes. Maybe too small bcfna.", nw::font::PackedFont::HEADER_SIZE);
        return false;
    }


    // Gets the buffer size necessary for building the font, and allocates the buffer.
    // Specifies the glyph group to load for the second argument of GetRequireBufferSize.
    // Specifies the ratio of the area to cache the expanded sheet for the third argument of GetRequireBufferSize.
    // Specifies the ratio in values 0.0 ~ 1.0.
    // Decreasing the ratio will decrease the amount of necessary memory, but when rendering many characters it will worsen the performance.
    // Increasing the ratio will have the opposite effect.
    const u32 fontBufferSize = nw::font::PackedFont::GetRequireBufferSize(
                                            readBuffer, glyphGroups, rate);
    if (fontBufferSize == 0)
    {
        MemFree(readBuffer);
        NN_LOG("fail to GetRequireBufferSize.");
        return false;
    }

    void* fontBuffer = DevMemAlloc(fontBufferSize, nw::font::GlyphDataAlignment);
    if (fontBuffer == NULL)
    {
        MemFree(readBuffer);
        NN_LOG("fail to allocate memory for font buffer(%d byte).", fontBufferSize);
        return false;
    }


    // Starts font building.
    // Specifies the glyph group to load for the fourth argument of InitStreamingConstruct.
    // Same glyph group used when obtaining the fontBufferSize must be used.
    pFont->InitStreamingConstruct(&context, fontBuffer, fontBufferSize, glyphGroups);

    // Loads resource successively from the file stream, and passes the resource to the StreamingConstruct.
    while (readSize > 0)
    {
        ret = pFont->StreamingConstruct(&context, readBuffer, static_cast<u32>(readSize));
        if (ret == nw::font::PackedFont::CONSTRUCT_ERROR)
        {
            MemFree(fontBuffer);
            MemFree(readBuffer);
            NN_LOG("fail to StreamingConstruct.");
            return false;
        }

        readSize = pFileReader->Read(readBuffer, readBufferSize);
    }

    if (readSize < 0)
    {
        // Error
        MemFree(fontBuffer);
        MemFree(readBuffer);
        NN_LOG("fail to read resource.");
        return false;
    }

    // Resource load completion / font building should also be completed.
    NW_ASSERT(ret == nw::font::PackedFont::CONSTRUCT_FINISH);

    // readBuffer will not be used anymore, so it may be released.
    // fontBuffer must not be released if the build is successful.
    MemFree(readBuffer);

    // The size associated with building is reported.
    NN_LOG("InitFontStreaming:\n"
             "  fileSize=%7lld readBufferSize=%7d fontBufferSize=%7d\n",
             pFileReader->GetSize(), readBufferSize, fontBufferSize);

    return true;
}

//---------------------------------------------------------------------------
//
//
//
//---------------------------------------------------------------------------
void
CleanupFont(nw::font::PackedFont* pFont)
{
    // Destroys fonts.
    void* buffer = pFont->Destroy();

    // If the font is already built, a pointer to the buffer allocated at the build operation will be returned.
    if( buffer != NULL )
    {
        MemFree(buffer);
    }

    // After font data is destroyed, it cannot be used as a font until another Construct.
}

//---------------------------------------------------------------------------
//
//
//
//
//
//---------------------------------------------------------------------------
void
SetupTextCamera(
    const nw::font::TextWriterResource&
                textWriterResource,
    int         width,
    int         height
)
{
#if defined(NW_PLATFORM_CTR)
    const GLsizei lcdWidth  = height;
    const GLsizei lcdHeight = width;
#else
    const GLsizei lcdWidth  = width;
    const GLsizei lcdHeight = height;
#endif
    glViewport(0, 0, lcdWidth, lcdHeight);

    // Set the projection matrix for an orthogonal projection
    {
        // Sets the origin at the upper left, and the Y and Z axes in opposite directions.
        nn::math::MTX44 proj;
        f32 znear   =  0.0f;
        f32 zfar    = -1.0f;
        f32 t       =  0;
        f32 b       =  static_cast<f32>(height);
        f32 l       =  0;
        f32 r       =  static_cast<f32>(width);
        nn::math::MTX44OrthoPivot(&proj, l, r, b, t, znear, zfar, nn::math::PIVOT_UPSIDE_TO_TOP);
        textWriterResource.SetProjectionMtx(proj);
    }

    // Set the model view matrix as an identity matrix
    {
        nn::math::MTX34 mv;
        nn::math::MTX34Identity(&mv);
        textWriterResource.SetViewMtx(mv);
    }
}

//---------------------------------------------------------------------------
//
//
//
//
//
//---------------------------------------------------------------------------
void
DrawAscii(
    const nw::font::Font*           pFont,
    nw::font::TextWriterResource*   pTextWriterResource,
    const wchar_t*                  title
)
{
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    nw::font::WideTextWriter writer;

    writer.SetFont(pFont);
    writer.SetTextWriterResource(pTextWriterResource);
    writer.SetupGX();
    writer.SetCursor(0, 0);

    writer.Printf(L"DEMO: %ls\n", title);

    // Displays text sample.
    writer.Print(L"Glyph Sample:\n");
    writer.MoveCursorX(5);
    writer.Print(L" !\"#$%&'()*+,-./0123456789:;<=>?\n");
    writer.Print(L"@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_\n");
    writer.Print(L"`abcdefghijklmnopqrstuvwxyz{|}~\n");

    glFlush();
}

}   // Namespace

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------
void
nnMain()
{
    nn::applet::Enable();

    // fs initialization
    nn::fs::Initialize();

    nw::demo::SimpleApp& demoApp = nw::demo::SimpleApp::GetInstance();

    demoApp.Initialize();

    nw::font::PackedFont fontNormal;
    nw::font::PackedFont fontStream;

    // Build font
    {
        nn::fs::FileReader fileReader(NW_DEMO_FILE_PATH(L"tahoma.bcfna"));

        fileReader.Seek(0, nn::fs::POSITION_BASE_BEGIN);
        const bool bSuccess = InitFont(&fontNormal, &fileReader, DEMO_LOAD_GROUPS, DEMO_CACHE_RATE);
        NW_ASSERTMSG(bSuccess, "Fail to load PackedFont by default.");

        fileReader.Seek(0, nn::fs::POSITION_BASE_BEGIN);
        const bool bSuccessStream = InitFontStreaming(&fontStream, &fileReader, DEMO_LOAD_GROUPS, DEMO_CACHE_RATE);
        NW_ASSERTMSG(bSuccessStream, "Fail to load PackedFont by streaming.");
    }

    // Build the render resource
    nw::font::TextWriterResource textWriterResource;
    InitShaders(&textWriterResource);

    textWriterResource.ActiveGlProgram();

    InitDraw();

    // Display font data
    volatile bool loop = true;
    while (loop)
    {
        demoApp.SetRenderingTarget(demoApp.DISPLAY0);
        {
            glBindFramebuffer(GL_FRAMEBUFFER, demoApp.GetFrameBufferObject());
            SetupTextCamera(textWriterResource, demoApp.DISPLAY0_WIDTH, demoApp.DISPLAY0_HEIGHT);
            DrawAscii(&fontNormal, &textWriterResource, L"PackedFont - Construct");
        }

        demoApp.SetRenderingTarget(demoApp.DISPLAY1);
        {
            glBindFramebuffer(GL_FRAMEBUFFER, demoApp.GetFrameBufferObject());
            SetupTextCamera(textWriterResource, demoApp.DISPLAY1_WIDTH, demoApp.DISPLAY1_HEIGHT);
            DrawAscii(&fontStream, &textWriterResource, L"PackedFont - StreamingConstruct");
        }

        demoApp.SwapBuffer(demoApp.DISPLAY_BOTH);
    }

    // Destroy font
    CleanupFont(&fontStream);
    CleanupFont(&fontNormal);
}
