/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/


//------------------------------------------------------------------
// Demo: VramCmdBuf
//
// Overview
//   This is a sample that uses the command buffer located in VRAM.
//
// Controls
//   None.
//
//------------------------------------------------------------------

#include <GLES2/gl2.h>
#include <GLES2/gl2ext.h>
#include <nn.h>
#include <nn/math.h>
#include <nn/fs.h>
#include <nn/applet.h>
#include <nw/sdk.h>
#include <nw/font/font_TextWriter.h>
#include <nw/font/font_ResFont.h>
#include <nw/font/font_RectDrawer.h>
#include <nw/font/font_DispStringBuffer.h>
#include <nw/demo.h>

// This is the section particularly related to using the VRAM command buffer in the code.
#define USE_VRAM_COMMAND_BUFFER 1

namespace
{

bool    s_InitAsciiString   = false;
u32     s_Counter           = 0;
u8      s_Color             = 0;

//---------------------------------------------------------------------------
//
//
//
//
//
//
//---------------------------------------------------------------------------
void*
MemAlloc(
    size_t  size,
    u8      alignment = 4
)
{
    return nw::demo::SimpleApp::Allocate(size, alignment);
}

//---------------------------------------------------------------------------
//
//
//
//
//
//
//---------------------------------------------------------------------------
void*
DevMemAlloc(
    size_t  size,
    u8      alignment = 4
)
{
    return nw::demo::SimpleApp::AllocateDeviceMemory(size, alignment);
}

//---------------------------------------------------------------------------
//
//
//
//---------------------------------------------------------------------------
void
MemFree(void* memory)
{
    nw::demo::SimpleApp::Free(memory);
}

//---------------------------------------------------------------------------
//
//
//
//---------------------------------------------------------------------------
void*
InitShaders(nw::font::RectDrawer* pDrawer)
{
    const wchar_t* shaders = NW_DEMO_FILE_PATH( NW_FONT_RECTDRAWER_SHADERBINARY );
    nn::fs::FileReader shaderReader(shaders);

    const u32 fileSize = (u32)shaderReader.GetSize();

    void* shaderBinary = MemAlloc(fileSize);
    NN_NULL_ASSERT(shaderBinary);

    s32 read = shaderReader.Read(shaderBinary, fileSize);
    NN_ASSERT(read == fileSize);

    const u32 vtxBufCmdBufSize =
        nw::font::RectDrawer::GetVertexBufferCommandBufferSize(shaderBinary, fileSize);
    void *const vtxBufCmdBuf = DevMemAlloc(vtxBufCmdBufSize);
    NN_NULL_ASSERT(vtxBufCmdBuf);
    pDrawer->Initialize(vtxBufCmdBuf, shaderBinary, fileSize);

    MemFree(shaderBinary);

    return vtxBufCmdBuf;
}

//---------------------------------------------------------------------------
//
//
//
//
//---------------------------------------------------------------------------
void
InitDraw(
    int     width,
    int     height
)
{
    // Color buffer information
    // Switches the width and height to match the LCD direction.
    const nw::font::ColorBufferInfo colBufInfo = { height, width, PICA_DATA_DEPTH24_STENCIL8_EXT };

    const u32 screenSettingCommands[] =
    {
        // Viewport settings
        NW_FONT_CMD_SET_VIEWPORT( 0, 0, colBufInfo.width, colBufInfo.height ),

        // Disable scissoring
        NW_FONT_CMD_SET_DISABLE_SCISSOR( colBufInfo ),

        // Disable W buffer
        // Set the depth range
        // Disable polygon offset
        NW_FONT_CMD_SET_WBUFFER_DEPTHRANGE_POLYGONOFFSET(
            0.0f,           // wScale : W buffer is disabled at 0.0
            0.0f,           // depth range near
            1.0f,           // depth range far
            0,              // polygon offset units : polygon offset is disabled with 0.0
            colBufInfo),
    };

    nngxAdd3DCommand(screenSettingCommands, sizeof(screenSettingCommands), true);

    static const u32 s_InitCommands[] =
    {
        // Disable culling
        NW_FONT_CMD_SET_CULL_FACE( NW_FONT_CMD_CULL_FACE_DISABLE ),

        // Disable stencil test
        NW_FONT_CMD_SET_DISABLE_STENCIL_TEST(),

        // Disable depth test
        // Make all elements of the color buffer writable
        NW_FONT_CMD_SET_DEPTH_FUNC_COLOR_MASK(
            false,  // isDepthTestEnabled
            0,      // depthFunc
            true,   // depthMask
            true,   // red
            true,   // Green
            true,   // Blue
            true),  // Alpha

        // Disable early depth test
        NW_FONT_CMD_SET_ENABLE_EARLY_DEPTH_TEST( false ),

        // Framebuffer access control
        NW_FONT_CMD_SET_FBACCESS(
            true,   // colorRead
            true,   // colorWrite
            false,  // depthRead
            false,  // depthWrite
            false,  // stencilRead
            false), // stencilWrite
    };

    nngxAdd3DCommand(s_InitCommands, sizeof(s_InitCommands), true);
}

//---------------------------------------------------------------------------
//
//
//
//
//
//
//---------------------------------------------------------------------------
bool
InitFont(
    nw::font::ResFont*  pFont,
    const wchar_t*      filePath
)
{
    // Loads font resources
    nn::fs::FileReader fontReader(filePath);

    s32 fileSize = (s32)fontReader.GetSize();
    if ( fileSize <= 0 )
    {
        return false;
    }

    void* buffer = DevMemAlloc(fileSize, nw::font::GlyphDataAlignment);
    if (buffer == NULL)
    {
        return false;
    }

    s32 readSize = fontReader.Read(buffer, fileSize);
    if (readSize != fileSize)
    {
        MemFree(buffer);
        return false;
    }

    // Sets the font resource.
    bool bSuccess = pFont->SetResource(buffer);
    NN_ASSERT(bSuccess);

    //--- Fails if the resource is already set, already loaded, or invalid.
    if (! bSuccess)
    {
        MemFree(buffer);
    }

    // Sets the render buffer.
    const u32 drawBufferSize = nw::font::ResFont::GetDrawBufferSize(buffer);
    void* drawBuffer = MemAlloc(drawBufferSize, 4);
    NN_NULL_ASSERT(drawBuffer);
    pFont->SetDrawBuffer(drawBuffer);

    return bSuccess;
}

//---------------------------------------------------------------------------
//
//
//
//---------------------------------------------------------------------------
void
CleanupFont(nw::font::ResFont* pFont)
{
    // Disable render buffer
    // When the render buffer is set, the pointer to the buffer passed to SetDrawBuffer at build time is returned.
    // 
    void *const drawBuffer = pFont->SetDrawBuffer(NULL);
    if (drawBuffer != NULL)
    {
        MemFree(drawBuffer);
    }

    // Returns a pointer to the resource passed when SetResource was called if a font is set.
    // 
    void *const resource = pFont->RemoveResource();
    if (resource != NULL)
    {
        MemFree(resource);
    }

    // After RemoveResource is executed, data cannot be used as a font until SetResource is executed again.
}

//---------------------------------------------------------------------------
//
//
//
//
//
//---------------------------------------------------------------------------
nw::font::DispStringBuffer*
AllocDispStringBuffer(int charMax)
{
    const u32 DrawBufferSize = nw::font::CharWriter::GetDispStringBufferSize(charMax);
    void *const bufMem = MemAlloc(DrawBufferSize);
    NN_NULL_ASSERT(bufMem);

    return nw::font::CharWriter::InitDispStringBuffer(bufMem, charMax);
}

//---------------------------------------------------------------------------
//
//
//
//
//
//---------------------------------------------------------------------------
void
SetupTextCamera(
    nw::font::RectDrawer*   pDrawer,
    int                     width,
    int                     height
)
{
    // Set the projection matrix for an orthogonal projection
    {
        // Sets the origin at the upper left, and the Y and Z axes in opposite directions.
        nn::math::MTX44 proj;
        f32 znear   =  0.0f;
        f32 zfar    = -1.0f;
        f32 t       =  0;
        f32 b       =  static_cast<f32>(height);
        f32 l       =  0;
        f32 r       =  static_cast<f32>(width);
        nn::math::MTX44OrthoPivot(&proj, l, r, b, t, znear, zfar, nn::math::PIVOT_UPSIDE_TO_TOP);
        pDrawer->SetProjectionMtx(proj);
    }

    // Set the model view matrix as an identity matrix
    {
        nn::math::MTX34 mv;
        nn::math::MTX34Identity(&mv);
        pDrawer->SetViewMtxForText(mv);
    }
}

//---------------------------------------------------------------------------
//
//
//
//
//
//
//
//---------------------------------------------------------------------------
void
DrawAscii(
    nw::font::RectDrawer*           pDrawer,
    nw::font::DispStringBuffer*     pDrawStringBuf,
    nw::font::ResFont*              pFont,
    int                             width,
    int                             height
)
{
    nw::font::TextWriter writer;
    writer.SetDispStringBuffer(pDrawStringBuf);
    writer.SetFont(pFont);

    writer.SetCursor(0, 0);

    // Create the string render command once because the string does not change.
    if (! s_InitAsciiString)
    {
#if USE_VRAM_COMMAND_BUFFER
        GLvoid* (*pGlAllocator)(GLenum, GLenum, GLuint, GLsizei) = NULL;
        nngxGetAllocator(&pGlAllocator, NULL);
        if (pGlAllocator)
        {
            // Declares the use of the command buffer located in VRAM.
            // Generates a 3D command sequence that corresponds to the command buffer.
            pDrawStringBuf->SetCommandBufferMode(pDrawStringBuf->CBM_VRAM);
        }
#endif

        writer.StartPrint();
            (void)writer.Print("DEMO: VramCmdBuf\n");
            (void)writer.Print("\n");

            // Display a sample of ASCII characters
            (void)writer.Print("All ASCII Character listing:\n");
            (void)writer.Print("\n");
            (void)writer.Print(" !\"#$%&'()*+,-./\n");
            (void)writer.Print("0123456789:;<=>?\n");
            (void)writer.Print("@ABCDEFGHIJKLMNO\n");
            (void)writer.Print("PQRSTUVWXYZ[\\]^_\n");
            (void)writer.Print("`abcdefghijklmno\n");
            (void)writer.Print("pqrstuvwxyz{|}~\n");
        writer.EndPrint();
        pDrawer->BuildTextCommand(&writer);

#if USE_VRAM_COMMAND_BUFFER
        if (pGlAllocator)
        {
            // Allocates command buffer area in the VRAM and transfers the generated command sequence.
            // 

            void* cmdBuf = pGlAllocator(
                NN_GX_MEM_VRAMA,
                NN_GX_MEM_COMMANDBUFFER,
                0,
                pDrawStringBuf->commandBufferSize * sizeof(u32));

            // Since the 3D command sequence that was generated is for command jump only, it is not used if area cannot be allocated in the VRAM.
            // 
            NW_NULL_ASSERT(cmdBuf);

            pDrawStringBuf->SetVramCommandBuffer(cmdBuf);
            pDrawStringBuf->SendToVram();

            // Waits until the transfer to VRAM completes.
            nngxRunCmdlist();
            nngxWaitCmdlistDone();
        }
#endif

        s_InitAsciiString = true;
    }

    // The character's color can be changed without regenerating the string render command.
    writer.SetTextColor(nw::ut::Color8(s_Color, 255, s_Color, 255));
    s_Color++;

#if USE_VRAM_COMMAND_BUFFER

    // Gets start address of buffer of unissued commands.
    nngxFlush3DCommand();
    GLint cmdStartAddr = 0;
    nngxGetCmdlistParameteri(NN_GX_CMDLIST_CURRENT_BUFADDR, &cmdStartAddr);

#endif // USE_VRAM_COMMAND_BUFFER

    pDrawer->DrawBegin();

        SetupTextCamera(pDrawer, width, height);
#if USE_VRAM_COMMAND_BUFFER
        writer.UseCommandBuffer(pDrawer);
#else
        writer.UseCommandBuffer();
#endif

    pDrawer->DrawEnd();

#if USE_VRAM_COMMAND_BUFFER

    GLint cmdFirstReturnAddr = (GLint)pDrawer->GetCommandJumpFirstReturnAddr();
    if (cmdFirstReturnAddr != 0)
    {
        // A dedicated 3D command issuing method is required if a command jump is used.

        // Command list execution has to have stopped when calling nngxFlush3DCommandPartially.
        // 
        nngxStopCmdlist();
        nngxWaitCmdlistDone();

        nngxFlush3DCommandPartially(cmdFirstReturnAddr - cmdStartAddr);
        NW_GL_ASSERT();

        // Completes the last jump command.
        // The last jump command is not completed until the 3D command is issued and the entire length of the 3D command sequence is confirmed.
        // 
        GLint cmdEndAddr = 0;
        nngxGetCmdlistParameteri(NN_GX_CMDLIST_CURRENT_BUFADDR, &cmdEndAddr);
        pDrawer->FixCommandJump((void*)cmdEndAddr);

        // The command buffer must flush the cache itself.
        nngxUpdateBuffer((void*)cmdStartAddr, cmdEndAddr - cmdStartAddr);

        // Restarts execution of the command list.
        nngxRunCmdlist();
    }

#endif // USE_VRAM_COMMAND_BUFFER
}

//---------------------------------------------------------------------------
//
//
//
//
//
//
//
//---------------------------------------------------------------------------
void
DrawCounter(
    nw::font::RectDrawer*           pDrawer,
    nw::font::DispStringBuffer*     pDrawStringBuf,
    nw::font::ResFont*              pFont,
    int                             width,
    int                             height
)
{
    nw::font::TextWriter writer;
    writer.SetDispStringBuffer(pDrawStringBuf);
    writer.SetFont(pFont);

    writer.SetCursor(0, 0);
    writer.SetFixedWidth(8.0f);
    writer.EnableFixedWidth(true);

    pDrawer->DrawBegin();

        SetupTextCamera(pDrawer, width, height);

        // Generates the counter render command each time to use it.
        writer.StartPrint();
            (void)writer.Printf("Counter      : %d\n", s_Counter);
        writer.EndPrint();
        pDrawer->BuildTextCommand(&writer);
        writer.UseCommandBuffer();

        // Generates a counter 1/60 render command each time to use it.
        writer.StartPrint();
            (void)writer.Printf("Counter 1/60 : %d\n", s_Counter / 60);
        writer.EndPrint();
        pDrawer->BuildTextCommand(&writer);
        writer.UseCommandBuffer();

    pDrawer->DrawEnd();

    s_Counter++;
}

}   // Namespace


//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------
void
nnMain()
{
    nn::applet::Enable();

    // fs initialization
    nn::fs::Initialize();

    nw::demo::SimpleApp& demoApp = nw::demo::SimpleApp::GetInstance();

    demoApp.Initialize();

    nw::font::ResFont font;

    // Build font
    {
        bool bSuccess = InitFont(&font, NW_DEMO_FILE_PATH(L"tahoma.bcfnt"));
        NN_ASSERTMSG(bSuccess, "Fail to load ResFont.");
    }

    // Build the render resource
    nw::font::RectDrawer drawer;

    void *const drawerBuf = InitShaders(&drawer);

    // Allocate buffer for render string
    nw::font::DispStringBuffer *const pDrawStringBuf0 = AllocDispStringBuffer(512);
    nw::font::DispStringBuffer *const pDrawStringBuf1 = AllocDispStringBuffer(128);

#if USE_VRAM_COMMAND_BUFFER
    pDrawStringBuf0->SetCommandBufferMode(nw::font::DispStringBuffer::CBM_VRAM);
#endif

    const nw::ut::FloatColor clearColor(0.3f, 0.3f, 0.3f, 1.0f);
    const f32 clearDepth = 0.0f;

    // Display font data
    volatile bool loop = true;
    while (loop)
    {
        demoApp.SetRenderingTarget(demoApp.DISPLAY0);
        {
            demoApp.GetFrameBufferObject().ClearBuffer(clearColor, clearDepth);

            InitDraw(demoApp.DISPLAY0_WIDTH, demoApp.DISPLAY0_HEIGHT);

            DrawAscii(&drawer, pDrawStringBuf0, &font, demoApp.DISPLAY0_WIDTH, demoApp.DISPLAY0_HEIGHT);
        }

        demoApp.SetRenderingTarget(demoApp.DISPLAY1);
        {
            demoApp.GetFrameBufferObject().ClearBuffer(clearColor, clearDepth);

            InitDraw(demoApp.DISPLAY1_WIDTH, demoApp.DISPLAY1_HEIGHT);

            DrawCounter(&drawer, pDrawStringBuf1, &font, demoApp.DISPLAY1_WIDTH, demoApp.DISPLAY1_HEIGHT);
        }

        demoApp.SwapBuffer(demoApp.DISPLAY_BOTH);
    }

#if USE_VRAM_COMMAND_BUFFER
    {
        void (*pGlDeallocator)(GLenum, GLenum, GLuint, GLvoid*) = NULL;
        nngxGetAllocator(NULL, &pGlDeallocator);
        if (pGlDeallocator)
        {
            void* cmdBuf = pDrawStringBuf0->SetVramCommandBuffer(NULL);
            pGlDeallocator(NN_GX_MEM_VRAMA, NN_GX_MEM_COMMANDBUFFER, 0, cmdBuf);
        }
    }
#endif

    // Free the render string buffer
    MemFree(pDrawStringBuf1);
    MemFree(pDrawStringBuf0);

    drawer.Finalize();

    MemFree(drawerBuf);

    // Destroy font
    CleanupFont(&font);
}
