/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#ifndef NW_FONT_FONT_H_
#define NW_FONT_FONT_H_

#if defined(_MSC_VER) && _MSC_VER >= 1500
#pragma once
#endif

#include <nn/types.h>
#include <nw/font/font_CharStrmReader.h>
#include <nw/font/font_ResourceFormat.h>
#include <climits>

#ifdef NW_PLATFORM_CTR
#if NN_CURRENT_VERSION_NUMBER >= NN_VERSION_NUMBER(4,0,0,0)
#pragma diag_suppress 1301 // padding inserted in struct.
#pragma diag_suppress 2530 // padding added to end of struct.
#endif
#endif
#define NW_FONT_MIN_ASSERT(exp, min)            NN_ASSERT((exp) >= (min))

#define NW_FONT_MINMAX_ASSERT(exp, min, max)    NN_ASSERT((exp) >= (min) && (exp) <= (max))


namespace nw {
namespace font {

class Font;

//---- Texture format type
typedef u16     TexFmt;

namespace internal {

union TextureSize
{
    struct
    {
        u16                 height;   //
        u16                 width;    //
    };

    u32                 size;   //
};

class TextureObject
{
public:
                            TextureObject();

    void                    Set(
                                u32         name,
                                const Font* pFont,
                                const void* pImage,
                                TexFmt      format,
                                u16         width,
                                u16         height);

    void                    Reset()
    {
        Set(0, NULL, NULL, FONT_SHEET_FORMAT_A8, 0, 0);
    }

    u32                     GetName() const
    {
        return m_Name;
    }

    void                    SetName(u32 texName)
    {
        m_Name = texName;
    }

    const Font*             GetFont() const
    {
        return m_pFont;
    }

    uptr                    GetImage() const
    {
        return m_pImage;
    }

    u8                      GetFormat() const
    {
        return m_Format;
    }

    const TextureSize       GetSize() const
    {
        return m_Size;
    }

    u32                     GetWrapFilter() const;

private:
    u32                     m_Name;     //
    const Font*             m_pFont;    //
    uptr                    m_pImage;   //
    TextureSize             m_Size;     //
    u8                      m_Format;   //
};

}   // namespace internal


//---- Character code type
typedef u16     CharCode;

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------
struct Glyph
{
    const void* pTexture;                   //
    CharWidths  widths;                     //
    u8          height;                     //
    u16         texWidth;                   //
    u16         texHeight;                  //
    u16         cellX;                      //
    u16         cellY;                      //
    TexFmt      texFormat;                  //
    u8          isSheetUpdated;             //
    const internal::TextureObject*
                pTextureObject;             //

    //
                Glyph();
};

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------
class Font
{
public:
    /* ------------------------------------------------------------------------
            Types
       ------------------------------------------------------------------------ */
    enum Type
    {
        TYPE_NULL,          //
        TYPE_ROM,           // RomFont
        TYPE_RESOURCE,      // ResFont
        TYPE_PAIR           // <CODE>PairFont</CODE>
    };


    /* ------------------------------------------------------------------------
            Constants
       ------------------------------------------------------------------------ */
    // Invalid character code
    static const CharCode   INVALID_CHARACTER_CODE = INVALID_CHAR_CODE;


    /* ------------------------------------------------------------------------
            Functions
       ------------------------------------------------------------------------ */

    //
    //

    //
                            Font();

    //
    virtual                 ~Font();

    //


    //
    //

    //
    //
    //
    //
    virtual int             GetWidth() const = 0;

    //
    //
    //
    //
    virtual int             GetHeight() const = 0;

    //
    //
    //
    //
    virtual int             GetAscent() const = 0;

    //
    //
    //
    //
    virtual int             GetDescent() const = 0;

    //
    //
    //
    //
    virtual int             GetMaxCharWidth() const = 0;

    //
    //
    //
    //
    virtual Type            GetType() const = 0;

    //
    //
    //
    //
    virtual TexFmt          GetTextureFormat() const = 0;

    //
    //
    //
    //
    virtual int             GetLineFeed() const = 0;

    //
    //
    //
    //
    virtual const CharWidths
                            GetDefaultCharWidths() const = 0;

    //


    //
    //

    //
    //
    //
    //
    virtual void            SetLineFeed(int linefeed) = 0;

    //
    //
    //
    //
    virtual void            SetDefaultCharWidths(
                                const CharWidths& widths
                            ) = 0;

    //
    //
    //
    //
    //
    //
    //
    virtual bool            SetAlternateChar(CharCode c) = 0;

    //


    //
    //

    //
    //
    //
    //
    //
    //
    virtual int             GetCharWidth(CharCode c) const = 0;

    //
    //
    //
    //
    //
    //
    virtual const CharWidths
                            GetCharWidths(CharCode c) const = 0;

    //
    //
    //
    //
    //
    virtual void            GetGlyph(
                                Glyph*      pGlyph,
                                CharCode    c
                            ) const = 0;

    //
    //
    //
    //
    //
    //
    virtual bool            HasGlyph(
                                CharCode    c
                            ) const = 0;

    //


    //
    //

    //
    //
    //
    //
    virtual CharacterCode   GetCharacterCode() const = 0;

    //
    //
    //
    //
    //
    //
    //
    //
    //
    const CharStrmReader    GetCharStrmReader(char dummy) const;

    //
    //
    //
    //
    //
    //
    //
    //
    //
    const CharStrmReader    GetCharStrmReader(wchar_t dummy) const;

    //


    //
    //

    //
    //
    //
    //
    virtual int             GetBaselinePos() const = 0;

    //
    //
    //
    //
    virtual int             GetCellHeight() const = 0;

    //
    //
    //
    //
    virtual int             GetCellWidth() const = 0;

    //

    //
    //

    //
    //
    //
    //
    //
    //
    //
    virtual void            EnableLinearFilter(
                                bool    atSmall,
                                bool    atLarge
                            ) = 0;

    //
    //
    //
    //
    //
    virtual bool            IsLinearFilterEnableAtSmall() const = 0;

    //
    //
    //
    //
    //
    virtual bool            IsLinearFilterEnableAtLarge() const = 0;

    //
    //
    //
    //
    virtual u32             GetTextureWrapFilterValue() const = 0;
    //
};

namespace internal {

inline
u32
TextureObject::GetWrapFilter() const
{
    return m_pFont->GetTextureWrapFilterValue();
}

void                    LoadTexture(
                            u16         texWidth,
                            u16         texHeight,
                            TexFmt      texFormat,
                            const void* pImage,
                            bool        isSmallLinearFilter,
                            bool        isLargeLinearFilter);

}   // namespace internal

}   // namespace font
}   // namespace nw
#ifdef NW_PLATFORM_CTR
#if NN_CURRENT_VERSION_NUMBER >= NN_VERSION_NUMBER(4,0,0,0)
#pragma diag_default 2530 // padding added to end of struct.
#pragma diag_default 1301 // padding inserted in struct.
#endif
#endif

#endif //  NW_FONT_FONT_H_
