/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#include "precompiled.h"

#include <nw/font/font_ArchiveFontBase.h>
#include <nw/ut/ut_Inlines.h>
#include <nw/assert.h>
#include <nw/math/math_Arithmetic.h>

#include <cstring>
#include <algorithm>

namespace nw {
namespace font {

const char ArchiveFontBase::LOAD_GLYPH_ALL[1] = "";


/* =======================================================================
        ArchiveFontBase public
   ======================================================================== */

/* ------------------------------------------------------------------------
        Constructors/destructors
   ------------------------------------------------------------------------ */

ArchiveFontBase::ArchiveFontBase()
:   m_pGlyphIndexAdjustArray(NULL)
{
}

ArchiveFontBase::~ArchiveFontBase()
{
}



/* ------------------------------------------------------------------------
        Font Class Override
   ------------------------------------------------------------------------ */

const CharWidths
ArchiveFontBase::GetCharWidths(CharCode c) const
{
    GlyphIndex index = GetGlyphIndex(c);

    // The glyph index passed to GetCharWidthsFromIndex is the index before correction, but when there is no sheet, use AdjustIndex to test whether there is a sheet because you have to get the character width of the replacement character
    //
    //
    if (AdjustIndex(index) == GLYPH_INDEX_NOT_FOUND)
    {
        index = GetFINF()->alterCharIndex;
    }

    return GetCharWidthsFromIndex(index);
}

bool
ArchiveFontBase::HasGlyph(CharCode c) const
{
    GlyphIndex index = FindGlyphIndex(c);
    if (index != GLYPH_INDEX_NOT_FOUND )
    {
        return ( GLYPH_INDEX_NOT_FOUND != AdjustIndex(index) );
    }
    return false;
}





/* =======================================================================
        ArchiveFontBase protected
   ======================================================================== */

/* ------------------------------------------------------------------------
        Construct/Destruct
   ------------------------------------------------------------------------ */

void
ArchiveFontBase::SetResourceBuffer(
    void*               pUserBuffer,
    FontInformation*    pFontInfo,
    u16*                pAdjustTable )
{
    NN_POINTER_ASSERT( pUserBuffer );
    NN_POINTER_ASSERT( pFontInfo );
    NN_POINTER_ASSERT( pAdjustTable );
    NN_ASSERT( m_pGlyphIndexAdjustArray == NULL );

    ResFontBase::SetResourceBuffer(pUserBuffer, pFontInfo);
    m_pGlyphIndexAdjustArray = pAdjustTable;

    GenTextureNames();
}

void*
ArchiveFontBase::RemoveResourceBuffer()
{
    m_pGlyphIndexAdjustArray = NULL;
    return ResFontBase::RemoveResourceBuffer();
}



/* ------------------------------------------------------------------------
        Glyph index
   ------------------------------------------------------------------------ */

u16
ArchiveFontBase::AdjustIndex(u16 index) const
{
    const FontTextureGlyph& tg  = *GetFINF()->pGlyph;
    const int glyphsPerSheet    = internal::GetCellsInASheet(tg);
    const int sheetNo           = index / glyphsPerSheet;

    const u16 adjustor = m_pGlyphIndexAdjustArray[sheetNo];

    return (adjustor == ADJUST_OFFSET_SHEET_NOT_LOADED) ?
                GLYPH_INDEX_NOT_FOUND: static_cast<u16>(index - adjustor);
}



/* ------------------------------------------------------------------------
        GLGR Block Scanning
   ------------------------------------------------------------------------ */

bool
ArchiveFontBase::IncludeName(
    const char* nameList,
    const char* name
)
{
    const u32 nameLen = std::strlen(name);
    const char* found = nameList - 1;

    // return $nameList =~ m/(^|, *)$name *(,|$)/;
    for (;;)
    {
        found = std::strstr(found+1, name);

        if (found == NULL)
        {
            return false;
        }

        // At the beginning?
        if (found != nameList)
        {
            // Does not have to be at top if the delimiter comes immediately before.
            const char* pos = found - 1;
            while (nameList < pos && *pos == ' ')
            {
                pos--;
            }

            if (*pos != ',')
            {
                continue;
            }
        }

        // Searches for the next delimiter
        const char* delimitor = std::strchr(found, ',');
        u32 foundLen;

        // Find the length of the string found.
        if (delimitor != NULL)
        {
            foundLen = static_cast<u32>(delimitor - found);
        }
        else
        {
            foundLen = std::strlen(found);
        }

        // Found if the length is the same or if only an empty space is added to the end.
        {
            const char* pos = found + nameLen;
            const char* end = found + foundLen;
            while (pos < end && *pos == ' ')
            {
                pos++;
            }
            if (pos == end)
            {
                return true;
            }
        }
    }

    // Does not reach here.
    // return false;
}

bool
ArchiveFontBase::IsValidResource(
    const void* bfnt,
    u32         dataSize
)
{
    NW_FONT_MIN_ASSERT( dataSize, sizeof(ut::BinaryFileHeader) + sizeof(ut::BinaryBlockHeader) + reinterpret_cast<u32>(reinterpret_cast<FontGlyphGroups*>(0)->nameOffsets));
    const ut::BinaryFileHeader* fileHeader = reinterpret_cast<const ut::BinaryFileHeader*>(bfnt);

    // Check the binary file header
    if (! IsValidBinaryFile(fileHeader, BINFILE_SIG_FONTA, FONT_FILE_VERSION, 2))
    {
        // Unsupported format.
        NN_LOG("Invalid font resource.");
        return false;
    }

    const FontGlyphGroupsBlock* ggBlock = reinterpret_cast<const FontGlyphGroupsBlock*>(fileHeader + 1);

    // Check the starting block type.
    if (ggBlock->header.kind != BINBLOCK_SIG_GLGR)
    {
        NN_LOG("invalid block header(%c%c%c%c) expect(%c%c%c%c)",
            (ggBlock->header.kind >> 24) & 0xFF,
            (ggBlock->header.kind >> 16) & 0xFF,
            (ggBlock->header.kind >>  8) & 0xFF,
            (ggBlock->header.kind >>  0) & 0xFF,
            (BINBLOCK_SIG_GLGR >> 24) & 0xFF,
            (BINBLOCK_SIG_GLGR >> 16) & 0xFF,
            (BINBLOCK_SIG_GLGR >>  8) & 0xFF,
            (BINBLOCK_SIG_GLGR >>  0) & 0xFF
        );
        return false;
    }

    // Check whether the starting block is completely read.
    if (sizeof(ut::BinaryFileHeader) + ggBlock->header.size > dataSize)
    {
        NN_LOG("First stream block must include all of GLGP block. "
                            "Please use larger stream buffer.");
        return false;
    }

    return true;
}



/* ------------------------------------------------------------------------
        Build Process
   ------------------------------------------------------------------------ */

ArchiveFontBase::ConstructResult
ArchiveFontBase::RequestData(
    ArchiveFontBase::ConstructContext*      pContext,
    ArchiveFontBase::CachedStreamReader*    pStream,
    u32                                     size
)
{
    pContext->streamOffset += pStream->GetOffset();
    bool bSuccess = pStream->RequestData(pContext, size);
    return bSuccess ? CONSTRUCT_MORE_DATA: CONSTRUCT_ERROR;
}

ArchiveFontBase::ConstructResult
ArchiveFontBase::ConstructOpDispatch(
    ArchiveFontBase::ConstructContext*      pContext,
    ArchiveFontBase::CachedStreamReader*    pStream
)
{
    if (! pContext->HasMoreBlock())
    {
        return CONSTRUCT_FINISH;
    }

    const u32 requireSize = sizeof(ut::BinaryBlockHeader);
    if (pStream->GetRemain() < requireSize)
    {
        return RequestData(pContext, pStream, requireSize);
    }

    pStream->CopyTo(&pContext->header, sizeof(pContext->header));

    switch (pContext->header.kind)
    {
    case BINBLOCK_SIG_GLGR: pContext->op = ConstructContext::ANALYZE_GLGR;    break;
    case BINBLOCK_SIG_FINF: pContext->op = ConstructContext::ANALYZE_FINF;    break;
    case BINBLOCK_SIG_CMAP: pContext->op = ConstructContext::ANALYZE_CMAP;    break;
    case BINBLOCK_SIG_CWDH: pContext->op = ConstructContext::ANALYZE_CWDH;    break;
    case BINBLOCK_SIG_TGLP: pContext->op = ConstructContext::ANALYZE_TGLP;    break;
    default:
        NN_ASSERTMSG(false, "The font has unknown block(%c%c%c%c)",
            (pContext->header.kind >> 24) & 0xFF,
            (pContext->header.kind >> 16) & 0xFF,
            (pContext->header.kind >>  8) & 0xFF,
            (pContext->header.kind >>  0) & 0xFF );
        pContext->op = ConstructContext::FATAL_ERROR;
        return CONSTRUCT_ERROR;
    }

    // Moves current block number forward.
    pContext->NextBlock();

    // Start of the block data must be 4-byte aligned.
    pContext->Align(4);

    return CONSTRUCT_CONTINUE;
}

ArchiveFontBase::ConstructResult
ArchiveFontBase::ConstructOpAnalyzeFileHeader(
    ArchiveFontBase::ConstructContext*      pContext,
    ArchiveFontBase::CachedStreamReader*    pStream
)
{
    const u32 requireSize = sizeof(ut::BinaryFileHeader);

    if (pStream->GetRemain() < requireSize)
    {
        return RequestData(pContext, pStream, requireSize);
    }
    if (pContext->GetRemain() < requireSize)
    {
        // Insufficient output buffer size
        NN_LOG("buffer size too small");
        return CONSTRUCT_ERROR;
    }

    // Temporarily copy
    // Because the buffer position will move with CopyTo(pContext, ...), use CopyTo(void*, ...)
    pStream->CopyTo(pContext->GetCurrentPtr<void*>(), requireSize);
    pContext->op = ConstructContext::DISPATCH;

    return CONSTRUCT_CONTINUE;
}

ArchiveFontBase::ConstructResult
ArchiveFontBase::ConstructOpAnalyzeGLGR(
    ArchiveFontBase::ConstructContext*      pContext,
    ArchiveFontBase::CachedStreamReader*    pStream
)
{
    u8* pTempFileHead;
    u8* pTempGLGRTail;

    // Allocate GLGR block
    {
        const ut::BinaryFileHeader& header = *pContext->GetCurrentPtr<const ut::BinaryFileHeader*>();
        const u32 requireSize = pContext->header.size - sizeof(ut::BinaryBlockHeader);

        if (header.signature != BINFILE_SIG_FONTA)
        {
            // Binary file header is not loaded, or this is not font resource
            NN_LOG("invalid font binary file.");
            return CONSTRUCT_ERROR;
        }

        if (pStream->GetRemain() < requireSize)
        {
            return RequestData(pContext, pStream, requireSize);
        }

        if (pContext->GetRemain() < requireSize)
        {
            // Insufficient output buffer size
            NN_LOG("buffer size too small");
            return CONSTRUCT_ERROR;
        }

        // Temporarily copy GLGR block
        // Because the buffer position will move with CopyTo(pContext, ...), use CopyTo(void*, ...)
        const u32 infoSize = sizeof(ut::BinaryFileHeader) + pContext->header.size;

        pTempFileHead      = pContext->GetCurrentPtr<u8*>();
        u8* blockHeaderPos = pTempFileHead + sizeof(ut::BinaryFileHeader);
        u8* blockBodyPos   = blockHeaderPos + sizeof(ut::BinaryBlockHeader);
        pTempGLGRTail      = blockHeaderPos + pContext->header.size;
        std::memcpy(blockHeaderPos, &pContext->header, sizeof(ut::BinaryBlockHeader));
        pStream->CopyTo(blockBodyPos, requireSize);

        //---- Validity check
        if (! IsValidResource(pTempFileHead, infoSize))
        {
            return CONSTRUCT_ERROR;
        }
    }

    u16 numSheet;
    u16 numGlyphsPerSheet;
    u32 numBlock;
    u16* pTempAdjustArray;
    u32 sizeAdjustTable;

    //---- Build glyph index adjustment table
    {
        FontGlyphGroupsAcs gg(pTempFileHead);

        numSheet          = gg.GetNumSheet();
        numGlyphsPerSheet = gg.GetNumGlyphsPerSheet();
        numBlock          = gg.GetNumBlock();

        sizeAdjustTable = math::RoundUp(gg.GetNumSheet() * sizeof(u16), 4);   // 4-byte alignment
        pTempAdjustArray = reinterpret_cast<u16*>(
            math::RoundDown(reinterpret_cast<uptr>(pTempFileHead + pContext->GetRemain() - sizeAdjustTable), 2));
        if (pContext->GetRemain() < (pTempGLGRTail - pTempFileHead) + sizeAdjustTable)
        {
            // Insufficient output buffer size
            NN_LOG("buffer size too small");
            return CONSTRUCT_ERROR;
        }

        // Use AdjustTable temporarily as a flag array in the following section.
        // Drop all flags.
        for (int sheetNo = 0; sheetNo < gg.GetNumSheet(); ++sheetNo)
        {
            pTempAdjustArray[sheetNo] = 0;
        }

        // Set corresponding flag for all sheets being used.
        for (int setNo = 0; setNo < gg.GetNumSet(); ++setNo)
        {
            const char* setName = gg.GetSetName(setNo);

            if ( IsNullString(pContext->GetGlyphGroups())
              || IncludeName(pContext->GetGlyphGroups(), setName)
            )
            {
                for (int sheetNo = 0; sheetNo < gg.GetNumSheet(); ++sheetNo)
                {
                    if (gg.IsUseSheet(setNo, sheetNo))
                    {
                        pTempAdjustArray[sheetNo] = 1;
                    }
                }
            }
        }

        // Build AdjustTable based on flag.
        u16 adjust = 0;
        for (int sheetNo = 0; sheetNo < gg.GetNumSheet(); ++sheetNo)
        {
            u16 &entry = pTempAdjustArray[sheetNo];

            if (entry == 1)
            {
                entry = adjust;
            }
            else
            {
                entry = 0xFFFF;
                adjust += gg.GetNumGlyphsPerSheet();
            }
        }
    }

    // In the following, the temporarily copied GLGR block is destroyed, and cannot be accessed anymore.

    // Build Context
    {
        u16* pTrueAdjustArray = pContext->GetCurrentPtr<u16*>();
        pContext->Advance(sizeAdjustTable);

        std::copy(pTempAdjustArray, pTempAdjustArray + numSheet,
                  NW_CHECKED_ARRAY_ITERATOR(pTrueAdjustArray, numSheet));

        pContext->SetGLGR( pTrueAdjustArray,
                           numSheet,
                           numGlyphsPerSheet,
                           numBlock );

        pContext->op = ConstructContext::DISPATCH;
    }

    return CONSTRUCT_CONTINUE;
}

ArchiveFontBase::ConstructResult
ArchiveFontBase::ConstructOpAnalyzeFINF(
    ArchiveFontBase::ConstructContext*      pContext,
    ArchiveFontBase::CachedStreamReader*    /* pStream */
)
{
    // Binary block header is not copied
    u32 copySize = pContext->header.size - sizeof(ut::BinaryBlockHeader);

    if (pContext->GetRemain() < copySize)
    {
        // Insufficient output buffer size
        NN_LOG("buffer size too small");
        return CONSTRUCT_ERROR;
    }

    // Start copy task
    pContext->pFINF = pContext->GetCurrentPtr<FontInformation*>();
    pContext->SetupCopyTask(copySize);
    return CONSTRUCT_CONTINUE;
}

ArchiveFontBase::ConstructResult
ArchiveFontBase::ConstructOpAnalyzeCMAP(
    ArchiveFontBase::ConstructContext*      pContext,
    ArchiveFontBase::CachedStreamReader*    /* pStream */
)
{
    // Update the previous CMAP block or FINF block links
    FontCodeMap* pDstCMAP = pContext->GetCurrentPtr<FontCodeMap*>();

    if (pContext->pPrevCMAP != NULL)
    {
        pContext->pPrevCMAP->pNext = pDstCMAP;
    }
    else
    {
        NN_POINTER_ASSERT( pContext->pFINF );
        pContext->pFINF->pMap = pDstCMAP;
    }
    pContext->pPrevCMAP = pDstCMAP;

    // Binary block header is not copied
    u32 copySize = pContext->header.size - sizeof(ut::BinaryBlockHeader);

    if (pContext->GetRemain() < copySize)
    {
        // Insufficient output buffer size
        NN_LOG("buffer size too small");
        return CONSTRUCT_ERROR;
    }

    // Copy task of the data body
    pContext->SetupCopyTask(copySize);
    return CONSTRUCT_CONTINUE;
}

ArchiveFontBase::ConstructResult
ArchiveFontBase::ConstructOpAnalyzeCWDH(
    ArchiveFontBase::ConstructContext*      pContext,
    ArchiveFontBase::CachedStreamReader*    /* pStream */
)
{
    // Update the previous CWDH block or FINF block links
    FontWidth* pDstCWDH = pContext->GetCurrentPtr<FontWidth*>();

    if (pContext->pPrevCWDH != NULL)
    {
        pContext->pPrevCWDH->pNext = pDstCWDH;
    }
    else
    {
        NN_POINTER_ASSERT(pContext->pFINF);
        pContext->pFINF->pWidth = pDstCWDH;
    }
    pContext->pPrevCWDH = pDstCWDH;

    // Binary block header is not copied
    u32 copySize = pContext->header.size - sizeof(ut::BinaryBlockHeader);

    if (pContext->GetRemain() < copySize)
    {
        // Insufficient output buffer size
        NN_LOG("buffer size too small");
        return CONSTRUCT_ERROR;
    }

    // Copy task of the data body
    pContext->SetupCopyTask(copySize);
    return CONSTRUCT_CONTINUE;
}

ArchiveFontBase::ConstructResult
ArchiveFontBase::ConstructOpAnalyzeTGLP(
    ArchiveFontBase::ConstructContext*      pContext,
    ArchiveFontBase::CachedStreamReader*    pStream
)
{
    const u32 requireSize = sizeof(FontTextureGlyph);
    const u32 copySize    = sizeof(FontTextureGlyph);

    if (pStream->GetRemain() < requireSize)
    {
        return RequestData(pContext, pStream, requireSize);
    }
    if (pContext->GetRemain() < copySize)
    {
        // Insufficient output buffer size
        NN_LOG("buffer size too small");
        return CONSTRUCT_ERROR;
    }
    NN_POINTER_ASSERT(pContext->pFINF);

    // Update the FINF block link.
    pContext->pFINF->pGlyph = pContext->GetCurrentPtr<FontTextureGlyph*>();

    // Copy the area outside of sheet.
    pStream->CopyTo(pContext, sizeof(FontTextureGlyph));

    // Check whether sheet is compressed.
    const bool bSheetCompressed = ((pContext->pFINF->pGlyph->sheetFormat & FONT_SHEET_FORMAT_COMPRESSED_FLAG) != 0);
    pContext->pFINF->pGlyph->sheetFormat &= FONT_SHEET_FORMAT_MASK;

    // Update sheet count.
    {
        u16 numLoadSheet = 0;
        for (int i = 0; i < pContext->GetNumSheets(); ++i)
        {
            if (pContext->IsRequireSheet(i))
            {
                numLoadSheet++;
            }
        }
        pContext->pFINF->pGlyph->sheetNum = numLoadSheet;
    }

    // Get the in-stream offset of sheet.
    u32 sheetOffset  = reinterpret_cast<u32>(pContext->pFINF->pGlyph->sheetImage);

    // Alignment is needed for the sheet.
    pContext->Align(GlyphDataAlignment);
    pContext->pFINF->pGlyph->sheetImage = pContext->GetCurrentPtr<u8*>();

    // Branch depending on whether the sheet is compressed.
    const ConstructContext::Operation nextOp
        = bSheetCompressed ? ConstructContext::PREPAIR_EXPAND_SHEET:
                             ConstructContext::PREPAIR_COPY_SHEET;

    // Skip to the location of sheet.
    u32 streamOffset = pContext->streamOffset + pStream->GetOffset();
    u32 skipSize     = sheetOffset - streamOffset;
    pContext->SetupSkipTask(skipSize, nextOp);

    return CONSTRUCT_CONTINUE;
}

ArchiveFontBase::ConstructResult
ArchiveFontBase::ConstructOpPrepairCopySheet(
    ArchiveFontBase::ConstructContext*      pContext,
    ArchiveFontBase::CachedStreamReader*    /* pStream */
)
{
    // Is process completed for all sheets?
    if (! pContext->HasMoreSheet())
    {
        pContext->op = ConstructContext::DISPATCH;
        return CONSTRUCT_CONTINUE;
    }

    const u32 copySize = pContext->pFINF->pGlyph->sheetSize;

    // Does the sheet need to be loaded?
    if (pContext->IsRequireSheet())
    {
        // Load.
        if (pContext->GetRemain() < copySize)
        {
            NN_LOG("buffer size too small");
            return CONSTRUCT_ERROR;
        }

        // Copy
        pContext->SetupCopyTask(copySize, ConstructContext::PREPAIR_COPY_SHEET);
    }
    else
    {
        // Do not load
        pContext->SetupSkipTask(copySize, ConstructContext::PREPAIR_COPY_SHEET);
    }

    // Update the current sheet number
    pContext->NextSheet();

    return CONSTRUCT_CONTINUE;
}

ArchiveFontBase::ConstructResult
ArchiveFontBase::ConstructOpPrepairExpandSheet(
    ArchiveFontBase::ConstructContext*      pContext,
    ArchiveFontBase::CachedStreamReader*    pStream
)
{
    // Is process completed for all sheets?
    if (! pContext->HasMoreSheet())
    {
        pContext->op = ConstructContext::DISPATCH;
        return CONSTRUCT_CONTINUE;
    }

    u32 compSize;
    u32 huffHead;

    const u32 requireSize = sizeof(compSize) + sizeof(huffHead);

    if (pStream->GetRemain() < requireSize)
    {
        return RequestData(pContext, pStream, requireSize);
    }

    // Load compressed data size.
    pStream->CopyTo(&compSize, sizeof(compSize));
    std::memcpy(&huffHead, pStream->Get(0), sizeof(huffHead));
        // Get(0) is not good, but this is to prevent stream position from moving

    const u32 srcSize = compSize;
    const u32 dstSize = nn::cx::GetUncompressedSize(&huffHead);

    // Does the sheet need to be loaded?
    if (pContext->IsRequireSheet())
    {
        // Load.

        if (pContext->GetRemain() < srcSize + sizeof(*pContext->GetUncompContext()))
        {
            NN_LOG("buffer size too small");
            return CONSTRUCT_ERROR;
        }

        // Uncompress data
        nn::cx::InitUncompContextHuffman( pContext->GetUncompContext(),
                                    pContext->GetCurrentPtr<u8*>());
        pContext->SetupExtendTask(srcSize, ConstructContext::PREPAIR_EXPAND_SHEET);

        // The output buffer's position is not updated with ExtendTask, so you need to advance by the decompressed size ahead of time.
        // 
        pContext->Advance(dstSize);
    }
    else
    {
        // Do not load
        pContext->SetupSkipTask(srcSize, ConstructContext::PREPAIR_EXPAND_SHEET);
    }

    // Update the current sheet number
    pContext->NextSheet();

    return CONSTRUCT_CONTINUE;
}

ArchiveFontBase::ConstructResult
ArchiveFontBase::ConstructOpCopy(
    ArchiveFontBase::ConstructContext*      pContext,
    ArchiveFontBase::CachedStreamReader*    pStream
)
{
    const u32 requireSize = pStream->ManagedCopy(pContext);

    if (requireSize == 0)
    {
        // Copy finished
        pContext->EndTask();
    }
    else
    {
        // More data is necessary.
        return RequestData(pContext, pStream, requireSize);
    }

    return CONSTRUCT_CONTINUE;
}

ArchiveFontBase::ConstructResult
ArchiveFontBase::ConstructOpSkip(
    ArchiveFontBase::ConstructContext*      pContext,
    ArchiveFontBase::CachedStreamReader*    pStream
)
{
    u32 skipSize = math::Min(pContext->TaskRemain(), pStream->GetRemain());

    pStream->Advance(skipSize);
    pContext->TaskProceed(skipSize);

    if (pContext->TaskRemain() == 0)
    {
        // Skip finished.
        pContext->EndTask();
    }
    else
    {
        // Skip continued.
        return RequestData(pContext, pStream, pContext->TaskRemain());
    }

    return CONSTRUCT_CONTINUE;
}

ArchiveFontBase::ConstructResult
ArchiveFontBase::ConstructOpExpand(
    ArchiveFontBase::ConstructContext*      pContext,
    ArchiveFontBase::CachedStreamReader*    pStream
)
{
    NN_ASSERT(! pStream->HasBufData());
    u32 extendSize = math::Min(pContext->TaskRemain(), pStream->GetRemain());

    s32 remain =
        nn::cx::ReadUncompHuffman( pContext->GetUncompContext(),
                                      reinterpret_cast<const u8*>(pStream->Get(extendSize)),
                                      extendSize );
    NN_ASSERT(remain >= 0);

    pContext->TaskProceed(extendSize);

    if (pContext->TaskRemain() == 0)
    {
        // Decompression finished.
        NN_ASSERT(remain == 0);
        pContext->EndTask();
    }
    else
    {
        // More data is necessary.
        NN_ASSERT(remain != 0);
        return RequestData(pContext, pStream, pContext->TaskRemain());
    }

    return CONSTRUCT_CONTINUE;
}

ArchiveFontBase::ConstructResult
ArchiveFontBase::ConstructOpFatalError(
    ArchiveFontBase::ConstructContext*      pContext,
    ArchiveFontBase::CachedStreamReader*    /* pStream */
)
{

    pContext->op = ConstructContext::FATAL_ERROR;
    return CONSTRUCT_ERROR;
}





/* =======================================================================
        CachedStreamReader
   ======================================================================== */

void
ArchiveFontBase::CachedStreamReader::Init()
{
    mStreamBegin        = NULL;
    mStreamPos          = NULL;
    mStreamEnd          = NULL;
    mpTempStrmBuf       = NULL;
    mpTempStrmBufPos    = NULL;
    mpTempStrmBufEnd    = NULL;
    mRequireSize        = 0;
}

void
ArchiveFontBase::CachedStreamReader::Attach(
    const void* stream,
    u32         streamSize
)
{
    mStreamBegin    = reinterpret_cast<const u8*>(stream);
    mStreamPos      = mStreamBegin;
    mStreamEnd      = mStreamBegin + streamSize;
}

u32
ArchiveFontBase::CachedStreamReader::GetRemain() const
{
    return static_cast<u32>((mStreamEnd - mStreamPos) + (mpTempStrmBufEnd - mpTempStrmBufPos));
}

void
ArchiveFontBase::CachedStreamReader::Advance(u32 dx)
{
    const u32 sizeTempStrm = mpTempStrmBufEnd - mpTempStrmBufPos;
    if (sizeTempStrm > dx)
    {
        mpTempStrmBufPos += dx;
    }
    else
    {
        mpTempStrmBufPos = mpTempStrmBufEnd;
        mStreamPos += (dx - sizeTempStrm);
    }
}

void
ArchiveFontBase::CachedStreamReader::CopyTo(
    ArchiveFontBase::ConstructContext*  pContext,
    u32                                 size
)
{
    NN_ASSERT(pContext->GetRemain() >= size);

    CopyTo(pContext->GetCurrentPtr<void*>(), size);
    pContext->Advance(size);
}

void
ArchiveFontBase::CachedStreamReader::CopyTo(
    void*   buf,
    u32     size
)
{
    NN_ASSERT(size <= GetRemain());
    const u32 sizeTempStrm = mpTempStrmBufEnd - mpTempStrmBufPos;
    u8* buf8 = reinterpret_cast<u8*>(buf);

    if (sizeTempStrm >= size)
    {
        std::memcpy(buf8, mpTempStrmBufPos, size);
        mpTempStrmBufPos += size;
    }
    else
    {
        const u32 streamCopySize = size - sizeTempStrm;
        std::memcpy(buf8, mpTempStrmBufPos, static_cast<u32>(sizeTempStrm));
        std::memcpy(buf8 + sizeTempStrm, mStreamPos, streamCopySize);

        mpTempStrmBufPos = mpTempStrmBufEnd;
        mStreamPos += streamCopySize;
    }
}

void
ArchiveFontBase::CachedStreamReader::MoveTo(
    void*   buf,
    u32     size
)
{
    NN_ASSERT(size <= GetRemain());
    const u32 sizeTempStrm = mpTempStrmBufEnd - mpTempStrmBufPos;
    u8* buf8 = reinterpret_cast<u8*>(buf);

    if (sizeTempStrm >= size)
    {
        std::memmove(buf8, mpTempStrmBufPos, size);
        mpTempStrmBufPos += size;
    }
    else
    {
        const u32 streamCopySize = size - sizeTempStrm;
        std::memmove(buf8, mpTempStrmBufPos, static_cast<u32>(sizeTempStrm));
        std::memmove(buf8 + sizeTempStrm, mStreamPos, streamCopySize);

        mpTempStrmBufPos = mpTempStrmBufEnd;
        mStreamPos += streamCopySize;
    }
}

u32
ArchiveFontBase::CachedStreamReader::ManagedCopy(ArchiveFontBase::ConstructContext* pContext)
{
    NN_ASSERT(! HasBufData());
    u32 copySize = math::Min(static_cast<u32>(mStreamEnd - mStreamPos), pContext->TaskRemain());

    std::memcpy(pContext->GetCurrentPtr<void*>(), mStreamPos, copySize);
    mStreamPos += copySize;
    pContext->Advance(copySize);
    pContext->TaskProceed(copySize);

    return pContext->TaskRemain();
}

const void*
ArchiveFontBase::CachedStreamReader::Get(u32 size)
{
    NN_ASSERT(size <= GetRemain());
    const void* p = mStreamPos;
    mStreamPos += size;
    return p;
}

bool
ArchiveFontBase::CachedStreamReader::RequestData(
    ArchiveFontBase::ConstructContext*  pContext,
    u32                                 size
)
{
    // Must be in state where data is insufficient for streaming.
    NN_ASSERT(GetRemain() < size);

    if (GetRemain() == 0)
    {
        // Do not create temporary buffer when there is no stream data at all.
        mpTempStrmBuf       = NULL;
        mpTempStrmBufPos    = NULL;
        mpTempStrmBufEnd    = NULL;
        mRequireSize        = 0;
    }
    else
    {
        // Check if there is free space necessary for temporary buffer.
        if (pContext->GetRemain() < size * 2)
        {
            // Error if there is none
            return false;
        }

        u8* tempBuf = pContext->GetCurrentPtr<u8*>() + size;
        u32 curSize = GetRemain();

        // CopyTo cannot be used because temporary buffer may already be created.
        MoveTo(tempBuf, curSize);

        mpTempStrmBuf       = tempBuf;
        mpTempStrmBufPos    = mpTempStrmBuf;
        mpTempStrmBufEnd    = mpTempStrmBuf + curSize;
        mRequireSize        = size;
    }

    return true;
}

u32
ArchiveFontBase::CachedStreamReader::GetOffset() const
{
    return static_cast<u32>((mStreamPos - mStreamBegin) + (mpTempStrmBufPos - mpTempStrmBuf));
}


/* =======================================================================
        ConstructContext
   ======================================================================== */

void
ArchiveFontBase::ConstructContext::Init(
    void*       outBuffer,
    u32         outBufferSize,
    const char* glyphGroups
)
{
    streamReader.Init();
    pGlyphGroups = glyphGroups;
    pGlyphIndexAdjustTable = NULL;

    targetBuffer        = reinterpret_cast<u8*>(outBuffer);
    targetBufferEnd     = targetBuffer + outBufferSize;
    targetBufferCur     = targetBuffer;

    extendContext   = reinterpret_cast<nn::cx::UncompContextHuffman*>(
                        math::RoundDown(
                            reinterpret_cast<uptr>(targetBufferEnd - sizeof(*extendContext)), 4 ));

    op2             = INVALID_OPERATION;
    opSize          = 0;
    resNumBlock     = 1;    // Necessary for dispatching to GLGR block
    currentBlock    = 0;

    sheetIndex      = 0;
    numSheets       = 0;
    glyphsPerSheet  = 0;
    streamOffset    = 0;

    pFINF           = NULL;
    pPrevCWDH       = NULL;
    pPrevCMAP       = NULL;
    op              = INVALID_OPERATION;
}

void
ArchiveFontBase::ConstructContext::SetGLGR(
    u16*        pAdjustTable,
    u16         nSheets,
    u16         nPerSheet,
    u32         numBlock
)
{
    pGlyphIndexAdjustTable  = pAdjustTable;
    numSheets               = nSheets;
    glyphsPerSheet          = nPerSheet;
    resNumBlock             = numBlock;
}

void
ArchiveFontBase::ConstructContext::Finish(
    void**              ppData,
    FontInformation**   ppFontInfo,
    u16**               ppAdjustTable
)
{
    *ppData         = targetBuffer;
    *ppAdjustTable  = pGlyphIndexAdjustTable;
    *ppFontInfo     = pFINF;
}


/* =======================================================================
        FontGlyphGroupsAcs
   ======================================================================== */

ArchiveFontBase::FontGlyphGroupsAcs::FontGlyphGroupsAcs(const void* bfnt)
{
    mpHeader = reinterpret_cast<const ut::BinaryFileHeader*>(bfnt);
    mpData = reinterpret_cast<const FontGlyphGroupsBlock*>(mpHeader + 1);

    mSizeSheetFlags = sizeof(u32) * math::DivUp(mpData->body.numSheet, (u16)32);
    mSizeCWDHFlags  = sizeof(u32) * math::DivUp(mpData->body.numCWDH, (u16)32);
    mSizeCMAPFlags  = sizeof(u32) * math::DivUp(mpData->body.numCMAP, (u16)32);

    // Since FontGlyphGroups is only partially defined in the header, you need to calculate with offsets to get an accurate size.
    // 
    const u32 sizePreBlock      = sizeof(ut::BinaryFileHeader) + sizeof(ut::BinaryBlockHeader);
    const u32 sizeStatic        = reinterpret_cast<u32>(reinterpret_cast<FontGlyphGroups*>(0)->nameOffsets);
    const u32 sizeNameOffsets   = sizeof(u16) * mpData->body.numSet;
    const u32 sizeSizeSheets    = sizeof(u32) * mpData->body.numSheet;
    const u32 sizeSizeCWDH      = sizeof(u32) * mpData->body.numCWDH;
    const u32 sizeSizeCMAP      = sizeof(u32) * mpData->body.numCMAP;
    const u32 sizeUseSheets     = mSizeSheetFlags * mpData->body.numSet;
    const u32 sizeUseCWDH       = mSizeCWDHFlags * mpData->body.numSet;
//            const u32 sizeUseCMAP       = mSizeCMAPFlags * numSet;

    const u32 offsetStatic      = math::RoundUp(sizePreBlock, 2);
    const u32 offsetNameOffsets = math::RoundUp(offsetStatic     + sizeStatic,        2);
    const u32 offsetSizeSheets  = math::RoundUp(offsetNameOffsets+ sizeNameOffsets,   4);
    const u32 offsetSizeCWDH    = math::RoundUp(offsetSizeSheets + sizeSizeSheets,    4);
    const u32 offsetSizeCMAP    = math::RoundUp(offsetSizeCWDH   + sizeSizeCWDH,      4);
    const u32 offsetUseSheets   = math::RoundUp(offsetSizeCMAP   + sizeSizeCMAP,      4);
    const u32 offsetUseCWDH     = math::RoundUp(offsetUseSheets  + sizeUseSheets,     4);
    const u32 offsetUseCMAP     = math::RoundUp(offsetUseCWDH    + sizeUseCWDH,       4);

    mpSizeSheetsArray   = reinterpret_cast<const u32*>(ut::AddOffsetToPtr(mpHeader, offsetSizeSheets));
    mpSizeCWDHArray     = reinterpret_cast<const u32*>(ut::AddOffsetToPtr(mpHeader, offsetSizeCWDH));
    mpSizeCMAPArray     = reinterpret_cast<const u32*>(ut::AddOffsetToPtr(mpHeader, offsetSizeCMAP));
    mpUseSheetArray     = reinterpret_cast<const u32*>(ut::AddOffsetToPtr(mpHeader, offsetUseSheets));
    mpUseCWDHArray      = reinterpret_cast<const u32*>(ut::AddOffsetToPtr(mpHeader, offsetUseCWDH));
    mpUseCMAPArray      = reinterpret_cast<const u32*>(ut::AddOffsetToPtr(mpHeader, offsetUseCMAP));
}

}   // namespace font
}   // namespace nw
