/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#define NW_DEBUG_CHECK_MEMORY_LEAK

#include <nn/os.h>
#include <nn/fs.h>
#include <nn/applet.h>

#include <nw/types.h>
#include <nw/demo.h>
#include <nw/dev.h>
#include <nw/gfx.h>
#include <nw/ut.h>

namespace
{

//----------------------------------------
// Memory-related

// Allocator to allocate device memory.
nw::demo::DemoAllocator s_DeviceAllocator;
nw::demo::DemoAllocator s_ParticleAllocator;

//----------------------------------------
// Filename definitions.
const wchar_t* SKY_SPHERE_FILE_NAME  = NW_DEMO_FILE_PATH(L"SkySphere.bcmdl");

//----------------------------------------
// Rendering-related
const int RENDER_TARGET_COUNT = 1;
typedef nw::ut::FixedSizeArray<nw::gfx::IRenderTarget*, RENDER_TARGET_COUNT> RenderTargetArray;

RenderTargetArray s_RenderTargets;
nw::demo::SceneSystem*  s_SceneSystem = NULL;
nw::demo::RenderSystem* s_RenderSystem = NULL;
nw::gfx::ParticleContext* s_ParticleContext = NULL;

nw::demo::GraphicsDrawing  s_GraphicsDrawing;

//----------------------------------------
// Resource-related
nw::demo::ResourceArray s_Resources;

//----------------------------------------
// Scene-related
nw::gfx::SceneNode* s_SceneRoot = NULL;
s32 s_FrameCount = 0;
nw::gfx::Camera* s_BaseCamera = NULL;
nw::gfx::Camera* s_LeftCamera = NULL;
nw::gfx::Camera* s_RightCamera = NULL;
const f32 s_fNearPlane = 0.1f;

const s32 s_BaseCameraIndex = 0;

//----------------------------------------
// Particle-related
nw::gfx::ParticleSceneUpdater* s_ParticleSceneUpdater = NULL;

/*
  
 */
void
InitializeGraphics()
{
    // Perform stereo settings to renderDescription.
    nw::demo::RenderSystem::Description renderDescription;

    renderDescription.reusableCommandBufferSize = 0x100000;
    renderDescription.reusableCommandRequestCount      = 512;
    renderDescription.upperScreenMode = nw::demo::UPPER_SCREEN_MODE_STEREO;

    s_RenderSystem = nw::demo::RenderSystem::Create(&s_DeviceAllocator, renderDescription);

    s_GraphicsDrawing.SetScreenSize(
        renderDescription.lowerScreenDescription.width,
        renderDescription.lowerScreenDescription.height
    );

    nw::demo::Utility::InitializeGraphicsDrawing(&s_DeviceAllocator, s_GraphicsDrawing);

    s_RenderTargets.push_back(
        nw::demo::Utility::CreateUpperScreenBuffer(&s_DeviceAllocator, renderDescription)
    );
    NW_ASSERT(!s_RenderTargets.empty());
    s_RenderSystem->GetRenderContext()->SetRenderTarget(s_RenderTargets.front());

    // Standard settings to sceneDescription are performed with the constructor.
    nw::demo::SceneSystem::Description sceneDescription;
    s_SceneSystem = nw::demo::SceneSystem::Create(&s_DeviceAllocator, sceneDescription);

    s_ParticleContext = nw::gfx::ParticleContext::Builder()
        .MaxEmission(1000)
        .Create(&s_DeviceAllocator);

    s_ParticleSceneUpdater = nw::gfx::ParticleSceneUpdater::Builder()
        .Create(&s_DeviceAllocator);

    // Sets the furthest background model for the demo to the rendering system.
    // With the gfx demo, glClear, etc. is not used, and a fill is applied with the background.
    s_RenderSystem->LoadSkyModel(SKY_SPHERE_FILE_NAME);

    NW_GL_ASSERT();
}

/*
  
 */
void
TerminateGraphics()
{
    nw::gfx::SafeDestroy(s_LeftCamera);

    nw::gfx::SafeDestroy(s_RightCamera);

    nw::gfx::SafeDestroy(s_ParticleSceneUpdater);

    nw::gfx::SafeDestroy(s_ParticleContext);

    nw::gfx::SafeDestroy(s_SceneSystem);

    nw::gfx::SafeDestroyAll(s_RenderTargets);

    s_GraphicsDrawing.Finalize();

    nw::gfx::SafeDestroy(s_RenderSystem);

    NW_GL_ASSERT();
}

/*
  
 */
void
BuildRootNodes()
{
    NW_ASSERT(s_SceneRoot == NULL);
    s_SceneRoot = nw::gfx::TransformNode::DynamicBuilder()
        .IsFixedSizeMemory(false)
        .Create(&s_DeviceAllocator);
    NW_NULL_ASSERT(s_SceneRoot);
}

/*
  
 */
void
BuildCameras()
{
    nw::demo::Utility::CreateStereoCameras(
        &s_BaseCamera,
        &s_LeftCamera,
        &s_RightCamera,
        &s_DeviceAllocator,
        nw::math::VEC3(0.0f, 13.0f, 70.f),
        nw::math::VEC3(0.0f, 13.0f, 0.0f),
        s_fNearPlane
    );

    s_SceneRoot->AttachChild(s_BaseCamera);
    s_SceneSystem->GetCameraController()->Register(s_BaseCamera);
}


/*
  
 */
void
InitializeScenes()
{
    BuildRootNodes();

    BuildCameras();

    // Traverse and initialize the scene tree.
    s_SceneSystem->InitializeScene(s_SceneRoot);
    s_SceneSystem->UpdateScene();

    // Set the camera.
    nw::gfx::SceneEnvironment& sceneEnvironment = s_RenderSystem->GetSceneEnvironment();
    sceneEnvironment.SetCamera(s_BaseCameraIndex, s_BaseCamera);
    nw::demo::Utility::SetCameraAspectRatio(s_BaseCamera, s_RenderTargets[0]);

    NW_GL_ASSERT();

    s_FrameCount = 0;
}

/*
  
 */
void
TerminateScenes()
{
    nw::gfx::SafeDestroyBranch(s_SceneRoot);
    nw::demo::SafeCleanupResources(s_Resources);

    NW_GL_ASSERT();

    s_Resources.clear();
}

/*
  
 */
void
UpdateScene()
{
    NW_ASSERT(0 < s_RenderTargets.size());

    s_SceneSystem->GetCameraController()->Update();

    s_SceneSystem->UpdateScene();

    s_BaseCamera->UpdateCameraMatrix();

    NW_NULL_ASSERT(s_ParticleSceneUpdater);
    s_ParticleSceneUpdater->UpdateNode(
        s_SceneSystem->GetSceneContext(),
        s_ParticleContext);

    s_RenderSystem->CalcStereoCamera(s_LeftCamera, s_RightCamera, s_BaseCamera, s_fNearPlane + 5.0f);

    s_FrameCount++;
}

/*
  
 */
void
ReportDemo()
{
    NW_PROFILE("ReportDemo");

    // This load is exempted from the load display.
    s_RenderSystem->SuspendLoadMeter();

    nw::demo::DebugUtility::CalcLoadMeter(s_RenderSystem);

    s_GraphicsDrawing.BeginDrawingShape();

    nw::demo::DebugUtility::DrawLoadMeter(
        s_RenderSystem,
        &s_GraphicsDrawing
    );

    s_GraphicsDrawing.EndDrawingShape();

    s_GraphicsDrawing.BeginDrawingString();

    nw::demo::DebugUtility::DrawLoadMeterText(
        s_RenderSystem,
        &s_GraphicsDrawing
    );

    s_GraphicsDrawing.EndDrawingString();

    s_RenderSystem->ResumeLoadMeter();
}




//----------------------------------------
// Variable unique to the demo

// Particle effect class
nw::demo::ParticleEffect* s_ParticleEffect = NULL;

// Number of placements
#define DEMO_FIRE_DRAW_NUM  (5)

// Particle handle class
nw::demo::ParticleHandle* s_PaticleHandle5M5E[DEMO_FIRE_DRAW_NUM];
nw::demo::ParticleHandle* s_PaticleHandle1M5E = NULL;
nw::demo::ParticleHandle* s_PaticleHandle1M1E = NULL;

// Counter
u32 s_Count = 0;

// Animation state
enum
{
    DEMO_ASSORTMENT_STATE_5MODEL_5EMITTER   = 0,
    DEMO_ASSORTMENT_STATE_1MODEL_5EMITTER   = 1,
    DEMO_ASSORTMENT_STATE_1MODEL_1EMITTER   = 2,
    DEMO_ASSORTMENT_STATE_MAX               = 3
};
u32 s_AnimationState = DEMO_ASSORTMENT_STATE_5MODEL_5EMITTER;

// Effect file to be loaded
const wchar_t* EFFECT_RESOURCE_FILE  = NW_DEMO_FILE_PATH(L"fire_particle_mem5.bcptl");

// Effect shader file to be loaded
const wchar_t* SHADER_RESOURCE_FILE_NAME = NW_DEMO_FILE_PATH(L"nwgfx_ParticleDefaultShader.bcsdr");


/*
  
 */
void
InitializeParticleNodeCombinationDemo()
{
    // Using combinations of the following particle models and emitters, this demo displays five flame effects on the screen.
    // 
    // 5 ParticleModel - 5 ParticleEmitter(5M5E)
    //   The handling of placement, etc., is easy, but there are drawbacks in terms of memory and processing speed.
    //
    // 1 ParticleModel - 5 ParticleEmitter(5M5E)
    //   The degree of freedom is lower, but the processing speed and amount of memory used is better.
    //
    // 1 ParticleModel - 1 ParticleEmitter(1M1E)
    //   The degree of freedom is lower, but the processing speed and amount of memory used is better.
    //   Also, an emission interval can only be applied to the data of a single frame with this method.
    //   For memory, 1M1E is better, for speed, 1M5E is better.

    s_AnimationState = DEMO_ASSORTMENT_STATE_5MODEL_5EMITTER;
    s_Count = 0;
    s_ParticleEffect = NULL;

    // Loads shader binaries.
    nw::demo::ParticleEffect::InitializeShaderBinary(SHADER_RESOURCE_FILE_NAME, &s_DeviceAllocator);

    // Generate the particle effect class.
    s_ParticleEffect = nw::demo::ParticleEffect::Create(&s_DeviceAllocator, &s_DeviceAllocator, false, s_ParticleContext);

    // Load the effect data and setup.
    nw::demo::ResourceSet* resourceSet = nw::demo::Utility::LoadResources(s_Resources, EFFECT_RESOURCE_FILE, &s_DeviceAllocator);

    // Sets up the resource set.
    s_ParticleEffect->Setup(resourceSet->resource, false);

    // Set resources to ParticleEffect.
    s_ParticleEffect->Register(resourceSet->resource);
    s_ParticleEffect->AddPool(6);

    // Prepare for 5M5E.
    for (s32 i = 0; i < DEMO_FIRE_DRAW_NUM; ++i)
    {
        s_PaticleHandle5M5E[i] = s_ParticleEffect->LeaseInstance();
        if (s_PaticleHandle5M5E[i])
        {
            s_PaticleHandle5M5E[i]->SetTranslate(10.0f * (i - 2), 0.f, 0.f);
        }

        // Add to the scene.
        s_SceneRoot->AttachChild(s_PaticleHandle5M5E[i]);
    }

    // Prepare for 1M5E.
    s_PaticleHandle1M5E = s_ParticleEffect->CreateMultiEmitterParticleHandle( DEMO_FIRE_DRAW_NUM );
    if (s_PaticleHandle1M5E)
    {
        for (s32 i = 0; i < s_PaticleHandle1M5E->GetParticleEmitterSize(); ++i)
        {
            s_PaticleHandle1M5E->GetParticleEmitter(i)->Transform().SetTranslate(10.0f * (i - 2), 0.f, 0.f);
        }
    }

    // Prepare for 1M1E.
    s_PaticleHandle1M1E = s_ParticleEffect->LeaseInstance();
    s_PaticleHandle1M1E->SetTranslate(0.f, 0.f, 0.f);
    s_SceneRoot->AttachChild(s_PaticleHandle1M1E);

    // Updates the scene.
    s_SceneSystem->InitializeScene(s_SceneRoot);
    s_SceneSystem->UpdateScene();
}

/*
  
 */
void
UpdateParticleNodeCombinationDemo()
{
    ++s_Count;

    // Change the combination pattern every 600 frames.
    if ( s_Count%600 == 0 )
    {
        s_Count = 0;

        switch(s_AnimationState)
        {
        case DEMO_ASSORTMENT_STATE_5MODEL_5EMITTER:
            {
                // Detach the 5M5E node.
                for (u32 i = 0; i < DEMO_FIRE_DRAW_NUM; ++i)
                {
                    s_SceneRoot->DetachChild(s_PaticleHandle5M5E[i]);
                }

                // Attach the 1M5E node.
                s_SceneRoot->AttachChild(s_PaticleHandle1M5E);

                // Updates the scene.
                s_SceneSystem->InitializeScene(s_SceneRoot);
                s_SceneSystem->UpdateScene();

                s_AnimationState = DEMO_ASSORTMENT_STATE_1MODEL_5EMITTER;
            }
            break;

        case DEMO_ASSORTMENT_STATE_1MODEL_5EMITTER:
            {
                // Detach the 1M5E node.
                s_SceneRoot->DetachChild(s_PaticleHandle1M5E);

                // Attach the 1M1E node.
                s_SceneRoot->AttachChild(s_PaticleHandle1M1E);

                // Updates the scene.
                s_SceneSystem->InitializeScene(s_SceneRoot);
                s_SceneSystem->UpdateScene();

                s_AnimationState = DEMO_ASSORTMENT_STATE_1MODEL_1EMITTER;
            }
            break;

        case DEMO_ASSORTMENT_STATE_1MODEL_1EMITTER:
            {
                // Detach the 1M1E node.
                s_SceneRoot->DetachChild(s_PaticleHandle1M1E);

                // Attach the 5M5E node.
                for (u32 i = 0; i < DEMO_FIRE_DRAW_NUM; ++i)
                {
                    s_SceneRoot->AttachChild(s_PaticleHandle5M5E[i]);
                }

                // Updates the scene.
                s_SceneSystem->InitializeScene(s_SceneRoot);
                s_SceneSystem->UpdateScene();

                s_AnimationState = DEMO_ASSORTMENT_STATE_5MODEL_5EMITTER;
            }
            break;

        default:
            break;
        }
    }

    // For 1M1E, you need perform changes and emission for every frame emission position.
    // Reset occurs after emission, so even for data that specifies a value other than 1 frame for an emission interval, note that emission will occur at every frame.
    // 
    if( s_AnimationState == DEMO_ASSORTMENT_STATE_1MODEL_1EMITTER )
    {
        nw::gfx::ParticleEmitter* emitter = s_PaticleHandle1M1E->GetParticleEmitter(0);
        if (emitter)
        {
            for (s32 i = 0; i < DEMO_FIRE_DRAW_NUM; ++i)
            {
                // Set the emission position.
                emitter->WorldMatrix().SetupTranslate(nn::math::VEC3(nw::math::VEC3(10.0f * (i - 2), 0.f, 0.f)));

                // Perform emission and then reset.
                emitter->UpdateParticleFrame();
                emitter->Emission(s_ParticleContext);
                emitter->Reset();
            }
        }
    }
}

/*
  
 */
void
ReportParticleNodeCombinationDemo()
{
    s_GraphicsDrawing.DrawString(10, 10, "Counter              : %d\n", (600 - s_Count) );

    switch(s_AnimationState)
    {
    case DEMO_ASSORTMENT_STATE_5MODEL_5EMITTER:
        {
            s_GraphicsDrawing.DrawString(10, 22, "Particle Model   Num : 5\n" );
            s_GraphicsDrawing.DrawString(10, 34, "Particle Emitter Num : 5\n" );
        }
        break;

    case DEMO_ASSORTMENT_STATE_1MODEL_5EMITTER:
        {
            s_GraphicsDrawing.DrawString(10, 22, "Particle Model   Num : 1\n" );
            s_GraphicsDrawing.DrawString(10, 34, "Particle Emitter Num : 5\n" );
        }
        break;

    case DEMO_ASSORTMENT_STATE_1MODEL_1EMITTER:
        {
            s_GraphicsDrawing.DrawString(10, 22, "Particle Model   Num : 1\n" );
            s_GraphicsDrawing.DrawString(10, 34, "Particle Emitter Num : 1\n" );
        }
        break;
    }
}

/*
  
 */
void
TerminateParticleNodeCombinationDemo()
{
//    s_ParticleEffect->DestroyParticleHandle(s_PaticleHandle1M1E);
    nw::demo::ParticleNode* prevTarget = s_ParticleEffect->GetActiveEffect(0);
    while (prevTarget != NULL)
    {
        s_ParticleEffect->ReleaseInstance(prevTarget);
        prevTarget = s_ParticleEffect->GetActiveEffect(0);
    }

    // Generated with CreateMultiEmitterParticleHandle, so the handle must be destroyed.
    s_ParticleEffect->DestroyParticleHandle(s_PaticleHandle1M5E);

    s_ParticleEffect->FreePool();

    nw::ut::SafeDestroy(s_ParticleEffect);

    // Destroy the loaded particle shader.
    nw::demo::ParticleEffect::FinalizeShaderBinary();
}




/*
  
 */
void
DemoScene()
{
    NW_ASSERT(!s_RenderTargets.empty());

    nw::gfx::RenderContext* renderContext = s_RenderSystem->GetRenderContext();

    InitializeScenes();

    // The initialization process for ParticleNodeCombinationDemo.
    InitializeParticleNodeCombinationDemo();

    bool isContinuing = true;

    while ( isContinuing )
    {
        nw::demo::DebugUtility::AdvanceAutoTestFrame();

        nw::demo::PadFactory::GetPad()->Update();


        // The fixed term process for ParticleNodeCombinationDemo.
        UpdateParticleNodeCombinationDemo();

        UpdateScene();

        renderContext->SetActiveCamera(s_BaseCameraIndex);
        s_RenderSystem->SubmitView(s_SceneSystem);

        s_RenderSystem->SetRenderTarget(s_RenderTargets[0]);
        s_RenderSystem->RenderStereoScene(s_LeftCamera, s_RightCamera);

        s_RenderSystem->ClearBySkyModel(s_BaseCamera);
        ReportDemo();

        // Display the report for ParticleNodeCombinationDemo.
        ReportParticleNodeCombinationDemo();

        s_RenderSystem->TransferBuffer(nw::demo::LOWER_SCREEN);

        s_RenderSystem->PresentBuffer(nw::demo::UPPER_SCREEN | nw::demo::LOWER_SCREEN | nw::demo::EXTENSION_SCREEN);

        renderContext->ResetState();

        if (nw::demo::Utility::IsTerminating())
        {
            isContinuing = false;
        }
    }

    // The finalization process for ParticleNodeCombinationDemo.
    TerminateParticleNodeCombinationDemo();

    TerminateScenes();
}

} // Namespace

/*
  
 */
void
nnMain()
{
    nn::applet::Enable();

    nw::demo::InitializeGraphicsSystem(&s_DeviceAllocator);
    nw::demo::InitializeDemoAllocator(&s_ParticleAllocator, nw::demo::DEMO_PARTICLE_MEMORY_SIZE);

    nw::demo::PadFactory::Initialize(&s_DeviceAllocator);

    NW_DEMO_TEST_LOOP(&s_DeviceAllocator, &s_ParticleAllocator, &s_RenderSystem)
    {
        InitializeGraphics();

        DemoScene();

        TerminateGraphics();
    }

    nw::demo::PadFactory::Finalize();

    nw::demo::FinalizeDemoAllocator(&s_ParticleAllocator);

    nw::demo::FinalizeGraphicsSystem();
}
