/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#define NW_DEBUG_CHECK_MEMORY_LEAK

#include <nn/os.h>
#include <nn/fs.h>
#include <nn/applet.h>

#include <nw/types.h>
#include <nw/demo.h>
#include <nw/dev.h>
#include <nw/gfx.h>
#include <nw/ut.h>

namespace
{

//----------------------------------------
// Memory-related

// Allocator to allocate device memory.
nw::demo::DemoAllocator s_DeviceAllocator;
nw::demo::DemoAllocator s_ParticleAllocator;

//----------------------------------------
// Filename definitions.
const wchar_t* SKY_SPHERE_FILE_NAME  = NW_DEMO_FILE_PATH(L"SkySphere.bcmdl");

//----------------------------------------
// Rendering-related
const int RENDER_TARGET_COUNT = 1;
typedef nw::ut::FixedSizeArray<nw::gfx::IRenderTarget*, RENDER_TARGET_COUNT> RenderTargetArray;

RenderTargetArray s_RenderTargets;
nw::demo::SceneSystem*  s_SceneSystem = NULL;
nw::demo::RenderSystem* s_RenderSystem = NULL;
nw::gfx::ParticleContext* s_ParticleContext = NULL;

nw::demo::GraphicsDrawing  s_GraphicsDrawing;



//----------------------------------------
// Scene-related
nw::gfx::SceneNode* s_SceneRoot = NULL;
s32 s_FrameCount = 0;
nw::gfx::Camera* s_BaseCamera = NULL;
nw::gfx::Camera* s_LeftCamera = NULL;
nw::gfx::Camera* s_RightCamera = NULL;
const f32 s_fNearPlane = 0.1f;

const s32 s_BaseCameraIndex = 0;

//----------------------------------------
// Particle-related
nw::gfx::ParticleSceneUpdater* s_ParticleSceneUpdater = NULL;

/*
  
 */
void
InitializeGraphics()
{
    // Perform stereo settings to renderDescription.
    nw::demo::RenderSystem::Description renderDescription;

    renderDescription.reusableCommandBufferSize = 0x100000;
    renderDescription.reusableCommandRequestCount      = 512;
    renderDescription.upperScreenMode = nw::demo::UPPER_SCREEN_MODE_STEREO;

    s_RenderSystem = nw::demo::RenderSystem::Create(&s_DeviceAllocator, renderDescription);

    s_GraphicsDrawing.SetScreenSize(
        renderDescription.lowerScreenDescription.width,
        renderDescription.lowerScreenDescription.height
    );

    nw::demo::Utility::InitializeGraphicsDrawing(&s_DeviceAllocator, s_GraphicsDrawing);

    s_RenderTargets.push_back(
        nw::demo::Utility::CreateUpperScreenBuffer(&s_DeviceAllocator, renderDescription)
    );
    NW_ASSERT(!s_RenderTargets.empty());
    s_RenderSystem->GetRenderContext()->SetRenderTarget(s_RenderTargets.front());

    // Standard settings to sceneDescription are performed with the constructor.
    nw::demo::SceneSystem::Description sceneDescription;
    s_SceneSystem = nw::demo::SceneSystem::Create(&s_DeviceAllocator, sceneDescription);

    s_ParticleContext = nw::gfx::ParticleContext::Builder()
        .Create(&s_DeviceAllocator);

    s_ParticleSceneUpdater = nw::gfx::ParticleSceneUpdater::Builder()
        .Create(&s_DeviceAllocator);

    // Sets the furthest background model for the demo to the rendering system.
    // With the gfx demo, glClear, etc. is not used, and a fill is applied with the background.
    s_RenderSystem->LoadSkyModel(SKY_SPHERE_FILE_NAME);

    NW_GL_ASSERT();
}

/*
  
 */
void
TerminateGraphics()
{
    nw::gfx::SafeDestroy(s_LeftCamera);

    nw::gfx::SafeDestroy(s_RightCamera);

    nw::gfx::SafeDestroy(s_ParticleSceneUpdater);

    nw::gfx::SafeDestroy(s_ParticleContext);

    nw::gfx::SafeDestroy(s_SceneSystem);

    nw::gfx::SafeDestroyAll(s_RenderTargets);

    s_GraphicsDrawing.Finalize();

    nw::gfx::SafeDestroy(s_RenderSystem);

    NW_GL_ASSERT();
}

/*
  
 */
void
BuildRootNodes()
{
    NW_ASSERT(s_SceneRoot == NULL);
    s_SceneRoot = nw::gfx::TransformNode::DynamicBuilder()
        .IsFixedSizeMemory(false)
        .Create(&s_DeviceAllocator);
    NW_NULL_ASSERT(s_SceneRoot);
}

/*
  
 */
void
BuildCameras()
{
    nw::demo::Utility::CreateStereoCameras(
        &s_BaseCamera,
        &s_LeftCamera,
        &s_RightCamera,
        &s_DeviceAllocator,
        nw::math::VEC3(0.0f, 15.0f, 80.f),
        nw::math::VEC3(0.0f, 15.0f, 0.0f),
        s_fNearPlane
    );

    s_SceneRoot->AttachChild(s_BaseCamera);
    s_SceneSystem->GetCameraController()->Register(s_BaseCamera);
}


/*
  
 */
void
InitializeScenes()
{
    BuildRootNodes();

    BuildCameras();

    // Traverse and initialize the scene tree.
    s_SceneSystem->InitializeScene(s_SceneRoot);
    s_SceneSystem->UpdateScene();

    // Set the camera.
    nw::gfx::SceneEnvironment& sceneEnvironment = s_RenderSystem->GetSceneEnvironment();
    sceneEnvironment.SetCamera(s_BaseCameraIndex, s_BaseCamera);
    nw::demo::Utility::SetCameraAspectRatio(s_BaseCamera, s_RenderTargets[0]);

    NW_GL_ASSERT();

    s_FrameCount = 0;
}

/*
  
 */
void
TerminateScenes()
{
    nw::gfx::SafeDestroyBranch(s_SceneRoot);

    NW_GL_ASSERT();
}

/*
  
 */
void
UpdateScene()
{
    NW_ASSERT(0 < s_RenderTargets.size());

    s_SceneSystem->GetCameraController()->Update();

    s_SceneSystem->UpdateScene();

    s_BaseCamera->UpdateCameraMatrix();

    NW_NULL_ASSERT(s_ParticleSceneUpdater);
    s_ParticleSceneUpdater->UpdateNode(
        s_SceneSystem->GetSceneContext(),
        s_ParticleContext);

    s_RenderSystem->CalcStereoCamera(s_LeftCamera, s_RightCamera, s_BaseCamera, s_fNearPlane + 5.0f);

    s_FrameCount++;
}

/*
  
 */
void
ReportDemo()
{
    NW_PROFILE("ReportDemo");

    // This load is exempted from the load display.
    s_RenderSystem->SuspendLoadMeter();

    nw::demo::DebugUtility::CalcLoadMeter(s_RenderSystem);

    s_GraphicsDrawing.BeginDrawingShape();

    nw::demo::DebugUtility::DrawLoadMeter(
        s_RenderSystem,
        &s_GraphicsDrawing
    );

    s_GraphicsDrawing.EndDrawingShape();

    s_GraphicsDrawing.BeginDrawingString();

    nw::demo::DebugUtility::DrawLoadMeterText(
        s_RenderSystem,
        &s_GraphicsDrawing
    );

    s_GraphicsDrawing.EndDrawingString();

    s_RenderSystem->ResumeLoadMeter();
}




//----------------------------------------
// Variable unique to the demo

// Particle effect class
nw::demo::ParticleEffect* s_ParticleEffect = NULL;

// Particle node class
nw::demo::ParticleNode* s_PaticleNode = NULL;

//----------------------------------------
// Resource-related
nw::demo::ResourceArray s_Resources;

// Effect file to be loaded
const wchar_t* EFFECT_RESOURCE_FILE  = NW_DEMO_FILE_PATH(L"missile_particle.bcptl");

// Effect shader file to be loaded
const wchar_t* SHADER_RESOURCE_FILE_NAME = NW_DEMO_FILE_PATH(L"nwgfx_ParticleDefaultShader.bcsdr");

// Emitting angle
f32 s_ShootAngle = 0.0f;

/*
  
 */
void
InitializeParticleMissileDemo()
{
    // This demo creates a missile with a particle using speed inheritance from a parent to a child.
    // Within the data there are 3 ParticleSets, including 1 parent and 2 children, and when the second ParticleSet is generated speed inheritance from the parent is performed.
    // 

    // ParticleSet0: The missile itself
    // ParticleSet1: Particle that tracks the missile (inheriting the parent's speed)
    // ParticleSet2: Particle played back after the missile vanishes

    s_ShootAngle = 0.0f;

    // Loads shader binaries.
    nw::demo::ParticleEffect::InitializeShaderBinary(SHADER_RESOURCE_FILE_NAME, &s_DeviceAllocator);

    // Generate the particle effect class.
    s_ParticleEffect = nw::demo::ParticleEffect::Create(&s_DeviceAllocator, &s_DeviceAllocator, false, s_ParticleContext);

    // Load the effect data and setup.
    nw::demo::ResourceSet* resourceSet = nw::demo::Utility::LoadResources(s_Resources, EFFECT_RESOURCE_FILE, &s_DeviceAllocator);

    // Sets up the resources.
    s_ParticleEffect->Setup(resourceSet->resource, false);

    // Set resources to ParticleEffect.
    s_ParticleEffect->Register(resourceSet->resource);
    s_ParticleEffect->AddPool(1);

    // Lease the particle node class' instance.
    s_PaticleNode = s_ParticleEffect->LeaseInstance();

    // Add to the scene.
    s_SceneRoot->AttachChild(s_PaticleNode);

    // Updates the scene.
    s_SceneSystem->InitializeScene(s_SceneRoot);
    s_SceneSystem->UpdateScene();
}

/*
  
 */
void
UpdateParticleMissileDemo()
{
    // The name of the ParticleSet to remove at any timing.
    const char* DESTROY_MISSILE_NAME0  = "Missile_Particle";        // The missile itself
    const char* DESTROY_MISSILE_NAME1  = "Missile_Flash_Particle";  // The light of the missile's nozzle

    // Adjust the emission angle with the pad.
    if (nw::demo::PadFactory::GetPad()->IsButtonPress(nw::demo::Pad::BUTTON_RIGHT))
    {
        s_ShootAngle -= nw::math::F_PI * 0.01f;
        if (s_ShootAngle < -nw::math::F_PI/2.f)
        {
            s_ShootAngle = -nw::math::F_PI/2.f;
        }
    }

    if (nw::demo::PadFactory::GetPad()->IsButtonPress(nw::demo::Pad::BUTTON_LEFT))
    {
        s_ShootAngle += nw::math::F_PI * 0.01f;
        if (s_ShootAngle > nw::math::F_PI/2.f)
        {
            s_ShootAngle = nw::math::F_PI/2.f;
        }
    }

    // Adjust the missile's emission angle.
    for (u32 i = 0; i < s_PaticleNode->GetParticleEmitterSize(); ++i)
    {
        nw::gfx::ParticleEmitter* emitter = s_PaticleNode->GetParticleEmitter(i);
        emitter->Transform().SetRotateXYZ( 0.0f, 0.0f, s_ShootAngle );
    }


    // Emit with Pad A.
    // You need to set the runtime work memory for only the number of missile that can be emitted at the same time on the resource-side.
    if (nw::demo::PadFactory::GetPad()->IsButtonDown(nw::demo::Pad::BUTTON_A))
    {
        s_PaticleNode->Emission(s_ParticleContext);
    }

    // Remove all the missiles with Pad B.
    if (nw::demo::PadFactory::GetPad()->IsButtonDown(nw::demo::Pad::BUTTON_B))
    {
        nw::gfx::ParticleSet* particleSet0 = s_PaticleNode->GetParticleSet(DESTROY_MISSILE_NAME0);
        nw::gfx::ParticleSet* particleSet1 = s_PaticleNode->GetParticleSet(DESTROY_MISSILE_NAME1);

        nw::gfx::ParticleCollection* collection0= particleSet0->GetParticleCollection();
        nw::gfx::ParticleCollection* collection1= particleSet1->GetParticleCollection();

        collection0->KillParticles();
        collection1->KillParticles();
    }


    // Remove the emitted missiles (particles) at optional timing (collisions, etc.).
    {
        nw::gfx::ParticleSet* particleSet0 = s_PaticleNode->GetParticleSet(DESTROY_MISSILE_NAME0);
        nw::gfx::ParticleSet* particleSet1 = s_PaticleNode->GetParticleSet(DESTROY_MISSILE_NAME1);

        nw::gfx::ParticleCollection* collection0= particleSet0->GetParticleCollection();
        nw::gfx::ParticleCollection* collection1= particleSet1->GetParticleCollection();

        // Get the valid number of particles for missiles.
        const int count0 = collection0->GetCount();
        if (count0 != 0)
        {
            // The index to the valid particles
            u16* activeIndex =
                (u16*)collection0->GetStreamPtr(
                nw::gfx::PARTICLEUSAGE_ACTIVEINDEX,nw::gfx::PARTICLE_BUFFER_FRONT);

            // Particle Position
            nw::math::VEC3* translate =
                (nw::math::VEC3*)collection0->GetStreamPtr(
                nw::gfx::PARTICLEUSAGE_TRANSLATE, nw::gfx::PARTICLE_BUFFER_FRONT);

            const bool isAscendingOrder = particleSet0->IsAscendingOrder();
            const int startIndex = (isAscendingOrder) ? 0 : collection0->GetCapacity() - 1;
            const int incrIndex = (isAscendingOrder) ? 1 : -1;
            activeIndex += startIndex;

            for (int i = 0; i < count0; ++i)
            {
                int index = *activeIndex;
                activeIndex += incrIndex;

                // Missiles are removed at a distance of 25.f from the emission position.
                if ( translate[index].Length() > 25.f )
                {
                    collection0->KillParticle(index);
                }
            }
        }


        // The same process occurs for missile nozzle light.
        const int count1 = collection0->GetCount();
        if (count1 != 0)
        {
            // The index to the valid particles
            u16* activeIndex =
                (u16*)collection1->GetStreamPtr(
                nw::gfx::PARTICLEUSAGE_ACTIVEINDEX,nw::gfx::PARTICLE_BUFFER_FRONT);

            // Particle Position
            nw::math::VEC3* translate =
                (nw::math::VEC3*)collection1->GetStreamPtr(
                nw::gfx::PARTICLEUSAGE_TRANSLATE, nw::gfx::PARTICLE_BUFFER_FRONT);

            const bool isAscendingOrder = particleSet1->IsAscendingOrder();
            const int startIndex = (isAscendingOrder) ? 0 : collection1->GetCapacity() - 1;
            const int incrIndex = (isAscendingOrder) ? 1 : -1;
            activeIndex += startIndex;

            for (int i = 0; i < count1; ++i)
            {
                int index = *activeIndex;
                activeIndex += incrIndex;

                // Missiles are removed at a distance of 25.f from the emission position.
                if ( translate[index].Length() > 25.f )
                {
                    collection1->KillParticle(index);
                }
            }
        }
    }
}

/*
  
 */
void
ReportParticleMissileDemo()
{
    s32 degree = s_ShootAngle * 114.6f;
    s_GraphicsDrawing.DrawString(10, 10, "Angle    :%d\n", degree);
    s_GraphicsDrawing.DrawString(10, 22, "Button A :Shoot\n", degree);
    s_GraphicsDrawing.DrawString(10, 36, "Button B :Destruction\n", degree);
    s_GraphicsDrawing.DrawString(10, 48, "Pad L<->R:Angle Adjust\n", degree);
}

/*
  
 */
void
TerminateParticleMissileDemo()
{
    nw::demo::ParticleNode* prevTarget = s_ParticleEffect->GetActiveEffect(0);
    while (prevTarget != NULL)
    {
        s_ParticleEffect->ReleaseInstance(prevTarget);
        prevTarget = s_ParticleEffect->GetActiveEffect(0);
    }

    s_ParticleEffect->FreePool();

    nw::ut::SafeDestroy(s_ParticleEffect);

    // Destroy the loaded particle shader.
    nw::demo::ParticleEffect::FinalizeShaderBinary();

    nw::demo::SafeCleanupResources(s_Resources);
    s_Resources.clear();
}




/*
  
 */
void
DemoScene()
{
    NW_ASSERT(!s_RenderTargets.empty());

    nw::gfx::RenderContext* renderContext = s_RenderSystem->GetRenderContext();

    InitializeScenes();

    // The initialization process for ParticleMissileDemo.
    InitializeParticleMissileDemo();

    bool isContinuing = true;

    while ( isContinuing )
    {
        nw::demo::DebugUtility::AdvanceAutoTestFrame();

        nw::demo::PadFactory::GetPad()->Update();


        // The fixed term process for ParticleMissileDemo.
        UpdateParticleMissileDemo();

        UpdateScene();

        renderContext->SetActiveCamera(s_BaseCameraIndex);
        s_RenderSystem->SubmitView(s_SceneSystem);

        s_RenderSystem->SetRenderTarget(s_RenderTargets[0]);
        s_RenderSystem->RenderStereoScene(s_LeftCamera, s_RightCamera);

        s_RenderSystem->ClearBySkyModel(s_BaseCamera);
        ReportDemo();

        // Display the report for ParticleMissileDemo.
        ReportParticleMissileDemo();

        s_RenderSystem->TransferBuffer(nw::demo::LOWER_SCREEN);

        s_RenderSystem->PresentBuffer(nw::demo::UPPER_SCREEN | nw::demo::LOWER_SCREEN | nw::demo::EXTENSION_SCREEN);

        renderContext->ResetState();

        if (nw::demo::Utility::IsTerminating())
        {
            isContinuing = false;
        }
    }

    // The finalization process for ParticleMissileDemo.
    TerminateParticleMissileDemo();

    TerminateScenes();
}

} // Namespace

/*
  
 */
void
nnMain()
{
    nn::applet::Enable();

    nw::demo::InitializeGraphicsSystem(&s_DeviceAllocator);
    nw::demo::InitializeDemoAllocator(&s_ParticleAllocator, nw::demo::DEMO_PARTICLE_MEMORY_SIZE);

    nw::demo::PadFactory::Initialize(&s_DeviceAllocator);

    NW_DEMO_TEST_LOOP(&s_DeviceAllocator, &s_ParticleAllocator, &s_RenderSystem)
    {
        InitializeGraphics();

        DemoScene();

        TerminateGraphics();
    }

    nw::demo::PadFactory::Finalize();

    nw::demo::FinalizeDemoAllocator(&s_ParticleAllocator);

    nw::demo::FinalizeGraphicsSystem();
}
