/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#define NW_DEBUG_CHECK_MEMORY_LEAK

#include <nn/os.h>
#include <nn/fs.h>
#include <nn/applet.h>

#include <nw/types.h>
#include <nw/demo.h>
#include <nw/dev.h>
#include <nw/gfx.h>
#include <nw/ut.h>

namespace
{

//----------------------------------------
// Memory-related

// Allocator to allocate device memory.
nw::demo::DemoAllocator s_DeviceAllocator;
nw::demo::DemoAllocator s_ParticleAllocator;

//----------------------------------------
// Filename definitions.
const wchar_t* SKY_SPHERE_FILE_NAME  = NW_DEMO_FILE_PATH(L"SkySphere.bcmdl");

//----------------------------------------
// Rendering-related
const int RENDER_TARGET_COUNT = 1;
typedef nw::ut::FixedSizeArray<nw::gfx::IRenderTarget*, RENDER_TARGET_COUNT> RenderTargetArray;

RenderTargetArray s_RenderTargets;
nw::demo::SceneSystem*  s_SceneSystem = NULL;
nw::demo::RenderSystem* s_RenderSystem = NULL;
nw::gfx::ParticleContext* s_ParticleContext = NULL;

nw::demo::GraphicsDrawing  s_GraphicsDrawing;

//----------------------------------------
// Resource-related
nw::demo::ResourceArray s_Resources;

//----------------------------------------
// Scene-related
nw::gfx::SceneNode* s_SceneRoot = NULL;
s32 s_FrameCount = 0;
nw::gfx::Camera* s_BaseCamera = NULL;
nw::gfx::Camera* s_LeftCamera = NULL;
nw::gfx::Camera* s_RightCamera = NULL;
const f32 s_fNearPlane = 0.1f;

const s32 s_BaseCameraIndex = 0;

//----------------------------------------
// Particle-related
nw::gfx::ParticleSceneUpdater* s_ParticleSceneUpdater = NULL;

/*
  
 */
void
InitializeGraphics()
{
    // Perform stereo settings to renderDescription.
    nw::demo::RenderSystem::Description renderDescription;

    renderDescription.reusableCommandBufferSize = 0x100000;
    renderDescription.reusableCommandRequestCount      = 512;
    renderDescription.upperScreenMode = nw::demo::UPPER_SCREEN_MODE_STEREO;

    s_RenderSystem = nw::demo::RenderSystem::Create(&s_DeviceAllocator, renderDescription);

    s_GraphicsDrawing.SetScreenSize(
        renderDescription.lowerScreenDescription.width,
        renderDescription.lowerScreenDescription.height
    );

    nw::demo::Utility::InitializeGraphicsDrawing(&s_DeviceAllocator, s_GraphicsDrawing);

    s_RenderTargets.push_back(
        nw::demo::Utility::CreateUpperScreenBuffer(&s_DeviceAllocator, renderDescription)
    );
    NW_ASSERT(!s_RenderTargets.empty());
    s_RenderSystem->GetRenderContext()->SetRenderTarget(s_RenderTargets.front());

    // Standard settings to sceneDescription are performed with the constructor.
    nw::demo::SceneSystem::Description sceneDescription;
    s_SceneSystem = nw::demo::SceneSystem::Create(&s_DeviceAllocator, sceneDescription);

    s_ParticleContext = nw::gfx::ParticleContext::Builder()
        .Create(&s_DeviceAllocator);

    s_ParticleSceneUpdater = nw::gfx::ParticleSceneUpdater::Builder()
        .Create(&s_DeviceAllocator);

    // Sets the furthest background model for the demo to the rendering system.
    // With the gfx demo, glClear, etc. is not used, and a fill is applied with the background.
    s_RenderSystem->LoadSkyModel(SKY_SPHERE_FILE_NAME);

    NW_GL_ASSERT();
}

/*
  
 */
void
TerminateGraphics()
{
    nw::gfx::SafeDestroy(s_LeftCamera);

    nw::gfx::SafeDestroy(s_RightCamera);

    nw::gfx::SafeDestroy(s_ParticleSceneUpdater);

    nw::gfx::SafeDestroy(s_ParticleContext);

    nw::gfx::SafeDestroy(s_SceneSystem);

    nw::gfx::SafeDestroyAll(s_RenderTargets);

    s_GraphicsDrawing.Finalize();

    nw::gfx::SafeDestroy(s_RenderSystem);

    NW_GL_ASSERT();
}

/*
  
 */
void
BuildRootNodes()
{
    NW_ASSERT(s_SceneRoot == NULL);
    s_SceneRoot = nw::gfx::TransformNode::DynamicBuilder()
        .IsFixedSizeMemory(false)
        .Create(&s_DeviceAllocator);
    NW_NULL_ASSERT(s_SceneRoot);
}

/*
  
 */
void
BuildCameras()
{
    nw::demo::Utility::CreateStereoCameras(
        &s_BaseCamera,
        &s_LeftCamera,
        &s_RightCamera,
        &s_DeviceAllocator,
        nw::math::VEC3(0.0f, 13.0f, 70.f),
        nw::math::VEC3(0.0f, 13.0f, 0.0f),
        s_fNearPlane
    );

    s_SceneRoot->AttachChild(s_BaseCamera);
    s_SceneSystem->GetCameraController()->Register(s_BaseCamera);
}


/*
  
 */
void
InitializeScenes()
{
    BuildRootNodes();

    BuildCameras();

    // Traverse and initialize the scene tree.
    s_SceneSystem->InitializeScene(s_SceneRoot);
    s_SceneSystem->UpdateScene();

    // Set the camera.
    nw::gfx::SceneEnvironment& sceneEnvironment = s_RenderSystem->GetSceneEnvironment();
    sceneEnvironment.SetCamera(s_BaseCameraIndex, s_BaseCamera);
    nw::demo::Utility::SetCameraAspectRatio(s_BaseCamera, s_RenderTargets[0]);

    NW_GL_ASSERT();

    s_FrameCount = 0;
}

/*
  
 */
void
TerminateScenes()
{
    nw::gfx::SafeDestroyBranch(s_SceneRoot);
    nw::demo::SafeCleanupResources(s_Resources);

    NW_GL_ASSERT();

    s_Resources.clear();
}

/*
  
 */
void
UpdateScene()
{
    NW_ASSERT(0 < s_RenderTargets.size());

    s_SceneSystem->GetCameraController()->Update();

    s_SceneSystem->UpdateScene();

    s_BaseCamera->UpdateCameraMatrix();

    NW_NULL_ASSERT(s_ParticleSceneUpdater);
    s_ParticleSceneUpdater->UpdateNode(
        s_SceneSystem->GetSceneContext(),
        s_ParticleContext);

    s_RenderSystem->CalcStereoCamera(s_LeftCamera, s_RightCamera, s_BaseCamera, s_fNearPlane + 5.0f);

    s_FrameCount++;
}

/*
  
 */
void
ReportDemo()
{
    NW_PROFILE("ReportDemo");

    // This load is exempted from the load display.
    s_RenderSystem->SuspendLoadMeter();

    nw::demo::DebugUtility::CalcLoadMeter(s_RenderSystem);

    s_GraphicsDrawing.BeginDrawingShape();

    nw::demo::DebugUtility::DrawLoadMeter(
        s_RenderSystem,
        &s_GraphicsDrawing
    );

    s_GraphicsDrawing.EndDrawingShape();

    s_GraphicsDrawing.BeginDrawingString();

    nw::demo::DebugUtility::DrawLoadMeterText(
        s_RenderSystem,
        &s_GraphicsDrawing
    );

    s_GraphicsDrawing.EndDrawingString();

    s_RenderSystem->ResumeLoadMeter();
}




//----------------------------------------
// Variable unique to the demo

// Particle effect class
nw::demo::ParticleEffect* s_ParticleEffect = NULL;

// Particle node class
nw::demo::ParticleNode* s_PaticleNode = NULL;

// Counter
u32 s_Count = 0;

// Animation state
enum
{
    DEMO_STATE_SCALING_PARTICLE_MODEL     = 0,
    DEMO_STATE_SCALING_PARTICLE_EMITTER   = 1,
    DEMO_STATE_TRANSLATE_PARTICLE_EMITTER = 2,
    DEMO_STATE_SCALING_PARTICLE           = 3,
    DEMO_STATE_MAX                        = 4
};
u32 s_AnimationState = DEMO_STATE_SCALING_PARTICLE_MODEL;

// Effect file to be loaded
const wchar_t* EFFECT_RESOURCE_FILE  = NW_DEMO_FILE_PATH(L"rain_follow_on.bcptl");

// Effect shader file to be loaded
const wchar_t* SHADER_RESOURCE_FILE_NAME = NW_DEMO_FILE_PATH(L"nwgfx_ParticleDefaultShader.bcsdr");

/*
  
 */
void
InitializeParticleScaleDemo()
{
    // This demo confirms the effect when scale is applied to particle models, particle emitters and particles, respectively.
    // 

    // Loads shader binaries.
    nw::demo::ParticleEffect::InitializeShaderBinary(SHADER_RESOURCE_FILE_NAME, &s_DeviceAllocator);

    // Generate the particle effect class.
    s_ParticleEffect = nw::demo::ParticleEffect::Create(&s_DeviceAllocator, &s_DeviceAllocator, false, s_ParticleContext);

    // Load the effect data and setup.
    nw::demo::ResourceSet* resourceSet = nw::demo::Utility::LoadResources(s_Resources, EFFECT_RESOURCE_FILE, &s_DeviceAllocator);

    // Sets up the resources.
    s_ParticleEffect->Setup(resourceSet->resource, true);

    // Set resources to ParticleEffect.
    s_ParticleEffect->Register(resourceSet->resource);
    s_ParticleEffect->AddPool(1);

    // Lease the particle node class' instance.
    s_PaticleNode = s_ParticleEffect->LeaseInstance();

    // Add to the scene.
    s_SceneRoot->AttachChild(s_PaticleNode);
    s_SceneSystem->InitializeScene(s_SceneRoot);
    s_SceneSystem->UpdateScene();
}

/*
  
 */
void
UpdateParticleScaleDemo()
{
    ++s_Count;

    // Change the target for applying the scale in 360 frame units.
    if (s_Count%360 == 0)
    {
        s_AnimationState++;

        // Restore the emitter scale.
        if (s_AnimationState == DEMO_STATE_SCALING_PARTICLE)
        {
            for (u32 i = 0; i < s_PaticleNode->GetParticleEmitterSize(); ++i)
            {
                nn::math::MTX34 transformMatrix;
                transformMatrix.SetupIdentity();
                transformMatrix.SetupScale(nn::math::VEC3(nw::math::VEC3(1.f, 1.f, 1.f)));

                //  Sets a transform based on a resource's transform information.
                s_PaticleNode->GetParticleEmitter(i)->SetResourceBasedTransform( transformMatrix );
            }
        }

        if (s_AnimationState == DEMO_STATE_MAX)
        {
            s_AnimationState = DEMO_STATE_SCALING_PARTICLE_MODEL;
        }
    }

    // Apply the scale animation to the ParticleModel.
    // The scale is applied to the effect overall.
    if (s_AnimationState == DEMO_STATE_SCALING_PARTICLE_MODEL)
    {
        f32 fsin = 0.5f * nn::math::SinDeg(s_Count);

        for (u32 i = 0; i < s_PaticleNode->GetParticleModelSize(); ++i)
        {
            s_PaticleNode->GetParticleModel(i)->Transform().SetScale( 1.f + fsin,  1.f + fsin, 1.f + fsin );
        }
    }

    // Apply the scale to the ParticleEmitter.
    // The size of each particle remains the same, and the scale is applied to the emitting source.
    if (s_AnimationState == DEMO_STATE_SCALING_PARTICLE_EMITTER)
    {
        f32 fsin = 10.0f * nn::math::SinDeg(s_Count);

        // Scale is applied such that the cloud expands in the XZ plane.
        for (u32 i = 0; i < s_PaticleNode->GetParticleEmitterSize(); ++i)
        {
            nn::math::MTX34 transformMatrix;
            transformMatrix.SetupIdentity();
            transformMatrix.SetupScale(nn::math::VEC3(nw::math::VEC3(3.f, 1.f, 3.f)));

            //  Sets a transform based on a resource's transform information.
            s_PaticleNode->GetParticleEmitter(i)->SetResourceBasedTransform( transformMatrix );
        }
    }

    // Apply the position to the ParticleEmitter.
    if (s_AnimationState == DEMO_STATE_TRANSLATE_PARTICLE_EMITTER)
    {
        f32 fsin = 20.0f * nn::math::SinDeg(s_Count);

        // The emitter is moved to the left and right.
        for (u32 i = 0; i < s_PaticleNode->GetParticleEmitterSize(); ++i)
        {
            nn::math::MTX34 transformMatrix;
            transformMatrix.SetupIdentity();
            transformMatrix.SetupTranslate(nn::math::VEC3(nw::math::VEC3(fsin, 0.f, 0.f)));

            //  Set the position information based on the transform information the resources have.
            s_PaticleNode->GetParticleEmitter(i)->SetResourceBasedTransform( transformMatrix );
        }
    }

    // Scale animation for the particles
    // The scale is applied to the size of each particle.
    // Within the SetParticleScaleOffset method, manipulations occur to the SetScaleOffset and the method within ParticleSet maintained by ParticleModel.
    // 
    if (s_AnimationState == DEMO_STATE_SCALING_PARTICLE)
    {
        f32 fsin = 0.5f * nn::math::SinDeg(s_Count);
        s_PaticleNode->SetParticleScaleOffset(1.f + fsin);
    }
}

/*
  
 */
void
ReportParticleScaleDemo()
{
    if(s_AnimationState== DEMO_STATE_SCALING_PARTICLE_MODEL)
    {
        s_GraphicsDrawing.DrawString(10, 10, "Particle Model Scaling");
    }

    if(s_AnimationState== DEMO_STATE_SCALING_PARTICLE_EMITTER)
    {
        s_GraphicsDrawing.DrawString(10, 10, "Particle Emitter Scaling");
    }

    if(s_AnimationState== DEMO_STATE_TRANSLATE_PARTICLE_EMITTER)
    {
        s_GraphicsDrawing.DrawString(10, 10, "Particle Emitter Translate");
    }

    if(s_AnimationState== DEMO_STATE_SCALING_PARTICLE)
    {
        s_GraphicsDrawing.DrawString(10, 10, "Particle Scaling");
    }
}

/*
  
 */
void
TerminateParticleScaleDemo()
{
    nw::demo::ParticleNode* prevTarget = s_ParticleEffect->GetActiveEffect(0);
    while (prevTarget != NULL)
    {
        s_ParticleEffect->ReleaseInstance(prevTarget);
        prevTarget = s_ParticleEffect->GetActiveEffect(0);
    }

    s_ParticleEffect->FreePool();

    nw::ut::SafeDestroy(s_ParticleEffect);

    // Destroy the loaded particle shader.
    nw::demo::ParticleEffect::FinalizeShaderBinary();
}




/*
  
 */
void
DemoScene()
{
    NW_ASSERT(!s_RenderTargets.empty());

    nw::gfx::RenderContext* renderContext = s_RenderSystem->GetRenderContext();

    InitializeScenes();

    // The initialization process for ParticleScaleDemo.
    InitializeParticleScaleDemo();

    nw::demo::DebugUtility::PostInitializeScenes();

    bool isContinuing = true;

    while ( isContinuing )
    {
        nw::demo::DebugUtility::AdvanceAutoTestFrame();

        nw::demo::PadFactory::GetPad()->Update();


        // The fixed term process for ParticleScaleDemo.
        UpdateParticleScaleDemo();

        UpdateScene();

        renderContext->SetActiveCamera(s_BaseCameraIndex);
        s_RenderSystem->SubmitView(s_SceneSystem);

        s_RenderSystem->SetRenderTarget(s_RenderTargets[0]);
        s_RenderSystem->RenderStereoScene(s_LeftCamera, s_RightCamera);

        s_RenderSystem->ClearBySkyModel(s_BaseCamera);
        ReportDemo();

        // Display the report for ParticleScaleDemo.
        ReportParticleScaleDemo();

        s_RenderSystem->TransferBuffer(nw::demo::LOWER_SCREEN);

        s_RenderSystem->PresentBuffer(nw::demo::UPPER_SCREEN | nw::demo::LOWER_SCREEN | nw::demo::EXTENSION_SCREEN);

        renderContext->ResetState();

        if (nw::demo::Utility::IsTerminating())
        {
            isContinuing = false;
        }
    }

    nw::demo::DebugUtility::PreTerminateScenes();

    // The finalization process for ParticleScaleDemo.
    TerminateParticleScaleDemo();

    TerminateScenes();
}

} // Namespace

/*
  
 */
void
nnMain()
{
    nn::applet::Enable();

    nw::demo::InitializeGraphicsSystem(&s_DeviceAllocator);
    nw::demo::InitializeDemoAllocator(&s_ParticleAllocator, nw::demo::DEMO_PARTICLE_MEMORY_SIZE);

    nw::demo::PadFactory::Initialize(&s_DeviceAllocator);

    NW_DEMO_TEST_LOOP(&s_DeviceAllocator, &s_ParticleAllocator, &s_RenderSystem)
    {
        InitializeGraphics();

        DemoScene();

        TerminateGraphics();
    }

    nw::demo::PadFactory::Finalize();

    nw::demo::FinalizeDemoAllocator(&s_ParticleAllocator);

    nw::demo::FinalizeGraphicsSystem();
}
