/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#define NW_DEBUG_CHECK_MEMORY_LEAK

#include <nn/os.h>
#include <nn/fs.h>
#include <nn/applet.h>

#include <nw/types.h>

#include <nw/demo.h>
#include <nw/dev.h>
#include <nw/gfx.h>
#include <nw/ut.h>
#include <nw/anim.h>

namespace
{

//----------------------------------------
// Shadow-related code
// UserFlag represents the shadow state.
enum UserFlags
{
    SHADOW_CASTER_SHIFT = 0,
    SHADOW_RECEIVER_SHIFT = 1,

    SHADOW_CASTER = 0x1 << SHADOW_CASTER_SHIFT,
    SHADOW_RECEIVER = 0x1 << SHADOW_RECEIVER_SHIFT
};

// The function object that adds only the Shadow's Caster to the render elements.
class IsShadowCasterModelFunctor : public nw::gfx::ISceneUpdater::IsVisibleModelFunctor
{
public:
    virtual bool IsVisible(const nw::gfx::Model* model)
    {
        return nw::ut::CheckFlag(model->GetUserParameter<u32>(), SHADOW_CASTER);
    }
};

IsShadowCasterModelFunctor s_ShadowCasterModel;

// The render command that performs the shadow settings.
class StartShadowRenderCommand : public nw::gfx::RenderCommand
{
    NW_DISALLOW_COPY_AND_ASSIGN(StartShadowRenderCommand);

public:
    StartShadowRenderCommand(
        nw::demo::CommandListSwapper* commandListSwapper,
        nw::gfx::IRenderTarget* renderTarget,
        s32 shadowCameraIndex,
        nw::gfx::Material* shadowMaterial)
        : m_CommandListSwapper(commandListSwapper),
          m_RenderTarget(renderTarget),
          m_ShadowCameraIndex(shadowCameraIndex),
          m_ShadowMaterial(shadowMaterial),
          m_CommandSize(0),
          m_ProfilerPoint(0)
    {}
    virtual ~StartShadowRenderCommand() {}

    virtual void Invoke(nw::gfx::RenderContext* renderContext)
    {
        m_CommandSize = m_CommandListSwapper->GetCommandBufferSize();
        m_ProfilerPoint = m_CommandListSwapper->AddGpuProfilingStartPoint(true);

        // Switch to the off-screen buffer.
        renderContext->SetRenderTarget(m_RenderTarget);
        renderContext->ClearBuffer(GL_COLOR_BUFFER_BIT, nw::ut::FloatColor(1.0f, 1.0f, 1.0f, 1.0f), 1.0f);

        // Set the material and camera for the shadow.
        renderContext->SetActiveCamera(m_ShadowCameraIndex);
        renderContext->SetMaterial(m_ShadowMaterial);
        nw::gfx::Camera* shadowCamera = renderContext->GetActiveCamera();
        renderContext->SetCameraMatrix(shadowCamera);
        renderContext->ActivateContext();

        // With 1Pass, using RENDERMODE_IGNORE_MATERIAL, settings for subsequent materials are not performed.
        renderContext->SetRenderMode(nw::gfx::RenderContext::RENDERMODE_IGNORE_MATERIAL | nw::gfx::RenderContext::RENDERMODE_IGNORE_SHADER);
    }

    s32 GetCommandSize() const { return this->m_CommandSize; }
    s32 GetProfilerPoint() const { return this->m_ProfilerPoint; }

private:
    nw::demo::CommandListSwapper* m_CommandListSwapper;
    nw::gfx::IRenderTarget* m_RenderTarget;
    s32 m_ShadowCameraIndex;
    nw::gfx::Material* m_ShadowMaterial;
    s32 m_CommandSize;
    s32 m_ProfilerPoint;
};

// The render command that performs the shadow post-processing.
class EndShadowRenderCommand : public nw::gfx::RenderCommand
{
    NW_DISALLOW_COPY_AND_ASSIGN(EndShadowRenderCommand);

public:
    EndShadowRenderCommand(nw::demo::CommandListSwapper* commandListSwapper, StartShadowRenderCommand* startCommand)
        : m_CommandListSwapper(commandListSwapper),
          m_StartCommand(startCommand),
          m_CommandSize(0)
    {}
    virtual ~EndShadowRenderCommand() {}

    virtual void Invoke(nw::gfx::RenderContext* renderContext)
    {
        // Restore render mode to default, so that material settings can be performed normally.
        renderContext->SetRenderMode(nw::gfx::RenderContext::RENDERMODE_DEFAULT);

        // Perform ResetState to destroy the render cache.
        renderContext->ResetState();

        m_CommandSize = m_CommandListSwapper->GetCommandBufferSize();
        int profilerPoint = m_StartCommand->GetProfilerPoint();
        m_CommandListSwapper->SetGpuProfilingEndPoint(profilerPoint);
    }

    s32 GetCommandSize() const { return this->m_CommandSize; }

private:
    nw::demo::CommandListSwapper* m_CommandListSwapper;
    StartShadowRenderCommand* m_StartCommand;
    s32 m_CommandSize;
};

// The render command that performs the render start settings.
class StartRenderCommand : public nw::gfx::RenderCommand
{
    NW_DISALLOW_COPY_AND_ASSIGN(StartRenderCommand);

public:
    StartRenderCommand(nw::demo::RenderSystem* renderSystem, nw::gfx::IRenderTarget* renderTarget, s32 baseCameraIndex)
        : m_RenderSystem(renderSystem),
          m_RenderTarget(renderTarget),
          m_BaseCameraIndex(baseCameraIndex),
          m_CommandSize(0)
    {}

    virtual ~StartRenderCommand() {}

    virtual void Invoke(nw::gfx::RenderContext* renderContext)
    {
        nw::demo::CommandListSwapper* commandListSwapper = m_RenderSystem->GetCommandListSwapper();

        // Switch the render target to the on-screen buffer
        renderContext->SetRenderTarget(m_RenderTarget);
        // Set the calculation camera for lights, environment maps, etc.
        renderContext->SetActiveCamera(m_BaseCameraIndex);

        // Begin creation of command for stereo display. The created command will be rendered multiple times, so it will be re-used.
        commandListSwapper->StartCommandSave();

        m_CommandSize = commandListSwapper->GetCommandBufferSize();

        m_RenderSystem->ClearBySkyModel(renderContext->GetActiveCamera());
    }

    s32 GetCommandSize() const { return this->m_CommandSize; }

private:
    nw::demo::RenderSystem* m_RenderSystem;
    nw::gfx::IRenderTarget* m_RenderTarget;
    s32 m_BaseCameraIndex;
    s32 m_CommandSize;
};

// The render command that performs the render end settings.
class EndRenderCommand : public nw::gfx::RenderCommand
{
    NW_DISALLOW_COPY_AND_ASSIGN(EndRenderCommand);

public:
    EndRenderCommand(
        nw::demo::RenderSystem* renderSystem,
        nw::gfx::Camera* leftCamera,
        nw::gfx::Camera* rightCamera,
        StartRenderCommand* startCommand)
        : m_RenderSystem(renderSystem),
          m_LeftCamera(leftCamera),
          m_RightCamera(rightCamera),
          m_StartCommand(startCommand)
    {}
    virtual ~EndRenderCommand() {}

    virtual void Invoke(nw::gfx::RenderContext* renderContext)
    {
        int commandSize = m_RenderSystem->GetCommandListSwapper()->GetCommandBufferSize();
        m_RenderSystem->AddLoadMeterCommandSize(commandSize - m_StartCommand->GetCommandSize());

        nw::demo::CommandListSwapper* commandListSwapper = m_RenderSystem->GetCommandListSwapper();

        // End creation of command for stereo display.
        commandListSwapper->EndCommandSave();

        // Start GPU process time measurement
        int profilerLeft = commandListSwapper->AddGpuProfilingStartPoint(true);

        // Re-use the saved command for left eye rendering.
        renderContext->SetCameraMatrix(m_LeftCamera);
        commandListSwapper->ReuseCommand(false);

        // End GPU process time measurement
        commandListSwapper->SetGpuProfilingEndPoint(profilerLeft);

        m_RenderSystem->TransferBuffer(nw::demo::UPPER_SCREEN);

        // Start GPU process time measurement
        int profilerRight = commandListSwapper->AddGpuProfilingStartPoint(true);

        // Re-use the saved command for right eye rendering.
        renderContext->SetCameraMatrix(m_RightCamera);
        commandListSwapper->ReuseCommand(false);

        // End GPU process time measurement
        commandListSwapper->SetGpuProfilingEndPoint(profilerRight);

        m_RenderSystem->TransferBuffer(nw::demo::EXTENSION_SCREEN);
    }
private:
    nw::demo::RenderSystem* m_RenderSystem;
    nw::gfx::Camera* m_LeftCamera;
    nw::gfx::Camera* m_RightCamera;
    StartRenderCommand* m_StartCommand;
};

enum CameraKind
{
    PERSPECTIVE_CAMERA,
    FRUSTUM_CAMERA,
    ORTHO_CAMERA
};

nw::gfx::Camera* s_ShadowCamera = NULL;
// The type of camera used for shadows.
CameraKind s_CameraKind = PERSPECTIVE_CAMERA;
// The parameters for the camera for shadows.
const f32 s_OrthoCameraHeight = 32.0f;
const f32 s_ShadowNear = 1.0f;
const f32 s_ShadowFar = 500.0f;

StartShadowRenderCommand* s_StartShadowRenderCommand = NULL;
EndShadowRenderCommand* s_EndShadowRenderCommand = NULL;
StartRenderCommand* s_StartRenderCommand = NULL;
EndRenderCommand* s_EndRenderCommand = NULL;

// The model for shadow material and shader settings.
// This model is used for material and shader settings.
// It is not actually rendered.
nw::gfx::Model* s_ShadowDummyModel = NULL;
// The texture for shadows.
nw::gfx::ResShadowTexture s_ShadowTexture;
// Shadow texture size.
const s32 s_ShadowTextureSize = 512;
// The zBias applied to the shadow.
const f32 s_ShadowBias = 1.3f;
// Shadow strength (thickness).
const f32 s_ShadowIntensity = 0.2f;

//----------------------------------------
// Memory-related

// Allocator to allocate device memory.
nw::demo::DemoAllocator s_DeviceAllocator;

//----------------------------------------
// Filename definitions.
const wchar_t* SKY_SPHERE_FILE_NAME  = NW_DEMO_FILE_PATH(L"SkySphere.bcmdl");

const wchar_t* MODEL_RESOURCE_FILES[] =
{
    NW_DEMO_FILE_PATH(L"MaleShadow.bcmdl"),
    NW_DEMO_FILE_PATH(L"SceneEnvironmentSetting.bcenv"),
    NW_DEMO_FILE_PATH(L"SpotLight.bcenv"),
    NW_DEMO_FILE_PATH(L"AmbientLight.bcenv"),
};

const wchar_t* SKELETAL_ANIM_RESOURCE_FILE = NW_DEMO_FILE_PATH(L"WalkAimAt.bcskla");

const wchar_t* SHADOW_MATERIAL_FILE_NAME = NW_DEMO_FILE_PATH(L"ShadowModel.bcmdl");

//----------------------------------------
// Rendering-related
const int RENDER_TARGET_COUNT = 2;
typedef nw::ut::FixedSizeArray<nw::gfx::IRenderTarget*, RENDER_TARGET_COUNT> RenderTargetArray;

RenderTargetArray s_RenderTargets;
nw::demo::SceneSystem*  s_SceneSystem = NULL;
nw::demo::RenderSystem* s_RenderSystem = NULL;

nw::demo::GraphicsDrawing  s_GraphicsDrawing;

//----------------------------------------
// Resource-related
nw::demo::ResourceArray s_Resources;

//----------------------------------------
// Scene-related
nw::gfx::SceneNode* s_SceneRoot = NULL;
nw::gfx::SceneNode* s_ModelRoot = NULL;
nw::gfx::TransformNode* s_LightRoot = NULL;
s32 s_FrameCount = 0;
nw::gfx::Camera* s_BaseCamera = NULL;
nw::gfx::Camera* s_LeftCamera = NULL;
nw::gfx::Camera* s_RightCamera = NULL;
nw::gfx::FragmentLight* s_SpotLight = NULL;

const f32 s_fNearPlane = 0.1f;
const f32 s_fFarPlane = 1000.0f;

//----------------------------------------
// Scene environment-related
const s32 ENVIRONMENT_SETTINGS_COUNT = 1;

typedef nw::ut::FixedSizeArray<nw::gfx::SceneEnvironmentSetting*, ENVIRONMENT_SETTINGS_COUNT> SceneEnvironmentSettingArray;
SceneEnvironmentSettingArray s_SceneEnvironmentSettings;

const s32 s_BaseCameraIndex = 0;
const s32 s_ShadowCameraIndex = 1;

//----------------------------------------
// Animation-related
nw::gfx::SkeletalModel* s_AnimModel = NULL;
const int MAX_ANIM_OBJECTS = 8;
nw::ut::FixedSizeArray<nw::gfx::AnimObject*, MAX_ANIM_OBJECTS> s_AnimObjects;

/*
  
 */
void
InitializeGraphics()
{
    // Perform stereo settings to renderDescription.
    nw::demo::RenderSystem::Description renderDescription;

    renderDescription.reusableCommandBufferSize = 0x100000;
    renderDescription.reusableCommandRequestCount      = 512;
    renderDescription.upperScreenMode = nw::demo::UPPER_SCREEN_MODE_STEREO;

    s_RenderSystem = nw::demo::RenderSystem::Create(&s_DeviceAllocator, renderDescription);

    s_GraphicsDrawing.SetScreenSize(
        renderDescription.lowerScreenDescription.width,
        renderDescription.lowerScreenDescription.height
    );

    nw::demo::Utility::InitializeGraphicsDrawing(&s_DeviceAllocator, s_GraphicsDrawing);

    s_RenderTargets.push_back(
        nw::demo::Utility::CreateUpperScreenBuffer(&s_DeviceAllocator, renderDescription)
    );

    bool isPerspectiveShadow = true;

    // With the Ortho camera, division using the r value is not performed.
    if (s_CameraKind == ORTHO_CAMERA)
    {
        isPerspectiveShadow = false;
    }

    // Generate a shadow map texture.
    s_ShadowTexture =
        nw::gfx::ResShadowTexture::DynamicBuilder()
        .Width(s_ShadowTextureSize)
        .Height(s_ShadowTextureSize)
        .LocationFlag(NN_GX_MEM_VRAMA | GL_NO_COPY_FCRAM_DMP)
        .ExecutingMemoryFill(true)
        .PerspectiveShadow(isPerspectiveShadow)
        .ShadowZBias(s_ShadowBias * s_ShadowNear / (s_ShadowFar - s_ShadowNear))
        .Create(&s_DeviceAllocator);
    s_ShadowTexture.Setup(&s_DeviceAllocator, nw::gfx::ResGraphicsFile(NULL));

    // Create the off-screen buffer.
    nw::gfx::IRenderTarget* offScreenTarget =
        nw::gfx::IRenderTarget::CreateOffScreenBuffer(&s_DeviceAllocator, s_ShadowTexture);

    s_RenderTargets.push_back(offScreenTarget);

    NW_ASSERT(!s_RenderTargets.empty());
    s_RenderSystem->GetRenderContext()->SetRenderTarget(s_RenderTargets[0]);

    // Standard settings to sceneDescription are performed with the constructor.
    nw::demo::SceneSystem::Description sceneDescription;
    s_SceneSystem = nw::demo::SceneSystem::Create(&s_DeviceAllocator, sceneDescription);

    // Sets the furthest background model for the demo to the rendering system.
    // With the gfx demo, glClear, etc. is not used, and a fill is applied with the background.
    s_RenderSystem->LoadSkyModel(SKY_SPHERE_FILE_NAME);

    NW_GL_ASSERT();
}

/*
  
 */
void
TerminateGraphics()
{
    s_ShadowTexture.Cleanup();
    if (s_ShadowTexture.IsValid())
    {
        s_ShadowTexture.DynamicDestroy();
    }

    nw::gfx::SafeDestroy(s_LeftCamera);

    nw::gfx::SafeDestroy(s_RightCamera);

    nw::gfx::SafeDestroy(s_SceneSystem);

    nw::gfx::SafeDestroyAll(s_RenderTargets);

    s_GraphicsDrawing.Finalize();

    nw::gfx::SafeDestroy(s_RenderSystem);

    NW_GL_ASSERT();
}

/*
  

  
  
  

  
 */
nw::gfx::TransformAnimEvaluator*
CreateTransformAnimEvaluator(
    const int maxMembers,
    const bool translateAnimEnabled,
    const wchar_t* filePath
)
{
    //----------------------------------------
    // Generate the animation resource.
    nw::demo::ResourceSet* resourceSet = nw::demo::Utility::LoadResources(s_Resources, filePath, &s_DeviceAllocator);
    if (resourceSet->resource.GetSkeletalAnimsCount() == 0)
    {
        return NULL;
    }
    nw::anim::ResAnim resAnim = resourceSet->resource.GetSkeletalAnims(0);

    if (!resAnim.IsValid())
    {
        return NULL;
    }

    //----------------------------------------
    // Creates a transform animation evaluator.
    //
    // When an animation is application to only one model, and for frame-format data, setting AllocCache to FALSE will lower the processing load.
    // 
    nw::gfx::TransformAnimEvaluator* evaluator = nw::gfx::TransformAnimEvaluator::Builder()
        .AnimData(resAnim)
        .MaxMembers(maxMembers)
        .AllocCache(false)
        .Create(&s_DeviceAllocator);

    // Set the invalidation flag for movement animation.
    evaluator->SetIsTranslateDisabled(!translateAnimEnabled);

    return evaluator;
}

/*
  

  
 */
void
InitializeSkeletalAnim(nw::gfx::SkeletalModel* model)
{
    nw::gfx::AnimGroup* animGroup = model->GetSkeletalAnimGroup();
    if (animGroup == NULL) // There is no animation group for skeletal animations.
    {
        return;
    }

    nw::gfx::ResSkeletalModel resModel = model->GetResSkeletalModel();
    nw::gfx::ResSkeleton resSkeleton = resModel.GetSkeleton();
    const int maxBones = resSkeleton.GetBonesCount();
    const bool translateAnimEnabled =
        nw::ut::CheckFlag(resSkeleton.GetFlags(), nw::gfx::ResSkeleton::FLAG_TRANSLATE_ANIMATION_ENABLED);

    //----------------------------------------
    // Generate animation evaluation.
    nw::gfx::TransformAnimEvaluator* evaluator = CreateTransformAnimEvaluator(
        maxBones, translateAnimEnabled, SKELETAL_ANIM_RESOURCE_FILE);
    if (evaluator == NULL)
    {
        return;
    }

    // Bind the animation.
    bool bindResult = evaluator->Bind(animGroup);

    //----------------------------------------
    // Set animation for model.
    model->SetSkeletalAnimObject(evaluator);
    s_AnimObjects.PushBack(evaluator);
}

/*
  
 */
void
TerminateAnim(void)
{
    for (int animIdx = 0; animIdx < s_AnimObjects.Size(); ++animIdx)
    {
        nw::gfx::SafeDestroy(s_AnimObjects[animIdx]);
    }
    s_AnimObjects.clear();
}

/*
  
 */
void
BuildRootNodes()
{
    NW_ASSERT(s_SceneRoot == NULL);
    s_SceneRoot = nw::gfx::TransformNode::DynamicBuilder()
        .IsFixedSizeMemory(false)
        .Create(&s_DeviceAllocator);
    NW_NULL_ASSERT(s_SceneRoot);

    NW_ASSERT(s_ModelRoot == NULL);
    s_ModelRoot = nw::gfx::TransformNode::DynamicBuilder()
        .IsFixedSizeMemory(false)
        .Create(&s_DeviceAllocator);
    s_SceneRoot->AttachChild(s_ModelRoot);
    NW_NULL_ASSERT(s_ModelRoot);

    NW_ASSERT(s_LightRoot == NULL);
    s_LightRoot = nw::gfx::TransformNode::DynamicBuilder()
        .IsFixedSizeMemory(false)
        .Create(&s_DeviceAllocator);
    s_SceneRoot->AttachChild(s_LightRoot);
    NW_NULL_ASSERT(s_LightRoot);
}

/*
  
 */
void
BuildCameras()
{
    nw::demo::Utility::CreateStereoCameras(
        &s_BaseCamera,
        &s_LeftCamera,
        &s_RightCamera,
        &s_DeviceAllocator,
        nw::math::VEC3(20.0f, 15.0f, 20.0f),
        nw::math::VEC3(0.0f, 10.0f, 0.0f),
        s_fNearPlane,
        s_fFarPlane
    );
    s_SceneRoot->AttachChild(s_BaseCamera);
    s_SceneSystem->GetCameraController()->Register(s_BaseCamera);

    // Create the camera for the shadow.
    switch (s_CameraKind)
    {
        case PERSPECTIVE_CAMERA:
            {
                s_ShadowCamera =
                    nw::demo::Utility::CreateCamera(
                        &s_DeviceAllocator,
                        nw::math::VEC3(20.0f, 15.0f, 20.0f),
                        nw::math::VEC3(0.0f, 10.0f, 0.0f),
                        s_ShadowNear,
                        s_ShadowFar,
                        NW_MATH_DEG_TO_RAD(45.0f),
                        nw::math::PIVOT_NONE,
                        1.0f / s_ShadowFar);
            }
            break;
        case FRUSTUM_CAMERA:
            {
                s_ShadowCamera =
                    nw::demo::Utility::CreateFrustumCamera(
                        &s_DeviceAllocator,
                        nw::math::VEC3(20.0f, 15.0f, 20.0f),
                        nw::math::VEC3(0.0f, 10.0f, 0.0f),
                        s_ShadowNear,
                        s_ShadowFar,
                        nw::math::VEC2(0.0f, 0.0f),
                        1.0f,
                        nw::math::PIVOT_NONE,
                        1.0f / s_ShadowFar);
            }
            break;
        case ORTHO_CAMERA:
            {
                s_ShadowCamera =
                    nw::demo::Utility::CreateOrthoCamera(
                        &s_DeviceAllocator,
                        nw::math::VEC3(20.0f, 15.0f, 20.0f),
                        nw::math::VEC3(0.0f, 10.0f, 0.0f),
                        s_ShadowNear,
                        s_ShadowFar,
                        nw::math::VEC2(0.0f, 0.0f),
                        s_OrthoCameraHeight,
                        nw::math::PIVOT_NONE);
            }
            break;
        default:
            {
                NW_FATAL_ERROR("Unsupported camera type.");
            }
            break;
    }

    NW_NULL_ASSERT(s_ShadowCamera);

    s_SceneRoot->AttachChild(s_ShadowCamera);
}

/*
  
 */
void
SetupShadowMaterial(nw::gfx::ResMaterial resMaterial)
{
    NW_ASSERT(resMaterial.GetTextureCoordinatorsCount() != 0);

    resMaterial.SetTextureCoordinatorsHash(0x0);
    resMaterial.SetTextureMappersHash(0x0);
    resMaterial.SetTextureCombinersHash(0x0);
    resMaterial.SetMaterialColorHash(0x0);

    nw::gfx::ResTextureCoordinator resTexCoord = resMaterial.GetTextureCoordinators(0);

    // Set the index for the camera for shadow maps and texture matrix calculation method for shadow maps.
    resTexCoord.SetReferenceCamera(1);
    resTexCoord.SetMappingMethod(nw::gfx::ResTextureCoordinator::MAPPINGMETHOD_SHADOW);

    NW_ASSERT(resMaterial.GetTextureMappersCount() != 0);
    nw::gfx::ResPixelBasedTextureMapper resTextureMapper = resMaterial.GetTextureMappers(0);

    // Replace the dummy texture with a shadow texture.
    nw::gfx::ResTexture dummyTexture = resTextureMapper.GetTexture().Dereference();
    dummyTexture.Cleanup();
    resTextureMapper.SetTexture(s_ShadowTexture);

    nw::gfx::ResFragmentShader resFragmentShader = resMaterial.GetFragmentShader();

    // Set the combiner for displaying shadows.
    nw::gfx::ResTextureCombiner combiner0 = resFragmentShader.GetTextureCombiners(0);
    combiner0.SetSourceRgb(0, nw::gfx::ResTextureCombiner::SOURCE_TEXTURE0);
    combiner0.SetSourceRgb(1, nw::gfx::ResTextureCombiner::SOURCE_CONSTANT);
    combiner0.SetOperandRgb(0, nw::gfx::ResTextureCombiner::OPERANDRGB_ONE_MINUS_SRC_COLOR);
    combiner0.SetOperandRgb(1, nw::gfx::ResTextureCombiner::OPERANDRGB_SRC_COLOR);
    combiner0.SetCombineRgb(nw::gfx::ResTextureCombiner::COMBINE_MODULATE);

    nw::gfx::ResTextureCombiner combiner1 = resFragmentShader.GetTextureCombiners(1);
    combiner1.SetSourceRgb(0, nw::gfx::ResTextureCombiner::SOURCE_TEXTURE1);
    combiner1.SetSourceRgb(1, nw::gfx::ResTextureCombiner::SOURCE_PREVIOUS);
    combiner1.SetOperandRgb(0, nw::gfx::ResTextureCombiner::OPERANDRGB_SRC_COLOR);
    combiner1.SetOperandRgb(1, nw::gfx::ResTextureCombiner::OPERANDRGB_ONE_MINUS_SRC_COLOR);
    combiner1.SetCombineRgb(nw::gfx::ResTextureCombiner::COMBINE_MODULATE);

    nw::gfx::ResTextureCombiner combiner2 = resFragmentShader.GetTextureCombiners(2);
    combiner2.SetSourceRgb(0, nw::gfx::ResTextureCombiner::SOURCE_FRAGMENT_PRIMARY_COLOR);
    combiner2.SetSourceRgb(1, nw::gfx::ResTextureCombiner::SOURCE_PREVIOUS);
    combiner2.SetSourceRgb(2, nw::gfx::ResTextureCombiner::SOURCE_FRAGMENT_SECONDARY_COLOR);
    combiner2.SetOperandRgb(0, nw::gfx::ResTextureCombiner::OPERANDRGB_SRC_COLOR);
    combiner2.SetOperandRgb(1, nw::gfx::ResTextureCombiner::OPERANDRGB_SRC_COLOR);
    combiner2.SetOperandRgb(2, nw::gfx::ResTextureCombiner::OPERANDRGB_SRC_COLOR);
    combiner2.SetCombineRgb(nw::gfx::ResTextureCombiner::COMBINE_MULT_ADD);

    // Sets the shadow color.
    nw::gfx::ResMaterialColor color = resMaterial.GetMaterialColor();
    color.SetConstant0(1.0f - s_ShadowIntensity, 1.0f - s_ShadowIntensity, 1.0f - s_ShadowIntensity);
}

/*
  
 */
void
BuildResources(nw::demo::ResourceSet* resourceSet)
{
    resourceSet->resource.ForeachTexture(nw::gfx::LocationFlagSetter(NN_GX_MEM_VRAMA | GL_NO_COPY_FCRAM_DMP));
    resourceSet->resource.ForeachIndexStream(nw::gfx::LocationFlagSetter(NN_GX_MEM_VRAMB | GL_NO_COPY_FCRAM_DMP));
    resourceSet->resource.ForeachVertexStream(nw::gfx::LocationFlagSetter(NN_GX_MEM_VRAMB | GL_NO_COPY_FCRAM_DMP));

    nw::gfx::Result result = resourceSet->resource.Setup(&s_DeviceAllocator);
    if (result.IsFailure())
    {
        NW_FATAL_ERROR("Fail to set up model. A result code is 0x%x", result.GetCode());
    }

    nw::ut::MoveArray<nw::gfx::SceneNode*> sceneNodeArray(&s_DeviceAllocator);

    nw::gfx::ResModelArray models = resourceSet->resource.GetModels();
    nw::gfx::ResModelArray::iterator modelsEnd = models.end();
    for (nw::gfx::ResModelArray::iterator modelResource = models.begin();
         modelResource != modelsEnd; ++modelResource)
    {
        nw::gfx::SceneNode* node = nw::demo::Utility::CreateSceneNode(
            &s_DeviceAllocator,
            (*modelResource)
        );

        // Set the off-screen buffer as a texture for plane's Texture0
        if (::std::strcmp((*modelResource).GetName(), "Plane") == 0)
        {
            nw::gfx::Model* model = nw::ut::DynamicCast<nw::gfx::Model*>(node);
            NW_ASSERT(model->GetMaterialCount() != 0);

            nw::gfx::Material* material = model->GetMaterial(0);
            nw::gfx::ResMaterial resMaterial = material->GetOriginal();
            NW_ASSERT(resMaterial.IsValid());
            SetupShadowMaterial(resMaterial);
        }
        else if (::std::strcmp((*modelResource).GetName(), "Male") == 0)
        {
            nw::gfx::SkeletalModel* skeletalModel = nw::ut::DynamicCast<nw::gfx::SkeletalModel*>(node);
            s_AnimModel = skeletalModel;

            nw::gfx::Material* material = skeletalModel->GetMaterial(0);
            NW_ASSERT(skeletalModel->GetMaterialCount() != 0);

            nw::gfx::ResMaterial resMaterial = material->GetOriginal();
            NW_ASSERT(resMaterial.IsValid());
            SetupShadowMaterial(resMaterial);

            // Set the flag to the model that generates the shadow.
            skeletalModel->SetUserParameter(SHADOW_CASTER);
        }

        sceneNodeArray.push_back(node);
    }

    nw::gfx::ResLightArray lights = resourceSet->resource.GetLights();
    nw::gfx::ResLightArray::iterator lightsEnd = lights.end();
    for (nw::gfx::ResLightArray::iterator lightResource = lights.begin();
         lightResource != lightsEnd; ++lightResource)
    {
        nw::gfx::SceneNode* node = nw::demo::Utility::CreateSceneNode(
            &s_DeviceAllocator,
            (*lightResource)
        );

        if (::std::strcmp((*lightResource).GetName(), "SpotLight") == 0)
        {
            s_SpotLight = nw::ut::DynamicCast<nw::gfx::FragmentLight*>(node);
            s_LightRoot->AttachChild(s_SpotLight);
        }

        sceneNodeArray.push_back(node);
    }

    // Resolve parent-child reference relationships
    nw::gfx::SceneHelper::ResolveReference(sceneNodeArray);

    // Add a model to the scene
    nw::gfx::SceneHelper::ForeachRootNodes(
        sceneNodeArray.Begin(),
        sceneNodeArray.End(),
        nw::gfx::AttachNode(s_ModelRoot)
    );

    nw::gfx::ResSceneEnvironmentSettingArray settings = resourceSet->resource.GetSceneEnvironmentSettings();
    nw::gfx::ResSceneEnvironmentSettingArray::iterator settingsEnd = settings.end();
    for (nw::gfx::ResSceneEnvironmentSettingArray::iterator settingResource = settings.begin();
        settingResource != settingsEnd; ++settingResource)
    {
        nw::gfx::SceneObject* sceneObject = nw::gfx::SceneBuilder()
            .Resource(*settingResource)
            .CreateObject(&s_DeviceAllocator, &s_DeviceAllocator);

        nw::gfx::SceneEnvironmentSetting* sceneEnvironmentSetting =
            nw::ut::DynamicCast<nw::gfx::SceneEnvironmentSetting*>(sceneObject);

        NW_NULL_ASSERT(sceneEnvironmentSetting);
        s_SceneEnvironmentSettings.push_back(sceneEnvironmentSetting);
    }
}

/*
  
 */
void
InitializeScenes()
{
    BuildRootNodes();

    BuildCameras();

    NW_FOREACH(const wchar_t* name, MODEL_RESOURCE_FILES)
    {
        BuildResources(nw::demo::Utility::LoadResources(s_Resources, name, &s_DeviceAllocator));
    }

    if (s_AnimModel != NULL)
    {
        InitializeSkeletalAnim(s_AnimModel);
    }

    // Traverse and initialize the scene tree.
    s_SceneSystem->InitializeScene(s_SceneRoot);
    s_SceneSystem->UpdateScene();

    // Resolves and sets references for the scene environment.
    s_RenderSystem->SetSceneEnvironmentSettings(s_SceneSystem, &s_SceneEnvironmentSettings);

    // Set the camera.
    nw::gfx::SceneEnvironment& sceneEnvironment = s_RenderSystem->GetSceneEnvironment();
    sceneEnvironment.SetCamera(s_BaseCameraIndex, s_BaseCamera);
    sceneEnvironment.SetCamera(s_ShadowCameraIndex, s_ShadowCamera);
    nw::demo::Utility::SetCameraAspectRatio(s_BaseCamera, s_RenderTargets[0]);

    // Load the shadow material and the model that includes the shader.
    nw::demo::ResourceSet* resourceSet = nw::demo::Utility::LoadResources(s_Resources, SHADOW_MATERIAL_FILE_NAME, &s_DeviceAllocator);
    nw::gfx::ResModel resShadowModel = resourceSet->resource.GetModels("ShadowModel");
    nw::gfx::Result result = resShadowModel.Setup(&s_DeviceAllocator, resourceSet->resource);
    NW_ASSERT(result.IsSuccess());
    nw::gfx::SceneNode* node = nw::demo::Utility::CreateSceneNode(
        &s_DeviceAllocator,
        resShadowModel
    );
    s_ShadowDummyModel = nw::ut::DynamicCast<nw::gfx::Model*>(node);
    NW_NULL_ASSERT(s_ShadowDummyModel);

    NW_ASSERT(s_ShadowDummyModel->GetMaterialCount() != 0);

    // Set the material for the shadow to the shadow mode.
    // Necessary for writing the shadow strength and depth information to the shadow map.
    nw::gfx::ResFragmentOperation fragmentOperation = s_ShadowDummyModel->GetMaterial(0)->GetOriginal().GetFragmentOperation();
    fragmentOperation.SetFragmentOperationMode(nw::gfx::ResFragmentOperation::FRAGMENT_OPERATION_MODE_SHADOW);

    // Set the shadow object's color.
    // The shadow map must be set to black (0.0f, 0.0f, 0.0f ,0.0f).
    // Settings for the shadow color occur in the combiner using constant colors.
    nw::gfx::ResMaterialColor materialColor = s_ShadowDummyModel->GetMaterial(0)->GetOriginal().GetMaterialColor();
    materialColor.SetDiffuse(0.0f, 0.0f, 0.0f, 0.0f);

    nw::demo::CommandListSwapper* commandListSwapper = s_RenderSystem->GetCommandListSwapper();
    NW_ASSERT(s_ShadowDummyModel->GetMaterialCount() != 0);
    nw::gfx::Material* shadowMaterial = s_ShadowDummyModel->GetMaterial(0);

    // Create a render command.
    void* startShadowCommandMemory = s_DeviceAllocator.Alloc(sizeof(StartShadowRenderCommand));
    NW_NULL_ASSERT(startShadowCommandMemory);
    s_StartShadowRenderCommand = new(startShadowCommandMemory) StartShadowRenderCommand(
        commandListSwapper,
        s_RenderTargets[1],
        s_ShadowCameraIndex,
        shadowMaterial);

    void* endShadowCommandMemory = s_DeviceAllocator.Alloc(sizeof(EndShadowRenderCommand));
    NW_NULL_ASSERT(endShadowCommandMemory);
    s_EndShadowRenderCommand = new(endShadowCommandMemory) EndShadowRenderCommand(commandListSwapper, s_StartShadowRenderCommand);

    void* startCommandMemory = s_DeviceAllocator.Alloc(sizeof(StartRenderCommand));
    NW_NULL_ASSERT(startCommandMemory);
    s_StartRenderCommand = new(startCommandMemory) StartRenderCommand(s_RenderSystem, s_RenderTargets[0], s_BaseCameraIndex);

    void* endCommandMemory = s_DeviceAllocator.Alloc(sizeof(EndRenderCommand));
    NW_NULL_ASSERT(endCommandMemory);
    s_EndRenderCommand = new(endCommandMemory) EndRenderCommand(s_RenderSystem, s_LeftCamera, s_RightCamera, s_StartRenderCommand);

    NW_GL_ASSERT();

    s_FrameCount = 0;
}

/*
  
 */
void
TerminateScenes()
{
    s_DeviceAllocator.Free(s_EndRenderCommand);
    s_DeviceAllocator.Free(s_StartRenderCommand);
    s_DeviceAllocator.Free(s_EndShadowRenderCommand);
    s_DeviceAllocator.Free(s_StartShadowRenderCommand);

    nw::ut::SafeDestroy(s_ShadowDummyModel);
    nw::gfx::SafeDestroyBranch(s_SceneRoot);
    nw::demo::SafeCleanupResources(s_Resources);
    nw::ut::SafeDestroyAll(s_SceneEnvironmentSettings);
    TerminateAnim();

    NW_GL_ASSERT();

    s_Resources.clear();
    s_SceneEnvironmentSettings.clear();
    s_ModelRoot = NULL;
    s_LightRoot = NULL;
}

/*
  
 */
void
UpdateScene()
{
    float radian = static_cast<float>((s_FrameCount) % (314 * 2)) * 0.01f;

    // Rotate the light root around the Y-axis.
    nw::gfx::TransformNode* lightNode =
        nw::ut::DynamicCast<nw::gfx::TransformNode*>(s_LightRoot);
    NW_NULL_ASSERT(lightNode);

    lightNode->Transform().SetRotateXYZ(0.0f, radian, 0.0f);

    s_SceneSystem->GetCameraController()->Update();

    s_SceneSystem->UpdateScene();

    s_BaseCamera->UpdateCameraMatrix();

    s_RenderSystem->CalcStereoCamera(s_LeftCamera, s_RightCamera, s_BaseCamera, s_fNearPlane + 5.0f);

    // To perform shadow rendering from the light direction, track the camera position and direction to the light.
    nw::math::VEC3 translate;
    translate.x = s_SpotLight->WorldMatrix().m[0][3];
    translate.y = s_SpotLight->WorldMatrix().m[1][3];
    translate.z = s_SpotLight->WorldMatrix().m[2][3];

    s_ShadowCamera->WorldMatrix().m[0][3] = translate.x;
    s_ShadowCamera->WorldMatrix().m[1][3] = translate.y;
    s_ShadowCamera->WorldMatrix().m[2][3] = translate.z;

    const nw::math::VEC3& direction = s_SpotLight->Direction();
    s_ShadowCamera->SetTargetPosition(translate + direction);

    s_ShadowCamera->UpdateCameraMatrix();

    ++s_FrameCount;
}

/*
  
 */
void
SubmitView()
{
    nw::gfx::RenderContext* renderContext = s_RenderSystem->GetRenderContext();
    nw::gfx::SceneEnvironment& sceneEnvironment = renderContext->GetSceneEnvironment();
    nw::gfx::SceneContext* sceneContext = s_SceneSystem->GetSceneContext();

    nw::gfx::RenderQueue* renderQueue = s_RenderSystem->GetRenderQueue();

    renderQueue->Reset();

    // Add the render target and render command to the render queue.
    // To make it easy to understand, perform Submit or Enqueue in the order to be processed.

    // Add the command that performs the setting for the shadow to the callback that executes after layer 0.
    s_RenderSystem->EnqueueRenderCommand(
        s_StartShadowRenderCommand, nw::gfx::ResMaterial::TRANSLUCENCY_KIND_OPAQUE, 0, 0);

    // Add only the model specified as the shadow caster to the render queue as layer 1.
    s_SceneSystem->GetSceneUpdater()->SubmitView(
        renderQueue,
        sceneContext,
        *s_ShadowCamera,
        1,
        1,
        &s_ShadowCasterModel,
        s_RenderSystem->GetRenderSortMode());

    // Add the command that performs shadow post-processing to the callback that executes after layer 1.
    s_RenderSystem->EnqueueRenderCommand(
        s_EndShadowRenderCommand, nw::gfx::ResMaterial::TRANSLUCENCY_KIND_OPAQUE, 0, 1);

    // Add the command that performs the settings for stereo rendering pre-processing to the callback that executes after layer １.
    s_RenderSystem->EnqueueRenderCommand(
        s_StartRenderCommand, nw::gfx::ResMaterial::TRANSLUCENCY_KIND_OPAQUE, 1, 1);

    // Add all of the scene context models to the render queue as layer 2.
    s_SceneSystem->GetSceneUpdater()->SubmitView(
        renderQueue,
        sceneContext,
        *s_BaseCamera,
        2,
        s_RenderSystem->GetRenderSortMode());

    // Add the command that performs the settings for stereo rendering post-processing to the callback that executes after layer 2.
    s_RenderSystem->EnqueueRenderCommand(
        s_EndRenderCommand, nw::gfx::ResMaterial::TRANSLUCENCY_KIND_OPAQUE, 0, 2);

    std::sort(
        renderQueue->Begin(),
        renderQueue->End(),
        nw::gfx::RenderElementCompare());
}

//----------------------------------------
struct RenderSceneInternalFunctor
: public std::unary_function<nw::gfx::RenderElement&, void>
{
    nw::gfx::RenderContext* m_RenderContext;
    nw::gfx::MeshRenderer* m_MeshRenderer;

    RenderSceneInternalFunctor(nw::gfx::RenderContext* renderContext, nw::gfx::MeshRenderer* meshRenderer)
    : m_RenderContext(renderContext), m_MeshRenderer(meshRenderer)
    {
        NW_NULL_ASSERT(renderContext);
        NW_NULL_ASSERT(meshRenderer);
    }

    void operator()(nw::gfx::RenderElement& element)
    {
        if (element.IsCommand())
        {
            nw::gfx::RenderCommand* command = element.GetCommand();
            NW_NULL_ASSERT(command);
            command->Invoke(this->m_RenderContext);
        }
        else
        {
            nw::gfx::ResMesh mesh = element.GetMesh();
            nw::gfx::Model* model = element.GetModel();
            model->PreRenderSignal()(model, mesh, this->m_RenderContext);
            this->m_MeshRenderer->RenderMesh(mesh, model);
            model->PostRenderSignal()(model, mesh, this->m_RenderContext);
        }
        NW_GL_ASSERT();
    }
};

/*
  
 */
void
RenderScene()
{
    nw::gfx::RenderContext* renderContext = s_RenderSystem->GetRenderContext();
    nw::gfx::RenderQueue* renderQueue = s_RenderSystem->GetRenderQueue();
    nw::gfx::MeshRenderer* meshRenderer = s_RenderSystem->GetMeshRenderer();

    std::for_each(
        renderQueue->Begin(),
        renderQueue->End(),
        RenderSceneInternalFunctor(renderContext, meshRenderer));

    NW_GL_ASSERT();
}

/*
  
 */
void
ReportDemo()
{
    NW_PROFILE("ReportDemo");

    // This load is exempted from the load display.
    s_RenderSystem->SuspendLoadMeter();

    s_RenderSystem->AddLoadMeterCommandSize(
        s_EndShadowRenderCommand->GetCommandSize() - s_StartShadowRenderCommand->GetCommandSize());

    nw::demo::DebugUtility::CalcLoadMeter(s_RenderSystem);

    s_GraphicsDrawing.BeginDrawingShape();

    nw::demo::DebugUtility::DrawLoadMeter(
        s_RenderSystem,
        &s_GraphicsDrawing
    );

    s_GraphicsDrawing.EndDrawingShape();

    s_GraphicsDrawing.BeginDrawingString();

    nw::demo::DebugUtility::DrawLoadMeterText(
        s_RenderSystem,
        &s_GraphicsDrawing
    );

    s_GraphicsDrawing.EndDrawingString();

    s_RenderSystem->ResumeLoadMeter();
}

/*
  
 */
void
DemoScene()
{
    NW_ASSERT(!s_RenderTargets.empty());

    nw::gfx::RenderContext* renderContext = s_RenderSystem->GetRenderContext();

    InitializeScenes();

    nw::demo::DebugUtility::PostInitializeScenes();

    bool isContinuing = true;

    while ( isContinuing )
    {
        nw::demo::DebugUtility::AdvanceAutoTestFrame();

        nw::demo::PadFactory::GetPad()->Update();

        UpdateScene();

        SubmitView();

        RenderScene();

        s_RenderSystem->ClearBySkyModel(s_BaseCamera);
        ReportDemo();
        s_RenderSystem->TransferBuffer(nw::demo::LOWER_SCREEN);

        s_RenderSystem->PresentBuffer(nw::demo::UPPER_SCREEN | nw::demo::LOWER_SCREEN | nw::demo::EXTENSION_SCREEN);

        renderContext->ResetState();

        if (nw::demo::Utility::IsTerminating())
        {
            isContinuing = false;
        }
    }

    nw::demo::DebugUtility::PreTerminateScenes();

    TerminateScenes();
}

} // Namespace

/*
  
 */
void
nnMain()
{
    nn::applet::Enable();

    nw::demo::InitializeGraphicsSystem(&s_DeviceAllocator);

    nw::demo::PadFactory::Initialize(&s_DeviceAllocator);

    NW_DEMO_TEST_LOOP(&s_DeviceAllocator, NULL, &s_RenderSystem)
    {

        InitializeGraphics();

        DemoScene();

        TerminateGraphics();

    }

    nw::demo::PadFactory::Finalize();

    nw::demo::FinalizeGraphicsSystem();
}
