/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#include "precompiled.h"

#include <nw/anim/anim_AnimBlend.h>
#include <nw/gfx/gfx_AnimAdder.h>
#include <nw/gfx/gfx_AnimGroup.h>

namespace nw
{
namespace gfx
{

NW_UT_RUNTIME_TYPEINFO_DEFINITION(AnimAdder        , AnimBlender);

//----------------------------------------------------------
const anim::AnimResult*
AnimAdder::GetResult(
    void* target,
    int memberIdx
) const
{
    //----------------------------------------------------------
    // When there is no blend operation or when there is no blending for a blend operation, overwrite
    // 
    const anim::AnimBlendOp* blendOp = m_AnimGroup->GetBlendOperation(memberIdx);
    if (blendOp == NULL || !blendOp->HasBlend())
    {
        for (int animIdx = m_AnimObjects.Size() - 1; animIdx >= 0; --animIdx)
        {
            if (m_AnimObjects[animIdx] == NULL)
            {
                continue;
            }
            const anim::AnimResult* childResult =
                m_AnimObjects[animIdx]->GetResult(target, memberIdx);
            if (childResult != NULL)
            {
                return childResult;
            }
        }
        return NULL;
    }

    //----------------------------------------------------------
    // If there is operational space for interpolation it should occur there, and after evacuating target's CONVERTED flag, it should be turned on
    // 
    anim::AnimResult* result = reinterpret_cast<anim::AnimResult*>(target);
    const bool convertedBak = result->IsEnabledFlags(anim::AnimResult::FLAG_CONVERTED);
    result->EnableFlags(anim::AnimResult::FLAG_CONVERTED, true);

    //----------------------------------------------------------
    // Loop for all child animations
    anim::AnimResult workResult;
    bool written = false;
    //for (int animIdx = 0; animIdx < m_AnimObjects.Size(); ++animIdx)
    for (int animIdx = m_AnimObjects.Size() - 1; animIdx >= 0; --animIdx)
    {
        if (m_AnimObjects[animIdx] == NULL)
        {
            continue;
        }
        const float childWeight = m_Weights[animIdx];
        if (!AnimWeightNearlyEqualZero(childWeight))
        {
            // Reflect the evacuated CONVERTED flag to work
            // If the child is a Blender that handles CONVERTED, expect the CONVERTED results to be returned as-is
            // When the child is an Evaluator, the flag is returned as off but there is no problem for conversion to occur within blendOp
            workResult.EnableFlags(anim::AnimResult::FLAG_CONVERTED, convertedBak);

            // Evaluations
            const anim::AnimResult* childResult =
                m_AnimObjects[animIdx]->GetResult(&workResult, memberIdx);
            if (childResult != NULL)
            {
                written = true;
                if (!blendOp->Blend(result, NULL, childResult, &childWeight))
                {
                    break;
                }
            }
        }
    }

    //----------------------------------------------------------
    // Restore target's CONVERTED flag
    if (!convertedBak)
    {
        result->DisableFlags(anim::AnimResult::FLAG_CONVERTED);
    }

    if (!written) // No valid child animation
    {
        return NULL;
    }

    //----------------------------------------------------------
    // Execute if there are processes after blending
    if (!convertedBak && blendOp->HasPostBlend())
    {
        blendOp->PostBlend(result, NULL);
    }
    return result;
}

} // namespace gfx
} // namespace nw
