/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#include "precompiled.h"

#include <nw/gfx/gfx_AnimBinding.h>
#include <nw/gfx/gfx_AnimEvaluator.h>
#include <nw/gfx/gfx_TransformAnimEvaluator.h>

namespace nw
{
namespace gfx
{

//----------------------------------------------------------
void
AnimBinding::Evaluate(anim::ResGraphicsAnimGroup::EvaluationTiming timing)
{
    for (int animGroupIdx = 0; animGroupIdx < m_AnimGroups.Size(); ++animGroupIdx)
    {
        AnimGroup* animGroup = m_AnimGroups[animGroupIdx];

        // Check the existence of animations and evaluation timing
        if (animGroup == NULL || animGroup->GetResGraphicsAnimGroup().GetEvaluationTiming() != timing)
        {
            continue;
        }

        for (int animObjectIdx = 0; animObjectIdx < m_AnimObjectCountPerGroup; ++animObjectIdx)
        {
            const int index = animGroupIdx * m_AnimObjectCountPerGroup + animObjectIdx;
            AnimObject* animObj = m_AnimObjects[index];
            if (animObj == NULL)
            {
                continue;
            }

            // For greater speed, branch on the existence of the blender, and on whether to Transform
            switch (animObj->GetAnimType())
            {
            case AnimObject::ANIMTYPE_SIMPLE:
                static_cast<AnimEvaluator*>(animObj)->UpdateCacheNonVirtual();
                EvaluateSimple(animGroup, static_cast<AnimEvaluator*>(animObj));
                break;
            case AnimObject::ANIMTYPE_TRANSFORM_SIMPLE:
                static_cast<TransformAnimEvaluator*>(animObj)->UpdateCacheNonVirtual();
                EvaluateTransformSimple(animGroup, static_cast<TransformAnimEvaluator*>(animObj));
                break;
            default:
                // If the animation evaluation result's internal cache is old, update
                animObj->UpdateCache();
                EvaluateBlender(animGroup, animObj);
            }
        }
    }
}

//----------------------------------------------------------
void AnimBinding::EvaluateSimple(AnimGroup* animGroup, AnimEvaluator* evaluator)
{
    NW_ASSERT(!(animGroup->GetResAnimGroup().GetFlags() & anim::ResAnimGroup::FLAG_IS_CALCULATED_TRANSFORM));

    // Evaluate general-use animations
    int lastTargetObjIdx = -1;
    bool targetObjSkipFlag = false;
    int animCount = evaluator->GetAnimData().GetMemberAnimSetCount();
    for (int animIdx = 0; animIdx < animCount; ++animIdx)
    {
        int memberIdx = evaluator->ReverseBindIndexTable()[animIdx];
        if (memberIdx == BaseAnimEvaluator::NotFoundIndex)
        {
            continue;
        }
        EvaluateMember(animGroup, memberIdx, evaluator, lastTargetObjIdx, targetObjSkipFlag);
    }
}

//----------------------------------------------------------
void AnimBinding::EvaluateTransformSimple(AnimGroup* animGroup, TransformAnimEvaluator* evaluator)
{
    NW_ASSERT(animGroup->GetResAnimGroup().GetFlags() & anim::ResAnimGroup::FLAG_IS_CALCULATED_TRANSFORM);

    // Evaluate transform animations
    int animCount = evaluator->GetAnimData().GetMemberAnimSetCount();
    for (int animIdx = 0; animIdx < animCount; ++animIdx)
    {
        int memberIdx = evaluator->ReverseBindIndexTable()[animIdx];
        if (memberIdx == BaseAnimEvaluator::NotFoundIndex)
        {
            continue;
        }
        EvaluateTransformMemberFast(animGroup, memberIdx, evaluator);
    }
}

//----------------------------------------------------------
void AnimBinding::EvaluateBlender(AnimGroup* animGroup, AnimObject* animObj)
{
    if (animGroup->GetResAnimGroup().GetFlags() & anim::ResAnimGroup::FLAG_IS_CALCULATED_TRANSFORM)
    {
        // Evaluate transform animations
        int memberCount = animGroup->GetMemberCount();
        for (int memberIdx = 0; memberIdx < memberCount; ++memberIdx)
        {
            EvaluateTransformMember(animGroup, memberIdx, animObj);
        }
    }
    else
    {
        // Evaluate general-use animations
        int lastTargetObjIdx = -1;
        bool targetObjSkipFlag = false;
        int memberCount = animGroup->GetMemberCount();
        for (int memberIdx = 0; memberIdx < memberCount; ++memberIdx)
        {
            EvaluateMember(animGroup, memberIdx, animObj, lastTargetObjIdx, targetObjSkipFlag);
        }
    }
}

//----------------------------------------------------------
void AnimBinding::EvaluateTransformMember(AnimGroup* animGroup, int memberIdx, AnimObject* animObj)
{
    AnimGroup::PreEvaluateCallback preEvaluateCallback = animGroup->GetPreEvaluateCallback();

    // Call the pre-evaluation callback
    if (preEvaluateCallback != NULL)
    {
        const int targetObjIdx = animGroup->GetTargetObjectIndex(memberIdx);
        if (!preEvaluateCallback(animGroup, targetObjIdx))
        {
            return;
        }
    }

    // Evaluations
    CalculatedTransform* target = static_cast<CalculatedTransform*>(animGroup->GetTargetPtr(memberIdx));
    animObj->GetResult(target, memberIdx);

    // Since there is no need to set via Setter, Setter is not called
    // The rotation matrix is generated within Setter, so calling it has a big affect on performance
}

//----------------------------------------------------------
void AnimBinding::EvaluateTransformMemberFast(AnimGroup* animGroup, int memberIdx, TransformAnimEvaluator* evaluator)
{
    AnimGroup::PreEvaluateCallback preEvaluateCallback = animGroup->GetPreEvaluateCallback();

    // Call the pre-evaluation callback
    if (preEvaluateCallback != NULL)
    {
        const int targetObjIdx = animGroup->GetTargetObjectIndex(memberIdx);
        if (!preEvaluateCallback(animGroup, targetObjIdx))
        {
            return;
        }
    }

    // Evaluations
    CalculatedTransform* target = static_cast<CalculatedTransform*>(animGroup->GetTargetPtr(memberIdx));
    evaluator->GetResultFast(target, memberIdx);

    // Since there is no need to set via Setter, Setter is not called
    // The rotation matrix is generated within Setter, so calling it has a big affect on performance
}

//----------------------------------------------------------
void AnimBinding::EvaluateMember(AnimGroup* animGroup, int memberIdx, AnimObject* animObj, int& lastTargetObjIdx, bool& targetObjSkipFlag)
{
    AnimGroup::PreEvaluateCallback preEvaluateCallback = animGroup->GetPreEvaluateCallback();

    // Call the pre-evaluation callback
    if (preEvaluateCallback != NULL)
    {
        const int targetObjIdx = animGroup->GetTargetObjectIndex(memberIdx);
        if (targetObjIdx != lastTargetObjIdx)
        {
            // For faster speeds, only call the pre-evaluation callback when the target object is different from the time before
            targetObjSkipFlag = !preEvaluateCallback(animGroup, targetObjIdx);
            lastTargetObjIdx = targetObjIdx;
        }
        if (targetObjSkipFlag)
        {
            return;
        }
    }

    // Evaluations
    anim::ResAnimGroupMember member = animGroup->GetResAnimGroupMember(memberIdx);
    void* target = NULL;
    ut::Offset texturePatternTarget;

    if (member.GetObjectType() == anim::ResAnimGroupMember::OBJECT_TYPE_TEXTURE_MAPPER &&
        member.GetMemberType() == anim::ResTextureMapperMember::MEMBER_TYPE_TEXTURE)
    {
        // Texture pattern animations only call Setter without overwriting target
        //
        // When the material buffer is valid, target indicates a dynamically allocated ReferenceTexture
        // If that is overwritten, at deallocation the allocator will attempt to deallocate a pointer that was not allocated and there will be a crash
        target = &texturePatternTarget;
    }
    else
    {
        target = animGroup->GetTargetPtr(memberIdx);
    }

    const anim::AnimBlendOp* blendOp = animGroup->GetBlendOperation(memberIdx);
    const anim::AnimResult* resultPtr = NULL;
    if (blendOp != NULL)
    {
        // When there are blend operations, write to the target via the AnimResult object
        // 
        anim::AnimResult result;
        resultPtr = animObj->GetResult(&result, memberIdx);
        if (resultPtr != NULL)
        {
            blendOp->Apply(target, resultPtr); // Write only the component for when the VALID flag is on
        }
    }
    else
    {
        // When there are no blend operations, directly overwrite the target
        resultPtr = animObj->GetResult(target, memberIdx);
    }

    // For members of transform, do not call setter.
    bool isTransformMember = (
        (member.GetObjectType() == anim::ResAnimGroupMember::OBJECT_TYPE_TRANSFORM) &&
        (member.GetMemberType() == anim::ResTransformMember::MEMBER_TYPE_TRANSFORM)
        );

    if (!isTransformMember && resultPtr)
    {
        // Call Setter for members needed for processes other than writing to memory
        // (Updating of the command cache, etc., occurs)
        //
        // Setter rewrites target as well so writing occurs twice.
        // We want to do it once, but when there is a component when the VALID flag is off, you have to get the target value.
        // Therefore the order becomes buffer allocation > Get > Apply > Set, which will likely make things slower.
        void* object = animGroup->GetTargetObject(memberIdx);
        member.SetValueForType(object, target);
    }
}

} // namespace gfx
} // namespace nw
