/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#include "precompiled.h"

#include <nw/anim/anim_AnimBlend.h>
#include <nw/gfx/gfx_AnimGroup.h>
#include <nw/gfx/gfx_AnimInterpolator.h>

namespace nw
{
namespace gfx
{

NW_UT_RUNTIME_TYPEINFO_DEFINITION(AnimInterpolator , AnimBlender);

//----------------------------------------------------------
const anim::AnimResult*
AnimInterpolator::GetResult(
    void* target,
    int memberIdx
) const
{
    //----------------------------------------------------------
    // When there is no blend operation or when there is no blending for a blend operation, overwrite
    // 
    const anim::AnimBlendOp* blendOp = m_AnimGroup->GetBlendOperation(memberIdx);
    if (blendOp == NULL || !blendOp->HasBlend())
    {
        for (int animIdx = m_AnimObjects.Size() - 1; animIdx >= 0; --animIdx)
        {
            if (m_AnimObjects[animIdx] == NULL)
            {
                continue;
            }
            const anim::AnimResult* childResult =
                m_AnimObjects[animIdx]->GetResult(target, memberIdx);
            if (childResult != NULL)
            {
                return childResult;
            }
        }
        return NULL;
    }

    if (m_IsOldMethod)
    {
        // TODO: Clean up the blocks here.

        //----------------------------------------------------------
        // Find the total weight of valid child animations
        float weightSum = 0.0f;
        const AnimObject* lastAnimObj = NULL;
        int validAnimCount = 0;
        for (int animIdx = 0; animIdx < m_AnimObjects.Size(); ++animIdx)
        {
            if (m_AnimObjects[animIdx] == NULL)
            {
                continue;
            }
            const float weight = m_Weights[animIdx];
            if (!AnimWeightNearlyEqualZero(weight))
            {
                if (m_AnimObjects[animIdx]->HasMemberAnim(memberIdx))
                {
                    weightSum += weight;
                    lastAnimObj = m_AnimObjects[animIdx];
                    ++validAnimCount;
                }
            }
        }

        if (validAnimCount == 0)
        {
            // When there are no valid child animations, NULL is returned without changing target
            // 
            return NULL;
        }
        else if (validAnimCount == 1)
        {
            // When there is only 1 valid child animation, return that animation's result
            // 
            return lastAnimObj->GetResult(target, memberIdx);
        }

        // When the weight total is not 1.0, calculate a normalized scale
        const float weightNormalizeScale = GetAnimWeightNormalizeScale(weightSum);

        //----------------------------------------------------------
        // If there is operational space for interpolation it should occur there, and after evacuating target's CONVERTED flag, it should be turned on
        // 
        anim::AnimResult* result = reinterpret_cast<anim::AnimResult*>(target);
        const bool convertedBak = result->IsEnabledFlags(anim::AnimResult::FLAG_CONVERTED);
        result->EnableFlags(anim::AnimResult::FLAG_CONVERTED, true);

        //----------------------------------------------------------
        // Loop for all child animations
        anim::AnimResult workResult;
        bool written = false;
        float compWeights[anim::AnimResult::MAX_COMPONENTS];
        //for (int animIdx = 0; animIdx < m_AnimObjects.Size(); ++animIdx)
        for (int animIdx = m_AnimObjects.Size() - 1; animIdx >= 0; --animIdx)
        {
            if (m_AnimObjects[animIdx] == NULL)
            {
                continue;
            }
            const float childWeight = m_Weights[animIdx] * weightNormalizeScale;
            if (!AnimWeightNearlyEqualZero(childWeight))
            {
                // Reflect the evacuated CONVERTED flag to work
                // If the child is a Blender that handles CONVERTED, expect the CONVERTED results to be returned as-is
                // When the child is an Evaluator, the flag is returned as off but there is no problem for conversion to occur within blendOp
                workResult.EnableFlags(anim::AnimResult::FLAG_CONVERTED, convertedBak);

                // Evaluations
                const anim::AnimResult* childResult =
                    m_AnimObjects[animIdx]->GetResult(&workResult, memberIdx);
                if (childResult != NULL)
                {
                    written = true;
                    if (!blendOp->Blend(result, compWeights, childResult, &childWeight))
                    {
                        break;
                    }
                }
            }
        }

        //----------------------------------------------------------
        // Restore target's CONVERTED flag
        if (!convertedBak)
        {
            result->DisableFlags(anim::AnimResult::FLAG_CONVERTED);
        }

        if (!written) // No valid child animation
        {
            return NULL;
        }

        //----------------------------------------------------------
        // Execute if there are processes after blending
        if (!convertedBak && blendOp->HasPostBlend())
        {
            blendOp->PostBlend(result, compWeights);
        }
        return result;
    }
    else
    {
        bool isValidAnim = false;
        for (int animIdx = 0; animIdx < m_AnimObjects.Size(); ++animIdx)
        {
            if (m_AnimObjects[animIdx] && m_AnimObjects[animIdx]->HasMemberAnim(memberIdx))
            {
                isValidAnim = true;
                break;
            }
        }

        // When there are no valid child animations, NULL is returned without changing target
        // 
        if (!isValidAnim)
        {
            return NULL;
        }

        // When normalization of the animation weight is needed, perform it.
        if (m_IsWeightDirty && m_IsWeightNormalizationEnabled)
        {
            NormalizeWeight();
        }

        //----------------------------------------------------------
        // If there is operational space for interpolation it should occur there, and after evacuating target's CONVERTED flag, it should be turned on
        // 
        anim::AnimResult* result = reinterpret_cast<anim::AnimResult*>(target);
        const bool convertedBak = result->IsEnabledFlags(anim::AnimResult::FLAG_CONVERTED);
        result->EnableFlags(anim::AnimResult::FLAG_CONVERTED, true);

        //----------------------------------------------------------
        // Loop for all child animations
        anim::AnimResult workResult;
        float compWeights[anim::AnimResult::MAX_COMPONENTS];
        //for (int animIdx = 0; animIdx < m_AnimObjects.Size(); ++animIdx)
        for (int animIdx = m_AnimObjects.Size() - 1; animIdx >= 0; --animIdx)
        {
            const float childWeight = m_NormalizedWeights[animIdx];
            if (!AnimWeightNearlyEqualZero(childWeight))
            {
                // Reflect the evacuated CONVERTED flag to work
                // If the child is a Blender that handles CONVERTED, expect the CONVERTED results to be returned as-is
                // When the child is an Evaluator, the flag is returned as off but there is no problem for conversion to occur within blendOp
                workResult.EnableFlags(anim::AnimResult::FLAG_CONVERTED, convertedBak);

                // Evaluations
                const anim::AnimResult* childResult = NULL;
                if (m_AnimObjects[animIdx] != NULL)
                {
                    childResult = m_AnimObjects[animIdx]->GetResult(&workResult, memberIdx);
                }

                if (childResult == NULL)
                {
                    NW_ASSERT(GetAnimGroup()->HasOriginalValue());
                    blendOp->ConvertToAnimResult(
                        &workResult,
                        GetAnimGroup()->GetOriginalValue(memberIdx));
                    childResult = &workResult;

                    NW_NULL_ASSERT(childResult);
                }

                if (!blendOp->Blend(result, compWeights, childResult, &childWeight))
                {
                    break;
                }

            }
        }

        //----------------------------------------------------------
        // Restore target's CONVERTED flag
        if (!convertedBak)
        {
            result->DisableFlags(anim::AnimResult::FLAG_CONVERTED);
        }

        //----------------------------------------------------------
        // Execute if there are processes after blending
        if (!convertedBak && blendOp->HasPostBlend())
        {
            blendOp->PostBlend(result, compWeights);
        }
        return result;
    }
}

} // namespace gfx
} // namespace nw
