/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#include "precompiled.h"

#include <nw/gfx/gfx_MaterialActivator.h>
#include <nw/gfx/gfx_MaterialState.h>
#include <nw/gfx/gfx_Model.h>
#include <nn/gx.h>

namespace nw
{
namespace gfx
{

NW_UT_RUNTIME_TYPEINFO_DEFINITION(DirectMaterialActivator, IMaterialActivator);

//----------------------------------------
DirectMaterialActivator*
DirectMaterialActivator::Create(os::IAllocator* allocator)
{
    void* memory = allocator->Alloc(sizeof(DirectMaterialActivator));

    if (memory == NULL)
    {
        return NULL;
    }
    else
    {
        return new(memory) DirectMaterialActivator(allocator);
    }
}

//----------------------------------------
void
DirectMaterialActivator::Activate(RenderContext* renderContext, const Material* material)
{
    NW_NULL_ASSERT(renderContext);
    const ShaderProgram* shaderProgram = renderContext->GetShaderProgram();
    NW_NULL_ASSERT(shaderProgram);

    const Model* owner = material->GetOwnerModel();
    NW_NULL_ASSERT(owner);

    const SceneEnvironment& sceneEnvironment = renderContext->GetSceneEnvironment();

    RenderContext::MaterialHash materialHash = renderContext->GetMaterialHash();

    if (ut::CheckFlag(m_ActivateFlags, Model::FLAG_BUFFER_SHADER_PARAMETER))
    {
        internal::MaterialState::ActivateShaderParameter(shaderProgram, material->GetShaderParameterResMaterial());
        materialHash.shaderParameter = 0x0;
    }

    const ResMaterial resShadingParameterResMaterial = material->GetShadingParameterResMaterial();

    // FragmentLightEnabled must be set.
    internal::MaterialState::ActivateFragmentLightParameters(sceneEnvironment, shaderProgram, resShadingParameterResMaterial);

    if (ut::CheckFlag(m_ActivateFlags, Model::FLAG_BUFFER_SHADING_PARAMETER))
    {
        internal::MaterialState::ActivateShadingParameters(sceneEnvironment, shaderProgram, resShadingParameterResMaterial);
        materialHash.shadingParameter = 0x0;
    }

    const ResMaterial resRasterizationMaterial = material->GetRasterizationResMaterial();
    if (ut::CheckFlag(m_ActivateFlags, Model::FLAG_BUFFER_RASTERIZATION))
    {
        internal::MaterialState::ActivateRasterization(resRasterizationMaterial.GetRasterization());
        materialHash.rasterization = 0x0;
    }

    if (ut::CheckFlag(m_ActivateFlags, Model::FLAG_BUFFER_TEXTURE_COORDINATOR))
    {
        internal::MaterialState::ActivateTextureCoordinators(renderContext, shaderProgram, material->GetTextureCoordinatorResMaterial());
        materialHash.textureCoordinator = 0x0;
    }

    if (ut::CheckFlag(m_ActivateFlags, Model::FLAG_BUFFER_TEXTURE_MAPPER))
    {
        internal::MaterialState::ActivateTextureMappers(material->GetTextureMapperResMaterial());
        materialHash.textureMapper = 0x0;
    }

    const ResMaterial resFragmentLightingResMaterial = material->GetFragmentLightingResMaterial();
    const ResFragmentLighting resFragmentLighting = resFragmentLightingResMaterial.GetFragmentShader().GetFragmentLighting();
    if (ut::CheckFlag(m_ActivateFlags, Model::FLAG_BUFFER_FRAGMENT_LIGHTING))
    {
        internal::MaterialState::ActivateFragmentLighting(sceneEnvironment, resFragmentLighting);
        materialHash.fragmentLighting = 0x0;
    }

    const ResMaterial resFramgnetLightingTableResMaterial = material->GetFragmentLightingTableResMaterial();
    if (ut::CheckFlag(m_ActivateFlags, Model::FLAG_BUFFER_FRAGMENT_LIGHTING_TABLE))
    {
        const ResFragmentLightingTable resFragmentLightingTable = resFramgnetLightingTableResMaterial.GetFragmentShader().GetFragmentLightingTable();
        internal::MaterialState::ActivateFragmentLightingTable(resFragmentLighting, resFragmentLightingTable);
        materialHash.fragmentLightingTable = 0x0;
    }

    const ResMaterial resMaterialColorResMaterial = material->GetMaterialColorResMaterial();
    const ResMaterialColor resMaterialColor = resMaterialColorResMaterial.GetMaterialColor();

    Material* cacheMaterial = renderContext->GetMaterialCache();
    if (ut::CheckFlag(m_ActivateFlags, Model::FLAG_BUFFER_MATERIAL_COLOR))
    {
        bool reflectionEnabled = ut::CheckFlag(resFragmentLighting.GetFlags(), ResFragmentLightingData::FLAG_REFLECTION_ENABLED);
        internal::MaterialState::ActivateMaterialColor(sceneEnvironment, shaderProgram, resMaterialColor, reflectionEnabled);
        materialHash.materialColor = 0x0;
    }

    const ResMaterial resTextureCombinerResMaterial = material->GetTextureCombinerResMaterial();
    if (ut::CheckFlag(m_ActivateFlags, Model::FLAG_BUFFER_TEXTURE_COMBINER))
    {
        const ResFragmentShader resTextureCombinerFragmentShader = resTextureCombinerResMaterial.GetFragmentShader();
        internal::MaterialState::ActivateTextureCombiners(resTextureCombinerFragmentShader, resMaterialColor);
        materialHash.textureCombiner = 0x0;
    }

    const ResMaterial resAlphaTestResMaterial = material->GetAlphaTestResMaterial();
    if (ut::CheckFlag(m_ActivateFlags, Model::FLAG_BUFFER_ALPHA_TEST))
    {
        const ResAlphaTest resAlphaTest = resAlphaTestResMaterial.GetFragmentShader().GetAlphaTest();
        internal::MaterialState::ActivateAlphaTest(resAlphaTest);
        materialHash.alphaTest = 0x0;
    }

    const ResMaterial resFragmentOperationResMaterial = material->GetFragmentOperationResMaterial();
    if (ut::CheckFlag(m_ActivateFlags, Model::FLAG_BUFFER_FRAGMENT_OPERATION))
    {
        const ResFragmentOperation resFragmentOperation = resFragmentOperationResMaterial.GetFragmentOperation();
        internal::MaterialState::ActivateFragmentOperation(resFragmentOperation);
        materialHash.fragmentOperation = 0x0;
    }

    renderContext->SetMaterialHash(materialHash);
}

//----------------------------------------
DirectMaterialActivator::DirectMaterialActivator(os::IAllocator* allocator)
    : IMaterialActivator(allocator),
      m_ActivateFlags(Model::MULTI_FLAG_BUFFER_MATERIAL)
{
}

//----------------------------------------
DirectMaterialActivator::~DirectMaterialActivator()
{
}

} // namespace gfx
} // namespace nw
