/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#include "precompiled.h"

#include <nw/gfx/gfx_SceneUpdater.h>

#include <nw/gfx/gfx_SceneUpdateHelper.h>
#include <nw/gfx/gfx_SceneContext.h>
#include <nw/gfx/gfx_SkeletalModel.h>
#include <nw/gfx/gfx_ParticleModel.h>
#include <nw/gfx/gfx_ParticleSet.h>
#include <nw/gfx/res/gfx_ResShape.h>

#include <nw/dev.h>

#if NN_CURRENT_VERSION_NUMBER >= NN_VERSION_NUMBER(4,0,0,0)
#pragma diag_suppress 1301 // padding inserted in struct.
#pragma diag_suppress 2530 // padding added to end of struct.
#endif
namespace nw
{
namespace gfx
{

NW_UT_RUNTIME_TYPEINFO_DEFINITION(SceneUpdater, ISceneUpdater);

//----------------------------------------
SceneUpdater*
SceneUpdater::Builder::Create(os::IAllocator* allocator)
{
    NW_NULL_ASSERT(allocator);

    if (!this->m_SkeletonUpdater)
    {
        this->m_SkeletonUpdater.Reset(SkeletonUpdater::Builder()
            .Create(allocator));
    }

    if (!this->m_WorldMatrixUpdater)
    {
        this->m_WorldMatrixUpdater.Reset(WorldMatrixUpdater::Builder()
            .Create(allocator));
    }

    if (!this->m_BillboardUpdater)
    {
        this->m_BillboardUpdater.Reset(BillboardUpdater::Create(allocator));
    }

    void* memory = allocator->Alloc(sizeof(SceneUpdater));
    NW_NULL_ASSERT(memory);

    return new(memory) SceneUpdater(
        allocator,
        this->m_SkeletonUpdater,
        this->m_WorldMatrixUpdater,
        this->m_BillboardUpdater);
}

//----------------------------------------
void
SceneUpdater::UpdateAll(SceneContext* sceneContext)
{
    NW_NULL_ASSERT(sceneContext);

    this->EvaluateAnim(sceneContext, anim::ResGraphicsAnimGroup::EVALUATION_BEFORE_WORLD_UPDATE);
    this->UpdateTransformNode(sceneContext);
    // TODO: In the future, render culling will be processed here.
    this->EvaluateAnim(sceneContext, anim::ResGraphicsAnimGroup::EVALUATION_AFTER_SCENE_CULLING);
    this->UpdateSkeletalModel(sceneContext);
    this->UpdateAnim(sceneContext);
}

//----------------------------------------
void
SceneUpdater::SubmitView(
    RenderQueue* renderQueue,
    SceneContext* sceneContext,
    const Camera& camera,
    u8 layerId,
    RenderSortMode renderSortMode)
{
    Model::IsVisibleModelDefaultFunctor isVisibleModel;

    SceneUpdateHelper::SubmitView(
        this->m_SkeletonUpdater.Get(),
        this->m_BillboardUpdater.Get(),
        renderQueue,
        sceneContext,
        camera,
        layerId,
        layerId,
        renderSortMode,
        this->m_DepthSortMode,
        &isVisibleModel);
}

//----------------------------------------
void
SceneUpdater::SubmitView(
    RenderQueue* renderQueue,
    SceneContext* sceneContext,
    const Camera& camera,
    u8 layerId,
    u8 particleLayerId,
    RenderSortMode renderSortMode)
{
    Model::IsVisibleModelDefaultFunctor isVisibleModel;

    SceneUpdateHelper::SubmitView(
        this->m_SkeletonUpdater.Get(),
        this->m_BillboardUpdater.Get(),
        renderQueue,
        sceneContext,
        camera,
        layerId,
        particleLayerId,
        renderSortMode,
        this->m_DepthSortMode,
        &isVisibleModel);
}

//----------------------------------------
void
SceneUpdater::SubmitView(
    RenderQueue* renderQueue,
    SceneContext* sceneContext,
    const Camera& camera,
    u8 layerId,
    u8 particleLayerId,
    SceneUpdater::IsVisibleModelFunctor* isVisibleModel,
    RenderSortMode renderSortMode)
{
    SceneUpdateHelper::SubmitView(
        this->m_SkeletonUpdater.Get(),
        this->m_BillboardUpdater.Get(),
        renderQueue,
        sceneContext,
        camera,
        layerId,
        particleLayerId,
        renderSortMode,
        this->m_DepthSortMode,
        isVisibleModel);
}

//----------------------------------------
void
SceneUpdater::UpdateTransformNode(SceneContext* sceneContext) const
{
    NW_NULL_ASSERT(sceneContext);

    SceneNodeArray::iterator end = sceneContext->GetSceneNodesEnd();
    for (SceneNodeArray::iterator node = sceneContext->GetSceneNodesBegin(); node != end; ++node)
    {
        (*node)->PreUpdateSignal()(*node, sceneContext);

        // Inherits flag from the parent node.
        (*node)->InheritTraversalResults();

        (*node)->UpdateTransform(this->m_WorldMatrixUpdater.Get(), sceneContext);
    }
}

//----------------------------------------
void
SceneUpdater::UpdateSkeletalModel(SceneContext* sceneContext, bool ignoreUpdatedFlag) const
{
    NW_NULL_ASSERT(sceneContext);

    SkeletalModelArray::iterator end = sceneContext->GetSkeletalModelsEnd();

    if (ignoreUpdatedFlag)
    {
        // A skeleton once updated will no longer update without performing a SetUpdated(false).
        // Since SubmitView performs a SetUpdated(false) in the library, the skeleton may not update properly depending on the use.
        // 
        // The update flag exists so that shared skeletons are not updated multiple times, so all flags are cleared at once here.
        // 

        for (SkeletalModelArray::iterator model = sceneContext->GetSkeletalModelsBegin(); model != end; ++model)
        {
            (*model)->GetSkeleton()->SetUpdated(false);
        }
    }

    for (SkeletalModelArray::iterator model = sceneContext->GetSkeletalModelsBegin(); model != end; ++model)
    {
        // Skips processing when using a fully baked animation.
        if ((*model)->GetFullBakedAnimEnabled()) { continue; }

        Skeleton* skeleton = (*model)->GetSkeleton();

        // Skips processing when the skeleton has already been updated.
        if (skeleton->IsUpdated()) { continue; }

        this->m_SkeletonUpdater->UpdateWorld(
            skeleton,
            *(this->m_WorldMatrixUpdater.Get()));

        skeleton->SetUpdated(true);
    }
}

//----------------------------------------
void
SceneUpdater::UpdateAnim(SceneContext* sceneContext) const
{
    NW_NULL_ASSERT(sceneContext);

    std::for_each(
        sceneContext->GetAnimatableNodesBegin(),
        sceneContext->GetAnimatableNodesEnd(),
        std::mem_fun(&SceneNode::UpdateFrame));
}

//----------------------------------------
void
SceneUpdater::EvaluateAnim(SceneContext* sceneContext, anim::ResGraphicsAnimGroup::EvaluationTiming timing) const
{
    NW_NULL_ASSERT(sceneContext);

    AnimatableNodeArray::iterator animatableNodesEnd = sceneContext->GetAnimatableNodesEnd();
    for (AnimatableNodeArray::iterator animNode = sceneContext->GetAnimatableNodesBegin();
        animNode != animatableNodesEnd; ++animNode)
    {
        // A node stored in SceneContext::AnimNodes has AnimBinding.
        (*animNode)->GetAnimBinding()->Evaluate(timing);
    }
}



} // namespace gfx
} // namespace nw
#if NN_CURRENT_VERSION_NUMBER >= NN_VERSION_NUMBER(4,0,0,0)
#pragma diag_default 2530 // padding added to end of struct.
#pragma diag_default 1301 // padding inserted in struct.
#endif
