/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#include "precompiled.h"

#include <nw/anim/anim_AnimBlend.h>
#include <nw/gfx/gfx_AnimGroup.h>
#include <nw/gfx/gfx_CalculatedTransform.h>
#include <nw/gfx/gfx_TransformAnimAdder.h>
#include <nw/gfx/gfx_TransformAnimBlendOp.h>
#include <nw/gfx/gfx_TransformAnimEvaluator.h>

namespace nw
{
namespace gfx
{

NW_UT_RUNTIME_TYPEINFO_DEFINITION(TransformAnimAdder       , AnimAdder);

//----------------------------------------------------------
const anim::AnimResult*
TransformAnimAdder::GetResult(
    void* target,
    int memberIdx
) const
{
    //----------------------------------------------------------
    // Get the blend operation
    const anim::AnimBlendOp* blendOp = m_AnimGroup->GetBlendOperation(memberIdx);

    //----------------------------------------------------------
    // If there is operational space for interpolation it should occur there, and after evacuating target's CONVERTED flag, it should be turned on
    // 
    CalculatedTransform* transform =
        reinterpret_cast<CalculatedTransform*>(target);
    const bit32 flagsBak = transform->GetFlags();
    const bool convertedBak = transform->IsEnabledFlags(CalculatedTransform::FLAG_CONVERTED_FOR_BLEND);
    transform->EnableFlags(CalculatedTransform::FLAG_CONVERTED_FOR_BLEND);

    //----------------------------------------------------------
    // Raise the target's IGNORE flag, and set to unwritten state
    transform->EnableFlags(CalculatedTransform::FLAG_IS_IGNORE_ALL);

    //----------------------------------------------------------
    // Loop for all child animations
    CalculatedTransform workResult;
    bool written = false;
    bool firstRotateFlag = true;
    math::MTX34 firstRotateMtx;
    //for (int animIdx = 0; animIdx < m_AnimObjects.Size(); ++animIdx)
    for (int animIdx = m_AnimObjects.Size() - 1; animIdx >= 0; --animIdx)
    {
        const AnimObject* animObj = m_AnimObjects[animIdx];
        if (animObj == NULL)
        {
            continue;
        }

        const float childWeight = m_Weights[animIdx];
        float srcWeights[3] = { childWeight, childWeight, childWeight };
        TransformAnimEvaluator::DisableSRTWeightsIfNeeded(srcWeights, animObj);

        if (!TransformAnimEvaluator::CheckWeightsNearlyZero(srcWeights))
        {
            // Reflect the evacuated CONVERTED flag to work
            // If the child is a Blender that handles CONVERTED, expect the CONVERTED results to be returned as-is
            // When the child is an Evaluator, the flag is returned as off but there is no problem for conversion to occur within blendOp
            workResult.EnableFlags(CalculatedTransform::FLAG_CONVERTED_FOR_BLEND, convertedBak);

            // Evaluations
            const anim::AnimResult* childResult =
                animObj->GetResult(&workResult, memberIdx);
            if (childResult != NULL)
            {
                written = true;
                const bool evaluatorFlag =
                    ut::DynamicCast<const TransformAnimEvaluator*>(animObj) != NULL;

                // Save the first animation's rotation matrix for when the row of the blend result's rotation matrix is a zero vector
                // 
                if (evaluatorFlag && firstRotateFlag &&
                    !workResult.IsEnabledFlags(CalculatedTransform::FLAG_IS_IGNORE_ROTATE))
                {
                    firstRotateFlag = false;
                    firstRotateMtx = workResult.TransformMatrix();
                }

                // Blend
                if (!blendOp->Blend(reinterpret_cast<anim::AnimResult*>(transform),
                    NULL, childResult, srcWeights))
                {
                    break;
                }
            }
        }
    }

    //----------------------------------------------------------
    // Restore target's CONVERTED flag
    if (!convertedBak)
    {
        transform->DisableFlags(CalculatedTransform::FLAG_CONVERTED_FOR_BLEND);
    }

    if (!written) // No valid child animation
    {
        transform->RestoreFlags(CalculatedTransform::FLAG_IS_IGNORE_ALL, flagsBak);
        return NULL;
    }

    //----------------------------------------------------------
    // Execute if there are processes after blending
    if (!convertedBak && blendOp->HasPostBlend())
    {
        if (!blendOp->PostBlend(reinterpret_cast<anim::AnimResult*>(transform), NULL))
        {
            // If the row of the blend result's rotation matrix is a zero vector, use the first animation's rotation matrix.
            // 
            // When there is a problem with the quality of the results, you must avoid this by slightly shifting the original data.
            // 
            transform->AdjustZeroRotateMatrix(!firstRotateFlag, firstRotateMtx);
        }
    }
    transform->UpdateScaleFlags();
    transform->UpdateRotateFlagsStrictly();
    transform->UpdateTranslateFlags();
    transform->UpdateCompositeFlags();

    // Cast AnimResult to the target and return (use with the parent's blender)
    return reinterpret_cast<anim::AnimResult*>(target);
}

} // namespace gfx
} // namespace nw
