﻿/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#include "precompiled.h"

#include <nw/gfx/gfx_WorldMatrixUpdater.h>
#include <nw/gfx/gfx_CalculatedTransform.h>
#include <nw/ut/ut_MoveArray.h>
#include <nw/math.h>
#include <nw/os/os_Memory.h>

#include <nw/dev.h>

namespace nw {
namespace gfx {

//----------------------------------------
WorldMatrixUpdater*
WorldMatrixUpdater::Builder::Create(
    os::IAllocator* allocator
)
{
    NW_NULL_ASSERT(allocator);

    void* updaterMemory = allocator->Alloc(sizeof(WorldMatrixUpdater));
    NW_NULL_ASSERT(updaterMemory);

    return new(updaterMemory) WorldMatrixUpdater(allocator);
}

//----------------------------------------
WorldMatrixUpdater::WorldMatrixUpdater(
    os::IAllocator* allocator
)
: GfxObject(allocator)
{
    NW_NULL_ASSERT(allocator);
}

//----------------------------------------
WorldMatrixUpdater::~WorldMatrixUpdater()
{
}

//----------------------------------------
void
WorldMatrixUpdater::CalculateWorldXsi(
    math::MTX34* transformMatrix,
    math::VEC3* scale,
    const CalculatedTransform& localTransform,
    const CalculatedTransform& parentWorldTransform,
    const CalculatedTransform& parentLocalTransform
) const
{
    NW_NULL_ASSERT(transformMatrix);
    NW_NULL_ASSERT(scale);

    NW_UNUSED_VARIABLE(parentLocalTransform);

    const math::MTX34& parentMatrix = parentWorldTransform.TransformMatrix();

    if (localTransform.IsEnabledFlagsOr(
        CalculatedTransform::FLAG_IS_IDENTITY | CalculatedTransform::FLAG_IS_ROTATE_TRANSLATE_ZERO))
    {
        // TR(i) = TR(i-1)
        math::MTX34Copy(transformMatrix, parentMatrix);
    }
    else
    {
        bool isParentScaleOne = parentWorldTransform.IsEnabledFlags(CalculatedTransform::FLAG_IS_SCALE_ONE);
        const math::VEC3& parentScale = parentWorldTransform.m_Scale;

        const math::MTX34& localMatrix = localTransform.TransformMatrix();
        math::VEC3 localTranslate = localTransform.TransformMatrix().GetColumn(3);

        if (localTransform.IsEnabledFlags(CalculatedTransform::FLAG_IS_ROTATE_ZERO))
        {
            // TR(i) = TR(i-1) * S(i-1) * t(i) * Rev(S(i-1))
            if (isParentScaleOne)
            {
                // TR(i) = TR(i-1) * t(i)
                math::MTX34MultTranslate(transformMatrix, &parentMatrix, &localTranslate);
            }
            else
            {
                // S(i-1) != 1
                math::VEC3Mult(&localTranslate, &localTranslate, &parentScale);
                math::MTX34MultTranslate(transformMatrix, &parentMatrix, &localTranslate);
            }
        }
        else
        {
            if (isParentScaleOne)
            {
                // TR(i) = TR(i-1) * t(i) * r(i)
                math::MTX34Mult(transformMatrix, &parentMatrix, &localMatrix);
            }
            else
            {
                // TR(i) = TR(i-1) * S(i-1) * t(i) * Rev(S(i-1)) * r(i)
                math::MTX34 scaledLocalMatrix;
                math::MTX34Copy(&scaledLocalMatrix, &localMatrix);
                scaledLocalMatrix.f._03 *= parentScale.x;
                scaledLocalMatrix.f._13 *= parentScale.y;
                scaledLocalMatrix.f._23 *= parentScale.z;
                math::MTX34Mult(transformMatrix, &parentMatrix, &scaledLocalMatrix);
            }
        }
    }

    if (parentWorldTransform.IsEnabledFlags(CalculatedTransform::FLAG_IS_SCALE_ONE))
    {
        scale->Set(localTransform.m_Scale);
    }
    else
    {
        math::VEC3Mult(scale, &parentWorldTransform.m_Scale, &localTransform.m_Scale);
    }
}

//----------------------------------------
void
WorldMatrixUpdater::CalculateWorldMayaSsc(
    math::MTX34* transformMatrix,
    math::VEC3* scale,
    const CalculatedTransform& localTransform,
    const CalculatedTransform& parentWorldTransform,
    const CalculatedTransform& parentLocalTransform
) const
{
    NW_NULL_ASSERT(transformMatrix);
    NW_NULL_ASSERT(scale);

    const math::MTX34& parentMatrix = parentWorldTransform.m_TransformMatrix;

    if (localTransform.IsEnabledFlagsOr(CalculatedTransform::FLAG_IS_IDENTITY | CalculatedTransform::FLAG_IS_ROTATE_TRANSLATE_ZERO))
    {
        // TR(i) = TR(i-1)
        math::MTX34Copy(transformMatrix, parentMatrix);
    }
    else
    {
        const math::MTX34& localMatrix = localTransform.m_TransformMatrix;
        math::VEC3 localTranslate = localTransform.m_TransformMatrix.GetColumn(3);

        if (localTransform.IsEnabledFlags(CalculatedTransform::FLAG_IS_ROTATE_ZERO))
        {
            if (parentLocalTransform.IsEnabledFlags(CalculatedTransform::FLAG_IS_SCALE_ONE))
            {
                // TR(i) = TR(i-1) * t(i)
                math::MTX34MultTranslate(transformMatrix, &parentMatrix, &localTranslate);
            }
            else
            {
                // TR(i) = TR(i-1) * S(i-1) * t(i) * Rev(S(i-1))
                math::MTX34Copy(transformMatrix, parentMatrix);
                math::MTX33 scaledParentRotate(*transformMatrix);
                const math::VEC3& parentScale = parentLocalTransform.m_Scale;
                this->ScaleMatrix(&scaledParentRotate, parentScale);
                math::VEC3Transform(&localTranslate, &scaledParentRotate, &localTranslate);
                this->AddTranslate(transformMatrix, localTranslate);
            }
        }
        else
        {
            if (parentLocalTransform.IsEnabledFlags(CalculatedTransform::FLAG_IS_SCALE_ONE))
            {
                // TR(i) = TR(i-1) * t(i) * r(i)
                math::MTX34MultTranslate(transformMatrix, &parentMatrix, &localTranslate);
                math::MTX33Mult(transformMatrix, transformMatrix, &localMatrix);
            }
            else
            {
                // TR(i) = TR(i-1) * S(i-1) * t(i) * Rev(S(i-1)) * r(i)
                math::MTX34Copy(transformMatrix, parentMatrix);
                math::MTX33 scaledParentRotate(*transformMatrix);
                const math::VEC3& parentScale = parentLocalTransform.m_Scale;
                this->ScaleMatrix(&scaledParentRotate, parentScale);
                math::VEC3Transform(&localTranslate, &scaledParentRotate, &localTranslate);
                this->AddTranslate(transformMatrix, localTranslate);
                math::MTX33Mult(transformMatrix, transformMatrix, &localMatrix);
            }
        }
    }

    if (parentWorldTransform.IsEnabledFlags(CalculatedTransform::FLAG_IS_SCALE_ONE))
    {
        scale->Set(localTransform.m_Scale);
    }
    else
    {
        math::VEC3Mult(scale, &parentWorldTransform.m_Scale, &localTransform.m_Scale);
    }
}

//----------------------------------------
void
WorldMatrixUpdater::CalculateWorldBasic(
    math::MTX34* transformMatrix,
    math::VEC3* scale,
    const CalculatedTransform& localTransform,
    const CalculatedTransform& parentWorldTransform,
    const CalculatedTransform& parentLocalTransform
) const
{
    NW_NULL_ASSERT(transformMatrix);
    NW_NULL_ASSERT(scale);

    const math::MTX34& parentMatrix = parentWorldTransform.m_TransformMatrix;

    if (localTransform.IsEnabledFlagsOr(CalculatedTransform::FLAG_IS_IDENTITY | CalculatedTransform::FLAG_IS_ROTATE_TRANSLATE_ZERO))
    {
        if (parentLocalTransform.IsEnabledFlags(CalculatedTransform::FLAG_IS_SCALE_ONE))
        {
            // TR(i) = TR(i-1)
            math::MTX34Copy(transformMatrix, parentMatrix);
        }
        else
        {
            // TR(i) = TR(i-1) * S(i-1)
            this->MultScale(transformMatrix, &parentMatrix, parentLocalTransform.m_Scale);
            this->CopyTranslate(transformMatrix, parentMatrix);
        }
    }
    else
    {
        if (localTransform.IsEnabledFlags(CalculatedTransform::FLAG_IS_ROTATE_ZERO))
        {
            if (parentLocalTransform.IsEnabledFlags(CalculatedTransform::FLAG_IS_SCALE_ONE))
            {
                math::VEC3 localTranslate = localTransform.m_TransformMatrix.GetColumn(3);

                // TR(i) = TR(i-1) * t(i)
                math::MTX34MultTranslate(transformMatrix, &parentMatrix, &localTranslate);
            }
            else
            {
                math::VEC3 localTranslate = localTransform.m_TransformMatrix.GetColumn(3);

                // TR(i) = TR(i-1) * S(i-1) * t(i)
                this->MultScale(transformMatrix, &parentMatrix, parentLocalTransform.m_Scale);
                this->CopyTranslate(transformMatrix, parentMatrix);
                math::MTX34MultTranslate(transformMatrix, transformMatrix, &localTranslate);
            }
        }
        else
        {
            if (parentLocalTransform.IsEnabledFlags(CalculatedTransform::FLAG_IS_SCALE_ONE))
            {
                math::VEC3 localTranslate = localTransform.m_TransformMatrix.GetColumn(3);

                // TR(i) = TR(i-1) * t(i) * r(i)
                math::MTX34MultTranslate(transformMatrix, &parentMatrix, &localTranslate);
                math::MTX33Mult(transformMatrix, transformMatrix, &localTransform.m_TransformMatrix);
            }
            else
            {
                math::VEC3 localTranslate = localTransform.m_TransformMatrix.GetColumn(3);

                // TR(i) = TR(i-1) * S(i-1) * t(i) * r(i)
                this->MultScale(transformMatrix, &parentMatrix, parentLocalTransform.m_Scale);
                this->CopyTranslate(transformMatrix, parentMatrix);
                math::MTX34MultTranslate(transformMatrix, transformMatrix, &localTranslate);
                math::MTX33Mult(transformMatrix, transformMatrix, &localTransform.m_TransformMatrix);
            }
        }
    }

    if (parentWorldTransform.IsEnabledFlags(CalculatedTransform::FLAG_IS_SCALE_ONE))
    {
        scale->Set(localTransform.m_Scale);
    }
    else
    {
        math::VEC3Mult(scale, &parentWorldTransform.m_Scale, &localTransform.m_Scale);
    }
}

} // namespace gfx
} // namespace nw
