/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

//------------------------------------------------------------------
// Demo: animShare
//
// Overview
//   This is a demo of animation sharing.
//
//   In the animShare.clan file, an animation is set only for Picture_00 pane, and animation sharing settings are performed on this animation so that it can be shared with each pane registered to Group_00 group.
//   
//   
//   All of the picture panes (Picture_00 - Picture_11) are registered in the "Group_00" group.
//   
//
// Controls
//   Moves the crosshairs with the +Control Pad.
//   When the crosshairs come in position over a pane, that pane is enlarged and the animation operates.
//   
//   While enlarged, if the crosshairs move away, the pane shrinks to its original size.
//   
//
//------------------------------------------------------------------

#include <nn/applet.h>
#include <GLES2/gl2.h>
#include <nw/lyt.h>
#include <nw/demo.h>

namespace
{

typedef nw::ut::MoveArray<u8> File;

/*
  
  
  
 
*/
void
InitDraw(int width, int height)
{
    // Color buffer information
    // Switches the width and height to match the LCD direction.
    const nw::font::ColorBufferInfo colBufInfo =
    {
        height, width, PICA_DATA_DEPTH24_STENCIL8_EXT
    };

    const u32 commands[] =
    {
        // Viewport settings
        NW_FONT_CMD_SET_VIEWPORT(0, 0, colBufInfo.width, colBufInfo.height),

        // Disable scissoring
        NW_FONT_CMD_SET_DISABLE_SCISSOR(colBufInfo),

        // Disable W buffer
        // Set the depth range
        // Disable polygon offset
        NW_FONT_CMD_SET_WBUFFER_DEPTHRANGE_POLYGONOFFSET(
            0.0f,           // wScale : W buffer is disabled at 0.0
            0.0f,           // depth range near
            1.0f,           // depth range far
            0,              // polygon offset units : polygon offset is disabled with 0.0
            colBufInfo),
    };

    nngxAdd3DCommand(commands, sizeof(commands), true);

    static const u32 constCommands[] =
    {
        // Clear the cache so frame-buffer access control can be changed
        // 0x111
        1, PICA_CMD_HEADER_SINGLE(PICA_REG_COLOR_DEPTH_BUFFER_CLEAR1),
        // 0x110
        1, PICA_CMD_HEADER_SINGLE(PICA_REG_COLOR_DEPTH_BUFFER_CLEAR0),

        // Disable culling
        NW_FONT_CMD_SET_CULL_FACE(NW_FONT_CMD_CULL_FACE_DISABLE),

        // Disable stencil test
        NW_FONT_CMD_SET_DISABLE_STENCIL_TEST(),

        // Disable depth test
        // Make all elements of the color buffer writable
        NW_FONT_CMD_SET_DEPTH_FUNC_COLOR_MASK(
            false,  // isDepthTestEnabled
            0,      // depthFunc
            true,   // depthMask
            true,   // red
            true,   // Green
            true,   // Blue
            true),  // Alpha

        // Disable early depth test
        NW_FONT_CMD_SET_ENABLE_EARLY_DEPTH_TEST(false),

        // Framebuffer access control
        NW_FONT_CMD_SET_FBACCESS(
            true,   // colorRead
            true,   // colorWrite
            false,  // depthRead
            false,  // depthWrite
            false,  // stencilRead
            false), // stencilWrite
    };

    nngxAdd3DCommand(constCommands, sizeof(constCommands), true);
}

/*
  
  
  
 
*/
void
SetupCamera(nw::lyt::DrawInfo& drawInfo, const nw::lyt::Layout& layout)
{
    nw::ut::Rect layoutRect = layout.GetLayoutRect();

    f32 znear = 0.f;
    f32 zfar = 500.f;

    // Set the projection matrix for an orthogonal projection
    // (Since the layout data is horizontal, change its direction.)
    nw::math::MTX44 projMtx;
    nw::math::MTX44Ortho(
        &projMtx,
        layoutRect.bottom,  // Left
        layoutRect.top,     // Right
        -layoutRect.right,  // Bottom
        -layoutRect.left,   // Top
        znear,
        zfar);
    drawInfo.SetProjectionMtx(projMtx);

    // Set the modelview matrix
    // (Since the layout data is horizontal, the upper part of the screen is the -X direction of the layout.)
    nw::math::VEC3 pos(0, 0, 1);
    nw::math::VEC3 up(-1, 0, 0);
    nw::math::VEC3 target(0, 0, 0);

    nw::math::MTX34 viewMtx;
    nw::math::MTX34LookAt(&viewMtx, &pos, &up, &target);
    drawInfo.SetViewMtx(viewMtx);
}

/*
  
  
  
 

*/
nw::math::VEC2
GetCursorPosition(nw::lyt::Layout& layout)
{
    const nw::ut::Rect& layoutRect = layout.GetLayoutRect();
    const nw::math::VEC2 center(
        (layoutRect.left  + layoutRect.right ) / 2,
        (layoutRect.top   + layoutRect.bottom) / 2
        );

    nw::math::VEC2 curPos;

    // Convert the cursor's coordinate system into the layout's coordinate system.
    const nw::ut::Rect cursorRect(-0.5, 0.5, 0.5, -0.5); // LTRB
    nw::demo::Pad& pad = *nw::demo::PadFactory::GetPad();
    const nw::math::VEC2 stick = pad.GetAnalogStick();
    curPos.x = center.x + stick.x / (cursorRect.right - cursorRect.left  ) * (layoutRect.right - layoutRect.left  );
    curPos.y = center.y + stick.y / (cursorRect.top   - cursorRect.bottom) * (layoutRect.top   - layoutRect.bottom);
    // NW_LOG("curPos %f, %f\n", curPos.x, curPos.y);

    return curPos;
}

/*
  
  
  
 
*/
void
ApplyCursorPosition(nw::lyt::Layout& layout, const nw::math::VEC2& curPos)
{
    nw::lyt::Pane* pCrossBar = layout.GetRootPane()->FindPaneByName("CrossBar", true);
    if (pCrossBar)
    {
        pCrossBar->SetTranslate(curPos);
    }
}

/*
  
  
  
  
  
  
  
  
  
  
 


*/
nw::lyt::AnimationLink*
FindLayoutAnimationLink(
    nw::lyt::Pane*                pPane,
    const nw::lyt::AnimResource&  animRes
)
{
    // search for pane animations
    if (nw::lyt::AnimationLink *const pAnimLink = pPane->FindAnimationLinkSelf(animRes))
    {
        return pAnimLink;
    }

    // Configure the material animation.
    const u8 materialNum = pPane->GetMaterialNum();
    for (u8 i = 0; i < materialNum; ++i)
    {
        if (nw::lyt::AnimationLink *const pAnimLink = pPane->GetMaterial(i)->FindAnimationLink(animRes))
        {
            return pAnimLink;
        }
    }

    // Search recursively.
    for (nw::lyt::PaneList::Iterator it = pPane->GetChildList().GetBeginIter(); it != pPane->GetChildList().GetEndIter(); ++it)
    {
        if (nw::lyt::AnimationLink* pAnimLink = FindLayoutAnimationLink(&(*it), animRes))
        {
            return pAnimLink;
        }
    }

    return NULL;
}

/*
  
  
  
  
  
  
  
 

*/
void
SetAnimationFrame(
    nw::lyt::Pane*                pPane,
    const nw::lyt::AnimResource&  animRes,
    f32                           frame
)
{
    // Configure the pane animation.
    if (nw::lyt::AnimationLink *const pAnimLink = pPane->FindAnimationLinkSelf(animRes))
    {
        pAnimLink->GetAnimTransform()->SetFrame(frame);
    }

    // Configure the material animation.
    const u8 materialNum = pPane->GetMaterialNum();
    for (u8 i = 0; i < materialNum; ++i)
    {
        if (nw::lyt::AnimationLink *const pAnimLink = pPane->GetMaterial(i)->FindAnimationLink(animRes))
        {
            pAnimLink->GetAnimTransform()->SetFrame(frame);
        }
    }

    // Performed recursively.
    for (nw::lyt::PaneList::Iterator it = pPane->GetChildList().GetBeginIter(); it != pPane->GetChildList().GetEndIter(); ++it)
    {
        SetAnimationFrame(&(*it), animRes, frame);
    }
}

/*
  
  
  
  
  
  
 

*/
void
UpdateAnimationFrame(
    nw::lyt::Pane*                pPane,
    const nw::lyt::AnimResource&  animRes,
    bool                          bContain
)
{
    // Searches for nw::lyt::AnimationLink objects to link with the specified animation resource.
    // The first occurrence that is found will be returned.
    nw::lyt::AnimationLink *const pAnimationLink = FindLayoutAnimationLink(pPane, animRes);
    if (! pAnimationLink)
    {
        return;
    }

    // Get nw::lyt::AnimTransform from nw::lyt::AnimationLink and get the current frame value and the maximum frame.
    // 
    nw::lyt::AnimTransform *const pAnimTrans = pAnimationLink->GetAnimTransform();
    f32 animFrame = pAnimTrans->GetFrame();                     // current frame
    const f32 animFrameMax = static_cast<f32>(pAnimTrans->GetFrameSize() - 1);    // maximum frame value

    if (bContain)
    {
        // When pointed to, advance the animation's frame value forward.
        // 
        const f32 step = 2.f;
        if (animFrame < animFrameMax)
        {
            animFrame = nw::ut::Min(animFrameMax, animFrame + step);
            // The animation frames for all panes and materials under pPane are all set to the same value.
            SetAnimationFrame(pPane, animRes, animFrame);
        }
    }
    else
    {
        // When not pointed, restore the animation's frame value.
        // 
        const f32 step = 1.f;
        if (animFrame > 0.f)        // when the current frame is not zero
        {
            animFrame = nw::ut::Max(0.f, animFrame - step);
            // The animation frames for all panes and materials under pPane are all set to the same value.
            SetAnimationFrame(pPane, animRes, animFrame);
        }
    }
}

/*
  
  
  
  
 
*/
void
UpdatePane(
    const nw::math::VEC2&         pos,
    nw::lyt::Group*               pGroup,
    const nw::lyt::AnimResource&  animRes
)
{
    // processing for the panes that are registered in the group
    nw::lyt::PaneLinkList& paneList = pGroup->GetPaneList();
    for (nw::lyt::PaneLinkList::Iterator it = paneList.GetBeginIter(); it != paneList.GetEndIter(); ++it)
    {
        // panes that are registered in the group
        nw::lyt::Pane *const pPane = it->target;

        // Updates the status of animation based on whether the pane is pointed to by the crosshair cursor.
        const bool bContain = nw::lyt::IsContain(pPane, pos);
        UpdateAnimationFrame(
            pPane,
            animRes,
            bContain);
    }
}

}   // Namespace

/*
  
 */
void
nnMain()
{
    nn::applet::Enable();

    nw::demo::SimpleApp& demoApp = nw::demo::SimpleApp::GetInstance();
    demoApp.Initialize();

    nw::lyt::Initialize(&demoApp.GetAllocator(), &demoApp.GetDeviceAllocator());

    nw::lyt::Layout* pLayout = new nw::lyt::Layout();

    // Loads layout binary resources (archives).
    File fileLayout = nw::demo::Utility::LoadFile(
        &demoApp.GetDeviceAllocator(), NW_DEMO_FILE_PATH(L"layout.arc"), 128);

    if (fileLayout.empty())
    {
        NW_FATAL_ERROR("can not open layout archive.\n");
    }

    // Specifies the binary resource root directory and generates the resource assessor.
    nw::lyt::ArcResourceAccessor* pResAccessor = new nw::lyt::ArcResourceAccessor;
    if (!pResAccessor->Attach(fileLayout.begin(), "."))
    {
        NW_FATAL_ERROR("can not attach layout archive.\n");
    }

    // Load the layout resource
    {
        const void* lytRes = pResAccessor->GetResource(0, "animShare.bclyt");
        NW_NULL_ASSERT(lytRes);
        pLayout->Build(lytRes, pResAccessor);
    }

    nw::lyt::AnimResource animRes;
    nw::lyt::AnimTransform* pAnimTrans = 0;
    // Loads the animation resource.
    {
        const void* lpaRes = pResAccessor->GetResource(0, "animShare_focus.bclan");
        NW_NULL_ASSERT(lpaRes);
        animRes.Set(lpaRes);
    }

    // Binds animations considering animation sharing settings and animation interval tag-related groups.
    // 
    pLayout->BindAnimationAuto(animRes, pResAccessor);

    // Animations that were bound using Layout::BindAnimationAuto() are disabled.
    // For this sample, since there is no need to switch the enable/disable of individual animations, enable animations overall.
    // 
    pLayout->GetRootPane()->SetAnimationEnable(animRes, true/*bEnable*/, true/*bRecursive*/);

    // Gets a pointer to the group in which the picture panes are registered.
    nw::lyt::Group *const pGroup_00 = pLayout->GetGroupContainer()->FindGroupByName("Group_00");

    f32 animFrame = 0;

    nw::lyt::GraphicsResource graphicsResource;
    // Loads the global resource file.
    {
        graphicsResource.StartSetup();
        const wchar_t* resourcePath = 0;
        for (int i = 0;
             (resourcePath = graphicsResource.GetResourcePath(i)) != NULL;
             ++i)
        {
            File file = nw::demo::Utility::LoadFile(&demoApp.GetAllocator(), resourcePath);

            if (file.empty())
            {
                NW_FATAL_ERROR("can not open lyt resource file.");
            }

            graphicsResource.SetResource(i, file.begin(), file.size(), false);
        }

        graphicsResource.FinishSetup();
    }

    nw::lyt::DrawInfo drawInfo;
    drawInfo.SetGraphicsResource(&graphicsResource);
    nw::math::VEC2 curPos(0.f, 0.f);

    nw::lyt::Drawer drawer;
    drawer.Initialize(graphicsResource);

    nw::demo::Pad& pad = *nw::demo::PadFactory::GetPad();

    const nw::ut::FloatColor clearColor(0.3f, 0.3f, 0.3f, 1.0f);
    const f32 clearDepth = 0.0f;

    bool loop = true;
    while (loop)
    {
        pad.Update();

        if (pad.IsButtonDown(pad.BUTTON_START))
        {
            loop = false;
        }

        demoApp.SetRenderingTarget(demoApp.DISPLAY0);
        {
            demoApp.GetFrameBufferObject().ClearBuffer(clearColor, clearDepth);

            InitDraw(demoApp.DISPLAY0_WIDTH, demoApp.DISPLAY0_HEIGHT);

            SetupCamera(drawInfo, *pLayout);

            pLayout->CalculateMtx(drawInfo);

            drawer.DrawBegin(drawInfo);
            drawer.Draw(pLayout, drawInfo);
            drawer.DrawEnd(drawInfo);

            curPos = GetCursorPosition(*pLayout);

            // Reflect the cursor position in the pane.
            ApplyCursorPosition(*pLayout, curPos);

            // Convert cursor position from the layout coordinates system to the view coordinates system.
            {
                nw::math::VEC3 tmp(curPos.x, curPos.y, 0.0f);
                nw::math::VEC3Transform(&tmp, &drawInfo.GetViewMtx(), &tmp);
                curPos = nw::math::VEC2(tmp.x, tmp.y);
            }

            // Updates the status of the panes depending on the cursor.
            UpdatePane(curPos, pGroup_00, animRes);

            pLayout->Animate();

        }

        demoApp.SetRenderingTarget(demoApp.DISPLAY1);
        {
            demoApp.GetFrameBufferObject().ClearBuffer(clearColor, clearDepth);
        }

        demoApp.SwapBuffer(demoApp.DISPLAY_BOTH);
    }

    drawer.Finalize();
    graphicsResource.Finalize();
    delete pLayout;
    delete pResAccessor;
}

