/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

//------------------------------------------------------------------
// Demo: bounding
// 
// Overview
//   This demo performs a hit check with a bounding pane.
// 
// Controls
//   Moves the crosshairs with the +Control Pad.
//   
// 
// 
//------------------------------------------------------------------

#include <nn/applet.h>
#include <nw/lyt.h>
#include <nw/demo.h>

namespace
{

typedef nw::ut::MoveArray<u8> File;

/*
  
  
  
 
*/
void
InitDraw(int width, int height)
{
    // Color buffer information
    // Switches the width and height to match the LCD direction.
    const nw::font::ColorBufferInfo colBufInfo =
    {
        height, width, PICA_DATA_DEPTH24_STENCIL8_EXT
    };

    const u32 commands[] =
    {
        // Viewport settings
        NW_FONT_CMD_SET_VIEWPORT(0, 0, colBufInfo.width, colBufInfo.height),

        // Disable scissoring
        NW_FONT_CMD_SET_DISABLE_SCISSOR(colBufInfo),

        // Disable W buffer
        // Set the depth range
        // Disable polygon offset
        NW_FONT_CMD_SET_WBUFFER_DEPTHRANGE_POLYGONOFFSET(
            0.0f,           // wScale : W buffer is disabled at 0.0
            0.0f,           // depth range near
            1.0f,           // depth range far
            0,              // polygon offset units : polygon offset is disabled with 0.0
            colBufInfo),
    };

    nngxAdd3DCommand(commands, sizeof(commands), true);

    static const u32 constCommands[] =
    {
        // Clear the cache so frame-buffer access control can be changed
        // 0x111
        1, PICA_CMD_HEADER_SINGLE(PICA_REG_COLOR_DEPTH_BUFFER_CLEAR1),
        // 0x110
        1, PICA_CMD_HEADER_SINGLE(PICA_REG_COLOR_DEPTH_BUFFER_CLEAR0),

        // Disable culling
        NW_FONT_CMD_SET_CULL_FACE(NW_FONT_CMD_CULL_FACE_DISABLE),

        // Disable stencil test
        NW_FONT_CMD_SET_DISABLE_STENCIL_TEST(),

        // Disable depth test
        // Make all elements of the color buffer writable
        NW_FONT_CMD_SET_DEPTH_FUNC_COLOR_MASK(
            false,  // isDepthTestEnabled
            0,      // depthFunc
            true,   // depthMask
            true,   // red
            true,   // Green
            true,   // Blue
            true),  // Alpha

        // Disable early depth test
        NW_FONT_CMD_SET_ENABLE_EARLY_DEPTH_TEST(false),

        // Framebuffer access control
        NW_FONT_CMD_SET_FBACCESS(
            true,   // colorRead
            true,   // colorWrite
            false,  // depthRead
            false,  // depthWrite
            false,  // stencilRead
            false), // stencilWrite
    };

    nngxAdd3DCommand(constCommands, sizeof(constCommands), true);
}

/*
  
  
  
 
*/
void
SetupCamera(nw::lyt::DrawInfo& drawInfo, const nw::lyt::Layout& layout)
{
    nw::ut::Rect layoutRect = layout.GetLayoutRect();

    f32 znear = 0.f;
    f32 zfar = 500.f;

    // Set the projection matrix for an orthogonal projection
    // (Since the layout data is horizontal, change its direction.)
    nw::math::MTX44 projMtx;
    nw::math::MTX44Ortho(
        &projMtx,
        layoutRect.bottom,  // Left
        layoutRect.top,     // Right
        -layoutRect.right,  // Bottom
        -layoutRect.left,   // Top
        znear,
        zfar);
    drawInfo.SetProjectionMtx(projMtx);

    // Set the modelview matrix
    // (Since the layout data is horizontal, the upper part of the screen is the -X direction of the layout.)
    nw::math::VEC3 pos(0, 0, 1);
    nw::math::VEC3 up(-1, 0, 0);
    nw::math::VEC3 target(0, 0, 0);

    nw::math::MTX34 viewMtx;
    nw::math::MTX34LookAt(&viewMtx, &pos, &up, &target);
    drawInfo.SetViewMtx(viewMtx);
}

/*
  
  
  
 

*/
const nw::math::VEC2
GetCursorPosition(nw::lyt::Layout* pLayout)
{
    const nw::ut::Rect& layoutRect = pLayout->GetLayoutRect();
    const nw::math::VEC2 center(
        (layoutRect.left  + layoutRect.right ) / 2,
        (layoutRect.top   + layoutRect.bottom) / 2
        );

    nw::math::VEC2 curPos;

    // Convert the cursor's coordinate system into the layout's coordinate system.
    const nw::ut::Rect cursorRect(-0.5, 0.5, 0.5, -0.5); // LTRB
    nw::demo::Pad& pad = *nw::demo::PadFactory::GetPad();
    const nw::math::VEC2 stick = pad.GetAnalogStick();
    curPos.x = center.x + stick.x / (cursorRect.right - cursorRect.left  ) * (layoutRect.right - layoutRect.left  );
    curPos.y = center.y + stick.y / (cursorRect.top   - cursorRect.bottom) * (layoutRect.top   - layoutRect.bottom);
    // NW_LOG("curPos %f, %f\n", curPos.x, curPos.y);
    return curPos;
}

/*
  
  
  
 
*/
void
ApplyCursorPosition(nw::lyt::Layout* pLayout, const nw::math::VEC2& curPos)
{
    nw::lyt::Pane* pCrossBar = pLayout->GetRootPane()->FindPaneByName("CrossBar", true);
    if (pCrossBar)
    {
        pCrossBar->SetTranslate(curPos);
    }
}

}   // Namespace

/*
  
 */
void
nnMain()
{
    nn::applet::Enable();

    nw::demo::SimpleApp& demoApp = nw::demo::SimpleApp::GetInstance();
    demoApp.Initialize();

    nw::lyt::Initialize(&demoApp.GetAllocator(), &demoApp.GetDeviceAllocator());

    nw::lyt::Layout* pLayout = new nw::lyt::Layout();

    // Loads layout binary resources (archives).
    File fileLayout = nw::demo::Utility::LoadFile(
        &demoApp.GetDeviceAllocator(), NW_DEMO_FILE_PATH(L"layout.arc"), 128);

    if (fileLayout.empty())
    {
        NW_FATAL_ERROR("can not open layout archive.\n");
    }

    // Specifies the binary resource root directory and generates the resource assessor.
    nw::lyt::ArcResourceAccessor* pResAccessor = new nw::lyt::ArcResourceAccessor;
    if (!pResAccessor->Attach(fileLayout.begin(), "."))
    {
        NW_FATAL_ERROR("can not attach layout archive.\n");
    }

    // Load the layout resource
    {
        const void* lytRes = pResAccessor->GetResource(0, "bounding.bclyt");
        NW_NULL_ASSERT(lytRes);
        pLayout->Build(lytRes, pResAccessor);
    }

    const int BoundPaneNum = 6;

    static const char *const paneNames[] = { "picture_0", "picture_1", "picture_2", "picture_3", "picture_4", "picture_5" };

    enum
    {
        ANIMTYPE_UNSELECT,
        ANIMTYPE_SELECT,

        ANIMTYPE_MAX
    };

    nw::lyt::AnimTransform* pAnimTrans[ANIMTYPE_MAX];

    const void* rlanRes = pResAccessor->GetResource(0, "bounding_unselect.bclan");
    NW_NULL_ASSERT(rlanRes);
    pAnimTrans[ANIMTYPE_UNSELECT] = pLayout->CreateAnimTransform(rlanRes, pResAccessor);
    rlanRes = pResAccessor->GetResource(0, "bounding_select.bclan");
    NW_NULL_ASSERT(rlanRes);
    pAnimTrans[ANIMTYPE_SELECT] = pLayout->CreateAnimTransform(rlanRes, pResAccessor);

    for (int i = 0; i < BoundPaneNum; ++i)
    {
        // Set an animation for a specific pane (and its children)
        if (nw::lyt::Pane* pPane = pLayout->GetRootPane()->FindPaneByName(paneNames[i]))
        {
            pPane->BindAnimation(pAnimTrans[ANIMTYPE_UNSELECT]);
            pPane->BindAnimation(pAnimTrans[ANIMTYPE_SELECT]);
        }
    }

    f32 animFrame[ANIMTYPE_MAX];
    for (int i = 0; i < ANIMTYPE_MAX; ++i)
    {
        animFrame[i] = 0.f;
    }

    nw::lyt::GraphicsResource graphicsResource;
    // Loads the global resource file.
    {
        graphicsResource.StartSetup();
        const wchar_t* resourcePath = 0;
        for (int i = 0;
             (resourcePath = graphicsResource.GetResourcePath(i)) != NULL;
             ++i)
        {
            File file = nw::demo::Utility::LoadFile(&demoApp.GetAllocator(), resourcePath);

            if (file.empty())
            {
                NW_FATAL_ERROR("can not read lyt resource file.");
            }

            graphicsResource.SetResource(i, file.begin(), file.size(), false);
        }

        graphicsResource.FinishSetup();
    }

    nw::lyt::DrawInfo drawInfo;
    drawInfo.SetGraphicsResource(&graphicsResource);
#if ! defined(NW_RELEASE)
    drawInfo.SetDebugDrawMode(true);
#endif

    nw::lyt::Drawer drawer;
    drawer.Initialize(graphicsResource);

    nw::demo::Pad& pad = *nw::demo::PadFactory::GetPad();

    nw::lyt::Pane* pHitPane = 0;
    nw::math::VEC2 curPos(0.f, 0.f);

    const nw::ut::FloatColor clearColor(0.3f, 0.3f, 0.3f, 1.0f);
    const f32 clearDepth = 0.0f;

    bool loop = true;
    while (loop)
    {
        pad.Update();

        if (pad.IsButtonDown(pad.BUTTON_START))
        {
            loop = false;
        }

        demoApp.SetRenderingTarget(demoApp.DISPLAY0);
        {
            demoApp.GetFrameBufferObject().ClearBuffer(clearColor, clearDepth);

            InitDraw(demoApp.DISPLAY0_WIDTH, demoApp.DISPLAY0_HEIGHT);

            SetupCamera(drawInfo, *pLayout);

            // Child animation
            for (nw::lyt::PaneList::Iterator it = pLayout->GetRootPane()->GetChildList().GetBeginIter(); it != pLayout->GetRootPane()->GetChildList().GetEndIter(); ++it)
            {
                // Enables the selected animation for the parent pane of the hit pane (picture pane).
                // Enables the unselected animation for the other parent panes.
                const bool bSelected = pHitPane && &(*it) == pHitPane->GetParent();
                it->SetAnimationEnable(pAnimTrans[ANIMTYPE_UNSELECT], ! bSelected);
                it->SetAnimationEnable(pAnimTrans[ANIMTYPE_SELECT]  ,   bSelected);
            }

            // Frame update.
            for (int i = 0; i < ANIMTYPE_MAX; ++i)
            {
                animFrame[i] += 1.0f;
                while (animFrame[i] >= pAnimTrans[i]->GetFrameSize())
                {
                    animFrame[i] -= pAnimTrans[i]->GetFrameSize();
                }
                pAnimTrans[i]->SetFrame(animFrame[i]);
            }

            pLayout->Animate();
            pLayout->CalculateMtx(drawInfo);

            curPos = GetCursorPosition(pLayout);

            ApplyCursorPosition(pLayout, curPos);

            // Convert cursor position from the layout coordinates system to the view coordinates system.
            {
                nw::math::VEC3 tmp(curPos.x, curPos.y, 0.0f);
                nw::math::VEC3Transform(&tmp, &drawInfo.GetViewMtx(), &tmp);
                curPos = nw::math::VEC2(tmp.x, tmp.y);
            }

            // Perform this after CalculateMtx() to use the matrix.
            pHitPane = nw::lyt::FindHitPane(pLayout, curPos);

            drawer.DrawBegin(drawInfo);
            drawer.Draw(pLayout, drawInfo);
            drawer.DrawEnd(drawInfo);
        }

        demoApp.SetRenderingTarget(demoApp.DISPLAY1);
        {
            demoApp.GetFrameBufferObject().ClearBuffer(clearColor, clearDepth);
        }

        demoApp.SwapBuffer(demoApp.DISPLAY_BOTH);
    }

    drawer.Finalize();
    graphicsResource.Finalize();

    delete pLayout;

    delete pResAccessor;
}

