/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

//------------------------------------------------------------------
// Demo: dblCmdBuf
// 
// Overview
//   This is a demo that sets the command buffer used for character string rendering to double buffering and is called by a command jump.
//   
// 
// 
// When using command jump, double buffering of the command buffer is required for executing CPU and GPU render processing in parallel.
// 
// 
// When command jump is used, it has the following affects:
// 
// (1) Copying the command buffer retained by DispStringBuffer to the current command list every time rendering takes place is no longer required
//    
// 
// (2) The command buffer retained by DispStringBuffer must undergo a cache flush each time a character string changes.
//    
// 
// (3) An explicit cache flush of the 3D command sequences accumulated in the current command list is required.
//     
//     Although the cache flush itself is a process that always takes place, the thread that carries out the process changes to a user thread.
//     
// 
// If there is a low frequency of changes in the character string to be rendered, the advantageous effects of (1) exceed the cost of (2), reducing the processing load.
// 
// 
//------------------------------------------------------------------

#define USE_DOUBLE_COMMAND_BUFFER

#include <GLES2/gl2.h>
#include <cctype>

#include <nn.h>
#include <nn/fs.h>
#include <nn/applet.h>
#include <nw/sdk.h>
#include <nw/demo.h>
#include <nw/lyt.h>

namespace
{

//
// Specify the double buffer mode through a TextBox constructor call.
//
class MyLayout : public nw::lyt::Layout
{
protected:
    virtual nw::lyt::Pane* BuildPaneObj(
        s32 kind,
        const void* dataPtr,
        const nw::lyt::ResBlockSet& resBlockSet)
    {
        switch (kind)
        {
#if defined(USE_DOUBLE_COMMAND_BUFFER)
        case nw::lyt::res::DATABLOCKKIND_TEXTBOX:
            {
                const nw::lyt::res::TextBox* pBlock =
                    static_cast<const nw::lyt::res::TextBox*>(dataPtr);
                return NewObj<nw::lyt::TextBox>(
                    pBlock,
                    resBlockSet,
                    nw::font::DispStringBuffer::CBM_DOUBLE_BUFFER);
            }
#endif

        default:
            return nw::lyt::Layout::BuildPaneObj(kind, dataPtr, resBlockSet);
        }
    }
};

/*
  
  
  
 
*/
void
InitDraw(int width, int height)
{
    // Color buffer information
    // Switches the width and height to match the LCD direction.
    const nw::font::ColorBufferInfo colBufInfo =
    {
        height, width, PICA_DATA_DEPTH24_STENCIL8_EXT
    };

    const u32 commands[] =
    {
        // Viewport settings
        NW_FONT_CMD_SET_VIEWPORT(0, 0, colBufInfo.width, colBufInfo.height),

        // Disable scissoring
        NW_FONT_CMD_SET_DISABLE_SCISSOR(colBufInfo),

        // Disable W buffer
        // Set the depth range
        // Disable polygon offset
        NW_FONT_CMD_SET_WBUFFER_DEPTHRANGE_POLYGONOFFSET(
            0.0f,           // wScale : W buffer is disabled at 0.0
            0.0f,           // depth range near
            1.0f,           // depth range far
            0,              // polygon offset units : polygon offset is disabled with 0.0
            colBufInfo),
    };

    nngxAdd3DCommand(commands, sizeof(commands), true);

    static const u32 constCommands[] =
    {
        // Clear the cache so frame-buffer access control can be changed
        // 0x111
        1, PICA_CMD_HEADER_SINGLE(PICA_REG_COLOR_DEPTH_BUFFER_CLEAR1),
        // 0x110
        1, PICA_CMD_HEADER_SINGLE(PICA_REG_COLOR_DEPTH_BUFFER_CLEAR0),

        // Disable culling
        NW_FONT_CMD_SET_CULL_FACE(NW_FONT_CMD_CULL_FACE_DISABLE),

        // Disable stencil test
        NW_FONT_CMD_SET_DISABLE_STENCIL_TEST(),

        // Disable depth test
        // Make all elements of the color buffer writable
        NW_FONT_CMD_SET_DEPTH_FUNC_COLOR_MASK(
            false,  // isDepthTestEnabled
            0,      // depthFunc
            true,   // depthMask
            true,   // red
            true,   // Green
            true,   // Blue
            true),  // Alpha

        // Disable early depth test
        NW_FONT_CMD_SET_ENABLE_EARLY_DEPTH_TEST(false),

        // Framebuffer access control
        NW_FONT_CMD_SET_FBACCESS(
            true,   // colorRead
            true,   // colorWrite
            false,  // depthRead
            false,  // depthWrite
            false,  // stencilRead
            false), // stencilWrite
    };

    nngxAdd3DCommand(constCommands, sizeof(constCommands), true);
}

/*
  
 */
class File
{
public:
    //
    //
    //
    File()
      : m_Buffer(NULL), m_Size(0)
    {
    }

    //
    //
    //
    //
    //
    //
    //
    //
    bool Read(const wchar_t* fileName, nw::os::IAllocator& allocator)
    {
        NW_ASSERT(m_Buffer == NULL);

        nn::fs::FileReader fileReader(fileName);
        if (fileReader.GetSize() <= 0)
        {
            return false;
        }

        this->m_Size = static_cast<u32>(fileReader.GetSize());
        this->m_Buffer = static_cast<u8*>(allocator.Alloc(this->m_Size, 128));
        if (this->m_Buffer == NULL)
        {
            this->m_Size = 0;
            return false;
        }

        fileReader.Read(this->m_Buffer, this->m_Size);
        fileReader.Finalize();

        return true;
    }

    //
    //
    //
    //
    //
    u8* Buffer()
    {
        return this->m_Buffer;
    }

    //
    //
    //
    //
    //
    u32 Size()
    {
        return this->m_Size;
    }

private:
    u8* m_Buffer;
    u32 m_Size;
};

/*
  
  
  
 
*/
void
SetupCamera(nw::lyt::DrawInfo& drawInfo, const nw::lyt::Layout& layout)
{
    nw::ut::Rect layoutRect = layout.GetLayoutRect();

    f32 znear = 0.f;
    f32 zfar = 500.f;

    // Set the projection matrix for an orthogonal projection
    // (Since the layout data is horizontal, change its direction.)
    nw::math::MTX44 projMtx;
    nw::math::MTX44OrthoPivot(
        &projMtx,
        layoutRect.left,   // Left
        layoutRect.right,  // Right
        layoutRect.bottom, // Bottom
        layoutRect.top,    // Top
        znear,
        zfar,
        nw::math::PIVOT_UPSIDE_TO_TOP);
    drawInfo.SetProjectionMtx(projMtx);

    // Set the modelview matrix
    // (Since the layout data is horizontal, the upper part of the screen is the -X direction of the layout.)
    nw::math::VEC3 pos(0, 0, 1);
    nw::math::VEC3 up(0, 1, 0);
    nw::math::VEC3 target(0, 0, 0);

    nw::math::MTX34 viewMtx;
    nw::math::MTX34LookAt(&viewMtx, &pos, &up, &target);
    drawInfo.SetViewMtx(viewMtx);
}

}   // namespace {anonymous}

/*
  
 */
void
nnMain()
{
    nn::applet::Enable();

    // Initializes the base library.
    nn::fs::Initialize();

    // Initialize applications
    nw::demo::SimpleApp& demoApp = nw::demo::SimpleApp::GetInstance();
    demoApp.Initialize();

    // Initializes the layout library.
    nw::lyt::Initialize(&demoApp.GetAllocator(), &demoApp.GetDeviceAllocator());

    nw::lyt::Layout* pLayout = new MyLayout();

    // Loads layout binary resources (archives).
    File fileLayout;
    if (!fileLayout.Read(NW_DEMO_FILE_PATH(L"layout.arc"), demoApp.GetDeviceAllocator()))
    {
        NW_FATAL_ERROR("can not open layout archive.\n");
    }

    // Specifies the binary resource root directory and generates the resource assessor.
    nw::lyt::ArcResourceAccessor* pResAccessor = new nw::lyt::ArcResourceAccessor;
    if (!pResAccessor->Attach(fileLayout.Buffer(), "."))
    {
        NW_FATAL_ERROR("can not attach layout archive.\n");
    }

    // Load the layout resource
    {
        const void* lytRes = pResAccessor->GetResource(0, "dblCmdBuf.bclyt");
        NW_NULL_ASSERT(lytRes);
        pLayout->Build(lytRes, pResAccessor);
    }

    nw::lyt::AnimTransform* pAnimTrans = 0;
    // Loads the animation resource.
    {
        const void* lpaRes = pResAccessor->GetResource(0, "dblCmdBuf.bclan");
        if (lpaRes)
        {
            pAnimTrans = pLayout->CreateAnimTransform(lpaRes, pResAccessor);
            pLayout->BindAnimation(pAnimTrans);
        }
    }

    f32 animFrame = 0;

    nw::lyt::GraphicsResource graphicsResource;

    // Loads the global resource file.
    {
        graphicsResource.StartSetup();
        const wchar_t* resourcePath = 0;
        for (int i = 0;
             (resourcePath = graphicsResource.GetResourcePath(i)) != NULL;
             ++i)
        {
            File file;
            if (!file.Read(resourcePath, demoApp.GetAllocator()))
            {
                NW_FATAL_ERROR("can not read lyt resource file.");
            }
            graphicsResource.SetResource(i, file.Buffer(), file.Size());
        }
        graphicsResource.FinishSetup();
    }

    nw::lyt::DrawInfo drawInfo;
    drawInfo.SetGraphicsResource(&graphicsResource);

    nw::lyt::Drawer drawer;
    drawer.Initialize(graphicsResource);

    nw::demo::Pad& pad = *nw::demo::PadFactory::GetPad();

    const nw::ut::FloatColor clearColor(0.3f, 0.3f, 0.3f, 1.0f);
    const f32 clearDepth = 0.0f;

    // This sample makes double use of the command list through the demo library features and it executes in parallel the issuing of rendering commands from CPU and rendering from GPU.
    // 
    // 
    // If demoApp.SwapBuffer is called at the end of a frame, the current command list changes and enters the stopped mode.
    // 
    // 
    // 
    // The first frame is put in the stopped mode here.
    nngxStopCmdlist();

    const int NUM = 10;

    nw::ut::MoveArray<wchar_t> strBuf;
    nw::lyt::Pane* pWindow[NUM];
    nw::lyt::TextBox* pTextBox[NUM];

    for (int i = 0; i < NUM; ++i)
    {
        char name[11];
        nw::ut::snprintf(name, sizeof(name), sizeof(name) - 1, "window_%02d", i);
        pWindow[i] = pLayout->GetRootPane()->FindPaneByName(name);
        nw::ut::snprintf(name, sizeof(name), sizeof(name) - 1, "textBox_%02d", i);
        pTextBox[i] = nw::ut::DynamicCast<nw::lyt::TextBox*>(
            pLayout->GetRootPane()->FindPaneByName(name));
    }

    if (pTextBox[0])
    {
        u32 strLen = pTextBox[0]->GetStringLength() + 1;
        strBuf = nw::ut::MoveArray<wchar_t>(strLen, &demoApp.GetAllocator());
        strBuf.Resize(strLen);
    }

#if defined(USE_DOUBLE_COMMAND_BUFFER)
    bool enableDoubleBuffer = true;
#else
    bool enableDoubleBuffer = false;
#endif
    u32 numVisible = NUM;
    u32 frameCount = 0;
    nn::os::Tick drawTime;
    nn::os::Tick flushTime;
    u32 cmdBufLen = 0;
    bool mainloop = true;
    while (mainloop)
    {
        pad.Update();

        if (pad.IsButtonDown(pad.BUTTON_START))
        {
            mainloop = false;
        }

#if defined(USE_DOUBLE_COMMAND_BUFFER)
        if (pad.IsButtonDown(pad.BUTTON_A))
        {
            enableDoubleBuffer = !enableDoubleBuffer;

            // If double buffering has been disabled, waits for the current GPU rendering to complete.
            // 
            if (!enableDoubleBuffer)
            {
                demoApp.m_pRenderSystem->WaitCommandList();
            }

            // Switches the DispStringBuffer mode.
            for (int i = 0; i < NUM; ++i)
            {
                if (pTextBox[i])
                {
                    if (enableDoubleBuffer)
                    {
                        pTextBox[i]->GetDispStringBuffer()->SetCommandBufferMode(
                            nw::font::DispStringBuffer::CBM_DOUBLE_BUFFER);
                    }
                    else
                    {
                        pTextBox[i]->GetDispStringBuffer()->SetCommandBufferMode(
                            nw::font::DispStringBuffer::CBM_DYNAMIC);
                    }
                }
            }
        }
#endif

        {
            u32 prevNumVisible = numVisible;

            if (pad.IsButtonDown(pad.BUTTON_X) && numVisible < NUM)
            {
                ++numVisible;
            }

            if (pad.IsButtonDown(pad.BUTTON_Y) && 1 < numVisible)
            {
                --numVisible;
            }

            if (prevNumVisible != numVisible)
            {
                for (int i = 0; i < NUM; ++i)
                {
                    if (pWindow[i])
                    {
                        pWindow[i]->SetVisible(i < numVisible);
                    }
                }
            }
        }

        demoApp.SetRenderingTarget(demoApp.DISPLAY0);
        {
            // Changes the character string to display.
            if (frameCount == 60)
            {
                u32 size = strBuf.Size();

                u32 len = nw::ut::snwprintf(
                    strBuf.begin(),
                    size,
                    size - 1,
                    L"num of win = %d         <X:+ Y:->\n"
                    L"dbl cmd buff = %s       <A>\n"
                    L"cpu draw = %lld [usec]\n"
                    L"cache flush = %lld [usec]\n"
                    L"cmd list len = %d [byte]\n\n",
                    numVisible,
                    enableDoubleBuffer? "on" : "off",
                    nn::os::Tick(drawTime / frameCount).ToTimeSpan().GetMicroSeconds(),
                    nn::os::Tick(flushTime / frameCount).ToTimeSpan().GetMicroSeconds(),
                    cmdBufLen);

                for (int i = 0; len + i < size - 1; ++i)
                {
                    int c = (i % 127) + 1;
                    if (!std::isgraph(c))
                    {
                        c = ' ';
                    }

                    strBuf[len + i] = c;
                }

                strBuf[size - 1] = '\0';

                for (int i = 0; i < NUM; ++i)
                {
                    pTextBox[i]->SetString(strBuf.begin());
                }

                frameCount = 0;
                drawTime = nn::os::Tick(0);
                flushTime = nn::os::Tick(0);
            }

            demoApp.GetFrameBufferObject().ClearBuffer(clearColor, clearDepth);

            SetupCamera(drawInfo, *pLayout);

            InitDraw(demoApp.DISPLAY0_WIDTH, demoApp.DISPLAY0_HEIGHT);
            pLayout->Animate();
            pLayout->CalculateMtx(drawInfo);

            // Gets start address of buffer of unissued commands.
            // Used later to flush the command buffer cache.
            nngxFlush3DCommand();
            GLint cmdStartAddr = 0;
            GLint cmdEndAddr = 0;
            nngxGetCmdlistParameteri(
                NN_GX_CMDLIST_CURRENT_BUFADDR, &cmdStartAddr);

            nn::os::Tick startDrawTime = nn::os::Tick::GetSystemCurrent();

            drawer.DrawBegin(drawInfo);
            drawer.Draw(pLayout, drawInfo);
            drawer.DrawEnd(drawInfo);

#if defined(USE_DOUBLE_COMMAND_BUFFER)
            // The command list must be in the stopped state when calling nngxFlush3DCommandPartially.
            // 
            // 
            // However, the current command list is actually always in the stopped state because this demo uses the double command list.
            // 
            GLint isRunning = 0;
            nngxGetCmdlistParameteri(NN_GX_CMDLIST_IS_RUNNING, &isRunning);
            NW_ASSERT(!isRunning); // Actually, always in the stopped state.
            if (isRunning)
            {
                nngxWaitCmdlistDone();
                nngxStopCmdlist();
            }

            GLint cmdFirstReturnAddr =
                (GLint)drawer.GetCommandJumpFirstReturnAddr();
            if (cmdFirstReturnAddr != 0)
            {
                // When using a command jump.
                nngxFlush3DCommandPartially(cmdFirstReturnAddr - cmdStartAddr);
                NW_GL_ASSERT();
            }
            else
            {
                // When not using a command jump.
                nngxFlush3DCommandNoCacheFlush();
                NW_GL_ASSERT();
            }

            nngxGetCmdlistParameteri(NN_GX_CMDLIST_CURRENT_BUFADDR, &cmdEndAddr);

            if (cmdFirstReturnAddr != 0)
            {
                // Completes the last jump command.
                // The last jump command is not completed until the 3D command is issued and the entire length of the 3D command sequence is confirmed.
                // 
                drawer.FixCommandJump(reinterpret_cast<void*>(cmdEndAddr));
            }

            drawTime += nn::os::Tick::GetSystemCurrent() - startDrawTime;

            // Must flush the cache itself.
            nn::os::Tick flushStartTick = nn::os::Tick::GetSystemCurrent();
            nngxUpdateBuffer((void*)cmdStartAddr, cmdEndAddr - cmdStartAddr);
            flushTime += nn::os::Tick::GetSystemCurrent() - flushStartTick;

            // Returns the command list execution status to what it was.
            if (isRunning)
            {
                nngxRunCmdlist();
            }
#endif

            // When the double command buffer has been enabled, the command list 3D command sequence is shortened so that command jump can be used.
            // 
            // 
            nngxGetCmdlistParameteri(
                NN_GX_CMDLIST_CURRENT_BUFADDR, &cmdEndAddr);
            cmdBufLen = cmdEndAddr - cmdStartAddr;

            ++frameCount;
        }

        demoApp.SetRenderingTarget(demoApp.DISPLAY1);
        {
            demoApp.GetFrameBufferObject().ClearBuffer(clearColor, clearDepth);
        }

        // Frame update.
        if (pAnimTrans)
        {
            animFrame += 1.0f;
            while (animFrame >= pAnimTrans->GetFrameSize())
            {
                animFrame -= pAnimTrans->GetFrameSize();
            }
            pAnimTrans->SetFrame(animFrame);
        }

        demoApp.SwapBuffer(demoApp.DISPLAY_BOTH);
    }

    // Returns the DispStringBuffer mode to what it originally was so that it will be deleted by the appropriate reallocator.
    for (int i = 0; i < NUM; ++i)
    {
        if (pTextBox[i])
        {
            pTextBox[i]->GetDispStringBuffer()->SetCommandBufferMode(
                nw::font::DispStringBuffer::CBM_DOUBLE_BUFFER);
        }
    }

    drawer.Finalize();
    graphicsResource.Finalize();
    delete pLayout;
    delete pResAccessor;

    demoApp.GetDeviceAllocator().Free(fileLayout.Buffer());
}

