/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

//------------------------------------------------------------------
// Demo: multiArc
// 
// Overview
//   This demo loads multiple archives as layout resources.
//   There are 2 layouts, layout1 and layout 2, and the resources are split into 3 archives.
// 
// 
//      layout1.arc: the layout and texture for layout1
//      layout2.arc: the layout and texture for layout2
//      layout3.arc: the jointly used font and window textures
// 
//------------------------------------------------------------------


#include <nn/applet.h>
#include <nw/lyt.h>
#include <nw/ut.h>
#include <nw/demo.h>

namespace
{

typedef nw::ut::MoveArray<u8> File;

/*
  
  
  
 
*/
void
InitDraw(int width, int height)
{
    // Color buffer information
    // Switches the width and height to match the LCD direction.
    const nw::font::ColorBufferInfo colBufInfo =
    {
        height, width, PICA_DATA_DEPTH24_STENCIL8_EXT
    };

    const u32 commands[] =
    {
        // Viewport settings
        NW_FONT_CMD_SET_VIEWPORT(0, 0, colBufInfo.width, colBufInfo.height),

        // Disable scissoring
        NW_FONT_CMD_SET_DISABLE_SCISSOR(colBufInfo),

        // Disable W buffer
        // Set the depth range
        // Disable polygon offset
        NW_FONT_CMD_SET_WBUFFER_DEPTHRANGE_POLYGONOFFSET(
            0.0f,           // wScale : W buffer is disabled at 0.0
            0.0f,           // depth range near
            1.0f,           // depth range far
            0,              // polygon offset units : polygon offset is disabled with 0.0
            colBufInfo),
    };

    nngxAdd3DCommand(commands, sizeof(commands), true);

    static const u32 constCommands[] =
    {
        // Clear the cache so frame-buffer access control can be changed
        // 0x111
        1, PICA_CMD_HEADER_SINGLE(PICA_REG_COLOR_DEPTH_BUFFER_CLEAR1),
        // 0x110
        1, PICA_CMD_HEADER_SINGLE(PICA_REG_COLOR_DEPTH_BUFFER_CLEAR0),

        // Disable culling
        NW_FONT_CMD_SET_CULL_FACE(NW_FONT_CMD_CULL_FACE_DISABLE),

        // Disable stencil test
        NW_FONT_CMD_SET_DISABLE_STENCIL_TEST(),

        // Disable depth test
        // Make all elements of the color buffer writable
        NW_FONT_CMD_SET_DEPTH_FUNC_COLOR_MASK(
            false,  // isDepthTestEnabled
            0,      // depthFunc
            true,   // depthMask
            true,   // red
            true,   // Green
            true,   // Blue
            true),  // Alpha

        // Disable early depth test
        NW_FONT_CMD_SET_ENABLE_EARLY_DEPTH_TEST(false),

        // Framebuffer access control
        NW_FONT_CMD_SET_FBACCESS(
            true,   // colorRead
            true,   // colorWrite
            false,  // depthRead
            false,  // depthWrite
            false,  // stencilRead
            false), // stencilWrite
    };

    nngxAdd3DCommand(constCommands, sizeof(constCommands), true);
}


//!---------------------------------------------------------------------------
//
//
//
//
//!---------------------------------------------------------------------------
void
SetupCamera(nw::lyt::DrawInfo& drawInfo, const nw::lyt::Layout& layout)
{
    nw::ut::Rect layoutRect = layout.GetLayoutRect();

    f32 znear = 0.f;
    f32 zfar = 500.f;

    // Set the projection matrix for an orthogonal projection
    // (Since the layout data is horizontal, change its direction.)
    nw::math::MTX44 projMtx;
    nw::math::MTX44Ortho(
        &projMtx,
        layoutRect.bottom,  // Left
        layoutRect.top,     // Right
        -layoutRect.right,  // Bottom
        -layoutRect.left,   // Top
        znear,
        zfar);
    drawInfo.SetProjectionMtx(projMtx);

    // Set the modelview matrix
    // (Since the layout data is horizontal, the upper part of the screen is the -X direction of the layout.)
    nw::math::VEC3 pos(0, 0, 1);
    nw::math::VEC3 up(-1, 0, 0);
    nw::math::VEC3 target(0, 0, 0);

    nw::math::MTX34 viewMtx;
    nw::math::MTX34LookAt(&viewMtx, &pos, &up, &target);
    drawInfo.SetViewMtx(viewMtx);
}

}   // Namespace

//!---------------------------------------------------------------------------
//
//!---------------------------------------------------------------------------
void
nnMain()
{
    nn::applet::Enable();

    nw::demo::SimpleApp& demoApp = nw::demo::SimpleApp::GetInstance();
    demoApp.Initialize();

    nw::lyt::Initialize(&demoApp.GetAllocator(), &demoApp.GetDeviceAllocator());

    File arcLytBuf1 = nw::demo::Utility::LoadFile(
        &demoApp.GetDeviceAllocator(), NW_DEMO_FILE_PATH(L"layout1.arc"), 128);
    NW_ASSERT(!arcLytBuf1.empty());

    File arcLytBuf2 = nw::demo::Utility::LoadFile(
        &demoApp.GetDeviceAllocator(), NW_DEMO_FILE_PATH(L"layout2.arc"), 128);
    NW_ASSERT(!arcLytBuf2.empty());

    File arcLytBuf3 = nw::demo::Utility::LoadFile(
        &demoApp.GetDeviceAllocator(), NW_DEMO_FILE_PATH(L"layout3.arc"), 128);
    NW_ASSERT(!arcLytBuf3.empty());

    // Specify the directory that will be used as the archive data and the resource root

    // adds the resource search target to the archive
    nw::lyt::ArcResourceLink* pArcResLink1 = new nw::lyt::ArcResourceLink;
    pArcResLink1->Set(arcLytBuf1.begin(), "arc");
    nw::lyt::ArcResourceLink* pArcResLink2 = new nw::lyt::ArcResourceLink;
    pArcResLink2->Set(arcLytBuf2.begin(), "arc");
    nw::lyt::ArcResourceLink* pArcResLink3 = new nw::lyt::ArcResourceLink;
    pArcResLink3->Set(arcLytBuf3.begin(), "arc");

    nw::lyt::MultiArcResourceAccessor* pResAccessor = new nw::lyt::MultiArcResourceAccessor();
    pResAccessor->Attach(pArcResLink1);
    pResAccessor->Attach(pArcResLink2);
    pResAccessor->Attach(pArcResLink3);

    nw::lyt::Layout* pLayout1 = new nw::lyt::Layout();
    nw::lyt::Layout* pLayout2 = new nw::lyt::Layout();

    // Load the layout resource
    {
        const void* lytRes = pResAccessor->GetResource(0, "layout1.bclyt");
        NW_NULL_ASSERT(lytRes);
        pLayout1->Build(lytRes, pResAccessor);
    }
    {
        const void* lytRes = pResAccessor->GetResource(0, "layout2.bclyt");
        NW_NULL_ASSERT(lytRes);
        pLayout2->Build(lytRes, pResAccessor);
    }

    nw::lyt::GraphicsResource graphicsResource;
    // Loads the global resource file.
    {
        graphicsResource.StartSetup();
        const wchar_t* resourcePath = 0;
        for (int i = 0;
             (resourcePath = graphicsResource.GetResourcePath(i)) != NULL;
             ++i)
        {
            File file = nw::demo::Utility::LoadFile(&demoApp.GetAllocator(), resourcePath);

            if (file.empty())
            {
                NW_FATAL_ERROR("can not read lyt resource file.");
            }

            graphicsResource.SetResource(i, file.begin(), file.size(), false);
        }

        graphicsResource.FinishSetup();
    }

    nw::lyt::DrawInfo drawInfo;
    drawInfo.SetGraphicsResource(&graphicsResource);

    nw::lyt::Drawer drawer;
    drawer.Initialize(graphicsResource);

    nw::demo::Pad& pad = *nw::demo::PadFactory::GetPad();

    const nw::ut::FloatColor clearColor(0.3f, 0.3f, 0.3f, 1.0f);
    const f32 clearDepth = 0.0f;

    bool loop = true;
    while (loop)
    {
        pad.Update();

        if (pad.IsButtonDown(pad.BUTTON_START))
        {
            loop = false;
        }

        demoApp.SetRenderingTarget(demoApp.DISPLAY0);

        {
            demoApp.GetFrameBufferObject().ClearBuffer(clearColor, clearDepth);

            InitDraw(demoApp.DISPLAY0_WIDTH, demoApp.DISPLAY0_HEIGHT);

            SetupCamera(drawInfo, *pLayout1);

            pLayout1->CalculateMtx(drawInfo);
            pLayout2->CalculateMtx(drawInfo);

            drawer.DrawBegin(drawInfo);
            drawer.Draw(pLayout1, drawInfo);
            drawer.Draw(pLayout2, drawInfo);
            drawer.DrawEnd(drawInfo);
        }

        demoApp.SetRenderingTarget(demoApp.DISPLAY1);

        {
            demoApp.GetFrameBufferObject().ClearBuffer(clearColor, clearDepth);
        }

        demoApp.SwapBuffer(demoApp.DISPLAY_BOTH);
    }

    drawer.Finalize();
    graphicsResource.Finalize();
    delete pLayout2;
    delete pLayout1;

    // removes the archive from the resource search target
#if 1
    pResAccessor->Detach(pArcResLink3);
    pResAccessor->Detach(pArcResLink2);
    pResAccessor->Detach(pArcResLink1);
#else
    pResAccessor->DetachAll();
#endif

    delete pResAccessor;

    delete pArcResLink3;
    delete pArcResLink2;
    delete pArcResLink1;
}

