/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

//------------------------------------------------------------------
// Demo: simple
//
// Overview
//   This demo displays layout data and plays animations.
//   The layout and animation data include the following.
//
//      simple.clyt: the layout data
//      simple.clan: the animation data
//                        - the pane's SRT animation
//                        - the texture pattern animation
//                        - the visibility animation
//                        - the vertex color animation
//                        - the material color  animation
//                        - the texture SRT animation
//
//------------------------------------------------------------------

#include <GLES2/gl2.h>

#include <nn.h>
#include <nn/fs.h>
#include <nn/applet.h>
#include <nw/sdk.h>
#include <nw/demo.h>
#include <nw/lyt.h>

namespace
{

/*
  

  
  
 */
void
InitDraw(int width, int height)
{
    // Color buffer information
    // Switches the width and height to match the LCD direction.
    const nw::font::ColorBufferInfo colBufInfo =
    {
        height, width, PICA_DATA_DEPTH24_STENCIL8_EXT
    };

    const u32 commands[] =
    {
        // Viewport settings
        NW_FONT_CMD_SET_VIEWPORT(0, 0, colBufInfo.width, colBufInfo.height),

        // Disable scissoring
        NW_FONT_CMD_SET_DISABLE_SCISSOR(colBufInfo),

        // Disable W buffer
        // Set the depth range
        // Disable polygon offset
        NW_FONT_CMD_SET_WBUFFER_DEPTHRANGE_POLYGONOFFSET(
            0.0f,           // wScale : W buffer is disabled at 0.0
            0.0f,           // depth range near
            1.0f,           // depth range far
            0,              // polygon offset units : polygon offset is disabled with 0.0
            colBufInfo),
    };

    nngxAdd3DCommand(commands, sizeof(commands), true);

    static const u32 constCommands[] =
    {
        // Clear the cache so frame-buffer access control can be changed
        // 0x111
        1, PICA_CMD_HEADER_SINGLE(PICA_REG_COLOR_DEPTH_BUFFER_CLEAR1),
        // 0x110
        1, PICA_CMD_HEADER_SINGLE(PICA_REG_COLOR_DEPTH_BUFFER_CLEAR0),

        // Disable culling
        NW_FONT_CMD_SET_CULL_FACE(NW_FONT_CMD_CULL_FACE_DISABLE),

        // Disable stencil test
        NW_FONT_CMD_SET_DISABLE_STENCIL_TEST(),

        // Disable depth test
        // Make all elements of the color buffer writable
        NW_FONT_CMD_SET_DEPTH_FUNC_COLOR_MASK(
            false,  // isDepthTestEnabled
            0,      // depthFunc
            true,   // depthMask
            true,   // red
            true,   // Green
            true,   // Blue
            true),  // Alpha

        // Disable early depth test
        NW_FONT_CMD_SET_ENABLE_EARLY_DEPTH_TEST(false),

        // Framebuffer access control
        NW_FONT_CMD_SET_FBACCESS(
            true,   // colorRead
            true,   // colorWrite
            false,  // depthRead
            false,  // depthWrite
            false,  // stencilRead
            false), // stencilWrite
    };

    nngxAdd3DCommand(constCommands, sizeof(constCommands), true);
}

/*
  
 */
class File
{
public:
    //
    //
    //
    File()
      : m_Buffer(NULL), m_Size(0)
    {
    }

    //
    //
    //
    //
    //
    //
    //
    //
    bool Read(const wchar_t* fileName, nw::os::IAllocator& allocator)
    {
        NW_ASSERT(m_Buffer == NULL);

        nn::fs::FileReader fileReader(fileName);
        if (fileReader.GetSize() <= 0)
        {
            return false;
        }

        this->m_Size = static_cast<u32>(fileReader.GetSize());
        this->m_Buffer = static_cast<u8*>(allocator.Alloc(this->m_Size, 128));
        if (this->m_Buffer == NULL)
        {
            this->m_Size = 0;
            return false;
        }

        fileReader.Read(this->m_Buffer, this->m_Size);
        fileReader.Finalize();

        return true;
    }

    //
    //
    //
    //
    //
    u8* Buffer()
    {
        return this->m_Buffer;
    }

    //
    //
    //
    //
    //
    u32 Size()
    {
        return this->m_Size;
    }

private:
    u8* m_Buffer;
    u32 m_Size;
};

/*
  

  
  
 */
void
SetupCamera(nw::lyt::DrawInfo& drawInfo, const nw::lyt::Layout& layout)
{
    nw::ut::Rect layoutRect = layout.GetLayoutRect();

    f32 znear = 0.f;
    f32 zfar = 500.f;

    // Set the projection matrix for an orthogonal projection
    // (Since the layout data is horizontal, change its direction.)
    nw::math::MTX44 projMtx;
    nw::math::MTX44OrthoPivot(
        &projMtx,
        layoutRect.left,   // Left
        layoutRect.right,  // Right
        layoutRect.bottom, // Bottom
        layoutRect.top,    // Top
        znear,
        zfar,
        nw::math::PIVOT_UPSIDE_TO_TOP);
    drawInfo.SetProjectionMtx(projMtx);

    // Set the modelview matrix
    // (Since the layout data is horizontal, the upper part of the screen is the -X direction of the layout.)
    nw::math::VEC3 pos(0, 0, 1);
    nw::math::VEC3 up(0, 1, 0);
    nw::math::VEC3 target(0, 0, 0);

    nw::math::MTX34 viewMtx;
    nw::math::MTX34LookAt(&viewMtx, &pos, &up, &target);
    drawInfo.SetViewMtx(viewMtx);
}

}   // namespace {anonymous}

/*
  
 */
void
nnMain()
{
    nn::applet::Enable();

    // Initializes the base library.
    nn::fs::Initialize();

    // Initialize applications
    nw::demo::SimpleApp& demoApp = nw::demo::SimpleApp::GetInstance();
    demoApp.Initialize();

    // Initializes the layout library.
    nw::lyt::Initialize(&demoApp.GetAllocator(), &demoApp.GetDeviceAllocator());

    nw::lyt::Layout* pLayout = new nw::lyt::Layout();

    // Loads layout binary resources (archives).
    File fileLayout;
    if (!fileLayout.Read(NW_DEMO_FILE_PATH(L"layout.arc"), demoApp.GetDeviceAllocator()))
    {
        NW_FATAL_ERROR("can not open layout archive.\n");
    }

    // Specifies the binary resource root directory and generates the resource assessor.
    nw::lyt::ArcResourceAccessor* pResAccessor = new nw::lyt::ArcResourceAccessor;
    if (!pResAccessor->Attach(fileLayout.Buffer(), "."))
    {
        NW_FATAL_ERROR("can not attach layout archive.\n");
    }

    // Load the layout resource
    {
        const void* lytRes = pResAccessor->GetResource(0, "simple.bclyt");
        NW_NULL_ASSERT(lytRes);
        pLayout->Build(lytRes, pResAccessor);
    }

    nw::lyt::AnimTransform* pAnimTrans = 0;
    // Loads the animation resource.
    {
        const void* lpaRes = pResAccessor->GetResource(0, "simple.bclan");
        NW_NULL_ASSERT(lpaRes);
        pAnimTrans = pLayout->CreateAnimTransform(lpaRes, pResAccessor);
        pLayout->BindAnimation(pAnimTrans);
    }

    f32 animFrame = 0;

    nw::lyt::GraphicsResource graphicsResource;

    // Loads the global resource file.
    {
        graphicsResource.StartSetup();
        const wchar_t* resourcePath = 0;
        for (int i = 0;
             (resourcePath = graphicsResource.GetResourcePath(i)) != NULL;
             ++i)
        {
            File file;
            if (!file.Read(resourcePath, demoApp.GetAllocator()))
            {
                NW_FATAL_ERROR("can not read lyt resource file.");
            }
            graphicsResource.SetResource(i, file.Buffer(), file.Size());
        }
        graphicsResource.FinishSetup();
    }

    nw::lyt::DrawInfo drawInfo;
    drawInfo.SetGraphicsResource(&graphicsResource);

    nw::lyt::Drawer drawer;
    drawer.Initialize(graphicsResource);

    const nw::ut::FloatColor clearColor(0.3f, 0.3f, 0.3f, 1.0f);
    const f32 clearDepth = 0.0f;

    bool mainloop = true;
    while (mainloop)
    {
        demoApp.SetRenderingTarget(demoApp.DISPLAY0);
        {
            demoApp.GetFrameBufferObject().ClearBuffer(clearColor, clearDepth);

            SetupCamera(drawInfo, *pLayout);

            InitDraw(demoApp.DISPLAY0_WIDTH, demoApp.DISPLAY0_HEIGHT);
            pLayout->Animate();
            pLayout->CalculateMtx(drawInfo);

            drawer.DrawBegin(drawInfo);
            drawer.Draw(pLayout, drawInfo);
            drawer.DrawEnd(drawInfo);
        }

        demoApp.SetRenderingTarget(demoApp.DISPLAY1);
        {
            demoApp.GetFrameBufferObject().ClearBuffer(clearColor, clearDepth);
        }

        // Frame update.
        animFrame += 1.0f;
        while (animFrame >= pAnimTrans->GetFrameSize())
        {
            animFrame -= pAnimTrans->GetFrameSize();
        }
        pAnimTrans->SetFrame(animFrame);

        demoApp.SwapBuffer(demoApp.DISPLAY_BOTH);
    }

    drawer.Finalize();
    graphicsResource.Finalize();
    delete pLayout;
    delete pResAccessor;

    demoApp.GetDeviceAllocator().Free(fileLayout.Buffer());
}

