/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

//------------------------------------------------------------------
// Demo: vramCmdBuf
//
// Overview
//   This demo places the command buffer used for character string rendering in VRAM and executes it with a command jump.
//   
//
//------------------------------------------------------------------

#define USE_VRAM_COMMAND_BUFFER

#include <GLES2/gl2.h>

#include <nn.h>
#include <nn/fs.h>
#include <nn/applet.h>
#include <nw/sdk.h>
#include <nw/demo.h>
#include <nw/lyt.h>

namespace
{

/*
  
  
  
 
*/
void
InitDraw(int width, int height)
{
    // Color buffer information
    // Switches the width and height to match the LCD direction.
    const nw::font::ColorBufferInfo colBufInfo =
    {
        height, width, PICA_DATA_DEPTH24_STENCIL8_EXT
    };

    const u32 commands[] =
    {
        // Viewport settings
        NW_FONT_CMD_SET_VIEWPORT(0, 0, colBufInfo.width, colBufInfo.height),

        // Disable scissoring
        NW_FONT_CMD_SET_DISABLE_SCISSOR(colBufInfo),

        // Disable W buffer
        // Set the depth range
        // Disable polygon offset
        NW_FONT_CMD_SET_WBUFFER_DEPTHRANGE_POLYGONOFFSET(
            0.0f,           // wScale : W buffer is disabled at 0.0
            0.0f,           // depth range near
            1.0f,           // depth range far
            0,              // polygon offset units : polygon offset is disabled with 0.0
            colBufInfo),
    };

    nngxAdd3DCommand(commands, sizeof(commands), true);

    static const u32 constCommands[] =
    {
        // Clear the cache so frame-buffer access control can be changed
        // 0x111
        1, PICA_CMD_HEADER_SINGLE(PICA_REG_COLOR_DEPTH_BUFFER_CLEAR1),
        // 0x110
        1, PICA_CMD_HEADER_SINGLE(PICA_REG_COLOR_DEPTH_BUFFER_CLEAR0),

        // Disable culling
        NW_FONT_CMD_SET_CULL_FACE(NW_FONT_CMD_CULL_FACE_DISABLE),

        // Disable stencil test
        NW_FONT_CMD_SET_DISABLE_STENCIL_TEST(),

        // Disable depth test
        // Make all elements of the color buffer writable
        NW_FONT_CMD_SET_DEPTH_FUNC_COLOR_MASK(
            false,  // isDepthTestEnabled
            0,      // depthFunc
            true,   // depthMask
            true,   // red
            true,   // Green
            true,   // Blue
            true),  // Alpha

        // Disable early depth test
        NW_FONT_CMD_SET_ENABLE_EARLY_DEPTH_TEST(false),

        // Framebuffer access control
        NW_FONT_CMD_SET_FBACCESS(
            true,   // colorRead
            true,   // colorWrite
            false,  // depthRead
            false,  // depthWrite
            false,  // stencilRead
            false), // stencilWrite
    };

    nngxAdd3DCommand(constCommands, sizeof(constCommands), true);
}

/*
  
 */
class File
{
public:
    //
    //
    //
    File()
      : m_Buffer(NULL), m_Size(0)
    {
    }

    //
    //
    //
    //
    //
    //
    //
    //
    bool Read(const wchar_t* fileName, nw::os::IAllocator& allocator)
    {
        NW_ASSERT(m_Buffer == NULL);

        nn::fs::FileReader fileReader(fileName);
        if (fileReader.GetSize() <= 0)
        {
            return false;
        }

        this->m_Size = static_cast<u32>(fileReader.GetSize());
        this->m_Buffer = static_cast<u8*>(allocator.Alloc(this->m_Size, 128));
        if (this->m_Buffer == NULL)
        {
            this->m_Size = 0;
            return false;
        }

        fileReader.Read(this->m_Buffer, this->m_Size);
        fileReader.Finalize();

        return true;
    }

    //
    //
    //
    //
    //
    u8* Buffer()
    {
        return this->m_Buffer;
    }

    //
    //
    //
    //
    //
    u32 Size()
    {
        return this->m_Size;
    }

private:
    u8* m_Buffer;
    u32 m_Size;
};

/*
  
  
  
 
*/
void
SetupCamera(nw::lyt::DrawInfo& drawInfo, const nw::lyt::Layout& layout)
{
    nw::ut::Rect layoutRect = layout.GetLayoutRect();

    f32 znear = 0.f;
    f32 zfar = 500.f;

    // Set the projection matrix for an orthogonal projection
    // (Since the layout data is horizontal, change its direction.)
    nw::math::MTX44 projMtx;
    nw::math::MTX44OrthoPivot(
        &projMtx,
        layoutRect.left,   // Left
        layoutRect.right,  // Right
        layoutRect.bottom, // Bottom
        layoutRect.top,    // Top
        znear,
        zfar,
        nw::math::PIVOT_UPSIDE_TO_TOP);
    drawInfo.SetProjectionMtx(projMtx);

    // Set the modelview matrix
    // (Since the layout data is horizontal, the upper part of the screen is the -X direction of the layout.)
    nw::math::VEC3 pos(0, 0, 1);
    nw::math::VEC3 up(0, 1, 0);
    nw::math::VEC3 target(0, 0, 0);

    nw::math::MTX34 viewMtx;
    nw::math::MTX34LookAt(&viewMtx, &pos, &up, &target);
    drawInfo.SetViewMtx(viewMtx);
}

}   // namespace {anonymous}

/*
  
 */
void
nnMain()
{
    nn::applet::Enable();

    // Initializes the base library.
    nn::fs::Initialize();

    // Initialize applications
    nw::demo::SimpleApp& demoApp = nw::demo::SimpleApp::GetInstance();
    demoApp.Initialize();

    // Initializes the layout library.
    nw::lyt::Initialize(&demoApp.GetAllocator(), &demoApp.GetDeviceAllocator());

    nw::lyt::Layout* pLayout = new nw::lyt::Layout();

    // Loads layout binary resources (archives).
    File fileLayout;
    if (!fileLayout.Read(NW_DEMO_FILE_PATH(L"layout.arc"), demoApp.GetDeviceAllocator()))
    {
        NW_FATAL_ERROR("can not open layout archive.\n");
    }

    // Specifies the binary resource root directory and generates the resource assessor.
    nw::lyt::ArcResourceAccessor* pResAccessor = new nw::lyt::ArcResourceAccessor;
    if (!pResAccessor->Attach(fileLayout.Buffer(), "."))
    {
        NW_FATAL_ERROR("can not attach layout archive.\n");
    }

    // Load the layout resource
    {
        const void* lytRes = pResAccessor->GetResource(0, "simple.bclyt");
        NW_NULL_ASSERT(lytRes);
        pLayout->Build(lytRes, pResAccessor);
    }

    nw::lyt::AnimTransform* pAnimTrans = 0;
    // Loads the animation resource.
    {
        const void* lpaRes = pResAccessor->GetResource(0, "simple.bclan");
        NW_NULL_ASSERT(lpaRes);
        pAnimTrans = pLayout->CreateAnimTransform(lpaRes, pResAccessor);
        pLayout->BindAnimation(pAnimTrans);
    }

    f32 animFrame = 0;

    nw::lyt::GraphicsResource graphicsResource;

    // Loads the global resource file.
    {
        graphicsResource.StartSetup();
        const wchar_t* resourcePath = 0;
        for (int i = 0;
             (resourcePath = graphicsResource.GetResourcePath(i)) != NULL;
             ++i)
        {
            File file;
            if (!file.Read(resourcePath, demoApp.GetAllocator()))
            {
                NW_FATAL_ERROR("can not read lyt resource file.");
            }
            graphicsResource.SetResource(i, file.Buffer(), file.Size());
        }
        graphicsResource.FinishSetup();
    }

    nw::lyt::DrawInfo drawInfo;
    drawInfo.SetGraphicsResource(&graphicsResource);

    nw::lyt::Drawer drawer;
    drawer.Initialize(graphicsResource);

    nw::demo::Pad& pad = *nw::demo::PadFactory::GetPad();

#if defined(USE_VRAM_COMMAND_BUFFER)
    // Generate a command list for fixed text and place it in VRAM.
    {
        GLvoid* (*pGlAllocator)(GLenum, GLenum, GLuint, GLsizei) = NULL;
        nngxGetAllocator(&pGlAllocator, NULL);
        if (pGlAllocator)
        {
            nw::lyt::Pane* pPane = pLayout->GetRootPane();
            while (pPane)
            {
                nw::lyt::TextBox* pTextBox = nw::ut::DynamicCast<nw::lyt::TextBox*>(pPane);
                pPane = nw::lyt::GetNextPane(pPane);

                if (!pTextBox)
                {
                    continue;
                }

                nw::font::DispStringBuffer* pDispStringBuff = pTextBox->GetDispStringBuffer();

                // Arrange to use the command buffer located in VRAM.
                pDispStringBuff->SetCommandBufferMode(pDispStringBuff->CBM_VRAM);

                // Generate a 3D command list.
                pTextBox->UpdateDrawCharData(&drawer);

                // Allocate a command buffer big enough to store the 3D command list from VRAM.
                void* cmdBuf = pGlAllocator(NN_GX_MEM_VRAMA, NN_GX_MEM_COMMANDBUFFER, 0, pDispStringBuff->commandBufferSize * sizeof(u32));

                // This does not work if the command buffer cannot be allocated because the generated 3D command list is for command jumps only.
                NW_NULL_ASSERT(cmdBuf);

                // Transfer to VRAM.
                pDispStringBuff->SetVramCommandBuffer(cmdBuf);
                pDispStringBuff->SendToVram();
            }
        }
    }
#endif

    const nw::ut::FloatColor clearColor(0.3f, 0.3f, 0.3f, 1.0f);
    const f32 clearDepth = 0.0f;

    // Waits until the transfer to VRAM completes.
    nngxRunCmdlist();
    nngxWaitCmdlistDone();

    // This sample makes double use of the command list through the demo library features and it executes in parallel the issuing of rendering commands from CPU and rendering from GPU.
    // 
    // 
    // If demoApp.SwapBuffer is called at the end of a frame, the current command list changes and enters the stopped mode.
    // 
    // 
    // 
    // The first frame is put in the stopped mode here.
    nngxStopCmdlist();

    bool mainloop = true;
    while (mainloop)
    {
        pad.Update();

        if (pad.IsButtonDown(pad.BUTTON_START))
        {
            mainloop = false;
        }

        demoApp.SetRenderingTarget(demoApp.DISPLAY0);
        {
            demoApp.GetFrameBufferObject().ClearBuffer(clearColor, clearDepth);

            SetupCamera(drawInfo, *pLayout);

            InitDraw(demoApp.DISPLAY0_WIDTH, demoApp.DISPLAY0_HEIGHT);
            pLayout->Animate();
            pLayout->CalculateMtx(drawInfo);


#if defined(USE_VRAM_COMMAND_BUFFER)
            // Gets start address of buffer of unissued commands.
            // Used later to flush the command buffer cache.
            nngxFlush3DCommand();
            GLint cmdStartAddr = 0;
            nngxGetCmdlistParameteri(NN_GX_CMDLIST_CURRENT_BUFADDR, &cmdStartAddr);
#endif

            drawer.DrawBegin(drawInfo);
            drawer.Draw(pLayout, drawInfo);
            drawer.DrawEnd(drawInfo);

#if defined(USE_VRAM_COMMAND_BUFFER)
            // Draw commands are issued differently if command jumps are used.
            GLint cmdFirstReturnAddr = (GLint)drawer.GetCommandJumpFirstReturnAddr();
            if (cmdFirstReturnAddr != 0)
            {
                // The command list must be in the stopped state when calling nngxFlush3DCommandPartially.
                // 
                // 
                // However, the current command list is actually always in the stopped state because this demo uses the double command list.
                // 
                GLint isRunning = 0;
                nngxGetCmdlistParameteri(NN_GX_CMDLIST_IS_RUNNING, &isRunning);
                NW_ASSERT(!isRunning); // Actually, always in the stopped state.
                if (isRunning)
                {
                    nngxWaitCmdlistDone();
                    nngxStopCmdlist();
                }

                // This instruction issues a 3D command if using command jumps.
                nngxFlush3DCommandPartially(cmdFirstReturnAddr - cmdStartAddr);
                NW_GL_ASSERT();

                // Completes the last jump command.
                // The last jump command is not completed until the 3D command is issued and the entire length of the 3D command sequence is confirmed.
                // 
                GLint cmdEndAddr;
                nngxGetCmdlistParameteri(NN_GX_CMDLIST_CURRENT_BUFADDR, &cmdEndAddr);
                drawer.FixCommandJump(reinterpret_cast<void*>(cmdEndAddr));

                // The command buffer must flush the cache itself.
                nngxUpdateBuffer((void*)cmdStartAddr, cmdEndAddr - cmdStartAddr);

                // Returns the command list execution status to what it was.
                if (isRunning)
                {
                    nngxRunCmdlist();
                }
            }
#endif
        }

        demoApp.SetRenderingTarget(demoApp.DISPLAY1);
        {
            demoApp.GetFrameBufferObject().ClearBuffer(clearColor, clearDepth);
        }

        // Frame update.
        animFrame += 1.0f;
        while (animFrame >= pAnimTrans->GetFrameSize())
        {
            animFrame -= pAnimTrans->GetFrameSize();
        }
        pAnimTrans->SetFrame(animFrame);

        demoApp.SwapBuffer(demoApp.DISPLAY_BOTH);
    }

#if defined(USE_VRAM_COMMAND_BUFFER)
    // Destroy the VRAM command buffer.
    {
        void (*pGlDeallocator)(GLenum, GLenum, GLuint, GLvoid*) = NULL;
        nngxGetAllocator(NULL, &pGlDeallocator);
        if (pGlDeallocator)
        {
            nw::lyt::Pane* pPane = pLayout->GetRootPane();
            while (pPane)
            {
                nw::lyt::TextBox* pTextBox = nw::ut::DynamicCast<nw::lyt::TextBox*>(pPane);
                pPane = nw::lyt::GetNextPane(pPane);

                if (!pTextBox)
                {
                    continue;
                }

                nw::font::DispStringBuffer* pDispStringBuff = pTextBox->GetDispStringBuffer();
                void* cmdBuf = pDispStringBuff->SetVramCommandBuffer(NULL);
                pGlDeallocator(NN_GX_MEM_VRAMA, NN_GX_MEM_COMMANDBUFFER, 0, cmdBuf);
            }
        }
    }
#endif

    drawer.Finalize();
    graphicsResource.Finalize();
    delete pLayout;
    delete pResAccessor;

    demoApp.GetDeviceAllocator().Free(fileLayout.Buffer());
}

