/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#ifndef NW_LYT_STOPWATCH_H_
#define NW_LYT_STOPWATCH_H_

#include <climits>
#include <math.h>
#include <nw/sdk.h>
#include <nw/ut/ut_LinkList.h>

#ifdef NW_PLATFORM_CTR
#include <nn/os/os_Tick.h>
#endif

namespace nw {
namespace lyt {
namespace internal {

// #define NW_LYT_PROFILE 1

#ifdef NW_LYT_PROFILE
#  define NW_LYT_JOIN(x, y) NW_LYT_JOIN_AGAIN(x, y)
#  define NW_LYT_JOIN_AGAIN(x, y) x ## y

#  define NW_LYT_STOPWATCH_MEASURE(no, label) \
    nw::lyt::internal::Measure NW_LYT_JOIN(measure, __LINE__) ( \
        nw::lyt::internal::Stopwatch::Instance<no>(label), no)

#  define NW_LYT_STOPWATCH_START(no, label) \
    if(nw::lyt::internal::Stopwatch::IsTarget(no)) \
        nw::lyt::internal::Stopwatch::Instance<no>().Start(label)

#  define NW_LYT_STOPWATCH_STOP(no) \
    if(nw::lyt::internal::Stopwatch::IsTarget(no)) \
        nw::lyt::internal::Stopwatch::Instance<no>().Stop()
#else
#  define NW_LYT_STOPWATCH_MEASURE(no, label)
#  define NW_LYT_STOPWATCH_START(no, label)
#  define NW_LYT_STOPWATCH_STOP(no)
#endif

#if defined(NW_LYT_PROFILE)

#ifdef NW_PLATFORM_CTR
using nn::fnd::TimeSpan;
using nn::os::Tick;
#endif

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------
class Stopwatch : public nw::ut::LinkListNode
{
public:
    static const s64 S64_MAX = (s64)(((u64)-1) >> 1);

    //
    typedef nw::ut::LinkList<Stopwatch, 0> List;

    //
    Stopwatch(int no = 0, const char* label = NULL);

    //
    //
    //
    //
    void Start(const char* label = NULL)
    {
#ifndef NW_PLATFORM_CTR
        NW_UNUSED_VARIABLE(label)
#else
        if (label != NULL)
        {
            this->SetLabel(label);
        }

        NW_ASSERT(!this->IsRunning());

        m_StartTick = Tick::GetSystemCurrent();
#endif
    }

    //
    void Stop()
    {
#ifdef NW_PLATFORM_CTR
        Tick stopTick = Tick::GetSystemCurrent();

        NW_ASSERT(this->IsRunning());

        Tick elapsed = stopTick - m_StartTick;
        m_StartTick = Tick();

        if (s64(elapsed) < m_MinSample)
        {
            m_MinSample = s64(elapsed);
        }

        if (s64(elapsed) > m_MaxSample)
        {
            m_MaxSample = s64(elapsed);
        }

        m_TotalTick += elapsed;
        ++m_Count;
#endif
    }

    void Reset();

    static void ResetAll();

    //
    void SetNo(int no)
    {
#ifndef NW_PLATFORM_CTR
        NW_UNUSED_VARIABLE(no);
#else
        m_No = no;
#endif
    }

    //
    void SetLabel(const char* label)
    {
#ifndef NW_PLATFORM_CTR
        NW_UNUSED_VARIABLE(label);
#else
        m_Label = label;
#endif
    }

#ifdef NW_PLATFORM_CTR
    //
    bool IsRunning() const
    {
        return m_StartTick != Tick();
    }

    //
    int GetNo() const
    {
        return m_No;
    }

    //
    const char* GetLabel() const
    {
        return m_Label;
    }

    //
    const TimeSpan GetElapsedTime() const
    {
        return m_TotalTick;
    }

    //
    int GetCount() const
    {
        return m_Count;
    }

    //
    TimeSpan GetMinSample() const
    {
        if (m_Count == 0)
        {
            return Tick(0).ToTimeSpan();
        }
        else
        {
            return Tick(m_MinSample).ToTimeSpan();
        }
    }

    //
    TimeSpan GetMaxSample() const
    {
        if (m_Count == 0)
        {
            return Tick(0).ToTimeSpan();
        }
        else
        {
            return Tick(m_MaxSample).ToTimeSpan();
        }
    }
#endif

    //
    //
    //
    //
    //
    template <int no>
    static Stopwatch& Instance(const char* label = NULL)
    {
        static Stopwatch obj(no);

        if (label != NULL)
        {
            obj.SetLabel(label);
        }

        return obj;
    }

    //
    static List& GetList()
    {
        static List list;
        return list;
    }

    //
    //
    //
    static void Enable();

    //
    //
    //
    static void Disable();

    //
    static bool IsTarget(int no)
    {
        return s_TargetNoMin <= no && no <= s_TargetNoMax;
    }

    //
    static void SetTargetNo(int min, int max)
    {
        s_TargetNoMin = min;
        s_TargetNoMax = max;
    }

    //
    static void SetTargetNo(int no)
    {
        SetTargetNo(no, no);
    }

    //
    static void Dump(const Stopwatch& stopwatch);

    //
    static void Dump();

    //
    static void DumpHeader();

protected:
#ifdef NW_PLATFORM_CTR
    //
    int m_Count;

    //
    int m_No;

    //
    const char* m_Label;

    //
    u8 padding[4];

    //
    Tick m_TotalTick;

    //
    Tick m_StartTick;

    //
    s64 m_MinSample;

    //
    s64 m_MaxSample;
#endif

    //
    static int s_TargetNoMin;

    //
    static int s_TargetNoMax;
};

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------
class Measure
{
public:
    Measure(Stopwatch& stopwatch, int no)
      : m_Stopwatch(stopwatch)
    {
#ifdef NW_PLATFORM_CTR
        if(Stopwatch::IsTarget(no))
        {
            m_Stopwatch.Start();
        }
#endif
    }

    ~Measure()
    {
#ifdef NW_PLATFORM_CTR
        if(m_Stopwatch.IsRunning())
        {
            m_Stopwatch.Stop();
        }
#endif
    }

protected:
    //
    Stopwatch& m_Stopwatch;
};

#endif // NW_LYT_PROFILE

} // namespace internal
} // namespace lyt
} // namespace nw

#endif // NW_LYT_STOPWATCH_H_
