/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#include <nn/applet.h>
#include <nw/viewer.h>
#include <nw/lyt.h>
#include <nw/demo.h>

#ifdef NW_PLATFORM_CTR
#if NN_CURRENT_VERSION_NUMBER >= NN_VERSION_NUMBER(4,0,0,0)
#pragma diag_suppress 2530 // padding added to end of struct.
#endif
#endif
namespace
{

nw::demo::RenderSystem* s_pRenderSystem = NULL;
nw::gfx::IRenderTarget* s_pRenderTarget = NULL;
nw::gfx::IRenderTarget* s_pRenderTargetLower = NULL;

//----------------------------------------
// Memory-related

// Allocator to allocate device memory.
nw::demo::DemoAllocator s_DeviceAllocator;

//----------------------------------------
typedef nw::ut::MoveArray<u8> File;

/*
  
 */
class LayoutResourceListener : public nw::viewer::lyt::IResourceListener
{
public:
    LayoutResourceListener()
        : mLoaded(false)
    {
    }

    virtual bool OnLoadResource(nw::viewer::lyt::FileResourceAccessor* accessor)
    {
        nw::viewer::lyt::FileResourceAccessor::FindInfoPtr found;

        NW_DEV_LOG("OnLoadResource(%p)\n", accessor);
        NW_DEV_LOG("  bclyt resources:\n");
        found = accessor->FindResource(nw::lyt::res::RESOURCETYPE_LAYOUT);
        for (; found->IsValid(); found->FindNext())
        {
            NW_DEV_LOG("    %s\n", found->GetName());
        }

        NW_DEV_LOG("  bclan resources:\n");
        found = accessor->FindResource(nw::lyt::res::RESOURCETYPE_ANIMATION);
        for (; found->IsValid(); found->FindNext())
        {
            NW_DEV_LOG("    %s\n", found->GetName());
        }

        mLoaded = true;
        return false;
    }

    virtual nw::lyt::Layout* OnBuildLayout(const void* res, const char* name, nw::viewer::lyt::FileResourceAccessor* accessor)
    {
        NW_DEV_LOG("OnBuildLayout(%p, \"%s\", %p)\n", res, name, accessor);
        return NULL;
    }

    virtual bool OnBindAnimation(nw::lyt::Layout* layout, const nw::lyt::AnimResource& res, nw::viewer::lyt::FileResourceAccessor* accessor)
    {
        const char* tagName = res.GetTagName();
        NW_DEV_LOG("OnBindAnimation(%p, %p, %p): %s\n", layout, &res, accessor, tagName? tagName : "<all>");
        return false;
    }

    virtual void OnLoadResourceDone(bool result)
    {
        NW_DEV_LOG("OnLoadResourceDone(%s)\n", result? "true" : "false");
    }

    virtual void OnLoadToUpper()
    {
        NW_DEV_LOG("OnLoadToUpper()\n");
    }

    virtual void OnLoadToLower()
    {
        NW_DEV_LOG("OnLoadToLower()\n");
    }

    // Sets to true if OnLoadResource is called.
    bool mLoaded;
};

/*
  
 */
void
InitializeGraphics()
{
    nw::demo::RenderSystem::Description renderDesc;

    nw::demo::DisplayBufferSwapper::Description& upperScreenDesc =
        renderDesc.upperScreenDescription;

    nw::demo::DisplayBufferSwapper::Description& lowerScreenDesc =
        renderDesc.lowerScreenDescription;

    s_pRenderSystem = nw::demo::RenderSystem::Create(
        &s_DeviceAllocator,
        renderDesc);

    NW_ASSERT(s_pRenderSystem != NULL);

    s_pRenderTarget = nw::gfx::IRenderTarget::Builder()
        .BufferSize(upperScreenDesc.height, upperScreenDesc.width)
        .ColorFormat(renderDesc.renderColorFormat)
        .Create(&s_DeviceAllocator);

    NW_ASSERT(s_pRenderTarget != NULL);

    s_pRenderTargetLower = nw::gfx::IRenderTarget::Builder()
        .BufferSize(lowerScreenDesc.height, lowerScreenDesc.width)
        .ColorFormat(renderDesc.renderColorFormat)
        .Create(&s_DeviceAllocator);

    NW_ASSERT(s_pRenderTargetLower != NULL);

    NW_GL_ASSERT();
}

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------
void
TerminateGraphics()
{
    nw::gfx::SafeDestroy(s_pRenderSystem);
    nw::gfx::SafeDestroy(s_pRenderTarget);
    nw::gfx::SafeDestroy(s_pRenderTargetLower);
}

/*
  
  
  
 
*/
void
InitDraw(int width, int height)
{
    // Color buffer information
    // Switches the width and height to match the LCD direction.
    const nw::font::ColorBufferInfo colBufInfo =
    {
        height, width, PICA_DATA_DEPTH24_STENCIL8_EXT
    };

    const u32 commands[] =
    {
        // Viewport settings
        NW_FONT_CMD_SET_VIEWPORT(0, 0, colBufInfo.width, colBufInfo.height),

        // Disable scissoring
        NW_FONT_CMD_SET_DISABLE_SCISSOR(colBufInfo),

        // Disable W buffer
        // Set the depth range
        // Disable polygon offset
        NW_FONT_CMD_SET_WBUFFER_DEPTHRANGE_POLYGONOFFSET(
            0.0f,           // wScale : W buffer is disabled at 0.0
            0.0f,           // depth range near
            1.0f,           // depth range far
            0,              // polygon offset units : polygon offset is disabled with 0.0
            colBufInfo),
    };

    nngxAdd3DCommand(commands, sizeof(commands), true);

    static const u32 constCommands[] =
    {
        // Disable culling
        NW_FONT_CMD_SET_CULL_FACE(NW_FONT_CMD_CULL_FACE_DISABLE),

        // Disable stencil test
        NW_FONT_CMD_SET_DISABLE_STENCIL_TEST(),

        // Disable depth test
        // Make all elements of the color buffer writable
        NW_FONT_CMD_SET_DEPTH_FUNC_COLOR_MASK(
            false,  // isDepthTestEnabled
            0,      // depthFunc
            true,   // depthMask
            true,   // red
            true,   // Green
            true,   // Blue
            true),  // Alpha

        // Disable early depth test
        NW_FONT_CMD_SET_ENABLE_EARLY_DEPTH_TEST(false),

        // Framebuffer access control
        NW_FONT_CMD_SET_FBACCESS(
            true,   // colorRead
            true,   // colorWrite
            false,  // depthRead
            false,  // depthWrite
            false,  // stencilRead
            false), // stencilWrite
    };

    nngxAdd3DCommand(constCommands, sizeof(constCommands), true);
}

//---------------------------------------------------------------------------
//
//
//
//
//
//
//---------------------------------------------------------------------------
void
SetupCamera(
    nw::lyt::DrawInfo&      drawInfo,
    const nw::lyt::Layout&  layout
)
{
    nw::ut::Rect layoutRect = layout.GetLayoutRect();

    f32 znear = 0.f;
    f32 zfar = 500.f;

    // Set the projection matrix for an orthogonal projection
    // (Since the layout data is horizontal, change its direction.)
    nw::math::MTX44 projMtx;
    nw::math::MTX44OrthoPivot(
        &projMtx,
        layoutRect.left,   // Left
        layoutRect.right,  // Right
        layoutRect.bottom, // Bottom
        layoutRect.top,    // Top
        znear,
        zfar,
        nw::math::PIVOT_UPSIDE_TO_TOP);
    drawInfo.SetProjectionMtx(projMtx);

    // Set the modelview matrix
    nw::math::VEC3 pos(0, 0, 200);
    nw::math::VEC3 up(0, 1, 0);
    nw::math::VEC3 target(0, 0, 0);

    nw::math::MTX34 viewMtx;
    nw::math::MTX34LookAt(&viewMtx, &pos, &up, &target);
    drawInfo.SetViewMtx(viewMtx);
}

/*
  
 */
f32 GetCurrentFrame()
{
    nw::viewer::lyt::LayoutInfo lytInfo =
        nw::viewer::lyt::LayoutInfo::GetFirstLayout();

    if (lytInfo.IsValid())
    {
        return lytInfo.GetCurrentFrame();
    }
    else
    {
        return 0.0f;
    }
}

/*
  
 */
void
DemoScene()
{
    // Class for accessing the embedded viewer.
    nw::viewer::sys::App& app = nw::viewer::sys::App::GetInstance();

    // Class for accessing the layout module.
    nw::viewer::lyt::ModuleLYT& module = app.GetModuleLYT();

    // Registers the listener of the preview event.
    LayoutResourceListener listener;
    module.SetResourceListener(&listener);

    // Initialization involving layout rendering.
    nw::lyt::GraphicsResource graphicsResource;
    nw::lyt::DrawInfo drawInfo;
    nw::lyt::Drawer drawer;

    // Loads the global resource file.
    {
        graphicsResource.StartSetup();
        const wchar_t* resourcePath = 0;
        for (int i = 0;
            (resourcePath = graphicsResource.GetResourcePath(i)) != NULL;
            ++i)
        {
            File resource =
                nw::demo::Utility::LoadFile(&s_DeviceAllocator, resourcePath);

            if (resource.empty())
            {
                NW_FATAL_ERROR("can not read lyt resource file.");
            }

            graphicsResource.SetResource(i, resource.begin(), resource.size(), false);
        }

        graphicsResource.FinishSetup();
    }

    bool autoAnimate = true;
    f32 frame = 0;
    drawInfo.SetGraphicsResource(&graphicsResource);
    drawer.Initialize(graphicsResource);

    bool isContinuing = true;
    while ( isContinuing )
    {
        nw::demo::DebugUtility::AdvanceAutoTestFrame();

        nw::demo::Pad& pad = *nw::demo::PadFactory::GetPad();
        pad.Update();
        if (pad.IsButtonDown(pad.BUTTON_A))
        {
            autoAnimate = !autoAnimate;
            frame = GetCurrentFrame();
        }
        if (pad.IsButtonPress(pad.BUTTON_X))
        {
            autoAnimate = false;
            frame = GetCurrentFrame() + 1.0f;
        }
        if (pad.IsButtonPress(pad.BUTTON_Y))
        {
            autoAnimate = false;
            frame = GetCurrentFrame() - 1.0f;
        }

        // Communications processing of the embedded viewer.
        app.PollStream();

        if (listener.mLoaded)
        {
            // Accesses the instance of the loaded layout.
            // (Displays layout information.)

            // Memo: The number of loaded layouts is always 1.
            NW_DEV_LOG("num of layout: %d\n", nw::viewer::lyt::LayoutInfo::GetLayoutNum());

            nw::viewer::lyt::LayoutInfo lytInfo = nw::viewer::lyt::LayoutInfo::GetFirstLayout();
            while (lytInfo.IsValid())
            {
                NW_DEV_LOG("layout name: %s\n", lytInfo.GetName());
                NW_DEV_LOG("  num of anim: %d\n", lytInfo.GetAnimNum());

                nw::viewer::lyt::AnimInfo animInfo = lytInfo.GetFirstAnim();
                while (animInfo.IsValid())
                {
                    NW_DEV_LOG("  animation name: %s\n", animInfo.GetName());
                    animInfo = animInfo.GetNext();
                }

                lytInfo = lytInfo.GetNext();
            }

            listener.mLoaded = false;
        }

        // Update the animation.
        {

            if (autoAnimate)
            {                // Entrusts frame updating to the viewer library.
                module.UpdateFrame();
            }

            // Animations are evaluated using demos to use the drawInfo that demo possess.
            nw::viewer::lyt::LayoutInfo lytInfo = nw::viewer::lyt::LayoutInfo::GetFirstLayout();
            while (lytInfo.IsValid())
            {
                if (!autoAnimate)
                {
                    lytInfo.SetCurrentFrame(frame);
                }

                nw::lyt::Layout* pLayout = lytInfo.GetLayout();

                pLayout->Animate();
                pLayout->CalculateMtx(drawInfo);

                lytInfo = lytInfo.GetNext();
            }
        }

        // Upper screen
        s_pRenderSystem->SetRenderTarget(s_pRenderTarget);
        NW_GL_ASSERT();
        {
            s_pRenderSystem->ClearBuffer(
                nw::ut::FloatColor(0.2f, 0.2f, 0.2f, 0.0f));

            InitDraw(
                nw::demo::SimpleApp::DISPLAY0_WIDTH,
                nw::demo::SimpleApp::DISPLAY0_HEIGHT);

            // The demo renders the layout possessed by the viewer library.
            nw::viewer::lyt::LayoutInfo lytInfo = nw::viewer::lyt::LayoutInfo::GetFirstLayout();
            while (lytInfo.IsValid() && lytInfo.GetLayout())
            {
                nw::lyt::Layout* pLayout = lytInfo.GetLayout();

                SetupCamera(drawInfo, *pLayout);
                NW_GL_ASSERT();

                drawer.DrawBegin(drawInfo);
                drawer.Draw(pLayout, drawInfo);
                drawer.DrawEnd(drawInfo);

                lytInfo = lytInfo.GetNext();
            }

            s_pRenderSystem->TransferBuffer(nw::demo::UPPER_SCREEN);
        }

        // Lower screen
        s_pRenderSystem->SetRenderTarget(s_pRenderTargetLower);
        {
            s_pRenderSystem->ClearBuffer(
                nw::ut::FloatColor(0.1f, 0.1f, 0.1f, 0.0f));

            s_pRenderSystem->TransferBuffer(nw::demo::LOWER_SCREEN);
        }

        s_pRenderSystem->PresentBuffer(nw::demo::NORMAL_SCREENS);
        NW_GL_ASSERT();

        if (nw::demo::Utility::IsTerminating())
        {
            isContinuing = false;
        }
    }

    drawer.Finalize();
}

}   // namespace {anonymous}

/*
  
 */
void
nnMain()
{
    nn::applet::Enable();

    nw::demo::InitializeGraphicsSystem(&s_DeviceAllocator);

    nw::demo::PadFactory::Initialize(&s_DeviceAllocator);

    void* hioWorkMemory = s_DeviceAllocator.Alloc(nn::hio::CTR::WORKMEMORY_SIZE);
    nn::hio::Initialize(hioWorkMemory);

    NW_DEMO_TEST_LOOP(&s_DeviceAllocator, NULL, &s_pRenderSystem)
    {
        InitializeGraphics();

        // Initializes the layout library.
        // The demos allocate all memory from device memory, so the same allocator gets passed.
        // 
        // To handle this and make a distinction with heap memory, be sure to specify device memory for the second argument.
        nw::lyt::Initialize(&s_DeviceAllocator, &s_DeviceAllocator);

        nw::viewer::sys::App& app = nw::viewer::sys::App::GetInstance();

        // Enables the layout module of the view library.
        // This processing occurs before app.Init.
        app.GetModuleLYT().SetModuleEnable();

        // Performs initialization.
        app.Init(&s_DeviceAllocator, &s_DeviceAllocator);

        DemoScene();

        TerminateGraphics();
    }

    nn::hio::Finalize();
    s_DeviceAllocator.Free(hioWorkMemory);

    nw::demo::PadFactory::Finalize();

    nw::demo::FinalizeGraphicsSystem();

    nn::applet::CloseApplication();
}
#ifdef NW_PLATFORM_CTR
#if NN_CURRENT_VERSION_NUMBER >= NN_VERSION_NUMBER(4,0,0,0)
#pragma diag_default 2530 // padding added to end of struct.
#endif
#endif
