/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#ifndef NW4R_Protocol3D_H_
#define NW4R_Protocol3D_H_

#pragma warn_padding on

/*
    Viewer-3DEditor Communication Protocol
	
    Communication message type
        Viewer Message  Message sent from the viewer to the editor
        Editor Message  Message sent from the editor to the viewer
        Error Message    Message sent when an error occurs
		
    When communicating, be sure to send the MessageHeader and then send the MessageData.
    MessageHeader is composed of a 2-byte MessageType and a 2-byte MessageSize.
	
    +---------------------------+--------------------+
    |      MessageHeader        |    MessageData     |
    +-------------+-------------+                    |
    | MessageType | MessageSize | (MessageSize byte) |
    | (2byte)     | (2byte)     |                    |
    +-------------+-------------+--------------------+
	
    MessageType comprises a 1-bit ErrorFlag, a 1-bit ReplyFlag, a 6-bit Category, and an 8-bit Command.
	
    
    15             8 7             0
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    |E|R| Category  |    Command    |
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    When Error is TRUE, Reply is always FALSE.
*/
namespace NW4R{
namespace Protocol3D{

	//-----------------------------------------------------------------------
	/* Version Information Definition */
	/* Protocol version. */
	const unsigned long	PROTOCOL_VERSION =	0x01040000;
	/* Major protocol version mask. */
	const unsigned long	PROTOCOL_VERSION_MAJOR_MASK =	0xFF000000;
	/* Minor protocol version mask. */
	const unsigned long	PROTOCOL_VERSION_MINOR_MASK =	0x00FF0000;
	/* Protocol revision mask.         */
	const unsigned long	PROTOCOL_VERSION_REVISION_MASK =	0x0000FF00;

	//-----------------------------------------------------------------------
	/* Message flag and mask definition */
	/* Flag for errors   */
	const unsigned short	HEADER_ERROR_FLAG =	0x8000;
	/* Flag for responses     */
	const unsigned short	HEADER_REPLY_FLAG =	0x4000;
	/* Mask for categories */
	const unsigned short	HEADER_CATEGORY_MASK =	0x3F00;
	/* Mask for commands */
	const unsigned short	HEADER_COMMAND_MASK =	0x00FF;

	//-----------------------------------------------------------------------
	/* Category definitions */
	/* System-related categories       */
	const unsigned short	CATEGORY_SYSTEM =	0x0000;
	/* Data management-related category     */
	const unsigned short	CATEGORY_DATA =	0x0100;
	/* Model State Change Category   */
	const unsigned short	CATEGORY_MODEL =	0x0200;
	/* Environment setting change category       */
	const unsigned short	CATEGORY_ENV =	0x0300;
	/* Scene state change category   */
	const unsigned short	CATEGORY_SCENE =	0x0400;
	/* Parameter edit category     */
	const unsigned short	CATEGORY_EDIT =	0x0500;
	/* Animation control category */
	const unsigned short	CATEGORY_ANIMCTRL =	0x0600;

	//-----------------------------------------------------------------------
	/* 3D protocol message shared header */
	typedef struct {
		/* Message types.    */
		unsigned short	type;
		/* Message size    */
		unsigned short	size;
	} Header;

	//-----------------------------------------------------------------------
	/* Editor system-related category message */
	/* Version request command */
	const unsigned short	ED_SYSTEM_VERSION =	0x00 + CATEGORY_SYSTEM + HEADER_REPLY_FLAG;
	/* State reset command   */
	const unsigned short	ED_SYSTEM_RESET =	0x01 + CATEGORY_SYSTEM;
	/* Synchronization command           */
	const unsigned short	ED_SYSTEM_SYNC =	0x02 + CATEGORY_SYSTEM + HEADER_REPLY_FLAG;
	/* Stop render update         */
	const unsigned short	ED_SYSTEM_FREEZE =	0x03 + CATEGORY_SYSTEM;
	/* NOP command            */
	const unsigned short	ED_SYSTEM_NOP =	0x04 + CATEGORY_SYSTEM;

	//-----------------------------------------------------------------------
	/* Editor data management-related category message */
	/* Model data load command         */
	const unsigned short	ED_DATA_LOAD_MDL =	0x00 + CATEGORY_DATA + HEADER_REPLY_FLAG;
	/* Animation data load command */
	const unsigned short	ED_DATA_LOAD_ANIM =	0x01 + CATEGORY_DATA + HEADER_REPLY_FLAG;
	/* Texture data load command     */
	const unsigned short	ED_DATA_LOAD_TEX =	0x02 + CATEGORY_DATA + HEADER_REPLY_FLAG;
	/* Palette data load command       */
	const unsigned short	ED_DATA_LOAD_PLTT =	0x03 + CATEGORY_DATA + HEADER_REPLY_FLAG;
	/* Scene data load command         */
	const unsigned short	ED_DATA_LOAD_SCENE =	0x04 + CATEGORY_DATA + HEADER_REPLY_FLAG;

	//-----------------------------------------------------------------------
	/* Editor model state change category message */
	/* Model display position setting command                */
	const unsigned short	ED_MODEL_SET_SRT =	0x00 + CATEGORY_MODEL;
	/* Action set setting command              */
	const unsigned short	ED_MODEL_ACTION_SET =	0x01 + CATEGORY_MODEL;
	/* Node show/hide setting change command       */
	const unsigned short	ED_MODEL_NODE_VISIBILITY =	0x02 + CATEGORY_MODEL;
	/* Node billboard setting change command        */
	const unsigned short	ED_MODEL_NODE_BILLBOARD =	0x03 + CATEGORY_MODEL;
	/* Material display face setting change command        */
	const unsigned short	ED_MODEL_MAT_DISPLAY_FACE =	0x04 + CATEGORY_MODEL;
	/* Material color change command              */
	const unsigned short	ED_MODEL_MAT_COLOR =	0x05 + CATEGORY_MODEL;
	/* TEV color register change command             */
	const unsigned short	ED_MODEL_MAT_TEV_COLOR =	0x06 + CATEGORY_MODEL;
	/* TEV constant register change command       */
	const unsigned short	ED_MODEL_MAT_TEV_CONSTANT =	0x07 + CATEGORY_MODEL;
	/* Light enable/disable change command              */
	const unsigned short	ED_MODEL_MAT_CHAN_LIGHT_ENABLE =	0x08 + CATEGORY_MODEL;
	/* Ambient color source change command    */
	const unsigned short	ED_MODEL_MAT_CHAN_AMB_SOURCE =	0x09 + CATEGORY_MODEL;
	/* Material color source change command      */
	const unsigned short	ED_MODEL_MAT_CHAN_MAT_SOURCE =	0x0A + CATEGORY_MODEL;
	/* TEV swap table change command           */
	const unsigned short	ED_MODEL_MAT_TEV_SWAP_TABLE =	0x0B + CATEGORY_MODEL;
	/* TEV stage count change command                       */
	const unsigned short	ED_MODEL_MAT_TEV_STAGE_NUM =	0x0C + CATEGORY_MODEL;
	/* TEV order change command                   */
	const unsigned short	ED_MODEL_MAT_TEV_ORDER =	0x0D + CATEGORY_MODEL;
	/* TEV swap change command                   */
	const unsigned short	ED_MODEL_MAT_TEV_SWAP =	0x0E + CATEGORY_MODEL;
	/* TEV color stage change command             */
	const unsigned short	ED_MODEL_MAT_TEV_COLOR_STAGE =	0x0F + CATEGORY_MODEL;
	/* TEV alpha stage change command           */
	const unsigned short	ED_MODEL_MAT_TEV_ALPHA_STAGE =	0x10 + CATEGORY_MODEL;
	/* Translucent/opaque flag change command           */
	const unsigned short	ED_MODEL_MAT_TRANSPARENCY_MODE =	0x11 + CATEGORY_MODEL;
	/* Alpha compare change command              */
	const unsigned short	ED_MODEL_MAT_PE_ALPHA_COMPARE =	0x12 + CATEGORY_MODEL;
	/* Destination alpha change command    */
	const unsigned short	ED_MODEL_MAT_PE_DST_ALPHA =	0x13 + CATEGORY_MODEL;
	/* Z compare change command                     */
	const unsigned short	ED_MODEL_MAT_PE_Z_COMPARE =	0x14 + CATEGORY_MODEL;
	/* Blend mode change command                */
	const unsigned short	ED_MODEL_MAT_PE_BLEND_MODE =	0x15 + CATEGORY_MODEL;
	/* Texture map information change command          */
	const unsigned short	ED_MODEL_MAT_TEX_MAP =	0x16 + CATEGORY_MODEL;
	/* Texture coordinate count change command              */
	const unsigned short	ED_MODEL_MAT_TEX_COORD_SIZE =	0x17 + CATEGORY_MODEL;
	/* Texture coordinate information change command            */
	const unsigned short	ED_MODEL_MAT_TEX_COORD =	0x18 + CATEGORY_MODEL;
	/* Texture matrix information change command            */
	const unsigned short	ED_MODEL_MAT_TEX_MATRIX =	0x19 + CATEGORY_MODEL;
	/* Texture image change command            */
	const unsigned short	ED_MODEL_MAT_TEX_MAP_IMAGE =	0x20 + CATEGORY_MODEL;
	/* Indirect matrix SRT change command         */
	const unsigned short	ED_MODEL_MAT_INDMTX_SRT =	0x21 + CATEGORY_MODEL;
	/* Fur lie matrix change command                 */
	const unsigned short	ED_MODEL_MAT_FUR_INDMTX_SR =	0x22 + CATEGORY_MODEL;

	//-----------------------------------------------------------------------
	/* Display setting change category message */
	/* Selected object emphasized display command          */
	const unsigned short	ED_ENV_SET_APPEAL =	0x00 + CATEGORY_ENV;
	/* Selected object emphasis setting change command      */
	const unsigned short	ED_ENV_CONFIG_APPEAL =	0x01 + CATEGORY_ENV;
	/* Operation target camera/light change command         */
	const unsigned short	ED_ENV_SET_CURRENT_CONTROL =	0x02 + CATEGORY_ENV;

	//-----------------------------------------------------------------------
	/* Editor scene state change category message */
	/* Active scene change command                */
	const unsigned short	ED_SCENE_SET_ACTIVE =	0x00 + CATEGORY_SCENE;
	/* ROTATE camera position get command              */
	const unsigned short	ED_SCENE_GET_CAMERA_ROT_POS =	0x01 + CATEGORY_SCENE + HEADER_REPLY_FLAG;
	/* AIM camera position get command                 */
	const unsigned short	ED_SCENE_GET_CAMERA_AIM_POS =	0x02 + CATEGORY_SCENE + HEADER_REPLY_FLAG;
	/* Light position get command                    */
	const unsigned short	ED_SCENE_GET_LIGHT_POS =	0x03 + CATEGORY_SCENE + HEADER_REPLY_FLAG;
	/* In-scene object setting swap command  */
	const unsigned short	ED_SCENE_SWAP_OBJECT =	0x04 + CATEGORY_SCENE;
	/* View camera position settings command              */
	const unsigned short	ED_SCENE_SET_VIEW_CAMERA_POS =	0x05 + CATEGORY_SCENE;

	//-----------------------------------------------------------------------
	/* Editor parameter edit category message */
	/* Edit object selection state change command  */
	const unsigned short	ED_EDIT_SELECTION =	0x00 + CATEGORY_EDIT;
	/* int value change command                     */
	const unsigned short	ED_EDIT_INT =	0x11 + CATEGORY_EDIT;
	/* bool value change command                    */
	const unsigned short	ED_EDIT_BOOL =	0x12 + CATEGORY_EDIT;
	/* float value change command                   */
	const unsigned short	ED_EDIT_FLOAT =	0x13 + CATEGORY_EDIT;
	/* vector value change command                  */
	const unsigned short	ED_EDIT_VECTOR =	0x14 + CATEGORY_EDIT;
	/* color value change command                   */
	const unsigned short	ED_EDIT_COLOR =	0x15 + CATEGORY_EDIT;
	/* colorS10 value change command                */
	const unsigned short	ED_EDIT_COLORS10 =	0x16 + CATEGORY_EDIT;
	/* int value const animation change command    */
	const unsigned short	ED_EDIT_CONSTANIM_INT =	0x21 + CATEGORY_EDIT;
	/* bool value const animation change command   */
	const unsigned short	ED_EDIT_CONSTANIM_BOOL =	0x22 + CATEGORY_EDIT;
	/* float value const animation change command  */
	const unsigned short	ED_EDIT_CONSTANIM_FLOAT =	0x23 + CATEGORY_EDIT;
	/* vector value const animation change command */
	const unsigned short	ED_EDIT_CONSTANIM_VECTOR =	0x24 + CATEGORY_EDIT;
	/* color value const animation change command  */
	const unsigned short	ED_EDIT_CONSTANIM_COLOR =	0x25 + CATEGORY_EDIT;
	/* int value step animation change command    */
	const unsigned short	ED_EDIT_STEPANIM_INT =	0x31 + CATEGORY_EDIT;
	/* bool value step animation change command   */
	const unsigned short	ED_EDIT_STEPANIM_BOOL =	0x32 + CATEGORY_EDIT;
	/* float value step animation change command  */
	const unsigned short	ED_EDIT_STEPANIM_FLOAT =	0x33 + CATEGORY_EDIT;
	/* vector value step animation change command */
	const unsigned short	ED_EDIT_STEPANIM_VECTOR =	0x34 + CATEGORY_EDIT;
	/* color value step animation change command  */
	const unsigned short	ED_EDIT_STEPANIM_COLOR =	0x35 + CATEGORY_EDIT;
	/* int value hermite animation change command    */
	const unsigned short	ED_EDIT_HERMTANIM_INT =	0x41 + CATEGORY_EDIT;
	/* bool value hermite animation change command   */
	const unsigned short	ED_EDIT_HERMTANIM_BOOL =	0x42 + CATEGORY_EDIT;
	/* float value hermite animation change command  */
	const unsigned short	ED_EDIT_HERMTANIM_FLOAT =	0x43 + CATEGORY_EDIT;
	/* vector value hermite animation change command */
	const unsigned short	ED_EDIT_HERMTANIM_VECTOR =	0x44 + CATEGORY_EDIT;
	/* color value hermite animation change command  */
	const unsigned short	ED_EDIT_HERMTANIM_COLOR =	0x45 + CATEGORY_EDIT;

	//-----------------------------------------------------------------------
	/* Editor animation control category message */
	/* Command to stop the animation with a specific frame    */
	const unsigned short	ED_ANIMCTRL_STOP =	0x00 + CATEGORY_ANIMCTRL;
	/* Command to play the animation from a specific frame  */
	const unsigned short	ED_ANIMCTRL_PLAY =	0x01 + CATEGORY_ANIMCTRL;

	//-----------------------------------------------------------------------
	/* Object type definition */
	/* node               */
	const unsigned short	OBJECT_TYPE_NODE =	0;
	/* shape             */
	const unsigned short	OBJECT_TYPE_SHP =	1;
	/* Material           */
	const unsigned short	OBJECT_TYPE_MAT =	2;
	/* model               */
	const unsigned short	OBJECT_TYPE_MDL =	3;
	/* Camera               */
	const unsigned short	OBJECT_TYPE_CAMERA =	4;
	/* Lights               */
	const unsigned short	OBJECT_TYPE_LIGHT =	5;
	/* Ambient light   */
	const unsigned short	OBJECT_TYPE_AMB_LIGHT =	6;
	/* Light set         */
	const unsigned short	OBJECT_TYPE_LIGHT_SET =	7;
	/* Fog               */
	const unsigned short	OBJECT_TYPE_FOG =	8;

	//-----------------------------------------------------------------------
	/*
         Version request response message
            Receive editor protocol version, and return viewer protocol version.
         */
	typedef struct {
		/* 4B: Version number */
		unsigned long	version;
	} EdSystemVersion;

	//-----------------------------------------------------------------------
	/*
        Synchronization message
              When receiving this synchronization command, the Viewer returns to the application the response command where the specified tagName name is stored.
              
              tagName can be freely used by the application.
         

*/
	typedef struct {
		/* 4B: Tag name. */
		unsigned long	tagName;
	} EdSysSync;

	//-----------------------------------------------------------------------
	/*
         Render update stop message
              When the Viewer receives this command as freezeFlg = true, 3D-related rendering does not update until the same command is received next with freezeFlg = false.
              
         

*/
	typedef struct {
		/* 4B: Rendering update stop flag */
		unsigned long	freezeFlg;
	} EdSysFreeze;

	//-----------------------------------------------------------------------
	/*
            Load model data message
            When fileSize is 0, the data that has that model ID is considered to be destroyed.
            When the model data is destroyed, the associated animation data is destroyed at the same time.
            
            When loading succeeds, return the BODY size is 0 success response message.
            If unsuccessful, return the ER_DATA_LOAD command.
         

*/
	typedef struct {
		/*   4B: Model ID         */
		unsigned long	id;
		/*   4B: File size.   */
		unsigned long	fileSize;
		/* 260B: Path name of the data   */
		unsigned char	pathName[260];
	} EdDataLoadMdl;

	//-----------------------------------------------------------------------
	/* Animation type */
	enum AnimType {
		/* Visibility animation       */
		ANIM_VIA =	0,
		/* Color animation             */
		ANIM_CLA =	1,
		/* Texture pattern animation */
		ANIM_TPA =	2,
		/* Texture SRT Animation      */
		ANIM_TSA =	3,
		/* Shape animation.           */
		ANIM_SHA =	4,
		/* Character animation         */
		ANIM_CHA =	5,
		/* Character animation 1        */
		ANIM_CHA1 =	6,
		/* Character animation 2        */
		ANIM_CHA2 =	7,
		/* Character animation 3        */
		ANIM_CHA3 =	8,
		/* Define enum as 4-byte type */
		AnimType_4BYTE_ENUM =	0xFFFFFFFF
	};

	//-----------------------------------------------------------------------
	/*
            Animation data load message
            When fileSize is 0, the animation data that has that action ID is considered to be destroyed.
            When loading succeeds, return the BODY size is 0 success response message.
            If unsuccessful, return the ER_DATA_LOAD command.
         

*/
	typedef struct {
		/*   4B: Animation type   */
		AnimType	animType;
		/*   4B: Model ID to bind   */
		unsigned long	modelID;
		/*   4B: Action ID           */
		unsigned long	actionID;
		/*   4B: File size.         */
		unsigned long	fileSize;
		/* 260B: Path name of the data         */
		unsigned char	pathName[260];
	} EdDataLoadAnim;

	//-----------------------------------------------------------------------
	/*
            Texture data load message
            When fileSize is 0, the texture data that has that ID is considered to be destroyed.
            When loading succeeds, return the BODY size is 0 success response message.
            If unsuccessful, return the ER_DATA_LOAD command.
         

*/
	typedef struct {
		/*   4B: Texture ID           */
		unsigned long	texID;
		/*   4B: File size.         */
		unsigned long	fileSize;
		/* 260B: Path name of the data         */
		unsigned char	pathName[260];
	} EdDataLoadTex;

	//-----------------------------------------------------------------------
	/*
            Palette data load message
            When fileSize is 0, the palette data that has that ID is considered to be destroyed.
            When loading succeeds, return the BODY size is 0 success response message.
            If unsuccessful, return the ER_DATA_LOAD command.
         

*/
	typedef struct {
		/*   4B: Palette ID             */
		unsigned long	plttID;
		/*   4B: File size.         */
		unsigned long	fileSize;
		/* 260B: Path name of the data         */
		unsigned char	pathName[260];
	} EdDataLoadPltt;

	//-----------------------------------------------------------------------
	/*
            Scene data load message
            When fileSize is 0, the scene is considered to be destroyed.
            When loading succeeds, return the BODY size is 0 success response message.
            If unsuccessful, return the ER_DATA_LOAD command.
         

*/
	typedef struct {
		/*   4B: Scene ID         */
		unsigned long	id;
		/*   4B: File size.   */
		unsigned long	fileSize;
		/* 260B: Path name of the data   */
		unsigned char	pathName[260];
	} EdDataLoadScene;

	//-----------------------------------------------------------------------
	/*
            Model display position setting message
         */
	typedef struct {
		/*   4B: Rotation X-component (unit: degrees)            */
		float	rotate_x;
		/*   4B: Rotation Y-component (unit: degrees)            */
		float	rotate_y;
		/*   4B: Rotation Z-component (unit: degrees)            */
		float	rotate_z;
		/*   4B: Translation X-component                */
		float	trans_x;
		/*   4B: Translation Y-component                */
		float	trans_y;
		/*   4B: Translation Z-component                */
		float	trans_z;
		/*   4B: Scale X-component                      */
		float	scale_x;
		/*   4B: Scale Y-component                      */
		float	scale_y;
		/*   4B: Scale Z-component                      */
		float	scale_z;
		/*   4B: Visibility                       */
		unsigned long	visibility;
		/*   4B: Bind target model                 */
		unsigned long	bind_model;
		/*   4B: Bind target node                 */
		long	bind_node;
		/*   4B: Character animation translation component ignore setting */
		unsigned long	ignore_anmchr_trans;
		/*   4B: Wireframe visibility     */
		unsigned long	wireframe_visibility;
	} EdModelSetSRT;

	//-----------------------------------------------------------------------
	/*
            Action set element definition
         */
	typedef struct {
		/*   4B: Animation ID  */
		unsigned long	actionID;
		/*   4B: loop count        */
		unsigned long	loopNum;
	} ActionSetElement;

	//-----------------------------------------------------------------------
	/*
            Action set setting message
            When num is 0, the action set data is considered to be destroyed.
         
*/
	typedef struct {
		/*   4B: Number of elements included in the action set          */
		unsigned long	num;
		/*   4B: Existence of loop in the entire action set      */
		unsigned long	loopFlag;
	} EdModelActionSet;

	//-----------------------------------------------------------------------
	/*
            Node show/hide setting change message
         */
	typedef struct {
		/*   4B: Node display flag   */
		unsigned long	visible;
	} EdModelNodeVisibility;

	//-----------------------------------------------------------------------
	/* Billboard setting definition */
	enum BillboardType {
		/* Billboard off        */
		BILLBOARD_OFF =	0,
		/* std Billboard        */
		BILLBOARD_STD =	1,
		/* persp_std billboard  */
		BILLBOARD_PERSP_STD =	2,
		/* rot billboard        */
		BILLBOARD_ROT =	3,
		/* persp_rot billboard  */
		BILLBOARD_PERSP_ROT =	4,
		/* y billboard          */
		BILLBOARD_Y =	5,
		/* persp_y billboard    */
		BILLBOARD_PERSP_Y =	6,
		/* Define enum as 4-byte type */
		BillboardType_4BYTE_ENUM =	0xFFFFFFFF
	};

	//-----------------------------------------------------------------------
	/*
            Node billboard setting change message
         */
	typedef struct {
		/*   2B: Node billboard setting   */
		BillboardType	billboardType;
	} EdModelNodeBillboard;

	//-----------------------------------------------------------------------
	/* Material display face setting definition */
	enum DisplayFaceType {
		/* Display surface only         */
		DISPLAY_FACE_FRONT =	0,
		/* Display back face only         */
		DISPLAY_FACE_BACK =	1,
		/* Display Both Faces             */
		DISPLAY_FACE_BOTH =	2,
		/* Do Not Display           */
		DISPLAY_FACE_NONE =	3,
		/* Define enum as 4-byte type */
		DisplayFaceType_4BYTE_ENUM =	0xFFFFFFFF
	};

	//-----------------------------------------------------------------------
	/*
            Material display face setting change message
         */
	typedef struct {
		/*   4B: Material display face setting   */
		DisplayFaceType	displayFace;
	} EdModelMatDisplayFace;

	//-----------------------------------------------------------------------
	/*
            Material color structure
         */
	typedef struct {
		/*   1B: R element   */
		unsigned char	r;
		/*   1B: G element   */
		unsigned char	g;
		/*   1B: B element   */
		unsigned char	b;
		/*   1B: A element   */
		unsigned char	a;
	} MatColor;

	//-----------------------------------------------------------------------
	/*
            S10 material color structure
         */
	typedef struct {
		/*   2B: R element   */
		short	r;
		/*   2B: G element   */
		short	g;
		/*   2B: B element   */
		short	b;
		/*   2B: A element   */
		short	a;
	} MatColorS10;

	//-----------------------------------------------------------------------
	/* Material color type */
	enum MatColorType {
		/* Material color ０         */
		MAT_COLOR_TYPE_MAT0 =	0,
		/* Ambient color ０       */
		MAT_COLOR_TYPE_AMB0 =	1,
		/* Material color １         */
		MAT_COLOR_TYPE_MAT1 =	2,
		/* Ambient color 1       */
		MAT_COLOR_TYPE_AMB1 =	3,
		/* Define enum as 4-byte type */
		MatColorType_4BYTE_ENUM =	0xFFFFFFFF
	};

	//-----------------------------------------------------------------------
	/*
            Material color change message
         */
	typedef struct {
		/*   4B: Material color type   */
		MatColorType	type;
		/*   4B: Color                   */
		MatColor	color;
	} EdModelMatColor;

	//-----------------------------------------------------------------------
	/*
            TEV color register change message
         */
	typedef struct {
		/*   4B: TEV color register index   */
		unsigned long	index;
		/*   8B: Color                        */
		MatColorS10	color;
	} EdModelMatTevColor;

	//-----------------------------------------------------------------------
	/*
            TEV constant register change message
         */
	typedef struct {
		/*   4B: TEV constant register index   */
		unsigned long	index;
		/*   4B: Color                              */
		MatColor	color;
	} EdModelMatTevConstant;

	//-----------------------------------------------------------------------
	/* Channel type */
	enum MatChannelType {
		/* Color ０         */
		CHANNEL_TYPE_COLOR0 =	0,
		/* Alpha ０       */
		CHANNEL_TYPE_ALPHA0 =	1,
		/* Color 1         */
		CHANNEL_TYPE_COLOR1 =	2,
		/* Alpha 1       */
		CHANNEL_TYPE_ALPHA1 =	3,
		/* Define enum as 4-byte type */
		MatChannelType_4BYTE_ENUM =	0xFFFFFFFF
	};

	//-----------------------------------------------------------------------
	/*
            Light enable/disable change message
         */
	typedef struct {
		/*   4B: Targeted channel   */
		MatChannelType	channel;
		/*   4B: Enable/disable flag         */
		unsigned long	enable;
	} EdModelMatChanLightEnable;

	//-----------------------------------------------------------------------
	/*
            Ambient color source change message
         */
	typedef struct {
		/*   4B: Targeted channel   */
		MatChannelType	channel;
		/*   4B: Vertex color use flag   */
		unsigned long	useVtxColor;
	} EdModelMatChanAmbSource;

	//-----------------------------------------------------------------------
	/*
            Material color source change message
         */
	typedef struct {
		/*   4B: Targeted channel   */
		MatChannelType	channel;
		/*   4B: Vertex color use flag   */
		unsigned long	useVtxColor;
	} EdModelMatChanMatSource;

	//-----------------------------------------------------------------------
	/*
            TEV swap table change message
         */
	typedef struct {
		/*   4B: Targeted swap table    */
		unsigned long	index;
		/*   1B: R Component                         */
		unsigned char	r;
		/*   1B: G Component                         */
		unsigned char	g;
		/*   1B: B Component                         */
		unsigned char	b;
		/*   1B: A Component                         */
		unsigned char	a;
	} EdModelMatTevSwapTable;

	//-----------------------------------------------------------------------
	/*
            TEV stage count change message
         */
	typedef struct {
		/*   4B: TEV stage count    */
		unsigned long	num;
	} EdModelMatTevStageNum;

	//-----------------------------------------------------------------------
	/*  */
	enum TevRasOrderType {
		/* Do not input rasterized color */
		TEV_RAS_ORDER_TYPE_COLOR_NULL,
		/* Color channel 0              */
		TEV_RAS_ORDER_TYPE_COLOR0A0,
		/* Color channel 1              */
		TEV_RAS_ORDER_TYPE_COLOR1A1,
		/* Zero                           */
		TEV_RAS_ORDER_TYPE_COLOR_ZERO,
		/* Define enum as 4-byte type */
		TevRasOrderType_4BYTE_ENUM =	0xFFFFFFFF
	};

	//-----------------------------------------------------------------------
	/*
            TEV order change message
         */
	typedef struct {
		/*   4B: Targeted TEV stage    */
		unsigned long	stageIndex;
		/*   4B: Input rasterization       */
		TevRasOrderType	rasOrder;
		/*   4B: Texture map input   */
		long	texMapOrder;
		/*   4B: Texture coordinate input     */
		long	texCoordOrder;
	} EdModelMatTevOrder;

	//-----------------------------------------------------------------------
	/*
            TEV swap change message
         */
	typedef struct {
		/*   4B: Targeted TEV stage      */
		unsigned long	stageIndex;
		/*   4B: Swap rasterization     */
		unsigned long	rasSwap;
		/*   4B: Swap texture color */
		unsigned long	texSwap;
	} EdModelMatTevSwap;

	//-----------------------------------------------------------------------
	/* TEV input color operand type */
	enum TevColorArg {
		/* RGB color from TEVREG0 register              */
		TEV_COLOR_ARG_C0,
		/* RGB color from TEVREG1 register              */
		TEV_COLOR_ARG_C1,
		/* RGB color from TEVREG2 register              */
		TEV_COLOR_ARG_C2,
		/* RGB color from TEVPREV register              */
		TEV_COLOR_ARG_CPREV,
		/* Alpha from TEVREG0 register               */
		TEV_COLOR_ARG_A0,
		/* Alpha from TEVREG1 register               */
		TEV_COLOR_ARG_A1,
		/* Alpha from TEVREG2 register               */
		TEV_COLOR_ARG_A2,
		/* Alpha from TEVPREV register               */
		TEV_COLOR_ARG_APREV,
		/* RGB color from texture                   */
		TEV_COLOR_ARG_TEXC,
		/* Alpha from texture                    */
		TEV_COLOR_ARG_TEXA,
		/* RGB color from rasterized vertex color */
		TEV_COLOR_ARG_RASC,
		/* Alpha from rasterized vertex color  */
		TEV_COLOR_ARG_RASA,
		/* 1.0                                         */
		TEV_COLOR_ARG_ONE,
		/* 0.5                                         */
		TEV_COLOR_ARG_HALF,
		/* 0.0                                         */
		TEV_COLOR_ARG_ZERO,
		/* Constant color input                              */
		TEV_COLOR_ARG_KONST,
		/* Define enum as 4-byte type */
		TevColorArg_4BYTE_ENUM =	0xFFFFFFFF
	};

	//-----------------------------------------------------------------------
	/* TEV constant color type */
	enum TevColorConstant {
		/* Constant color register 0         */
		TEV_COLOR_CONSTANT_K0,
		/* Constant color register 0 (RRR)    */
		TEV_COLOR_CONSTANT_K0_R,
		/* Constant color register 0 (GGG)    */
		TEV_COLOR_CONSTANT_K0_G,
		/* Constant color register 0 (BBB)    */
		TEV_COLOR_CONSTANT_K0_B,
		/* Constant color register 0 (AAA)    */
		TEV_COLOR_CONSTANT_K0_A,
		/* Constant color register 1         */
		TEV_COLOR_CONSTANT_K1,
		/* Constant color register 1 (RRR)    */
		TEV_COLOR_CONSTANT_K1_R,
		/* Constant color register 1 (GGG)    */
		TEV_COLOR_CONSTANT_K1_G,
		/* Constant color register 1 (BBB)    */
		TEV_COLOR_CONSTANT_K1_B,
		/* Constant color register 1 (AAA)    */
		TEV_COLOR_CONSTANT_K1_A,
		/* Constant color register 2         */
		TEV_COLOR_CONSTANT_K2,
		/* Constant color register 2 (RRR)    */
		TEV_COLOR_CONSTANT_K2_R,
		/* Constant color register 2 (GGG)    */
		TEV_COLOR_CONSTANT_K2_G,
		/* Constant color register 2 (BBB)    */
		TEV_COLOR_CONSTANT_K2_B,
		/* Constant color register 2 (AAA)    */
		TEV_COLOR_CONSTANT_K2_A,
		/* Constant color register 3         */
		TEV_COLOR_CONSTANT_K3,
		/* Constant color register 3 (RRR)    */
		TEV_COLOR_CONSTANT_K3_R,
		/* Constant color register 3 (GGG)    */
		TEV_COLOR_CONSTANT_K3_G,
		/* Constant color register 3 (BBB)    */
		TEV_COLOR_CONSTANT_K3_B,
		/* Constant color register 3 (AAA)    */
		TEV_COLOR_CONSTANT_K3_A,
		/* Fixed constant 8/8                */
		TEV_COLOR_CONSTANT_8_8,
		/* Fixed constant 7/8                */
		TEV_COLOR_CONSTANT_7_8,
		/* Fixed constant 6/8                */
		TEV_COLOR_CONSTANT_6_8,
		/* Fixed constant 5/8                */
		TEV_COLOR_CONSTANT_5_8,
		/* Fixed constant 4/8                */
		TEV_COLOR_CONSTANT_4_8,
		/* Fixed constant 3/8                */
		TEV_COLOR_CONSTANT_3_8,
		/* Fixed constant 2/8                */
		TEV_COLOR_CONSTANT_2_8,
		/* Fixed constant 1/8                */
		TEV_COLOR_CONSTANT_1_8,
		/* Define enum as 4-byte type */
		TevColorConstant_4BYTE_ENUM =	0xFFFFFFFF
	};

	//-----------------------------------------------------------------------
	/* TEV calculation type */
	enum TevColorOp {
		/* add      */
		TEV_COLOR_OP_ADD,
		/* Sub      */
		TEV_COLOR_OP_SUB,
		/* r8_gt    */
		TEV_COLOR_OP_COMP_R8_GT,
		/* r8_eq    */
		TEV_COLOR_OP_COMP_R8_EQ,
		/* gr16_gt  */
		TEV_COLOR_OP_COMP_GR16_GT,
		/* gr16_eq  */
		TEV_COLOR_OP_COMP_GR16_EQ,
		/* bgr24_gt */
		TEV_COLOR_OP_COMP_BGR24_GT,
		/* bgr24_eq */
		TEV_COLOR_OP_COMP_BGR24_EQ,
		/* rgb8_gt  */
		TEV_COLOR_OP_COMP_RGB8_GT,
		/* rgb8_eq  */
		TEV_COLOR_OP_COMP_RGB8_EQ,
		/* a8_gt    */
		TEV_COLOR_OP_COMP_A8_GT,
		/* a8_eq    */
		TEV_COLOR_OP_COMP_A8_EQ,
		/* Define enum as 4-byte type */
		TevColorOp_4BYTE_ENUM =	0xFFFFFFFF
	};

	//-----------------------------------------------------------------------
	/* TEV bias value */
	enum TevColorBias {
		/* 0      */
		TEV_BIAS_ZERO,
		/* 0.5    */
		TEV_BIAS_ADD_HALF,
		/* -0.5   */
		TEV_BIAS_SUB_HALF,
		/* Define enum as 4-byte type */
		TevColorBias_4BYTE_ENUM =	0xFFFFFFFF
	};

	//-----------------------------------------------------------------------
	/* TEV scale value */
	enum TevColorScale {
		/* 1      */
		TEV_SCALE_1,
		/* 2      */
		TEV_SCALE_2,
		/* 4      */
		TEV_SCALE_4,
		/* 1/2    */
		TEV_SCALE_DIVIDE_2,
		/* Define enum as 4-byte type */
		TevColorScale_4BYTE_ENUM =	0xFFFFFFFF
	};

	//-----------------------------------------------------------------------
	/*
            TEV color stage change message
         */
	typedef struct {
		/*   4B: Targeted TEV stage      */
		unsigned long	stageIndex;
		/*   4B: Input operand a-component      */
		TevColorArg	a;
		/*   4B: Input operand b-component      */
		TevColorArg	b;
		/*   4B: Input operand c-component      */
		TevColorArg	c;
		/*   4B: Input operand d-component      */
		TevColorArg	d;
		/*   4B: TEV constant color              */
		TevColorConstant	constant;
		/*   4B: TEV calculation                    */
		TevColorOp	op;
		/*   4B: TEV bias value              */
		TevColorBias	bias;
		/*   4B: TEV scale value              */
		TevColorScale	scale;
		/*   4B: TEV clamp                */
		unsigned long	clamp;
		/*   4B: Output register number           */
		unsigned long	outReg;
	} EdModelMatTevColorStage;

	//-----------------------------------------------------------------------
	/*
            TEV alpha stage change message
         */
	typedef struct {
		/*   4B: Targeted TEV stage      */
		unsigned long	stageIndex;
		/*   4B: Input operand a-component      */
		TevColorArg	a;
		/*   4B: Input operand b-component      */
		TevColorArg	b;
		/*   4B: Input operand c-component      */
		TevColorArg	c;
		/*   4B: Input operand d-component      */
		TevColorArg	d;
		/*   4B: TEV constant color              */
		TevColorConstant	constant;
		/*   4B: TEV calculation                    */
		TevColorOp	op;
		/*   4B: TEV bias value              */
		TevColorBias	bias;
		/*   4B: TEV scale value              */
		TevColorScale	scale;
		/*   4B: TEV clamp                */
		unsigned long	clamp;
		/*   4B: Output register number           */
		unsigned long	outReg;
	} EdModelMatTevAlphaStage;

	//-----------------------------------------------------------------------
	/*
            Translucent/opaque flag change message
         */
	typedef struct {
		/*   4B: Translucent flag      */
		unsigned long	xluFlag;
	} EdModelMatTransparencyMode;

	//-----------------------------------------------------------------------
	/* PE compare function */
	enum PeCompareFunc {
		/* never          */
		PE_COMPARE_FUNC_NEVER,
		/* less           */
		PE_COMPARE_FUNC_LESS,
		/* less equal     */
		PE_COMPARE_FUNC_LEQUAL,
		/* equal          */
		PE_COMPARE_FUNC_EQUAL,
		/* not equal      */
		PE_COMPARE_FUNC_NEQUAL,
		/* greater equal  */
		PE_COMPARE_FUNC_GEQUAL,
		/* greater        */
		PE_COMPARE_FUNC_GREATER,
		/* always         */
		PE_COMPARE_FUNC_ALWAYS,
		/* Define enum as 4-byte type */
		PeCompareFunc_4BYTE_ENUM =	0xFFFFFFFF
	};

	//-----------------------------------------------------------------------
	/* PE alpha calculation */
	enum PeAlphaOp {
		/* and        */
		PE_COMPARE_OP_AND,
		/* or         */
		PE_COMPARE_OP_OR,
		/* xor        */
		PE_COMPARE_OP_XOR,
		/* xnor       */
		PE_COMPARE_OP_XNOR,
		/* Define enum as 4-byte type */
		PeAlphaOp_4BYTE_ENUM =	0xFFFFFFFF
	};

	//-----------------------------------------------------------------------
	/*
            Alpha compare change message
         */
	typedef struct {
		/*   4B: Comparison subfunction0             */
		PeCompareFunc	comp0;
		/*   4B: Reference value for subfunction0    */
		unsigned long	ref0;
		/*   4B: Operator for the combination of subfunction0 and subfunction1 */
		PeAlphaOp	op;
		/*   4B: Comparison subfunction1             */
		PeCompareFunc	comp1;
		/*   4B: Reference value for subfunction1    */
		unsigned long	ref1;
	} EdModelMatPeAlphaCompare;

	//-----------------------------------------------------------------------
	/*
            Destination alpha change message
         */
	typedef struct {
		/*   4B: Destination alpha enable flag */
		unsigned long	enable;
		/*   4B: Alpha value for destination alpha       */
		unsigned long	alpha;
	} EdModelMatPeDstAlpha;

	//-----------------------------------------------------------------------
	/*
            Z compare change message
         */
	typedef struct {
		/*   4B: Z-comparison enable flag         */
		unsigned long	enable;
		/*   4B: Z-comparison function                       */
		PeCompareFunc	func;
		/*   4B: Z-buffer change enable flag */
		unsigned long	update;
		/*   4B: Z-test timing flag         */
		unsigned long	before_tex;
	} EdModelMatPeZCompare;

	//-----------------------------------------------------------------------
	/* PE blend mode */
	enum PeBlendType {
		/* Do not perform either blend processing or logical processing */
		PE_BLEND_TYPE_NONE,
		/* Blend processing mode                       */
		PE_BLEND_TYPE_BLEND,
		/* Logical processing mode                           */
		PE_BLEND_TYPE_LOGIC,
		/* Subtraction processing mode                           */
		PE_BLEND_TYPE_SUBTRACT,
		/* Define enum as 4-byte type */
		PeBlendType_4BYTE_ENUM =	0xFFFFFFFF
	};

	//-----------------------------------------------------------------------
	/* PE blend coefficient */
	enum PeBlendFactor {
		/* 0.0                            */
		PE_BLEND_FACTOR_ZERO,
		/* 1.0                            */
		PE_BLEND_FACTOR_ONE,
		/* Source color                   */
		PE_BLEND_FACTOR_SRC_COLOR,
		/* 1.0 - Source color             */
		PE_BLEND_FACTOR_INV_SRC_COLOR,
		/* Source alpha                 */
		PE_BLEND_FACTOR_SRC_ALPHA,
		/* 1.0 - Source alpha           */
		PE_BLEND_FACTOR_INV_SRC_ALPHA,
		/* Framebuffer alpha       */
		PE_BLEND_FACTOR_DST_ALPHA,
		/* 1.0-framebuffer alpha */
		PE_BLEND_FACTOR_INV_DST_ALPHA,
		/* Framebuffer color         */
		PE_BLEND_FACTOR_DST_COLOR =	PE_BLEND_FACTOR_SRC_COLOR,
		/* 1.0-framebuffer color   */
		PE_BLEND_FACTOR_INV_DST_COLOR =	PE_BLEND_FACTOR_INV_SRC_COLOR,
		/* Define enum as 4-byte type */
		PeBlendFactor_4BYTE_ENUM =	0xFFFFFFFF
	};

	//-----------------------------------------------------------------------
	/* PE blend calculation */
	enum PeBlendOp {
		/*  dst = 0                 */
		PE_BLEND_OP_CLEAR,
		/*  dst = 1                 */
		PE_BLEND_OP_SET,
		/*  dst = src               */
		PE_BLEND_OP_COPY,
		/*  dst = ~src              */
		PE_BLEND_OP_INV_COPY,
		/*  dst = dst               */
		PE_BLEND_OP_NOOP,
		/*  dst = ~dst              */
		PE_BLEND_OP_INV,
		/*  dst = src and dst       */
		PE_BLEND_OP_AND,
		/*  dst = ~(src and dst)    */
		PE_BLEND_OP_NAND,
		/*  dst = src | dst         */
		PE_BLEND_OP_OR,
		/*  dst = ~(src | dst)      */
		PE_BLEND_OP_NOR,
		/*  dst = src ^ dst         */
		PE_BLEND_OP_XOR,
		/*  dst = ~(src ^ dst)      */
		PE_BLEND_OP_EQUIV,
		/*  dst = src and ~dst      */
		PE_BLEND_OP_REV_AND,
		/*  dst = ~src and dst      */
		PE_BLEND_OP_INV_AND,
		/*  dst = src | ~dst        */
		PE_BLEND_OP_REV_OR,
		/*  dst = ~src | dst        */
		PE_BLEND_OP_INV_OR,
		/* Define enum as 4-byte type */
		PeBlendOp_4BYTE_ENUM =	0xFFFFFFFF
	};

	//-----------------------------------------------------------------------
	/*
            Blend mode change message
         */
	typedef struct {
		/*   4B: Blending mode          */
		PeBlendType	type;
		/*   4B: Source coefficient              */
		PeBlendFactor	src_factor;
		/*   4B: Destination coefficient  */
		PeBlendFactor	dst_factor;
		/*   4B: Blending            */
		PeBlendOp	op;
	} EdModelMatPeBlendMode;

	//-----------------------------------------------------------------------
	/* Texture wrap settings */
	enum TexWrap {
		/*  Clamp (no repeat, no mirror) */
		TEX_WRAP_CLAMP,
		/*  Repeat (repeat, no mirror) */
		TEX_WRAP_REPEAT,
		/*  Mirror (repeat, mirror)   */
		TEX_WRAP_MIRROR,
		/* Define enum as 4-byte type */
		TexWrap_4BYTE_ENUM =	0xFFFFFFFF
	};

	//-----------------------------------------------------------------------
	/* Texture filter mode */
	enum TexFilter {
		/*  Point sampling, no mipmap           */
		TEX_FILTER_NEAR,
		/*  Bilinear filter, no mipmap             */
		TEX_FILTER_LINEAR,
		/*  Point sampling, mipmap not interpolated   */
		TEX_FILTER_NEAR_MIP_NEAR,
		/*  Bilinear filter, mipmap not interpolated     */
		TEX_FILTER_LINEAR_MIP_NEAR,
		/*  Point sampling, linear mipmap         */
		TEX_FILTER_NEAR_MIP_LINEAR,
		/*  Trilinear filter                             */
		TEX_FILTER_LINEAR_MIP_LINEAR,
		/* Define enum as 4-byte type */
		TexFilter_4BYTE_ENUM =	0xFFFFFFFF
	};

	//-----------------------------------------------------------------------
	/* Maximum anisotropic filter used */
	enum TexMaxAniso {
		/*  Do not use anisotropic filter           */
		TEX_MAX_ANISO_1,
		/*  Apply maximum of filters for two samples          */
		TEX_MAX_ANISO_2,
		/*  Apply maximum of filters for four samples          */
		TEX_MAX_ANISO_4,
		/* Define enum as 4-byte type */
		TexMaxAniso_4BYTE_ENUM =	0xFFFFFFFF
	};

	//-----------------------------------------------------------------------
	/*
            Texture map information change message
         */
	typedef struct {
		/*   4B: Texture map index                          */
		unsigned long	index;
		/*   4B: Texture ID to be used                                */
		unsigned long	texID;
		/*   4B: Palette ID to be used. 0xffffffff if not used            */
		unsigned long	plttID;
		/*   4B: Wrap settings in the texture horizontal (s) direction                   */
		TexWrap	wrap_s;
		/*   4B: Wrap settings in the texture vertical (t) direction                   */
		TexWrap	wrap_t;
		/*   4B: Specify whether to use mipmap              */
		unsigned long	mipmap;
		/*   4B: Filter used when reducing texels                */
		TexFilter	filterMin;
		/*   4B: Filter used when expanding texels                */
		TexFilter	filterMag;
		/*   4B: Bias added to the calculated LOD value                   */
		float	lodBias;
		/*   4B: bias clamp settings                                  */
		unsigned long	biasClamp;
		/*   4B: Flag to specify whether to calculate LOD using texel interpolation   */
		unsigned long	doEdgeLod;
		/*   4B: Maximum anisotropic filter used              */
		TexMaxAniso	maxAniso;
	} EdModelMatTexMap;

	//-----------------------------------------------------------------------
	/*
            Texture coordinate count change message
         */
	typedef struct {
		/*   4B: Number of texture coordinates to set                */
		unsigned long	size;
	} EdModelMatTexCoordSize;

	//-----------------------------------------------------------------------
	/* Feature used to generate texture coordinates */
	enum TexCoordFunc {
		/*  Convert source attributes with a 2x4 matrix           */
		TEX_COORD_FUNC_MTX24,
		/*  Convert source attributes with a 3x4 matrix           */
		TEX_COORD_FUNC_MTX34,
		/*  R (red)/G (green) elements from color channel output */
		TEX_COORD_FUNC_SRTG,
		/* Define enum as 4-byte type */
		TexCoordFunc_4BYTE_ENUM =	0xFFFFFFFF
	};

	//-----------------------------------------------------------------------
	/* Attributes used to generate texture coordinates */
	enum TexCoordSrcParam {
		/*  Position coordinates                        */
		TEX_COORD_SRC_PARAM_POS,
		/*  Normal Vectors                    */
		TEX_COORD_SRC_PARAM_NRM,
		/*  Binormal vector  */
		TEX_COORD_SRC_PARAM_BINRM,
		/*  Tangent vector    */
		TEX_COORD_SRC_PARAM_TANGENT,
		/*  Input texture coordinate 0             */
		TEX_COORD_SRC_PARAM_TEX0,
		/*  Input texture coordinate 1             */
		TEX_COORD_SRC_PARAM_TEX1,
		/*  Input texture coordinate 2             */
		TEX_COORD_SRC_PARAM_TEX2,
		/*  Input texture coordinate 3             */
		TEX_COORD_SRC_PARAM_TEX3,
		/*  Input texture coordinate 4             */
		TEX_COORD_SRC_PARAM_TEX4,
		/*  Input texture coordinate 5             */
		TEX_COORD_SRC_PARAM_TEX5,
		/*  Input texture coordinate 6             */
		TEX_COORD_SRC_PARAM_TEX6,
		/*  Input texture coordinate 7             */
		TEX_COORD_SRC_PARAM_TEX7,
		/*  Vertex color 0                     */
		TEX_COORD_SRC_PARAM_COLOR0,
		/*  Vertex color 1                     */
		TEX_COORD_SRC_PARAM_COLOR1,
		/* Define enum as 4-byte type */
		TexCoordSrcParam_4BYTE_ENUM =	0xFFFFFFFF
	};

	//-----------------------------------------------------------------------
	/* Matrix used to generate texture coordinates */
	enum TexCoordMtx {
		/*  Texture matrix 0                   */
		TEX_COORD_MTX_TEXMTX0,
		/*  Texture matrix 1                   */
		TEX_COORD_MTX_TEXMTX1,
		/*  Texture matrix 2                   */
		TEX_COORD_MTX_TEXMTX2,
		/*  Texture matrix 3                   */
		TEX_COORD_MTX_TEXMTX3,
		/*  Texture matrix 4                   */
		TEX_COORD_MTX_TEXMTX4,
		/*  Texture matrix 5                   */
		TEX_COORD_MTX_TEXMTX5,
		/*  Texture matrix 6                   */
		TEX_COORD_MTX_TEXMTX6,
		/*  Texture matrix 7                   */
		TEX_COORD_MTX_TEXMTX7,
		/*  Texture matrix 8                   */
		TEX_COORD_MTX_TEXMTX8,
		/*  Texture matrix 9                   */
		TEX_COORD_MTX_TEXMTX9,
		/*  unit matrix                          */
		TEX_COORD_MTX_IDENTITY,
		/* Define enum as 4-byte type */
		TexCoordMtx_4BYTE_ENUM =	0xFFFFFFFF
	};

	//-----------------------------------------------------------------------
	/*
            Texture coordinate information change message
         */
	typedef struct {
		/*   4B: Texture coordinate index                */
		unsigned long	index;
		/*   4B: Function used to generate texture coordinates    */
		TexCoordFunc	func;
		/*   4B: Source attributes used to generate texture coordinates  */
		TexCoordSrcParam	srcParam;
		/*   4B: Matrix used to generate texture coordinates        */
		TexCoordMtx	mtx;
	} EdModelMatTexCoord;

	//-----------------------------------------------------------------------
	/* Texture mapping method */
	enum TexMapMethod {
		/*  Texture coordinates to set in each vertex                                                 */
		TEX_MAP_METHOD_TEX_COORD,
		/*  Environment mapping from the camera used for display                                       */
		TEX_MAP_METHOD_ENV_CAMERA,
		/*  Environment mapping from light specified by ref_light                                      */
		TEX_MAP_METHOD_ENV_LIGHT,
		/*  Environment mapping from the position between the light specified by ref_light and the camera used for display  */
		TEX_MAP_METHOD_ENV_SPEC,
		/*  Projection mapping from the camera specified by ref_camera                                     */
		TEX_MAP_METHOD_PROJECTION,
		/* Define enum as 4-byte type */
		TexMapMethod_4BYTE_ENUM =	0xFFFFFFFF
	};

	//-----------------------------------------------------------------------
	/*
            Texture matrix information change message
         */
	typedef struct {
		/*   4B: Texture matrix index                */
		unsigned long	index;
		/*   4B: Texture mapping method              */
		TexMapMethod	mapping_method;
		/*   4B: Camera referenced when mapping is performed  */
		long	ref_camera;
		/*   4B: Light referenced when mapping is performed  */
		long	ref_light;
		/*   4B: Scale S-component                           */
		float	scale_s;
		/*   4B: Scale T-component                           */
		float	scale_t;
		/*   4B: Rotation component (unit: degrees)                  */
		float	rotate;
		/*   4B: Translation S-component                     */
		float	trans_s;
		/*   4B: Translation T-component                     */
		float	trans_t;
	} EdModelMatTexMatrix;

	//-----------------------------------------------------------------------
	/*
            Texture image information change message
         */
	typedef struct {
		/*   4B: Texture map index                          */
		unsigned long	index;
		/*   4B: Texture ID to be used                                */
		unsigned long	texID;
		/*   4B: Palette ID to be used. 0xffffffff if not used            */
		unsigned long	plttID;
	} EdModelMatTexMapImage;

	//-----------------------------------------------------------------------
	/*
            Indirect matrix SRT change message
         */
	typedef struct {
		/*   4B: Indirect matrix ID                    */
		unsigned long	index;
		/*   4B: Scale S-component                           */
		float	scale_s;
		/*   4B: Scale T-component                           */
		float	scale_t;
		/*   4B: Rotation component (unit: degrees)                  */
		float	rotate;
		/*   4B: Translation S-component                     */
		float	trans_s;
		/*   4B: Translation T-component                     */
		float	trans_t;
	} EdModelMatIndMtxSRT;

	//-----------------------------------------------------------------------
	/*
            Fur lie matrix change command
         */
	typedef struct {
		/*   4B: Indirect matrix ID                    */
		unsigned long	index;
		/*   4B: Scale S-component                           */
		float	scale_s;
		/*   4B: Scale T-component                           */
		float	scale_t;
		/*   4B: Rotation component (unit: degrees)                  */
		float	rotate;
	} EditModelMatFurIndMtxSR;

	//-----------------------------------------------------------------------
	/*
            Structure to specify selected object
         */
	typedef struct {
		/*  4B: Model ID/scene ID                                                */
		unsigned long	id;
		/*  2B: Selected object type                                         */
		unsigned short	objectType;
		/*  2B: Object index value (ignored when OBJECT_TYPE_MDL)  */
		unsigned short	index;
	} AppealObject;

	//-----------------------------------------------------------------------
	/*
            Selected object emphasized display message
         */
	typedef struct {
		/*   4B: Number of variable length selected object data to follow afterwards   */
		unsigned long	size;
	} EdEnvSetAppeal;

	//-----------------------------------------------------------------------
	/*
            Selected object emphasis setting change message
         */
	typedef struct {
		/*  2B: Flash interval (units: game frames)                    */
		unsigned short	interval;
		/*  2B: Number of times to flash                                         */
		unsigned short	count;
		/*  2B: Flag whether to recursively flash the child node     */
		unsigned short	node_recursive_flg;
		/*  2B: Flag whether to move the camera focus       */
		unsigned short	focus_camera_flg;
	} EdEnvConfigAppeal;

	//-----------------------------------------------------------------------
	/*
            Operation target camera/light change message
         */
	typedef struct {
		/*  2B: Operation target type (OBJECT_TYPE_CAMERA/OBJECT_TYPE_LIGHT)  */
		unsigned short	type;
		/*  2B: Operation target index                                    */
		unsigned short	index;
	} EdEnvCurrentControl;

	//-----------------------------------------------------------------------
	/*
            Active scene change message
         */
	typedef struct {
		/*  4B: ID of scene to make active                   */
		unsigned long	id;
	} EdSceneSetActive;

	//-----------------------------------------------------------------------
	/*
            ROTATE camera position get reply message
         */
	typedef struct {
		/*  4B: Camera position X-component             */
		float	posX;
		/*  4B: Camera position Y-component             */
		float	posY;
		/*  4B: Camera position Z-component             */
		float	posZ;
		/*  4B: X-axis rotation.                     */
		float	rotX;
		/*  4B: Y-axis rotation                     */
		float	rotY;
		/*  4B: rotation on the Z-axis                     */
		float	rotZ;
	} EdSceneGetCameraRotPosReply;

	//-----------------------------------------------------------------------
	/*
            AIM camera position get reply message
         */
	typedef struct {
		/*  4B: Camera position X-component             */
		float	posX;
		/*  4B: Camera position Y-component             */
		float	posY;
		/*  4B: Camera position Z-component             */
		float	posZ;
		/*  4B: Camera focal point X-component           */
		float	aimX;
		/*  4B: Camera focal point Y-component           */
		float	aimY;
		/*  4B: Camera focal point Z-component           */
		float	aimZ;
		/*  4B: Rotational angle for the line of sight direction    */
		float	twist;
	} EdSceneGetCameraAimPosReply;

	//-----------------------------------------------------------------------
	/*
            Light position get reply message
         */
	typedef struct {
		/*  4B: Light position X-component                   */
		float	posX;
		/*  4B: Light position Y-component                   */
		float	posY;
		/*  4B: Light position Z-component                   */
		float	posZ;
		/*  4B: Light focal point X-component                 */
		float	aimX;
		/*  4B: Light focal point Y-component                 */
		float	aimY;
		/*  4B: Light focal point Z-component                 */
		float	aimZ;
	} EdSceneGetLightPosReply;

	//-----------------------------------------------------------------------
	/*
            In-scene object setting swap message
         */
	typedef struct {
		/*  4B: Scene ID                            */
		unsigned long	id;
		/*  2B: Object type to perform swap  */
		unsigned short	objectType;
		/*  2B: Index of swap target １          */
		unsigned short	index1;
		/*  2B: Index of swap target 2          */
		unsigned short	index2;
		/*  2B: Reserved                                */
		unsigned short	padding_;
	} EdSceneSwapObject;

	//-----------------------------------------------------------------------
	/*
            View camera position settings message
         */
	typedef struct {
		/*  4B: Camera position X-component             */
		float	posX;
		/*  4B: Camera position Y-component             */
		float	posY;
		/*  4B: Camera position Z-component             */
		float	posZ;
		/*  4B: Camera focal point X-component           */
		float	aimX;
		/*  4B: Camera focal point Y-component           */
		float	aimY;
		/*  4B: Camera focal point Z-component           */
		float	aimZ;
		/*  4B: Rotational angle for the line of sight direction    */
		float	twist;
	} EdSceneSetViewCameraPos;

	//-----------------------------------------------------------------------
	/*
            Structure to specify object to be edited
         */
	typedef struct {
		/*  4B: ID                                                */
		unsigned long	id;
		/*  4B: Index value of object to be edited          */
		long	index;
	} EditSelectedObject;

	//-----------------------------------------------------------------------
	/* Edit object type definition */
	enum EditObjectType {
		/* model             */
		EDIT_OBJECT_TYPE_MODEL =	0,
		/* node             */
		EDIT_OBJECT_TYPE_NODE =	1,
		/* Material         */
		EDIT_OBJECT_TYPE_MAT =	2,
		/* Camera             */
		EDIT_OBJECT_TYPE_CAMERA =	3,
		/* Ambient light */
		EDIT_OBJECT_TYPE_AMB_LIGHT =	4,
		/* Lights             */
		EDIT_OBJECT_TYPE_LIGHT =	5,
		/* Light set       */
		EDIT_OBJECT_TYPE_LIGHT_SET =	6,
		/* Fog             */
		EDIT_OBJECT_TYPE_FOG =	7,
		/* Define enum as 4-byte type */
		EditObjectType_4BYTE_ENUM =	0xFFFFFFFF
	};

	//-----------------------------------------------------------------------
	/*
            Edit object selection state change message
         */
	typedef struct {
		/*   2B: Edit object type                     */
		unsigned short	objectType;
		/*   2B: Number of variable length edit objects to follow   */
		unsigned short	size;
	} EdEditSelection;

	//-----------------------------------------------------------------------
	/* Edit target types. */
	enum EditTargetType {
		/* Scene target                           */
		EDIT_TARGET_SCENE =	0x0100,
		/* Camera Target                           */
		EDIT_TARGET_CAMERA =	0x0200,
		/* Ambient light target               */
		EDIT_TARGET_AMB_LIGHT =	0x0300,
		/* Light Target                           */
		EDIT_TARGET_LIGHT =	0x0400,
		/* Light set target                     */
		EDIT_TARGET_LIGHT_SET =	0x0500,
		/* Fog Target                           */
		EDIT_TARGET_FOG =	0x0600,
		/* Model Target                           */
		EDIT_TARGET_MODEL =	0x0700,
		/* Node Target                           */
		EDIT_TARGET_NODE =	0x0800,
		/* Shape target                         */
		EDIT_TARGET_SHP =	0x0900,
		/* Material Target                       */
		EDIT_TARGET_MAT =	0x0A00,
		/* Color animation target             */
		EDIT_TARGET_COLOR_ANIM =	0x0B00,
		/* Texture SRT animation target      */
		EDIT_TARGET_TEXSRT_ANIM =	0x0C00,
		/* Indirect SRT animation target  */
		EDIT_TARGET_INDSRT_ANIM =	0x0D00,
		/* Texture target (unused)               */
		EDIT_TARGET_TEX =	0x0E00,
		/* Palette target (unused)               */
		EDIT_TARGET_PLT =	0x0F00,
		/* Target mask                           */
		EDIT_TARGET_MASK =	0xFF00,
		/* Define enum as 4-byte type */
		EditTargetType_4BYTE_ENUM =	0xFFFFFFFF
	};

	//-----------------------------------------------------------------------
	/* Model display position edit target */
	enum EditTargetModel {
		/* Model rotation X-component             */
		EDIT_TARGET_MODEL_ROT_X =	0x0700 + 0x01,
		/* Model rotation Y-component             */
		EDIT_TARGET_MODEL_ROT_Y =	0x0700 + 0x02,
		/* Model rotation Z-component             */
		EDIT_TARGET_MODEL_ROT_Z =	0x0700 + 0x03,
		/* Model translation X-component         */
		EDIT_TARGET_MODEL_TRANS_X =	0x0700 + 0x04,
		/* Model translation Y-component         */
		EDIT_TARGET_MODEL_TRANS_Y =	0x0700 + 0x05,
		/* Model translation Z-component         */
		EDIT_TARGET_MODEL_TRANS_Z =	0x0700 + 0x06,
		/* Model scale X-component               */
		EDIT_TARGET_MODEL_SCALE_X =	0x0700 + 0x07,
		/* Model scale Y-component               */
		EDIT_TARGET_MODEL_SCALE_Y =	0x0700 + 0x08,
		/* Model scale Z-component               */
		EDIT_TARGET_MODEL_SCALE_Z =	0x0700 + 0x09,
		/* Model scale component                */
		EDIT_TARGET_MODEL_SCALE_XYZ =	0x0700 + 0x0A,
		/* Model's visibility                */
		EDIT_TARGET_MODEL_VISIBILITY =	0x0700 + 0x0B,
		/* Model for the bind target model          */
		EDIT_TARGET_MODEL_BIND_MODEL =	0x0700 + 0x0C,
		/* Model for the bind target node          */
		EDIT_TARGET_MODEL_BIND_NODE =	0x0700 + 0x0D,
		/* Current action change              */
		EDIT_TARGET_MODEL_CURRENT_ACTION =	0x0700 + 0x0E,
		/* Enables/disables volume culling        */
		EDIT_TARGET_MODEL_VOLUME_ENABLE =	0x0700 + 0x0F,
		/* Center coordinate X for volume culling       */
		EDIT_TARGET_MODEL_VOLUME_CENTER_X =	0x0700 + 0x10,
		/* Center coordinate Y for volume culling       */
		EDIT_TARGET_MODEL_VOLUME_CENTER_Y =	0x0700 + 0x11,
		/* Center coordinate Z for volume culling       */
		EDIT_TARGET_MODEL_VOLUME_CENTER_Z =	0x0700 + 0x12,
		/* Size X-component for volume culling     */
		EDIT_TARGET_MODEL_VOLUME_SIZE_X =	0x0700 + 0x13,
		/* Size Y-component for volume culling     */
		EDIT_TARGET_MODEL_VOLUME_SIZE_Y =	0x0700 + 0x14,
		/* Size Z-component for volume culling     */
		EDIT_TARGET_MODEL_VOLUME_SIZE_Z =	0x0700 + 0x15,
		/* Size component for volume culling      */
		EDIT_TARGET_MODEL_VOLUME_SIZE_XYZ =	0x0700 + 0x16,
		/* Minimum X-coordinate for volume culling       */
		EDIT_TARGET_MODEL_VOLUME_MIN_X =	0x0700 + 0x17,
		/* Minimum Y-coordinate for volume culling       */
		EDIT_TARGET_MODEL_VOLUME_MIN_Y =	0x0700 + 0x18,
		/* Minimum Z-coordinate for volume culling       */
		EDIT_TARGET_MODEL_VOLUME_MIN_Z =	0x0700 + 0x19,
		/* Maximum X-coordinate for volume culling       */
		EDIT_TARGET_MODEL_VOLUME_MAX_X =	0x0700 + 0x1A,
		/* Maximum Y-coordinate for volume culling       */
		EDIT_TARGET_MODEL_VOLUME_MAX_Y =	0x0700 + 0x1B,
		/* Maximum Z-coordinate for volume culling       */
		EDIT_TARGET_MODEL_VOLUME_MAX_Z =	0x0700 + 0x1C,
		/* Character animation translation component ignore setting  */
		EDIT_TARGET_MODEL_IGNORE_ANMCHR_TRANS =	0x0700 + 0x1D,
		/* Wireframe visibility      */
		EDIT_TARGET_MODEL_WIREFRAME_VISIBILITY =	0x0700 + 0x1E,
		/* Define enum as 4-byte type */
		EditTargetModel_4BYTE_ENUM =	0xFFFFFFFF
	};

	//-----------------------------------------------------------------------
	/* Camera edit target */
	enum EditTargetCamera {
		/* Camera type    */
		EDIT_TARGET_CAMERA_TYPE =	0x0200 + 0x01,
		/* Projection type      */
		EDIT_TARGET_CAMERA_PROJECTION_TYPE =	0x0200 + 0x02,
		/* position            */
		EDIT_TARGET_CAMERA_POS =	0x0200 + 0x03,
		/* Viewpoint          */
		EDIT_TARGET_CAMERA_AIM =	0x0200 + 0x04,
		/* Rotation angle          */
		EDIT_TARGET_CAMERA_ROT =	0x0200 + 0x05,
		/* Twist angle        */
		EDIT_TARGET_CAMERA_TWIST =	0x0200 + 0x06,
		/* aspect ratio    */
		EDIT_TARGET_CAMERA_ASPECT =	0x0200 + 0x07,
		/* Near clip    */
		EDIT_TARGET_CAMERA_NEAR =	0x0200 + 0x08,
		/* Far clip  */
		EDIT_TARGET_CAMERA_FAR =	0x0200 + 0x09,
		/* Perspective projection angle    */
		EDIT_TARGET_CAMERA_PERSP_FOVY =	0x0200 + 0x0A,
		/* Orthogonal projection height      */
		EDIT_TARGET_CAMERA_ORTHO_HEIGHT =	0x0200 + 0x0B,
		/* Define enum as 4-byte type */
		EditTargetCamera_4BYTE_ENUM =	0xFFFFFFFF
	};

	//-----------------------------------------------------------------------
	/* Camera type */
	enum TargetCameraType {
		/* Rotate camera type */
		TARGET_CAMERA_TYPE_ROTATE =	0x00,
		/* Focus camera type */
		TARGET_CAMERA_TYPE_AIM =	0x01,
		/* Define enum as 4-byte type */
		TargetCameraType_4BYTE_ENUM =	0xFFFFFFFF
	};

	//-----------------------------------------------------------------------
	/* Projection type */
	enum TargetCameraProjectionType {
		/* Parallel projection camera */
		TARGET_CAMERA_PROJECTION_TYPE_ORTHO =	0x00,
		/* Perspective projection camera */
		TARGET_CAMERA_PROJECTION_TYPE_PERSP =	0x01,
		/* Define enum as 4-byte type */
		TargetCameraProjectionType_4BYTE_ENUM =	0xFFFFFFFF
	};

	//-----------------------------------------------------------------------
	/* Ambient light edit target */
	enum EditTargetAmbLight {
		/* Color                   */
		EDIT_TARGET_AMB_LIGHT_COLOR =	0x0300 + 0x01,
		/* RBGA component enable/disable state   */
		EDIT_TARGET_AMB_LIGHT_EFFECT =	0x0300 + 0x02,
		/* Define enum as 4-byte type */
		EditTargetAmbLight_4BYTE_ENUM =	0xFFFFFFFF
	};

	//-----------------------------------------------------------------------
	/* RBGA component enable/disable state */
	enum TargetAmbLightEffect {
		/* RGBA         */
		TARGET_AMB_LIGHT_EFFECT_RGBA =	0x00,
		/* RGB          */
		TARGET_AMB_LIGHT_EFFECT_RGB =	0x01,
		/* A            */
		TARGET_AMB_LIGHT_EFFECT_ALPHA =	0x02,
		/* Define enum as 4-byte type */
		TargetAmbLightEffect_4BYTE_ENUM =	0xFFFFFFFF
	};

	//-----------------------------------------------------------------------
	/* Light edit target */
	enum EditTargetLight {
		/* Light Type             */
		EDIT_TARGET_LIGHT_TYPE =	0x0400 + 0x01,
		/* Specular existence       */
		EDIT_TARGET_LIGHT_SPECULAR =	0x0400 + 0x02,
		/* Light amount distribution function             */
		EDIT_TARGET_LIGHT_SPOT_FUNC =	0x0400 + 0x03,
		/* Brightness reduction function         */
		EDIT_TARGET_LIGHT_DIST_FUNC =	0x0400 + 0x04,
		/* Enable               */
		EDIT_TARGET_LIGHT_ENABLE =	0x0400 + 0x05,
		/* position                     */
		EDIT_TARGET_LIGHT_POS =	0x0400 + 0x06,
		/* Viewpoint                   */
		EDIT_TARGET_LIGHT_AIM =	0x0400 + 0x07,
		/* Color                   */
		EDIT_TARGET_LIGHT_COLOR =	0x0400 + 0x08,
		/* Specular color         */
		EDIT_TARGET_LIGHT_SPEC_COLOR =	0x0400 + 0x09,
		/* Shininess parameter         */
		EDIT_TARGET_LIGHT_SHININESS =	0x0400 + 0x0A,
		/* Cutoff Angle           */
		EDIT_TARGET_LIGHT_CUTOFF =	0x0400 + 0x0B,
		/* Distance between light and reference point   */
		EDIT_TARGET_LIGHT_REF_DISTANCE =	0x0400 + 0x0C,
		/* Reference point brightness           */
		EDIT_TARGET_LIGHT_REF_BRIGHTNESS =	0x0400 + 0x0D,
		/* RBGA component enable/disable state   */
		EDIT_TARGET_LIGHT_EFFECT =	0x0400 + 0x0E,
		/* Light camera tracking state   */
		EDIT_TARGET_LIGHT_BIND =	0x0400 + 0x0F,
		/* Define enum as 4-byte type */
		EditTargetLight_4BYTE_ENUM =	0xFFFFFFFF
	};

	//-----------------------------------------------------------------------
	/* Light Type */
	enum TargetLightType {
		/* Point light source          */
		TARGET_LIGHT_TYPE_POINT =	0x00,
		/* Parallel light source        */
		TARGET_LIGHT_TYPE_DIRECTIONAL =	0x01,
		/* Spot light  */
		TARGET_LIGHT_TYPE_SPOT =	0x02,
		/* Define enum as 4-byte type */
		TargetLightType_4BYTE_ENUM =	0xFFFFFFFF
	};

	//-----------------------------------------------------------------------
	/* Light amount distribution function type */
	enum TargetLightSpotFunc {
		/* Off          */
		TARGET_LIGHT_SPOT_FUNC_OFF =	0x00,
		/* flat         */
		TARGET_LIGHT_SPOT_FUNC_FLAT =	0x01,
		/* cos          */
		TARGET_LIGHT_SPOT_FUNC_COS =	0x02,
		/* cos2         */
		TARGET_LIGHT_SPOT_FUNC_COS2 =	0x03,
		/* sharp        */
		TARGET_LIGHT_SPOT_FUNC_SHARP =	0x04,
		/* ring1        */
		TARGET_LIGHT_SPOT_FUNC_RING1 =	0x05,
		/* ring2        */
		TARGET_LIGHT_SPOT_FUNC_RING2 =	0x06,
		/* Define enum as 4-byte type */
		TargetLightSpotFunc_4BYTE_ENUM =	0xFFFFFFFF
	};

	//-----------------------------------------------------------------------
	/* Brightness reduction function type */
	enum TargetLightDistFunc {
		/* Off             */
		TARGET_LIGHT_DIST_FUNC_OFF =	0x00,
		/* gentle          */
		TARGET_LIGHT_DIST_FUNC_GENTLE =	0x01,
		/* medium          */
		TARGET_LIGHT_DIST_FUNC_MEDIUM =	0x02,
		/* steep           */
		TARGET_LIGHT_DIST_FUNC_STEEP =	0x03,
		/* Define enum as 4-byte type */
		TargetLightDistFunc_4BYTE_ENUM =	0xFFFFFFFF
	};

	//-----------------------------------------------------------------------
	/* RBGA component enable/disable state */
	enum TargetLightEffect {
		/* RGBA         */
		TARGET_LIGHT_EFFECT_RGBA =	0x00,
		/* RGB          */
		TARGET_LIGHT_EFFECT_RGB =	0x01,
		/* A            */
		TARGET_LIGHT_EFFECT_ALPHA =	0x02,
		/* Define enum as 4-byte type */
		TargetLightEffect_4BYTE_ENUM =	0xFFFFFFFF
	};

	//-----------------------------------------------------------------------
	/* Light camera tracking state */
	enum TargetLightBind {
		/* Do not follow     */
		TARGET_LIGHT_BIND_NONE =	0x00,
		/* Headlight   */
		TARGET_LIGHT_BIND_HEAD =	0x01,
		/* Rim light     */
		TARGET_LIGHT_BIND_REAR =	0x02,
		/* Relative position fixed   */
		TARGET_LIGHT_BIND_RELATIVE =	0x03,
		/* Define enum as 4-byte type */
		TargetLightBind_4BYTE_ENUM =	0xFFFFFFFF
	};

	//-----------------------------------------------------------------------
	/* Light set edit target */
	enum EditTargetLightSet {
		/* Ambient light index    */
		EDIT_TARGET_LIGHT_SET_AMB_LIGHT_INDEX =	0x0500 + 0x01,
		/* Light index                */
		EDIT_TARGET_LIGHT_SET_LIGHT_INDEX =	0x0500 + 0x02,
		/* Define enum as 4-byte type */
		EditTargetLightSet_4BYTE_ENUM =	0xFFFFFFFF
	};

	//-----------------------------------------------------------------------
	/* Fog edit target */
	enum EditTargetFog {
		/* Fog type    */
		EDIT_TARGET_FOG_TYPE =	0x0600 + 0x01,
		/* Color          */
		EDIT_TARGET_FOG_COLOR =	0x0600 + 0x02,
		/* Fog start Z value   */
		EDIT_TARGET_FOG_START =	0x0600 + 0x03,
		/* Fog end Z value   */
		EDIT_TARGET_FOG_END =	0x0600 + 0x04,
		/* Define enum as 4-byte type */
		EditTargetFog_4BYTE_ENUM =	0xFFFFFFFF
	};

	//-----------------------------------------------------------------------
	/* Fog type */
	enum TargetFogType {
		/* No              */
		TARGET_FOG_NONE =	0x00,
		/* linear              */
		TARGET_FOG_PERSP_LIN =	0x01,
		/* Index              */
		TARGET_FOG_PERSP_EXP =	0x02,
		/* Square of index        */
		TARGET_FOG_PERSP_EXP2 =	0x03,
		/* Reciprocal of index        */
		TARGET_FOG_PERSP_REVEXP =	0x04,
		/* Reciprocal of square of index  */
		TARGET_FOG_PERSP_REVEXP2 =	0x05,
		/* linear              */
		TARGET_FOG_ORTHO_LIN =	0x06,
		/* Index              */
		TARGET_FOG_ORTHO_EXP =	0x07,
		/* Square of index        */
		TARGET_FOG_ORTHO_EXP2 =	0x08,
		/* Reciprocal of index        */
		TARGET_FOG_ORTHO_REVEXP =	0x09,
		/* Reciprocal of square of index  */
		TARGET_FOG_ORTHO_REVEXP2 =	0x0A,
		/* Define enum as 4-byte type */
		TargetFogType_4BYTE_ENUM =	0xFFFFFFFF
	};

	//-----------------------------------------------------------------------
	/* Material edit target */
	enum EditTargetMat {
		/* Light set index                                   */
		EDIT_TARGET_MAT_LIGHTSET_INDEX =	0x0A00 + 0x01,
		/* Fog index                                         */
		EDIT_TARGET_MAT_FOG_INDEX =	0x0A00 + 0x02,
		/* Shape display face                                         */
		EDIT_TARGET_MAT_DISPLAY_FACE =	0x0A00 + 0x03,
		/* Material color                                         */
		EDIT_TARGET_MAT_COLOR =	0x0A00 + 0x04,
		/* TEV color register                                        */
		EDIT_TARGET_MAT_TEV_COLOR =	0x0A00 + 0x05,
		/* TEV constant register                                  */
		EDIT_TARGET_MAT_TEV_CONSTANT =	0x0A00 + 0x06,
		/* Light enable/disable                                         */
		EDIT_TARGET_MAT_CHAN_LIGHT_ENABLE =	0x0A00 + 0x07,
		/* Ambient color source                               */
		EDIT_TARGET_MAT_CHAN_AMB_SOURCE =	0x0A00 + 0x08,
		/* Material color source                                 */
		EDIT_TARGET_MAT_CHAN_MAT_SOURCE =	0x0A00 + 0x09,
		/* TEV swap table R-component                                 */
		EDIT_TARGET_MAT_TEV_SWAP_TABLE_R =	0x0A00 + 0x0A,
		/* TEV swap table G-component                                 */
		EDIT_TARGET_MAT_TEV_SWAP_TABLE_G =	0x0A00 + 0x0B,
		/* TEV swap table B-component                                 */
		EDIT_TARGET_MAT_TEV_SWAP_TABLE_B =	0x0A00 + 0x0C,
		/* TEV swap table A-component                                 */
		EDIT_TARGET_MAT_TEV_SWAP_TABLE_A =	0x0A00 + 0x0D,
		/* TEV stage count                                                  */
		EDIT_TARGET_MAT_TEV_STAGE_NUM =	0x0A00 + 0x0E,
		/* Texture coordinate input                                     */
		EDIT_TARGET_MAT_TEV_ORDER_TEX_COORD =	0x0A00 + 0x0F,
		/* Texture map input                                   */
		EDIT_TARGET_MAT_TEV_ORDER_TEX_MAP =	0x0A00 + 0x10,
		/* Input rasterization                                       */
		EDIT_TARGET_MAT_TEV_ORDER_RAS_ORDER =	0x0A00 + 0x11,
		/* Swap rasterization                                   */
		EDIT_TARGET_MAT_TEV_SWAP_RAS_SEL =	0x0A00 + 0x12,
		/* Swap texture color                               */
		EDIT_TARGET_MAT_TEV_SWAP_TEX_SEL =	0x0A00 + 0x13,
		/* Input operand a-component                                    */
		EDIT_TARGET_MAT_TEV_COLOR_STAGE_IN_A =	0x0A00 + 0x14,
		/* Input operand b-component                                    */
		EDIT_TARGET_MAT_TEV_COLOR_STAGE_IN_B =	0x0A00 + 0x15,
		/* Input operand c-component                                    */
		EDIT_TARGET_MAT_TEV_COLOR_STAGE_IN_C =	0x0A00 + 0x16,
		/* Input operand d-component                                    */
		EDIT_TARGET_MAT_TEV_COLOR_STAGE_IN_D =	0x0A00 + 0x17,
		/* TEV constant color                                            */
		EDIT_TARGET_MAT_TEV_COLOR_STAGE_CONSTANT =	0x0A00 + 0x18,
		/* TEV calculation                                                  */
		EDIT_TARGET_MAT_TEV_COLOR_STAGE_OP =	0x0A00 + 0x19,
		/* TEV bias value                                            */
		EDIT_TARGET_MAT_TEV_COLOR_STAGE_BIAS =	0x0A00 + 0x1A,
		/* TEV scale value                                            */
		EDIT_TARGET_MAT_TEV_COLOR_STAGE_SCALE =	0x0A00 + 0x1B,
		/* TEV clamp                                              */
		EDIT_TARGET_MAT_TEV_COLOR_STAGE_CLAMP =	0x0A00 + 0x1C,
		/* Output register number                                         */
		EDIT_TARGET_MAT_TEV_COLOR_STAGE_OUT_REG =	0x0A00 + 0x1D,
		/* Input operand a-component                                    */
		EDIT_TARGET_MAT_TEV_ALPHA_STAGE_IN_A =	0x0A00 + 0x1E,
		/* Input operand b-component                                    */
		EDIT_TARGET_MAT_TEV_ALPHA_STAGE_IN_B =	0x0A00 + 0x1F,
		/* Input operand c-component                                    */
		EDIT_TARGET_MAT_TEV_ALPHA_STAGE_IN_C =	0x0A00 + 0x20,
		/* Input operand d-component                                    */
		EDIT_TARGET_MAT_TEV_ALPHA_STAGE_IN_D =	0x0A00 + 0x21,
		/* TEV constant color                                            */
		EDIT_TARGET_MAT_TEV_ALPHA_STAGE_CONSTANT =	0x0A00 + 0x22,
		/* TEV calculation                                                  */
		EDIT_TARGET_MAT_TEV_ALPHA_STAGE_OP =	0x0A00 + 0x23,
		/* TEV bias value                                            */
		EDIT_TARGET_MAT_TEV_ALPHA_STAGE_BIAS =	0x0A00 + 0x24,
		/* TEV scale value                                            */
		EDIT_TARGET_MAT_TEV_ALPHA_STAGE_SCALE =	0x0A00 + 0x25,
		/* TEV clamp                                              */
		EDIT_TARGET_MAT_TEV_ALPHA_STAGE_CLAMP =	0x0A00 + 0x26,
		/* Output register number                                         */
		EDIT_TARGET_MAT_TEV_ALPHA_STAGE_OUT_REG =	0x0A00 + 0x27,
		/* Translucent flag                                             */
		EDIT_TARGET_MAT_TRANSPARENCY_MODE =	0x0A00 + 0x28,
		/* Alpha comparison subfunction0                                */
		EDIT_TARGET_MAT_PE_ALPHA_COMPARE_COMP0 =	0x0A00 + 0x29,
		/* Reference value for subfunction0                               */
		EDIT_TARGET_MAT_PE_ALPHA_COMPARE_REF0 =	0x0A00 + 0x2A,
		/* Operator for the combination of subfunction0 and subfunction1  */
		EDIT_TARGET_MAT_PE_ALPHA_COMPARE_OP =	0x0A00 + 0x2B,
		/* Alpha comparison subfunction1                                */
		EDIT_TARGET_MAT_PE_ALPHA_COMPARE_COMP1 =	0x0A00 + 0x2C,
		/* Reference value for subfunction1                               */
		EDIT_TARGET_MAT_PE_ALPHA_COMPARE_REF1 =	0x0A00 + 0x2D,
		/* Z-comparison enable flag                                  */
		EDIT_TARGET_MAT_PE_Z_COMPARE_ENABLE =	0x0A00 + 0x2E,
		/* Z-comparison function                                                */
		EDIT_TARGET_MAT_PE_Z_COMPARE_FUNC =	0x0A00 + 0x2F,
		/* Z-buffer change enable flag                          */
		EDIT_TARGET_MAT_PE_Z_COMPARE_UPDATE =	0x0A00 + 0x30,
		/* Z-test timing flag                                  */
		EDIT_TARGET_MAT_PE_Z_COMPARE_BEFORE_TEX =	0x0A00 + 0x31,
		/* Blending mode                                           */
		EDIT_TARGET_MAT_PE_BLEND_MODE_TYPE =	0x0A00 + 0x32,
		/* Source coefficient                                               */
		EDIT_TARGET_MAT_PE_BLEND_MODE_SRC =	0x0A00 + 0x33,
		/* Destination coefficient                                   */
		EDIT_TARGET_MAT_PE_BLEND_MODE_DST =	0x0A00 + 0x34,
		/* Blending                                             */
		EDIT_TARGET_MAT_PE_BLEND_MODE_OP =	0x0A00 + 0x35,
		/* Wrap settings in the texture horizontal (s) direction                        */
		EDIT_TARGET_MAT_TEXMAP_WRAP_S =	0x0A00 + 0x36,
		/* Wrap settings in the texture vertical (t) direction                        */
		EDIT_TARGET_MAT_TEXMAP_WRAP_T =	0x0A00 + 0x37,
		/* Mipmap enabled/disabled                                  */
		EDIT_TARGET_MAT_TEXMAP_MIPMAP =	0x0A00 + 0x38,
		/* Filter used when reducing texels                     */
		EDIT_TARGET_MAT_TEXMAP_FILTER_MIN =	0x0A00 + 0x39,
		/* Filter used when expanding texels                     */
		EDIT_TARGET_MAT_TEXMAP_FILTER_MAG =	0x0A00 + 0x3A,
		/* Bias added to the calculated LOD value                        */
		EDIT_TARGET_MAT_TEXMAP_LOD_BIAS =	0x0A00 + 0x3B,
		/* bias clamp settings                                       */
		EDIT_TARGET_MAT_TEXMAP_BIAS_CLAMP =	0x0A00 + 0x3C,
		/* Flag to specify whether to calculate LOD using texel interpolation        */
		EDIT_TARGET_MAT_TEXMAP_DO_EDGE_LOD =	0x0A00 + 0x3D,
		/* Maximum anisotropic filter used                   */
		EDIT_TARGET_MAT_TEXMAP_MAX_ANISO =	0x0A00 + 0x3E,
		/* Number of texture coordinates to set                             */
		EDIT_TARGET_MAT_TEXCOORD_SIZE =	0x0A00 + 0x3F,
		/* Function used to generate texture coordinates                     */
		EDIT_TARGET_MAT_TEXCOORD_FUNC =	0x0A00 + 0x40,
		/* Source attributes used to generate texture coordinates                   */
		EDIT_TARGET_MAT_TEXCOORD_SRC_PARAM =	0x0A00 + 0x41,
		/* Matrix used to generate texture coordinates                         */
		EDIT_TARGET_MAT_TEXCOORD_MTX =	0x0A00 + 0x42,
		/* Texture mapping method                               */
		EDIT_TARGET_MAT_TEXMTX_MAPPING_METHOD =	0x0A00 + 0x43,
		/* Camera index to reference                                 */
		EDIT_TARGET_MAT_TEXMTX_REF_CAMERA =	0x0A00 + 0x44,
		/* Light index to reference                                 */
		EDIT_TARGET_MAT_TEXMTX_REF_LIGHT =	0x0A00 + 0x45,
		/* Scale S-component                                            */
		EDIT_TARGET_MAT_TEXMTX_SCALE_S =	0x0A00 + 0x46,
		/* Scale T-component                                            */
		EDIT_TARGET_MAT_TEXMTX_SCALE_T =	0x0A00 + 0x47,
		/* Scale ST-component                                           */
		EDIT_TARGET_MAT_TEXMTX_SCALE_ST =	0x0A00 + 0x48,
		/* Rotation component (unit: degrees)                                   */
		EDIT_TARGET_MAT_TEXMTX_ROTATE =	0x0A00 + 0x49,
		/* Translation S-component                                      */
		EDIT_TARGET_MAT_TEXMTX_TRANS_S =	0x0A00 + 0x4A,
		/* Translation T-component                                      */
		EDIT_TARGET_MAT_TEXMTX_TRANS_T =	0x0A00 + 0x4B,
		/* Indirect texture mapping method                 */
		EDIT_TARGET_MAT_INDMTX_INDIRECT_METHOD =	0x0A00 + 0x4C,
		/* Light number to reference when performing normal mapping representation         */
		EDIT_TARGET_MAT_INDMTX_REF_LIGHT =	0x0A00 + 0x4D,
		/* Indirect texture matrix scale S-component              */
		EDIT_TARGET_MAT_INDMTX_SCALE_S =	0x0A00 + 0x4E,
		/* Indirect texture matrix scale T-component              */
		EDIT_TARGET_MAT_INDMTX_SCALE_T =	0x0A00 + 0x4F,
		/* Indirect texture matrix scale ST-component             */
		EDIT_TARGET_MAT_INDMTX_SCALE_ST =	0x0A00 + 0x50,
		/* Indirect texture matrix rotation component (unit: degrees)      */
		EDIT_TARGET_MAT_INDMTX_ROTATE =	0x0A00 + 0x51,
		/* Indirect texture matrix translation S-component        */
		EDIT_TARGET_MAT_INDMTX_TRANS_S =	0x0A00 + 0x52,
		/* Indirect texture matrix translation T-component        */
		EDIT_TARGET_MAT_INDMTX_TRANS_T =	0x0A00 + 0x53,
		/* Indirect stage count                               */
		EDIT_TARGET_MAT_INDSTG_NUM =	0x0A00 + 0x54,
		/* Texture map ID to be used                               */
		EDIT_TARGET_MAT_INDSTG_TEXMAP_ORDER =	0x0A00 + 0x55,
		/* Texture coordinate ID to be used                                 */
		EDIT_TARGET_MAT_INDSTG_COORD_ORDER =	0x0A00 + 0x56,
		/* Scale factor for the S-coordinate                            */
		EDIT_TARGET_MAT_INDSTG_SCALE_S =	0x0A00 + 0x57,
		/* Scale factor for the T-coordinate                            */
		EDIT_TARGET_MAT_INDSTG_SCALE_T =	0x0A00 + 0x58,
		/* Indirect stage to be used with this TEV stage          */
		EDIT_TARGET_MAT_TEV_IND_STAGE_IND_STAGE =	0x0A00 + 0x59,
		/* Which bits to remove from the color indirectly referenced          */
		EDIT_TARGET_MAT_TEV_IND_STAGE_FORMAT =	0x0A00 + 0x5A,
		/* Whether to apply bias to each element of the indirect offset   */
		EDIT_TARGET_MAT_TEV_IND_STAGE_BIAS_SEL =	0x0A00 + 0x5B,
		/* Which matrix and scale value to multiply with the offset       */
		EDIT_TARGET_MAT_TEV_IND_STAGE_MATRIX_SEL =	0x0A00 + 0x5C,
		/* Wrap coefficient for the S component of regular texture coordinates              */
		EDIT_TARGET_MAT_TEV_IND_STAGE_WRAP_S =	0x0A00 + 0x5D,
		/* Wrap coefficient for the T component of regular texture coordinates              */
		EDIT_TARGET_MAT_TEV_IND_STAGE_WRAP_T =	0x0A00 + 0x5E,
		/* Whether to add the texture coordinate results from the previous TEV stage      */
		EDIT_TARGET_MAT_TEV_IND_STAGE_ADD_PREV =	0x0A00 + 0x5F,
		/* Whether to use texture coordinates changed by mipmap LOD calculations      */
		EDIT_TARGET_MAT_TEV_IND_STAGE_UTC_LOD =	0x0A00 + 0x60,
		/* Which offset elements to set to "bump alpha"         */
		EDIT_TARGET_MAT_TEV_IND_STAGE_ALPHA_SEL =	0x0A00 + 0x61,
		/* Maximum length to which fur can grow                                 */
		EDIT_TARGET_MAT_FUR_LENGTH =	0x0A00 + 0x62,
		/* Number of fur layers                                       */
		EDIT_TARGET_MAT_FUR_LAYER_SIZE =	0x0A00 + 0x63,
		/* Type of spacing between fur layers                             */
		EDIT_TARGET_MAT_FUR_LAYER_INTERVAL =	0x0A00 + 0x64,
		/* Parameter for adjusting the fur layers' transparency curve             */
		EDIT_TARGET_MAT_FUR_ALPHA_CURVE =	0x0A00 + 0x65,
		/* Parameter for adjusting the manner in which specular light hits             */
		EDIT_TARGET_MAT_FUR_SPECULAR_CURVE =	0x0A00 + 0x66,
		/* Color to apply so that it becomes the hair tip                                   */
		EDIT_TARGET_MAT_FUR_TIP_COLOR =	0x0A00 + 0x67,
		/* Color to apply to be close to the hair root                             */
		EDIT_TARGET_MAT_FUR_ROOT_COLOR =	0x0A00 + 0x68,
		/* Hair brightness                                               */
		EDIT_TARGET_MAT_FUR_BRIGHTNESS =	0x0A00 + 0x69,
		/* Lie granularity                                             */
		EDIT_TARGET_MAT_FUR_TEXMTX_SCALE =	0x0A00 + 0x70,
		/* Diffusion function setting                                             */
		EDIT_TARGET_MAT_CHAN_DIFF_FUNC =	0x0A00 + 0x71,
		/* Attenuation function setting                                             */
		EDIT_TARGET_MAT_CHAN_ATTN_FUNC =	0x0A00 + 0x72,
		/* Constant alpha write enable flag                   */
		EDIT_TARGET_MAT_PE_DST_ALPHA_ENABLE =	0x0A00 + 0x73,
		/* Alpha value setting for constant alpha write                     */
		EDIT_TARGET_MAT_PE_DST_ALPHA_ALPHA =	0x0A00 + 0x74,
		/* Define enum as 4-byte type */
		EditTargetMat_4BYTE_ENUM =	0xFFFFFFFF
	};

	//-----------------------------------------------------------------------
	/* Texture mapping method */
	enum TargetMatTexMtxMappingMethodType {
		/* Vertex texture coordinates              */
		TARGET_MAT_TEXMTX_MAPPING_METHOD_TEXCOORD =	0x00,
		/* Environment map from the camera          */
		TARGET_MAT_TEXMTX_MAPPING_METHOD_ENV_CAMERA =	0x01,
		/* Environment map from lights          */
		TARGET_MAT_TEXMTX_MAPPING_METHOD_ENV_LIGHT =	0x02,
		/* Environment map from half vectors  */
		TARGET_MAT_TEXMTX_MAPPING_METHOD_ENV_SPEC =	0x03,
		/* Projection map                      */
		TARGET_MAT_TEXMTX_MAPPING_METHOD_PROJECTION =	0x04,
		/* Define enum as 4-byte type */
		TargetMatTexMtxMappingMethodType_4BYTE_ENUM =	0xFFFFFFFF
	};

	//-----------------------------------------------------------------------
	/* Indirect texture mapping method */
	enum TargetMatIndMtxIndirectMethodType {
		/* Indirect warp                                    */
		TARGET_MAT_INDMTX_MAPPING_METHOD_WARP =	0x00,
		/* Normal mapping using indirect                    */
		TARGET_MAT_INDMTX_MAPPING_METHOD_NRMMAP =	0x01,
		/* Add bright spot to the expression of normal mapping using indirect  */
		TARGET_MAT_INDMTX_MAPPING_METHOD_NRMMAP_SPECULAR =	0x02,
		/* Define enum as 4-byte type */
		TargetMatIndMtxIndirectMethodType_4BYTE_ENUM =	0xFFFFFFFF
	};

	//-----------------------------------------------------------------------
	/* Indirect texture mapping method */
	enum TargetMatIndStgScale {
		/* No scale. */
		TARGET_MAT_INDSTG_SCALE_ITS_1 =	0x00,
		/* 1/2                */
		TARGET_MAT_INDSTG_SCALE_ITS_2 =	0x01,
		/* 1/4                */
		TARGET_MAT_INDSTG_SCALE_ITS_4 =	0x02,
		/* 1/8                */
		TARGET_MAT_INDSTG_SCALE_ITS_8 =	0x03,
		/* 1/16               */
		TARGET_MAT_INDSTG_SCALE_ITS_16 =	0x04,
		/* 1/32               */
		TARGET_MAT_INDSTG_SCALE_ITS_32 =	0x05,
		/* 1/64               */
		TARGET_MAT_INDSTG_SCALE_ITS_64 =	0x06,
		/* 1/128              */
		TARGET_MAT_INDSTG_SCALE_ITS_128 =	0x07,
		/* 1/256              */
		TARGET_MAT_INDSTG_SCALE_ITS_256 =	0x08,
		/* Define enum as 4-byte type */
		TargetMatIndStgScale_4BYTE_ENUM =	0xFFFFFFFF
	};

	//-----------------------------------------------------------------------
	/* Which bits to remove from the color indirectly referenced */
	enum TargetMatTevIndFormat {
		/* Lower 8 bits           */
		TARGET_MAT_TEV_IND_FORMAT_ITF_8 =	0x00,
		/* Lower 5 bits           */
		TARGET_MAT_TEV_IND_FORMAT_ITF_5 =	0x01,
		/* Lower 4 bits           */
		TARGET_MAT_TEV_IND_FORMAT_ITF_4 =	0x02,
		/* Lower 3 bits           */
		TARGET_MAT_TEV_IND_FORMAT_ITF_3 =	0x03,
		/* Define enum as 4-byte type */
		TargetMatTevIndFormat_4BYTE_ENUM =	0xFFFFFFFF
	};

	//-----------------------------------------------------------------------
	/* Bias of each indirect offset element */
	enum TargetMatTevIndBiasSel {
		/* No               */
		TARGET_MAT_TEV_IND_BIASSEL_ITB_NONE =	0x00,
		/* Only S element          */
		TARGET_MAT_TEV_IND_BIASSEL_ITB_S =	0x01,
		/* Only T element          */
		TARGET_MAT_TEV_IND_BIASSEL_ITB_T =	0x02,
		/* Only U element          */
		TARGET_MAT_TEV_IND_BIASSEL_ITB_U =	0x03,
		/* S element and T element       */
		TARGET_MAT_TEV_IND_BIASSEL_ITB_ST =	0x04,
		/* S element and U element       */
		TARGET_MAT_TEV_IND_BIASSEL_ITB_SU =	0x05,
		/* T element and U element       */
		TARGET_MAT_TEV_IND_BIASSEL_ITB_TU =	0x06,
		/* all S, T, and U elements  */
		TARGET_MAT_TEV_IND_BIASSEL_ITB_STU =	0x07,
		/* Define enum as 4-byte type */
		TargetMatTevIndBiasSel_4BYTE_ENUM =	0xFFFFFFFF
	};

	//-----------------------------------------------------------------------
	/* Which matrix and scale value to multiply with the offset */
	enum TargetMatTevIndMtxSel {
		/* Zero matrix                                              */
		TARGET_MAT_TEV_IND_MTXSEL_ITM_OFF =	0x00,
		/* Indirect matrix 0 and matrix scale value 0  */
		TARGET_MAT_TEV_IND_MTXSEL_ITM_0 =	0x01,
		/* Indirect matrix 1 and matrix scale value 1  */
		TARGET_MAT_TEV_IND_MTXSEL_ITM_1 =	0x02,
		/* Indirect matrix 2 and matrix scale value 2  */
		TARGET_MAT_TEV_IND_MTXSEL_ITM_2 =	0x03,
		/* Dynamic S matrix and matrix scale value 0    */
		TARGET_MAT_TEV_IND_MTXSEL_ITM_S0 =	0x04,
		/* Dynamic S matrix and matrix scale value 1    */
		TARGET_MAT_TEV_IND_MTXSEL_ITM_S1 =	0x05,
		/* Dynamic S matrix and matrix scale value 1    */
		TARGET_MAT_TEV_IND_MTXSEL_ITM_S2 =	0x06,
		/* Dynamic T matrix and matrix scale value 2    */
		TARGET_MAT_TEV_IND_MTXSEL_ITM_T0 =	0x07,
		/* Dynamic T matrix and matrix scale value 0    */
		TARGET_MAT_TEV_IND_MTXSEL_ITM_T1 =	0x08,
		/* Dynamic T matrix and matrix scale value 0    */
		TARGET_MAT_TEV_IND_MTXSEL_ITM_T2 =	0x09,
		/* Define enum as 4-byte type */
		TargetMatTevIndMtxSel_4BYTE_ENUM =	0xFFFFFFFF
	};

	//-----------------------------------------------------------------------
	/* Wrap coefficient to apply to the regular texture coordinates */
	enum TargetMatTevIndWrap {
		/* No wrap               */
		TARGET_MAT_TEV_IND_WRAP_ITW_OFF =	0x00,
		/* Wrap at 256            */
		TARGET_MAT_TEV_IND_WRAP_ITW_256 =	0x01,
		/* Wrap at 128            */
		TARGET_MAT_TEV_IND_WRAP_ITW_128 =	0x02,
		/* Wrap at 64             */
		TARGET_MAT_TEV_IND_WRAP_ITW_64 =	0x03,
		/* Wrap at 32             */
		TARGET_MAT_TEV_IND_WRAP_ITW_32 =	0x04,
		/* Wrap at 16             */
		TARGET_MAT_TEV_IND_WRAP_ITW_16 =	0x05,
		/* Set all output coordinates to zero   */
		TARGET_MAT_TEV_IND_WRAP_ITW_0 =	0x06,
		/* Define enum as 4-byte type */
		TargetMatTevIndWrap_4BYTE_ENUM =	0xFFFFFFFF
	};

	//-----------------------------------------------------------------------
	/* Which offset elements to set to indirect "bump alpha" */
	enum TargetMatTevIndAlphaSel {
		/* No bump alpha                 */
		TARGET_MAT_TEV_IND_ALPHASEL_ITBA_OFF =	0x00,
		/* S element provides bump alpha  */
		TARGET_MAT_TEV_IND_ALPHASEL_ITBA_S =	0x01,
		/* T element provides bump alpha  */
		TARGET_MAT_TEV_IND_ALPHASEL_ITBA_T =	0x02,
		/* U element provides bump alpha  */
		TARGET_MAT_TEV_IND_ALPHASEL_ITBA_U =	0x03,
		/* Define enum as 4-byte type */
		TargetMatTevIndAlphaSel_4BYTE_ENUM =	0xFFFFFFFF
	};

	//-----------------------------------------------------------------------
	/* Diffusion function setting type */
	enum TargetMatChanDiffFunc {
		/* Disable diffuse lighting                               */
		TARGET_MAT_CHAN_DIFFFUNC_NONE =	0x00,
		/* Enable use of diffuse lighting, and clamp between 0.0 and 1.0  */
		TARGET_MAT_CHAN_DIFFFUNC_CLAMP =	0x01,
		/* Enable use of diffuse lighting                           */
		TARGET_MAT_CHAN_DIFFFUNC_SIGN =	0x02,
		/* Define enum as 4-byte type */
		TargetMatChanDiffFunc_4BYTE_ENUM =	0xFFFFFFFF
	};

	//-----------------------------------------------------------------------
	/* Attenuation function setting type */
	enum TargetMatChanAttnFunc {
		/* No light attenuation    */
		TARGET_MAT_CHAN_ATTNFUNC_NONE =	0x00,
		/* Spot light    */
		TARGET_MAT_CHAN_ATTNFUNC_SPOT =	0x01,
		/* Specular light  */
		TARGET_MAT_CHAN_ATTNFUNC_SPEC =	0x02,
		/* Define enum as 4-byte type */
		TargetMatChanAttnFunc_4BYTE_ENUM =	0xFFFFFFFF
	};

	//-----------------------------------------------------------------------
	/* Type of spacing between fur layers */
	enum TargetMatFurLayerInterval {
		/* Overlap fur layers in equal intervals      */
		TARGET_MAT_FUR_LAYERINTERVAL_UNIFORM =	0x00,
		/* Overlap tightly so that the interval is the tip of the hair  */
		TARGET_MAT_FUR_LAYERINTERVAL_TIP =	0x01,
		/* Define enum as 4-byte type */
		TargetMatFurLayerInterval_4BYTE_ENUM =	0xFFFFFFFF
	};

	//-----------------------------------------------------------------------
	/* Camera animation edit target */
	enum EditTargetColorAnim {
		/* R-component of material color 0         */
		EDIT_TARGET_COLOR_ANIM_MAT_COLOR0_R =	0x0B00 + 0x01,
		/* G-component of material color 0         */
		EDIT_TARGET_COLOR_ANIM_MAT_COLOR0_G =	0x0B00 + 0x02,
		/* B-component of material color 0         */
		EDIT_TARGET_COLOR_ANIM_MAT_COLOR0_B =	0x0B00 + 0x03,
		/* A-component of material color 0         */
		EDIT_TARGET_COLOR_ANIM_MAT_COLOR0_A =	0x0B00 + 0x04,
		/* R-component of material color 1         */
		EDIT_TARGET_COLOR_ANIM_MAT_COLOR1_R =	0x0B00 + 0x11,
		/* G-component of material color 1         */
		EDIT_TARGET_COLOR_ANIM_MAT_COLOR1_G =	0x0B00 + 0x12,
		/* B-component of material color 1         */
		EDIT_TARGET_COLOR_ANIM_MAT_COLOR1_B =	0x0B00 + 0x13,
		/* A-component of material color 1         */
		EDIT_TARGET_COLOR_ANIM_MAT_COLOR1_A =	0x0B00 + 0x14,
		/* R-component of ambient color 0       */
		EDIT_TARGET_COLOR_ANIM_AMB_COLOR0_R =	0x0B00 + 0x21,
		/* G-component of ambient color 0       */
		EDIT_TARGET_COLOR_ANIM_AMB_COLOR0_G =	0x0B00 + 0x22,
		/* B-component of ambient color 0       */
		EDIT_TARGET_COLOR_ANIM_AMB_COLOR0_B =	0x0B00 + 0x23,
		/* A-component of ambient color 0       */
		EDIT_TARGET_COLOR_ANIM_AMB_COLOR0_A =	0x0B00 + 0x24,
		/* R-component of ambient color 1       */
		EDIT_TARGET_COLOR_ANIM_AMB_COLOR1_R =	0x0B00 + 0x31,
		/* G-component of ambient color 1       */
		EDIT_TARGET_COLOR_ANIM_AMB_COLOR1_G =	0x0B00 + 0x32,
		/* B-component of ambient color 1       */
		EDIT_TARGET_COLOR_ANIM_AMB_COLOR1_B =	0x0B00 + 0x33,
		/* A-component of ambient color 1       */
		EDIT_TARGET_COLOR_ANIM_AMB_COLOR1_A =	0x0B00 + 0x34,
		/* R-component of TEV color register 0        */
		EDIT_TARGET_COLOR_ANIM_TEV_COLOR0_R =	0x0B00 + 0x41,
		/* G-component of TEV color register 0        */
		EDIT_TARGET_COLOR_ANIM_TEV_COLOR0_G =	0x0B00 + 0x42,
		/* B-component of TEV color register 0        */
		EDIT_TARGET_COLOR_ANIM_TEV_COLOR0_B =	0x0B00 + 0x43,
		/* A-component of TEV color register 0        */
		EDIT_TARGET_COLOR_ANIM_TEV_COLOR0_A =	0x0B00 + 0x44,
		/* R-component of TEV color register 1        */
		EDIT_TARGET_COLOR_ANIM_TEV_COLOR1_R =	0x0B00 + 0x51,
		/* G-component of TEV color register 1        */
		EDIT_TARGET_COLOR_ANIM_TEV_COLOR1_G =	0x0B00 + 0x52,
		/* B-component of TEV color register 1        */
		EDIT_TARGET_COLOR_ANIM_TEV_COLOR1_B =	0x0B00 + 0x53,
		/* A-component of TEV color register 1        */
		EDIT_TARGET_COLOR_ANIM_TEV_COLOR1_A =	0x0B00 + 0x54,
		/* R-component of TEV color register 2        */
		EDIT_TARGET_COLOR_ANIM_TEV_COLOR2_R =	0x0B00 + 0x61,
		/* G-component of TEV color register 2        */
		EDIT_TARGET_COLOR_ANIM_TEV_COLOR2_G =	0x0B00 + 0x62,
		/* B-component of TEV color register 2        */
		EDIT_TARGET_COLOR_ANIM_TEV_COLOR2_B =	0x0B00 + 0x63,
		/* A-component of TEV color register 2        */
		EDIT_TARGET_COLOR_ANIM_TEV_COLOR2_A =	0x0B00 + 0x64,
		/* R-component of TEV constant register 0  */
		EDIT_TARGET_COLOR_ANIM_TEV_CONST0_R =	0x0B00 + 0x71,
		/* G-component of TEV constant register 0  */
		EDIT_TARGET_COLOR_ANIM_TEV_CONST0_G =	0x0B00 + 0x72,
		/* B-component of TEV constant register 0  */
		EDIT_TARGET_COLOR_ANIM_TEV_CONST0_B =	0x0B00 + 0x73,
		/* A-component of TEV constant register 0  */
		EDIT_TARGET_COLOR_ANIM_TEV_CONST0_A =	0x0B00 + 0x74,
		/* R-component of TEV constant register 1  */
		EDIT_TARGET_COLOR_ANIM_TEV_CONST1_R =	0x0B00 + 0x81,
		/* G-component of TEV constant register 1  */
		EDIT_TARGET_COLOR_ANIM_TEV_CONST1_G =	0x0B00 + 0x82,
		/* B-component of TEV constant register 1  */
		EDIT_TARGET_COLOR_ANIM_TEV_CONST1_B =	0x0B00 + 0x83,
		/* A-component of TEV constant register 1  */
		EDIT_TARGET_COLOR_ANIM_TEV_CONST1_A =	0x0B00 + 0x84,
		/* R-component of TEV constant register 2  */
		EDIT_TARGET_COLOR_ANIM_TEV_CONST2_R =	0x0B00 + 0x91,
		/* G-component of TEV constant register 2  */
		EDIT_TARGET_COLOR_ANIM_TEV_CONST2_G =	0x0B00 + 0x92,
		/* B-component of TEV constant register 2  */
		EDIT_TARGET_COLOR_ANIM_TEV_CONST2_B =	0x0B00 + 0x93,
		/* A-component of TEV constant register 2  */
		EDIT_TARGET_COLOR_ANIM_TEV_CONST2_A =	0x0B00 + 0x94,
		/* R-component of TEV constant register 3  */
		EDIT_TARGET_COLOR_ANIM_TEV_CONST3_R =	0x0B00 + 0xA1,
		/* G-component of TEV constant register 3  */
		EDIT_TARGET_COLOR_ANIM_TEV_CONST3_G =	0x0B00 + 0xA2,
		/* B-component of TEV constant register 3  */
		EDIT_TARGET_COLOR_ANIM_TEV_CONST3_B =	0x0B00 + 0xA3,
		/* A-component of TEV constant register 3  */
		EDIT_TARGET_COLOR_ANIM_TEV_CONST3_A =	0x0B00 + 0xA4,
		/* Define enum as 4-byte type */
		EditTargetColorAnim_4BYTE_ENUM =	0xFFFFFFFF
	};

	//-----------------------------------------------------------------------
	/* Texture SRT animation edit target */
	enum EditTargetTexSRTAnim {
		/* Scale S-component for texture matrix        */
		EDIT_TARGET_TEXSRT_ANIM_SCALE_S =	0x0C00 + 0x01,
		/* Scale R-component for texture matrix        */
		EDIT_TARGET_TEXSRT_ANIM_SCALE_R =	0x0C00 + 0x02,
		/* Scale T-component for texture matrix        */
		EDIT_TARGET_TEXSRT_ANIM_SCALE_T =	0x0C00 + 0x03,
		/* Rotation T-component for texture matrix        */
		EDIT_TARGET_TEXSRT_ANIM_ROTATE =	0x0C00 + 0x04,
		/* Translation S-component for texture matrix  */
		EDIT_TARGET_TEXSRT_ANIM_TRANS_S =	0x0C00 + 0x05,
		/* Translation T-component for texture matrix  */
		EDIT_TARGET_TEXSRT_ANIM_TRANS_T =	0x0C00 + 0x06,
		/* Define enum as 4-byte type */
		EditTargetTexSRTAnim_4BYTE_ENUM =	0xFFFFFFFF
	};

	//-----------------------------------------------------------------------
	/* Indirect SRT animation edit target */
	enum EditTargetIndSRTAnim {
		/* Scale S-component for indirect matrix        */
		EDIT_TARGET_INDSRT_ANIM_SCALE_S =	0x0D00 + 0x01,
		/* Scale R-component for indirect matrix        */
		EDIT_TARGET_INDSRT_ANIM_SCALE_R =	0x0D00 + 0x02,
		/* Scale T-component for indirect matrix        */
		EDIT_TARGET_INDSRT_ANIM_SCALE_T =	0x0D00 + 0x03,
		/* Rotation T-component for indirect matrix        */
		EDIT_TARGET_INDSRT_ANIM_ROTATE =	0x0D00 + 0x04,
		/* Translation S-component for indirect matrix  */
		EDIT_TARGET_INDSRT_ANIM_TRANS_S =	0x0D00 + 0x05,
		/* Translation T-component for indirect matrix  */
		EDIT_TARGET_INDSRT_ANIM_TRANS_T =	0x0D00 + 0x06,
		/* Define enum as 4-byte type */
		EditTargetIndSRTAnim_4BYTE_ENUM =	0xFFFFFFFF
	};

	//-----------------------------------------------------------------------
	/*
            int value change message
         */
	typedef struct {
		/*  2B: Edit target                                      */
		short	target;
		/*  2B: Edit target index (when it must be specified)    */
		short	target_index;
		/*  4B: Value                                                  */
		long	val;
	} EdEditInt;

	//-----------------------------------------------------------------------
	/*
            bool value change message
         */
	typedef struct {
		/*  2B: Edit target                                      */
		short	target;
		/*  2B: Edit target index (when it must be specified)    */
		short	target_index;
		/*  4B: Value                                                  */
		unsigned long	val;
	} EdEditBool;

	//-----------------------------------------------------------------------
	/*
            float value change message
         */
	typedef struct {
		/*  2B: Edit target                                      */
		short	target;
		/*  2B: Edit target index (when it must be specified)    */
		short	target_index;
		/*  4B: Value                                                  */
		float	val;
	} EdEditFloat;

	//-----------------------------------------------------------------------
	/*
            Vector value change message
         */
	typedef struct {
		/*  2B: Edit target                                      */
		short	target;
		/*  2B: Edit target index (when it must be specified)    */
		short	target_index;
		/*  4B: X element                                               */
		float	x;
		/*  4B: Y element                                               */
		float	y;
		/*  4B: Z element                                               */
		float	z;
	} EdEditVector;

	//-----------------------------------------------------------------------
	/*
            Color value change message
         */
	typedef struct {
		/*  2B: Edit target                                      */
		short	target;
		/*  2B: Edit target index (when it must be specified)    */
		short	target_index;
		/*  1B: R element                                               */
		unsigned char	r;
		/*  1B: G element                                               */
		unsigned char	g;
		/*  1B: B element                                               */
		unsigned char	b;
		/*  1B: A element                                               */
		unsigned char	a;
	} EdEditColor;

	//-----------------------------------------------------------------------
	/*
            ColorS10 value change message
         */
	typedef struct {
		/*  2B: Edit target                                      */
		short	target;
		/*  2B: Edit target index (when it must be specified)    */
		short	target_index;
		/*  2B: R element                                               */
		short	r;
		/*  2B: G element                                               */
		short	g;
		/*  2B: B element                                               */
		short	b;
		/*  2B: A element                                               */
		short	a;
	} EdEditColorS10;

	//-----------------------------------------------------------------------
	/*
            int value const animation change message
         */
	typedef struct {
		/*  2B: Edit target                                      */
		short	target;
		/*  2B: Edit target index (when it must be specified)    */
		short	target_index;
		/*  4B: Value                                                  */
		long	val;
	} EdEditConstAnimInt;

	//-----------------------------------------------------------------------
	/*
            bool value const animation change message
         */
	typedef struct {
		/*  2B: Edit target                                      */
		short	target;
		/*  2B: Edit target index (when it must be specified)    */
		short	target_index;
		/*  4B: Value                                                  */
		unsigned long	val;
	} EdEditConstAnimBool;

	//-----------------------------------------------------------------------
	/*
            float value const animation change message
         */
	typedef struct {
		/*  2B: Edit target                                      */
		short	target;
		/*  2B: Edit target index (when it must be specified)    */
		short	target_index;
		/*  4B: Value                                                  */
		float	val;
	} EdEditConstAnimFloat;

	//-----------------------------------------------------------------------
	/*
            Vector value const animation change message
         */
	typedef struct {
		/*  2B: Edit target                                      */
		short	target;
		/*  2B: Edit target index (when it must be specified)    */
		short	target_index;
		/*  4B: X element                                               */
		float	x;
		/*  4B: Y element                                               */
		float	y;
		/*  4B: Z element                                               */
		float	z;
	} EdEditConstAnimVector;

	//-----------------------------------------------------------------------
	/*
            Color value const animation change message
         */
	typedef struct {
		/*  2B: Edit target                                      */
		short	target;
		/*  2B: Edit target index (when it must be specified)    */
		short	target_index;
		/*  1B: R element                                               */
		unsigned char	r;
		/*  1B: G element                                               */
		unsigned char	g;
		/*  1B: B element                                               */
		unsigned char	b;
		/*  1B: A element                                               */
		unsigned char	a;
	} EdEditConstAnimColor;

	//-----------------------------------------------------------------------
	/*
            int value hermite animation change message
         */
	typedef struct {
		/*  2B: Edit target                                      */
		short	target;
		/*  2B: Edit target index (when it must be specified)    */
		short	target_index;
		/*  4B: Number of keys                                          */
		long	key_num;
		/*  4B: frame                                            */
		float	frame;
		/*  4B: Slope                                            */
		float	slope;
		/*  4B: Value                                                  */
		long	val;
	} EdEditHermiteAnimInt;

	//-----------------------------------------------------------------------
	/*
            bool value Hermite animation change message
         */
	typedef struct {
		/*  2B: Edit target                                      */
		short	target;
		/*  2B: Edit target index (when it must be specified)    */
		short	target_index;
		/*  4B: Number of keys                                          */
		long	key_num;
		/*  4B: frame                                            */
		float	frame;
		/*  4B: Slope                                            */
		float	slope;
		/*  4B: Value                                                  */
		unsigned long	val;
	} EdEditHermiteAnimBool;

	//-----------------------------------------------------------------------
	/*
            float value Hermite animation change message
         */
	typedef struct {
		/*  2B: Edit target                                      */
		short	target;
		/*  2B: Edit target index (when it must be specified)    */
		short	target_index;
		/*  4B: Number of keys                                          */
		long	key_num;
		/*  4B: frame                                            */
		float	frame;
		/*  4B: Slope                                            */
		float	slope;
		/*  4B: Value                                                  */
		float	val;
	} EdEditHermiteAnimFloat;

	//-----------------------------------------------------------------------
	/*
            Color value Hermite animation change message
         */
	typedef struct {
		/*  2B: Edit target                                      */
		short	target;
		/*  2B: Edit target index (when it must be specified)    */
		short	target_index;
		/*  4B: Number of keys                                          */
		long	key_num;
		/*  4B: frame                                            */
		float	frame;
		/*  4B: Slope                                            */
		float	slope;
		/*  1B: R element                                               */
		unsigned char	r;
		/*  1B: G element                                               */
		unsigned char	g;
		/*  1B: B element                                               */
		unsigned char	b;
		/*  1B: A element                                               */
		unsigned char	a;
	} EdEditHermiteAnimColor;

	//-----------------------------------------------------------------------
	/*
            Animation stop command structure
         */
	typedef struct {
		/*  4B: Frame specification to stop animation  */
		float	frame;
	} EdAnimCtrlStop;

	//-----------------------------------------------------------------------
	/*
            Animation playback command structure
         */
	typedef struct {
		/*  4B: Start frame specification to playback animation  */
		float	frame;
	} EdAnimCtrlPlay;

	//-----------------------------------------------------------------------
	/* Viewer system-related category message */
	/* Disconnection request command */
	const unsigned short	VW_SYSTEM_DISCONNECT =	0x00 + CATEGORY_SYSTEM;

	//-----------------------------------------------------------------------
	/* System error category message */
	/* Undefined command error */
	const unsigned short	ERR_SYSTEM_UNKNOWN_COMMAND =	0x0 + CATEGORY_SYSTEM + HEADER_ERROR_FLAG;

	//-----------------------------------------------------------------------
	/* Data error category message */
	/* Data load error         */
	const unsigned short	ERR_DATA_LOAD =	0x0 + CATEGORY_DATA + HEADER_ERROR_FLAG;

	//-----------------------------------------------------------------------
	/*
            Undefined command error message
            Returns the received undefined command with the command parameter.
         
*/
	typedef struct {
		/*   2B: Received undefined command   */
		unsigned short	command;
	} ErrSystemUnknownCommand;

	//-----------------------------------------------------------------------
	/* Load data error type */
	enum ErrDataLoadCode {
		/* Failed to load file.         */
		ERR_DATA_LOAD_FILE_NOT_FOUND =	0,
		/* Failed to allocate memory.   */
		ERR_DATA_LOAD_MEMORY_ALLOCATE =	1,
		/* No valid data exists.       */
		ERR_DATA_LOAD_INVALID_DATA =	2,
		/* An error occurred during MCS communication.            */
		ERR_DATA_LOAD_MCS_STREAM_FAIL =	3,
		/* Failed to bind to model data. */
		ERR_DATA_LOAD_BIND_TO_MODEL =	4,
		/* Define enum as 4-byte type */
		ErrDataLoadCode_4BYTE_ENUM =	0xFFFFFFFF
	};

	//-----------------------------------------------------------------------
	/*
            Data load error.
         */
	typedef struct {
		/*   4B: Error Code           */
		ErrDataLoadCode	code;
	} ErrDataLoad;

} /* namespace Protocol3D */
} /* namespace NW4R */

#pragma warn_padding reset

#endif /* NW4R_Protocol3D_H_ */
