/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#ifndef NW4R_ProtocolEffect_H_
#define NW4R_ProtocolEffect_H_

#pragma warn_padding on

/*

    ==========================================================================
    Note: This protocol header is automatically generated from $(NW4R_ROOT)/ToolUtil/ProtocolUtil/Protocol/ProtocolEffect/ProtocolEffect.xml.
   
   
    
    When editing this protocol, you must not directly edit the .h or .cs files but rather edit ProtocolEffect.xml and then re-convert.
    ==========================================================================
	
    Viewer - EffectMaker Communication Protocol
	
    Communication message type
        Viewer Message  Message sent from the viewer to the editor
        Editor Message  Message sent from the editor to the viewer
        Error Message    Message sent when an error occurs
		
    When communicating, be sure to send the MessageHeader and then send the MessageData.
    MessageHeader is composed of a 2-byte MessageType and a 2-byte MessageSize.
	
    +---------------------------+--------------------+
    |      MessageHeader        |    MessageData     |
    +-------------+-------------+                    |
    | MessageType | MessageSize | (MessageSize byte) |
    | (2byte)     | (2byte)     |                    |
    +-------------+-------------+--------------------+
	
    MessageType comprises a 1-bit ErrorFlag, a 1-bit ReplyFlag, a 6-bit Category, and an 8-bit Command.
	
    
    15             8 7             0
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    |E|R| Category  |    Command    |
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    When Error is TRUE, Reply is always FALSE.
 */
namespace NW4R{
namespace ProtocolEffect{

	//-----------------------------------------------------------------------
	/*
            Version Information Definition
            Please match usage rules for using the version number with those for each module and tool
         */
	/* Protocol version. */
	const unsigned long	PROTOCOL_VERSION =	0x00010010;

	//-----------------------------------------------------------------------
	/* Message flag and mask definition */
	/* Flag for errors   */
	const unsigned short	HEADER_ERROR_FLAG =	0x8000;
	/* Flag for responses     */
	const unsigned short	HEADER_REPLY_FLAG =	0x4000;
	/* Mask for categories */
	const unsigned short	HEADER_CATEGORY_MASK =	0x3F00;
	/* Mask for commands */
	const unsigned short	HEADER_COMMAND_MASK =	0x00FF;

	//-----------------------------------------------------------------------
	/* Category definitions */
	/* System-related categories        */
	const unsigned short	CATEGORY_SYSTEM =	0x0000;
	/* 3D model-related categories        */
	const unsigned short	CATEGORY_3D =	0x0100;
	/* Effect-related categories      */
	const unsigned short	CATEGORY_EFFECT =	0x0200;
	/* MCS communication test categories       */
	const unsigned short	CATEGORY_TEST =	0x3F00;
	/* 2D layout-related categories        */
	const unsigned short	CATEGORY_2D =	0x0300;

	//-----------------------------------------------------------------------
	/* Effect protocol message shared header */
	typedef struct {
		/* Message types.    */
		unsigned short	type;
		/* Message size    */
		unsigned short	size;
	} Header;

	//-----------------------------------------------------------------------
	/* System-related category message */
	/* Version request command        */
	const unsigned short	TOOL_SYSTEM_VERSION =	0x00 + CATEGORY_SYSTEM + HEADER_REPLY_FLAG;
	/* Command for synchronization                */
	const unsigned short	TOOL_SYSTEM_SYNC =	0x01 + CATEGORY_SYSTEM + HEADER_REPLY_FLAG;
	/* Module reset command    */
	const unsigned short	TOOL_SYSTEM_RESET =	0x02 + CATEGORY_SYSTEM;

	//-----------------------------------------------------------------------
	/* 3D model-related category message */
	/* Command to load model data            */
	const unsigned short	TOOL_3D_LOAD_MODEL =	0x00 + CATEGORY_3D + HEADER_REPLY_FLAG;
	/* Command to get loaded model list  */
	const unsigned short	TOOL_3D_GET_MODEL_LIST =	0x01 + CATEGORY_3D + HEADER_REPLY_FLAG;
	/* Command to get model node list          */
	const unsigned short	TOOL_3D_GET_NODE_LIST =	0x02 + CATEGORY_3D + HEADER_REPLY_FLAG;
	/* Command to get model action list          */
	const unsigned short	TOOL_3D_GET_ACTION_LIST =	0x03 + CATEGORY_3D + HEADER_REPLY_FLAG;

	//-----------------------------------------------------------------------
	/* Effect-related category message */
	/* Effect data load command            */
	const unsigned short	TOOL_EFFECT_LOAD =	0x00 + CATEGORY_EFFECT + HEADER_REPLY_FLAG;
	/* Effect selection command            */
	const unsigned short	TOOL_EFFECT_SELECT =	0x01 + CATEGORY_EFFECT;
	/* Command to start/stop effect            */
	const unsigned short	TOOL_EFFECT_RUNSTOP =	0x02 + CATEGORY_EFFECT;
	/* Command to forcibly switch to the effect mode            */
	const unsigned short	TOOL_EFFECT_FOCUS =	0x03 + CATEGORY_EFFECT;
	/* 3D linked preview data load command            */
	const unsigned short	TOOL_EFFECT_3DPREVIEW =	0x04 + CATEGORY_EFFECT + HEADER_REPLY_FLAG;
	/* 3D linked preview frame count command            */
	const unsigned short	TOOL_EFFECT_3DPREVIEW_FRAME =	0x05 + CATEGORY_EFFECT;
	/* Real time transfer command            */
	const unsigned short	TOOL_EFFECT_REAL_PARAM =	0x06 + CATEGORY_EFFECT;
	/* Real time transfer command            */
	const unsigned short	TOOL_EFFECT_REAL_PARAM_REPLY =	0x06 + CATEGORY_EFFECT + HEADER_REPLY_FLAG;
	/* PAD control command            */
	const unsigned short	TOOL_EFFECT_PADCONTROL =	0x07 + CATEGORY_EFFECT;
	/* Camera control command            */
	const unsigned short	TOOL_EFFECT_CAMERACONTROL =	0x08 + CATEGORY_EFFECT;
	/* Node focus command            */
	const unsigned short	TOOL_EFFECT_FOCUS_NODE =	0x09 + CATEGORY_EFFECT;
	/* Action selection command            */
	const unsigned short	TOOL_EFFECT_ACTION =	0x0a + CATEGORY_EFFECT;
	/* Preview frame count command            */
	const unsigned short	TOOL_EFFECT_PREVIEW_FRAME =	0x0b + CATEGORY_EFFECT;
	/* 3D linked preview SRT command            */
	const unsigned short	TOOL_EFFECT_3DPREVIEW_SRT =	0x0c + CATEGORY_EFFECT;
	/* 3D linked preview start/end command            */
	const unsigned short	TOOL_EFFECT_3DPREVIEW_START_END =	0x0d + CATEGORY_EFFECT;
	/* Project delete command            */
	const unsigned short	TOOL_EFFECT_UNLOAD_PROJECT =	0x0e + CATEGORY_EFFECT;
	/* Effect name get command            */
	const unsigned short	TOOL_EFFECT_GETEFFECT =	0x0f + CATEGORY_EFFECT+ HEADER_REPLY_FLAG;
	/* Child display skip command            */
	const unsigned short	TOOL_EFFECT_SKIP_CHILD =	0x10 + CATEGORY_EFFECT;
	/* Current frame command            */
	const unsigned short	TOOL_EFFECT_CURRENT_FRAME =	0x11 + CATEGORY_EFFECT;
	/* Viewer control command            */
	const unsigned short	TOOL_EFFECT_VIWERCONTROL =	0x12 + CATEGORY_EFFECT;
	/* Command to return number of 3D links     */
	const unsigned short	TOOL_EFFECT_COUNT_3DPREVIEW =	0x13 + CATEGORY_EFFECT + HEADER_REPLY_FLAG;
	/* Texture LOD setting command  */
	const unsigned short	TOOL_EFFECT_TEXTURE_LOD =	0x14 + CATEGORY_EFFECT;
	/* Real time parameter emitter */
	const unsigned long	REAL_PARAM_EMITTER =	0x01;
	/* Real time parameter particle */
	const unsigned long	REAL_PARAM_PARTICLE =	0x02;
	/* Real time parameter track */
	const unsigned long	REAL_PARAM_TRACK =	0x03;
	/* Real time parameter viewer */
	const unsigned long	REAL_PARAM_VIEWER =	0x04;

	//-----------------------------------------------------------------------
	/* 2D layout-related category message */
	/* Command to get loaded layout list  */
	const unsigned short	TOOL_2D_GET_LAYOUT_LIST =	0x01 + CATEGORY_2D + HEADER_REPLY_FLAG;
	/* Command to get layout pane list          */
	const unsigned short	TOOL_2D_GET_PANE_LIST =	0x02 + CATEGORY_2D + HEADER_REPLY_FLAG;
	/* Command to get pane tag list          */
	const unsigned short	TOOL_2D_GET_TAG_LIST =	0x03 + CATEGORY_2D + HEADER_REPLY_FLAG;

	//-----------------------------------------------------------------------
	/* Test category message */
	/* EFB color change command            */
	const unsigned short	TOOL_TEST_EFB_COLOR =	0x00 + CATEGORY_TEST;

	//-----------------------------------------------------------------------
	/*
         Version request response message
            Receive tool protocol version, and return viewer protocol version.
         */
	typedef struct {
		/* 4B: Version number */
		unsigned long	version;
	} ToolSystemVersion;

	//-----------------------------------------------------------------------
	/*
        Synchronization message
              When receiving this synchronization command, the Viewer returns to the application the response command that echoes the specified tagName name.
              
              The tagName can be freely set on the tool-side.
         

*/
	typedef struct {
		/* 4B: Tag name. */
		unsigned long	tagName;
	} ToolSystemSync;

	//-----------------------------------------------------------------------
	/*
            Load model data message
            When fileSize is 0, the data that has that model ID is considered to be destroyed.
            When loading succeeds, Tool3DLoadModelReply is returned as a successful response message.
            If unsuccessful, return the ER_DATA_LOAD command.
         

*/
	typedef struct {
		/*   4B: File size.        */
		unsigned long	fileSize;
		/* 260B: Path name for the data on the PC  */
		unsigned char	pathName[260];
	} Tool3DLoadModel;

	//-----------------------------------------------------------------------
	/*
            Model data load response message
         */
	typedef struct {
		/*   4B: Management ID of the loaded model   */
		unsigned long	id;
	} Tool3DLoadModelReply;

	//-----------------------------------------------------------------------
	/*
            Loaded model list response message
            A variable length response message with a size number of model IDs in succession after this.
         */
	typedef struct {
		/*                 4B: Number of models loaded to the Viewer            */
		unsigned long	modelCount;
	} Tool3DGetModelListReply;

	//-----------------------------------------------------------------------
	/*
            Model list element structure
         */
	typedef struct {
		/*           4B: Model management ID                        */
		unsigned long	id;
		/*           4B: Model name string length (including NULL)           */
		unsigned long	nameSize;
		/*           4B: Current animation total number of frames  */
		float	totalFrames;
	} Tool3DGetModelListReplyElement;

	//-----------------------------------------------------------------------
	/*
            Model node list get message
         */
	typedef struct {
		/*   4B: Model management ID            */
		unsigned long	id;
	} Tool3DGetNodeList;

	//-----------------------------------------------------------------------
	/*
            Model node list response message
            When no model exists, nameLength is returned as 0
         */
	typedef struct {
		/*                4B: Number of nodes in the model     */
		long	nodeCount;
	} Tool3DGetNodeListReply;

	//-----------------------------------------------------------------------
	/*
            Node list element structure
         */
	typedef struct {
		/*           4B: Node index                                  */
		long	idx;
		/*           4B: Parent node index (negative value for a root node)  */
		long	parentIdx;
		/*           4B: Node name string length (including NULL)                           */
		unsigned long	nameSize;
	} Tool3DGetNodeListReplyElement;

	//-----------------------------------------------------------------------
	/*
            Get model action list message
         */
	typedef struct {
		/*   4B: Model management ID            */
		unsigned long	id;
	} Tool3DGetActionList;

	//-----------------------------------------------------------------------
	/*
            Model action list response message
            When no model exists, nameLength is returned as 0
         */
	typedef struct {
		/*                4B: Number of model actions     */
		long	actionCount;
	} Tool3DGetActionListReply;

	//-----------------------------------------------------------------------
	/*
            Action list element structure
         */
	typedef struct {
		/*           4B: Number of frames in the action                                  */
		long	frame;
		/*           4B: Action name string length (including NULL)                           */
		unsigned long	nameSize;
	} Tool3DGetActionListReplyElement;

	//-----------------------------------------------------------------------
	/* Effect load option */
	enum EffectLoadCode {
		/* Emitter load */
		TOOL_EFFECT_LOAD_EMITTER =	1,
		/* Load Texture */
		TOOL_EFFECT_LOAD_TEXTURE =	2,
		/* Load both */
		TOOL_EFFECT_LOAD_BOTH =	3,
		/* Define enum as 4-byte type */
		EffectLoadCode_4BYTE_ENUM =	0xFFFFFFFF
	};

	//-----------------------------------------------------------------------
	/*
            Load effect data message
         */
	typedef struct {
		/*   4B: File size.        */
		unsigned long	fileSize;
		/*   4B: Options        */
		unsigned long	option;
		/* 260B: Path name for the data on the PC  */
		unsigned char	pathName[260];
	} ToolEffectLoadRec;

	//-----------------------------------------------------------------------
	/*
            Load effect data response message
         */
	typedef struct {
		/*   4B: Management ID of the loaded effect   */
		unsigned long	id;
	} ToolEffectLoadReply;

	//-----------------------------------------------------------------------
	/*
            Effect selection message
         */
	typedef struct {
		/* 260B: Effect Name  */
		unsigned char	effectName[260];
	} ToolEffectSelectRec;

	//-----------------------------------------------------------------------
	/*
            Action selection message
         */
	typedef struct {
		/* 260B: Effect Name  */
		unsigned char	actionName[260];
	} ToolEffectActionRec;

	//-----------------------------------------------------------------------
	/*
            Effect's start/stop message
         */
	typedef struct {
		/* 1B: enum RunStopEffectCode  */
		unsigned char	sw;
		/* 1B:   */
		unsigned char	sw2;
		/* 1B:   */
		unsigned char	sw3;
		/* 1B:   */
		unsigned char	sw4;
		/* 1B:  */
		unsigned long	dispusage;
	} ToolEffectRunStopRec;

	//-----------------------------------------------------------------------
	/* Effect's start/stop message parameters */
	enum RunStopEffectCode {
		/* Stop all */
		TOOL_EFFECT_RUNSTOP_STOP =	0,
		/* Startup */
		TOOL_EFFECT_RUNSTOP_RUN =	1,
		/* Normal effect stop */
		TOOL_EFFECT_RUNSTOP_STOP_NORMAL =	2,
		/* 3D linked preview stop */
		TOOL_EFFECT_RUNSTOP_STOP_3D =	3,
		/* 2D linked preview stop */
		TOOL_EFFECT_RUNSTOP_STOP_2D =	4,
		/* Start with POSE */
		TOOL_EFFECT_RUNSTOP_POSE =	5,
		/* Reset (set to frame 0) */
		TOOL_EFFECT_RUNSTOP_RESET =	6,
		/* Information display off */
		TOOL_EFFECT_RUNSTOP_DISABLE_INFO =	7,
		/* Information display on */
		TOOL_EFFECT_RUNSTOP_ENABLE_INFO =	8,
		/* Define enum as 4-byte type */
		RunStopEffectCode_4BYTE_ENUM =	0xFFFFFFFF
	};

	//-----------------------------------------------------------------------
	/*
            Project deletion message
         */
	typedef struct {
		/* 260B: Project name  */
		unsigned char	projectName[260];
	} ToolEffectUnloadProjectRec;

	//-----------------------------------------------------------------------
	/*
            3D linked preview data message
         */
	typedef struct {
		/*   4B: File size.        */
		unsigned long	fileSize;
		/* 260B: Path name for the data on the PC  */
		unsigned char	pathName[260];
	} ToolEffect3DPreviewDataRec;

	//-----------------------------------------------------------------------
	/*
            3D linked preview data load response message
         */
	typedef struct {
		/*   4B: Number of preview data loaded normally   */
		unsigned long	count;
	} ToolEffect3DPreviewReply;

	//-----------------------------------------------------------------------
	/*
            3D linked preview frame count message
         */
	typedef struct {
		/* 3D linked preview frame count */
		unsigned long	frame[1];
	} ToolEffect3DPreviewFrameRec;

	//-----------------------------------------------------------------------
	/*
            3D linked preview SRT message
         */
	typedef struct {
		/* Scale       */
		float	sca[3];
		/* Rotation */
		float	rot[3];
		/* Translate */
		float	tra[3];
		/* Effect ID */
		unsigned long	id;
		/* 1B:   */
		unsigned char	option;
		/* 1B:   */
		unsigned char	padd1;
		/* 1B:   */
		unsigned char	padd2;
		/* 1B:   */
		unsigned char	padd3;
	} ToolEffect3DPreviewSRTRec;

	//-----------------------------------------------------------------------
	/*
            3D linked preview start/end message
         */
	typedef struct {
		/* 3D, layout selection/flag      */
		unsigned long	opt;
		/* Start       */
		unsigned long	start;
		/* End       */
		unsigned long	end;
		/* 260B: Link name  */
		unsigned char	previewName[260];
	} ToolEffect3DPreviewStartEndRec;

	//-----------------------------------------------------------------------
	/*
      Preview frame count message
       */
	typedef struct {
		/* Preview frame count */
		unsigned long	frame[1];
	} ToolEffectPreviewFrameRec;

	//-----------------------------------------------------------------------
	/*
            Real time transfer data message
         */
	typedef struct {
		/*  */
		unsigned long	type;
		/*  */
		unsigned long	kind;
		/*  */
		unsigned long	index;
		/*  */
		unsigned long	page;
		/*  */
		unsigned long	dispusage;
		/*  */
		unsigned long	allsize;
		/* Variable data length */
		unsigned long	size;
	} ToolEffectRealtimeDataRec;

	//-----------------------------------------------------------------------
	/*
            Real time transfer data message response message
         */
	typedef struct {
		/*   4B:  */
		unsigned long	id;
	} ToolEffectRealtimeDataReply;

	//-----------------------------------------------------------------------
	/*
      3D link count get message
       */
	typedef struct {
		/* Type  */
		unsigned long	type;
	} ToolEffectCount3DPreviewRec;

	//-----------------------------------------------------------------------
	/*
            3D link count get message response message
         */
	typedef struct {
		/*   4B:  */
		unsigned long	count;
	} ToolEffectCount3DPreviewReply;

	//-----------------------------------------------------------------------
	/*
            Skip child message
         */
	typedef struct {
		/*  */
		unsigned long	skip;
	} ToolEffectSkipChildRec;

	//-----------------------------------------------------------------------
	/*
            PAD control message
         */
	typedef struct {
		/* 2B:   */
		unsigned short	button;
		/* 1B:   */
		char	stickX;
		/* 1B:   */
		char	stickY;
		/* 1B:   */
		char	substickX;
		/* 1B:   */
		char	substickY;
		/* 1B:   */
		unsigned char	triggerLeft;
		/* 1B:   */
		unsigned char	triggerRight;
		/* 1B:   */
		unsigned char	analogA;
		/* 1B:   */
		unsigned char	analogB;
		/* 2B:   */
		unsigned short	press;
		/* 2B:   */
		unsigned short	trig;
		/* 2B:   */
		unsigned short	release;
		/* 2B:   */
		unsigned short	repeat;
		/* 2B:   */
		unsigned short	repeatAccel;
	} ToolEffectPadControlRec;

	//-----------------------------------------------------------------------
	/*
            Camera control message
         */
	typedef struct {
		/* 1B:   */
		unsigned short	usage;
		/* 2B:   */
		short	angleV;
		/* 2B:   */
		short	angleH;
		/* 2B:   */
		short	padding;
		/* 4B:   */
		float	distance;
	} ToolEffectCameraControlRec;

	//-----------------------------------------------------------------------
	/*
            Node focus message
         */
	typedef struct {
		/* 4B:   */
		unsigned long	model;
		/* 4B:   */
		long	node;
		/* 1B:   */
		unsigned char	camera_focus;
		/* 1B:   */
		unsigned char	padd1;
		/* 1B:   */
		unsigned char	padd2;
		/* 1B:   */
		unsigned char	padd3;
	} ToolEffectFocusNodeRec;

	//-----------------------------------------------------------------------
	/*  */
	typedef struct {
		/*  */
		unsigned long	getactive;
	} ToolEffectGetSelectedEffectRec;

	//-----------------------------------------------------------------------
	/*
            Selected effect get response message
         */
	typedef struct {
		/*                 4B: Project name length */
		unsigned long	projectNameLen;
		/*                 4B: Effect name length */
		unsigned long	effectNameLen;
		/* 260B: Project name  */
		unsigned char	projectName[260];
		/* 260B: Effect Name  */
		unsigned char	effectName[260];
	} ToolEffectGetSelectetEffectReply;

	//-----------------------------------------------------------------------
	/*
            Current frame message
         */
	typedef struct {
		/*  Current frame */
		unsigned long	currentFrame;
		/*  Project name length */
		unsigned long	psize;
		/*  Effect name length */
		unsigned long	esize;
	} ToolEffectCurrentFrameRec;

	//-----------------------------------------------------------------------
	/* Viewer control code */
	enum ViewerControlCode {
		/*  */
		TOOL_EFFECT_VIWERCONTROL_A =	1,
		/*  */
		TOOL_EFFECT_VIWERCONTROL_B =	2,
		/*  */
		TOOL_EFFECT_VIWERCONTROL_Play =	3,
		/*  */
		TOOL_EFFECT_VIWERCONTROL_Pose =	4,
		/*  */
		TOOL_EFFECT_VIWERCONTROL_Step =	5,
		/*  */
		TOOL_EFFECT_VIWERCONTROL_Reset =	6,
		/* Define enum as 4-byte type */
		ViewerControlCode_4BYTE_ENUM =	0xFFFFFFFF
	};

	//-----------------------------------------------------------------------
	/*
            Viewer control message
         */
	typedef struct {
		/*  */
		unsigned long	kind;
	} ToolEffectViewerControlRec;

	//-----------------------------------------------------------------------
	/*
            Texture LOD data message
         */
	typedef struct {
		/*   4B:         */
		unsigned long	min_filt;
		/*   4B:        */
		unsigned long	mag_filt;
		/* 4B:   */
		float	lod_bias;
		/* 260B: Texture instruction */
		unsigned char	texname[260];
	} ToolEffectSetTextureLODRec;

	//-----------------------------------------------------------------------
	/*
            Loaded layout list response message
            A variable length response message with a size number of IDs in succession after this.
         */
	typedef struct {
		/*                 4B: Number of layouts loaded to the Viewer            */
		unsigned long	layoutCount;
	} Tool2DGetLayoutListReply;

	//-----------------------------------------------------------------------
	/*
            Layout list element structure
         */
	typedef struct {
		/*           4B: Layout management ID                        */
		unsigned long	id;
		/*           4B: Layout name string length (including NULL)           */
		unsigned long	nameSize;
		/*           4B: Current animation total number of frames  */
		float	totalFrames;
	} Tool2DGetLayoutListReplyElement;

	//-----------------------------------------------------------------------
	/*
            Layout pane list get message
         */
	typedef struct {
		/*   4B: Pane management ID            */
		unsigned long	id;
	} Tool2DGetPaneList;

	//-----------------------------------------------------------------------
	/*
            Layout pane list response message
            When no layout exists, nameLength is returned as 0
         */
	typedef struct {
		/*                4B: Number of panes in the layout     */
		long	paneCount;
	} Tool2DGetPaneListReply;

	//-----------------------------------------------------------------------
	/*
            Pane list element structure
         */
	typedef struct {
		/*           4B: Pane index                                  */
		long	idx;
		/*           4B: Parent pane index				      */
		long	parentIdx;
		/*           4B: Pane name string length (including NULL)                           */
		unsigned long	nameSize;
	} Tool2DGetPaneListReplyElement;

	//-----------------------------------------------------------------------
	/*
            Layout tag list get message
         */
	typedef struct {
		/*   4B: Management ID            */
		unsigned long	id;
	} Tool2DGetTagList;

	//-----------------------------------------------------------------------
	/*
            Layout tag list response message
            When no layout exists, nameLength is returned as 0
         */
	typedef struct {
		/*                4B: Number of layout tags     */
		long	tagCount;
	} Tool2DGetTagListReply;

	//-----------------------------------------------------------------------
	/*
            Tag list element structure
         */
	typedef struct {
		/*           4B: Number of frames for tags                                  */
		long	frame;
		/*           4B: Tag name string length (including NULL)                           */
		unsigned long	nameSize;
	} Tool2DGetTagListReplyElement;

	//-----------------------------------------------------------------------
	/*
            EFB color change message
         */
	typedef struct {
		/* 1B: R Component */
		unsigned char	r;
		/* 1B: G Component */
		unsigned char	g;
		/* 1B: B Component */
		unsigned char	b;
		/* 1B: A Component */
		unsigned char	a;
	} ToolTestEfbColor;

	//-----------------------------------------------------------------------
	/* Viewer-initiated system-related category message */
	/* Viewer-initiated test command */
	const unsigned short	VIEWER_TEST_NOP =	0x00 + CATEGORY_TEST;

	//-----------------------------------------------------------------------
	/* System error category message */
	/* Undefined command error */
	const unsigned short	ERR_SYSTEM_UNKNOWN_COMMAND =	0x0 + CATEGORY_SYSTEM + HEADER_ERROR_FLAG;

	//-----------------------------------------------------------------------
	/* Data error category message */
	/* Data load error         */
	const unsigned short	ERR_3D_LOAD =	0x0 + CATEGORY_3D + HEADER_ERROR_FLAG;

	//-----------------------------------------------------------------------
	/*
            Undefined command error message
            Returns the received undefined command with the command parameter.
         
*/
	typedef struct {
		/*   2B: Received undefined command   */
		unsigned short	command;
	} ErrSystemUnknownCommand;

	//-----------------------------------------------------------------------
	/* Load data error type */
	enum Err3DLoadCode {
		/* Failed to load file.         */
		ERR_3D_LOAD_FILE_NOT_FOUND =	0,
		/* Failed to allocate memory.   */
		ERR_3D_LOAD_MEMORY_ALLOCATE =	1,
		/* No valid data exists.       */
		ERR_3D_LOAD_INVALID_DATA =	2,
		/* An error occurred during MCS communication.            */
		ERR_3D_LOAD_MCS_STREAM_FAIL =	3,
		/* Define enum as 4-byte type */
		Err3DLoadCode_4BYTE_ENUM =	0xFFFFFFFF
	};

	//-----------------------------------------------------------------------
	/*
            Data load error.
         */
	typedef struct {
		/*   4B: Error Code           */
		Err3DLoadCode	code;
	} Err3DLoad;

} /* namespace ProtocolEffect */
} /* namespace NW4R */

#pragma warn_padding reset

#endif /* NW4R_ProtocolEffect_H_ */
