/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#ifndef NW4R_ProtocolSample_H_
#define NW4R_ProtocolSample_H_

/*
    ==========================================================================
    Note: This protocol header is automatically generated from $(NW4R_ROOT)/ToolUtil/ProtocolUtil/Protocol/ProtocolSample/ProtocolSample.xml.
    
    
    When editing this protocol, you must not directly edit the .h or .cs files but rather edit ProtocolSample.xml and then re-convert.
    
    ==========================================================================
    Viewer - ProtocolSampleApplication Communication Protocol
    Communication message type
        Viewer Message  Message sent from the viewer to the editor
        Editor Message  Message sent from the editor to the viewer
        Error Message    Message sent when an error occurs
    When communicating, be sure to send the MessageHeader and then send the MessageData.
    MessageHeader is composed of a 2-byte MessageType and a 2-byte MessageSize.
    +---------------------------+--------------------+
    |      MessageHeader        |    MessageData     |
    +-------------+-------------+                    |
    | MessageType | MessageSize | (MessageSize byte) |
    | (2byte)     | (2byte)     |                    |
    +-------------+-------------+--------------------+
    MessageType comprises a 1-bit ErrorFlag, a 1-bit ReplyFlag, a 6-bit Category, and an 8-bit Command.
    
    15             8 7             0
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    |E|R| Category  |    Command    |
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    When Error is TRUE, Reply is always FALSE.
 






*/
namespace NW4R{
namespace ProtocolSample{

	//-----------------------------------------------------------------------
	/*
            Version Information Definition
            Please match usage rules for using the version number with those for each module and tool
         */
	/* Protocol version. */
	const unsigned long	PROTOCOL_VERSION =	0x00010000;

	//-----------------------------------------------------------------------
	/* Message flag and mask definition */
	/* Flag for errors   */
	const unsigned short	HEADER_ERROR_FLAG =	0x8000;
	/* Flag for responses     */
	const unsigned short	HEADER_REPLY_FLAG =	0x4000;
	/* Mask for categories */
	const unsigned short	HEADER_CATEGORY_MASK =	0x3F00;
	/* Mask for commands */
	const unsigned short	HEADER_COMMAND_MASK =	0x00FF;

	//-----------------------------------------------------------------------
	/* Category definitions */
	/* System-related categories        */
	const unsigned short	CATEGORY_SYSTEM =	0x0000;
	/* 3D model-related categories        */
	const unsigned short	CATEGORY_3D =	0x0100;
	/* MCS communication test categories       */
	const unsigned short	CATEGORY_TEST =	0x3F00;

	//-----------------------------------------------------------------------
	/* Sample protocol message shared header */
	typedef struct {
		/* Message types.    */
		unsigned short	type;
		/* Message size    */
		unsigned short	size;
	} Header;

	//-----------------------------------------------------------------------
	/* System-related category message */
	/* Version request command        */
	const unsigned short	TOOL_SYSTEM_VERSION =	0x00 + CATEGORY_SYSTEM + HEADER_REPLY_FLAG;
	/* Command for synchronization                */
	const unsigned short	TOOL_SYSTEM_SYNC =	0x01 + CATEGORY_SYSTEM + HEADER_REPLY_FLAG;
	/* Module reset command    */
	const unsigned short	TOOL_SYSTEM_RESET =	0x02 + CATEGORY_SYSTEM;

	//-----------------------------------------------------------------------
	/* 3D model-related category message */
	/* Command to load model data            */
	const unsigned short	TOOL_3D_LOAD_MODEL =	0x00 + CATEGORY_3D + HEADER_REPLY_FLAG;
	/* Command to get loaded model list  */
	const unsigned short	TOOL_3D_GET_MODEL_LIST =	0x01 + CATEGORY_3D + HEADER_REPLY_FLAG;
	/* Command to get model node list          */
	const unsigned short	TOOL_3D_GET_NODE_LIST =	0x02 + CATEGORY_3D + HEADER_REPLY_FLAG;

	//-----------------------------------------------------------------------
	/* Test category message */
	/* EFB color change command            */
	const unsigned short	TOOL_TEST_EFB_COLOR =	0x00 + CATEGORY_TEST;

	//-----------------------------------------------------------------------
	/*
         Version request response message
            Receive tool protocol version, and return viewer protocol version.
         */
	typedef struct {
		/* 4B: Version number */
		unsigned long	version;
	} ToolSystemVersion;

	//-----------------------------------------------------------------------
	/*
        Synchronization message
              When receiving this synchronization command, the Viewer returns to the application the response command that echoes the specified tagName name.
              
              The tagName can be freely set on the tool-side.
         

*/
	typedef struct {
		/* 4B: Tag name. */
		unsigned long	tagName;
	} ToolSystemSync;

	//-----------------------------------------------------------------------
	/*
            Load model data message
            When fileSize is 0, the data that has that model ID is considered to be destroyed.
            When loading succeeds, Tool3DLoadModelReply is returned as a successful response message.
            If unsuccessful, return the ER_DATA_LOAD command.
         

*/
	typedef struct {
		/*   4B: File size.        */
		unsigned long	fileSize;
		/* 260B: Path name for the data on the PC  */
		unsigned char	pathName[260];
	} Tool3DLoadModel;

	//-----------------------------------------------------------------------
	/*
            Model data load response message
         */
	typedef struct {
		/*   4B: Management ID of the loaded model   */
		unsigned long	id;
	} Tool3DLoadModelReply;

	//-----------------------------------------------------------------------
	/*
            Loaded model list response message
            A variable length response message with a size number of model IDs in succession after this.
         */
	typedef struct {
		/*                 4B: Number of models loaded to the Viewer            */
		unsigned long	modelCount;
	} Tool3DGetModelListReply;

	//-----------------------------------------------------------------------
	/*
            Model list element structure
         */
	typedef struct {
		/*           4B: Model management ID               */
		unsigned long	id;
		/*           4B: Model name string length (including NULL)  */
		unsigned long	nameSize;
	} Tool3DGetModelListReplyElement;

	//-----------------------------------------------------------------------
	/*
            Model node list get message
         */
	typedef struct {
		/*   4B: Model management ID            */
		unsigned long	id;
	} Tool3DGetNodeList;

	//-----------------------------------------------------------------------
	/*
            Model node list response message
            When no model exists, nameLength is returned as 0
         */
	typedef struct {
		/*                4B: Number of nodes in the model     */
		long	nodeCount;
	} Tool3DGetNodeListReply;

	//-----------------------------------------------------------------------
	/*
            Node list element structure
         */
	typedef struct {
		/*           4B: Node index                                  */
		long	idx;
		/*           4B: Parent node index (negative value for a root node)  */
		long	parentIdx;
		/*           4B: Node name string length (including NULL)                           */
		unsigned long	nameSize;
	} Tool3DGetNodeListReplyElement;

	//-----------------------------------------------------------------------
	/*
         EFB color change message
         */
	typedef struct {
		/* 1B: R Component */
		unsigned char	r;
		/* 1B: G Component */
		unsigned char	g;
		/* 1B: B Component */
		unsigned char	b;
		/* 1B: A Component */
		unsigned char	a;
	} ToolTestEfbColor;

	//-----------------------------------------------------------------------
	/* Viewer-initiated system-related category message */
	/* Viewer-initiated test command */
	const unsigned short	VIEWER_TEST_NOP =	0x00 + CATEGORY_TEST;

	//-----------------------------------------------------------------------
	/* System error category message */
	/* Undefined command error */
	const unsigned short	ERR_SYSTEM_UNKNOWN_COMMAND =	0x0 + CATEGORY_SYSTEM + HEADER_ERROR_FLAG;

	//-----------------------------------------------------------------------
	/* Data error category message */
	/* Data load error         */
	const unsigned short	ERR_3D_LOAD =	0x0 + CATEGORY_3D + HEADER_ERROR_FLAG;

	//-----------------------------------------------------------------------
	/*
            Undefined command error message
            Returns the received undefined command with the command parameter.
         
*/
	typedef struct {
		/*   2B: Received undefined command   */
		unsigned short	command;
	} ErrSystemUnknownCommand;

	//-----------------------------------------------------------------------
	/* Load data error type */
	enum Err3DLoadCode {
		/* Failed to load file.         */
		ERR_3D_LOAD_FILE_NOT_FOUND =	0,
		/* Failed to allocate memory.   */
		ERR_3D_LOAD_MEMORY_ALLOCATE =	1,
		/* No valid data exists.       */
		ERR_3D_LOAD_INVALID_DATA =	2,
		/* An error occurred during MCS communication.            */
		ERR_3D_LOAD_MCS_STREAM_FAIL =	3,
		/* Define enum as 4-byte type */
		Err3DLoadCode_4BYTE_ENUM =	0xFFFFFFFF
	};

	//-----------------------------------------------------------------------
	/*
            Data load error.
         */
	typedef struct {
		/*   4B: Error Code           */
		Err3DLoadCode	code;
	} Err3DLoad;

} /* namespace ProtocolSample */
} /* namespace NW4R */

#endif /* NW4R_ProtocolSample_H_ */
