/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#ifndef NW4C_VIEWER_G3D_DIRECTIONAL_OBJ_H_
#define NW4C_VIEWER_G3D_DIRECTIONAL_OBJ_H_



#if NN_CURRENT_VERSION_NUMBER >= NN_VERSION_NUMBER(4,0,0,0)
#pragma diag_suppress 1301 // padding inserted in struct.
#pragma diag_suppress 2530 // padding added to end of struct.
#endif
namespace vw {
namespace g3d {


enum TOUCH_PANEL_MODE
{
    TOUCH_PANEL_MODE_NORMAL,
    TOUCH_PANEL_MODE_DRAG,
    TOUCH_PANEL_MODE_INERTIA
};


float DegreeToRadian(float deg);    // TODO: Transfer to an appropriate location.
void ComputeRotateRad(f32 rotateX, f32 rotateY, nw::math::VEC3& result);    // TODO: Transfer to an appropriate location.

/*---------------------------------------------------------------------------
  Name:         DirectionalInfo

  Description:  Directional object setting information structure.
  ---------------------------------------------------------------------------*/
struct DirectionalInfo
{
    f32                 Near;
    f32                 Far;
    f32                 Fovy;
    nw::math::VEC3      Position;
    nw::math::VEC3      TargetPos;
    nw::math::VEC3      UpVector;
    f32                 RotationInit;
    f32                 RotationStep;
    f32                 RotationMax;
    f32                 TranslationInit;
    f32                 TranslationStep;
    f32                 TranslationMax;
    f32                 TiltInit;
    f32                 DollyInit;
    f32                 DollyAccel;
    f32                 DollyDecel;
    f32                 DollyLimit;
    f32                 DollyThreshold;
    f32                 TiltAccel;
    f32                 TiltDecel;
    f32                 TiltLimit;
    f32                 SlideAccel;
    f32                 SlideDecel;
    f32                 SlideLimit;
    bool                TumbleXInverse;
    bool                TumbleYInverse;
    bool                TrackXInverse;
    bool                TrackYInverse;
    bool                DollyInverse;
    bool                DollyTargetInverse;
    bool                TiltXInverse;
    bool                TiltYInverse;
    f32                 FocusScale;

    // Stereoscopic vision parameters
    f32                 depthLevel;
    f32                 factor;
    bool                realSwitch;

    TOUCH_PANEL_MODE    TouchPanelMode;

    f32                 depthLevelBias;

    DirectionalInfo();
};


/*---------------------------------------------------------------------------
  Name:         PadEvent

  Description:  Pad event structure
  ---------------------------------------------------------------------------*/
struct PadEvent
{
    static const u32 HANDLE_ROTATE = 0x00000001;
    static const u32 HANDLE_DOLLY  = 0x00000002;
    static const u32 HANDLE_TILT   = 0x00000004;
    static const u32 HANDLE_SLIDE  = 0x00000008;
    static const u32 HANDLE_TRACK  = 0x00000010;
    static const u32 HANDLE_ZOOM   = 0x00000020;
    static const u32 HANDLE_RESET  = 0x00000040;

    u32             flags;            // Event to be processed.
    f32             rotateH;          // Horizontal rotation angle
    f32             rotateV;          // Vertical rotation angle
    f32             dolly;            // Dolly direction
    f32             tiltH;            // Horizontal tilt angle
    f32             tiltV;            // Vertical tilt angle
    f32             slideX;           // Slide X direction
    f32             slideZ;           // Slide Z direction
    f32             trackH;           // Horizontal slide movement direction
    f32             trackV;           // Vertical slide movement direction
    f32             zoom;             // Zoom direction
    bool            moveTarget;       // Whether to move the gaze point when there is a dolly.
    bool            moveInViewSpace;  // Whether to move using view space.

    PadEvent();
};

/*---------------------------------------------------------------------------
  Name:         DirectionalObj

  Description:  Object having a direction (light, camera)
  ---------------------------------------------------------------------------*/
class DirectionalObj
{
    //-----------------------------------------------------------------------
    // list of friend classes and methods
    //-----------------------------------------------------------------------
    /* NOTHING */

public:
    //-----------------------------------------------------------------------
    // public variables
    //-----------------------------------------------------------------------
    enum UpdateType
    {
        UPDATE_TYPE_AIM = 0,        // Aim mode
        UPDATE_TYPE_LOOKAT,         // LooktAt mode
        UPDATE_TYPE_ROTATE,         // Rotate mode

        NUM_UPDATE_TYPE
    };

    //-----------------------------------------------------------------------
    // public methods
    //-----------------------------------------------------------------------
    DirectionalObj();
    virtual ~DirectionalObj();

    void                    SetUpdateType( const u32 updateType );
    u32                     GetUpdateType() const;

    void                    SetPadEventFromHid( vw::system::HidStatus& hidStatus, const bool isVeiwTypeRotate );
    void                    SetPadEvent( PadEvent& event );
    PadEvent                GetPadEvent() const;

    void                    SetTranslate( const f32 x, const f32 y, const f32 z );
    void                    SetTranslate( const nw::math::VEC3& translate );
    nw::math::VEC3          GetTranslate() const;

    void                    SetTargetPos( const f32 x, const f32 y, const f32 z );
    void                    SetTargetPos( const nw::math::VEC3& targetPos );
    nw::math::VEC3          GetTargetPos() const;

    void                    SetUpward( const f32 x, const f32 y, const f32 z );
    void                    SetUpward( const nw::math::VEC3& upward );
    nw::math::VEC3          GetUpward() const;

    void                    SetRotate( const f32 x, const f32 y, const f32 z );
    void                    SetRotate( const nw::math::VEC3& rotate );
    nw::math::VEC3          GetRotate() const;

    void                    SetTwist( const f32 twist );
    f32                     GetTwist() const;

    void                    SetOrientation( const nw::math::MTX33& orientation );
    nw::math::MTX33         GetOrientation() const;

    void                    SetRotateSpeed( const f32 rotateSpeed );
    f32                     GetRotateSpeed() const;

    void                    SetTransSpeed( const f32 transSpeed );
    f32                     GetTransSpeed() const;

    void                    SetTrackSpeed( const f32 trackSpeed );
    f32                     GetTrackSpeed() const;

    void                    SetTiltXSpeed( const f32 tiltXSpeed );
    f32                     GetTiltXSpeed() const;

    void                    SetTiltYSpeed( const f32 tiltYSpeed );
    f32                     GetTiltYSpeed() const;

    static void             SetDirectionalInfo( DirectionalInfo info );
    static DirectionalInfo& GetDirectionalInfo();

    void                    SetIsProjTypeOrtho( const bool isOrtho );
    bool                    GetIsProjTypeOrtho() const;

    void                    SetInitUpward( const nw::math::VEC3& initUpward );
    nw::math::VEC3          GetInitUpward() const;

    void                    Rotate( bool isLight );
    void                    Dolly( bool isLight );
    void                    Slide( bool isLight );
    void                    Track( bool isLight );
    void                    Tilt( bool isLight );

protected:
    //-----------------------------------------------------------------------
    // protected variables
    //-----------------------------------------------------------------------
    /* NOTHING */

    //-----------------------------------------------------------------------
    // protected methods
    //-----------------------------------------------------------------------
    /* NOTHING */

private:
    //-----------------------------------------------------------------------
    // Private variables
    //-----------------------------------------------------------------------
    nw::math::VEC3          m_Translate;
    nw::math::VEC3          m_TargetPos;
    nw::math::VEC3          m_Upward;
    nw::math::VEC3          m_Rotate;
    nw::math::VEC3          m_InitUpward;
    nw::math::MTX33         m_Orientation;

    f32                     m_Twist;
    f32                     m_RotateSpeed;
    f32                     m_TransSpeed;
    f32                     m_TrackSpeed;
    f32                     m_TiltXSpeed;
    f32                     m_TiltYSpeed;

    bool                    m_ProjTypeOrtho;

    PadEvent                m_PadEvent;

    u32                     m_UpdateType;
};

}   /* namespace g3d */
}   /* namespace vw */
#if NN_CURRENT_VERSION_NUMBER >= NN_VERSION_NUMBER(4,0,0,0)
#pragma diag_default 2530 // padding added to end of struct.
#pragma diag_default 1301 // padding inserted in struct.
#endif

/* NW4C_VIEWER_G3D_DIRECTIONAL_OBJ_H_ */
#endif
