/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

/*

  

  */

#ifndef NW4C_VIEWER_SYSTEM_BUTTON_H_
#define NW4C_VIEWER_SYSTEM_BUTTON_H_

#if NN_CURRENT_VERSION_NUMBER >= NN_VERSION_NUMBER(4,0,0,0)
#pragma diag_suppress 1301 // padding inserted in struct.
#pragma diag_suppress 2530 // padding added to end of struct.
#endif
namespace vw     {
namespace system {

//---------------------------------------------------------------------
//Double-click determination (also check for single click)
class DoubleClickChecker
{
public:
	DoubleClickChecker()
	: m_IsDoubleClick( false ),
	  m_IsSingleClick( false ),
	  m_IsTouch( false ),
	  m_HitRect( 0, 0, 0, 0 ),
	  m_PrevClickFrame( 10000 ), //The default value is large enough
	  m_PrevTouch( 0 )
	{
	}
	void Clear()
	{
		m_IsDoubleClick = false;
		m_IsSingleClick = false;
		m_IsTouch       = false;
		m_PrevClickFrame = 10000;
		m_PrevTouch = 0;
	}
	void SetHitRect( const nw::ut::Rect& rect ) { m_HitRect = rect; }
	void Update( const nn::hid::TouchPanelStatus& tp )
	{
		m_IsDoubleClick = false;
		m_IsSingleClick = false;
		m_IsTouch = false;
		if ( tp.touch && IsPosInRect( tp.x, tp.y, m_HitRect ) )
		{ //Being touched
			m_IsTouch = true;
			if ( ! m_PrevTouch )
			{ //Clicked this time
				m_IsSingleClick = true;
				if ( m_PrevClickFrame <= ACCEPT_FRAME )
				{ //Accepted range for double-click
					m_IsDoubleClick = true;
				}
				m_PrevClickFrame = 0;
			}
		}
		++ m_PrevClickFrame;
		m_PrevTouch = tp.touch;
	}
	bool IsDoubleClick() const { return m_IsDoubleClick; }
	bool IsSingleClick() const { return m_IsSingleClick; }
	bool IsTouch()		 const { return m_IsTouch;		 }

private:
	static const int ACCEPT_FRAME = 12;
	bool			m_IsDoubleClick;
	bool			m_IsSingleClick;
	bool			m_IsTouch;
	nw::ut::Rect	m_HitRect;
	int 			m_PrevClickFrame;
	u8			 	m_PrevTouch;
};


//---------------------------------------------------------------------
//Abstract button (
class Button {
public:
	nw::math::VEC2	pos;
	nw::math::VEC2	size;
	f32				tpChkMargin;
	int				repeatTime1st;
	int				repeatTime2nd;
	bool			visible;
	bool			enable;

	Button()
	: pos( 0, 0 ),
	  size( 32, 16 ),
	  tpChkMargin( 2.0f ),
	  repeatTime1st( 30 ),
	  repeatTime2nd(  8 ),
	  visible( true ),
	  enable( true ),
	  m_IsRepeat( false ),
	  m_PressTimer( 0 )
	{
	}
	void Update( const nn::hid::TouchPanelStatus& tp )
	{
		if ( visible && enable )
		{
			const nw::math::VEC2 tmVec2 = nw::math::VEC2( tpChkMargin, tpChkMargin );
			const nw::math::VEC2 tpos1 = pos - tmVec2;
			const nw::math::VEC2 tpos2 = pos + size + tmVec2;
			nw::ut::Rect hitRect( tpos1.x, tpos1.y, tpos2.x, tpos2.y );
			m_DoubleClickChecker.SetHitRect( hitRect );
			m_DoubleClickChecker.Update( tp );
		}
		else
		{
			m_DoubleClickChecker.Clear();
		}
		if ( IsPress() ) {
			++ m_PressTimer;
			//Double speed every second
			const int accelRepeatTime2nd = nw::ut::Max( 1, repeatTime2nd >> (m_PressTimer / 60) );
			m_IsRepeat = (   IsDown()
						|| ( m_PressTimer == repeatTime1st )
						|| ( m_PressTimer > repeatTime1st &&
							 (m_PressTimer - repeatTime1st) % accelRepeatTime2nd == 0
						   )
						 );
		} else {
			m_PressTimer = 0;
			m_IsRepeat = false;
		}
	}
	virtual void Disp() = 0;

	bool IsPress() 			const { return m_DoubleClickChecker.IsTouch(); }
	bool IsDown() 			const { return m_DoubleClickChecker.IsSingleClick(); }
	bool IsDoubleClick() 	const { return m_DoubleClickChecker.IsDoubleClick(); }
	bool IsRepeat() 		const { return m_IsRepeat; }

private:
	bool m_IsRepeat;
	int	 m_PressTimer;
	DoubleClickChecker m_DoubleClickChecker;
};


//---------------------------------------------------------------------
//Button with text
class TextButton : public Button {
public:
	//============ property =============
	/* Inherit from Button
	nw::math::VEC2	pos;
	nw::math::VEC2	size;
	f32				tpChkMargin;
	int				repeatTime1st;
	int				repeatTime2nd;
	bool			visible;
	bool			enable;
	*/
	const char * 		txtStr;
	f32					txtFontSize;

	nw::ut::Color8		txtCol;
	nw::ut::Color8		backCol;
	f32					wakuLineWidth;
	nw::ut::Color8		wakuCol;
	nw::ut::Color8		disableTxtCol;
	//===================================
	TextButton()
	:
	  txtStr( "Btn" ),
	  txtFontSize( 10.0f ),
	  txtCol( nw::ut::Color8::WHITE ),
	  backCol( 0, 0, 0, 255 ),
	  wakuLineWidth( 1.0f ),
	  wakuCol( nw::ut::Color8::WHITE ),
	  disableTxtCol( nw::ut::Color8::GRAY )
	{
	}
	//void Update( const nn::hid::TouchPanelStatus& tp ) Inherit from Button
	virtual void Disp() //Inherit from Button
	{
		using namespace nw::math;
		if ( ! visible ) return;
		AutoDrawAllParamBak paramBak;
		//Background
		FillRectangle( pos, size, backCol );
		if ( enable ) { //Frame
			SetLineWidth( wakuLineWidth );
			DrawRectangle( pos, size, wakuCol );
		}
		//Text
		SetTextColor( (enable) ? txtCol : disableTxtCol );
		SetFontSize( txtFontSize );
		{ //Centering
			const f32 txtWidth  = FontSizeToDx( txtFontSize ) * std::strlen( txtStr );
			const f32 txtHeight = FontSizeToDy( txtFontSize );
			SetTextPos( pos + size / 2 - nw::math::VEC2( txtWidth / 2, txtHeight / 2 ) );
		}
		Print( txtStr );

	}
	/* Inherit from Button
	bool IsPress() const;
	bool IsDown() const;
	bool IsDoubleClick() const;
	bool IsRepeat() const;
	*/
private:
};

//---------------------------------------------------------------------
//Arrow button
class ArrowButton : public Button {
public:
	/* Inherit from Button
	nw::math::VEC2	pos;
	nw::math::VEC2	size;
	f32				tpChkMargin;
	int				repeatTime1st;
	int				repeatTime2nd;
	bool			visible;
	bool			enable;
	*/
	enum ArrowDir { DIR_LEFT = 0, DIR_UP, DIR_RIGHT, DIR_DOWN, DIR_NUM };
	//====== Properties ======
	ArrowDir 	 	arrowDir;
	nw::ut::Color8 	arrowCol;
	f32 			margin;
	//==========================
	ArrowButton()
	: arrowDir( DIR_RIGHT ),
	  arrowCol( nw::ut::Color8(255, 255, 0, 255) ),
	  margin( 0.0f ) //Rendered smaller, by a margin amount, than the actual button size.
	{
		size = nw::math::VEC2( 10, 10 );
	}
	//void Update( const nn::hid::TouchPanelStatus& tp ) Inherit from Button
	virtual void Disp() //Inherit from Button
	{
		using namespace nw::math;
		if ( ! visible || ! enable ) return;
		VEC2 p1, p2, p3;
		switch( arrowDir ) {
			case DIR_LEFT :
							p1 = pos + VEC2( size.x  ,   margin );
							p2 = pos + VEC2( margin*2, size.y/2 );
							p3 = pos + VEC2( size.x  , size.y - margin );
							break;
			case DIR_UP :
							p1 = pos + VEC2( margin, size.y );
							p2 = pos + VEC2( size.x/2, margin * 2 );
							p3 = pos + VEC2( size.x - margin, size.y );;
							break;
			case DIR_RIGHT :
							p1 = pos + VEC2( 0, margin );
							p2 = pos + VEC2( size.x - margin * 2, size.y/2 );
							p3 = pos + VEC2( 0, size.y - margin );
							break;
			case DIR_DOWN :
							p1 = pos + VEC2( margin, 0 );
							p2 = pos + VEC2( size.x/2, size.y - margin * 2 );
							p3 = pos + VEC2( size.x - margin, 0 );
							break;
			default : NW_ASSERT(0); break;
		}
		FillTriangle( p1, p2, p3, arrowCol );
	}
	/* Inherit from Button
	bool IsPress() const;
	bool IsDown() const;
	bool IsDoubleClick() const;
	bool IsRepeat() const;
	*/
private:
};

//---------------------------------------------------------------------
//Increase/decrease button (text button + arrow buttons x 2)
class IncDecButton {
public:
	IncDecButton()
	:
	  m_RepeatOn( true ),
	  m_Visible( true ),
	  m_Enable( true ),
	  m_IncEnable( true ),
	  m_DecEnable( true ),
	  m_IncDec( 0 )
	{
		SetArea( 100, 200, 100 );
		SetText( "N/A" );
		m_DecButton.arrowDir = ArrowButton::DIR_LEFT;
		m_DecButton.margin = 1.0f;
		m_IncButton.arrowDir = ArrowButton::DIR_RIGHT;
		m_IncButton.margin = 1.0f;
	}
	void SetArea( f32 x1, f32 x2, f32 y, f32 fontSize=10.0f ) //Height/width automatically calculated from font size
	{
		using namespace nw::math;
		m_X1 = x1;
		m_X2 = x2;
		m_X0 = (x1 + x2)/2;
		m_Y  = y;
		m_FontSize = fontSize;
		m_FontDx = FontSizeToDx( fontSize );
		m_Dy = fontSize;
		m_DecButton.pos = VEC2( x1, y );
		m_IncButton.pos = VEC2( x2-m_Dy, y );
		m_DecButton.size = m_IncButton.size = VEC2( m_Dy, m_Dy );
		m_DoubleClickChecker.SetHitRect( nw::ut::Rect(x1+fontSize, y, x2-fontSize, y+m_Dy) );
	}
	void SetRepeatOn( bool repeatOn ) { m_RepeatOn = repeatOn; }
	void SetRepeatFastOn( bool repeatFast ) {
		if ( repeatFast ) SetRepeatOn( true );
		const int repeatTime2nd = (repeatFast) ? 4 : 8;
		m_DecButton.repeatTime2nd = repeatTime2nd;
		m_IncButton.repeatTime2nd = repeatTime2nd;
	}
	void SetVisible( bool visible )
	{
		m_Visible = visible;
		m_DecButton.visible = visible;
		m_IncButton.visible = visible;
	}
	void SetEnable( bool enable )
	{
		m_Enable = enable;
		m_DecButton.enable = enable;
		m_IncButton.enable = enable;
	}
	void SetIncDecEnable( bool incEnable, bool decEnable )
			{ m_IncEnable = incEnable; m_DecEnable = decEnable; }
	void Update( const nn::hid::TouchPanelStatus& tp )
	{
		m_IncDec = 0;
		if ( ! m_Visible || ! m_Enable ) {
			m_DoubleClickChecker.Clear();
			return;
		}
		nn::hid::TouchPanelStatus disableTp;
		disableTp.touch = 0;
		m_DecButton.Update( (m_DecEnable) ? tp : disableTp );
		m_IncButton.Update( (m_IncEnable) ? tp : disableTp );
		if ( m_RepeatOn ) {
			m_IncDec = m_IncButton.IsRepeat()  - m_DecButton.IsRepeat();
		} else {
			m_IncDec = m_IncButton.IsDown() - m_DecButton.IsDown();
		}
		m_DoubleClickChecker.Update(tp);
	}
	int GetIncDec() const { return m_IncDec; }
	bool IsDoubleClick() const { return m_DoubleClickChecker.IsDoubleClick(); }

	void SetText(const char *str) //The string buffer is statically allocated with the user side.
	{
		m_Str = str;
		m_StrLen = std::strlen( str );
	}
	void Disp()
	{
		if ( ! m_Visible ) return;
		AutoDrawAllParamBak autoBak;
		//Text
		using namespace nw::ut;
		SetFontSize( m_FontSize );
		SetTextPos( m_X0 - m_StrLen * m_FontDx / 2, m_Y );
		SetTextColor( (m_Enable) ? Color8::WHITE : Color8::GRAY );
		Print( m_Str );
		//Increase/decrease button
		if ( m_DecEnable ) m_DecButton.Disp();
		if ( m_IncEnable ) m_IncButton.Disp();
	}
private:
	ArrowButton	m_DecButton;
	ArrowButton	m_IncButton;
	f32 m_X1;
	f32 m_X2;
	f32 m_X0;
	f32 m_Y;
	f32 m_FontSize;
	f32 m_FontDx;
	f32 m_Dy;
	const char* m_Str;
	int m_StrLen;
	bool m_RepeatOn;
	bool m_Visible;
	bool m_Enable;
	bool m_IncEnable;
	bool m_DecEnable;
	int m_IncDec;
	DoubleClickChecker m_DoubleClickChecker;
};
//---------------------------------------------------------------------
//On/Off Button
class OnOffButton
{
public:
	OnOffButton()
	: m_Initialized( false ),
	  m_DefaultIsOn( false ),
	  m_InternalVal( m_DefaultIsOn ),
	  m_pVal( & m_InternalVal ),
	  m_IsChanged( false )
	{
	}
	void Initialize( const char* str, bool defaultIsOn, bool* pVal=NULL ) //If pVal=NULL, use internal variable
	{
		m_TextButton.txtStr = str;
		m_DefaultIsOn = defaultIsOn;
		if ( pVal ) m_pVal = pVal;
		(*m_pVal) = defaultIsOn;
		m_Initialized = true;
	}
	void SetIsOn( bool isOn ) { (*m_pVal) = isOn; }
	void SetEnable( bool enable ) { m_TextButton.enable = enable; }
	void Update( const nn::hid::TouchPanelStatus& tp )
	{
		NW_ASSERT( m_Initialized );
		const bool isOnBak = IsOn();
		m_TextButton.Update( tp );
		if ( m_TextButton.IsDown() )
		{
			SetIsOn( ! IsOn() );
		}
		if ( m_TextButton.IsDoubleClick() )
		{
			SetIsOn( m_DefaultIsOn );
		}
		m_IsChanged = (IsOn() != isOnBak);
	}
	bool IsOn() const { NW_POINTER_ASSERT( m_pVal ); return *m_pVal; }
	void Disp( const nw::math::VEC2& pos, const nw::math::VEC2& size, f32 fontSize )
	{
		NW_ASSERT( m_Initialized );
		m_TextButton.pos  = pos;
		m_TextButton.size = size;
		m_TextButton.txtFontSize = fontSize;
		if ( m_TextButton.enable ) {
			const bool isOn = *m_pVal;
			m_TextButton.txtCol  = (isOn) ? nw::ut::Color8::BLACK : nw::ut::Color8::WHITE;
			m_TextButton.backCol = (isOn) ? nw::ut::Color8::WHITE : nw::ut::Color8::BLACK;
		}
		else
		{
			m_TextButton.txtCol  = nw::ut::Color8::GRAY;
			m_TextButton.backCol = nw::ut::Color8::BLACK;
		}
		m_TextButton.Disp();
	}
	bool IsChanged() const { return m_IsChanged; }
private:
	bool		m_Initialized;
	TextButton	m_TextButton;
	bool		m_DefaultIsOn;
	bool		m_InternalVal;
	bool*		m_pVal;
	bool		m_IsChanged;
};

}   /* namespace system */
}   /* namespace vw     */
#if NN_CURRENT_VERSION_NUMBER >= NN_VERSION_NUMBER(4,0,0,0)
#pragma diag_default 2530 // padding added to end of struct.
#pragma diag_default 1301 // padding inserted in struct.
#endif

#endif // BUTTON_H_
