/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/
#ifndef NW4CVIEWER_DRAW3DHELPER_H_
#define NW4CVIEWER_DRAW3DHELPER_H_

#include <viewer.h>


namespace vw {

namespace os {
class IAllocator;
}

namespace font {
class Font;
}

namespace system {



/*
  
 */
class CommandCache
{
public:
    CommandCache();
    ~CommandCache();
    u32* Begin();
    void End( u32* command );
    u32* MakeCommand( u32* command ) const;
    void Release();
    u32  GetSize() const { return m_Size; }
    static void Finalize();

private:
    void ResizeBuffer(u32 size);

private:
    u32* m_Buffer;
    u32  m_Size;
    static u32* m_Data;
};

/*
  
 */
class XYZTC
{
public:
    struct Vertex
    {
        float x,y,z;
        float u,v;
        nw::ut::Color8 color;

        Vertex& setPos( float x_, float y_, float z_ )
        {
            x=x_, y=y_, z=z_;
            return *this;
        }
        Vertex& setUV( float u_, float v_ )
        {
            u=u_, v=v_;
            return *this;
        }
        Vertex& setColor( nw::ut::Color8 col )
        {
            color = col;
            return *this;
        }
    };
    XYZTC();
    ~XYZTC();
    bool CreateBuffer( u32 vertexNum, u32 indexNum );
    void ReleaseBuffer();
    void WriteBuffer(const Vertex* vertex, const u16* index);
    void UpdateVertex(u32 start, u32 size);
    void UpdateIndex(u32 start, u32 size);
    void UpdateData();
    void SetColor( float r, float g, float b, float a = 1.0f );
    void Draw() const;
    void Draw(u32 drawVtxNum);
    Vertex* GetVertexData() const{return m_VertexData;}
    u16*    GetIndexData()  const{return m_IndexData;}
    u32     GetVertexSize() const{return m_VertexSize;}
    u32     GetIndexSize()  const{return m_IndexSize;}

private:
    void CreateCommand();

private:
    nn::gr::Vertex              m_VertexStream;
    nn::gr::Vertex::IndexStream m_IndexStream;
    u32                         m_Size;
    u8*                         m_Data;
    Vertex*                     m_VertexData;
    u32                         m_VertexSize;
    u16*                        m_IndexData;
    u32                         m_IndexSize;
    CommandCache                m_Command;
};

/*
  
 */
class Draw3DHelper
{
public:
    static Draw3DHelper& GetInstance();
    static void Initialize();
    static void Finalize();
    void SetScreenSize(u32 width, u32 height);
    void Begin();
    void SetWorld( const nn::math::MTX34& world );
    void SetView ( const nn::math::MTX34& view );
    void SetProj ( const nn::math::MTX44& proj );
    void End();
    const nn::gr::BindSymbolVSInput* GetSymbolPos() const{return &m_SymbolPos;}
    const nn::gr::BindSymbolVSInput* GetSymbolTex() const{return &m_SymbolTex;}
    const nn::gr::BindSymbolVSInput* GetSymbolClr() const{return &m_SymbolClr;}
    u32*& GetCommand(){return m_Command;}

    enum State{
        ALPHABLEND = 1 << 0,
        DEPTHWRITE = 1 << 1,
        CULLFACE   = 1 << 2,
        DEPTHTEST  = 1 << 3,
        ALPHATEST  = 1 << 4
    };
    typedef u32 STATE;
    void Enable(STATE state);   // (ALPHABLEND | DEPTHWRITE | CULLFACE | DEPTHTEST | ALPHATEST)
    void Disable(STATE state);  // (ALPHABLEND | DEPTHWRITE | CULLFACE | DEPTHTEST | ALPHATEST)
    STATE GetEnable();
    STATE GetDisable();

private:
    Draw3DHelper();
    ~Draw3DHelper();
    void CreateCommand_();

private:
    static Draw3DHelper*        m_pInstance;
    u8*                         m_ShaderBinary;
    nn::gr::Shader              m_Shader;
    nn::gr::BindSymbolVSInput   m_SymbolPos;
    nn::gr::BindSymbolVSInput   m_SymbolTex;
    nn::gr::BindSymbolVSInput   m_SymbolClr;
    nn::gr::BindSymbolVSFloat   m_SymbolWorldMtx;
    nn::gr::BindSymbolVSFloat   m_SymbolViewMtx;
    nn::gr::BindSymbolVSFloat   m_SymbolProjMtx;
    u32*                        m_Command;
    const u32*                  m_StartAddr;

    nn::gr::Viewport            m_Viewport;
    nn::gr::Scissor             m_Scissor;
    nn::gr::CTR::RenderState    m_RenderState;
};

namespace Draw3DUtil
{
    // Sets the number of primitive segments for shape rendering
    void SetupSphere( u32 COL_NUM, u32 ROW_NUM );
    void SetupDisc( u32 div );
    void SetupCylinder( u32 div );

    void Finalize();

    // Shape generating function
    bool CreateSphere   ( XYZTC* model, const nw::math::VEC3& radius, nw::ut::Color8 color, u32 COL_NUM=8, u32 ROW_NUM=8 );
    bool CreatCube      ( XYZTC* model, const nw::math::VEC3& radius, nw::ut::Color8 color );
    bool CreateDisc     ( XYZTC* model, const nw::math::VEC3& radius, nw::ut::Color8 color, u32 div=10 );
    bool CreateCylinder ( XYZTC* model, const nw::math::VEC3& radius, nw::ut::Color8 color, u32 div=10 );
    bool CreateRectangle( XYZTC* model, const nw::math::VEC3& radius, nw::ut::Color8 color );
    bool CreateCone     ( XYZTC* model, const nw::math::VEC3& radius, nw::ut::Color8 color, u32 div=10 );

    // Shape rendering function
    void DrawSphere   ( const nw::math::MTX34& world, const nw::math::VEC3& radius, const nw::ut::FloatColor& color );
    void DrawCube     ( const nw::math::MTX34& world, const nw::math::VEC3& radius, const nw::ut::FloatColor& color );
    void DrawDisc     ( const nw::math::MTX34& world, const nw::math::VEC3& radius, const nw::ut::FloatColor& color );
    void DrawCylinder ( const nw::math::MTX34& world, const nw::math::VEC3& radius, const nw::ut::FloatColor& color );
    void DrawRectangle( const nw::math::MTX34& world, const nw::math::VEC3& radius, const nw::ut::FloatColor& color );
    void DrawCone     ( const nw::math::MTX34& world, const nw::math::VEC3& radius, const nw::ut::FloatColor& color );
}


} // namespace system
} // namespace vw

#endif // NW4CVIEWER_DRAW3DHELPER_H_
