/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*

	
	
	
  


  *---------------------------------------------------------------------------*/

#ifndef NW4C_VIEWER_SYSTEM_DRAWFAST_H_
#define NW4C_VIEWER_SYSTEM_DRAWFAST_H_

namespace vw     {
namespace system {

//======================================================================//
//							Text display
//======================================================================//
//Text color
void SetTextColor( nw::ut::Color8 color=nw::ut::Color8::WHITE ); //Reset to WHITE when Begin().
nw::ut::Color8 GetTextColor();

//Font size
#define DRAWFAST_DEFAULT_FONT_SIZE	(12.0f)
#define DRAWFAST_DEFAULT_FONT_DX	(9.5f)
#define DRAWFAST_DEFAULT_FONT_DY	(14.0f)
#define DRAWFAST_FONT_SIZE_TO_DX_RATIO	(DRAWFAST_DEFAULT_FONT_DX / DRAWFAST_DEFAULT_FONT_SIZE)
#define DRAWFAST_FONT_SIZE_TO_DY_RATIO	(DRAWFAST_DEFAULT_FONT_DY / DRAWFAST_DEFAULT_FONT_SIZE)
#define DRAWFAST_FONT_DX_TO_SIZE_RATIO	(DRAWFAST_DEFAULT_FONT_SIZE / DRAWFAST_DEFAULT_FONT_DX)
#define DRAWFAST_FONT_DY_TO_SIZE_RATIO	(DRAWFAST_DEFAULT_FONT_SIZE / DRAWFAST_DEFAULT_FONT_DY)

void SetFontSize( f32 fontSize=12.0f ); //Reset to 12.0f when Begin().
f32 GetFontSize();

inline f32 FontSizeToDx( f32 fontSize ) { return fontSize 	* DRAWFAST_FONT_SIZE_TO_DX_RATIO; }
inline f32 FontDxToSize( f32 dx ) 		{ return dx 		* DRAWFAST_FONT_DX_TO_SIZE_RATIO; }
inline f32 FontSizeToDy( f32 fontSize ) //Carriage return with margins (if no margins, same as FontSize)
										{ return fontSize 	* DRAWFAST_FONT_SIZE_TO_DY_RATIO; }
inline f32 FontDyToSize( f32 dy ) 		{ return dy 		* DRAWFAST_FONT_DY_TO_SIZE_RATIO; }
inline nw::math::VEC2 FontSizeToPixelSize( f32 fontSize )
		{ return nw::math::VEC2( FontSizeToDx(fontSize), FontSizeToDy(fontSize) ); }
inline f32 FontDxDyToSize( f32 dx, f32 dy ) { return nw::ut::Min( FontDxToSize(dx), FontDyToSize(dy) ); }
inline f32 FontPixelSizeToSize( const nw::math::VEC2& pixelSize )
										{ return FontDxDyToSize( pixelSize.x, pixelSize.y ); }
//Text display start position
void SetTextPos( f32 x=0.0f, f32 y=0.0f ); //Reset to (0.0f, 0.0f) when Begin().
inline void SetTextPos( const nw::math::VEC2 & pos ) { SetTextPos( pos.x, pos.y ); }
nw::math::VEC2 GetTextPos();


//◇ Render Function (If not between Begin() - End(), an assert failure occurs.)
//- Supports only ASCII characters.
void Printf( const char *format, ... );
void VPrintf( const char *format, std::va_list vlist );
void Print( const char *str ); //If only strings, this is 10-20% faster.

//Get display size
nw::math::VEC2 StrToCharSize( const char* str ); //String -> display character size

inline nw::math::VEC2 CharSizeToPixelSize( //Display character size -> display pixel size
		const nw::math::VEC2& charSize, f32 fontSize )
{
	return nw::math::VEC2( charSize.x * FontSizeToDx( fontSize ),
						   charSize.y * FontSizeToDy( fontSize ) );
}

inline nw::math::VEC2 StrToPixelSize( const char* str, f32 fontSize ) //String -> display pixel size
{
	return CharSizeToPixelSize( StrToCharSize( str ), fontSize );
}

//======================================================================//
//							  Shapes display
//======================================================================//
//◇ Property (Can be called even outside Begin() - End().

//Line width (If less than 2.0, the width may appear to change depending on the coordinates.)
void SetLineWidth( f32 lineWidth=2.0f );
f32  GetLineWidth();

//◇ Render Function (If not between Begin() - End(), an assert failure occurs.)

//Straight line
void DrawLine( f32 x1, f32 y1, f32 x2, f32 y2, const nw::ut::Color8 & color );
void DrawLine( const nw::math::VEC2 & p1, const nw::math::VEC2 & p2, const nw::ut::Color8 & color );

//Square line
void DrawRectangle( f32 posh, f32 posv, f32 sizeh, f32 sizev, const nw::ut::Color8 & color );
void DrawRectangle( const nw::math::VEC2& pos, const nw::math::VEC2& size, const nw::ut::Color8 & color );

//Square
void FillRectangle( f32 posh, f32 posv, f32 sizeh, f32 sizev, const nw::ut::Color8 & color );
void FillRectangle( const nw::math::VEC2& pos, const nw::math::VEC2& size, const nw::ut::Color8 & color );

//Triangle
void FillTriangle( const nw::math::VEC2& pt1, const nw::math::VEC2& pt2, const nw::math::VEC2& pt3, const nw::ut::Color8 & color );

//======================================================================//
//						Start/stop rendering
//======================================================================//
//Called within demolib, so unnecessary to call.
void Begin();
void End();

//======================================================================//
//						Encapsulates all parameters.
//======================================================================//
struct DrawAllParam {
	nw::ut::Color8 	textColor;
	f32 			fontSize;
	nw::math::VEC2	textPos;
	f32 			lineWidth;
	DrawAllParam()
	: textColor( nw::ut::Color8::WHITE ),
	  fontSize( 12.0f ),
	  textPos( nw::math::VEC2(0.0f, 0.0f) ),
	  lineWidth( 2.0f )
	{}
	void SetByCurrentParam() //Set the current parameters in this structure
	{
		textColor = GetTextColor();
		fontSize  = GetFontSize();
		textPos	  = GetTextPos();
		lineWidth = GetLineWidth();
	}
	void SetToCurrentParam() //Reflect the content of this structure in the parameters
	{
		SetTextColor( textColor );
		SetFontSize( fontSize );
		SetTextPos( textPos.x, textPos.y );
		SetLineWidth( lineWidth );
	}
	//This method is easier to understand when evacuating/restoring.
	void Push() { SetByCurrentParam(); }
	void Pop()  { SetToCurrentParam(); }
};

//Render all parameters, render auto evacuate & restore
class AutoDrawAllParamBak {
public:
  	AutoDrawAllParamBak()  { m_Bak.Push(); }
  	~AutoDrawAllParamBak() { m_Bak.Pop();  }
private:
  	DrawAllParam m_Bak;
};


} // namespace vw::sys
} // namespace vw

#endif // NW4C_VIEWER_SYSTEM_DRAWFAST_H_
