/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#ifndef NW4R_VIEWER_SYSTEM_MEMORY_H_
#define NW4R_VIEWER_SYSTEM_MEMORY_H_


#include <viewer/system/AppId.h>

namespace vw     {
namespace system {

namespace Memory {

    class Allocator : public nw::os::IAllocator
    {
    public:
        virtual void* Alloc(size_t size, u8 alignment) ;
        virtual void Free(void* memory);
        void Dump();
    };
}

const s32 MEMORY_DEFAULT_ALIGN = 4;
const u16 MEMORY_DEFAULT_ID = MODULE_NUM + 1;

// Initialize heap.
//
//                             For 0, set the default size (32 MB).
void  Memory_Init( size_t mainMemorySize = 0 );

void  Memory_Init( nw::os::IAllocator* pAllocator, nw::os::IAllocator* pDeviceAllocator );

// Perform finalization for heap.
void  Memory_Finalize();

// Allocates memory of the specified size.
void* Memory_Alloc( u16 id, u32 size, u32 align = MEMORY_DEFAULT_ALIGN );

// Allocates device memory of the specified size.
void* Memory_AllocDevice( u16 id, u32 size, u32 align = MEMORY_DEFAULT_ALIGN );

// Returns memory obtained from the heap.
void  Memory_Free( void* addr );
void  Memory_Free( const void* addr );

// Checks the amount of heap memory currently available.
u32   Memory_FreeSize( void );

// Dump and output the heap state to the console.
void  Memory_Dump();

// Gets a pointer to the allocator.
nw::os::IAllocator*   Memory_GetAllocator( u16 id );

// Gets a pointer to the allocator for device memory.
nw::os::IAllocator*   Memory_GetAllocatorDevice( u16 id );



/*---------------------------------------------------------------------------*
  Name:         Memory_AllocSys
  Description:  Allocates memory with SYSTEM_ID attributes.
  Arguments:    size  Allocation size
                align Required alignment
  Returns:      Returns a pointer to the allocated region.
 *---------------------------------------------------------------------------*/
inline void*
Memory_AllocSys( u32 size, u32 align = MEMORY_DEFAULT_ALIGN ) { return Memory_Alloc( SYSTEM_ID, size, align ); }
inline void*
Memory_AllocSysDevice( u32 size, u32 align = MEMORY_DEFAULT_ALIGN ) { return Memory_AllocDevice( SYSTEM_ID, size, align ); }

inline nw::os::IAllocator*
Memory_GetSysAllocator() { return Memory_GetAllocator(SYSTEM_ID ); }
inline nw::os::IAllocator*
Memory_GetSysAllocatorDevice() { return Memory_GetAllocatorDevice(SYSTEM_ID ); }


}   /* namespace system */


namespace g3d {

inline void*
Memory_Alloc( u32 size, u32 align = system::MEMORY_DEFAULT_ALIGN ) { return vw::system::Memory_Alloc( system::G3D_ID, size, align ); }

inline void*
Memory_AllocDevice( u32 size, u32 align = system::MEMORY_DEFAULT_ALIGN ) { return vw::system::Memory_AllocDevice( system::G3D_ID, size, align ); }

inline void
Memory_Free( void* addr ) { vw::system::Memory_Free( addr ); }
inline void
Memory_Free( const void* addr ) { vw::system::Memory_Free( addr ); }

inline nw::os::IAllocator*
Memory_GetAllocator()     { return vw::system::Memory_GetAllocator( system::G3D_ID ); }

inline nw::os::IAllocator*
Memory_GetAllocatorDevice() { return vw::system::Memory_GetAllocatorDevice( system::G3D_ID ); }

} /* namespace g3d */



namespace effect {

inline void*
Memory_Alloc( u32 size, u32 align = system::MEMORY_DEFAULT_ALIGN ) { return vw::system::Memory_Alloc( system::EFFECT_ID, size, align ); }

inline void*
Memory_AllocDevice( u32 size, u32 align = system::MEMORY_DEFAULT_ALIGN ) { return vw::system::Memory_AllocDevice( system::EFFECT_ID, size, align ); }

inline void
Memory_Free( void* addr ) { vw::system::Memory_Free( addr ); }
inline void
Memory_Free( const void* addr ) { vw::system::Memory_Free( addr ); }

inline nw::os::IAllocator*
Memory_GetAllocator()     { return vw::system::Memory_GetAllocator( system::EFFECT_ID ); }

inline nw::os::IAllocator*
Memory_GetAllocatorDevice() { return vw::system::Memory_GetAllocatorDevice( system::EFFECT_ID ); }

} /* namespace effect */



namespace g2d {

inline void*
Memory_Alloc( u32 size, u32 align = system::MEMORY_DEFAULT_ALIGN ) { return vw::system::Memory_Alloc( system::G2D_ID, size, align ); }

inline void*
Memory_AllocDevice( u32 size, u32 align = system::MEMORY_DEFAULT_ALIGN ) { return vw::system::Memory_AllocDevice( system::G2D_ID, size, align ); }

inline void
Memory_Free( void* addr ) { vw::system::Memory_Free( addr ); }
inline void
Memory_Free( const void* addr ) { vw::system::Memory_Free( addr ); }

inline nw::os::IAllocator*
Memory_GetAllocator()     { return vw::system::Memory_GetAllocator( system::G2D_ID ); }

inline nw::os::IAllocator*
Memory_GetAllocatorDevice() { return vw::system::Memory_GetAllocatorDevice( system::G2D_ID ); }

} /* namespace g2d */



namespace sound {

inline void*
Memory_Alloc( u32 size, u32 align = system::MEMORY_DEFAULT_ALIGN ) { return vw::system::Memory_Alloc( system::SOUND_ID, size, align ); }

inline void*
Memory_AllocDevice( u32 size, u32 align = system::MEMORY_DEFAULT_ALIGN ) { return vw::system::Memory_AllocDevice( system::SOUND_ID, size, align ); }

inline void
Memory_Free( void* addr ) { vw::system::Memory_Free( addr ); }
inline void
Memory_Free( const void* addr ) { vw::system::Memory_Free( addr ); }

inline nw::os::IAllocator*
Memory_GetAllocator()     { return vw::system::Memory_GetAllocator( system::SOUND_ID ); }

inline nw::os::IAllocator*
Memory_GetAllocatorDevice() { return vw::system::Memory_GetAllocatorDevice( system::SOUND_ID ); }

} /* namespace sound */



#if defined(USE_SAMPLE)

namespace sample {

inline void*
Memory_Alloc( u32 size, u32 align = system::MEMORY_DEFAULT_ALIGN ) { return vw::system::Memory_Alloc( system::SAMPLE_ID, size, align ); }

inline void*
Memory_AllocDevice( u32 size, u32 align = system::MEMORY_DEFAULT_ALIGN ) { return vw::system::Memory_AllocDevice( system::SAMPLE_ID, size, align ); }

inline void
Memory_Free( void* addr ) { vw::system::Memory_Free( addr ); }
inline void
Memory_Free( const void* addr ) { vw::system::Memory_Free( addr ); }

inline nw::os::IAllocator*
Memory_GetAllocator()     { return vw::system::Memory_GetAllocator( system::SAMPLE_ID ); }

inline nw::os::IAllocator*
Memory_GetAllocatorDevice() { return vw::system::Memory_GetAllocatorDevice( system::SAMPLE_ID ); }

} /* namespace sample */

#endif

}   /* namespace vw     */



/* NW4R_VIEWER_SYSTEM_MEMORY_H_ */
#endif
