/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#ifndef NW4C_VIEWER_SYSTEM_PAD_H_
#define NW4C_VIEWER_SYSTEM_PAD_H_

#include <viewer/system/AppId.h>

#if NN_CURRENT_VERSION_NUMBER >= NN_VERSION_NUMBER(4,0,0,0)
#pragma diag_suppress 2530 // padding added to end of struct.
#endif
namespace vw     {
namespace system {

#define PAD_BUTTON_LEFT         nn::hid::CTR::BUTTON_LEFT
#define PAD_BUTTON_RIGHT        nn::hid::CTR::BUTTON_RIGHT
#define PAD_BUTTON_DOWN         nn::hid::CTR::BUTTON_DOWN
#define PAD_BUTTON_UP           nn::hid::CTR::BUTTON_UP
#define PAD_TRIGGER_R           nn::hid::CTR::BUTTON_R
#define PAD_TRIGGER_L           nn::hid::CTR::BUTTON_L
#define PAD_BUTTON_A            nn::hid::CTR::BUTTON_A
#define PAD_BUTTON_B            nn::hid::CTR::BUTTON_B
#define PAD_BUTTON_X            nn::hid::CTR::BUTTON_X
#define PAD_BUTTON_Y            nn::hid::CTR::BUTTON_Y
#define PAD_BUTTON_SELECT       nn::hid::CTR::PSEUDO_BUTTON_SELECT
#define PAD_BUTTON_START        nn::hid::CTR::BUTTON_START

#define PAD_STICK_LEFT          nn::hid::CTR::BUTTON_EMULATION_LEFT
#define PAD_STICK_RIGHT         nn::hid::CTR::BUTTON_EMULATION_RIGHT
#define PAD_STICK_DOWN          nn::hid::CTR::BUTTON_EMULATION_DOWN
#define PAD_STICK_UP            nn::hid::CTR::BUTTON_EMULATION_UP

struct PADInfo
{
    u32     button;      // Or-ed PAD_BUTTON_* bits
    f32     stickX;      // -1.0f <= stickX       <= 1.0f
    f32     stickY;      // -1.0f <= stickY       <= 1.0f
    u32     press;       // Input value
    u32     trig;        // Push trigger input
    u32     release;     // Release trigger input
    u32     repeat;      // Drag repeat input
    u32     repeatAccel; // accelerated Repeat input
    u32     disabled;    // Flag of whether invalid
};

void    Pad_Init( void );

void    Pad_Read( void );

void    Pad_SetForceRightAngle( bool force );

bool    Pad_GetForceRightAngle();

void    Pad_GetKeyInfo   ( u16 appId, PADInfo* keyInfo );

u32     Pad_IsPressed           ( u16 appId, u32 button );

u32     Pad_IsPressedHold       ( u16 appId, u32 button );

u32     Pad_IsTriggered         ( u16 appId, u32 button );

u32     Pad_IsReleased          ( u16 appId, u32 button );

u32     Pad_IsRepeated          ( u16 appId, u32 button );

u32     Pad_IsRepeatedAccel     ( u16 appId, u32 button );

u32     Pad_IsClicked           ( u16 appId, u32 button );

u32     Pad_IsLongPressed       ( u16 appId, u32 button );

void    Pad_GetStickXY          ( u16 appId, f32* stickX, f32* stickY );

bool    Pad_IsEnable            ( u16 appId );

void    Pad_ClearState          ( u16 appId );


bool    Pad_GetRawPADInfo_( nn::hid::CTR::PadStatus* pad );

bool    Pad_GetRawDEBUGPADInfo_( nn::hid::CTR::DebugPadStatus* dpad );

enum {
    PAD_BUTTONMASK_STICK    = 0x00100000 // Masking flag for the Control Stick
};

void    Pad_SetButtonMask( u32 mask );


inline bool
Pad_IsTriggerToMenu( u16 appId )
{
    return bool( (Pad_IsTriggered( appId, PAD_BUTTON_START ) != 0) &&
                 (! Pad_IsTriggered( appId, PAD_TRIGGER_L) )                    &&
                 (! Pad_IsTriggered( appId, PAD_TRIGGER_R) )                    &&
                 (! Pad_IsPressed( appId, u16(~(PAD_BUTTON_START | PAD_STICK_LEFT | PAD_STICK_RIGHT | PAD_STICK_DOWN | PAD_STICK_UP) ) ) ) );
}


/*---------------------------------------------------------------------------*
  Name:         Pad
  Description:  Pad-related functions wrapper class
 *---------------------------------------------------------------------------*/
class Pad
{
public:
    /* ctor */  explicit Pad ( u16 appId  ) : mAppId( appId ) {}
    virtual /* dtor */ ~Pad() {}
    u32         IsPressed      ( u32 button ) const { return Pad_IsPressed( mAppId, button );   }
    u32         IsPressedHold  ( u32 button ) const { return Pad_IsPressedHold( mAppId, button ); }
    u32         IsTriggered    ( u32 button ) const { return Pad_IsTriggered( mAppId, button ); }
    u32         IsReleased     ( u32 button ) const { return Pad_IsReleased( mAppId, button );  }
    u32         IsRepeated     ( u32 button ) const { return Pad_IsRepeated( mAppId, button );  }
    u32         IsRepeatedAccel( u32 button ) const { return Pad_IsRepeatedAccel( mAppId, button ); }
    u32         IsClicked      ( u32 button ) const { return Pad_IsClicked( mAppId, button );   }
    u32         IsLongPressed  ( u32 button ) const { return Pad_IsLongPressed( mAppId, button );      }

    void        GetStickXY     ( f32* stickX, f32* stickY ) const { Pad_GetStickXY( mAppId, stickX, stickY );    }
    void        GetKeyInfo     ( PADInfo* keyInfo )         const { Pad_GetKeyInfo( mAppId, keyInfo );           }

    bool        IsTriggerToMenu( void ) const { return Pad_IsTriggerToMenu( mAppId ); }
    bool        IsEnable       ( void ) const { return Pad_IsEnable( mAppId ); }
private:
    u16         mAppId;     // Module ID
};


}   /* namespace system */

namespace g3d {

class Pad : public system::Pad
{
    /* ctor */ Pad() : system::Pad( G3D_ID ) {}
    static Pad      GetInstance() { return Pad(); }
};

inline void Pad_GetKeyInfo   ( system::PADInfo* keyInfo ) { system::Pad_GetKeyInfo( G3D_ID, keyInfo ); }
inline u32  Pad_IsPressed    ( u32 button ) { return system::Pad_IsPressed( G3D_ID, button ); }
inline u32  Pad_IsTriggered  ( u32 button ) { return system::Pad_IsTriggered( G3D_ID, button ); }
inline u32  Pad_IsReleased   ( u32 button ) { return system::Pad_IsReleased( G3D_ID, button ); }
inline u32  Pad_IsRepeated   ( u32 button ) { return system::Pad_IsRepeated( G3D_ID, button ); }
inline u32  Pad_IsRepeatedAccel( u32 button ) { return system::Pad_IsRepeatedAccel( G3D_ID, button ); }
inline u32  Pad_IsClicked    ( u32 button ) { return system::Pad_IsClicked( G3D_ID, button ); }
inline u32  Pad_IsPressedHold( u32 button ) { return system::Pad_IsPressedHold( G3D_ID, button );    }
inline void Pad_GetStickXY   ( f32* stickX, f32* stickY ) { system::Pad_GetStickXY( G3D_ID, stickX, stickY ); }
inline bool Pad_IsTriggerToMenu( void ) { return system::Pad_IsTriggerToMenu( G3D_ID ); }
inline bool Pad_IsEnable     ( void ) { return system::Pad_IsEnable( G3D_ID ); }
inline u32  Pad_IsLongPressed( u32 button ) { return system::Pad_IsLongPressed( G3D_ID, button ); }

} /* namespace g3d */

namespace effect {

class Pad : public system::Pad
{
    /* ctor */ Pad() : system::Pad( EFFECT_ID ) {}
    static Pad      GetInstance() { return Pad(); }
};

inline void Pad_GetKeyInfo   ( system::PADInfo* keyInfo ) { system::Pad_GetKeyInfo( EFFECT_ID, keyInfo ); }
inline u32  Pad_IsPressed    ( u32 button ) { return system::Pad_IsPressed( EFFECT_ID, button ); }
inline u32  Pad_IsTriggered  ( u32 button ) { return system::Pad_IsTriggered( EFFECT_ID, button ); }
inline u32  Pad_IsReleased   ( u32 button ) { return system::Pad_IsReleased( EFFECT_ID, button ); }
inline u32  Pad_IsRepeated   ( u32 button ) { return system::Pad_IsRepeated( EFFECT_ID, button ); }
inline u32  Pad_IsRepeatedAccel( u32 button ) { return system::Pad_IsRepeatedAccel( EFFECT_ID, button ); }
inline u32  Pad_IsClicked    ( u32 button ) { return system::Pad_IsClicked( EFFECT_ID, button ); }
inline u32  Pad_IsPressedHold( u32 button ) { return system::Pad_IsPressedHold( EFFECT_ID, button );    }
inline void Pad_GetStickXY   ( f32* stickX, f32* stickY ) { system::Pad_GetStickXY( EFFECT_ID, stickX, stickY ); }
inline bool Pad_IsTriggerToMenu( void ) { return system::Pad_IsTriggerToMenu( EFFECT_ID ); }
inline bool Pad_IsEnable     ( void ) { return system::Pad_IsEnable( EFFECT_ID ); }
inline u32  Pad_IsLongPressed( u32 button ) { return system::Pad_IsLongPressed( EFFECT_ID, button ); }

} /* namespace effect */

namespace g2d {

class Pad : public system::Pad
{
    /* ctor */ Pad() : system::Pad( G2D_ID ) {}
    static Pad      GetInstance() { return Pad(); }
};

inline void Pad_GetKeyInfo   ( system::PADInfo* keyInfo ) { system::Pad_GetKeyInfo( G2D_ID, keyInfo ); }
inline u32  Pad_IsPressed    ( u32 button ) { return system::Pad_IsPressed( G2D_ID, button ); }
inline u32  Pad_IsTriggered  ( u32 button ) { return system::Pad_IsTriggered( G2D_ID, button ); }
inline u32  Pad_IsReleased   ( u32 button ) { return system::Pad_IsReleased( G2D_ID, button ); }
inline u32  Pad_IsRepeated   ( u32 button ) { return system::Pad_IsRepeated( G2D_ID, button ); }
inline u32  Pad_IsRepeatedAccel( u32 button ) { return system::Pad_IsRepeatedAccel( G2D_ID, button ); }
inline u32  Pad_IsClicked    ( u32 button ) { return system::Pad_IsClicked( G2D_ID, button ); }
inline u32  Pad_IsPressedHold( u32 button ) { return system::Pad_IsPressedHold( G2D_ID, button );    }
inline void Pad_GetStickXY   ( f32* stickX, f32* stickY ) { system::Pad_GetStickXY( G2D_ID, stickX, stickY ); }
inline bool Pad_IsTriggerToMenu( void ) { return system::Pad_IsTriggerToMenu( G2D_ID ); }
inline bool Pad_IsEnable     ( void ) { return system::Pad_IsEnable( G2D_ID ); }
inline u32  Pad_IsLongPressed( u32 button ) { return system::Pad_IsLongPressed( G2D_ID, button ); }

} /* namespace g2d */

namespace sound {

class Pad : public system::Pad
{
    /* ctor */ Pad() : system::Pad( SOUND_ID ) {}
    static Pad      GetInstance() { return Pad(); }
};

inline void Pad_GetKeyInfo   ( system::PADInfo* keyInfo ) { system::Pad_GetKeyInfo( SOUND_ID, keyInfo ); }
inline u32  Pad_IsPressed    ( u32 button ) { return system::Pad_IsPressed( SOUND_ID, button ); }
inline u32  Pad_IsTriggered  ( u32 button ) { return system::Pad_IsTriggered( SOUND_ID, button ); }
inline u32  Pad_IsReleased   ( u32 button ) { return system::Pad_IsReleased( SOUND_ID, button ); }
inline u32  Pad_IsRepeated   ( u32 button ) { return system::Pad_IsRepeated( SOUND_ID, button ); }
inline u32  Pad_IsRepeatedAccel( u32 button ) { return system::Pad_IsRepeatedAccel( SOUND_ID, button ); }
inline u32  Pad_IsClicked    ( u32 button ) { return system::Pad_IsClicked( SOUND_ID, button ); }
inline u32  Pad_IsPressedHold( u32 button ) { return system::Pad_IsPressedHold( SOUND_ID, button ); }
inline void Pad_GetStickXY   ( f32* stickX, f32* stickY ) { system::Pad_GetStickXY( SOUND_ID, stickX, stickY ); }
inline bool Pad_IsTriggerToMenu( void ) { return system::Pad_IsTriggerToMenu( SOUND_ID ); }
inline bool Pad_IsEnable     ( void ) { return system::Pad_IsEnable( SOUND_ID ); }
inline u32  Pad_IsLongPressed( u32 button ) { return system::Pad_IsLongPressed( SOUND_ID, button ); }

} /* namespace sound */



}   /* namespace vw     */
#if NN_CURRENT_VERSION_NUMBER >= NN_VERSION_NUMBER(4,0,0,0)
#pragma diag_default 2530 // padding added to end of struct.
#endif



/* NWCR_VIEWER_SYSTEM_PAD_H_ */
#endif
