/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#ifndef NW4R_VIEWER_SYSTEM_PERFORM_H_
#define NW4R_VIEWER_SYSTEM_PERFORM_H_

#include <nn/os.h>

#if NN_CURRENT_VERSION_NUMBER >= NN_VERSION_NUMBER(4,0,0,0)
#pragma diag_suppress 2530 // padding added to end of struct.
#endif
namespace vw     {
namespace system {

/*===========================================================================*
  Name:         CpuTime

  Description:  CPU time class
 *===========================================================================*/
class CpuTime
{
public:
    CpuTime() : mStartTick(0), mCumulativeTime(0.0f){}

    // Start measuring!
    void Begin()
    {
        mStartTick = nn::os::Tick::GetSystemCurrent();
    }

    // End measurement!
    void End()
    {
        const f32 NANO_TO_MILLI = 1.0f / (1000.0f * 1000.0f);
        nn::os::Tick endTick = nn::os::Tick::GetSystemCurrent();
        s64 span = static_cast<nn::os::Tick>(endTick - mStartTick).ToTimeSpan().GetNanoSeconds();
        mCumulativeTime += static_cast<f32>(span) * NANO_TO_MILLI;
        mStartTick = 0;
    }

    // Reset
    void Reset()
    {
        mCumulativeTime = 0.0f;
    }

    // Get accumulated time
    f32 GetCumulativeTime() const
    {
        return mCumulativeTime;
    }

    // Addition
    void AddTime(f32 time)
    {
        mCumulativeTime += time;
    }

private:
    s64 mStartTick;      // !< Start time
    f32 mCumulativeTime; // !< Accumulated time
};

/*===========================================================================*
  Name:         PerformCounter

  Description:  Performance management
 *===========================================================================*/
class PerformCounter
{
public:
    PerformCounter() : mCumulativeTime(0.0f),mFrameMaxTime(0.0f),mAvgTime(0.0f),mMaxTime(0.0f),mFrameTime(0.0f),mFrameCount(0){}

    // Add measurement time
    void AddTime(f32 time)
    {
        mFrameTime = time;
        mCumulativeTime += time;

        // Maximum value update
        if(mFrameMaxTime < time)
        {
            mFrameMaxTime = time;
        }
        mFrameCount++;
    }

    // Average time calculation
    void CalcTime()
    {
        mAvgTime = mCumulativeTime / static_cast<f32>(mFrameCount);
        mMaxTime = mFrameMaxTime;
        mCumulativeTime = 0.0f;
        mFrameMaxTime = 0.0f;
        mFrameCount = 0;
    }

    // Get average time
    f32 GetAvgTime() const
    {
        return mAvgTime;
    }

    // Get maximum time
    f32 GetMaxTime() const
    {
        return mMaxTime;
    }

    // Get most recent frame time
    f32 GetFrameTime() const
    {
        return mFrameTime;
    }

private:
    f32 mCumulativeTime;
    f32 mFrameMaxTime;
    f32 mAvgTime;
    f32 mMaxTime;
    f32 mFrameTime;
    u32 mFrameCount;
};

/*===========================================================================*
  Name:         G3dPerformManager

  Description:  G3D performance management
 *===========================================================================*/
class PerformMgr
{
public:
    // Gets the instance.
    static PerformMgr* GetInstance();

    // Initialization
    static void Initialize();

    // Finalize.
    static void Finalize();

    // Get GPU process time
    f32 GetGpuTime() const;

    // Get CPU process time
    f32 GetCpuTime() const;

    // Get Calc process time
    f32 GetCalcTime() const;

    // Get Draw process time
    f32 GetDrawTime() const;

    // Get GPU maximum processing time
    f32 GetMaxGpuTime() const { return mGpuCounter.GetMaxTime(); }

    // Get CPU maximum processing time
    f32 GetMaxCpuTime() const { return mCpuCounter.GetMaxTime(); }

    // Get Calc maximum processing time
    f32 GetMaxCalcTime() const { return mCalcCounter.GetMaxTime(); }

    // Get Draw maximum processing time
    f32 GetMaxDrawTime() const { return mDrawCounter.GetMaxTime(); }

    // Start Calc processing time measurement
    void StartCalcMeasure() { mCalc.Begin(); }

    // Stop Calc processing time measurement
    void StopCalcMeasure() { mCalc.End(); }

    // Start Draw processing time measurement
    void StartDrawMeasure() { mDraw.Begin(); }

    // Stop Draw processing time measurement
    void StopDrawMeasure() { mDraw.End(); }

    // Get CalcTime
    CpuTime& GetCalc() { return mCalc; }

    // Get DrawTime
    CpuTime& GetDraw() { return mDraw; }

    // Get UpdateTime
    CpuTime& GetUpdate() { return mUpdate; }

    // Get AllTime
    CpuTime& GetAllTimer() { return mAll; }

    // 1 frame measurement finalization
    void EndFrame();

    // Set calculation interval for measurement time
    void SetMeasureCount(u32 count);

    // Start 1 frame process time measurement
    void SetTimerStart();

    // End 1 frame process time measurement
    void SetTimerEnd();

    // Get 1 frame process time
    f32 GetSystemTime();

private:
    PerformMgr();
    ~PerformMgr();
    // Average time calculation
    void CalcTime();

private:
    static PerformMgr*  m_pInstance;
    CpuTime             mCalc;              //
    CpuTime             mDraw;              //
    CpuTime             mSystem;            //
    CpuTime             mAll;               //
    CpuTime             mUpdate;            //
    PerformCounter      mCalcCounter;       //
    PerformCounter      mDrawCounter;       //
    PerformCounter      mCpuCounter;        //
    PerformCounter      mGpuCounter;        //
    u32                 mCount;             //
    u32                 mMeasureCount;      //
    f32                 mSystemTime;        //
    f32                 mCalcTmp;
    f32                 mDrawTmp;
    f32                 mCpuTmp;
    f32                 mGpuTmp;
};

}   /* namespace system */
}   /* namespace vw     */
#if NN_CURRENT_VERSION_NUMBER >= NN_VERSION_NUMBER(4,0,0,0)
#pragma diag_default 2530 // padding added to end of struct.
#endif

/* NW4R_VIEWER_SYSTEM_PERFORM_H_ */
#endif
