/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

/*

  

  */

#ifndef NW4C_VIEWER_SYSTEM_WINSYS_H_
#define NW4C_VIEWER_SYSTEM_WINSYS_H_

#ifndef VW_USER_MULTIPATH_RENDERING
#define VW_USER_MULTIPATH_RENDERING
#endif

#if NN_CURRENT_VERSION_NUMBER >= NN_VERSION_NUMBER(4,0,0,0)
#pragma diag_suppress 1301 // padding inserted in struct.
#endif
namespace vw     {
namespace system {

class WinSys {
public:

	class Entry {
	public:
		virtual ~Entry() {}
		virtual const char* GetTitle() const = 0;
		virtual void SetPos( const nw::math::VEC2& pos ) = 0;
		virtual nw::math::VEC2 GetPos() const = 0;
		virtual void SetSize( const nw::math::VEC2& size ) { (void)size; }
		virtual nw::math::VEC2 GetSize() const = 0;
		virtual void SetOpenSize( const nw::math::VEC2& size ) = 0;
		virtual nw::math::VEC2 GetOpenSize() const = 0;
		virtual bool IsOpen() const = 0;
		virtual void SetIsOpen( bool isOpen ) = 0;
		virtual bool IsResizable() const = 0;
		virtual void Update( const nn::hid::TouchPanelStatus& tp ) = 0;
		virtual void Disp() = 0;

		void Resize( const nw::math::VEC2& size ) { SetSize( size ); }
		nw::ut::Rect GetRect() const
		{
			nw::math::VEC2 pos  = GetPos();
			nw::math::VEC2 size = GetSize();
			return nw::ut::Rect( pos.x, pos.y, pos.x + size.x, pos.y + size.y );
		}
	};

	WinSys()
	: m_EntryNum( 0 ),
	  m_IsMoving( false ),
	  m_pMovingWindow( NULL ),
	  m_ResizeStat( RESIZE_STAT_NONE ),
	  m_pResizingWindow( NULL )
	{
	}

	~WinSys() {}

	void EntryWindow( Entry* pEntry )
	{
		NW_POINTER_ASSERT( pEntry );

		if ( m_EntryNum >= ENTRY_MAX )
		{
			nwosPanic( "entry max over.\n" );
			return;
		}

		m_pEntries[ m_EntryNum ++ ] = pEntry;
	}

	void DeleteWindow( Entry* pEntry )
	{
		for( int i=0; i<m_EntryNum; ++i )
		{
			if ( m_pEntries[i] == pEntry )
			{
				for( int j=i; j<m_EntryNum-1; ++j )
				{
					m_pEntries[j] = m_pEntries[j+1];
				}
				-- m_EntryNum;
				return;
			}
		}
		nwosPanic( "Cannot find entry window.\n");
	}

	void Update( const nn::hid::TouchPanelStatus& tp_ )
	{
		using namespace nw::math;
		nn::hid::TouchPanelStatus tp = tp_;

		for( int i=0; i<m_EntryNum; ++i )
		{
			const nw::ut::Rect winRect = m_pEntries[i]->GetRect();
			const VEC2 tpPos = VEC2( tp.x, tp.y );
			const bool isTpTrg = (tp.touch && ! m_PrvTp.touch); //Start touch?

			if ( m_IsMoving && m_pEntries[i] == m_pMovingWindow )
			{ //Moving
				if ( tp.touch )
				{ //Continue moving
					NW_ASSERT( m_PrvTp.touch );
					const VEC2 moveDiffPos = tpPos - VEC2( m_PrvTp.x, m_PrvTp.y );
					const VEC2 srcPos = VEC2( winRect.left, winRect.top );
					const VEC2 dstPos = srcPos + moveDiffPos;
					m_pEntries[i]->SetPos( dstPos );
				}
				else
				{ //End moving
					m_IsMoving = false;
					m_pMovingWindow = NULL;
				}
			}
			else if ( m_ResizeStat != RESIZE_STAT_NONE && m_pEntries[i] == m_pResizingWindow )
			{ //Resizing
				if ( tp.touch )
				{ //Continue resizing
					DoResize( tpPos );
				}
				else
				{ //End resizing
					m_ResizeStat = RESIZE_STAT_NONE;
					m_pResizingWindow = NULL;
				}
			}
			else
			{ //Neither moving nor resizing.
				if ( isTpTrg ) //Touched this time?
				{
					m_ResizeStat = (m_pEntries[i]->IsResizable())
										? CheckResizePos( tpPos, winRect )
										: RESIZE_STAT_NONE;
					if ( m_ResizeStat != RESIZE_STAT_NONE ) //Start resizing?
					{
						m_pResizingWindow = m_pEntries[i];
					}
					else
					{
						nw::ut::Rect moveStatableRect = winRect;
						moveStatableRect.bottom = moveStatableRect.top + 12.0f;
						if ( IsPosInRect( tpPos, moveStatableRect ) ) //Start moving?
						{
							m_IsMoving = true;
							m_pMovingWindow = m_pEntries[i];
						}
					}
				}
			}
			//Changing movement or size?
			const bool isMoveOrResize =    m_pEntries[i] == m_pMovingWindow
										|| m_pEntries[i] == m_pResizingWindow;
			//Touching window?
			const bool isTouchWindow = (tp.touch && IsPosInRect( tpPos, winRect ));

			//Move window to top if changing movement or size or if the window is being touched
			if ( isMoveOrResize || isTouchWindow )
			{
				ToTop( i );
			}

			//Clear touch judgment if changing movement or size -> Cannot make touch bar-related determinations on the window side.
			//if ( isMoveOrResize )
			//{
			//	tp.touch = 0;
			//	tp.x = 0;
			//	tp.y = 0;
			//}

			//Call for each frame process on window side
			m_pEntries[i]->Update( tp );

			//If the window is touched, moved, or resized, the touch determination for the lower window is disabled
			if ( isMoveOrResize || isTouchWindow  )
			{
				tp.touch = 0;
				tp.x = 0;
				tp.y = 0;
			}

		} // of for
		m_PrvTp = tp_;
	}

	void Disp()
	{
		for( int i=m_EntryNum-1; i>=0; --i )
		{
			m_pEntries[i]->Disp();
		}
	}

	int GetEntryNum() const { return m_EntryNum; }

	Entry& GetEntry( int index )
	{
		NW_MINMAX_ASSERT( index, 0, GetEntryNum() - 1 );
		Entry* pEntry = m_pEntries[ index ];
		NW_POINTER_ASSERT( pEntry );
		return *pEntry;
	}

	void ToTop( int index )
	{
		NW_MINMAX_ASSERT( index, 0, GetEntryNum() - 1 );
		if ( index == 0 ) return; //Always on top
		Entry* const t = m_pEntries[index];
		for( int i=index; i>=1; --i )
		{
			m_pEntries[i] = m_pEntries[i-1];
		}
		m_pEntries[0] = t;
	}

private:
	static const int ENTRY_MAX = 64;
	enum ResizeStat {
		RESIZE_STAT_NONE 			= 0,
		RESIZE_STAT_LEFT			,
		RESIZE_STAT_RIGHT			,
		RESIZE_STAT_TOP				,
		RESIZE_STAT_LEFT_TOP 		,
		RESIZE_STAT_RIGHT_TOP		,
		RESIZE_STAT_BOTTOM			,
		RESIZE_STAT_LEFT_BOTTOM		,
		RESIZE_STAT_RIGHT_BOTTOM	,
		RESIZE_STAT_NUM
	};
	Entry* 						m_pEntries[ ENTRY_MAX ]; //[0] top-most window - [m_EntryNum-1] bottom-most window
	int 						m_EntryNum;
	nn::hid::TouchPanelStatus 	m_PrvTp;
	bool 						m_IsMoving;
	Entry* 						m_pMovingWindow;
	ResizeStat 					m_ResizeStat;
	Entry* 						m_pResizingWindow;

	bool IsPosInRect( const nw::math::VEC2& pos, const nw::ut::Rect& rect ) const
	{
		return rect.left <= pos.x && pos.x <= rect.right
			&& rect.top  <= pos.y && pos.y <= rect.bottom;
	}

	ResizeStat CheckResizePos( const nw::math::VEC2& pos, const nw::ut::Rect& rect ) const
	{
		const int m = 4.0f;
		const nw::ut::Rect   leftLineRect = nw::ut::Rect( rect.left  - m, rect.top       ,
													      rect.left  + m, rect.bottom     );
		const nw::ut::Rect  rightLineRect = nw::ut::Rect( rect.right - m, rect.top       ,
													      rect.right + m, rect.bottom     );
		const nw::ut::Rect    topLineRect = nw::ut::Rect( rect.left     , rect.top    - m - 4.0f,
													   	  rect.right    , rect.top    + m - 4.0f );
		const nw::ut::Rect bottomLineRect = nw::ut::Rect( rect.left     , rect.bottom - m,
													  	  rect.right    , rect.bottom + m );
		const bool isOnLeftLine   = IsPosInRect( pos,   leftLineRect );
		const bool isOnRightLine  = IsPosInRect( pos,  rightLineRect );
		const bool isOnTopLine    = IsPosInRect( pos,    topLineRect );
		const bool isOnBottomLine = IsPosInRect( pos, bottomLineRect );
		const int hStat = (isOnLeftLine ) ? 1 : ((isOnRightLine ) ? 2 : 0);
		const int vStat = (isOnTopLine  ) ? 1 : ((isOnBottomLine) ? 2 : 0);
		return static_cast<ResizeStat>( hStat + vStat * 3 );
	}

	void DoResize( const nw::math::VEC2& tpPos )
	{
		nw::ut::Rect dstRect = m_pResizingWindow->GetRect();
		const int hStat = m_ResizeStat % 3;
		const int vStat = m_ResizeStat / 3;
		if ( hStat == 1 ) dstRect.left   = tpPos.x;
		if ( hStat == 2 ) dstRect.right  = tpPos.x;
		if ( vStat == 1 ) dstRect.top    = tpPos.y;
		if ( vStat == 2 ) dstRect.bottom = tpPos.y;
		const nw::math::VEC2 dstPos1( dstRect.left , dstRect.top    );
		const nw::math::VEC2 dstPos2( dstRect.right, dstRect.bottom );
		m_pResizingWindow->SetPos( dstPos1 );
		m_pResizingWindow->Resize( dstPos2 - dstPos1 );
	}

};

//======================================================================//
//							Window that can be entered
//======================================================================//
#include "Window.h"
#include "Button.h"

class EntriableWindow : public WinSys::Entry {
public:
	EntriableWindow(
		const char* titleStr,					//Window title string
		f32 defaultFontSize,					//Font size initial value
		const nw::math::VEC2& defaultClientSize //Initial value for client region size that does not include margins
	)
	: m_DefaultFontSize( defaultFontSize ),
	  m_DefaultClientSize( defaultClientSize ),
	  m_ItemNum( 1 ),
	  m_IsShouldOpen( true ),
	  m_OpenRate( 1.0f ),
	  m_PrevClickFrame( 10000 ), //The default value is large enough
	  m_PrevTouch( 0 ),
	  m_PosSizeLogOn( false ) //If this is set to TRUE, it is convenient when adjusting the position and size of all windows
	  						  //Maybe easy to display the WinSysSave content
	{
		m_Window.titleStr = titleStr;
		iSetPos( m_Window.pos ); //For calculating client coordinates with margins
		SetPos( m_Window.pos );
		m_Window.SetClientSize( m_DefaultClientSize );
		Resize( m_Window.size ); //For calculating client coordinates with margins and font sizes
		m_CloseSize = m_Window.GetMinimumSize();
	}
	virtual ~EntriableWindow() {}
	void ToClose( bool isRapid )
	{
		m_IsShouldOpen = false;
		if ( isRapid )
		{
			m_OpenRate = 0.0f;
			Resize( m_CloseSize );
		}
	}
	//================ Inherit from WinSys::Entry ================//
	virtual const char* GetTitle() const { return m_Window.titleStr; }
	virtual void SetPos( const nw::math::VEC2& pos )
	{
		m_TargetPos = pos;
	}
	virtual nw::math::VEC2 GetPos() const { return m_TargetPos; }
	virtual void SetSize( const nw::math::VEC2& size )
	{
		m_Window.Resize( size );
		if ( IsResizable() )
		{
			m_TargetOpenSize = m_OpenSize = m_Window.size; //Save as the open size only when the user adjusts the size
		}
		if ( m_PosSizeLogOn )
		{
			NW_LOG( "%s: size=(%d, %d)\n", m_Window.titleStr, (int)size.x, (int)size.y  );
		}
		m_CSize = m_Window.GetClientSize();
		m_FontSize = nw::ut::Min( m_DefaultFontSize * m_CSize.x / m_DefaultClientSize.x,
								  m_DefaultFontSize * m_CSize.y / m_DefaultClientSize.y  );
		m_CMSize = m_CSize - GetMarginVec2() * 2;
	}
	virtual nw::math::VEC2 GetSize() const { return m_Window.size; }
	virtual void SetOpenSize( const nw::math::VEC2& size )
	{
		m_TargetOpenSize = size;
	}
	virtual nw::math::VEC2 GetOpenSize() const { return m_TargetOpenSize; }
	virtual bool IsOpen() const { return m_IsShouldOpen; }
	virtual void SetIsOpen( bool isOpen ) { m_IsShouldOpen = isOpen; }
	virtual bool IsResizable() 		const 	{ return m_IsShouldOpen && m_OpenRate == 1.0f; }
	virtual void Update( const nn::hid::TouchPanelStatus& tp )
	{
		//So that movement is smooth when changing placement
		if ( m_TargetPos != m_Window.pos )
		{
			const nw::math::VEC2 diffPos = m_TargetPos - m_Window.pos;
			const nw::math::VEC2 nextPos = (( diffPos.LengthSquare() < 1.0f )
												? m_TargetPos
												: (m_Window.pos + diffPos / 2 ));
			iSetPos( nextPos );
		}
		if ( m_TargetOpenSize != m_OpenSize )
		{
			const nw::math::VEC2 diffSize = m_TargetOpenSize - m_OpenSize;
			const nw::math::VEC2 nextSize = (( diffSize.LengthSquare() < 1.0f )
												? m_TargetOpenSize
												: (m_OpenSize + diffSize / 2 ));
			m_OpenSize = nextSize;
			m_OpenRate = nw::ut::Min( m_OpenRate, 0.99f ); //Force size calculation
		}

		//Toggle open/close when the title bar is double-clicked
		m_TitleDoubleClickChecker.SetHitRect( m_Window.GetTitleRect() );
		m_TitleDoubleClickChecker.Update( tp );
		if ( m_TitleDoubleClickChecker.IsDoubleClick() )
		{
			m_IsShouldOpen = ! m_IsShouldOpen;
		}
		//Self-adjust size with open/close
		const f32 targetOpenRate = (m_IsShouldOpen) ? 1.0f : 0.0f;
		if ( targetOpenRate != m_OpenRate )
		{							  //Never converges if 1.0, so (below)					(below) open/close speed
			const f32 nextOpenRate = nw::ut::Clamp( m_OpenRate + (targetOpenRate * 1.1f - m_OpenRate) * 0.25f, 0.0f, 1.0f );
			Resize( (m_OpenSize - m_CloseSize) * nextOpenRate + m_CloseSize );
			m_OpenRate = nextOpenRate; //Because cannot distinguish if it is a user update or auto open/close when m_OpenRate is not updated after the Resize function.
		}
		//Derived class-side update process callback
		ClientUpdate( tp );
	}
	virtual void Disp()
	{
		//Display window itself
		m_Window.Disp();

		//If the client region is too small, the client region display is omitted
		const nw::math::VEC2 csize = m_Window.GetClientSize();
		if ( csize.y < 5.0f ) return;

		//Render parameter: auto evacuate & restore
		AutoDrawAllParamBak autoDrawAllParamBak;

		//Default render parameter set
		SetTextColor( nw::ut::Color8::WHITE );
		SetTextPos( GetCMPos() );
		SetFontSize( GetFontSize() );
		SetLineWidth( 2.0f );

		//Derived class-side render callback
		ClientDisp();
	}
	//=========================================================//
	void SetPosSizeLogOn( bool posSizeLogOn=true ) { m_PosSizeLogOn = posSizeLogOn; } //Position size log output on
	const nw::math::VEC2& GetWinSize()  const { return m_Window.size; } //Get the window size

protected:
	//================== Implementation on the sub-class side ===================//
	virtual void ClientUpdate( const nn::hid::TouchPanelStatus& tp ) { (void)tp; } //Frame processing for each client
	virtual void ClientDisp() = 0; //Client region rendering (upper left position GetCMPos function, size GetCMSize function)

	//====================  Coordinates Calculation Related ======================//

	//--------- Must be referenced on the client implementation side.
	const nw::math::VEC2& GetCMPos()   	const { return m_CMPos		; } //Upper left coordinate of client region with margins
	const nw::math::VEC2& GetCMSize()  	const { return m_CMSize		; } //Size of client region with margins

	//--------- Convenient if referenced on the client implementation side.
	f32					  GetMargin()   const { return 4.0f; } //Margin standard value
	nw::math::VEC2		  GetMarginVec2() const //Margin standard 2-D
							{ const f32 margin = GetMargin(); return nw::math::VEC2( margin, margin ); }
		  nw::math::VEC2  GetCPos()     const { return m_CPos; } //Upper left coordinate of client region without margins
		  nw::math::VEC2  GetCSize()    const { return m_CSize; } //Size of client region without margins
		  nw::math::VEC2  GetCPos2()    const { return m_CPos + m_CSize; } //Lower right coordinate of client region without margins
	      nw::math::VEC2  GetCMPos2()   const { return m_CMPos + m_CMSize; } //Upper right coordinate of client region with margins
	f32 				  GetFontSize() const { return m_FontSize	; } //Font size for which size changes have been considered
	nw::math::VEC2		  GetFSize() const //No-margin font size 2D
							{ return nw::math::VEC2( m_FontSize * 8.0f / 12.0f, m_FontSize ); }
	nw::math::VEC2		  GetFMSize() const //Margin font size 2D
							{ return FontSizeToPixelSize( m_FontSize ); }
	f32					  GetScale() const { return m_FontSize / m_DefaultFontSize; } //Get magnification/reduction percentage
	//=========================================================//
private:
	static const int DOUBLE_CLICK_MAX_FRAME = 12; //Maximum number of frames considered as a double-click
	vw::system::Window 			m_Window;
	f32 			m_DefaultFontSize;
	nw::math::VEC2 	m_DefaultClientSize;
	nw::math::VEC2	m_CPos;
	nw::math::VEC2	m_CSize;
	nw::math::VEC2	m_CMPos;
	nw::math::VEC2	m_CMSize;
	f32				m_FontSize;
	int				m_ItemNum;
	bool			m_IsShouldOpen; //Should window be opened?
	f32				m_OpenRate;		//Degree to open (0.0f: completely closed - 1.0f: completely opened)
	nw::math::VEC2	m_OpenSize;		//Window size when opened
	nw::math::VEC2	m_CloseSize;	//Window size when closed
	int		 	  m_PrevClickFrame; //Number of frames since the title bar was previously clicked (for detecting double-clicks)
	u8			 	m_PrevTouch;	//For click detection
	bool			m_PosSizeLogOn; //Position size log output on (convenient when adjusting position/size)
	nw::math::VEC2	m_TargetPos;	//So that it is smooth when switching placements
	nw::math::VEC2	m_TargetOpenSize; //So that it is smooth when switching placements
	vw::system::DoubleClickChecker m_TitleDoubleClickChecker; //For title double-click detection

	void iSetPos( const nw::math::VEC2& pos )
	{
		if ( m_PosSizeLogOn )
		{
			NW_LOG( "%s: pos=(%d, %d)\n", m_Window.titleStr, (int)pos.x, (int)pos.y );
		}
		m_Window.pos = pos;
		m_CPos = m_Window.GetClientPos();
		m_CMPos = m_CPos + GetMarginVec2();
	}

};

}   /* namespace system */
}   /* namespace vw     */
#if NN_CURRENT_VERSION_NUMBER >= NN_VERSION_NUMBER(4,0,0,0)
#pragma diag_default 1301 // padding inserted in struct.
#endif

#endif // NW4C_VIEWER_SYSTEM_WINSYS_H_
