/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#include <viewer/system.h>
#include <viewer/system/Perform.h>
#include "../include/InfoDrawer.h"
#include "../../system/include/PerformDrawer.h"
#include "../../system/include/LogSystem.h"
#include "../../g3d/include/InfoDrawer.h"



namespace vw     {
namespace effect {

InfoDrawer* InfoDrawer::mInstance = NULL;

/*===========================================================================*
  Name:         InfoDrawer
  Description:  G3D information display management
 *===========================================================================
*/
InfoDrawer* InfoDrawer::GetInstance()
{
    NW_ASSERT( mInstance != NULL );
    return mInstance;
}

// Constructor
InfoDrawer::InfoDrawer()
{
    mRenderInfoFlag = 0;    // Displays all defaults.
}

// Destructor
InfoDrawer::~InfoDrawer()
{
}

// Initialize
void
InfoDrawer::Initialize()
{
    if (mInstance != NULL) return;

    void *buf = Memory_GetAllocator()->Alloc(sizeof(InfoDrawer));
    if (buf == NULL) return;

    mInstance = new(buf) InfoDrawer();
}

// Finalize
void
InfoDrawer::Finalize()
{
    if (mInstance == NULL) return;

    mInstance->~InfoDrawer();
    Memory_GetAllocator()->Free(mInstance);

    mInstance = NULL;
}

// HID input
void
InfoDrawer::Main(vw::system::HidStatus &hidStatus)
{

}

void
InfoDrawer::Draw(const DrawInfo& info)
{
    // Displays performance information.
    vw::system::PerformDrawer::GetInstance()->DrawMeter();

    // Displays performance bar.
    vw::system::PerformDrawer::GetInstance()->DrawGraph();

    // Display seek bar.
    f32 FrameCount  = info.LoopFlag? info.FrameCount    : 0.0f;
    f32 barMaxFrame = info.LoopFlag? info.MaxFrameCount : 0.0f;
    this->DrawSeekBar(FrameCount, barMaxFrame);
    vw::system::FontDrawer& font = vw::system::FontDrawer::GetInstance();
    font.BeginDraw();

    // Display the mode
    vw::common::CharArray ModeName("Effect/", vw::system::Memory_GetSysAllocator());
    ModeName += info.SubModeName;

    s32 NumLog = vw::system::LogSystem::GetInstance()->GetNumLog();
    if( NumLog > 0 )
    {
        char str[20];
        sprintf(str, "(%d)", NumLog);
        ModeName += str;
    }

#if defined( NN_PCGX )
    font.DrawString(2,20,"%s", ModeName.Get(), NULL);
#else
    font.DrawString(2,20,"%s", ModeName.Get());
#endif

    // Display current number of particles
    font.BeginDraw();
    font.DrawString(148,160,"    Particle : %5d", info.ParticleCount);
    // Display maximum number of particles after playback starts
    font.DrawString(148,180,"Max Particle : %5d", info.MaxParticle);

    // Display playback frame
    font.DrawString(125,200,"%6.1f", FrameCount );
    // Display maximum frames and frame rate
    font.DrawString(175,200, "/%6.1f[RATE%4.1f]\n", info.MaxFrameCount, info.StepFrame);

    // Display loop flag
    font.DrawString(148,140, info.LoopFlag? "        Loop :    On" :
                                            "        Loop :   Off");
    font.EndDraw();

    // FPS display
    font.BeginDraw();
    font.SetColor( ~0 );
    font.DrawString(2, 40, "FPS : %3.1f", vw::system::Video_GetFPS());
    font.EndDraw();
}

// Seek bar display
void
InfoDrawer::DrawSeekBar(f32 frameCount, f32 maxFrame)
{
    f32 seek = 0.0f;
    if( maxFrame > 0.0f )
    {
        seek = frameCount / maxFrame;
    }
    if( seek > 1.0f)
    {
        seek = 1.0f;
    }

    vw::system::Draw2DHelper& draw2D = vw::system::Draw2DHelper::GetInstance();
    draw2D.BeginDraw();

    const nw::ut::Color8 white = nw::ut::Color8(~0);
    draw2D.SetLineWidth(2);
    draw2D.DrawLine(10,230,310,230,white);
    s32 x = 300 * seek;
    draw2D.FillRectangle(6+x,223,8,14,white);

    draw2D.EndDraw();
}

// Switches display modes.
void
InfoDrawer::ChangeRenderInfoFlag()
{
    // Switches in the order: Show all > Hide emitter shape > Hide all.
    // 
    mRenderInfoFlag = ( mRenderInfoFlag + 1 ) % 3;

    switch( mRenderInfoFlag )
    {
        case 0:
        {
            vw::g3d::InfoDrawer::GetInstance()->SetRenderInfoFlag( 0 );
            bool* renderFlag = vw::g3d::InfoDrawer::GetInstance()->GetRenderFlag();
            renderFlag[ vw::g3d::SceneRenderer::RenderParticleEmitterAxis ] = true;
            renderFlag[ vw::g3d::SceneRenderer::RenderParticleEmitterForm ] = true;
        }
        break;

        case 1:
        {
            bool* renderFlag = vw::g3d::InfoDrawer::GetInstance()->GetRenderFlag();
            renderFlag[ vw::g3d::SceneRenderer::RenderParticleEmitterAxis ] = false;
            renderFlag[ vw::g3d::SceneRenderer::RenderParticleEmitterForm ] = false;
        }
        break;

        case 2:
        {
            vw::g3d::InfoDrawer::GetInstance()->SetRenderInfoFlag( 1 );
        }
        break;
    }
}



}   /* namespace effect */
}   /* namespace vw  */
