/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#include <viewer/system.h>
#include "../../effect/include/ParamSetter.h"
#include "../../effect/include/AnimManager.h"

using namespace ProtocolEM4C;
using namespace vw::g3d;

namespace vw  {
namespace effect {

const u32   BODY_BUFFER_SIZE    =   32 * 1024;

static void     ExecCommand                    ( const Header& header, void* body );
static void     ExecCategoryFile               ( const Header& header, void* body );
static void     ExecCategoryObject             ( const Header& header, void* body );
static void     ExecCategoryEmitter            ( const Header& header, void* body );
static void     ExecCategoryParticleForm       ( const Header& header, void* body );
static void     ExecCategoryParticleModel      ( const Header& header, void* body );
static void     ExecCategoryParticleInitializer( const Header& header, void* body );
static void     ExecCategoryParticleUpdater    ( const Header& header, void* body );
static void     ExecCategoryParticleSet        ( const Header& header, void* body );
static void     ExecCategoryMaterial           ( const Header& header, void* body );
static void     ExecCategoryBoneLinkage        ( const Header& header, void* body );

static void     UnknownCommandError( u16 type );

// Buffer for reading the body portion
static u32      sBodyBuffer[ BODY_BUFFER_SIZE / sizeof(u32) ];
static bool     sProtocolClearFlg = true;
static bool     sHeaderRead = false;

/*---------------------------------------------------------------------------*
  Name:         ProtocolExec_Main

  Description:  3D protocol processing main routine

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
ProtocolExec_Main()
{
    static Header header;

    const u16 channel = ProtocolEM4C::Constants::CHANNEL_ID_EFFECTMAKER_ACTIVE;

    if (vw::system::Stream_Polling( channel ))
    {
        system::App::GetInstance().SetFocusID( system::EFFECT_ID );
    }
    vw::system::Stream_Polling( ProtocolEM4C::Constants::CHANNEL_ID_EFFECTMAKER_PASSIVE );

    while( vw::system::Stream_IsConnected( channel ) )
    {
        // Checks the header read complete flag
        if ( !sHeaderRead )
        {
            u32 ReadableBytes = vw::system::Stream_GetReadableBytes( channel );
            if( ReadableBytes < sizeof(Header) ) return;

            vw::system::Stream_Read( channel, sBodyBuffer, sizeof(Header) );

            if( !vw::system::Stream_IsConnected( channel ) ){
                NW_VIEWER_LOG("Connection lost.\n");
                sHeaderRead = false;
                return;
            }

            // Copy header information
            header.type = ((Header*)sBodyBuffer)->type;
            header.size = ((Header*)sBodyBuffer)->size;
            header.arraySize = ((Header*)sBodyBuffer)->arraySize;

            // Enables the header read complete flag
            sHeaderRead = true;
        }

        if ( sHeaderRead )
        {
            u32 ReadableBytes = vw::system::Stream_GetReadableBytes( channel );
            if ( ReadableBytes < header.size + header.arraySize ) return;

            // Load the data itself
            u32 rSize = vw::system::Stream_Read( channel, sBodyBuffer, header.size + header.arraySize );
            if ( rSize != header.size + header.arraySize ) return;

            ExecCommand( header, sBodyBuffer );

            // Reset the header read complete flag
            sHeaderRead = false;

            // Sends response notification to the host side
            int reply = 1;
            vw::system::Stream_Write(channel, &reply, sizeof(int) );
        }
    }
}

/*---------------------------------------------------------------------------*
  Name:         ExecCommand

  Description:  Command analysis

  Arguments:    header  Reference to the header portion
                body    Pointer to the body portion. NULL if the message size is 0

  Returns:      None.
 *---------------------------------------------------------------------------*/
static void
ExecCommand( const Header& header, void* body )
{
    switch ( header.type & Constants::MESSAGE_CATEGORY_EM4C_MASK )
    {
    //---------------------------------
    // File-related category
    case MessageCategory_File:
        ExecCategoryFile( header, body );
        break;

    //---------------------------------
    // Model State Change Category
    case MessageCategory_Object:
        ExecCategoryObject( header, body );
        break;

    //---------------------------------
    // Emitter State Change Category
    case MessageCategory_Emitter:
        ExecCategoryEmitter( header, body );
        break;

    //---------------------------------
    // State Change Category for Emitter Shape
    case MessageCategory_ParticleForm:
        ExecCategoryParticleForm( header, body );
        break;

    //---------------------------------
    // State Change Category for Particle Model
    case MessageCategory_ParticleModel:
        ExecCategoryParticleModel( header, body );
        break;

    //---------------------------------
    // State Change Category for Particle Initializer
    case MessageCategory_ParticleInitializer:
        ExecCategoryParticleInitializer( header, body );
        break;

    //---------------------------------
    // State Change Category for Particle Updater
    case MessageCategory_ParticleUpdater:
        ExecCategoryParticleUpdater( header, body );
        break;

    //---------------------------------
    // State Change Category for Particle Set
    case MessageCategory_ParticleSet:
        ExecCategoryParticleSet( header, body );
        break;

    //---------------------------------
    // Material Management-related Category
    case MessageCategory_Material:
        ExecCategoryMaterial( header, body );
        break;

    //---------------------------------
    // 3D linked category
    case MessageCategory_BoneLinkage:
        ExecCategoryBoneLinkage( header, body );
        break;

    //---------------------------------
    default:
        UnknownCommandError( header.type );
    }
}

/*---------------------------------------------------------------------------*
  Name:         ExecCategoryFile

  Description:  Execute system-related category commands

  Arguments:    header  Reference to the header portion
                body    Pointer to the body portion. NULL if the message size is 0

  Returns:      None.
 *---------------------------------------------------------------------------*/
static void
ExecCategoryFile( const Header& header, void* body )
{
    vw::effect::ParamSetter &paramSetter = vw::effect::ParamSetter::GetInstance();

    switch ( header.type )
    {
    //---------------------------------
    //  Load command
    case MESSAGEPACKET_File_LoadBinaryData::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_File_LoadBinaryData----\n");
        {
            MESSAGEPACKET_File_LoadBinaryData* loadBinaryData =
                reinterpret_cast<MESSAGEPACKET_File_LoadBinaryData*>(body);
            vw::g3d::ResourceManager::GetInstance()->LoadBinaryData(
                (vw::g3d::ResourceSet::BinaryFileType)loadBinaryData->Type,
                loadBinaryData->Id,
                ProtocolEM4C::Constants::CHANNEL_ID_EFFECTMAKER_ACTIVE,
                loadBinaryData->FilePath.Value
            );
            vw::effect::AnimManager::GetInstance()->OnTargetModuleChanged( vw::system::App::GetInstance().GetFocusID() );
        }
        break;
    case MESSAGEPACKET_File_Close::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_File_Close\n--");
        {
            MESSAGEPACKET_File_Close* closeBinaryData = reinterpret_cast<MESSAGEPACKET_File_Close*>(body);
            vw::g3d::ResourceManager::GetInstance()->DeleteResource(
                (vw::g3d::ResourceSet::BinaryFileType)closeBinaryData->Type,
                closeBinaryData->Id
            );
            paramSetter.ClearSelectObject();
        }
        break;
    case MESSAGEPACKET_File_CloseAll::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_File_CloseAll----\n");
        vw::g3d::ResourceManager::GetInstance()->ClearResources(ProtocolEM4C::Constants::CHANNEL_ID_EFFECTMAKER_ACTIVE);
        paramSetter.ClearSelectObject();
        break;
    //---------------------------------
    default:
        UnknownCommandError( header.type );
    }
}

/*---------------------------------------------------------------------------*
  Name:         ExecCategoryObject

  Description:  Execute system-related category commands

  Arguments:    header  Reference to the header portion
                body    Pointer to the body portion. NULL if the message size is 0

  Returns:      None.
 *---------------------------------------------------------------------------*/
static void
ExecCategoryObject( const Header& header, void* body )
{
    vw::effect::ParamSetter &paramSetter = vw::effect::ParamSetter::GetInstance();
    switch(header.type)
    {
    //---------------------------------
    //  Select command
    case MESSAGEPACKET_Object_SelectEM::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Object_SelectEM----\n");
        {
            // Set to selected object
            MESSAGEPACKET_Object_SelectEM* selectEMObject = reinterpret_cast<MESSAGEPACKET_Object_SelectEM*>( body );
            NW_VIEWER_LOG("object seleced EM. Id=%d Type=%d Name=%s\n",
                selectEMObject->Id, selectEMObject->ObjectType, selectEMObject->TargetName.Value );

            // Search for resource index in names
            int selectedIndex = -1;
            if (selectEMObject->TargetType == ProtocolEM4C::EditTargetType_Material) {
                selectedIndex = paramSetter.GetResMaterialIndex(selectEMObject->Id, selectEMObject->TargetName.ToString());
            } else if (selectEMObject->TargetType == ProtocolEM4C::EditTargetType_Emitter) {
                selectedIndex = paramSetter.GetResParticleEmitterIndex(selectEMObject->Id, selectEMObject->TargetName.ToString());
            } else if (selectEMObject->TargetType == ProtocolEM4C::EditTargetType_ParticleSet) {
                selectedIndex = paramSetter.GetResParticleSetIndex(selectEMObject->Id, selectEMObject->TargetName.ToString());
            } else if (selectEMObject->TargetType == ProtocolEM4C::EditTargetType_ParticleModel) {
                selectedIndex = 0;
            }

            // Add selection information
            nw::ut::MoveArray<EditSelectedObject> selectedObject( Memory_GetAllocator() );
            if (selectedIndex != -1) {
                EditSelectedObject edit;
                edit.Id = selectEMObject->Id;
                edit.Index = selectedIndex;
                selectedObject.push_back(edit);
            }

            // Add selection information to module
            paramSetter.SetSelectObject(selectEMObject->TargetType, selectedObject);
        }
        break;
    case MESSAGEPACKET_Object_StartUpdateAll::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Object_StartUpdateAll----\n");
        {
            paramSetter.StartUpdateAll();
        }
        break;
    case MESSAGEPACKET_Object_StepUpdateAll::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Object_StepUpdateAll----\n");
        {
            paramSetter.StepUpdateAll();
        }
        break;
    case MESSAGEPACKET_Object_StopUpdateAll::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Object_StopUpdateAll----\n");
        {
            paramSetter.StopUpdateAll();
        }
        break;
    case MESSAGEPACKET_Object_MoveFirstFrame::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Object_MoveFirstFrame----\n");
        {
            paramSetter.MoveFirstFrame();
        }
        break;
    case MESSAGEPACKET_Object_SetFrameRange::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Object_SetFrameRange----\n");
        {
            paramSetter.SetFrameRange(reinterpret_cast<MESSAGEPACKET_Object_SetFrameRange*>(body));
        }
    default:
        UnknownCommandError( header.type );
    }
}

/*---------------------------------------------------------------------------*
  Name:         ExecCategoryEmitter for Emitter

  Description:  Execute system-related category commands

  Arguments:    header  Reference to the header portion
                body    Pointer to the body portion. NULL if the message size is 0

  Returns:      None.
 *---------------------------------------------------------------------------*/
static void
ExecCategoryEmitter( const Header& header, void* body )
{
    vw::effect::ParamSetter &paramSetter = vw::effect::ParamSetter::GetInstance();

    switch( header.type)
    {
    //---------------------------------
    //  Emitter: Positioning
    case MESSAGEPACKET_Emitter_SetTransform::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Emitter_SetTransform----\n");
        paramSetter.SetEmitterTransform(reinterpret_cast<MESSAGEPACKET_Emitter_SetTransform*>(body));
        break;

    //---------------------------------
    //  Emitter: Emission
    case MESSAGEPACKET_Emitter_SetEmission::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Emitter_SetEmission----\n");
        paramSetter.SetEmitterEmission(reinterpret_cast<MESSAGEPACKET_Emitter_SetEmission*>(body));
        break;

    default:
        UnknownCommandError( header.type );
    }

    // callback call
    paramSetter.CallParticleEmitterCallBack();
}

/*---------------------------------------------------------------------------*
  Name:         ExecCategoryParticleForm For Emitter shape

  Description:  Execute system-related category commands

  Arguments:    header  Reference to the header portion
                body    Pointer to the body portion. NULL if the message size is 0

  Returns:      None.
 *---------------------------------------------------------------------------*/
static void
ExecCategoryParticleForm( const Header& header, void* body )
{
    vw::effect::ParamSetter &paramSetter = vw::effect::ParamSetter::GetInstance();

    switch( header.type)
    {
    //---------------------------------
    //  Cylinder
    case MESSAGEPACKET_ParticleForm_CylinderForm::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_ParticleForm_CylinderForm----\n");
        paramSetter.SetCylinderForm(reinterpret_cast<MESSAGEPACKET_ParticleForm_CylinderForm*>(body));
        break;

    //---------------------------------
    //  Sphere
    case MESSAGEPACKET_ParticleForm_SphereForm::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_ParticleForm_SphereForm----\n");
        paramSetter.SetSphereForm(reinterpret_cast<MESSAGEPACKET_ParticleForm_SphereForm*>(body));
        break;

    //---------------------------------
    //  Disc
    case MESSAGEPACKET_ParticleForm_DiscForm::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_ParticleForm_DiscForm----\n");
        paramSetter.SetDiscForm(reinterpret_cast<MESSAGEPACKET_ParticleForm_DiscForm*>(body));
        break;

    //---------------------------------
    //  Rectangle
    case MESSAGEPACKET_ParticleForm_RectangleForm::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_ParticleForm_RectangleForm----\n");
        paramSetter.SetRectangleForm(reinterpret_cast<MESSAGEPACKET_ParticleForm_RectangleForm*>(body));
        break;

    //---------------------------------
    //  Cube
    case MESSAGEPACKET_ParticleForm_CubeForm::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_ParticleForm_CubeForm----\n");
        paramSetter.SetCubeForm(reinterpret_cast<MESSAGEPACKET_ParticleForm_CubeForm*>(body));
        break;

    default:
        UnknownCommandError( header.type );
    }

    // callback call
    paramSetter.CallParticleEmitterCallBack();
}

/*---------------------------------------------------------------------------*
  Name:         ExecCategoryModel For models

  Description:  Execute system-related category commands

  Arguments:    header  Reference to the header portion
                body    Pointer to the body portion. NULL if the message size is 0

  Returns:      None.
 *---------------------------------------------------------------------------*/
static void
ExecCategoryParticleModel( const Header& header, void* body )
{
    vw::effect::ParamSetter &paramSetter = vw::effect::ParamSetter::GetInstance();

    switch( header.type)
    {
    //---------------------------------
    //  Emitter: Positioning
    case MESSAGEPACKET_ParticleModel_SetTransform::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_ParticleModel_SetTransform----\n");
        paramSetter.SetParticleModelTransform(reinterpret_cast<MESSAGEPACKET_ParticleModel_SetTransform*>(body));
        break;

    default:
        UnknownCommandError( header.type );
    }

    // callback call
    paramSetter.CallParticleModelCallBack();
}

/*---------------------------------------------------------------------------*
  Name:         ExecCategoryParticleInitializer Particle Initializer

  Description:  Execute system-related category commands

  Arguments:    header  Reference to the header portion
                body    Pointer to the body portion. NULL if the message size is 0

  Returns:      None.
 *---------------------------------------------------------------------------*/
static void
ExecCategoryParticleInitializer( const Header& header, void* body )
{
    vw::effect::ParamSetter &paramSetter = vw::effect::ParamSetter::GetInstance();

    switch( header.type)
    {
    //---------------------------------
    //  Diffusion from Y-axis
    case MESSAGEPACKET_ParticleInitializer_YAxisVelocityInitializer::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_ParticleInitializer_YAxisVelocityInitializer----\n");
        paramSetter.SetYAxisVelocityInitializer(
            reinterpret_cast<MESSAGEPACKET_ParticleInitializer_YAxisVelocityInitializer*>(body)
        );
        break;

    //---------------------------------
    //  Spread out from origin
    case MESSAGEPACKET_ParticleInitializer_OriginVelocityInitializer::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_ParticleInitializer_OriginVelocityInitializer----\n");
        paramSetter.SetOriginVelocityInitializer(
            reinterpret_cast<MESSAGEPACKET_ParticleInitializer_OriginVelocityInitializer*>(body)
        );
        break;

    //---------------------------------
    //  Random direction
    case MESSAGEPACKET_ParticleInitializer_RandomVelocityInitializer::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_ParticleInitializer_RandomVelocityInitializer----\n");
        paramSetter.SetRandomVelocityInitializer(
            reinterpret_cast<MESSAGEPACKET_ParticleInitializer_RandomVelocityInitializer*>(body)
        );
        break;

    //---------------------------------
    //  Specified Direction
    case MESSAGEPACKET_ParticleInitializer_DirectionalVelocityInitializer::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_ParticleInitializer_DirectionalVelocityInitializer----\n");
        paramSetter.SetDirectionalVelocityInitializer(
            reinterpret_cast<MESSAGEPACKET_ParticleInitializer_DirectionalVelocityInitializer*>(body)
        );
        break;

    //---------------------------------
    //  Random specified direction
    case MESSAGEPACKET_ParticleInitializer_RandomDirectionalVelocityInitializer::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_ParticleInitializer_RandomDirectionalVelocityInitializer----\n");
        paramSetter.SetRandomDirectionalVelocityInitializer(
            reinterpret_cast<MESSAGEPACKET_ParticleInitializer_RandomDirectionalVelocityInitializer*>(body)
        );
        break;

    //---------------------------------
    //  Randomize results
    case MESSAGEPACKET_ParticleInitializer_Vector3MultRandomInitializer::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_ParticleInitializer_Vector3MultRandomInitializer----\n");
        paramSetter.SetVector3MultRandomInitializer(
            reinterpret_cast<MESSAGEPACKET_ParticleInitializer_Vector3MultRandomInitializer*>(body)
        );
        break;

    //---------------------------------
    //  Random in the range
    case MESSAGEPACKET_ParticleInitializer_FloatRangeRandomInitializer::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_ParticleInitializer_FloatRangeRandomInitializer----\n");
        paramSetter.SetFloatRangeRandomInitializer(
            reinterpret_cast<MESSAGEPACKET_ParticleInitializer_FloatRangeRandomInitializer*>(body)
        );
        break;

    //---------------------------------
    //  Float immediate
    case MESSAGEPACKET_ParticleInitializer_FloatImmediateInitializer::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_ParticleInitializer_FloatImmediateInitializer----\n");
        paramSetter.SetFloatImmediateInitializer(
            reinterpret_cast<MESSAGEPACKET_ParticleInitializer_FloatImmediateInitializer*>(body)
        );
        break;

    //---------------------------------
    //  Float random
    case MESSAGEPACKET_ParticleInitializer_FloatRandomInitializer::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_ParticleInitializer_FloatRandomInitializer----\n");
        paramSetter.SetFloatRandomInitializer(
            reinterpret_cast<MESSAGEPACKET_ParticleInitializer_FloatRandomInitializer*>(body)
        );
        break;

    //---------------------------------
    //  Vec3 random
    case MESSAGEPACKET_ParticleInitializer_Vector3Random1Initializer::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_ParticleInitializer_Vector3Random1Initializer----\n");
        paramSetter.SetVector3Random1Initializer(
            reinterpret_cast<MESSAGEPACKET_ParticleInitializer_Vector3Random1Initializer*>(body)
        );
        break;

    //---------------------------------
    //  Vec3 random
    case MESSAGEPACKET_ParticleInitializer_Vector3Random3Initializer::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_ParticleInitializer_Vector3Random3Initializer----\n");
        paramSetter.SetVector3Random3Initializer(
            reinterpret_cast<MESSAGEPACKET_ParticleInitializer_Vector3Random3Initializer*>(body)
        );
        break;

    //---------------------------------
    //  Vec3 immediate
    case MESSAGEPACKET_ParticleInitializer_Vector3ImmediateInitializer::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_ParticleInitializer_Vector3ImmediateInitializer----\n");
        paramSetter.SetVector3ImmediateInitializer(
            reinterpret_cast<MESSAGEPACKET_ParticleInitializer_Vector3ImmediateInitializer*>(body)
        );
        break;

    //---------------------------------
    //  Vec2 immediate
    case MESSAGEPACKET_ParticleInitializer_Vector2ImmediateInitializer::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_ParticleInitializer_Vector2ImmediateInitializer----\n");
        paramSetter.SetVector2ImmediateInitializer(
            reinterpret_cast<MESSAGEPACKET_ParticleInitializer_Vector2ImmediateInitializer*>(body)
        );
        break;

    default:
        UnknownCommandError( header.type );
    }

    // callback call
    paramSetter.CallParticleSetCallBack();
}

/*---------------------------------------------------------------------------*
  Name:         ExecCategoryParticleUpdater Particle updater

  Description:  Execute system-related category commands

  Arguments:    header  Reference to the header portion
                body    Pointer to the body portion. NULL if the message size is 0

  Returns:      None.
 *---------------------------------------------------------------------------*/
static void
ExecCategoryParticleUpdater( const Header& header, void* body )
{
    vw::effect::ParamSetter &paramSetter = vw::effect::ParamSetter::GetInstance();

    switch( header.type)
    {
    //---------------------------------
    //  Accelerate in specified direction
    case MESSAGEPACKET_ParticleUpdater_GravityUpdater::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_ParticleUpdater_GravityUpdater----\n");
        paramSetter.SetGravityUpdater(reinterpret_cast<MESSAGEPACKET_ParticleUpdater_GravityUpdater*>(body));
        break;

    //---------------------------------
    //  Spin
    case MESSAGEPACKET_ParticleUpdater_SpinUpdater::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_ParticleUpdater_SpinUpdater----\n");
        paramSetter.SetSpinUpdater(reinterpret_cast<MESSAGEPACKET_ParticleUpdater_SpinUpdater*>(body));
        break;

    //---------------------------------
    //  Random
    case MESSAGEPACKET_ParticleUpdater_RandomUpdater::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_ParticleUpdater_RandomUpdater----\n");
        paramSetter.SetRandomUpdater(reinterpret_cast<MESSAGEPACKET_ParticleUpdater_RandomUpdater*>(body));
        break;

    //---------------------------------
    //  Acceleration
    case MESSAGEPACKET_ParticleUpdater_AccelarationUpdater::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_ParticleUpdater_AccelarationUpdater----\n");
        paramSetter.SetAccelarationUpdater(reinterpret_cast<MESSAGEPACKET_ParticleUpdater_AccelarationUpdater*>(body));
        break;

    //---------------------------------
    //  UpVector rotation
    case MESSAGEPACKET_ParticleUpdater_RotateUpVectorUpdater::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_ParticleUpdater_RotateUpVectorUpdater----\n");
        paramSetter.SetRotateUpVectorUpdater(reinterpret_cast<MESSAGEPACKET_ParticleUpdater_RotateUpVectorUpdater*>(body));
        break;

    //---------------------------------
    //  Float immediate
    case MESSAGEPACKET_ParticleUpdater_FloatImmediateUpdater::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_ParticleUpdater_FloatImmediateUpdater----\n");
        paramSetter.SetFloatImmediateUpdater(reinterpret_cast<MESSAGEPACKET_ParticleUpdater_FloatImmediateUpdater*>(body));
        break;

    //---------------------------------
    //  Vec3 addition
    case MESSAGEPACKET_ParticleUpdater_Vector3AdditiveUpdater::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_ParticleUpdater_Vector3AdditiveUpdater----\n");
        paramSetter.SetVector3AdditiveUpdater(reinterpret_cast<MESSAGEPACKET_ParticleUpdater_Vector3AdditiveUpdater*>(body));
        break;

    //---------------------------------
    //  Vec3 immediate
    case MESSAGEPACKET_ParticleUpdater_Vector3ImmediateUpdater::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_ParticleUpdater_Vector3ImmediateUpdater----\n");
        paramSetter.SetVector3ImmediateUpdater(reinterpret_cast<MESSAGEPACKET_ParticleUpdater_Vector3ImmediateUpdater*>(body));
        break;

    //---------------------------------
    //  Vec2 immediate
    case MESSAGEPACKET_ParticleUpdater_Vector2ImmediateUpdater::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_ParticleUpdater_Vector2ImmediateUpdater----\n");
        paramSetter.SetVector2ImmediateUpdater(reinterpret_cast<MESSAGEPACKET_ParticleUpdater_Vector2ImmediateUpdater*>(body));
        break;

    //---------------------------------
    //  Texture Pattern
    case MESSAGEPACKET_ParticleUpdater_TexturePatternUpdater::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_ParticleUpdater_TexturePatternUpdater----\n");
        paramSetter.SetTexturePatternUpdater(reinterpret_cast<MESSAGEPACKET_ParticleUpdater_TexturePatternUpdater*>(body));
        break;

    //---------------------------------
    //  Child
    case MESSAGEPACKET_ParticleUpdater_ChildUpdater::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_ParticleUpdater_ChildUpdater----\n");
        paramSetter.SetChildUpdater(reinterpret_cast<MESSAGEPACKET_ParticleUpdater_ChildUpdater*>(body));
        break;

    //---------------------------------
    //  Float 4-key animation
    case MESSAGEPACKET_ParticleUpdater_FloatImmediate4KeyUpdater::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_ParticleUpdater_FloatImmediate4KeyUpdater----\n");
        paramSetter.SetFloatImmediate4KeyUpdater(reinterpret_cast<MESSAGEPACKET_ParticleUpdater_FloatImmediate4KeyUpdater*>(body));
        break;

    //---------------------------------
    //  Vec3 4-key animation
    case MESSAGEPACKET_ParticleUpdater_Vector3Immediate4KeyUpdater::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_ParticleUpdater_Vector3Immediate4KeyUpdater----\n");
        paramSetter.SetVector3Immediate4KeyUpdater(reinterpret_cast<MESSAGEPACKET_ParticleUpdater_Vector3Immediate4KeyUpdater*>(body));
        break;

    //---------------------------------
    //  Magnetic Force
    case MESSAGEPACKET_ParticleUpdater_MagnetUpdater::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_ParticleUpdater_MagnetUpdater----\n");
        paramSetter.SetMagnetUpdater(reinterpret_cast<MESSAGEPACKET_ParticleUpdater_MagnetUpdater*>(body));
        break;

    //---------------------------------
    //  Collision
    case MESSAGEPACKET_ParticleUpdater_CollisionUpdater::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_ParticleUpdater_CollisionUpdater----\n");
        paramSetter.SetCollisionUpdater(reinterpret_cast<MESSAGEPACKET_ParticleUpdater_CollisionUpdater*>(body));
        break;

    default:
        UnknownCommandError( header.type );
    }

    // callback call
    paramSetter.CallParticleSetCallBack();
}

/*---------------------------------------------------------------------------*
  Name:         ExecCategoryParticleSet Particle set

  Description:  Execute system-related category commands

  Arguments:    header  Reference to the header portion
                body    Pointer to the body portion. NULL if the message size is 0

  Returns:      None.
 *---------------------------------------------------------------------------*/
static void
ExecCategoryParticleSet( const Header& header, void* body )
{
    vw::effect::ParamSetter &paramSetter = vw::effect::ParamSetter::GetInstance();

    switch( header.type)
    {
    //---------------------------------
    //  Shape builder
    case MESSAGEPACKET_ParticleSet_ShapeBuilder::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_ParticleSet_ShapeBuilder----\n");
        paramSetter.SetShapeBuilder(reinterpret_cast<MESSAGEPACKET_ParticleSet_ShapeBuilder*>(body));
        break;

    default:
        UnknownCommandError( header.type );
    }

    // callback call
    paramSetter.CallParticleSetCallBack();
}

/*---------------------------------------------------------------------------*
  Name:         ExecCategoryMaterial

  Description:  Execute system-related category commands

  Arguments:    header  Reference to the header portion
                body    Pointer to the body portion. NULL if the message size is 0

  Returns:      None.
 *---------------------------------------------------------------------------*/
static void
ExecCategoryMaterial( const Header& header, void* body )
{
    vw::effect::ParamSetter &paramSetter = vw::effect::ParamSetter::GetInstance();

    switch ( header.type )
    {
    //---------------------------------
    // Material blender
    case MESSAGEPACKET_Material_MaterialBlender::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Material_MaterialBlender----\n");
        paramSetter.SetMaterialBlender(reinterpret_cast<MESSAGEPACKET_Material_MaterialBlender*>(body));
        break;

    //---------------------------------
    // Material color
    case MESSAGEPACKET_Material_MaterialColor::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Material_MaterialColor----\n");
        paramSetter.SetMaterialColor(reinterpret_cast<MESSAGEPACKET_Material_MaterialColor*>(body));
        break;

    //---------------------------------
    // Material mapper
    case MESSAGEPACKET_Material_MaterialMapper::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Material_MaterialMapper----\n");
        paramSetter.SetMaterialMapper(reinterpret_cast<MESSAGEPACKET_Material_MaterialMapper*>(body));
        break;

    //---------------------------------
    // Material combiner
    case MESSAGEPACKET_Material_MaterialCombiner::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Material_MaterialCombiner----\n");
        paramSetter.SetMaterialCombiner(reinterpret_cast<MESSAGEPACKET_Material_MaterialCombiner*>(body));
        break;

    //---------------------------------
    default:
        UnknownCommandError( header.type );
    }

    // callback call
    paramSetter.CallMaterialCallBack();
}

/*---------------------------------------------------------------------------*
  Name:         ExecCategoryBoneLinkage For 3D link

  Description:  Execute system-related category commands

  Arguments:    header  Reference to the header portion
                body    Pointer to the body portion. NULL if the message size is 0

  Returns:      None.
 *---------------------------------------------------------------------------*/
static void
ExecCategoryBoneLinkage( const Header& header, void* body )
{
    vw::effect::ParamSetter &paramSetter = vw::effect::ParamSetter::GetInstance();
    const u32 channel = ProtocolEM4C::Constants::CHANNEL_ID_EFFECTMAKER_PASSIVE;

    switch( header.type)
    {
    //---------------------------------
    //  Get model list
    case MESSAGEPACKET_BoneLinkage_ModelNameList::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_BoneLinkage_ModelNameList----\n");
        {
            nw::ut::MoveArray<unsigned int> hashBuffer( Memory_GetAllocator() );
            nw::ut::MoveArray<char> strBuffer( Memory_GetAllocator() );
            paramSetter.GatherSkeletalModelList( hashBuffer, strBuffer );

            // Set header
            ProtocolEM4C::Header sendHeader;
            sendHeader.type = header.type;
            sendHeader.size = header.size;
            sendHeader.arraySize = hashBuffer.size() * sizeof(unsigned int) + strBuffer.size();

            // Size check
            int totalDataSize = sizeof( sendHeader ) + sendHeader.size + sendHeader.arraySize;
            if ( totalDataSize >= BODY_BUFFER_SIZE )
            {
                NW_VIEWER_LOG( "MESSAGEPACKET_BoneLinkage_ModelNameList | Buffer is not enough\n" );
                return;
            }

            // Set packet
            ProtocolEM4C::MESSAGEPACKET_BoneLinkage_ModelNameList modelListPacket;
            modelListPacket.ModifiedProperties = 0xFFFFFFFF;
            modelListPacket.ModifiedPropertyElements = 0xFFFFFFFF;
            modelListPacket.PanelID = reinterpret_cast<MESSAGEPACKET_BoneLinkage_ModelNameList*>(body)->PanelID;
            modelListPacket.Count = hashBuffer.size();
            NW_ASSERT( sendHeader.size == sizeof( modelListPacket ) );

            // Copy data
            u8* buf = (u8*)sBodyBuffer;
            memcpy( buf, &sendHeader, sizeof( sendHeader ) );
            buf += sizeof( sendHeader );
            memcpy( buf, &modelListPacket, sendHeader.size );
            buf += sendHeader.size;
            if ( sendHeader.arraySize != 0 ) {
                memcpy( buf, &(hashBuffer.front()), hashBuffer.size() * sizeof(unsigned int) );
                buf += hashBuffer.size() * sizeof(unsigned int);
                memcpy( buf, &(strBuffer.front()), strBuffer.size() );
            }

            // Sending Data
            vw::system::Stream_Write( channel, sBodyBuffer, totalDataSize );
        }
        break;

    //---------------------------------
    //  Get bone list
    case MESSAGEPACKET_BoneLinkage_BoneNameList::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_BoneLinkage_BoneNameList----\n");
        {
            nw::ut::MoveArray<char> strBuffer( Memory_GetAllocator() );
            int boneCount = 0;
            paramSetter.GatherBoneList(
                reinterpret_cast<MESSAGEPACKET_BoneLinkage_BoneNameList*>(body)->ModelNameHashID,
                strBuffer,
                boneCount
            );

            // Set header
            ProtocolEM4C::Header sendHeader;
            sendHeader.type = header.type;
            sendHeader.size = header.size;
            sendHeader.arraySize = strBuffer.size();

            // Size check
            int totalDataSize = sizeof( sendHeader ) + sendHeader.size + sendHeader.arraySize;
            if ( totalDataSize >= BODY_BUFFER_SIZE )
            {
                NW_VIEWER_LOG( "MESSAGEPACKET_BoneLinkage_BoneNameList | Buffer is not enough\n" );
                return;
            }

            // Set packet
            ProtocolEM4C::MESSAGEPACKET_BoneLinkage_BoneNameList boneListPacket;
            boneListPacket.ModifiedProperties = 0xFFFFFFFF;
            boneListPacket.ModifiedPropertyElements = 0xFFFFFFFF;
            boneListPacket.PanelID = reinterpret_cast<MESSAGEPACKET_BoneLinkage_BoneNameList*>(body)->PanelID;
            boneListPacket.ModelNameHashID = reinterpret_cast<MESSAGEPACKET_BoneLinkage_BoneNameList*>(body)->ModelNameHashID;
            boneListPacket.Count = boneCount;
            NW_ASSERT( sendHeader.size == sizeof( boneListPacket ) );

            // Copy data
            u8* buf = (u8*)sBodyBuffer;
            memcpy( buf, &sendHeader, sizeof( sendHeader ) );
            buf += sizeof( sendHeader );
            memcpy( buf, &boneListPacket, sendHeader.size );
            buf += sendHeader.size;
            if ( sendHeader.arraySize != 0 ) {
                memcpy( buf, &(strBuffer.front()), strBuffer.size() );
            }

            // Sending Data
            vw::system::Stream_Write( channel, sBodyBuffer, totalDataSize );
        }
        break;

    //---------------------------------
    //  Set constraints
    case MESSAGEPACKET_BoneLinkage_SetBoneLink::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_BoneLinkage_SetBoneLink----\n");
        paramSetter.SetBoneLink(reinterpret_cast<MESSAGEPACKET_BoneLinkage_SetBoneLink*>(body));
        break;

    default:
        UnknownCommandError( header.type );
    }
}

/*---------------------------------------------------------------------------*
  Name:         UnknownCommandError

  Description:  Sends undefined command error

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
static void
UnknownCommandError( u16 type )
{
    // NW4R_WARNING( false, RESTXT_ON_UNKNOWN_COMMAND_x, type );

    //if ( type & HEADER_REPLY_FLAG )
    // Returns error message to those with enabled reply flags
    //{
    //    ErrSystemUnknownCommand reply;
    //    reply.command = type;
    //
    //    Stream3d_Reply( ERR_SYSTEM_UNKNOWN_COMMAND, reply );
    //}
}

}   /* namespace effect */
}   /* namespace vw  */

