/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#include <nn/math.h>
#include <viewer/system.h>
#include "../include/CameraMgr.h"
#include "../include/Resources.h"

namespace vw  {
namespace g2d {

/*---------------------------------------------------------------------------*
  Name:         CameraMgr::CameraMgr
  Description:  Constructor
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
/* ctor */
CameraMgr::CameraMgr()
 : mRect(-320.f, 240.f, 320.f, -240.f)  // ltrb
 , mbWideLocation( false )
 , mbPerspective( false )
{
    // With a right-handed coordinate system, near > far, but with MTXORTHO, it is assumed that far - near is positive.
    // 
    mFovy    =  45.f;
    mNear    = -1000;
    mFar     =  1000;
    mMinNear =  0.05f;

    mWideScale = 4.f / 3.f;
    mTestScale = 1.0f;

    mStereoBasePlane = 0.0f;
    mStereoFactor = 1.0f;
    mRealSwitch = false;
}

void
CameraMgr::Init()
{
    mStereoCamera.Initialize();

    this->CalcProjMtx();
}

f32
CameraMgr::GetDistance() const
{
    f32 fovy = this->GetFovy();

    return this->GetLayoutHeight() * 0.5f / nn::math::TanDeg(fovy * 0.5f);
}

/*---------------------------------------------------------------------------*
  Name:         CameraMgr::CalcPerspMtx
  Description:  Calculates the projection matrix for perspective projection.
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
CameraMgr::CalcPerspMtx()
{
    //
    // NOTE: The formula must match InfoSubMenu::Draw().
    //

    const f32 widthScale = mbWideLocation ? mWideScale: 1.f;

    // Get the layout Rect.
    // When mbWideLocation is TRUE, Rect also reflects that effect.
    f32 left  = mRect.left  * widthScale;
    f32 right = mRect.right * widthScale;

    const f32 width  = right - left;
    const f32 height = mRect.top - mRect.bottom;
    const f32 aspect = width / height;

    const f32 dist = this->GetDistance();
    const f32 depthLevel = nw::ut::Max(1.0f, dist - this->GetStereoBasePlane());

    // Perspective projection matrix. For the lower screen.
    nn::math::MTX44PerspectivePivotDeg(
        &mPerspMtx,
        mFovy,
        aspect,
        nw::ut::Max(mNear + dist, mMinNear),
        mFar + dist,
        nn::math::PIVOT_UPSIDE_TO_TOP);

    // Perspective projection matrix for calculating stereoscopy.
    nn::math::MTX44 proj;
    nn::math::MTX44PerspectiveDeg(
        &proj,
        mFovy,
        aspect,
        nw::ut::Max(mNear + dist, mMinNear),
        mFar + dist);

    nn::math::MTX44 projL;
    nn::math::MTX44 viewL = nn::math::MTX44::Identity();
    nn::math::MTX44 projR;
    nn::math::MTX44 viewR = nn::math::MTX44::Identity();;

    mStereoCamera.CalculateMatrices(
        &projL,
        (nn::math::MTX34*)&viewL,
        &projR,
        (nn::math::MTX34*)&viewR,
        &proj,
        &nn::math::MTX34::Identity(),
        depthLevel,
        mStereoFactor,
        mRealSwitch);

    // Applies stereoscopic view correction to the perspective projection matrix.
    nn::math::MTX44Mult(&mPerspMtxL, &projL, &viewL);
    nn::math::MTX44Mult(&mPerspMtxR, &projR, &viewR);
}

/*---------------------------------------------------------------------------*
  Name:         CameraMgr::CalcOrthoMtx
  Description:  Calculates the projection matrix for orthogonal projection.
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
CameraMgr::CalcOrthoMtx()
{
    //
    // NOTE: The formula must match InfoSubMenu::Draw().
    //

    const f32 widthScale = mbWideLocation ? mWideScale: 1.f;

    // Get the layout Rect.
    // When mbWideLocation is TRUE, Rect also reflects that effect.
    f32 left  = mRect.left  * widthScale;
    f32 right = mRect.right * widthScale;

    const f32 width  = this->GetLayoutWidth() * widthScale;
    const f32 height = this->GetLayoutHeight();
    const f32 aspect = width / height;

    const f32 dist = this->GetDistance();
    const f32 depthLevel = nw::ut::Max(1.0f, dist - this->GetStereoBasePlane());

    // Perspective projection matrix for calculating stereoscopy.
    nn::math::MTX44 proj;
    nn::math::MTX44PerspectiveDeg(
        &proj,
        mFovy,
        aspect,
        nw::ut::Max(mNear + dist, mMinNear),
        mFar + dist);

    // Calculates the amount of parallax during parallel projection.
    nn::math::MTX44 projL, projR;
    nn::math::MTX34 viewL, viewR;
    mStereoCamera.CalculateMatrices(
        &projL,
        &viewL,
        &projR,
        &viewR,
        &proj,
        &nn::math::MTX34::Identity(),
        depthLevel,
        mStereoFactor,
        false);

    mOrthoParallax = mStereoCamera.GetCoefficientForParallax() * NN_GX_DISPLAY0_HEIGHT;

    // Parallel projection matrix.
    nn::math::MTX44OrthoPivot(
        &mOrthoMtx,
        left,
        right,
        mRect.bottom,
        mRect.top,
        nw::ut::Max(mNear + dist, mMinNear),
        mFar + dist,
        nn::math::PIVOT_UPSIDE_TO_TOP);
}

/*---------------------------------------------------------------------------*
  Name:         CameraMgr::GetViewMtx
  Description:  Gets the view matrix.
  Arguments:    mtx     Region to get the view matrix
                persp   Specifies whether a perspective projection
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
CameraMgr::GetViewMtx( nn::math::MTX34* pMtx ) const
{
    NW_NULL_ASSERT( pMtx );

    nn::math::VEC3 pos(0, 0, this->GetDistance());
    nn::math::VEC3 up(0, 1, 0);
    nn::math::VEC3 target(0, 0, 0);
    nn::math::MTX34LookAt(pMtx, pos, up, target);

#if 0 // FIXME!
    // For wide test
    // Scales down vertical/horizontal, scales up the horizontal for wide
    if (mTestScale != 1.0f)
    {
        nw4r::math::MTX34 scale;

        system::VideoSetting setting;

        system::Video_GetSetting( &setting );

        MTXScale(scale, setting.efbHeight * mRect.GetWidth() / (-mRect.GetHeight() * setting.efbWidth) * mTestScale, mTestScale, 1.f);
        MTXConcat(scale, mtx, mtx);
    }
#endif
}

/*---------------------------------------------------------------------------*
  Name:         CameraMgr::SetupProjection
  Description:  Performs projection settings.
  Arguments:    persp   Specifies whether a perspective projection.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
CameraMgr::SetupProjection( vw::system::ScreenKind screen, bool persp )
{
    nw::lyt::DrawInfo& drawInfo = vw::g2d::res::GetDrawInfo();
    nw::lyt::Drawer& drawer = vw::g2d::res::GetDrawer();

    if ( persp )
    {
        switch (screen)
        {
        case vw::system::UPPER_SCREEN:
            drawInfo.SetProjectionMtx(mPerspMtxL);
            break;

        case vw::system::EXTENSION_SCREEN:
            drawInfo.SetProjectionMtx(mPerspMtxR);
            break;

        case vw::system::LOWER_SCREEN:
            drawInfo.SetProjectionMtx(mPerspMtx);
            break;

        default:
            NW_ASSERT(false);
        }

        drawer.SetParallax(0.0f, 0.0f);
    }
    else
    {
        drawInfo.SetProjectionMtx(mOrthoMtx);
        const f32 dist = this->GetDistance();
        const f32 depthLevel = nw::ut::Max(1.0f, dist - this->GetStereoBasePlane());

        switch (screen)
        {
        case vw::system::UPPER_SCREEN:
            drawer.SetParallax(-mOrthoParallax, depthLevel);
            break;

        case vw::system::EXTENSION_SCREEN:
            drawer.SetParallax(mOrthoParallax, depthLevel);
            break;

        case vw::system::LOWER_SCREEN:
            drawer.SetParallax(0, 0);
            break;

        default:
            NW_ASSERT(false);
        }
    }
}


}   /* namespace g2d */
}   /* namespace vw  */



