/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#include <viewer/system.h>
#include "../include/MenuSubMode.h"
#include "../include/Module.h"
#include "../include/Resources.h"
#include "../../system/include/HidManager.h"
#include "../../system/include/VersionDraw.h"

namespace vw  {
namespace g2d {

#define DECL_MENU_FUNC( type ) \
class type##Menu : public MenuBase \
{ \
public: \
    virtual void Apply(); \
    virtual const char* GetMenuName() const { return #type; } \
    virtual MenuType GetMenuType() const { return Menu_##type; } \
};

#define DECL_MENU_CONF( type ) \
class type##Menu : public MenuBase \
{ \
public: \
    virtual void Prev(); \
    virtual void Next(); \
    virtual bool Draw(); \
    virtual const char* GetMenuName() const { return #type; } \
    virtual MenuType GetMenuType() const { return Menu_##type; } \
};

class MenuBase : public IMenu
{
protected:
    void DrawString(const char* format, ...)
    {
        const s32 aw = 5;
        const s32 ah = 100;

        std::va_list vargs;
        va_start(vargs, format);

        vw::system::FontDrawer& font = vw::system::FontDrawer::GetInstance();
        font.BeginDraw();
        font.DrawString(aw + 8, ah + 10, format, vargs);
        font.EndDraw();

        va_end(vargs);
    }
};

DECL_MENU_FUNC( Reset )
DECL_MENU_FUNC( ScreenCapture )
#ifdef NW_TEST_LYT_VRAM
DECL_MENU_CONF( TexMemArea )
#endif
DECL_MENU_CONF( ZCompare )
DECL_MENU_CONF( Fovy )
DECL_MENU_CONF( Near )
DECL_MENU_CONF( Far )
DECL_MENU_CONF( MinNear )
DECL_MENU_CONF( Stereo_BasePlane )
DECL_MENU_CONF( Stereo_Factor )
DECL_MENU_CONF( Stereo_RealSwitch )

void ResetMenu::Apply()
{
    ModuleG2dImpl& module = ModuleG2dImpl::GetInstance();
    module.Reset();
}

void ScreenCaptureMenu::Apply()
{
    vw::system::GetCapture().Capture();
}

// Version that allows capture of only the top screen
class ScreenCaptureUpperMenu : public ScreenCaptureMenu
{
    typedef ScreenCaptureMenu BaseClass;

public:
    virtual void Apply()
    {
        ModuleG2dImpl& module = ModuleG2dImpl::GetInstance();

        if (module.IsToUpper())
        {
            BaseClass::Apply();
        }
    }

    virtual bool Draw()
    {
        ModuleG2dImpl& module = ModuleG2dImpl::GetInstance();
        this->DrawString("%s%s", this->GetMenuName(), module.IsToUpper()? "" : " (Disable)");
        return true;
    }
};

#ifdef NW_TEST_LYT_VRAM
void TexMemAreaMenu::Prev()
{
    ModuleG2dImpl& module = ModuleG2dImpl::GetInstance();
    nw::lyt::MemArea texMemArea = (nw::lyt::MemArea)((module.GetTexMemArea() + nw::lyt::MEMAREA_MAX - 1) % nw::lyt::MEMAREA_MAX);
    module.SetTexMemArea(texMemArea);
}

void TexMemAreaMenu::Next()
{
    ModuleG2dImpl& module = ModuleG2dImpl::GetInstance();
    nw::lyt::MemArea texMemArea = (nw::lyt::MemArea)((module.GetTexMemArea() + 1) % nw::lyt::MEMAREA_MAX);
    module.SetTexMemArea(texMemArea);
}

bool TexMemAreaMenu::Draw()
{
    const char* labels[] =
    {
        "FCRAM",
        "VRAMA",
        "VRAMB",
    };

    NW_STATIC_ASSERT(sizeof(labels) / sizeof(labels[0]) == nw::lyt::MEMAREA_MAX);

    ModuleG2dImpl& module = ModuleG2dImpl::GetInstance();
    nw::lyt::MemArea texMemArea = module.GetTexMemArea();
    this->DrawString("/g2d/tex_mem_area = %s", labels[texMemArea]);
    return true;
}
#endif

void ZCompareMenu::Prev()
{
    ModuleG2dImpl& module = ModuleG2dImpl::GetInstance();
    module.SetZCompareEnable(! module.IsZCompareEnable());
}

void ZCompareMenu::Next()
{
    this->Prev();
}

bool ZCompareMenu::Draw()
{
    ModuleG2dImpl& module = ModuleG2dImpl::GetInstance();
    this->DrawString("/g2d/z_compare = %s", module.IsZCompareEnable()? "ON" : "OFF");
    return true;
}

void FovyMenu::Prev()
{
    const f32 step = -1.0f;

    CameraMgr& cameraMgr = res::GetCameraMgr();
    f32 fovy = cameraMgr.GetFovy();
    fovy += step;
    if (ModuleG2dImpl::FovyMin <= fovy)
    {
        cameraMgr.SetFovy(fovy);
    }
}

void FovyMenu::Next()
{
    const f32 step = 1.0f;

    CameraMgr& cameraMgr = res::GetCameraMgr();
    f32 fovy = cameraMgr.GetFovy();
    fovy += step;
    if (fovy <= ModuleG2dImpl::FovyMax)
    {
        cameraMgr.SetFovy(fovy);
    }
}

bool FovyMenu::Draw()
{
    CameraMgr& cameraMgr = res::GetCameraMgr();
    f32 fovy = cameraMgr.GetFovy();
    this->DrawString("/g2d/camera/fovy = %.1f", fovy);
    return true;
}

void NearMenu::Prev()
{
    const f32 step = -10.0f;

    CameraMgr& cameraMgr = res::GetCameraMgr();
    f32 near = cameraMgr.GetNear();
    near += step;
    cameraMgr.SetNear(near);
}

void NearMenu::Next()
{
    const f32 step = 10.0f;

    CameraMgr& cameraMgr = res::GetCameraMgr();
    f32 near = cameraMgr.GetNear();
    near += step;
    if (near < cameraMgr.GetFar())
    {
        cameraMgr.SetNear(near);
    }
}

bool NearMenu::Draw()
{
    CameraMgr& cameraMgr = res::GetCameraMgr();
    f32 near = cameraMgr.GetNear();
    this->DrawString("/g2d/camera/near = %.1f", near);
    return true;
}

void FarMenu::Prev()
{
    const f32 step = -10.0f;

    CameraMgr& cameraMgr = res::GetCameraMgr();
    f32 far = cameraMgr.GetFar();
    far += step;
    if (cameraMgr.GetNear() < far && cameraMgr.GetMinNear() < far)
    {
        cameraMgr.SetFar(far);
    }
}

void FarMenu::Next()
{
    const f32 step = 10.0f;

    CameraMgr& cameraMgr = res::GetCameraMgr();
    f32 far = cameraMgr.GetFar();
    far += step;
    cameraMgr.SetFar(far);
}

bool FarMenu::Draw()
{
    CameraMgr& cameraMgr = res::GetCameraMgr();
    f32 far = cameraMgr.GetFar();
    this->DrawString("/g2d/camera/far = %.1f", far);
    return true;
}

void MinNearMenu::Prev()
{
    const f32 step = -0.01f;

    CameraMgr& cameraMgr = res::GetCameraMgr();
    f32 near = cameraMgr.GetMinNear();
    near += step;
    if (0.0f < near)
    {
        cameraMgr.SetMinNear(near);
    }
}

void MinNearMenu::Next()
{
    const f32 step = 0.01f;

    CameraMgr& cameraMgr = res::GetCameraMgr();
    f32 near = cameraMgr.GetMinNear();
    near += step;
    if (near < cameraMgr.GetFar())
    {
        cameraMgr.SetMinNear(near);
    }
}

bool MinNearMenu::Draw()
{
    CameraMgr& cameraMgr = res::GetCameraMgr();
    f32 near = cameraMgr.GetMinNear();
    this->DrawString("/g2d/camera/min_near = %f", near);
    return true;
}

void Stereo_BasePlaneMenu::Prev()
{
    const f32 step = -1.0f;

    CameraMgr& cameraMgr = res::GetCameraMgr();
    f32 value = cameraMgr.GetStereoBasePlane();
    value += step;
    cameraMgr.SetStereoBasePlane(value);
}

void Stereo_BasePlaneMenu::Next()
{
    const f32 step = 1.0f;

    CameraMgr& cameraMgr = res::GetCameraMgr();
    f32 value = cameraMgr.GetStereoBasePlane();
    value += step;
    cameraMgr.SetStereoBasePlane(value);
}

bool Stereo_BasePlaneMenu::Draw()
{
    CameraMgr& cameraMgr = res::GetCameraMgr();
    f32 value = cameraMgr.GetStereoBasePlane();
    this->DrawString("/g2d/stereo/base_plane = %.1f", value);
    return true;
}

void Stereo_FactorMenu::Prev()
{
    const f32 step = -0.1f;

    CameraMgr& cameraMgr = res::GetCameraMgr();
    f32 factor = cameraMgr.GetStereoFactor();
    factor += step;
    if (factor < 0.0f)
    {
        factor = 0.0f;
    }

    cameraMgr.SetStereoFactor(factor);
}

void Stereo_FactorMenu::Next()
{
    const f32 step = 0.1f;

    CameraMgr& cameraMgr = res::GetCameraMgr();
    f32 factor = cameraMgr.GetStereoFactor();
    factor += step;
    if (1.0f < factor)
    {
        factor = 1.0f;
    }

    cameraMgr.SetStereoFactor(factor);
}

bool Stereo_FactorMenu::Draw()
{
    CameraMgr& cameraMgr = res::GetCameraMgr();
    this->DrawString("/g2d/stereo/factor = %.1f", cameraMgr.GetStereoFactor());
    return true;
}

void Stereo_RealSwitchMenu::Prev()
{
    CameraMgr& cameraMgr = res::GetCameraMgr();
    cameraMgr.SetRealSwitch(! cameraMgr.GetRealSwitch());
}

void Stereo_RealSwitchMenu::Next()
{
    this->Prev();
}

bool Stereo_RealSwitchMenu::Draw()
{
    CameraMgr& cameraMgr = res::GetCameraMgr();
    this->DrawString("/g2d/stereo/real_switch = %s", cameraMgr.GetRealSwitch()? "ON": "OFF");
    return true;
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
MenuSubModeBase::MenuSubModeBase()
: mMenu(Memory_GetAllocator())
, mCurrentMenuIndex(0)
{
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
MenuSubModeBase::~MenuSubModeBase()
{
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
bool
MenuSubModeBase::Main(vw::system::HidStatus& hidStatus)
{
    bool bProcessed = false;

    switch (system::GetPadEvent(hidStatus, system::GetMenuSubModePadBindings()))
    {
    case system::MNU_MENU_APPLY:
        mMenu[mCurrentMenuIndex]->Apply();
        bProcessed = true;
        break;
    case system::MNU_SWITCH_PREV_MENU:
        mCurrentMenuIndex--;
        if(mCurrentMenuIndex < 0) mCurrentMenuIndex = mMenu.size() - 1;
        bProcessed = true;
        break;
    case system::MNU_SWITCH_NEXT_MENU:
        mCurrentMenuIndex++;
        if(mCurrentMenuIndex > mMenu.size() - 1) mCurrentMenuIndex = 0;
        bProcessed = true;
        break;
    case system::MNU_SWITCH_PREV_MENU_ITEM:
        mMenu[mCurrentMenuIndex]->Prev();
        bProcessed = true;
        break;
    case system::MNU_SWITCH_NEXT_MENU_ITEM:
        mMenu[mCurrentMenuIndex]->Next();
        bProcessed = true;
        break;
    default:
        break;
    }

    return bProcessed;
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
MenuSubModeBase::Draw()
{
    const char* name = mMenu[mCurrentMenuIndex]->GetMenuName();
    if( name == 0 ) name = "N/A";

    vw::system::Draw2DHelper& draw2D = vw::system::Draw2DHelper::GetInstance();
    draw2D.BeginDraw();
    s32 aw = 5;
    s32 ah = 100;
    s32 w = 320-aw*2;
    s32 h = 240-ah*2;
    draw2D.FillRectangle(aw,ah,w,h, nw::ut::Color8(64,64,96,255));
    draw2D.EndDraw();

    if (!mMenu[mCurrentMenuIndex]->Draw())
    {
        vw::system::FontDrawer& font = vw::system::FontDrawer::GetInstance();
        font.BeginDraw();
        font.DrawString(aw + 8, 110, name);
        font.EndDraw();
    }

    // draw version
    vw::system::Version_Draw();
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
const char*
MenuSubModeBase::GetModeName() const
{
    return "Menu";
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
SubModeType
MenuSubModeBase::GetModeType() const
{
    return SubMode_Menu;
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
MenuSubMode::MenuSubMode() : MenuSubModeBase()
{
     nw::os::IAllocator* pAllocator = Memory_GetAllocator();

#define ADD_MENU( type ) \
{ \
    void* buf = pAllocator->Alloc(sizeof(type##Menu)); \
    type##Menu* menu = new(buf) type##Menu(); \
    mMenu.push_back( menu ); \
} \

    ADD_MENU( Reset )
    ADD_MENU( ScreenCaptureUpper )
#ifdef NW_TEST_LYT_VRAM
    ADD_MENU( TexMemArea )
#endif
    ADD_MENU( ZCompare )
    ADD_MENU( Fovy )
    ADD_MENU( Near )
    ADD_MENU( Far )
    ADD_MENU( MinNear )
    ADD_MENU( Stereo_BasePlane )
    ADD_MENU( Stereo_Factor )
    ADD_MENU( Stereo_RealSwitch )
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
MenuSubMode::~MenuSubMode()
{
    nw::os::IAllocator* pAllocator = Memory_GetAllocator();

    NW_FOREACH(IMenu* menu, mMenu)
    {
        menu->~IMenu();
        pAllocator->Free( menu );
    }

    mMenu.clear();
}


}   /* namespace g2d */
}   /* namespace vw  */
