/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#include <nn.h>
#include <viewer/common.h>
#include <viewer/g2d.h>
#include <protocol/ProtocolLayout.h>
#include <viewer/system.h>

#include "../include/ProtocolExecCTR.h"
#include "../include/Module.h"
#include "../include/Resources.h"

#include <cstring>

using namespace NW4R::ProtocolLayout;

namespace vw {
namespace g2d {

class ProtocolExecCTRImpl : public ProtocolExecCTR
{
    struct Packet
    {
        int size;
        u8 request[4];
    };

public:
    static const s16 CHANNEL = 0x07;
    static const u32 BODY_BUFFER_SIZE = 1024;

    ProtocolExecCTRImpl()
      : ProtocolExecCTR()
    {}

    virtual void Initialize()
    {
        mConnected = false;
        mProtocolClear = false;

        vw::system::Stream_CreateChannel(CHANNEL, BODY_BUFFER_SIZE);
    }

    virtual void Finalize()
    {
    }

    //
    //
    //
    virtual void Main()
    {
        if (vw::system::Stream_Polling(CHANNEL))
        {
            system::App::GetInstance().SetFocusID(system::G2D_ID);
        }

        while (vw::system::Stream_IsConnected(CHANNEL))
        {
            if (!mConnected)
            {
                mConnected = true;
                mProtocolClear = false;
            }

            Packet packet;

            // Read the header.
            {
                s32 readSize = this->Read(&packet, sizeof(packet));
                if (readSize == 0)
                {
                    return;
                }

                if (readSize == -1)
                {
                    NW_DEV_LOG("Connection lost.\n");
                    mConnected = false;
                    return;
                }
            }

            int request =
                packet.request[0] << 24 |
                packet.request[1] << 16 |
                packet.request[2] << 8 |
                packet.request[3] << 0;

            if (mProtocolClear)
            {
                if (request != 'idnt')
                {
                    continue;
                }

                mProtocolClear = false;
            }

            ModuleG2dImpl& module = ModuleG2dImpl::GetInstance();

            switch (request)
            {
            case 'idnt':
                {
                    static const wchar_t* ident = L"SimpleLayoutViewer";
                    const int idSize = sizeof(wchar_t) * 18;
                    int result = 0;
                    int size = sizeof(packet) + sizeof(result) + idSize;
                    packet.size = size;
                    this->Write(&packet, sizeof(packet));
                    this->Write(&result, sizeof(result));
                    this->Write(ident, idSize);
                }
                break;

            case 'load':
            {
                // Response packet
                this->Write(&packet, sizeof(packet));

                // Load the layout archive
                (void) module.Load(NULL, NULL, 0);

                // Make G2D mode active
                vw::system::App::GetInstance().SetFocusID(G2D_ID);
            }
            break;

            case 'uppr':
                // Response packet
                this->Write(&packet, sizeof(packet));
                (void) module.ToUpper();
                break;

            case 'lowr':
                // Response packet
                this->Write(&packet, sizeof(packet));
                (void) module.ToLower();
                break;

            default:
                // Ignore until ident is received.
                mProtocolClear = true;
            }
        }

        mConnected = false;
    }

protected:
    s32 Read(void* data, size_t size)
    {
        if (size == 0)
        {
            return 0;
        }

        u32 readableBytes = vw::system::Stream_GetReadableBytes(CHANNEL);
        if (readableBytes < size)
        {
            return 0;
        }

        u32 readSize = vw::system::Stream_Read(CHANNEL, data, size);

        if (!vw::system::Stream_IsConnected(CHANNEL))
        {
            return -1;
        }

        if (readSize != size)
        {
            return 0;
        }
        else
        {
            return s32(readSize);
        }
    }

    s32 Write(const void* data, size_t size)
    {
        if (size == 0)
        {
            return 0;
        }

        u32 writeSize = vw::system::Stream_Write(CHANNEL, data, size);

        if (!vw::system::Stream_IsConnected(CHANNEL))
        {
            return -1;
        }

        if (writeSize != size)
        {
            return 0;
        }
        else
        {
            return s32(writeSize);
        }
    }

    bool mConnected;
    bool mProtocolClear;
#if defined(NW_PLATFORM_CTR)
    u8 padding[2];
#endif
};

ProtocolExecCTR& ProtocolExecCTR::GetInstance()
{
    static ProtocolExecCTRImpl instance;
    return instance;
}

} // namespace g2d
} // namespace vw
