/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#include <nw/lyt/lyt_DrawInfo.h>
#include <nw/viewer/lyt/LayoutAccess.h>
#include "../include/ResLayout.h"
#include "../include/Resources.h"

namespace nw  {
namespace viewer {
namespace lyt {

namespace res = vw::g2d::res;
using vw::g2d::AnimCtrl;
using vw::g2d::AnimManager;
using vw::g2d::ResLayout;

/*---------------------------------------------------------------------------*
  Name:         GetNextSibling_
  Description:  Gets the next sibling pane.
  Arguments:    pPane   Pointer to the current pane.
  Returns:      Pointer to the next sibling pane.
 *---------------------------------------------------------------------------*/
static nw::lyt::Pane*
GetNextSibling_( nw::lyt::Pane* pPane )
{
    NW_NULL_ASSERT( pPane );

    nw::lyt::Pane* pParent = pPane->GetParent();

    if ( ! pParent )
    {
        return NULL;
    }

    nw::lyt::PaneList::Iterator it = pParent->GetChildList().GetBeginIter();
    while ( it != pParent->GetChildList().GetEndIter() )
    {
        if ( &(*it) == pPane )
        {
            ++it;
            if ( it != pParent->GetChildList().GetEndIter() )
            {
                return &(*it);
            }
            else
            {
                return NULL;
            }
        }
        ++it;
    }
    return NULL;
}

/*---------------------------------------------------------------------------*
  Name:         GetChild_
  Description:  Gets the idx-th child pane.
  Arguments:    pPane   Pointer to the current pane.
                idx     Index for the child pane to get.
  Returns:      Pointer to the idx-th child pane
 *---------------------------------------------------------------------------*/
static nw::lyt::Pane*
GetChild_( nw::lyt::Pane* pPane, u32 idx )
{
    NW_NULL_ASSERT( pPane );

    nw::lyt::PaneList::Iterator it = pPane->GetChildList().GetBeginIter();
    u32 i = 0;

    while ( true )
    {
        if ( it == pPane->GetChildList().GetEndIter() )
        {
            return NULL;
        }
        if ( i == idx )
        {
            return &(*it);
        }
        ++it;
        ++i;
    }
}


/*---------------------------------------------------------------------------*
  Name:         GetNextPane_
  Description:  Gets the next pane.
  Arguments:    pPane   Pointer to the current pane.
  Returns:      Pointer to the next pane.
 *---------------------------------------------------------------------------*/
static nw::lyt::Pane*
GetNextPane_( nw::lyt::Pane* pPane )
{
    nw::lyt::Pane* p;

    p = GetChild_( pPane, 0 );
    if ( p )
    {
        return p;
    }

    do
    {
        p = GetNextSibling_( pPane );
        if ( p )
        {
            return p;
        }
        pPane = pPane->GetParent();
    } while ( pPane );
    return NULL;
}

#if 1 // Temporary workaround until the const version of GetPaneList is available.
/*---------------------------------------------------------------------------*
   Temporary workaround until the const version of GetPaneList is available.
 *---------------------------------------------------------------------------*/
static const nw::lyt::PaneList&
GetConstChildList_( const nw::lyt::Pane& pane )
{
    nw::lyt::Pane& ncPane = const_cast<nw::lyt::Pane&>( pane );
    return ncPane.GetChildList();
}
#endif


const nw::lyt::Layout&
LayoutInfo::ref_lyt() const
{
    NW_ASSERT( this->IsValid() );
    return *( ref().GetLayout() );
}

const nw::lyt::Layout*
LayoutInfo::ptr_lyt() const
{
    return ref().GetLayout();
}


/*---------------------------------------------------------------------------*
  Name:         LayoutInfo::GetLayoutNum

  Description:  Gets the number of layouts.

  Arguments:    None.

  Returns:      Number of loaded layouts.
 *---------------------------------------------------------------------------*/
/* static */ u32
LayoutInfo::GetLayoutNum()
{
    return (res::GetResLayout().IsValid()) ? 1U : 0U;
}

/*---------------------------------------------------------------------------*
  Name:         LayoutInfo::GetLayout

  Description:  Gets the layout information.

  Arguments:    id      Layout ID specification

  Returns:      Layout information
 *---------------------------------------------------------------------------*/
/* static */ LayoutInfo
LayoutInfo::GetLayout( u32 id )
{
    if ( res::GetResLayout().IsValid() &&
         res::GetResLayout().GetID() == id )
    {
        return LayoutInfo( &res::GetResLayout() );
    }
    else
    {
        return LayoutInfo( NULL );
    }
}

/*---------------------------------------------------------------------------*
  Name:         LayoutInfo::GetFirstLaytou

  Description:  Gets the first layout information.

  Arguments:    None.

  Returns:      Layout information
 *---------------------------------------------------------------------------*/
/* static */ LayoutInfo
LayoutInfo::GetFirstLayout()
{
    return LayoutInfo( &res::GetResLayout() );
}


/*---------------------------------------------------------------------------*
  Name:         LayoutInfo::GetName

  Description:  Gets the layout resource name.

  Arguments:    None.

  Returns:      Resource name
 *---------------------------------------------------------------------------*/
const char*
LayoutInfo::GetName() const
{
    return ref().GetName();
}


/*---------------------------------------------------------------------------*
  Name:         LayoutInfo::GetID

  Description:  Gets the layout's ID.

  Arguments:    None.

  Returns:      Layout ID
 *---------------------------------------------------------------------------*/
u32
LayoutInfo::GetID() const
{
    return ref().GetID();
}


/*---------------------------------------------------------------------------*
  Name:         LayoutInfo::GetLayoutSize

  Description:  Gets the layout size information.

  Arguments:    None.

  Returns:      Size information
 *---------------------------------------------------------------------------*/
const nw::lyt::Size&
LayoutInfo::GetLayoutSize() const
{
    return ref_lyt().GetLayoutSize();
}

/*---------------------------------------------------------------------------*
  Name:         LayoutInfo::GetLayoutRect

  Description:  Gets the layout's rectangular information.

  Arguments:    None.

  Returns:
 *---------------------------------------------------------------------------*/
const nw::ut::Rect
LayoutInfo::GetLayoutRect() const
{
    return ref_lyt().GetLayoutRect();
}

/*---------------------------------------------------------------------------*
  Name:         LayoutInfo::IsValid

  Description:  Determines if the layout is valid.

  Arguments:    None.

  Returns:
 *---------------------------------------------------------------------------*/
bool
LayoutInfo::IsValid() const
{
    return (mpLayout && mpLayout->IsValid());
}

/*---------------------------------------------------------------------------*
  Name:         LayoutInfo::GetNext

  Description:  Gets the next layout.

  Arguments:    None.

  Returns:
 *---------------------------------------------------------------------------*/
LayoutInfo
LayoutInfo::GetNext() const
{
    return LayoutInfo( NULL );
}

/*---------------------------------------------------------------------------*
  Name:         LayoutInfo::GetRootPaneInfo

  Description:  Gets the root pane information.

  Arguments:    None.

  Returns:
 *---------------------------------------------------------------------------*/
PaneInfo
LayoutInfo::GetRootPaneInfo() const
{
    return PaneInfo( ref_lyt().GetRootPane() );
}

/*---------------------------------------------------------------------------*
  Name:         LayoutInfo::GetPaneInfo

  Description:  Gets the pane information.

  Arguments:    name

  Returns:
 *---------------------------------------------------------------------------*/
PaneInfo
LayoutInfo::GetPaneInfo( const char* name ) const
{
    nw::lyt::Pane* pPane = ref_lyt().GetRootPane();

    while ( pPane )
    {
        if ( std::strcmp( name, pPane->GetName() ) == 0 )
        {
            return PaneInfo( pPane );
        }
        pPane = GetNextPane_( pPane );
    }
    return PaneInfo( NULL );
}


/*---------------------------------------------------------------------------*
  Name:         LayoutInfo::GetPaneInfo

  Description:  Gets the pane information.

  Arguments:    idx

  Returns:
 *---------------------------------------------------------------------------*/
PaneInfo
LayoutInfo::GetPaneInfo( u32 idx ) const
{
    nw::lyt::Pane* pPane = ref_lyt().GetRootPane();

    for ( u32 i = 0; i < idx; i++ )
    {
        pPane = GetNextPane_( pPane );
        if ( pPane == NULL )
        {
            break;
        }
    }

    return PaneInfo( pPane );
}


/*---------------------------------------------------------------------------*
  Name:         LayoutInfo::GetPaneNum

  Description:  Gets the total number of panes in the tree.

  Arguments:    None.

  Returns:      Total number of panes.
 *---------------------------------------------------------------------------*/
s32
LayoutInfo::GetPaneNum() const
{
    nw::lyt::Pane* pPane    = ref_lyt().GetRootPane();
    s32              paneNum = 0;

    while ( pPane )
    {
        ++paneNum;
        pPane = GetNextPane_( pPane );
    }
    return paneNum;
}


/*---------------------------------------------------------------------------*
  Name:         LayoutInfo::GetAnimNum

  Description:  Gets the number of animations.

  Arguments:    None.

  Returns:
 *---------------------------------------------------------------------------*/
u32
LayoutInfo::GetAnimNum() const
{
    return ref().GetAnimManager().GetAnimNum();
}

/*---------------------------------------------------------------------------*
  Name:         LayoutInfo::GetAnim

  Description:  Gets the animation information.

  Arguments:    idx

  Returns:
 *---------------------------------------------------------------------------*/
AnimInfo
LayoutInfo::GetAnim( u32 idx ) const
{
    return AnimInfo( ptr(), idx );
}

/*---------------------------------------------------------------------------*
  Name:         LayoutInfo::GetAnim

  Description:  Gets the animation information.

  Arguments:    name

  Returns:
 *---------------------------------------------------------------------------*/
AnimInfo
LayoutInfo::GetAnim( const char* name ) const
{
    const AnimManager& animManager = ref().GetAnimManager();

    for ( s32 i = 0; i < animManager.GetAnimNum(); i++ )
    {
        const AnimCtrl* pAnimCtrl = animManager.GetAnimCtrl(i);

        if ( std::strcmp( name, pAnimCtrl->GetName() ) == 0 )
        {
            return AnimInfo( ptr(), u32( i ) );
        }
    }
    return AnimInfo( NULL, 0 );
}

/*---------------------------------------------------------------------------*
  Name:         LayoutInfo::GetFirstAnim

  Description:  Gets the first animation.

  Arguments:    None.

  Returns:
 *---------------------------------------------------------------------------*/
AnimInfo
LayoutInfo::GetFirstAnim() const
{
    return this->GetAnim( u32(0) );
}

/*---------------------------------------------------------------------------*
  Name:         LayoutInfo::GetCurrentAnim

  Description:  Gets the information for the animation currently being displayed.

  Arguments:    None.

  Returns:
 *---------------------------------------------------------------------------*/
AnimInfo
LayoutInfo::GetCurrentAnim() const
{
    return AnimInfo( ptr(), u32( ref().GetAnimManager().GetAnimIdx() ) );
}

/*---------------------------------------------------------------------------*
  Name:         LayoutInfo::SetCurrentAnim

  Description:  Specifies the animation to be displayed.

  Arguments:    name

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
LayoutInfo::SetCurrentAnim( const char* name )
{
    AnimManager& animManager = ref().GetAnimManager();

    for ( s32 i = 0; i < animManager.GetAnimNum(); i++ )
    {
        AnimCtrl* pAnimCtrl = animManager.GetAnimCtrl(i);

        if ( std::strcmp( name, pAnimCtrl->GetName() ) == 0 )
        {
            animManager.SetAnim(i);
            return;
        }
    }
}

/*---------------------------------------------------------------------------*
  Name:         LayoutInfo::SetCurrentAnim

  Description:  Specifies the animation to be displayed.

  Arguments:    idx

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
LayoutInfo::SetCurrentAnim( u32 idx )
{
    ref().GetAnimManager().SetAnim( s32( idx ) );
}


/*---------------------------------------------------------------------------*
  Name:         LayoutInfo::HasAnimation

  Description:  Gets whether it is an animation.

  Arguments:    None.

  Returns:
 *---------------------------------------------------------------------------*/
bool
LayoutInfo::HasAnimation() const
{
    return ref().GetAnimManager().HasAnim();
}

/*---------------------------------------------------------------------------*
  Name:         LayoutInfo::GetCurrentFrame

  Description:  Gets the frame of the current animation.

  Arguments:    None.

  Returns:
 *---------------------------------------------------------------------------*/
f32
LayoutInfo::GetCurrentFrame() const
{
    return ref().GetAnimManager().GetCurrentFrame();
}

/*---------------------------------------------------------------------------*
  Name:         LayoutInfo::SetCurrentFrame

  Description:  Sets the frame of the current animation.

  Arguments:    frame   Specifies the frame of the current animation.

  Returns:
 *---------------------------------------------------------------------------*/
void
LayoutInfo::SetCurrentFrame(f32 frame)
{
    return ref().GetAnimManager().SetFrame(frame);
}

/*---------------------------------------------------------------------------*
  Name:         LayoutInfo::GetTotalFrame

  Description:  Gets the current total frames.

  Arguments:    None.

  Returns:
 *---------------------------------------------------------------------------*/
f32
LayoutInfo::GetTotalFrame() const
{
    return ref().GetAnimManager().GetAnimCtrl()->GetMaxFrame();
}

/*---------------------------------------------------------------------------*
  Name:         LayoutInfo::GetAbsoluteFrame

  Description:  Gets the current animation's absolute time.

  Arguments:    None.

  Returns:
 *---------------------------------------------------------------------------*/
f32
LayoutInfo::GetAbsoluteFrame() const
{
    return ref().GetAnimManager().GetAbsoluteFrame();
}


/*---------------------------------------------------------------------------*
  Name:         PaneInfo::GetName

  Description:  Gets the pane name.

  Arguments:    None.

  Returns:
 *---------------------------------------------------------------------------*/
const char*
PaneInfo::GetName() const
{
    return ref().GetName();
}

/*---------------------------------------------------------------------------*
  Name:         PaneInfo::GetMtx

  Description:  Gets the local matrix.

  Arguments:    None.

  Returns:
 *---------------------------------------------------------------------------*/
const nw::math::MTX34&
PaneInfo::GetMtx() const
{
    return ref().GetMtx();
}

/*---------------------------------------------------------------------------*
  Name:         PaneInfo::GetGlobalMtx

  Description:  Gets the global matrix.

  Arguments:    None.

  Returns:
 *---------------------------------------------------------------------------*/
const nw::math::MTX34&
PaneInfo::GetGlobalMtx() const
{
    return ref().GetGlobalMtx();
}

/*---------------------------------------------------------------------------*
  Name:         PaneInfo::GetPaneRect

  Description:  Gets the rectangular information.

  Arguments:    None.

  Returns:
 *---------------------------------------------------------------------------*/
const nw::ut::Rect
PaneInfo::GetPaneRect() const
{
    return ref().GetPaneRect();
}

/*---------------------------------------------------------------------------*
  Name:         PaneInfo::GetSize

  Description:  Gets the pane size.

  Arguments:    None.

  Returns:
 *---------------------------------------------------------------------------*/
const nw::lyt::Size&
PaneInfo::GetSize() const
{
    return ref().GetSize();
}

/*---------------------------------------------------------------------------*
  Name:         PaneInfo::GetSRT

  Description:  Gets the pane's SRT information.

  Arguments:    pScale  Areas for obtaining scale information
                pRot    Area for obtaining rotate information
                pTrans  Area for obtaining translate information

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
PaneInfo::GetSRT( nw::math::VEC2* pScale, nw::math::VEC3* pRot, nw::math::VEC3* pTrans ) const
{
    if ( pScale )
    {
        *pScale = ref().GetScale();
    }

    if ( pTrans )
    {
        *pTrans = ref().GetTranslate();
    }

    if ( pRot )
    {
        *pRot = ref(). GetRotate();
    }
}


/*---------------------------------------------------------------------------*
  Name:         PaneInfo::IsVisible

  Description:  Gets the pane's visibility information.

  Arguments:    None.

  Returns:      Returns the pane's visibility information.
 *---------------------------------------------------------------------------*/
bool
PaneInfo::IsVisible()    const
{
    return ref().IsVisible();
}


/*---------------------------------------------------------------------------*
  Name:         PaneInfo::GetParent

  Description:  Gets the parent pane.

  Arguments:    None.

  Returns:      Parent pane information
 *---------------------------------------------------------------------------*/
PaneInfo
PaneInfo::GetParent() const
{
    return PaneInfo( ref().GetParent() );
}

/*---------------------------------------------------------------------------*
  Name:         PaneInfo::GetChild

  Description:  Gets the idx-th child pane.

  Arguments:    idx  Child pane index specification

  Returns:      Child pane information
 *---------------------------------------------------------------------------*/
PaneInfo
PaneInfo::GetChild( u32 idx ) const
{
    return PaneInfo( GetChild_( mpPane, idx ) );
}

/*---------------------------------------------------------------------------*
  Name:         PaneInfo::GetChildNum

  Description:  Gets the number of child panes.

  Arguments:    None.

  Returns:      Number of child panes
 *---------------------------------------------------------------------------*/
u32
PaneInfo::GetChildNum() const
{
#if 1 // Temporary workaround until the const version of GetPaneList is available.
    const nw::lyt::PaneList& list = GetConstChildList_( ref() );
    nw::lyt::PaneList::ConstIterator it = list.GetBeginIter();
#else
    nw::lyt::PaneList::ConstIterator it = ref().GetChildList().GetBeginIter();
#endif
    u32 num = 0;

#if 1 // Temporary workaround until the const version of GetPaneList is available.
    while ( it != list.GetEndIter() )
#else
    while ( it != ref().GetChildList().GetEndIter() )
#endif
    {
        ++it;
        ++num;
    }
    return num;
}


const AnimCtrl&
AnimInfo::ref() const
{
    return *ref_lyt().GetAnimManager().GetAnimCtrl(mIdx);
}

/*---------------------------------------------------------------------------*
  Name:         PaneInfo::GetNext

  Description:  Gets the next pane in the layout according to certain rules.
                 Can make rounds to all panes by traveling in sequence with the root pane.

  Arguments:    None.

  Returns:      Next pane information
 *---------------------------------------------------------------------------*/
PaneInfo
PaneInfo::GetNext() const
{
    return PaneInfo( GetNextPane_( mpPane ) );
}


/*---------------------------------------------------------------------------*
  Name:         AnimInfo::GetName

  Description:  Gets the animation name.

  Arguments:    None.

  Returns:      Animation name
 *---------------------------------------------------------------------------*/
const char*
AnimInfo::GetName() const
{
    return ref().GetName();
}

/*---------------------------------------------------------------------------*
  Name:         AnimInfo::IsValid

  Description:  Gets whether this animation information is valid.

  Arguments:    None.

  Returns:      If the animation is valid. true.
                If invalid, false.
 *---------------------------------------------------------------------------*/
bool
AnimInfo::IsValid() const
{
    if ( (! mpLayout) || (! ptr_lyt()->IsValid()) )
    {
        return false;
    }
    if ( mIdx >= ref_lyt().GetAnimManager().GetAnimNum() )
    {
        return false;
    }
    return true;
}

/*---------------------------------------------------------------------------*
  Name:         AnimInfo::IsCurrent

  Description:  Gets whether the animation is the one currently being displayed.

  Arguments:    None.

  Returns:      If it is the one currently being displayed, true.
                If it is not, false.
 *---------------------------------------------------------------------------*/
bool
AnimInfo::IsCurrent() const
{
    return (ref_lyt().GetAnimManager().GetAnimIdx() == mIdx)? true : false;
}


/*---------------------------------------------------------------------------*
  Name:         AnimInfo::GetFrameMax

  Description:  Gets the maximum number frames of this animation.

  Arguments:    None.

  Returns:      Maximum number of animation frames.
 *---------------------------------------------------------------------------*/
f32
AnimInfo::GetFrameMax() const
{
    return ref().GetMaxFrame();
}

/*---------------------------------------------------------------------------*
  Name:         AnimInfo::IsLoop

  Description:  Gets whether it is the end loop.

  Arguments:    None.

  Returns:      If it is an animation that loops at the end, true.
                If it is not, false.
 *---------------------------------------------------------------------------*/
bool
AnimInfo::IsLoop() const
{
    return ref().IsLoop();
}

/*---------------------------------------------------------------------------*
  Name:         AnimInfo::GetNext

  Description:  Gets the next index's animation information.

  Arguments:    None.

  Returns:      Next animation information
 *---------------------------------------------------------------------------*/
AnimInfo
AnimInfo::GetNext() const
{
    return AnimInfo( mpLayout, u32( mIdx + 1 ) );
}

/*---------------------------------------------------------------------------*
  Name:         AnimInfo::GetAnimResource

  Description:  Gets AnimResource.

  Arguments:    None.

  Returns:      References AnimResource.
 *---------------------------------------------------------------------------*/
const nw::lyt::AnimResource&
AnimInfo::GetAnimResource() const
{
    return ref().GetAnimResource();
}


}   /* namespace lyt */
}   /* namespace viewer  */
}   /* namespace nw */



