/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#ifndef NW4C_VIEWER_G3D_ANIM_CONTROLLER_H_
#define NW4C_VIEWER_G3D_ANIM_CONTROLLER_H_

#if NN_CURRENT_VERSION_NUMBER >= NN_VERSION_NUMBER(4,0,0,0)
#pragma diag_suppress 1301 // padding inserted in struct.
#endif
namespace vw {
namespace g3d {

class AnimSetLinker;

class IAnimController
{
public:
    // Frame advance.
    virtual void StepForward() = 0;
    // Frame reverse
    virtual void StepBackward() = 0;
    // Play/Pause
    virtual void PauseAndPlay() = 0;
    virtual void Pause() = 0;
    virtual void Play() = 0;
    // Update
    virtual void Update() = 0;
    // Increase/decrease playback speed
    virtual void SpeedUp(f32 step) = 0;
    // Stop (pause playback, move frame to 0)
    virtual void Stop() = 0;
    // Resets.
    virtual void Reset() = 0;
    // Set/get frame
    virtual void SetFrame(f32 frame) = 0;
    virtual f32 GetFrame() = 0;
    // Set/get playback rate
    virtual void SetStepFrame(f32 step) = 0;
    virtual f32 GetStepFrame() = 0;
    // Animation enable/disable switch
    virtual bool IsActive() = 0;
    // When the focus module is changed
    virtual void OnTargetModuleChanged(u16 id) = 0;
};

class AnimController : public IAnimController
{
public:
    // Frame advance.
    virtual void StepForward();
    // Frame reverse
    virtual void StepBackward();
    // Play/Pause
    virtual void PauseAndPlay();
    virtual void Pause();
    virtual void Play();
    // Update
    virtual void Update();
    // Increase/decrease playback speed
    virtual void SpeedUp(f32 step);
    // Stop (pause playback, move frame to 0)
    virtual void Stop();
    // Resets.
    virtual void Reset();
    // Set/get frame
    virtual void SetFrame(f32 frame);
    virtual f32 GetFrame();
    // Set/get playback rate
    virtual void SetStepFrame(f32 step);
    virtual f32 GetStepFrame(){ return m_StepFrame; }
    // Animation enable/disable switch
    virtual bool IsActive();
    virtual void OnTargetModuleChanged( u16 id ){}

    class Builder
    {
    public:
        Builder() : m_pNode(NULL) {}

        AnimController*
        Create(nw::os::IAllocator *pAllocator)
        {
            NW_NULL_ASSERT( m_pNode );
            void *buf = pAllocator->Alloc(sizeof(AnimController));
            if (buf == NULL) return NULL;
            return new(buf) AnimController( m_pNode, pAllocator );
        }

        Builder&
        Node( nw::gfx::SceneNode *node )
        {
            m_pNode = node;
            return *this;
        }

    private:
        nw::gfx::SceneNode* m_pNode;
    };

    void
    Destroy()
    {
        this->~AnimController();
        m_pAllocator->Free( this );
    }

    void
    SetStartFrame(f32 frame)
    {
        m_StartFrame = frame;
    }

    virtual AnimSetLinker*
    GetAnimSetLinker()
    {
        return m_pLinker;
    }

protected:
    // Linker
    vw::g3d::AnimSetLinker *m_pLinker;
    // Playback permission flag
    bool m_IsActive;
    // Playback rate
    f32 m_StepFrame;
    // Start frame
    f32 m_StartFrame;
    // Allocator
    nw::os::IAllocator *m_pAllocator;

    AnimController( nw::gfx::SceneNode *node, nw::os::IAllocator* pAllocator );

    ~AnimController();
};


}
}
#if NN_CURRENT_VERSION_NUMBER >= NN_VERSION_NUMBER(4,0,0,0)
#pragma diag_default 1301 // padding inserted in struct.
#endif

#endif
