/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#ifndef NW4R_VIEWER_G3D_MATERIAL_H_
#define NW4R_VIEWER_G3D_MATERIAL_H_

#include <dolphin.h>
#include <nw4r/ut/list.h>
#include "ActionSet.h"

namespace vw  {
namespace g3d {

// Set node visibility
void    PatchNodeVisibility( bool visible );

// Node billboard setting
void    PatchNodeBillboardMode( nw4r::g3d::ResNode::Billboard billboardMode );

// Material display face setting
void    PatchMatCullMode( GXCullMode mode );

// Set channel enable flag
void    PatchMatChanEnable( GXChannelID chan, GXBool    enable );

// Set channel's material color
void    PatchMatChanMatColor( GXChannelID chan, GXColor mat_color );

// Set channel's ambient color
void    PatchMatChanAmbColor( GXChannelID chan, GXColor amb_color );

// Set channel's ambient source
void    PatchMatChanCtrlAmbSrc( GXChannelID chan, GXColorSrc ambSrc );

// Set channel's material source
void    PatchMatChanCtrlMatSrc( GXChannelID chan, GXColorSrc matSrc );

// Set channel diffusion function
void    PatchMatChanDiffFn( GXChannelID chan, GXDiffuseFn diff_fn );

// Set channel attenuation function
void    PatchMatChanAttnFn( GXChannelID chan, GXAttnFn attn_fn );

// Set TEV color register
void    PatchMatTevColor( GXTevRegID id, GXColor color );
void    PatchMatTevColorS10( GXTevRegID id, GXColorS10 color );

// Set TEV constant register
void    PatchMatTevKColor( GXTevKColorID id, GXColor color );

// Set TEV swap table
void    PatchMatTevSwapModeTable( GXTevSwapSel   id,
                                  GXTevColorChan red,
                                  GXTevColorChan green,
                                  GXTevColorChan blue,
                                  GXTevColorChan alpha );

// Set TEV stages
void    PatchMatTevStages( u8 nStages );

// TEV order change command (excluding texture coordinates)
void    PatchMatTevOrder( GXTevStageID  stage,
                          GXTexCoordID  coord,
                          GXTexMapID    map,
                          GXChannelID   color );

// Set TEV swap mode
void    PatchMatTevSwapMode( GXTevStageID   stage,
                             GXTevSwapSel   ras_sel,
                             GXTevSwapSel   tex_sel );

typedef struct
{
    GXTevColorArg a;    // a element
    GXTevColorArg b;    // b element
    GXTevColorArg c;    // c element
    GXTevColorArg d;    // d element
}
MatTevColorIn;

typedef struct
{
    GXTevAlphaArg a;    // a element
    GXTevAlphaArg b;    // b element
    GXTevAlphaArg c;    // c element
    GXTevAlphaArg d;    // d element
}
MatTevAlphaIn;

typedef struct
{
    GXTevOp       op;      // TEV calculation
    GXTevBias     bias;    // TEV bias value
    GXTevScale    scale;   // TEV scale value
    GXBool        clamp;   // Clump.
    GXTevRegID    out_reg; // Output register
}
MatTevOp;

// TEV color calculation change command
void    PatchMatTevColorOp( GXTevStageID   stage,    // TEV stage
                            MatTevColorIn  *in,      // TEV input color parameter
                            MatTevOp       *op,      // TEV calculation parameter
                            GXTevKColorSel kCol );   // Set constant color register

// TEV alpha calculation change command
void    PatchMatTevAlphaOp( GXTevStageID   stage,    // TEV stage
                            MatTevAlphaIn  *in,      // TEV input alpha parameter
                            MatTevOp       *op,      // TEV calculation parameter
                            GXTevKAlphaSel kAlpha ); // Set constant alpha register

// Translucent/opaque flag change
void    PatchMatTransparencyMode( bool xlu );        // Translucent flag

// Alpha compare change
void    PatchMatPixAlphaCompare( GXCompare  comp0,   // Comparison subfunction0
                                 u8         ref0,    // Reference value for subfunction0
                                 GXAlphaOp  op,      // Operator for the combination of subfunction0 and subfunction1
                                 GXCompare  comp1,   // Comparison subfunction1
                                 u8         ref1 );  // Reference value for subfunction1

// Z compare change
void    PatchMatPixZCompare( GXBool     compare_enable,  // Z-comparison enable flag
                             GXCompare  func,            // Z-comparison condition
                             GXBool     update_enable,   // Z-buffer change enable flag
                             GXBool     before_tex   );  // Z-test timing flag

// Blend mode change
void    PatchMatPixBlendMode( GXBlendMode   type,        // Blending mode
                              GXBlendFactor src_factor,  // Source coefficient
                              GXBlendFactor dst_factor,  // Destination coefficient
                              GXLogicOp     op        ); // Blending

// Texture map information change
void    PatchMatTexMap( GXTexMapID      texMap,     // Texture map ID
                        u32             texID,      // Texture ID
                        u32             plttID,     // Palette ID
                        GXTexWrapMode   wrap_s,     // Wrap settings in the texture horizontal (s) direction
                        GXTexWrapMode   wrap_t,     // Wrap settings in the texture vertical (t) direction
                        GXBool          mipmap,     // Specify whether to use mipmap
                        GXTexFilter     min_filt,   // Filter used when reducing texels
                        GXTexFilter     mag_filt,   // Filter used when expanding texels
                        f32             lod_bias,   // Bias added to the calculated LOD value
                        GXBool          bias_clamp, // bias clamp settings
                        GXBool          do_edge_lod,// Flag to specify whether to calculate LOD using texel interpolation
                        GXAnisotropy    max_aniso );// Maximum anisotropic filter used

// Texture coordinate count change
void    PatchMatNumTexGens( u8 num );       // Number of texture coordinates to be generated

// Texture coordinate information change
void    PatchMatTexCoord( GXTexCoordID  dst_coord,  // Texture coordinate index
                          GXTexGenType  func,       // Function used to generate texture coordinates
                          GXTexGenSrc   src_param,  // Source attributes used to generate texture coordinates
                          u32           mtx  );     // Matrix used to generate texture coordinates

// Texture coordinate information change
void    PatchMatTexMatrix( u32          id,             // Texture matrix index
                           u8           mapping_method, // Texture mapping method
                           s8           ref_camera,     // Camera referenced with texture mapping
                           s8           ref_light,      // Light referenced with texture mapping
                           float        scale_s,        // Scale S-component
                           float        scale_t,        // Scale T-component
                           float        rotate,         // Rotation component
                           float        trans_s,        // Translation S-component
                           float        trans_t );      // Translation T-component


// Texture image change
void    PatchMatTexMapImage( GXTexMapID texMap,     // Texture map ID
                             u32        texID,      // Texture ID
                             u32        plttID );   // Palette ID


// Change number of indirect texture stages
void    PatchMatNumIndStages( u8 nStages );         // Number of indirect stages


// Change order of indirect texture
void    PatchMatIndTexOrder( GXIndTexStageID    indStage,   // indirect stages
                             GXTexCoordID       tex_coord,  // Texture coordinate ID
                             GXTexMapID         tex_map );  // Texture map

// Change scale factor of indirect texture
void    PatchMatIndTexCoordScale( GXIndTexStageID   indStage,   // indirect stages
                                    GXIndTexScale   scale_s,    // Scale factor for the S-coordinate
                                    GXIndTexScale   scale_t );  // Scale factor for the T-coordinate

// Indirect texture operation
void    PatchMatTevIndirect( GXTevStageID       tev_stage,   // TEV stage
                             GXIndTexStageID    ind_stage,   // indirect stages
                             GXIndTexFormat     format,      // Indirect color format
                             GXIndTexBiasSel    bias_sel,    // Whether to apply bias to the indirect offset
                             GXIndTexMtxID      matrix_sel,  // Which matrix and scale to multiply
                             GXIndTexWrap       wrap_s,      // Wrap coefficient to apply to the S component
                             GXIndTexWrap       wrap_t,      // Wrap coefficient to apply to the T component
                             GXBool             add_prev,    // Whether to add the texture coordinate results from the previous TEV stage
                             GXBool             utc_lod,     // Whether to use texture coordinates not changed in mipmap LOD calculations
                             GXIndTexAlphaSel   alpha_sel ); // Which offset elements to set to indirect "bump alpha"

// Indirect matrix SRT change
void    PatchMatIndMtxSRT( GXIndTexMtxID     mtxID,         // Indirect matrix index
                           float             scale_s,       // Scale S-component
                           float             scale_t,       // Scale T-component
                           float             rotate,        // Rotation component
                           float             trans_s,       // Translation S-component
                           float             trans_t   );   // Translation T-component





// Set matrix in the model
void    PatchEnvModelSRT( Vec* scale, Vec* rot, Vec* trans, bool visibility, u32 ref_model, s32 ref_node );
void    PatchEnvModelActionSet( VwResActSet::ActSetElement* list, u32 actNum, bool loop );


void    PatchMatTexPlttInfo( GXTexMapID texMap, u32 texID, u32 plttID );
void    PatchMatTexPlttInfoNum( u32 num );

}   /* namespace g3d */
}   /* namespace vw  */


/* NW4R_VIEWER_G3D_MATERIAL_H_ */
#endif
