/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#ifndef NW4C_VIEWER_G3D_RESOURCE_MANAGER_H_
#define NW4C_VIEWER_G3D_RESOURCE_MANAGER_H_

#include <nw/gfx.h>
#include <viewer/g3d/Protocol4C.h>

#if NN_CURRENT_VERSION_NUMBER >= NN_VERSION_NUMBER(4,0,0,0)
#pragma diag_suppress 1301 // padding inserted in struct.
#endif
namespace vw {
namespace g3d {

bool IsContained(nw::gfx::ResGraphicsFile file, void* res_ptr);

class ResourceSet
{
friend class ResourceManager;
#define RES_USE_DATA_SIZE 4
public:
    enum BinaryFileType // Binary file type
    {
        Type_Model = Protocol4C::BinaryDataType_Bcmdl,   // model
        Type_Texture = Protocol4C::BinaryDataType_Bctex,   // Texture
        Type_LookupTableSet = Protocol4C::BinaryDataType_Bclts,  // Lookup table
        Type_SkeletalAnim = Protocol4C::BinaryDataType_Bcskla, // Skeletal animations
        Type_MaterialAnim = Protocol4C::BinaryDataType_Bcmata, // Material Animations
        Type_SceneEnvironmentSetting = Protocol4C::BinaryDataType_Bcscn,   // Scene
        Type_Camera = Protocol4C::BinaryDataType_Bccam,   // Camera
        Type_Light= Protocol4C::BinaryDataType_Bclgt,   // Lights
        Type_Fog = Protocol4C::BinaryDataType_Bcfog,   // Fog
        Type_VisibilityAnim = Protocol4C::BinaryDataType_Bcmdla, // Visibility animation
        Type_Particle = Protocol4C::BinaryDataType_Bcptl,   // Particle
        Type_AnimSnd,            // Special type for AnimationSound
        Type_UserData,           // Special type for replacing UserData
        Type_Packed            // Integrated data for CS or button loading
    };

    ResourceSet(
        nw::ut::MoveArray<u8>& buffer,
        BinaryFileType type,
        unsigned int hashID,
        int channelID
    );
    ResourceSet(const ResourceSet& resourceSet);
    ~ResourceSet();

    nw::gfx::ResGraphicsFile GetResource();
    BinaryFileType GetBinaryType() { return m_type; }
    unsigned int GetHashID() { return m_hashID; }
    int GetChannelID() { return m_channelID; }
    float* GetUserData() { return m_userData; }
    bool IsContained(void* res_ptr);

private:
    ResourceSet();

    // Link node
    nw::ut::LinkListNode m_Link;
    // Pointer to start of file itself
    nw::ut::MoveArray<u8> m_buffer;
    // File type
    BinaryFileType m_type;
    // File hash ID
    unsigned int m_hashID;
    // Channel number generated by resource
    int m_channelID;
    // Multipurpose user data
    float m_userData[ RES_USE_DATA_SIZE ];
};



const char* GetResourceSetTypeName(ResourceSet::BinaryFileType type);



class ResourceManager
{
public:
    typedef nw::ut::LinkList<ResourceSet, offsetof(ResourceSet, m_Link)> ResourceList;
    static ResourceManager* GetInstance();

    static void Initialize();
    static void Finalize();

    void AddResource(ResourceSet& resourceSet);
    void DeleteResource(const ResourceSet::BinaryFileType type, const unsigned int hashID);
    void ClearResources(const int channelID);
    void ClearResources();

    void LoadBinaryData(
        const ResourceSet::BinaryFileType type,
        const unsigned int hashID,
        const int channelID,
        const void *filename,
        const bool isWChar = true
    );

    ResourceList::iterator Find(const ResourceSet::BinaryFileType type, const unsigned int hashID);
    ResourceList::iterator Find(void* ptr);

    ResourceList::iterator GetResourcesBegin();
    ResourceList::iterator GetResourcesEnd();
    int GetResourcesCount() { return m_Resources.size(); }

    ResourceList::iterator GetFirstResourceByChannel(const int channelID);
    ResourceList::iterator DeleteResource(ResourceList::iterator);

    nw::gfx::res::ResImageLookupTable GetDefaultLookupTable();
    nw::gfx::res::ResTexture GetDefaultTexture();
    nw::gfx::res::ResShader  GetDefaultShader();
    nw::gfx::res::ResShader  GetDefaultParticleShader();

    void DumpMemory();

private:
    ResourceManager();
    ~ResourceManager();

    void LoadDefaultLookupTableSet();
    void LoadDefaultTexture();
    void LoadDefaultShader();
    void LoadDefaultParticleShader();
    void SetupScene();

    ResourceList m_Resources;
    ResourceSet m_DefaultLookupTableSet;
    ResourceSet m_DefaultTexture;
    ResourceSet m_DefaultShader;
    ResourceSet m_DefaultParticleShader;
    nw::os::IAllocator* m_pAllocator;
    nw::os::IAllocator* m_pDeviceAllocator;

    static ResourceManager* m_pInstance;
};

}
}
#if NN_CURRENT_VERSION_NUMBER >= NN_VERSION_NUMBER(4,0,0,0)
#pragma diag_default 1301 // padding inserted in struct.
#endif

#endif
