/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#ifndef NW4C_VIEWER_G3D_SCENE_MANAGER_H_
#define NW4C_VIEWER_G3D_SCENE_MANAGER_H_

#include "SceneElem.h"
#include "AnimElem.h"
#include "ParticleSceneElem.h"

#if NN_CURRENT_VERSION_NUMBER >= NN_VERSION_NUMBER(4,0,0,0)
#pragma diag_suppress 1301 // padding inserted in struct.
#pragma diag_suppress 2530 // padding added to end of struct.
#endif
namespace vw {
namespace g3d {

class ResourceSet;

#define DECL_RES_LINK_LIST( type, resType, name ) \
    typedef nw::ut::LinkList<type, offsetof(type, m_Link)> name##List; \
    private: name##List m_##name##Array; \
    public: name##List::iterator \
    Get##name##sBegin() { return m_##name##Array.begin(); } \
    public: name##List::iterator \
    Get##name##sEnd() { return m_##name##Array.end(); } \
    public: u32 \
    Get##name##sCount() { return m_##name##Array.size(); } \
    public: name##List::iterator Find##name( unsigned int hashID, unsigned int index = 0 ); \
    public: name##List::iterator Find##name( resType object ); \
    public: resType FindRes##name( unsigned int hashID, unsigned int index = 0 ); \

#define DECL_SCENE_LINK_LIST( type, name ) \
    DECL_RES_LINK_LIST( type, nw::gfx::res::Res##name, name ) \
    public: name##List::iterator Find##name( nw::gfx::name* object ); \
    public: nw::gfx::name* Find##name##Node( unsigned int hashID, unsigned int index = 0 );

class SceneManager
{
public:
    struct Description
    {
        s32 maxSceneNodes;  //
        s32 maxEmission;    //

        Description() :
            maxSceneNodes(32),
            maxEmission(3500)
        {}
    };

    // Function to get instance
    static SceneManager* GetInstance();

    // Instance initialization, finalization
    static void Initialize();
    static void Finalize();

    // Resource file initialization
    void InitializeResource();

    // Add/delete resource file
    void AddElem(ResourceSet* resourceSet, bool solveAnimReferences = false);
    void DeleteElem(ResourceSet* resourceSet);

    // Initialize/update context
    void InitContext(const Description& description);
    void UpdateSceneContext();

    // Function to get resource
    nw::gfx::TransformNode* GetSceneRoot();
    nw::gfx::SceneContext* GetSceneContext();
    nw::gfx::ParticleContext* GetParticleContext();

    // Declarations for each resource array
    DECL_SCENE_LINK_LIST( ModelElem, Model )
    DECL_RES_LINK_LIST( TextureElem, nw::gfx::res::ResTexture, Texture )
    DECL_RES_LINK_LIST( LookupTableSetElem, nw::gfx::res::ResLookupTableSet, LookupTableSet )
    DECL_RES_LINK_LIST( AnimElem, nw::anim::res::ResAnim, Anim )

    DECL_SCENE_LINK_LIST( CameraElem, Camera )
    DECL_SCENE_LINK_LIST( LightElem, Light )
    DECL_SCENE_LINK_LIST( FogElem, Fog )
    DECL_SCENE_LINK_LIST( SceneEnvironmentSettingElem, SceneEnvironmentSetting )

    DECL_SCENE_LINK_LIST( ParticleModelElem, ParticleModel )
    DECL_SCENE_LINK_LIST( ParticleEmitterElem, ParticleEmitter )

    // Function to get default camera and lights
    CameraElem* GetDefaultCamera(){ return m_pDefaultCamera; }
    LightElem* GetDefaultAmbientLight(){ return m_pDefaultAmbientLight; }
    LightElem* GetDefaultFragmentLight(){ return m_pDefaultFragmentLight; }
    nw::gfx::LightSet* GetDefaultLightSet(){ return m_pDefaultLightSet; }

    // Function to get sky light set/fog
    nw::gfx::LightSet* GetVoidLightSet(){ return m_pVoidLightSet; }
    nw::gfx::Fog* GetVoidFog() { return m_pVoidFog; }

    // Function for camera operations external to the Viewer
    void AddExternalCamera(nw::gfx::Camera* camera);
    void RemoveExternalCamera(nw::gfx::Camera* camera);
    nw::ut::MoveArray<CameraElem*>& GetExternalCameraArray(){ return m_ExternalCameraArray; }

    // Function for light operations external to the Viewer
    void AddExternalLight(nw::gfx::Light* light);
    void RemoveExternalLight(nw::gfx::Light* light);
    nw::ut::MoveArray<LightElem*>& GetExternalLightArray(){ return m_ExternalLightArray; }

    bool GetForceUseCameraWScale() { return m_ForceUseCameraWScale; }
    void SetForceUseCameraWScale( const bool flag );
    void SetCameraWScale(nw::gfx::Camera* objCamera);

private:
    SceneManager();
    ~SceneManager();
    void CheckError( nw::gfx::SceneNode* sceneNode );
    void CheckError( nw::ut::MoveArray<nw::gfx::SceneNode*>& sceneNodeArray );
    void SetupResource( nw::ut::MoveArray<nw::gfx::SceneNode*>& sceneNodeArray );
    void AddDefaultLight();
    void AddDefaultCamera();
    void AddDefaultFog();

    nw::gfx::TransformNode* m_pSceneRoot;
    nw::gfx::SceneInitializer* m_pSceneInitializer;
    nw::gfx::SceneTraverser* m_pSceneTraverser;
    nw::gfx::SceneContext* m_pSceneContext;
    nw::gfx::ParticleContext* m_pParticleContext;
    LightElem* m_pDefaultAmbientLight;
    LightElem* m_pDefaultFragmentLight;
    nw::gfx::LightSet* m_pDefaultLightSet;
    nw::gfx::LightSet* m_pVoidLightSet;
    nw::gfx::Fog* m_pVoidFog;
    CameraElem* m_pDefaultCamera;
    nw::ut::MoveArray<CameraElem*> m_ExternalCameraArray;
    nw::ut::MoveArray<LightElem*> m_ExternalLightArray;
    nw::os::IAllocator* m_pAllocator;
    nw::os::IAllocator* m_pDeviceAllocator;
    bool m_ForceUseCameraWScale;

    static SceneManager* m_pInstance;
};

}
}
#if NN_CURRENT_VERSION_NUMBER >= NN_VERSION_NUMBER(4,0,0,0)
#pragma diag_default 2530 // padding added to end of struct.
#pragma diag_default 1301 // padding inserted in struct.
#endif

#endif
