/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#ifndef NW4C_VIEWER_G3D_SCENE_RENDERER_H_
#define NW4C_VIEWER_G3D_SCENE_RENDERER_H_

#if NN_CURRENT_VERSION_NUMBER >= NN_VERSION_NUMBER(4,0,0,0)
#pragma diag_suppress 1301 // padding inserted in struct.
#pragma diag_suppress 2530 // padding added to end of struct.
#endif
namespace vw {
namespace system { class XYZTC; class Texture;}

namespace g3d {

class Grid;

class SceneRenderer
{
public:
    enum RenderFlag
    {
        RenderQueue = 0,
        RenderGrid,
        RenderLightDir,
        RenderParticleEmitterForm,
        RenderTargetCameraAxis,
        RenderModelAxis,
        RenderBoneAxis,
        RenderParticleEmitterAxis,
        RenderConstraintTargetBoneAxis,
        RenderCameraIcon,
        RenderLightIcon,
        RenderTargetLine,
        RenderViewFrustum,
        RenderBoundingBox,
        RenderDepthLevelSurface,
        RenderNum
    };

    enum AxisType
    {
        AxisModel = 0,
        AxisBone,
        AxisParticleEmitter,
        AxisConstraintTargetBone,
        AxisNum
    };

    struct DrawAxisInfo
    {
        AxisType axisType;
        nw::gfx::CalculatedTransform* axisTransform;
        void* key1;
        void* key2;
    };

    struct BlinkAnimTarget
    {
        enum BlinkAnimType {
            BLINK_MODEL,
            BLINK_MESH,
            BLINK_TEXTURE,
            BLINK_MATERIAL,
            BLINK_NUM
        };
        unsigned int hashID;
        unsigned int index;
        BlinkAnimType type;
    };

    struct BlinkAnimInfo
    {
        nw::gfx::Model* model;
        nw::gfx::res::ResMesh mesh;
    };

    // Get an instance
    static SceneRenderer* GetInstance();

    // Initialization
    static void Initialize();

    // Finalize.
    static void Finalize();

    // Initializes when graphics are used.
    void InitializeGraphics();

    // Update scene
    void UpdateScene();

    // Stereo scene rendering
    void RenderStereoScene();

    // Rendering
    void RenderScene(nw::gfx::IRenderTarget* renderTarget, nw::gfx::Camera* camera);

    // 3D information display
    void DrawInfo(nw::gfx::Camera* viewCamera, nw::gfx::Camera* baseCamera, bool isRender[RenderNum]);

    // Gets the grid management class.
    Grid* GetGrid() { return m_Grid; }

    // Sets the blinking animation interval.
    void SetBlinkSpan(u32 step);

    // Sets the number of animation blinks.
    void SetBlinkNum(u32 num);

    // Gets first iterator of the blinking animation.
    nw::ut::MoveArray<BlinkAnimInfo>::iterator GetBlinkAnimationsBegin();

    // Gets end iterator of the blinking animation.
    nw::ut::MoveArray<BlinkAnimInfo>::iterator GetBlinkAnimationsEnd();

    // Sets blinking animation.
    void SetBlinkAnimation(nw::ut::MoveArray<BlinkAnimTarget>& targets);

    // Resets blinking animation.
    void ResetBlinkAnimation();

    // Adds axis rendering.
    void AddDrawAxis(DrawAxisInfo& axis);

    // Deletes axis rendering of particular types.
    void DeleteDrawAxis(AxisType type);

    // Deletes axis rendering of particular objects.
    void DeleteDrawAxis(void* key1, void* key2);

    // Deletes all rendering.
    void ClearDrawAxis();

    // Sets the sort mode.
    void SetRenderSortMode(nw::gfx::ISceneUpdater::RenderSortMode mode);

    // Gets the world matrix updater.
    nw::gfx::WorldMatrixUpdater* GetWorldMatrixUpdater() { return m_pWorldMatrixUpdater; }

    // Gets the skeleton updater.
    nw::gfx::SkeletonUpdater* GeSkeletonUpdater() { return m_pSkeletonUpdater; }

    // Sets and gets the axis display width.
    void SetManipulatorScale(f32 scale) { m_ManipulatorScale = scale; }
    f32  GetManipulatorScale() const { return m_ManipulatorScale; }

    // Sets and gets the display width of the camera gaze point axis.
    void SetCameraManipulatorScale( f32 scale ) { m_CameraManipulatorScale = scale; }
    f32  GetCameraManipulatorScale() const { return m_CameraManipulatorScale; }

    // Sets and gets the alpha value of the stereoscopic vision display surface.
    void SetDepthLevelSurfaceAlpha( f32 alpha ) { m_DepthLevelSurfaceAlpha = alpha; }
    f32  GetDepthLevelSurfaceAlpha() const { return m_DepthLevelSurfaceAlpha; }
    // Gets the minimum and maximum alpha values of the stereoscopic display surface.
    static f32 GetDepthLevelSurfaceMinAlpha() { return 0.f; }
    static f32 GetDepthLevelSurfaceMaxAlpha() { return 1.f; }

    // Gets the render flag of the results for determining whether rendering is possible.
    void GetCorrectedRenderFlag( bool* RenderFlag );

private:
    SceneRenderer();
    ~SceneRenderer();

    void SetupDraw(bool* RenderFlag, nw::gfx::Camera* activeCamera);
    void DrawScene(nw::gfx::Camera* camera, bool isRender[RenderNum], bool copyCommand);
    void DrawInfo(nw::gfx::Camera* camera, nw::gfx::Camera* targetCamera, nw::gfx::Camera* viewCamera, bool isRender[RenderNum]);
    void DrawInfoDepthLevelSurface(nw::gfx::Camera* camera, nw::gfx::Camera* viewCamera, bool isTransparent);
    void DrawIcons(nw::gfx::Camera* camera, nw::gfx::Camera* viewCamera, bool isRender[RenderNum]);
    void DrawEmitterForm();
    void DrawLightIcon(nw::gfx::Camera* camera);
    void DrawCameraIcon(nw::gfx::Camera* camera, nw::gfx::Camera* viewCamera);
    void DrawAxis(bool isRenderAxis[AxisNum], nw::gfx::Camera* viewCamera);
    void DrawTargetCameraAxis(nw::gfx::Camera* targetCamera, nw::gfx::Camera* viewCamera);
    void DrawLightDir(nw::gfx::Camera* camera);
    void DrawTargetLine(nw::gfx::Camera* camera, nw::gfx::Camera* viewCamera);
    void DrawViewFrustum(nw::gfx::Camera* camera, nw::gfx::Camera* viewCamera);
    void DrawBoundingBox(nw::gfx::Camera* camera, nw::gfx::Camera* viewCamera);
    void DrawDepthLevelSurface(nw::gfx::Camera* camera, nw::gfx::Camera* viewCamera);
    void InitAxis();
    void CreateCursor(vw::system::XYZTC* cursor);
    void UpdateLightDir();

    void AddBlinkAnim(BlinkAnimInfo& info);
    void AddBlinkModel(unsigned int hashID);
    void AddBlinkMesh(unsigned int hashID, unsigned int index);
    void AddBlinkTexture(unsigned int hashID);
    void AddBlinkMaterial(unsigned int hashID, unsigned int index);
    void FlipSelectedObjectsVisibility(f32 time);
    void SubmitView(
        nw::gfx::SceneContext* sceneContext,
        const nw::gfx::Camera& camera,
        nw::gfx::ISceneUpdater::IsVisibleModelFunctor* isVisibleModel
    );

    nw::gfx::RenderKeyFactory* m_pPriorMaterialRenderKeyFactory;
    nw::gfx::RenderKeyFactory* m_pPriorDepthRenderKeyFactory;
    nw::gfx::RenderQueue* m_pRenderQueue;
    nw::gfx::SceneUpdater* m_pSceneUpdater;
    nw::gfx::WorldMatrixUpdater* m_pWorldMatrixUpdater;
    nw::gfx::SkeletonUpdater* m_pSkeletonUpdater;
    nw::gfx::MeshRenderer* m_pMeshRenderer;
    Grid* m_Grid;
    vw::system::XYZTC* m_Axis;
    vw::system::XYZTC* m_Cursor;
    vw::system::XYZTC* m_HemiSphereCursor;
    vw::system::XYZTC* m_LineModel;
    vw::system::XYZTC* m_ViewFrustum;
    vw::system::XYZTC* m_BoundingBoxLine;
    vw::system::XYZTC* m_DepthLevelSurface;
    vw::system::Texture* m_LightTexture;
    vw::system::Texture* m_CameraTexture;
    nw::ut::MoveArray<BlinkAnimInfo> m_BlinkTargets;
    s32 m_BlinkStartTick;
    u32 m_BlinkSpan;
    u32 m_BlinkNum;
    nw::ut::MoveArray<DrawAxisInfo> m_DrawAxis;
    nw::os::IAllocator* m_pAllocator;
    nw::os::IAllocator* m_pDeviceAllocator;

    f32 m_ManipulatorScale;
    f32 m_CameraManipulatorScale;
    f32 m_DepthLevelSurfaceAlpha;

    static SceneRenderer* m_pInstance;
};

}
}
#if NN_CURRENT_VERSION_NUMBER >= NN_VERSION_NUMBER(4,0,0,0)
#pragma diag_default 2530 // padding added to end of struct.
#pragma diag_default 1301 // padding inserted in struct.
#endif

#endif
