/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#include <viewer/system.h>
#include "../../../include/viewer/g3d/Camera.h"
#include "../../../include/viewer/g3d/CameraMgr.h"

namespace vw  {
namespace g3d {


/*---------------------------------------------------------------------------*
  Name:         Camera::Camera

  Description:  Constructor

  Arguments:    camera input camera

  Returns:      None.
 *---------------------------------------------------------------------------*/
Camera::Camera(nw::gfx::Camera* camera, const char *name) : m_Camera(camera)
{
    strcpy(mName, name);
    m_RotateSpeed   = 0.0f;
    m_TransSpeed    = 0.0f;
    m_TrackSpeed    = 0.0f;
    m_TouchPanelX   = -1;
    m_TouchPanelY   = -1;
    m_TouchLeftMode = false;
    m_TiltXSpeed    = 0.0f;
    m_TiltYSpeed    = 0.0f;
    m_SlideXSpeed   = 0.0f;
    m_SlideYSpeed   = 0.0f;

    this->Preset();
}

/*---------------------------------------------------------------------------*
  Name:         Camera::GetCamera

  Description:  Gets camera

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
nw::gfx::Camera*
Camera::GetCamera()
{
    return m_Camera;
}

/*---------------------------------------------------------------------------*
  Name:         Camera::Reset

  Description:  In the end, returns the position to the pre-set position.

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
Camera::Reset()
{
    nw::gfx::ResCameraViewUpdater       resViewUpdater             = m_Camera->GetViewUpdater()->GetResource();
    nw::gfx::ResAimTargetViewUpdater    resAimTargetViewUpdater    = nw::gfx::ResDynamicCast<nw::gfx::ResAimTargetViewUpdater>(resViewUpdater);
    nw::gfx::ResLookAtTargetViewUpdater resLookAtTargetViewUpdater = nw::gfx::ResDynamicCast<nw::gfx::ResLookAtTargetViewUpdater>(resViewUpdater);
    nw::gfx::ResRotateViewUpdater       resRotateViewUpdater       = nw::gfx::ResDynamicCast<nw::gfx::ResRotateViewUpdater>(resViewUpdater);
    if (resAimTargetViewUpdater.IsValid())
    {
        resAimTargetViewUpdater.SetTargetPosition( m_CameraTarget );
        resAimTargetViewUpdater.SetTwist( m_Twist );
    }
    else if (resLookAtTargetViewUpdater.IsValid())
    {
        resLookAtTargetViewUpdater.SetTargetPosition( m_CameraTarget );
        resLookAtTargetViewUpdater.SetUpwardVector( m_CameraUp );
    }
    else if(resRotateViewUpdater.IsValid())
    {
        resRotateViewUpdater.SetViewRotate( m_Rotate );
        InitializeVirtualTarget();
    }
    m_Camera->Transform().SetTranslate( m_CameraPos );

    nw::gfx::ResCameraProjectionUpdater resProjUpdater  = m_Camera->GetProjectionUpdater()->GetResource();
    nw::gfx::ResOrthoProjectionUpdater  resOrthoUpdater = nw::gfx::ResDynamicCast<nw::gfx::ResOrthoProjectionUpdater>( resProjUpdater );
    if( resOrthoUpdater.IsValid() )
    {
        resOrthoUpdater.SetRectWithYFlip(m_OrthoRect.left, m_OrthoRect.right, m_OrthoRect.bottom, m_OrthoRect.top);
    }
}

/*---------------------------------------------------------------------------*
  Name:         Camera::Preset

  Description:  Stores the current settings so that the status can return if there is a reset.

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
Camera::Preset()
{
    m_Camera->Transform().GetTranslate(&m_CameraPos);

    nw::gfx::ResCameraViewUpdater       resViewUpdater             = m_Camera->GetViewUpdater()->GetResource();
    nw::gfx::ResAimTargetViewUpdater    resAimTargetViewUpdater    = nw::gfx::ResDynamicCast<nw::gfx::ResAimTargetViewUpdater>( resViewUpdater );
    nw::gfx::ResLookAtTargetViewUpdater resLookAtTargetViewUpdater = nw::gfx::ResDynamicCast<nw::gfx::ResLookAtTargetViewUpdater>( resViewUpdater );
    nw::gfx::ResRotateViewUpdater       resRotateViewUpdater       = nw::gfx::ResDynamicCast<nw::gfx::ResRotateViewUpdater>( resViewUpdater );
    if ( resAimTargetViewUpdater.IsValid() )
    {
        m_CameraTarget = resAimTargetViewUpdater.GetTargetPosition();
        m_Twist = resAimTargetViewUpdater.GetTwist();
    }
    else if ( resLookAtTargetViewUpdater.IsValid() )
    {
        m_CameraTarget = resLookAtTargetViewUpdater.GetTargetPosition();
        m_CameraUp     = resLookAtTargetViewUpdater.GetUpwardVector();
    }
    else if(resRotateViewUpdater.IsValid())
    {
        m_Rotate = resRotateViewUpdater.GetViewRotate();
        InitializeVirtualTarget();
    }

    nw::gfx::ResCameraProjectionUpdater resProjUpdater  = m_Camera->GetProjectionUpdater()->GetResource();
    nw::gfx::ResOrthoProjectionUpdater  resOrthoUpdater = nw::gfx::ResDynamicCast<nw::gfx::ResOrthoProjectionUpdater>( resProjUpdater );
    if( resOrthoUpdater.IsValid() )
    {
        m_OrthoRect = resOrthoUpdater.GetRect();
    }
}

/*---------------------------------------------------------------------------*
  Name:         Camera::MoveCamera3D

  Description:  Camera movement main routine

  Arguments:    hidStatus HID input

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
Camera::MoveCamera3D(vw::system::HidStatus& hidStatus)
{
    DirectionalObj dirObj;
    u32 updateType;

    bool isViewTypeRotate = ( m_Camera->GetResCamera().GetViewType() == nw::gfx::ResCamera::VIEWTYPE_ROTATE );
    bool isProjTypeOrtho = ( m_Camera->GetResCamera().GetProjectionType() == nw::gfx::ResCamera::PROJTYPE_ORTHO );
    nw::gfx::ResCameraViewUpdater       resViewUpdater             = m_Camera->GetViewUpdater()->GetResource();
    nw::gfx::ResAimTargetViewUpdater    resAimTargetViewUpdater    = nw::gfx::ResDynamicCast<nw::gfx::ResAimTargetViewUpdater>(resViewUpdater);
    nw::gfx::ResLookAtTargetViewUpdater resLookAtTargetViewUpdater = nw::gfx::ResDynamicCast<nw::gfx::ResLookAtTargetViewUpdater>(resViewUpdater);
    nw::gfx::ResRotateViewUpdater       resRotateViewUpdater       = nw::gfx::ResDynamicCast<nw::gfx::ResRotateViewUpdater>(resViewUpdater);

    nw::math::VEC3  cameraPos   = m_Camera->Transform().GetTranslate();
    nw::math::MTX33 orientation = this->GetOrientation( CameraMgr::GetInstance()->GetViewCamera() );

    nw::math::VEC3 upward;

    // Sets the values required for the calculation in the directional object.

    // Aim mode
    if ( resAimTargetViewUpdater.IsValid() )
    {
        updateType = DirectionalObj::UPDATE_TYPE_AIM;
        nw::math::VEC3 targetPos = resAimTargetViewUpdater.GetTargetPosition();
        f32            twist     = resAimTargetViewUpdater.GetTwist();

        dirObj.SetUpdateType( updateType );
        dirObj.SetTargetPos( targetPos );
        dirObj.SetTwist( twist );
    }
    // LooktAt mode
    else if ( resLookAtTargetViewUpdater.IsValid() )
    {
        updateType = DirectionalObj::UPDATE_TYPE_LOOKAT;
        nw::math::VEC3 targetPos = resLookAtTargetViewUpdater.GetTargetPosition();
        upward    = resLookAtTargetViewUpdater.GetUpwardVector();

        dirObj.SetUpdateType( updateType );
        dirObj.SetTargetPos( targetPos );
        dirObj.SetUpward( upward );
        dirObj.SetInitUpward( m_CameraUp );
    }
    // Rotate mode
    else if ( resRotateViewUpdater.IsValid() )
    {
        updateType = DirectionalObj::UPDATE_TYPE_ROTATE;
        nw::math::VEC3 rotate = resRotateViewUpdater.GetViewRotate();

        dirObj.SetTargetPos( m_VirtualTarget );
        dirObj.SetUpdateType( updateType );
        dirObj.SetRotate( rotate );
    }
    dirObj.SetTranslate( cameraPos );
    dirObj.SetOrientation( orientation );
    dirObj.SetRotateSpeed( m_RotateSpeed );
    dirObj.SetTransSpeed( m_TransSpeed );
    dirObj.SetTrackSpeed( m_TrackSpeed );
    dirObj.SetIsProjTypeOrtho( isProjTypeOrtho );
    dirObj.SetPadEventFromHid( hidStatus, isViewTypeRotate );

#if defined( NN_PCGX )
    PadEvent padEvent = dirObj.GetPadEvent();
    PCGXAddCameraEvent(&padEvent);
    dirObj.SetPadEvent(padEvent);
#endif

    // Rotational processing
    Rotate( dirObj );

    // Movement forward and aft
    Dolly( dirObj );

    // Tilt manipulation
    Tilt( dirObj );

    // Sliding movement
    Slide( dirObj );

    // Track movement
    Track( dirObj );

    // Zoom processing
    Zoom( dirObj );

    // Update processing
    if ( resAimTargetViewUpdater.IsValid() )
    {
        resAimTargetViewUpdater.SetTargetPosition( dirObj.GetTargetPos() );
    }
    else if ( resLookAtTargetViewUpdater.IsValid() )
    {
        resLookAtTargetViewUpdater.SetTargetPosition( dirObj.GetTargetPos() );
        resLookAtTargetViewUpdater.SetUpwardVector( dirObj.GetUpward() );
    }
    else if ( resRotateViewUpdater.IsValid() )
    {
        resRotateViewUpdater.SetViewRotate( dirObj.GetRotate() );
        m_VirtualTarget = dirObj.GetTargetPos();
    }
    m_Camera->Transform().SetTranslate( dirObj.GetTranslate() );
    m_RotateSpeed = dirObj.GetRotateSpeed();
    m_TransSpeed = dirObj.GetTransSpeed();
    m_TrackSpeed = dirObj.GetTrackSpeed();

    // Reset process
    Reset( dirObj );
}

void
DecelValue(f32& value, f32 decel)
{
    if(value < 0.0f)
    {
        value += decel;
        if(value > 0.0f) value = 0.0f;
    }
    else if(value > 0.0f)
    {
        value -= decel;
        if(value < 0.0f) value = 0.0f;
    }
}

/*---------------------------------------------------------------------------*
  Name:         Camera::Rotate

  Description:  Camera rotational operation.

  Arguments:    directionalObj

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
Camera::Rotate( vw::g3d::DirectionalObj& directionalObj )
{
    directionalObj.Rotate( false );
}

/*---------------------------------------------------------------------------*
  Name:         Camera::Dolly

  Description:  Camera forward and aft movement

  Arguments:    directionalObj

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
Camera::Dolly( vw::g3d::DirectionalObj& directionalObj )
{
    directionalObj.Dolly( false );
}

/*---------------------------------------------------------------------------*
  Name:         Camera::Tilt

  Description:  Camera head shake processing

  Arguments:    directionalObj

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
Camera::Tilt( vw::g3d::DirectionalObj& directionalObj )
{
   directionalObj.Tilt( false );
}

/*---------------------------------------------------------------------------*
  Name:         Camera::Slide

  Description:  Camera slide movement

  Arguments:    directionalObj

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
Camera::Slide( vw::g3d::DirectionalObj& directionalObj )
{
    directionalObj.Slide( false );
}

/*---------------------------------------------------------------------------*
  Name:         Camera::Track

  Description:  Camera track movement

  Arguments:    directionalObj

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
Camera::Track( vw::g3d::DirectionalObj& directionalObj )
{
    directionalObj.Track( false );
}

/*---------------------------------------------------------------------------*
  Name:         Camera::Zoom

  Description:  Camera zoom

  Arguments:    directionalObj

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
Camera::Zoom( vw::g3d::DirectionalObj& directionalObj )
{
    if(m_Camera->GetResCamera().GetProjectionType() == nw::gfx::ResCamera::PROJTYPE_PERSPECTIVE)
        return;

    f32 zoom = 0.0f;
    if ( directionalObj.GetPadEvent().flags & PadEvent::HANDLE_ZOOM )
    {
        zoom = directionalObj.GetPadEvent().zoom;
    }
    if( zoom == 0.0f )
    {
        return;
    }

    nw::gfx::ResCameraProjectionUpdater resProjUpdater  = m_Camera->GetProjectionUpdater()->GetResource();
    nw::gfx::ResOrthoProjectionUpdater  resOrthoUpdater = nw::gfx::ResDynamicCast<nw::gfx::ResOrthoProjectionUpdater>( resProjUpdater );
    if( resOrthoUpdater.IsValid() )
    {
        nw::gfx::ResProjectionRect rect = resOrthoUpdater.GetRect();

        nw::math::VEC2 center = rect.GetCenter();
        f32 height = rect.GetHeight();
        f32 aspect = rect.GetAspectRatio();

        height += zoom;
        height = nw::ut::Clamp(height, 1.0f, 1000.0f);

        rect.Set( aspect, height, center );
        nw::ut::Rect rc = rect;

        resOrthoUpdater.SetRectWithYFlip(rc.left, rc.right, rc.bottom, rc.top);
    }
}

/*---------------------------------------------------------------------------*
  Name:         Camera::Reset

  Description:  Camera reset processing

  Arguments:    directionalObj

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
Camera::Reset( vw::g3d::DirectionalObj& directionalObj )
{
    if ( directionalObj.GetPadEvent().flags & PadEvent::HANDLE_RESET )
    {
        this->Reset();
    }
}

/*---------------------------------------------------------------------------*
  Name:         Camera::InitializeVirtualTarget

  Description:  Initializes the virtual gaze point used by the camera.

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
Camera::InitializeVirtualTarget()
{
    const float PI = 3.14159265359f;    // Transfers to the TODO utility.
    nw::math::VEC3 rotateRad = nw::math::VEC3(vw::g3d::DegreeToRadian(m_Rotate.x), PI + DegreeToRadian(m_Rotate.y), DegreeToRadian(m_Rotate.z));
    nw::math::VEC3 dir;
    ComputeRotateRad(rotateRad.x, rotateRad.y, dir);
    m_VirtualTarget = m_CameraPos + VIRTUAL_TARGET_DISTANCE * dir;
}

/*---------------------------------------------------------------------------*
  Name:         Camera::GetOrientation

  Description:  Gets the camera's orientation matrix.

  Arguments:    camera

  Returns:      Orientation matrix
 *---------------------------------------------------------------------------*/
nw::math::MTX33
Camera::GetOrientation( nw::gfx::Camera* camera )
{
    nw::math::MTX34 orientation;

    nw::math::VEC3 viewPos, rotate;
    camera->Transform().GetTranslate( &viewPos );

    switch( camera->GetResCamera().GetViewType() )
    {
    case nw::gfx::ResCamera::VIEWTYPE_AIM:
        {
            nw::gfx::ResAimTargetViewUpdater resAimTargetViewUpdater =
                nw::gfx::ResStaticCast<nw::gfx::ResAimTargetViewUpdater>( camera->GetViewUpdater()->GetResource() );

            nw::math::MTX34LookAtRad(
                &orientation,
                &viewPos,
                resAimTargetViewUpdater.GetTwist(),
                &resAimTargetViewUpdater.GetTargetPosition());

            nw::math::MTX34Inverse(&orientation, &orientation);
        }
        break;
    case nw::gfx::ResCamera::VIEWTYPE_LOOKAT:
        {
            nw::gfx::ResLookAtTargetViewUpdater resLookAtTargetViewUpdater =
                nw::gfx::ResStaticCast<nw::gfx::ResLookAtTargetViewUpdater>( camera->GetViewUpdater()->GetResource() );

            nw::math::MTX34LookAt(
                &orientation,
                &viewPos,
                &resLookAtTargetViewUpdater.GetUpwardVector(),
                &resLookAtTargetViewUpdater.GetTargetPosition());

            nw::math::MTX34Inverse(&orientation, &orientation);
        }
        break;
    case nw::gfx::ResCamera::VIEWTYPE_ROTATE:
        {
            nw::gfx::ResRotateViewUpdater resRotateViewUpdater =
                nw::gfx::ResStaticCast<nw::gfx::ResRotateViewUpdater>( camera->GetViewUpdater()->GetResource() );

            nw::math::MTX34CameraRotateRad(
                &orientation,
                &viewPos,
                &resRotateViewUpdater.GetViewRotate());

            nw::math::MTX34Inverse(&orientation, &orientation);
        }
        break;
    }

    nw::math::MTX33 rotateMatrix;
    nw::math::MTX34ToMTX33(&rotateMatrix, &orientation);
    return rotateMatrix;
}

/*---------------------------------------------------------------------------*
  Name:         Camera::Focus

  Description:  Focuses the camera

  Arguments:    pos focal position

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
Camera::Focus(nw::math::VEC3 pos)
{
    nw::math::VEC3 targetPosition;
    nw::math::VEC3 targetDelta;
    nw::math::VEC3 Translate;
    m_Camera->Transform().GetTranslate(&Translate);

    if(m_Camera->GetResCamera().GetViewType() == nw::gfx::ResCamera::VIEWTYPE_LOOKAT)
    {
        nw::gfx::LookAtTargetViewUpdater* viewUpdater =
            nw::ut::DynamicCast<nw::gfx::LookAtTargetViewUpdater*>(m_Camera->GetViewUpdater());

        nw::gfx::ResLookAtTargetViewUpdater resViewUpdater =
            nw::gfx::ResStaticCast<nw::gfx::ResLookAtTargetViewUpdater>(viewUpdater->GetResource());

        targetPosition = resViewUpdater.GetTargetPosition();
        targetDelta = pos - targetPosition;
        resViewUpdater.SetTargetPosition(pos);

        m_Camera->Transform().SetTranslate(Translate + targetDelta);
    }
    else if(m_Camera->GetResCamera().GetViewType() == nw::gfx::ResCamera::VIEWTYPE_AIM)
    {
        nw::gfx::AimTargetViewUpdater* viewUpdater =
            nw::ut::DynamicCast<nw::gfx::AimTargetViewUpdater*>(m_Camera->GetViewUpdater());

        nw::gfx::ResAimTargetViewUpdater resViewUpdater =
            nw::gfx::ResStaticCast<nw::gfx::ResAimTargetViewUpdater>(viewUpdater->GetResource());

        targetPosition = resViewUpdater.GetTargetPosition();
        targetDelta = pos - targetPosition;
        resViewUpdater.SetTargetPosition(pos);

        m_Camera->Transform().SetTranslate(Translate + targetDelta);
    }
}

/*---------------------------------------------------------------------------*
  Name:         Camera::Focus

  Description:  Focuses the camera

  Arguments:    aabb focusAABB

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
Camera::Focus(nw::math::AABB aabb)
{
    nw::math::VEC3 aabbCenter = (aabb.Pmax+aabb.Pmin)*0.5f;
    nw::math::VEC3 aabbSize   = (aabb.Pmax-aabb.Pmin);

    Focus(aabbCenter);

    f32 fovf = 0.659734f;
    if (m_Camera->GetResCamera().GetProjectionType() == nw::gfx::ResCamera::PROJTYPE_PERSPECTIVE)
    {
        nw::gfx::ResPerspectiveProjectionUpdater resProjectionUpdater =
            nw::gfx::ResDynamicCast<nw::gfx::ResPerspectiveProjectionUpdater>(
            m_Camera->GetProjectionUpdater()->GetResource());

        fovf = resProjectionUpdater.GetFovy();
    }
    else if(m_Camera->GetResCamera().GetProjectionType() == nw::gfx::ResCamera::PROJTYPE_ORTHO)
    {
        nw::gfx::ResOrthoProjectionUpdater resOrthoUpdater =
            nw::gfx::ResDynamicCast<nw::gfx::ResOrthoProjectionUpdater>(
            m_Camera->GetProjectionUpdater()->GetResource());

        fovf = 0.659734f;
    }

    nw::math::VEC3 Translate;
    m_Camera->Transform().GetTranslate(&Translate);
    nw::math::VEC3 direction = aabbCenter - Translate;
    f32 dist = VEC3Len(&direction);
    f32 radius = VEC3Len(&aabbSize);
    f32 distance = nw::math::CosRad(fovf) / nw::math::SinRad(fovf) * (radius*(1.0f/DirectionalObj::GetDirectionalInfo().FocusScale));
    if(distance > 0.01f)
    {
        nw::math::VEC3Normalize(&direction,&direction);
        m_Camera->Transform().SetTranslate(aabbCenter - direction * distance);
    }
}

void
Camera::SetPreset(const nw::math::VEC3& cameraPos,
                  const nw::math::VEC3& cameraTarget,
                  const nw::math::VEC3& cameraUp)
{
    m_CameraPos = cameraPos;
    m_CameraTarget = cameraTarget;
    m_CameraUp = cameraUp;
}

}   /* namespace g3d */
}   /* namespace vw  */
