/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#include <viewer/system.h>
#include "../../../include/viewer/g3d/CameraMgr.h"
#include "../include/SceneManager.h"
#include "../include/ErrorCheck.h"

namespace vw  {
namespace g3d {

CameraMgr* CameraMgr::m_Instance = NULL;

/*static*/CameraMgr*
CameraMgr::GetInstance()
{
    NW_ASSERT( m_Instance != NULL );
    return m_Instance;
}

bool IsExist(
    nw::ut::MoveArray<Camera>::iterator beginIter,
    nw::ut::MoveArray<Camera>::iterator endIter,
    nw::gfx::Camera *camera
) {
    for (; beginIter!=endIter; ++beginIter)
    {
        if (beginIter->GetCamera()==camera)
        {
            return true;
        }
    }
    return false;
}

bool IsExist(
    SceneElemSelector::CameraList::iterator beginIter,
    SceneElemSelector::CameraList::iterator endIter,
    nw::gfx::Camera *camera
) {
    for (; beginIter!=endIter; ++beginIter)
    {
        if ((*beginIter)->GetSceneNode()==camera)
        {
            return true;
        }
    }
    return false;
}

/*---------------------------------------------------------------------------*
  Name:         CameraMgr::MoveCamera3D
  Description:  Sets up the camera array.
  Arguments:    scene Scene class
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
CameraMgr::Setup(SceneElemSelector::CameraList::iterator camerasBegin,
                 SceneElemSelector::CameraList::iterator camerasEnd)
{
    // Deletes excluding the ones that already exist.
    {
        nw::ut::MoveArray<Camera>::iterator it = m_CameraArray.begin();
        while( it != m_CameraArray.end() )
        {
            if (!IsExist(camerasBegin, camerasEnd, it->GetCamera()))
            {
                nw::ut::MoveArray<Camera>::iterator next = it;
                ++next;
                it = m_CameraArray.erase(it, next);
            }
            else
            {
                ++it;
            }
        }
    }

    // Adds new camera.
    for(SceneElemSelector::CameraList::iterator it = camerasBegin; it != camerasEnd; ++it)
    {
        if (!IsExist(m_CameraArray.begin(), m_CameraArray.end(), (*it)->GetSceneNode()))
        {
            m_CameraArray.push_back( Camera( (*it)->GetSceneNode(), (*it)->GetName() ) );
        }
    }

    m_ViewIdx = 0;
    m_TargetIdx = 0;

    // If there is one that matches a saved name, that one is made the ViewCamera.
    for (int i=0; i<m_CameraArray.size(); ++i)
    {
        if (!strcmp(m_CameraArray[i].GetName(), m_PrevViewCameraName))
        {
            m_ViewIdx = i;
            break;
        }
    }

    // If there is one that matches a saved name, that one is made the TargetCamera.
    for (int i=0; i<m_CameraArray.size(); ++i)
    {
        if (!strcmp(m_CameraArray[i].GetName(), m_PrevTargetCameraName))
        {
            m_TargetIdx = i;
            break;
        }
    }
}


/*---------------------------------------------------------------------------*
  Name:         CameraMgr::MoveCamera3D
  Description:  Gets the current target camera.
  Arguments:    hidStatus Specifies HID input.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
CameraMgr::MoveCamera3D(vw::system::HidStatus& hidStatus)
{
    Camera* target = &m_CameraArray[m_TargetIdx];

    // Target camera operation
    target->MoveCamera3D(hidStatus);
}

void
CameraMgr::UpdateCameraMatrix()
{
    // Updates default camera aspect ratio
    vw::g3d::CameraElem* defaultCamera = vw::g3d::SceneManager::GetInstance()->GetDefaultCamera();
    nw::gfx::Camera* pCamera = defaultCamera->GetSceneNode();
    nw::gfx::ResPerspectiveProjectionUpdater resProjectionUpdater =
        nw::gfx::ResDynamicCast<nw::gfx::ResPerspectiveProjectionUpdater>(
        pCamera->GetProjectionUpdater()->GetResource());
    bool lowerPreview = vw::system::System_GetSetting()->lowerPreview;
    f32 aspectRatio = lowerPreview? 320.0f / 240.0f :
                                    400.0f / 240.0f ;
    resProjectionUpdater.SetAspectRatio(aspectRatio);

    // update camera
    for(size_t i=0; i<m_CameraArray.size(); ++i)
    {
        nw::gfx::Camera* camera = m_CameraArray[i].GetCamera();
        if ( camera->GetAnimGroup() && camera->GetAnimObject() )
        {
            ErrorCheck::CheckCameraNear( camera );
            ErrorCheck::CheckCameraNearFar( camera );
            ErrorCheck::CheckCameraPosition( camera );
            ErrorCheck::CheckCameraUpVectorZero( camera );
            ErrorCheck::CheckCameraUpVectorParallel( camera );
        }
        camera->UpdateCameraMatrix();
    }

    Camera* target = &m_CameraArray[m_ViewIdx];
    if(target)
    {
        // set WScale
        nw::gfx::Camera* camera = target->GetCamera();
        m_LeftViewCamera->SetWScale(camera->GetWScale());
        m_RightViewCamera->SetWScale(camera->GetWScale());

        // stereo camera
        nw::gfx::ResCameraProjectionUpdater resProj = camera->GetProjectionUpdater()->GetResource();
        const DirectionalInfo& info = DirectionalObj::GetDirectionalInfo();
        f32 depthLevel = resProj.GetNear() + 5.0f;
        f32 depthRange = info.factor;
        if(info.depthLevel != 0.0f)
        {
            depthLevel = info.depthLevel;
        }
        else
        {
            depthLevel += info.depthLevelBias;
        }

        nn::math::MTX44& projOriginal = camera->ProjectionMatrix();
        nn::math::MTX34& viewOriginal = camera->ViewMatrix();
        nn::math::MTX44& projL        = m_LeftViewCamera->ProjectionMatrix();
        nn::math::MTX34& viewL        = m_LeftViewCamera->ViewMatrix();
        nn::math::MTX44& projR        = m_RightViewCamera->ProjectionMatrix();
        nn::math::MTX34& viewR        = m_RightViewCamera->ViewMatrix();

        // When using the Ortho camera for stereo display, you have to prepare pictures yourself.
        if (camera->GetResCamera().GetProjectionType() == nw::gfx::ResCamera::PROJTYPE_ORTHO)
        {
            // At present, stereo display with the Ortho camera is not yet supported.
            nw::math::MTX44Copy(&projL, &projOriginal);
            nw::math::MTX34Copy(&viewL, &viewOriginal);
            nw::math::MTX44Copy(&projR, &projOriginal);
            nw::math::MTX34Copy(&viewR, &viewOriginal);

            // With CalculateMatrices(....) rotation occurs together with the screen, so for Ortho, set pivot.
            // 
            camera->GetProjectionUpdater()->SetPivotDirection(nw::math::PIVOT_UPSIDE_TO_TOP);
            m_LeftViewCamera->GetProjectionUpdater()->SetPivotDirection(nw::math::PIVOT_UPSIDE_TO_TOP);
            m_RightViewCamera->GetProjectionUpdater()->SetPivotDirection(nw::math::PIVOT_UPSIDE_TO_TOP);
        }
        else
        {
            m_StereoCamera.CalculateMatrices(
                &projL,
                &viewL,
                &projR,
                &viewR,
                &projOriginal,
                &viewOriginal,
                depthLevel,
                depthRange,
                info.realSwitch);
        }
    }
}

void
CameraMgr::Focus(nw::math::VEC3 pos)
{
    Camera* target = &m_CameraArray[m_ViewIdx];
    target->Focus(pos);
}

void
CameraMgr::Focus(nw::math::AABB aabb)
{
    Camera* target = &m_CameraArray[m_ViewIdx];
    target->Focus(aabb);
}

/* ctor */
CameraMgr::CameraMgr()
    : m_ViewIdx(0)
    , m_TargetIdx(0)
#if defined( NN_PCGX )
{
    ZeroMemory(m_PrevViewCameraName, 512);
    ZeroMemory(m_PrevTargetCameraName, 512);
#else
     , m_PrevViewCameraName("\0")
     , m_PrevTargetCameraName("\0")
{
#endif
    // create stereo camera
    nw::os::IAllocator* pAllocator = Memory_GetAllocator();
    m_StereoCamera.Initialize();
    m_LeftViewCamera  = nw::gfx::Camera::DynamicBuilder().MaxChildren(0).MaxCallbacks(0).Create(pAllocator);
    m_RightViewCamera = nw::gfx::Camera::DynamicBuilder().MaxChildren(0).MaxCallbacks(0).Create(pAllocator);
    m_CameraArray = nw::ut::MoveArray<g3d::Camera>(pAllocator);
}

/* dtor */
CameraMgr::~CameraMgr()
{
    m_StereoCamera.Finalize();
    nw::ut::SafeDestroy( m_LeftViewCamera );
    nw::ut::SafeDestroy( m_RightViewCamera );
}

// Initialize
void
CameraMgr::Initialize()
{
    if (m_Instance != NULL) return;

    void *buf = Memory_GetAllocator()->Alloc(sizeof(CameraMgr));
    if (buf == NULL) return;

    m_Instance = new(buf) CameraMgr();
}

// Finalize
void
CameraMgr::Finalize()
{
    if (m_Instance == NULL) return;

    m_Instance->~CameraMgr();
    Memory_GetAllocator()->Free(m_Instance);

    m_Instance = NULL;
}

// IsFirstViewCamera()
bool
CameraMgr::IsFirstViewCamera()
{
    return ( m_ViewIdx == 0 );
}

// IsLastViewCamera()
bool
CameraMgr::IsLastViewCamera()
{
    return ( m_ViewIdx == m_CameraArray.size() - 1 );
}

// SwitchToFirstViewCamera()
void
CameraMgr::SwitchToFirstViewCamera()
{
    m_ViewIdx = 0;
    strcpy( m_PrevViewCameraName, m_CameraArray[m_ViewIdx].GetName() );
}

// SwitchToLastViewCamera()
void
CameraMgr::SwitchToLastViewCamera()
{
    m_ViewIdx = m_CameraArray.size() - 1;
    strcpy( m_PrevViewCameraName, m_CameraArray[m_ViewIdx].GetName() );
}

// SwitchToPrevViewCamera()
void
CameraMgr::SwitchToPrevViewCamera()
{
    m_ViewIdx = m_ViewIdx == 0 ? m_CameraArray.size() - 1 : m_ViewIdx - 1;
    strcpy( m_PrevViewCameraName, m_CameraArray[m_ViewIdx].GetName() );
}

// SwitchToNextViewCamera()
void
CameraMgr::SwitchToNextViewCamera()
{
    m_ViewIdx = m_ViewIdx == m_CameraArray.size() - 1 ? 0 : m_ViewIdx + 1;
    strcpy( m_PrevViewCameraName, m_CameraArray[m_ViewIdx].GetName() );
}

// IsFirstTargetCamera()
bool
CameraMgr::IsFirstTargetCamera()
{
    return ( m_TargetIdx == 0 );
}

// IsLastTargetCamera()
bool
CameraMgr::IsLastTargetCamera()
{
    return ( m_TargetIdx == m_CameraArray.size() - 1 );
}

// SwitchToFirstTargetCamera()
void
CameraMgr::SwitchToFirstTargetCamera()
{
    m_TargetIdx = 0;
    strcpy( m_PrevTargetCameraName, m_CameraArray[m_TargetIdx].GetName() );
}

// SwitchToLastTargetCamera()
void
CameraMgr::SwitchToLastTargetCamera()
{
    m_TargetIdx = m_CameraArray.size() - 1;
    strcpy( m_PrevTargetCameraName, m_CameraArray[m_TargetIdx].GetName() );
}

// SwitchToPrevTargetCamera()
void
CameraMgr::SwitchToPrevTargetCamera()
{
    m_TargetIdx = m_TargetIdx == 0 ? m_CameraArray.size() - 1 : m_TargetIdx - 1;
    strcpy( m_PrevTargetCameraName, m_CameraArray[m_TargetIdx].GetName() );
}

// SwitchToNextTargetCamera()
void
CameraMgr::SwitchToNextTargetCamera()
{
    m_TargetIdx = m_TargetIdx == m_CameraArray.size() - 1 ? 0 : m_TargetIdx + 1;
    strcpy( m_PrevTargetCameraName, m_CameraArray[m_TargetIdx].GetName() );
}

// SynchronizeViewCameraToTargetCamera
void
CameraMgr::SynchronizeViewCameraToTargetCamera()
{
    m_ViewIdx = m_TargetIdx;
    strcpy( m_PrevViewCameraName, m_CameraArray[m_ViewIdx].GetName() );
}

// SynchronizeTargetCameraToViewCamera
void
CameraMgr::SynchronizeTargetCameraToViewCamera()
{
    m_TargetIdx = m_ViewIdx;
    strcpy( m_PrevTargetCameraName, m_CameraArray[m_TargetIdx].GetName() );
}

// ReloadSetting
void
CameraMgr::ReloadSetting()
{
    nw::gfx::Camera* camera = SceneManager::GetInstance()->GetDefaultCamera()->GetSceneNode();
    if( NULL == camera )
        return;

    nw::gfx::ResLookAtTargetViewUpdater resViewUpdater =
            nw::gfx::ResDynamicCast<nw::gfx::ResLookAtTargetViewUpdater>(
            camera->GetViewUpdater()->GetResource());
    if( resViewUpdater.IsValid() )
    {
        const nn::math::VEC3 cameraTarget = DirectionalObj::GetDirectionalInfo().TargetPos;
        const nn::math::VEC3 cameraUp     = DirectionalObj::GetDirectionalInfo().UpVector;
        resViewUpdater.SetTargetPosition(cameraTarget);
        resViewUpdater.SetUpwardVector(cameraUp);
    }

    nw::gfx::ResPerspectiveProjectionUpdater resProjectionUpdater =
            nw::gfx::ResStaticCast<nw::gfx::ResPerspectiveProjectionUpdater>(
            camera->GetProjectionUpdater()->GetResource());
    if( resProjectionUpdater.IsValid() )
    {
        resProjectionUpdater.SetNear( g3d::DirectionalObj::GetDirectionalInfo().Near );
        resProjectionUpdater.SetFar( g3d::DirectionalObj::GetDirectionalInfo().Far );
        resProjectionUpdater.SetFovy( g3d::DirectionalObj::GetDirectionalInfo().Fovy );

        f32 aspectRatio = 240.0f / 400.0f;
        resProjectionUpdater.SetAspectRatio( aspectRatio );
    }

    const nn::math::VEC3 cameraPos = DirectionalObj::GetDirectionalInfo().Position;
    camera->Transform().SetTranslate(cameraPos);

    g3d::Camera* g3dCamera = NULL;
    for( u32 i=0; i<m_CameraArray.Size(); ++i )
    {
        if( camera == m_CameraArray[i].GetCamera() )
            g3dCamera = &m_CameraArray[i];
    }

    if( g3dCamera )
    {
        g3dCamera->SetPreset( DirectionalObj::GetDirectionalInfo().Position,
                              DirectionalObj::GetDirectionalInfo().TargetPos,
                              DirectionalObj::GetDirectionalInfo().UpVector );
    }
}

}   /* namespace g3d */
}   /* namespace vw  */
