/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#include <viewer/system.h>
#include "../include/Config3D.h"

#include "../include/ModuleG3dImpl.h"
#include "../include/SubModeManager.h"
#include "../include/InfoDrawer.h"
#include "../include/SceneManager.h"
#include "../include/SceneRenderer.h"
#include "../include/AnimManager.h"
#include "../include/Grid.h"
#include "../../../include/viewer/g3d/CameraMgr.h"
#include "../../../include/viewer/g3d/LightMgr.h"
#include "../include/Setting3DMisc.h"

using namespace vw::common;
using namespace vw::system;

namespace vw  {
namespace g3d {

static void     Config3D_Camera_( void );
static void     Config3D_Light_( void );
static void     Config3D_Fog_( void );
static void     Config3D_AmbLight_( void );
static void     Config3D_Information_( void );
static void     Config3D_Setting_( void );

static void     Config3D_WriteBack_Camera_( void );
static void     Config3D_WriteBack_Light_( void );
static void     Config3D_WriteBack_Fog_( void );
static void     Config3D_WriteBack_AmbLight_( void );
static void     Config3D_WriteBack_Information_( void );
static void     Config3D_WriteBack_Setting_( void );

/*---------------------------------------------------------------------------*
  Name:         Config3D_Execute
  Description:  3D-related configuration settings
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
Config3D_Execute( void )
{
    if ( GetConfigFile() == NULL )
    {
        return;
    }

    Config3D_Camera_();
    Config3D_Information_();
    Config3D_Light_();
    Config3D_AmbLight_();
    Config3D_Setting_();
}

/*---------------------------------------------------------------------------*
  Name:         Config3D_Camera_
  Description:  Camera initial setting
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------*/
static void
Config3D_Camera_( void )
{
    f32     fValue;

    DirectionalInfo info;
    info.RotationInit = 0.0f;
    info.RotationStep = 0.0f;
    info.RotationMax = 0.0f;
    info.TranslationInit = 0.0f;
    info.TranslationStep = 0.0f;
    info.TranslationMax = 0.0f;
    info.TiltInit = 0.0f;
    info.DollyInit = 0.0f;
    info.TiltAccel = 0.0f;
    info.TiltDecel = 0.0f;
    info.TiltLimit = 0.0f;
    info.SlideAccel = 0.0f;
    info.SlideDecel = 0.0f;
    info.SlideLimit = 0.0f;
    info.DollyAccel = 0.0f;
    info.DollyDecel = 0.0f;


    // Initial value for rotation speed.
    if ( GetConfigFile()->GetFloat("/g3d/camera/rotation_init", &fValue ) )
    {
        if ( fValue > 0 )
        {
            info.RotationInit = fValue;
        }
    }
    // Acceleration of rotation speed in one frame.
    if ( GetConfigFile()->GetFloat("/g3d/camera/rotation_step", &fValue ) )
    {
        if ( fValue > 0 )
        {
            info.RotationStep = fValue;
        }
    }
    // Maximum rotation speed.
    if ( GetConfigFile()->GetFloat("/g3d/camera/rotation_max", &fValue ) )
    {
        if ( fValue > 0 )
        {
            info.RotationMax = fValue;
        }
    }

    // Initial value for parallel translation speed.
    if ( GetConfigFile()->GetFloat("/g3d/camera/translation_init", &fValue ) )
    {
        if ( fValue > 0 )
        {
            info.TranslationInit = fValue;
        }
    }

    // Acceleration of parallel translation speed in one frame.
    if ( GetConfigFile()->GetFloat("/g3d/camera/translation_step", &fValue ) )
    {
        if ( fValue > 0 )
        {
            info.TranslationStep = fValue;
        }
    }

    // Maximum value for parallel translation speed.
    if ( GetConfigFile()->GetFloat("/g3d/camera/translation_max", &fValue ) )
    {
        if ( fValue > 0 )
        {
            info.TranslationMax = fValue;
        }
    }

    // Initial value for tilt speed
    if ( GetConfigFile()->GetFloat("/g3d/camera/tilt_init", &fValue ) )
    {
        if ( fValue > 0 )
        {
            info.TiltInit = fValue;
        }
    }

    // Initial value for dolly speed.
    if ( GetConfigFile()->GetFloat("/g3d/camera/dolly_init", &fValue ) )
    {
        if ( fValue > 0 )
        {
            info.DollyInit = fValue;
        }
    }

    // near clipping setting
    if ( GetConfigFile()->GetFloat("/g3d/camera/near", &fValue ) )
    {
        if ( fValue > 0 )
        {
            info.Near = fValue;
        }
    }

    // far clipping setting
    if ( GetConfigFile()->GetFloat("/g3d/camera/far", &fValue ) )
    {
        if ( fValue > 0 )
        {
            info.Far = fValue;
        }
    }

    // Default field of view angle for camera.
    if ( GetConfigFile()->GetFloat("/g3d/camera/fovy", &fValue ) )
    {
        if ( fValue > 0 )
        {
            info.Fovy = fValue;
        }
    }

    // Default position for camera.
    if ( GetConfigFile()->GetFloat("/g3d/camera/position/x", &fValue ) )
    {
        info.Position.x = fValue;
    }
    if ( GetConfigFile()->GetFloat("/g3d/camera/position/y", &fValue ) )
    {
        info.Position.y = fValue;
    }
    if ( GetConfigFile()->GetFloat("/g3d/camera/position/z", &fValue ) )
    {
        info.Position.z = fValue;
    }

    // Target position for default camera
    if ( GetConfigFile()->GetFloat("/g3d/camera/target_pos/x", &fValue ) )
    {
        info.TargetPos.x = fValue;
    }
    if ( GetConfigFile()->GetFloat("/g3d/camera/target_pos/y", &fValue ) )
    {
        info.TargetPos.y = fValue;
    }
    if ( GetConfigFile()->GetFloat("/g3d/camera/target_pos/z", &fValue ) )
    {
        info.TargetPos.z = fValue;
    }

    // Default up vector for camera.
    if ( GetConfigFile()->GetFloat("/g3d/camera/up_vector/x", &fValue ) )
    {
        info.UpVector.x = fValue;
    }
    if ( GetConfigFile()->GetFloat("/g3d/camera/up_vector/y", &fValue ) )
    {
        info.UpVector.y = fValue;
    }
    if ( GetConfigFile()->GetFloat("/g3d/camera/up_vector/z", &fValue ) )
    {
        info.UpVector.z = fValue;
    }
    if(nw::math::VEC3Len(&info.UpVector) == 0.0f)
    {
        info.UpVector.y = 1.0f;
    }

    s32  iValue;
    // Flag for inverting camera rotation operations up/down.
    if ( GetConfigFile()->GetInt("/g3d/camera/tumble_x_inverse", &iValue ) )
    {
        info.TumbleXInverse = ( iValue != 0 );
    }
    // Flag for inverting camera rotation operations left/right.
    if ( GetConfigFile()->GetInt("/g3d/camera/tumble_y_inverse", &iValue ) )
    {
        info.TumbleYInverse = ( iValue != 0 );
    }

    // Flag for inverting camera parallel translation operations up/down.
    if ( GetConfigFile()->GetInt("/g3d/camera/track_x_inverse", &iValue ) )
    {
        info.TrackXInverse = ( iValue != 0 );
    }
    // Flag for inverting camera parallel translation operations left/right.
    if ( GetConfigFile()->GetInt("/g3d/camera/track_y_inverse", &iValue ) )
    {
        info.TrackYInverse = ( iValue != 0 );
    }

    // Flag for inverting camera dolly operations.
    if ( GetConfigFile()->GetInt("/g3d/camera/dolly_inverse", &iValue ) )
    {
        info.DollyInverse = ( iValue != 0 );
    }
    // Inversion flag for camera operation forward/backward movement
    if ( GetConfigFile()->GetInt("/g3d/camera/dolly_target_inverse", &iValue ) )
    {
        info.DollyTargetInverse = ( iValue != 0 );
    }

    // Flag for inverting camera tilt operations up/down.
    if ( GetConfigFile()->GetInt("/g3d/camera/tilt_x_inverse", &iValue ) )
    {
        info.TiltXInverse = ( iValue != 0 );
    }
    // Flag for inverting camera tilt operations left/right.
    if ( GetConfigFile()->GetInt("/g3d/camera/tilt_y_inverse", &iValue ) )
    {
        info.TiltYInverse = ( iValue != 0 );
    }

    // Acceleration amount for pixels during forward/backward movement
    if ( GetConfigFile()->GetFloat("/g3d/camera/dolly_accel", &fValue ) )
    {
        if ( fValue > 0 )
        {
            info.DollyAccel = fValue;
        }
    }
    // Deceleration amount in one frame during forward/backward movement
    if ( GetConfigFile()->GetFloat("/g3d/camera/dolly_decel", &fValue ) )
    {
        if ( fValue > 0 )
        {
            info.DollyDecel = fValue;
        }
    }
    // Limiting speed in one frame during forward/backward movement
    if ( GetConfigFile()->GetFloat("/g3d/camera/dolly_limit", &fValue ) )
    {
        if ( fValue > 0 )
        {
            info.DollyLimit = fValue;
        }
    }
    // Specifies the distance threshold between the viewpoint and the camera so the viewpoint does not shift while moving back and forth.
    if ( GetConfigFile()->GetFloat("/g3d/camera/dolly_threshold", &fValue ) )
    {
        if ( fValue > 0 )
        {
            info.DollyThreshold = fValue;
        }
    }

    // Acceleration amount for a pixel in tilt rotation
    if ( GetConfigFile()->GetFloat("/g3d/camera/tilt_accel", &fValue ) )
    {
        if ( fValue > 0 )
        {
            info.TiltAccel = fValue;
        }
    }
    // Deceleration amount in one frame for tilt rotation
    if ( GetConfigFile()->GetFloat("/g3d/camera/tilt_decel", &fValue ) )
    {
        if ( fValue > 0 )
        {
            info.TiltDecel = fValue;
        }
    }
    // Limiting speed in one frame for tilt rotation
    if ( GetConfigFile()->GetFloat("/g3d/camera/tilt_limit", &fValue ) )
    {
        if ( fValue > 0 )
        {
            info.TiltLimit = fValue;
        }
    }


    // Acceleration amount for pixels during slide movement
    if ( GetConfigFile()->GetFloat("/g3d/camera/slide_accel", &fValue ) )
    {
        if ( fValue > 0 )
        {
            info.SlideAccel = fValue;
        }
    }
    // Deceleration amount in one frame during slide movement
    if ( GetConfigFile()->GetFloat("/g3d/camera/slide_decel", &fValue ) )
    {
        if ( fValue > 0 )
        {
            info.SlideDecel = fValue;
        }
    }
    // Limiting speed for one frame during slide movement
    if ( GetConfigFile()->GetFloat("/g3d/camera/slide_limit", &fValue ) )
    {
        if ( fValue > 0 )
        {
            info.SlideLimit = fValue;
        }
    }

    // Touch panel operation mode
    const char* light_kind = GetConfigFile()->GetString("/g3d/camera/touch_panel_mode");
    if ( light_kind )
    {
        switch ( *light_kind )
        {
        case 'N':
        case 'n':
            info.TouchPanelMode = TOUCH_PANEL_MODE_NORMAL;
            break;
        case 'D':
        case 'd':
            info.TouchPanelMode = TOUCH_PANEL_MODE_DRAG;
            break;
        case 'I':
        case 'i':
            info.TouchPanelMode = TOUCH_PANEL_MODE_INERTIA;
            break;
        }
    }

    // Focus scale
    if ( GetConfigFile()->GetFloat("/g3d/camera/focus_scale", &fValue ) )
    {
        if (fValue < 0.01f )
        {
            fValue = 0.01f;
        }
        if (fValue > 10.0f )
        {
            fValue = 10.0f;
        }
        info.FocusScale = fValue;
    }

    // Set the virtual space distance from the camera to the location for placement on the LCD
    if ( GetConfigFile()->GetFloat("/g3d/camera/depth_level", &fValue ) )
    {
        info.depthLevel = fValue;
    }

    if ( GetConfigFile()->GetFloat("/g3d/camera/depth_level_bias", &fValue ) )
    {
        info.depthLevelBias = fValue;
    }

    // Stereoscopic adjustment factor (0 = No parallax. 1 = Standard.)
    if ( GetConfigFile()->GetFloat("/g3d/camera/factor", &fValue ) )
    {
        if (fValue < 0.01f )
        {
            fValue = 0.01f;
        }
        if (fValue > 1.0f )
        {
            fValue = 1.0f;
        }
        info.factor = fValue;
    }

    // Permit change in field-of-view angle for more realism?
    if ( GetConfigFile()->GetInt("/g3d/camera/real_switch", &iValue ) )
    {
        info.realSwitch = ( iValue != 0 );
    }

    DirectionalObj::SetDirectionalInfo(info);
    CameraMgr::GetInstance()->ReloadSetting();

    // Force WScale?
    if ( GetConfigFile()->GetInt("/g3d/camera/force_wscale", &iValue ) )
    {
        SceneManager::GetInstance()->SetForceUseCameraWScale( iValue != 0  );
    }
}


/*---------------------------------------------------------------------------*
  Name:         Config3D_Information_
  Description:  Initial settings for 3D information display system
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------*/
static void
Config3D_Information_( void )
{
    s32   iValue;
    f32   fValue;
    u32   uiValue;

    bool grid_visible = true;
    nw::ut::FloatColor grid_color;
    f32 grid_scale = 1.0f;
    s32 grid_division = 10;
    f32 grid_linewidth = 1.0f;

    vw::g3d::Grid* grid =  SceneRenderer::GetInstance()->GetGrid();
    if( grid )
    {
        // Display enable flag of the grid
        if ( GetConfigFile()->GetInt("/g3d/information/grid_visible", &iValue ) )
        {
            grid_visible = ( iValue != 0 );
        }

        // Grid color
        if ( GetConfigFile()->GetUInt("/g3d/information/grid_color", &uiValue) )
        {
            grid_color = nw::ut::FloatColor(uiValue);
        }

        // Grid line size
        if ( GetConfigFile()->GetFloat("/g3d/information/grid_scale", &fValue ) )
        {
            if ( fValue < 0.001f )  { fValue = 0.001f;  }
            if ( fValue > 1000.0f ) { fValue = 1000.0f; }
            grid_scale = fValue;
        }

        // Number of grid lines
        if ( GetConfigFile()->GetInt("/g3d/information/grid_num", &iValue ) )
        {
            if ( iValue > 0 )
            {
                if ( iValue > 20 )
                {
                    iValue = 20;
                }
                grid_division = iValue;
            }
        }

        // Line width for grid.
        if ( GetConfigFile()->GetFloat("/g3d/information/grid_linewidth", &fValue ) )
        {
            grid_linewidth = fValue;
        }

        // set grid
        SceneRenderer::GetInstance()->GetGrid()->SetMeshColor(grid_color);
        SceneRenderer::GetInstance()->GetGrid()->CreateBuffer(grid_scale,grid_division);
        SceneRenderer::GetInstance()->GetGrid()->SetLineWidth(grid_linewidth);

        // Processing when grid is not displayed
        if(!grid_visible)
        {
            InfoDrawer::GetInstance()->SetRenderInfoFlag(1);
        }
    }

    // Coefficient for adjusting axis display width (excluding the camera view axis)
    f32 maniScale = 1.0f;
    if(GetConfigFile()->GetFloat("/g3d/information/manipulator_scale", &fValue))
    {
        maniScale = fValue;
    }
    SceneRenderer::GetInstance()->SetManipulatorScale(maniScale);

    // Coefficient for adjusting the display width of the camera view axis.
    maniScale = 1.0f;
    if ( GetConfigFile()->GetFloat("/g3d/information/camera_manipulator_scale", &fValue) )
    {
        maniScale = fValue;
    }
    SceneRenderer::GetInstance()->SetCameraManipulatorScale( maniScale );
}


/*---------------------------------------------------------------------------*
  Name:         Config3D_Light_
  Description:  Light initial value
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------*/
static void
Config3D_Light_( void )
{
    s32  iValue;
    u32  uiValue;
    f32  fValue;

    LightElem* lightElem = SceneManager::GetInstance()->GetDefaultFragmentLight();
    nw::gfx::res::ResFragmentLight resFragmentLight = nw::ut::ResDynamicCast<nw::gfx::res::ResFragmentLight>(
                                                                             lightElem->GetSceneNode()->GetResSceneObject());

    // Light type of light 0
    const char* light_kind = GetConfigFile()->GetString("/g3d/light/light0_kind");
    if ( light_kind )
    {
        switch ( *light_kind )
        {
        case 'D':
        case 'd':
            resFragmentLight.SetLightKind(nw::gfx::ResFragmentLight::KIND_DIRECTIONAL);
            break;
        case 'S':
        case 's':
            resFragmentLight.SetLightKind(nw::gfx::ResFragmentLight::KIND_SPOT);
            break;
        case 'P':
        case 'p':
            resFragmentLight.SetLightKind(nw::gfx::ResFragmentLight::KIND_POINT);
            break;
        }
    }

    // Light 0 light color
    if ( GetConfigFile()->GetUInt("/g3d/light/light0_ambient_color", &uiValue ) )
    {
        resFragmentLight.SetAmbient( nw::ut::Color8(uiValue) );
    }
    if ( GetConfigFile()->GetUInt("/g3d/light/light0_diffuse_color", &uiValue ) )
    {
        resFragmentLight.SetDiffuse( nw::ut::Color8(uiValue) );
    }
    if ( GetConfigFile()->GetUInt("/g3d/light/light0_specular0_color", &uiValue ) )
    {
        resFragmentLight.SetSpecular0( nw::ut::Color8(uiValue) );
    }
    if ( GetConfigFile()->GetUInt("/g3d/light/light0_specular1_color", &uiValue ) )
    {
        resFragmentLight.SetSpecular1( nw::ut::Color8(uiValue) );
    }

    // Directional vector for light 0.
    nw::math::VEC3 direction;
    if ( GetConfigFile()->GetFloat("/g3d/light/light0_direction/x", &fValue ) )
    {
        direction.x = fValue;
    }
    if ( GetConfigFile()->GetFloat("/g3d/light/light0_direction/y", &fValue ) )
    {
        direction.y = fValue;
    }
    if ( GetConfigFile()->GetFloat("/g3d/light/light0_direction/z", &fValue ) )
    {
        direction.z = fValue;
    }
    if( nw::math::VEC3Len( &direction ) > 0.0f )
        direction.Normalize();
    resFragmentLight.SetDirection( direction );

    // Set light position
    f32 dist = 35.0f;
    lightElem->GetSceneNode()->Transform().SetTranslate(-direction * dist);

    // G3D light reset
    LightMgr::GetInstance()->ReloadDefaultLight( lightElem->GetSceneNode() );
}


/*---------------------------------------------------------------------------*
  Name:         Config3D_AmbLight_
  Description:  Sets default scene ambient light.
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
static void
Config3D_AmbLight_( void )
{
    u32  uiValue;
    if ( GetConfigFile()->GetUInt("/g3d/light/global_ambient_color", &uiValue ) )
    {
        LightElem* lightElem = SceneManager::GetInstance()->GetDefaultAmbientLight();
        nw::gfx::res::ResAmbientLight ambientLight = nw::ut::ResDynamicCast<nw::gfx::res::ResAmbientLight>(
                                                                            lightElem->GetSceneNode()->GetResSceneObject());
        ambientLight.SetAmbient( nw::ut::Color8(uiValue) );
    }
}


/*---------------------------------------------------------------------------*
  Name:         Config3D_Setting_
  Description:  Performs other miscellaneous 3D-related settings.
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
static void
Config3D_Setting_( void )
{
    s32   iValue;
    f32   fValue;
    u32   uiValue;
    // Default mode setting
    {
        const char* cmode = GetConfigFile()->GetString("/g3d/config/sub_mode");
        if( cmode )
        {
            if(0 == strcmp(cmode, "CAM"))
            {
                SubModeManager::GetInstance()->SetMode(SubMode_Camera);
            }
            else if(0 == strcmp(cmode, "MNU"))
            {
                SubModeManager::GetInstance()->SetMode(SubMode_Menu);
            }
            else if(0 == strcmp(cmode, "LOG"))
            {
                SubModeManager::GetInstance()->SetMode(SubMode_Log);
            }
        }
    }
    if ( GetConfigFile()->GetUInt("/g3d/config/blink_step_frame", &uiValue) )
    {
        SceneRenderer::GetInstance()->SetBlinkSpan(uiValue);
    }
    if ( GetConfigFile()->GetUInt("/g3d/config/num_blink", &uiValue) )
    {
        SceneRenderer::GetInstance()->SetBlinkNum(uiValue);
    }
    if ( GetConfigFile()->GetUInt("/g3d/config/loop_onetime_anim", &uiValue) )
    {
        AnimManager::GetInstance()->SetLoopFlag(uiValue);
    }

    // Default playback rate setting
    if ( GetConfigFile()->GetFloat("/g3d/animation/default_rate", &fValue ) )
    {
        AnimManager::GetInstance()->SetDefaultRate( fValue );
    }

    // 3DEditor Active channel number
    if ( GetConfigFile()->GetInt("/g3d/config/3DEditor_active", &iValue ) )
    {
        ModuleG3dImpl::GetInstance()->Set3DEditorActiveChannel( iValue );
    }

    // 3DEditor Passive channel number
    if ( GetConfigFile()->GetInt("/g3d/config/3DEditor_passive", &iValue ) )
    {
        ModuleG3dImpl::GetInstance()->Set3DEditorPassiveChannel( iValue );
    }

    // 3DEditor Survival channel number
    if ( GetConfigFile()->GetInt("/g3d/config/3DEditor_survival", &iValue ) )
    {
        ModuleG3dImpl::GetInstance()->Set3DEditorSurvivalChannel( iValue );
    }

    // Permission flag for communication with CS
    if ( GetConfigFile()->GetInt("/g3d/config/cs_connection", &iValue ) )
    {
        ModuleG3dImpl::GetInstance()->SetCSconnect( iValue );
    }

    const char* alloc_type;

    // Texture priority placement memory
    alloc_type = GetConfigFile()->GetString("/g3d/mem_location/tex");
    if ( alloc_type )
    {
        if ( std::strncmp( alloc_type, "VRAM", 4 ) == 0 )
        {
            Setting::GetInstance().SetTexAllocType( Setting::ALLOCTYPE_VRAM );
        }
        else if ( std::strncmp( alloc_type, "FCRAM", 5 ) == 0 )
        {
            Setting::GetInstance().SetTexAllocType( Setting::ALLOCTYPE_FCRAM );
        }
    }

    // Model priority placement memory
    alloc_type = GetConfigFile()->GetString("/g3d/mem_location/mdl");
    if ( alloc_type )
    {
        if ( std::strncmp( alloc_type, "VRAM", 4 ) == 0 )
        {
            Setting::GetInstance().SetModelAllocType( Setting::ALLOCTYPE_VRAM );
        }
        else if ( std::strncmp( alloc_type, "FCRAM", 5 ) == 0 )
        {
            Setting::GetInstance().SetModelAllocType( Setting::ALLOCTYPE_FCRAM );
        }
    }
}



/*---------------------------------------------------------------------------*
  Name:         Config3D_WriteBack
  Description:  Light initial value
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------*/
void
Config3D_WriteBack( void )
{
/*    Config3D_WriteBack_Camera_();
    Config3D_WriteBack_Information_();
    Config3D_WriteBack_Light_();
    Config3D_WriteBack_Fog_();
    Config3D_WriteBack_AmbLight_();
    Config3D_WriteBack_Setting_();*/
}



///*---------------------------------------------------------------------------*
//  Name:         Config3D_WriteBack_Camera_
//  Description:  Outputs camera settings
//  Arguments:    None.
//  Returns:      None.
// *---------------------------------------------------------------------------*/
//static void
//Config3D_WriteBack_Camera_( void )
//{
//    // Default camera distance (no change)
//
//    // near clipping setting
//    GetConfigFile()->SetFloat("/g3d/projection/near", GetProjection().GetNear() );
//    // far clipping setting
//    GetConfigFile()->SetFloat("/g3d/projection/far", GetProjection().GetFar() );
//}
//
//
///*---------------------------------------------------------------------------*
//  Name:         Config3D_WriteBack_Information_
//  Description:  Initial settings for 3D information display system
//  Arguments:    None.
//  Returns:      None.
// *---------------------------------------------------------------------------*/
//static void
//Config3D_WriteBack_Information_( void )
//{
//    // Display enable flag of the camera viewpoint
//    GetConfigFile()->SetInt("/g3d/information/camera_aim_visible", GetStage().IsVisibleCameraTarget() );
//
//    // Display enable flag of the coordinate axis
//    GetConfigFile()->SetInt("/g3d/information/axis_visible", GetGrid().GetAxisVisible() );
//
//    // Display enable flag of the grid
//    GetConfigFile()->SetInt("/g3d/information/grid_visible", GetGrid().GetMeshVisible() );
//
//    // Grid line size
//    GetConfigFile()->SetFloat("/g3d/information/grid_scale", GetGrid().GetScale() );
//
//    // Number of grid lines
//    GetConfigFile()->SetInt("/g3d/information/grid_num", s32( GetGrid().GetDivision() ) );
//
//    // Grid color
//    GetConfigFile()->SetUInt16("/g3d/information/grid_color", u32( nw4r::ut::Color( GetGrid().GetMeshColor() ) ) );
//
//    // Wireframe Z-comparison enable flag
//    GetConfigFile()->SetInt("/g3d/information/wireframe_z_compare_enable", GetStage().IsZCompareOnWireFrame());
//}
//
//
//
///*---------------------------------------------------------------------------*
//  Name:         Config3D_WriteBack_Light_
//  Description:  Light initial value
//  Arguments:    None.
//  Returns:      None.
// *---------------------------------------------------------------------------*/
//static void
//Config3D_WriteBack_Light_( void )
//{
//    const char* LIGHT_TYPE_TABLE[] = { "POINT", "DIRECTIONAL", "SPOT" };
//    const char* DISTFUNC_TABLE[]   = { "OFF", "GENTLE", "MEDIUM", "STEEP" };
//    const char* SPOTFUNC_TABLE[]   = { "OFF", "FLAT", "COS", "COS2", "SHARP", "RING1", "RING2" };
//
//    // Light 3 information display flag
//    GetConfigFile()->SetInt("/g3d/light/light0_visible", GetDefaultLight( 0 )->IsVisible()? 1:0 );
//    // Light 0 camera tracking flag
//    GetConfigFile()->SetInt("/g3d/light/light0_follow_camera", GetDefaultLight( 0 )->IsFollowCam()? 1:0 );
//    // Light 0 light type
//    GetConfigFile()->SetString("/g3d/light/light0_type", LIGHT_TYPE_TABLE[ GetDefaultLight( 0 )->GetLightType() ]);
//    /  Light 0 light color
//    GetConfigFile()->SetUInt16("/g3d/light/light0_diffuse_color", u32( nw4r::ut::Color( GetDefaultLight( 0 )->GetColor() ) ) );
//    GetConfigFile()->SetUInt16("/g3d/light/light0_specular_color", u32( nw4r::ut::Color( GetDefaultLight( 0 )->GetSpecColor() ) ) );
//    // Light 0 degree of brilliance parameter
//    GetConfigFile()->SetFloat("/g3d/light/light0_shininess", GetDefaultLight( 0 )->GetShininess());
//    // Light 0 distance attenuation parameter
//    GetConfigFile()->SetFloat("/g3d/light/light0_ref_distance", GetDefaultLight( 0 )->GetRefDistance());
//    GetConfigFile()->SetFloat("/g3d/light/light0_ref_brightness", GetDefaultLight( 0 )->GetRefBrightness());
//    GetConfigFile()->SetString("/g3d/light/light0_dist_func", DISTFUNC_TABLE[ GetDefaultLight( 0 )->GetDistFn() ]);
//    // Light 0 angular attenuation parameter
//    GetConfigFile()->SetFloat("/g3d/light/light0_cut_off", GetDefaultLight( 0 )->GetCutOff() );
//    GetConfigFile()->SetString("/g3d/light/light0_spot_func", SPOTFUNC_TABLE[ GetDefaultLight( 0 )->GetSpotFn() ]);
//
//    // Light 1 information display flag
//    GetConfigFile()->SetInt("/g3d/light/light1_visible", GetDefaultLight( 1 )->IsVisible()? 1:0 );
//    // Light 1 camera tracking flag
//    GetConfigFile()->SetInt("/g3d/light/light1_follow_camera", GetDefaultLight( 1 )->IsFollowCam()? 1:0 );
//    // Light 1 light type
//    GetConfigFile()->SetString("/g3d/light/light1_type", LIGHT_TYPE_TABLE[ GetDefaultLight( 1 )->GetLightType() ]);
//    /  Light 1 light color
//    GetConfigFile()->SetUInt16("/g3d/light/light1_diffuse_color", u32( nw4r::ut::Color( GetDefaultLight( 1 )->GetColor() ) ) );
//    GetConfigFile()->SetUInt16("/g3d/light/light1_specular_color", u32( nw4r::ut::Color( GetDefaultLight( 1 )->GetSpecColor() ) ) );
//    // Light 1 degree of brilliance parameter
//    GetConfigFile()->SetFloat( "/g3d/light/light1_shininess", GetDefaultLight( 1 )->GetShininess() );
//    // Light 1 distance attenuation parameter
//    GetConfigFile()->SetFloat("/g3d/light/light1_ref_distance", GetDefaultLight( 1 )->GetRefDistance());
//    GetConfigFile()->SetFloat("/g3d/light/light1_ref_brightness", GetDefaultLight( 1 )->GetRefBrightness());
//    GetConfigFile()->SetString("/g3d/light/light1_dist_func", DISTFUNC_TABLE[ GetDefaultLight( 1 )->GetDistFn() ]);
//    // Light 1 angular attenuation parameter
//    GetConfigFile()->SetFloat("/g3d/light/light1_cut_off", GetDefaultLight( 1 )->GetCutOff() );
//    GetConfigFile()->SetString("/g3d/light/light1_spot_func", SPOTFUNC_TABLE[ GetDefaultLight( 1 )->GetSpotFn() ]);
//
//    // Light 2 information display flag
//    GetConfigFile()->SetInt("/g3d/light/light2_visible", GetDefaultLight( 2 )->IsVisible()? 1:0 );
//    // Light 2 camera tracking flag
//    GetConfigFile()->SetInt("/g3d/light/light2_follow_camera", GetDefaultLight( 2 )->IsFollowCam()? 1:0 );
//    // Light 2 light type
//    GetConfigFile()->SetString("/g3d/light/light2_type", LIGHT_TYPE_TABLE[ GetDefaultLight( 2 )->GetLightType() ]);
//    /  Light 2 light color
//    GetConfigFile()->SetUInt16("/g3d/light/light2_diffuse_color", u32( nw4r::ut::Color( GetDefaultLight( 2 )->GetColor() ) ) );
//    GetConfigFile()->SetUInt16("/g3d/light/light2_specular_color", u32( nw4r::ut::Color( GetDefaultLight( 2 )->GetSpecColor() ) ) );
//    // Light 2 degree of brilliance parameter
//    GetConfigFile()->SetFloat( "/g3d/light/light2_shininess", GetDefaultLight( 2 )->GetShininess() );
//    // Light 2 distance attenuation parameter
//    GetConfigFile()->SetFloat("/g3d/light/light2_ref_distance", GetDefaultLight( 2 )->GetRefDistance());
//    GetConfigFile()->SetFloat("/g3d/light/light2_ref_brightness", GetDefaultLight( 2 )->GetRefBrightness());
//    GetConfigFile()->SetString("/g3d/light/light2_dist_func", DISTFUNC_TABLE[ GetDefaultLight( 2 )->GetDistFn() ]);
//    // Light 2 angular attenuation parameter
//    GetConfigFile()->SetFloat("/g3d/light/light2_cut_off", GetDefaultLight( 2 )->GetCutOff() );
//    GetConfigFile()->SetString("/g3d/light/light2_spot_func", SPOTFUNC_TABLE[ GetDefaultLight( 2 )->GetSpotFn() ]);
//
//    // Light 3 information display flag
//    GetConfigFile()->SetInt("/g3d/light/light3_visible", GetDefaultLight( 3 )->IsVisible()? 1:0 );
//    // Light 3 camera tracking flag
//    GetConfigFile()->SetInt("/g3d/light/light3_follow_camera", GetDefaultLight( 3 )->IsFollowCam()? 1:0 );
//    // Light 3 light type
//    GetConfigFile()->SetString("/g3d/light/light3_type", LIGHT_TYPE_TABLE[ GetDefaultLight( 3 )->GetLightType() ]);
//    /  Light 3 light color
//    GetConfigFile()->SetUInt16("/g3d/light/light3_diffuse_color", u32( nw4r::ut::Color( GetDefaultLight( 3 )->GetColor() ) ) );
//    GetConfigFile()->SetUInt16("/g3d/light/light3_specular_color", u32( nw4r::ut::Color( GetDefaultLight( 3 )->GetSpecColor() ) ) );
//    // Light 3 degree of brilliance parameter
//    GetConfigFile()->SetFloat( "/g3d/light/light3_shininess", GetDefaultLight( 3 )->GetShininess() );
//    // Light 3 distance attenuation parameter
//    GetConfigFile()->SetFloat("/g3d/light/light3_ref_distance", GetDefaultLight( 3 )->GetRefDistance());
//    GetConfigFile()->SetFloat("/g3d/light/light3_ref_brightness", GetDefaultLight( 3 )->GetRefBrightness());
//    GetConfigFile()->SetString("/g3d/light/light3_dist_func", DISTFUNC_TABLE[ GetDefaultLight( 3 )->GetDistFn() ]);
//    // Light 3 angular attenuation parameter
//    GetConfigFile()->SetFloat("/g3d/light/light3_cut_off", GetDefaultLight( 3 )->GetCutOff() );
//    GetConfigFile()->SetString("/g3d/light/light3_spot_func", SPOTFUNC_TABLE[ GetDefaultLight( 3 )->GetSpotFn() ]);
//
//}
//
//
///*---------------------------------------------------------------------------*
//  Name:         Config3D_WriteBack_Fog_
//  Description:  Writes back the default scene fog settings to the INI file.
//  Arguments:    None.
//  Returns:      None.
// *---------------------------------------------------------------------------*/
//static void
//Config3D_WriteBack_Fog_( void )
//{
//    const char* FOG_TYPE_TABLE[] = { "NONE", "NONE", "LINE", "NONE", "EXP", "EXP2", "REVEXP", "REVEXP2" };
//
//    nw4r::g3d::Fog fog = GetResource3D().GetScnDefault().GetFog();
//    u8 fogType = u8( fog.ref().type );
//
//    // Fog function
//    GetConfigFile()->SetString("/g3d/fog/type", FOG_TYPE_TABLE[ fogType & 0x7 ] );
//    // Fog projection type
//    GetConfigFile()->SetString("/g3d/fog/projection_mode", (fogType & 0x8)? "ORTHO" : "PERSP" );
//    // Fog start Z value
//    GetConfigFile()->SetFloat("/g3d/fog/start_z", fog.ref().startz);
//    // Fog start Z value
//    GetConfigFile()->SetFloat("/g3d/fog/end_z",  fog.ref().endz);
//    // Fog color
//    GetConfigFile()->SetUInt16("/g3d/fog/color", u32( nw4r::ut::Color( fog.ref().color ) ) );
//}
//
///*---------------------------------------------------------------------------*
//  Name:         Config3D_WriteBack_AmbLight_
//  Description:  Writes back the default scene ambient light settings to the INI file.
//  Arguments:    None.
//  Returns:      None.
// *---------------------------------------------------------------------------*/
//static void
//Config3D_WriteBack_AmbLight_( void )
//{
//    GXColor col = GetResource3D().GetScnDefault().GetAmbLightColor();
//    // Ambient color
//    GetConfigFile()->SetUInt16("/g3d/ambient_light/color", u32( nw4r::ut::Color( col ) ) );
//}
//
//
///*---------------------------------------------------------------------------*
//  Name:         Config3D_WriteBack_Setting_
//  Description:  Writes back the other 3D-related settings to the INI file.
//  Arguments:    None.
//  Returns:      None.
// *---------------------------------------------------------------------------*/
//static void
//Config3D_WriteBack_Setting_( void )
//{
//    u8 texLocation  = Setting::GetInstance().GetAllocType( Setting::RESTYPE_BRTEX  );
//    u8 plttLocation = Setting::GetInstance().GetAllocType( Setting::RESTYPE_BRPLTT );
//    u8 mdlLocation  = Setting::GetInstance().GetAllocType( Setting::RESTYPE_BRMDL  );
//
//    GetConfigFile()->SetString("/g3d/mem_location/tex",  (texLocation  == Setting::ALLOCTYPE_MEM1)? "MEM1" : "MEM2" );
//    GetConfigFile()->SetString("/g3d/mem_location/pltt", (plttLocation == Setting::ALLOCTYPE_MEM1)? "MEM1" : "MEM2" );
//    GetConfigFile()->SetString("/g3d/mem_location/mdl",  (mdlLocation  == Setting::ALLOCTYPE_MEM1)? "MEM1" : "MEM2" );
//}


}   /* namespace g3d */
}   /* namespace vw  */

