/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/


//-----------------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------------
#include <viewer/system.h>
#include "../../../include/viewer/g3d/DirectionalObj.h"
#include "../../../include/viewer/g3d/CameraMgr.h"


namespace vw {
namespace g3d {

//----------------------------------------------------------------------------
// Global Variables
//----------------------------------------------------------------------------
static DirectionalInfo sDirectionalInfo;


//----------------------------------------------------------------------------
// Internal Functions
//----------------------------------------------------------------------------

//----------------------------------------------------------------------------
//
//
//----------------------------------------------------------------------------
float DegreeToRadian(float deg)
{
    return deg*0.01745329252f;
}

//----------------------------------------------------------------------------
//
//
//
//
//
//----------------------------------------------------------------------------
static bool ComputeAngle
(
    const nw::math::VEC3&   cameraPos,
    const nw::math::VEC3&   targetPos,
    const f32               twist,
    nw::math::VEC3&         result
)
{
    nw::math::VEC3 lookReverse = ( cameraPos - targetPos );
    f32 dist = nw::math::VEC3Len( &lookReverse );
    if ( dist == 0.0f ) { return false; }
    nw::math::VEC3Normalize( &lookReverse, &lookReverse );

    result.x = nw::math::AsinFIdx( -lookReverse.y );
    result.y = nw::math::Atan2FIdx( -lookReverse.x, lookReverse.z );
    result.z = twist;

    return true;
}

//----------------------------------------------------------------------------
//
//
//
//
//----------------------------------------------------------------------------
static void ComputeRotateFIdx
(
    const f32           rotateX,
    const f32           rotateY,
    nw::math::VEC3&     result
)
{
    f32 xs = nw::math::SinFIdx( rotateX );
    f32 xc = nw::math::CosFIdx( rotateX );

    f32 ys = nw::math::SinFIdx( rotateY );
    f32 yc = nw::math::CosFIdx( rotateY );

    result.x = xc * ys;
    result.y = xs;
    result.z = xc * yc;
}

//----------------------------------------------------------------------------
//
//
//
//
//----------------------------------------------------------------------------
void ComputeRotateRad
(
    const f32           rotateX,
    const f32           rotateY,
    nw::math::VEC3&     result
)
{
    f32 xs = nw::math::SinRad( rotateX );
    f32 xc = nw::math::CosRad( rotateX );

    f32 ys = nw::math::SinRad( rotateY );
    f32 yc = nw::math::CosRad( rotateY );

    result.x = xc * ys;
    result.y = xs;
    result.z = xc * yc;
}

//----------------------------------------------------------------------------
//
//
//
//----------------------------------------------------------------------------
static PadEvent MakePadEventFromHid
(
    vw::system::HidStatus& hidStatus,
    const bool isViewTypeRotate
)
{
    // Getting the control value
    bool isHoldR = hidStatus.GetHoldR();
    bool isHoldL = hidStatus.GetHoldL();
    bool isHoldX = hidStatus.GetHoldX();
    bool isHoldY = hidStatus.GetHoldY();

    bool isHoldStart = hidStatus.GetHoldStart();
    bool isHoldRight = hidStatus.GetHoldRight();
    bool isHoldLeft  = hidStatus.GetHoldLeft();
    bool isHoldDown  = hidStatus.GetHoldDown();
    bool isHoldUp    = hidStatus.GetHoldUp();
    bool isTrigStart = hidStatus.GetTriggerStart();

    f32 stickX = hidStatus.padReader->NormalizeStick( hidStatus.pad.stick.x );
    f32 stickY = hidStatus.padReader->NormalizeStick( hidStatus.pad.stick.y );

    f32 dbgLeftStickX  = hidStatus.debugPad.leftStickX;
    f32 dbgLeftStickY  = hidStatus.debugPad.leftStickY;
    f32 dbgRightStickX = hidStatus.debugPad.rightStickX;
    f32 dbgRightStickY = hidStatus.debugPad.rightStickY;

    // Pad event assembly
    PadEvent padEvent;

    // Rotate
    bool isHandleRotate = false;
    if ( ( !isHoldL ) && ( !isHoldR ) )
    {
        if ( stickX != 0.0f ) padEvent.rotateH = stickX;
        if ( stickY != 0.0f ) padEvent.rotateV = stickY;

        if ( dbgLeftStickX != 0.0f ) padEvent.rotateH = dbgLeftStickX;
        if ( dbgLeftStickY != 0.0f ) padEvent.rotateV = dbgLeftStickY;

        padEvent.flags |= PadEvent::HANDLE_ROTATE;
    }

    // Dolly
    if ( ( !isHoldL ) && ( !isHoldR ) )
    {
        if ( isHoldX ) padEvent.dolly =  1.0f;
        if ( isHoldY ) padEvent.dolly = -1.0f;
        padEvent.moveTarget = false;
        padEvent.flags |= PadEvent::HANDLE_DOLLY;
    }
    else if ( isHoldL && ( !isHoldR ) )
    {
        if ( isHoldX ) padEvent.dolly =  1.0f;
        if ( isHoldY ) padEvent.dolly = -1.0f;
        padEvent.moveTarget = true;
        padEvent.flags |= PadEvent::HANDLE_DOLLY;
    }

    // Tilt
    bool isHandleTilt = false;
    if ( ( hidStatus.debugPad.hold & nn::hid::CTR::DEBUG_PAD_TRIGGER_ZL )
     && !( hidStatus.debugPad.hold & nn::hid::CTR::DEBUG_PAD_TRIGGER_ZR ) )
    {
        if ( dbgRightStickX != 0.0f ) padEvent.tiltH = dbgRightStickX * sDirectionalInfo.TiltInit;
        if ( dbgRightStickY != 0.0f ) padEvent.tiltV = dbgRightStickY * sDirectionalInfo.TiltInit;
        isHandleTilt = true;
    }
    if ( isHandleTilt )
    {
        padEvent.flags |= PadEvent::HANDLE_TILT;
    }

    // Slide
    if ( isHoldL && ( !isHoldR ) )
    {
        padEvent.slideX =  stickX;
        padEvent.slideZ = -stickY;

        if ( dbgLeftStickX != 0.0f ) padEvent.slideX =  dbgLeftStickX;
        if ( dbgLeftStickY != 0.0f ) padEvent.slideZ = -dbgLeftStickY;

        padEvent.flags |= PadEvent::HANDLE_SLIDE;
    }

    // Track
    bool isHandleTrack = false;
    if ( ( !isHoldL ) && ( !isHoldR ) )
    {
        if ( dbgRightStickX != 0.0f ) padEvent.trackH = dbgRightStickX;
        if ( dbgRightStickY != 0.0f ) padEvent.trackV = dbgRightStickY;
        isHandleTrack = true;
    }
    if ( ( !isHoldStart ) && ( !isHoldR ) )
    {
        if ( isHoldRight ) padEvent.trackH =  1.0f;
        if ( isHoldLeft )  padEvent.trackH = -1.0f;

        if ( isHoldDown )  padEvent.trackV = -1.0f;
        if ( isHoldUp )    padEvent.trackV =  1.0f;
        isHandleTrack = true;
    }
    if ( isHandleTrack )
    {
        // In the device viewer, parallel translation operations relative to the X-Y plane of the camera space do not occur.
        padEvent.moveInViewSpace = false;

        padEvent.flags |= PadEvent::HANDLE_TRACK;
    }

    // Zoom
    if ( !isHoldR )
    {
        if ( isHoldX ) padEvent.zoom = -1.0f;
        if ( isHoldY ) padEvent.zoom =  1.0f;
        padEvent.flags |= PadEvent::HANDLE_ZOOM;
    }

    // Reset
    if ( isHoldL && ( !isHoldR ) )
    {
        if ( isTrigStart )
        {
            padEvent.flags |= PadEvent::HANDLE_RESET;
        }
    }

    return padEvent;
}


/////////////////////////////////////////////////////////////////////////////
// DirectionalInfo structure
/////////////////////////////////////////////////////////////////////////////

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------
DirectionalInfo::DirectionalInfo()
{
    Fovy = DegreeToRadian(37.8f);
    Near = 0.01f;
    Far  = 10000.0f;
    Position.Set(50.0f, 30.0f, 50.0f);
    TargetPos.Set(0.0f, 0.0f, 0.0f);
    UpVector.Set(0.0f, 1.0f, 0.0f);
    RotationInit        = 0.05f;
    RotationStep        = 0.0f;
    RotationMax         = 0.1f;
    TranslationInit     = 0.004f;
    TranslationStep     = 0.001f;
    TranslationMax      = 0.01f;
    TiltInit            = 0.005f;
    DollyInit           = 0.01f;
    DollyAccel          = 0.01f;
    DollyDecel          = 0.01f;
    DollyLimit          = 0.01f;
    DollyThreshold      = 1.0f;
    TiltAccel           = 0.01f;
    TiltDecel           = 0.01f;
    TiltLimit           = 4.0f;
    SlideAccel          = 0.02f;
    SlideDecel          = 0.1f;
    SlideLimit          = 10.0f;
    TumbleXInverse      = false;
    TumbleYInverse      = false;
    TrackXInverse       = false;
    TrackYInverse       = false;
    DollyInverse        = false;
    DollyTargetInverse  = false;
    TiltXInverse        = false;
    TiltYInverse        = false;
    FocusScale          = 0.5f;
    depthLevel          = 0.0f;
    factor              = 1.0f;
    realSwitch          = true;
    TouchPanelMode      = TOUCH_PANEL_MODE_INERTIA;
    depthLevelBias      = 0.f;
}

/////////////////////////////////////////////////////////////////////////////
// PadEvent structure
/////////////////////////////////////////////////////////////////////////////

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------
PadEvent::PadEvent()
{
    flags      = 0x00000000;
    rotateH    = 0.0f;
    rotateV    = 0.0f;
    dolly      = 0.0f;
    tiltH      = 0.0f;
    tiltV      = 0.0f;
    slideX     = 0.0f;
    slideZ     = 0.0f;
    trackH     = 0.0f;
    trackV     = 0.0f;
    zoom       = 0.0f;
    moveTarget = false;
    moveInViewSpace = false;
}

/////////////////////////////////////////////////////////////////////////////
// DirectionalObj class
/////////////////////////////////////////////////////////////////////////////

//---------------------------------------------------------------------------
//
//
//---------------------------------------------------------------------------
void DirectionalObj::SetDirectionalInfo( vw::g3d::DirectionalInfo info )
{
    info.RotationInit = nw::ut::Clamp(info.RotationInit, 0.0001f, 100.0f);
    info.RotationStep = nw::ut::Clamp(info.RotationStep, 0.0001f, 100.0f);
    info.RotationMax  = nw::ut::Clamp(info.RotationMax, 0.0001f, 100.0f);
    sDirectionalInfo  = info;

    // Converting the rotation angles to radians
    sDirectionalInfo.RotationInit = DegreeToRadian(info.RotationInit);
    sDirectionalInfo.RotationStep = DegreeToRadian(info.RotationStep);
    sDirectionalInfo.RotationMax  = DegreeToRadian(info.RotationMax);
    sDirectionalInfo.TiltInit     = DegreeToRadian(info.TiltInit);
    sDirectionalInfo.Fovy         = DegreeToRadian(info.Fovy);
    sDirectionalInfo.TiltAccel    = DegreeToRadian(info.TiltAccel);
    sDirectionalInfo.TiltDecel    = DegreeToRadian(info.TiltDecel);
    sDirectionalInfo.TiltLimit    = DegreeToRadian(info.TiltLimit);
}

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------
DirectionalInfo& DirectionalObj::GetDirectionalInfo()
{
    return sDirectionalInfo;
}

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------
DirectionalObj::DirectionalObj()
{
    m_Translate  = nw::math::VEC3( 0.0f, 0.0f, 0.0f );
    m_TargetPos  = nw::math::VEC3( 0.0f, 0.0f, 0.0f );
    m_Upward     = nw::math::VEC3( 0.0f, 1.0f, 0.0f );
    m_InitUpward = nw::math::VEC3( 0.0f, 1.0f, 0.0f );
    m_Rotate     = nw::math::VEC3( 0.0f, 0.0f, 0.0f );

    m_Orientation    = nw::math::MTX33::Identity();
    m_Twist          = 0.0f;
    m_RotateSpeed    = 0.0f;
    m_TransSpeed     = 0.0f;
    m_TrackSpeed     = 0.0f;
    m_TiltXSpeed     = 1.0f;
    m_TiltYSpeed     = 1.0f;
    m_ProjTypeOrtho  = false;
    m_UpdateType     = DirectionalObj::UPDATE_TYPE_LOOKAT;
}

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------
DirectionalObj::~DirectionalObj()
{
    /* DO_NOTHING */
}

//---------------------------------------------------------------------------
//
//
//
//
//---------------------------------------------------------------------------
void DirectionalObj::SetPadEventFromHid
(
    vw::system::HidStatus& hidStatus,
    const bool isViewTypeRotate
)
{
    // Creates and sets pad events.
    m_PadEvent = MakePadEventFromHid( hidStatus, isViewTypeRotate );
}

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------
void DirectionalObj::SetPadEvent( PadEvent& event )
{
    m_PadEvent = event;
}

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------
PadEvent DirectionalObj::GetPadEvent() const
{
    return m_PadEvent;
}

//---------------------------------------------------------------------------
//
//
//
//
//---------------------------------------------------------------------------
void DirectionalObj::SetTranslate( const f32 x, const f32 y, const f32 z )
{
    m_Translate.Set( x, y, z );
}

//---------------------------------------------------------------------------
//
//
//---------------------------------------------------------------------------
void DirectionalObj::SetTranslate( const nw::math::VEC3& translate )
{
    m_Translate = translate;
}

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------
nw::math::VEC3 DirectionalObj::GetTranslate() const
{
    return m_Translate;
}

//---------------------------------------------------------------------------
//
//
//
//
//---------------------------------------------------------------------------
void DirectionalObj::SetTargetPos( const f32 x, const f32 y, const f32 z )
{
    m_TargetPos.Set( x, y, z );
}

//---------------------------------------------------------------------------
//
//
//---------------------------------------------------------------------------
void DirectionalObj::SetTargetPos( const nw::math::VEC3& targetPos )
{
    m_TargetPos = targetPos;
}

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------
nw::math::VEC3 DirectionalObj::GetTargetPos() const
{
    return m_TargetPos;
}

//---------------------------------------------------------------------------
//
//
//
//
//---------------------------------------------------------------------------
void DirectionalObj::SetUpward( const f32 x, const f32 y, const f32 z )
{
    m_Upward.Set( x, y, z );
}

//---------------------------------------------------------------------------
//
//
//---------------------------------------------------------------------------
void DirectionalObj::SetUpward( const nw::math::VEC3& upward )
{
    m_Upward = upward;
}

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------
nw::math::VEC3 DirectionalObj::GetUpward() const
{
    return m_Upward;
}

//---------------------------------------------------------------------------
//
//
//
//
//---------------------------------------------------------------------------
void DirectionalObj::SetRotate( const f32 x, const f32 y, const f32 z )
{
    m_Rotate.Set( x, y, z );
}

//---------------------------------------------------------------------------
//
//
//---------------------------------------------------------------------------
void DirectionalObj::SetRotate( const nw::math::VEC3& rotate )
{
    m_Rotate = rotate;
}

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------
nw::math::VEC3 DirectionalObj::GetRotate() const
{
    return m_Rotate;
}

//---------------------------------------------------------------------------
//
//
//---------------------------------------------------------------------------
void DirectionalObj::SetTwist( const f32 twist )
{
    m_Twist = twist;
}

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------
f32 DirectionalObj::GetTwist() const
{
    return m_Twist;
}

//---------------------------------------------------------------------------
//
//
//---------------------------------------------------------------------------
void DirectionalObj::SetOrientation( const nw::math::MTX33& orientation )
{
    m_Orientation = orientation;
}

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------
nw::math::MTX33 DirectionalObj::GetOrientation() const
{
    return m_Orientation;
}

//---------------------------------------------------------------------------
//
//
//---------------------------------------------------------------------------
void DirectionalObj::SetRotateSpeed( const f32 rotateSpeed )
{
    m_RotateSpeed = rotateSpeed;
}

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------
f32 DirectionalObj::GetRotateSpeed() const
{
    return m_RotateSpeed;
}

//---------------------------------------------------------------------------
//
//
//---------------------------------------------------------------------------
void DirectionalObj::SetTransSpeed( const f32 transSpeed )
{
    m_TransSpeed = transSpeed;
}

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------
f32 DirectionalObj::GetTransSpeed() const
{
    return m_TransSpeed;
}

//---------------------------------------------------------------------------
//
//
//---------------------------------------------------------------------------
void DirectionalObj::SetTrackSpeed( const f32 trackSpeed )
{
    m_TrackSpeed = trackSpeed;
}

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------
f32 DirectionalObj::GetTrackSpeed() const
{
    return m_TrackSpeed;
}

//---------------------------------------------------------------------------
//
//
//---------------------------------------------------------------------------
void DirectionalObj::SetTiltXSpeed( const f32 tiltXSpeed )
{
    m_TiltXSpeed = tiltXSpeed;
}

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------
f32 DirectionalObj::GetTiltXSpeed() const
{
    return m_TiltXSpeed;
}

//---------------------------------------------------------------------------
//
//
//---------------------------------------------------------------------------
void DirectionalObj::SetTiltYSpeed( const f32 tiltYSpeed )
{
    m_TiltYSpeed = tiltYSpeed;
}

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------
f32 DirectionalObj::GetTiltYSpeed() const
{
    return m_TiltYSpeed;
}

//---------------------------------------------------------------------------
//
//
//---------------------------------------------------------------------------
void DirectionalObj::SetIsProjTypeOrtho( const bool isOrtho )
{
    m_ProjTypeOrtho = isOrtho;
}

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------
bool DirectionalObj::GetIsProjTypeOrtho() const
{
    return m_ProjTypeOrtho;
}

//---------------------------------------------------------------------------
//
//
//---------------------------------------------------------------------------
void DirectionalObj::SetUpdateType( const u32 updateType )
{
    m_UpdateType = updateType;
}

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------
u32 DirectionalObj::GetUpdateType() const
{
    return m_UpdateType;
}

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------
void DirectionalObj::SetInitUpward(const nw::math::VEC3 &initUpward)
{
    m_InitUpward = initUpward;
    if ( nw::math::VEC3Len( m_InitUpward ) != 0.0f )
    { nw::math::VEC3Normalize( &m_InitUpward, &m_InitUpward ); }
}

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------
nw::math::VEC3 DirectionalObj::GetInitUpward() const
{
    return m_InitUpward;
}

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------
void DirectionalObj::Rotate( bool isLight )
{
   f32 rotationX = 0.0f;
   f32 rotationY = 0.0f;

   // Pad processing
   if ( m_PadEvent.flags & PadEvent::HANDLE_ROTATE )
   {
       rotationX = m_PadEvent.rotateH;
       rotationY = m_PadEvent.rotateV;
   }
   if ( 0.0f != rotationX || 0.0f != rotationY )
   {
       f32 rotSpeed = m_RotateSpeed;
       if ( isLight ) rotSpeed = 0.1f;
       rotationX *= rotSpeed;
       rotationY *= rotSpeed;
       m_RotateSpeed += sDirectionalInfo.RotationStep;

       // Adds restrictions so that values do not exceed the maximum.
       if ( m_RotateSpeed > sDirectionalInfo.RotationMax ) m_RotateSpeed = sDirectionalInfo.RotationMax;
   }
   else
   {
       // Returns to initial values.
       m_RotateSpeed = sDirectionalInfo.RotationInit;
       return;
   }

   // Inverts the manipulation system.
   if ( sDirectionalInfo.TumbleXInverse ) rotationX = -rotationX;
   if ( sDirectionalInfo.TumbleYInverse ) rotationY = -rotationY;

   u32 updateType = m_UpdateType;
   if ( isLight ) { updateType = DirectionalObj::UPDATE_TYPE_AIM; }

   switch( updateType )
   {
       // Aim mode
       case DirectionalObj::UPDATE_TYPE_AIM:
       {
           nw::math::VEC3 rotate;

           // Distance calculation
           f32 targetDist = nw::math::VEC3Dist( m_TargetPos, m_Translate );

           if ( targetDist == 0.0f ) return;

           // Finds the angles.
           f32 twist = ( isLight == true ) ? 0.0f : m_Twist;
           if ( !ComputeAngle( m_Translate, m_TargetPos, twist, rotate ) ) return;

           // Changes the amount of rotation using values input from the pad.
           rotate.y -= NN_MATH_RAD_TO_FIDX( rotationX );
           rotate.x -= NN_MATH_RAD_TO_FIDX( rotationY );

           // Restricted to a range of -90 degrees C to +90 degrees C.
           const f32 ROTATE_X_LIMIT = 63.9f;
           rotate.x = nw::ut::Clamp( rotate.x, -ROTATE_X_LIMIT, ROTATE_X_LIMIT );

           // Finds the amount of rotation.
           nw::math::VEC3 turn;
           ComputeRotateFIdx( rotate.x, rotate.y, turn );

           // Normalization
           if ( nw::math::VEC3Len( &turn ) != 0.0f ) { nw::math::VEC3Normalize( &turn, &turn ); }

           // Rotates the object position.
           m_Translate.Set(
               m_TargetPos.x - targetDist * turn.x,
               m_TargetPos.y - targetDist * turn.y,
               m_TargetPos.z + targetDist * turn.z );
       }
       break;

       // LooktAt mode
       case DirectionalObj::UPDATE_TYPE_LOOKAT:
       {
           nw::math::VEC3 rotate;
           // Finds the angles.
           if ( !ComputeAngle( m_Translate, m_TargetPos, 0.0f, rotate ) ) return;
           rotate.y -= NN_MATH_RAD_TO_FIDX( rotationX );
           rotate.x -= NN_MATH_RAD_TO_FIDX( rotationY );

           // Calculates the directional vectors.
           nw::math::VEC3 dir = m_TargetPos - m_Translate;

           f32 dist = nw::math::VEC3Len( &dir );
           if ( dist == 0.0f ) return;
           nw::math::VEC3Normalize( &dir, &dir );

           // Saves the distance before the calculation.
           f32 targetDist = dist;

           // Finds the light vector from the line of sight vector and up vector.
           nw::math::VEC3 right;
           nw::math::VEC3Cross( &right, &m_Upward, &dir );
           right.y = 0.0f;

           dist = nw::math::VEC3Len( &right );
           if ( dist == 0.0f ) return;
           nw::math::VEC3Normalize( &right, &right );

           // Decrements to cause rotation center on the point of origin.
           nw::math::VEC3 translate = m_Translate;
           translate -= m_TargetPos;

           // Finds the rotation matrix
           nw::math::MTX34 rotateV;
           nw::math::MTX34 rotateH;
           MTX34RotAxisRad( &rotateH, &right, rotationY );
           MTX34RotXYZRad( &rotateV, 0.0f, rotationX, 0.0f );

           // Rotates the position.
           nw::math::VEC3Transform( &translate, &rotateH, &translate );
           nw::math::VEC3Transform( &translate, &rotateV, &translate );

           // Maintains the original distance so that the calculated result is on target.
           if ( nw::math::VEC3Len( &translate ) != 0.0f )
           {
               nw::math::VEC3Normalize( &translate, &translate );
               translate *= targetDist;
           }

           // Returns the decremented amount to what it was.
           translate += m_TargetPos;

           // Object position setting
           m_Translate = translate;

           // Also rotates the up vector.
           nw::math::VEC3 upVector;
           nw::math::VEC3Transform( &upVector, &rotateH, &m_Upward );
           nw::math::VEC3Transform( &upVector, &rotateV, &upVector );

           dist = nw::math::VEC3Len( &upVector );
           if ( dist == 0.0f ) return;
           nw::math::VEC3Normalize( &upVector, &upVector );

           // Up vector settings
           m_Upward = upVector;
       }
       break;

       // Rotate mode
       case DirectionalObj::UPDATE_TYPE_ROTATE:
       {
           // Processes the camera only.
           if( !isLight )
           {
               nw::math::VEC3 rotate;

               // Finds the virtual gaze point.
               nw::math::VEC3 virTargetPos = m_TargetPos;

               // Offsets the virtual gaze point if it overlaps the camera position.
               if ( nw::math::VEC3Dist( &m_Translate, &virTargetPos ) == 0.0f )
               {
                   nw::math::VEC3 dir( 0.0f, 1.0f, 0.0f );

                   // Sets the virtual gaze point.
                   virTargetPos = m_Translate + Camera::VIRTUAL_TARGET_DISTANCE * dir;
               }

               // Distance calculation
               f32 targetDist = nw::math::VEC3Dist( virTargetPos, m_Translate );

               // Finds the angles.
               if ( !ComputeAngle( m_Translate, virTargetPos, 0.0f, rotate ) ) return;

               // Changes the amount of rotation using values input from the pad.
               rotate.y -= NN_MATH_RAD_TO_FIDX( rotationX );
               rotate.x -= NN_MATH_RAD_TO_FIDX( rotationY );

               // Restricted to a range of -90 degrees C to +90 degrees C.
               const f32 ROTATE_X_LIMIT = 63.9f;
               rotate.x = nw::ut::Clamp( rotate.x, -ROTATE_X_LIMIT, ROTATE_X_LIMIT );

               // Finds the amount of rotation.
               nw::math::VEC3 turn;
               ComputeRotateFIdx( rotate.x, rotate.y, turn );

               // Normalization
               nw::math::VEC3SafeNormalize( &turn, &turn, turn );

               // Rotates the object position.
               m_Translate.Set(
                   virTargetPos.x - targetDist * turn.x,
                   virTargetPos.y - targetDist * turn.y,
                   virTargetPos.z + targetDist * turn.z );

              // Sets the gaze point.
              m_TargetPos = virTargetPos;

              // Rotates to the direction of the virtual gaze point.
              if ( !ComputeAngle( m_Translate, virTargetPos, 0.0f, rotate ) ) return;
              m_Rotate.x = NN_MATH_FIDX_TO_RAD( rotate.x );
              m_Rotate.y = NN_MATH_FIDX_TO_RAD( -rotate.y );
              m_Rotate.z = 0.0f;
           }
       }
       break;

       // Methods other than those above.
       default:
       {
           /* DO_NOTHING */
       }
       break;
   }
}

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------
void DirectionalObj::Dolly( bool isLight )
{
    if ( m_ProjTypeOrtho ) return;

    f32 dolly = 0.0f;
    bool moveTarget = false;

    // Processes pad inputs.
    if ( m_PadEvent.flags & PadEvent::HANDLE_DOLLY )
    {
        dolly = m_PadEvent.dolly;
        moveTarget = m_PadEvent.moveTarget;
    }
    if ( dolly == 0.0f ) return;

    // Inverts the manipulation system.
    if ( sDirectionalInfo.DollyInverse ) dolly = -dolly;

    nw::math::VEC3 translate;
    switch( m_UpdateType )
    {
        // Aim/LookAt mode
        case DirectionalObj::UPDATE_TYPE_AIM:
        case DirectionalObj::UPDATE_TYPE_LOOKAT:
        {
            // Calculates the directional vectors.
            nw::math::VEC3 dir = m_TargetPos - m_Translate;

            // Distance calculation
            f32 dist = nw::math::VEC3Dist( m_TargetPos, m_Translate );

            // Zero check
            if ( dist == 0.0f )
            {
                if( !isLight ) return;

                dist = 10.0f;
                dir = nw::math::VEC3( 0.0f, 0.0f, 1.0f );
            }

            // Normalization
            nw::math::VEC3Normalize( &dir, &dir );

            // Calculates the amount of movement.
            translate = dolly * dir * dist * sDirectionalInfo.DollyInit;

            // Moves the target position.
            nw::math::VEC3 target = m_TargetPos;
            if ( moveTarget ) target = target + translate;

            // Restricts the distance so that it does not go below the threshold.
            if ( dist < sDirectionalInfo.DollyThreshold )
            { target = m_Translate + dir * sDirectionalInfo.DollyThreshold; }

            // Changes the target position.
            m_TargetPos = target;
        }
        break;

        case DirectionalObj::UPDATE_TYPE_ROTATE:
        {
            nw::math::VEC3 rotate = m_Rotate;
            nw::math::VEC3 turn;

            // Calculates amount of rotation.
            ComputeRotateRad( rotate.x, -rotate.y, turn );
            nw::math::VEC3SafeNormalize( &turn, &turn, turn );

            // Calculates the amount of movement.
            const f32 dist = dolly * sDirectionalInfo.DollyInit * 50.0f;
            translate.Set(
                dist * turn.x,
                dist * turn.y,
               -dist * turn.z );

            // Restricts the distance so that it does not go below the threshold.
            {
                f32 targetDist = nw::math::VEC3Dist( m_TargetPos, m_Translate );

                // Moves the target position.
                if ( targetDist < Camera::VIRTUAL_TARGET_DISTANCE )
                {
                    m_TargetPos += translate;
                }
            }
        }
        break;

        default:
        {
            /* DO_NOTHING */
        }
        break;
    }

    // Moves the object position.
    m_Translate = m_Translate + translate;
}

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------
void DirectionalObj::Slide( bool isLight )
{
    f32 slideX = 0.0f;
    f32 slideZ = 0.0f;

    // Processes pad inputs.
    if ( m_PadEvent.flags & PadEvent::HANDLE_SLIDE )
    {
        slideX = m_PadEvent.slideX;
        slideZ = m_PadEvent.slideZ;
    }
    if ( ( 0.0f != slideX ) || ( 0.0f != slideZ ) )
    {
        f32 transSpeed = m_TransSpeed;
        if ( isLight ) transSpeed = 1.0f;
        slideX *= transSpeed;
        slideZ *= transSpeed;
        m_TransSpeed += sDirectionalInfo.TranslationStep;

        // Adds restrictions so that values do not exceed the maximum.
        if ( m_TransSpeed > sDirectionalInfo.TranslationInit ) m_TransSpeed = sDirectionalInfo.TranslationInit;
    }
    else
    {
        // Returns to initial values.
        m_TransSpeed = sDirectionalInfo.TranslationInit;
        return;
    }

    // Inverts the manipulation system.
    if ( sDirectionalInfo.TrackXInverse ) slideX = -slideX;
    if ( sDirectionalInfo.TrackYInverse ) slideZ = -slideZ;

    f32 scale = 1.0f;

    nw::math::VEC3 xAxis;
    nw::math::VEC3 zAxis;

    // If light:
    if ( isLight )
    {
        nw::math::VEC3 viewPos;

        nw::gfx::Camera* camera = CameraMgr::GetInstance()->GetViewCamera();
        camera->Transform().GetTranslate( &viewPos );
        scale = nw::math::VEC3Dist( m_Translate, viewPos ) * 0.01f;
    }
    // Camera processing
    else
    {
        if ( ( m_UpdateType == DirectionalObj::UPDATE_TYPE_AIM )
          || ( m_UpdateType == DirectionalObj::UPDATE_TYPE_LOOKAT ) )
        {
            scale = nw::math::VEC3Dist( m_TargetPos, m_Translate ) * 0.01f;
        }
    }
    xAxis = m_Orientation.GetColumn( 0 );
    zAxis = m_Orientation.GetColumn( 2 );

    if ( xAxis.y == 1.0f || zAxis.y == 1.0f ) return;

    xAxis.y = 0.0f;
    zAxis.y = 0.0f;

    f32 dist = nw::math::VEC3Len( &xAxis );
    if ( dist == 0.0f ) return;

    dist = nw::math::VEC3Len( &zAxis );
    if ( dist == 0.0f ) return;

    // Finds the amount of slide.
    nw::math::VEC3Normalize( &xAxis, &xAxis );
    nw::math::VEC3Normalize( &zAxis, &zAxis );
    nw::math::VEC3 slide = ( xAxis * slideX + zAxis * slideZ ) * scale;

    if ( ( m_UpdateType == DirectionalObj::UPDATE_TYPE_AIM )
      || ( m_UpdateType == DirectionalObj::UPDATE_TYPE_LOOKAT ) )
    {
        // Slides the target position.
        m_TargetPos = m_TargetPos + slide;
    }
    else if ( !isLight && ( m_UpdateType == DirectionalObj::UPDATE_TYPE_ROTATE ) )
    {
        // Slides the target position.
        m_TargetPos = m_TargetPos + slide;
    }

    // Slides the object position.
    m_Translate = m_Translate + slide;
}

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------
void DirectionalObj::Track( bool isLight )
{
    f32 trackX = 0.0f;
    f32 trackY = 0.0f;
    bool moveInViewSpace = false;

    // Processes pad inputs.
    if ( m_PadEvent.flags & PadEvent::HANDLE_TRACK )
    {
        trackX = m_PadEvent.trackH;
        trackY = m_PadEvent.trackV;
        moveInViewSpace = m_PadEvent.moveInViewSpace;
    }

    // Parallel translation acceleration (Processes acceleration if parallel translation is entered.)
    if ( ( 0.0f != trackX ) || ( 0.0f != trackY ) )
    {
        f32 trackSpeed = m_TrackSpeed;
        if ( isLight ) trackSpeed = 1.0f;
        trackX *= trackSpeed;
        trackY *= trackSpeed;
        m_TrackSpeed += sDirectionalInfo.TranslationStep;

        // Adds restrictions if values are at or above the maximum.
        if ( m_TrackSpeed > sDirectionalInfo.TranslationMax ) m_TrackSpeed = sDirectionalInfo.TranslationMax;
    }
    else
    {
        // Returns to initial values.
        m_TrackSpeed = sDirectionalInfo.TranslationInit;
        return;
    }

    // Inverts the manipulation system.
    if ( sDirectionalInfo.TrackXInverse ) trackX = -trackX;
    if ( sDirectionalInfo.TrackYInverse ) trackY = -trackY;

    nw::math::VEC3 track;
    nw::math::VEC3 slide;
    // Regular track manipulation.
    if ( !moveInViewSpace )
    {
        track = trackX * m_Orientation.GetColumn( 0 );
        slide = nw::math::VEC3( 0.0f, trackY, 0.0f );
    }
    // Parallel translation relative to the XY plane of the camera space.
    else
    {
        track = trackX * m_Orientation.GetColumn( 0 );
        slide = trackY * m_Orientation.GetColumn( 1 );
    }
    nw::math::VEC3 translate = track + slide;

    if ( nw::math::VEC3Len( translate ) != 0.0f )
    {
        if ( ( m_UpdateType == DirectionalObj::UPDATE_TYPE_AIM )
           ||( m_UpdateType == DirectionalObj::UPDATE_TYPE_LOOKAT ) )
        {
            // Moves the target position.
            f32 scale = nw::math::VEC3Dist( m_TargetPos, m_Translate ) * 0.01f;
            translate *= scale;
            m_TargetPos = m_TargetPos + translate;
        }
        else if ( !isLight && ( m_UpdateType == DirectionalObj::UPDATE_TYPE_ROTATE ) )
        {
            // Moves the target position.
            m_TargetPos = m_TargetPos + translate;
        }

        // Moves the object position.
        m_Translate = m_Translate + translate;
    }
}

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------
void DirectionalObj::Tilt( bool isLight )
{
    f32 tiltX = 0.0f;
    f32 tiltY = 0.0f;

    // Processes pad inputs.
    if ( m_PadEvent.flags & PadEvent::HANDLE_TILT )
    {
        tiltX = m_PadEvent.tiltH * m_TiltXSpeed;
        tiltY = m_PadEvent.tiltV * m_TiltYSpeed;
    }
    if ( tiltX == 0.0f && tiltY == 0.0f ) return;

    // Inverts the manipulation system.
    if ( sDirectionalInfo.TiltXInverse ) tiltX = -tiltX;
    if ( sDirectionalInfo.TiltYInverse ) tiltY = -tiltY;

    nw::math::VEC3 rotate;
    switch( m_UpdateType )
    {
        // Aim mode
        case DirectionalObj::UPDATE_TYPE_AIM:
        {
            // Finds the angles formed.
            if ( !ComputeAngle( m_TargetPos, m_Translate, m_Twist, rotate ) ) return;
        }
        break;

        // LooktAt mode
        case DirectionalObj::UPDATE_TYPE_LOOKAT:
        {
            // Finds the angles formed.
            if ( !ComputeAngle( m_TargetPos, m_Translate, 0.0f, rotate ) ) return;
        }
        break;

        // Rotate mode
        case DirectionalObj::UPDATE_TYPE_ROTATE:
        {
            rotate   = m_Rotate;
            rotate.y -= tiltX;
            rotate.x += tiltY;
            m_Rotate = rotate;

            if ( !isLight )
            {
                // Calculates the directional vectors.
                nw::math::VEC3 turn;
                ComputeRotateRad( rotate.x, -rotate.y, turn );
                nw::math::VEC3SafeNormalize( &turn, &turn, turn );

                // Calculates the distance to the target position.
                float dist = nw::math::VEC3Dist( &m_Translate, &m_TargetPos );

                // Sets the target position.
                m_TargetPos.x = m_Translate.x + turn.x * dist;
                m_TargetPos.y = m_Translate.y + turn.y * dist;
                m_TargetPos.z = m_Translate.z - turn.z * dist;
            }
        }
        break;

        default:
        {
            /* DO_NOTHING */
        }
        break;
     }

    if ( m_UpdateType != DirectionalObj::UPDATE_TYPE_ROTATE )
    {
        // Increases or reduces the rotation angle using the values obtained from the pad.
        rotate.x += NN_MATH_RAD_TO_FIDX( tiltY );
        rotate.y += NN_MATH_RAD_TO_FIDX( tiltX );

        if ( m_UpdateType == DirectionalObj::UPDATE_TYPE_AIM )
        {
            // Restricts vertical rotation.
            const f32 ROTATE_X_LIMIT = 63.9f;
            rotate.x = nw::ut::Clamp(rotate.x, -ROTATE_X_LIMIT, ROTATE_X_LIMIT);

            f32 xs=nw::math::SinFIdx(rotate.x), xc=nw::math::CosFIdx(rotate.x);
            f32 ys=nw::math::SinFIdx(rotate.y), yc=nw::math::CosFIdx(rotate.y);
            f32 targetDistance = nw::math::VEC3Dist( m_TargetPos, m_Translate );

            // Gaze point calculation
            nw::math::VEC3 dir = nw::math::VEC3( -xc * ys,
                                                 -xs,
                                                  xc * yc );
            nw::math::VEC3SafeNormalize( &dir, &dir, dir );
            dir *= targetDistance;

            // Sets the gaze point.
            m_TargetPos = dir + m_Translate;
        }
        else if ( m_UpdateType == DirectionalObj::UPDATE_TYPE_LOOKAT )
        {
            // Calculates the gaze point and viewing point.
            f32 dist = nw::math::VEC3Dist( m_TargetPos, m_Translate );
            if ( dist == 0.0f ) { return; }

            // Calculates the line of sight vector.
            nw::math::VEC3 dir = m_TargetPos - m_Translate;
            nw::math::VEC3Normalize( &dir, &dir );

            // Up vector
            nw::math::VEC3 upward = m_Upward;

            // Light vector calculation
            nw::math::VEC3 right;
            nw::math::VEC3Cross( &right, &upward, &dir );
            nw::math::VEC3SafeNormalize( &right, &right, right );

            // Axes
            nw::math::VEC3 yAxis = nw::math::VEC3( 0.0f, 1.0f, 0.0f );
            nw::math::VEC3 xAxis = nw::math::VEC3( 1.0f, 0.0f, 0.0f );

            nw::math::MTX34 rotateH, rotateV;
            nw::math::MTX34 rotY;

            nw::math::MTX34Identity( &rotateH );
            nw::math::MTX34Identity( &rotateV );

            // Vertical rotation axis calculation
            nw::math::MTX34RotAxisFIdx( &rotY, yAxis, -rotate.y );
            nw::math::VEC3Transform( &xAxis, &rotY, &xAxis );
            nw::math::VEC3SafeNormalize( &xAxis, &xAxis, xAxis );

            // Sets the calculated vertical rotation axis when the light vector is zero.
            if ( nw::math::VEC3Len( &right ) == 0.0f )
            {
                right = xAxis;

                // End if the calculated axis is zero.
                if ( nw::math::VEC3Len( &right ) == 0.0f ) return;
            }

            // Vertical rotation
            if ( tiltY != 0.0f )
            {
                nw::math::MTX34RotAxisRad( &rotateV, &right, -tiltY );
                nw::math::VEC3Transform( &dir, &rotateV, &dir );
                nw::math::VEC3SafeNormalize( &dir, &dir, dir );
            }

            // Horizontal rotation
            if ( tiltX != 0.0f )
            {
                nw::math::MTX34RotAxisRad( &rotateH, yAxis, -tiltX );
                nw::math::VEC3Transform( &dir, &rotateH, &dir );
                nw::math::VEC3SafeNormalize( &dir, &dir, dir );
            }

            if ( !isLight )
            {
                // Rotates the light vector.
                nw::math::VEC3Transform( &right, &rotateH, &right );
                nw::math::VEC3SafeNormalize( &right, &right, right );

                // Up vector calculation
                nw::math::VEC3Cross( &upward, &dir, &right );
                nw::math::VEC3SafeNormalize( &upward, &upward, upward );

                // Sign check
                int sigB = ( m_Upward.y >= 0.0f ) ? 1 : -1;
                int sigA = ( upward.y   >= 0.0f ) ? 1 : -1;
                if ( sigB != sigA ) return;

                // Updates the up vector.
                m_Upward = upward;
            }

            // Viewing point settings
            m_TargetPos = ( dir * dist ) + m_Translate;
        }
    }
}

}   /* namespace g3d */
}   /* namespace vw */
