/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#include <viewer/system.h>
#include <viewer/g3d/CameraMgr.h>
#include "../include/Grid.h"

namespace vw  {
namespace g3d {

 /* ctor */
Grid::Grid() : mScale( 1.0f ), mDivision( 8 ), mMeshColor( 0.0f,0.0f,0.0f,1.0f ), mInit(false), mDistanceFromCamera(0.f)
{
}

void
Grid::CreateBuffer( f32 scale, s32 division )
{
    mScale = scale;
    mDivision = division;
    mInit = false;
    mDistanceFromCamera = 0.0f;
}

/*---------------------------------------------------------------------------*
  Name:         Grid::Draw
  Description:  Renders the grid.
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------*/
void
Grid::Draw( nw::gfx::Camera* pCamera, nw::gfx::Camera* pViewCamera )
{
    // Remember the position versus the camera and adjust the line width appropriately.
    f32 dist = 0.0f;
    f32 lineWidth = 0.2f;

    nw::gfx::ResCameraViewUpdater       resViewUpdater             = pViewCamera->GetViewUpdater()->GetResource();
    nw::gfx::ResAimTargetViewUpdater    resAimTargetViewUpdater    = nw::gfx::ResDynamicCast<nw::gfx::ResAimTargetViewUpdater>( resViewUpdater );
    nw::gfx::ResLookAtTargetViewUpdater resLookAtTargetViewUpdater = nw::gfx::ResDynamicCast<nw::gfx::ResLookAtTargetViewUpdater>( resViewUpdater );

    // Only for other than Rotate.
    if ( resAimTargetViewUpdater.IsValid() || resLookAtTargetViewUpdater.IsValid() )
    {
        nn::math::VEC3 p = pViewCamera->GetPosition();
        nn::math::VEC3 t = pViewCamera->GetTargetPosition();
        dist = (p - t).Length();

        if(nn::math::Abs(dist - mDistanceFromCamera) > (f32)1e-4)
        {
            mInit = false;
            lineWidth = dist * 0.01f;
            // Assign by clamping values.
            lineWidth = nn::math::Clamp(lineWidth, 0.01f, 0.2f);
            lineWidth *= mLineWidth;
            mDistanceFromCamera = dist;
        }
    }

    if( !mInit )
    {
        CreateGrid(mScale, mDivision, lineWidth);
    }

    vw::system::Draw3DHelper& draw3d = vw::system::Draw3DHelper::GetInstance();
    vw::system::Draw3DHelper::STATE enableState  = draw3d.GetEnable();
    vw::system::Draw3DHelper::STATE disableState = draw3d.GetDisable();

    draw3d.Begin();
    draw3d.SetView(pCamera->ViewMatrix());
    draw3d.SetProj(pCamera->ProjectionMatrix());
    draw3d.Enable( vw::system::Draw3DHelper::DEPTHTEST | vw::system::Draw3DHelper::DEPTHWRITE );

    nw::math::MTX34 world;
    nw::math::MTX34Identity(&world);
    vw::system::Draw3DHelper::GetInstance().SetWorld(world);
    mGridModel.Draw();

    draw3d.End();
    draw3d.Disable( disableState );
    draw3d.Enable( enableState );
}

void
Grid::CreateGrid(f32 scale, s32 division, f32 lineWidth)
{
    mInit = true;

    const f32 x0 = -scale * 0.5f * (f32)division;
    const f32 x1 =  scale * 0.5f * (f32)division;
    const f32 y0 = -scale * 0.5f * (f32)division;
    const f32 y1 =  scale * 0.5f * (f32)division;
    const f32 halfWidth = lineWidth * 0.5f;
    const nw::ut::Color8 color = mMeshColor;

    u32 vertexNum = (division + 1) * 6 * 2;
    mGridModel.CreateBuffer(vertexNum, vertexNum);
    vw::system::XYZTC::Vertex* vertex = mGridModel.GetVertexData();
    u16*                        index = mGridModel.GetIndexData();

    struct Inner
    {
        static void set( vw::system::XYZTC::Vertex& xyzc, f32 x, f32 y, f32 z, u32 color)
        {
            xyzc.x = x;
            xyzc.y = y;
            xyzc.z = z;
            xyzc.color = color;
        }
    };

    // Horizontal Line
    u32 iCount = 0;
    f32 y = y0;
    for (s32 i=0;i<division+1;i++)
    {
        // Triangle 1
        Inner::set( vertex[iCount++], x0,0.0f,y-halfWidth, color );
        Inner::set( vertex[iCount++], x1,0.0f,y-halfWidth, color );
        Inner::set( vertex[iCount++], x1,0.0f,y+halfWidth, color );
        // Triangle 2
        Inner::set( vertex[iCount++], x0,0.0f,y-halfWidth, color );
        Inner::set( vertex[iCount++], x1,0.0f,y+halfWidth, color );
        Inner::set( vertex[iCount++], x0,0.0f,y+halfWidth, color );
        y += scale;
    }

    // Vertical Line
    f32 x = x0;
    for (s32 i=0; i<division+1; i++)
    {
        // Triangle 1
        Inner::set( vertex[iCount++], x-halfWidth,0.0f,y0, color );
        Inner::set( vertex[iCount++], x-halfWidth,0.0f,y1, color );
        Inner::set( vertex[iCount++], x+halfWidth,0.0f,y1, color );
        // Triangle 1
        Inner::set( vertex[iCount++], x-halfWidth,0.0f,y0, color );
        Inner::set( vertex[iCount++], x+halfWidth,0.0f,y1, color );
        Inner::set( vertex[iCount++], x+halfWidth,0.0f,y0, color );
        x += scale;
    }

    // setup index
    for( s32 i=0; i<vertexNum; ++i)
    {
        index[i] = i;
    }

    // update data
    mGridModel.UpdateData();
}

}   /* namespace g3d */
}   /* namespace vw  */

