/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#include <viewer/system.h>
#include "../../../include/viewer/g3d/Light.h"
#include "../../../include/viewer/g3d/LightMgr.h"
#include "../../../include/viewer/g3d/Camera.h"
#include "../../../include/viewer/g3d/CameraMgr.h"

namespace vw  {
namespace g3d {


/*---------------------------------------------------------------------------*
  Name:         Light::Light
  Description:  Constructor
  Arguments:    light Input light
  Returns:      None.
 *---------------------------------------------------------------------------


*/
Light::Light(nw::gfx::Light* light, const char *name) : m_Light(light), m_CanTranslate(false), m_CanRotate(false)
{
    strcpy(m_Name, name);

    SetupLight();
}

/*---------------------------------------------------------------------------*
  Name:         Light::GetLight
  Description:  Gets light
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
nw::gfx::Light*
Light::GetLight()
{
    return m_Light;
}

/*---------------------------------------------------------------------------*
  Name:         Light::SetupLight
  Description:  Light initialization
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
Light::SetupLight()
{
    s32 kind = nw::gfx::ResLight::KIND_UNUSED;

    nw::gfx::res::ResSceneObject resObject = m_Light->GetResSceneObject();
    switch (resObject.GetTypeInfo())
    {
    case nw::gfx::res::ResFragmentLight::TYPE_INFO:
        {
            nw::gfx::res::ResFragmentLight fragmentLight = nw::ut::ResDynamicCast<nw::gfx::res::ResFragmentLight>(resObject);
            if ( fragmentLight.GetLightKind() != nw::gfx::ResLight::KIND_DIRECTIONAL ) m_CanTranslate = true;
            if ( fragmentLight.GetLightKind() != nw::gfx::ResLight::KIND_POINT) m_CanRotate = true;

            kind = fragmentLight.GetLightKind();
        }
        break;
    case nw::gfx::res::ResVertexLight::TYPE_INFO:
        {
            nw::gfx::res::ResVertexLight vertexLight = nw::ut::ResDynamicCast<nw::gfx::res::ResVertexLight>(resObject);
            if ( vertexLight.GetLightKind() != nw::gfx::ResLight::KIND_DIRECTIONAL ) m_CanTranslate = true;
            if ( vertexLight.GetLightKind() != nw::gfx::ResLight::KIND_POINT) m_CanRotate = true;

            kind = vertexLight.GetLightKind();
        }
        break;
    case nw::gfx::res::ResHemiSphereLight::TYPE_INFO:
        {
            m_CanRotate = true;
        }
        break;
    }

    // Virtual focal point setup
    const f32 dist = 10.0f;
    if( kind == nw::gfx::ResLight::KIND_DIRECTIONAL ||
        nw::gfx::res::ResHemiSphereLight::TYPE_INFO == m_Light->GetResSceneObject().GetTypeInfo())
    {
        m_TargetPos = nw::math::VEC3(0.0f, 0.0f, 0.0f);
        nw::math::VEC3 dir = this->GetDirection();
        m_Light->Transform().SetTranslate( dir * dist );
    }
    else if (kind == nw::gfx::ResLight::KIND_POINT)
    {
        nw::math::VEC3 lightPos = this->GetPosition();
        nw::math::VEC3 lightDir(0.0f, 1.0f, 0.0f);
        m_TargetPos = lightDir * dist;
    }
    else if (kind == nw::gfx::ResLight::KIND_SPOT)
    {
        nw::math::VEC3 lightPos = this->GetPosition();
        nw::math::VEC3 lightDir = this->GetDirection();
        m_TargetPos = lightPos + lightDir * dist;
    }
#if defined( NN_PCGX )
    else
    {
        m_TargetPos = nw::math::VEC3(0.0f, 0.0f, 0.0f);
    }
#endif

    if( m_CanRotate && kind != nw::gfx::ResLight::KIND_SPOT )
    {
        nw::math::VEC3 direction = this->GetDirection();

        m_Light->Transform().SetTranslate( -direction * dist );
    }
}

/*---------------------------------------------------------------------------*
  Name:         Light::MoveLight3D
  Description:  Light operation main routine
  Arguments:    hidStatus Specifies HID input.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
Light::MoveLight3D(vw::system::HidStatus& hidStatus)
{
    DirectionalObj dirObj;

    nw::gfx::Camera* camera = CameraMgr::GetInstance()->GetViewCamera();
    nw::math::MTX33 orientation;

    // Sets the values required for the calculation in the directional object.
    if ( camera != NULL )
    {
        orientation = Camera::GetOrientation( camera );
        dirObj.SetOrientation( orientation );
    }
    dirObj.SetPadEventFromHid( hidStatus, false );
    dirObj.SetTranslate( this->GetPosition() );
    dirObj.SetTargetPos( m_TargetPos );
    dirObj.SetIsProjTypeOrtho( false );
    dirObj.SetUpdateType( DirectionalObj::UPDATE_TYPE_LOOKAT );
    dirObj.SetRotateSpeed( 1.0f );
    dirObj.SetTransSpeed( 1.0f );
    dirObj.SetTrackSpeed( 1.0f );

#if defined( NN_PCGX )
    PadEvent padEvent = dirObj.GetPadEvent();
    PCGXAddLightEvent(&padEvent);
    dirObj.SetPadEvent(padEvent);
#endif

    // Rotational processing
    Rotate( dirObj );
    if ( m_CanTranslate )
    {
        // Dolly process
        Dolly( dirObj );

        // Slide process
        Slide( dirObj );

        // Track process
        Track( dirObj );

        // Tilt process
        Tilt( dirObj );
    }

    // Calculation direction
    nw::math::VEC3 direction = dirObj.GetTargetPos() - dirObj.GetTranslate();
    f32 dist = nw::math::VEC3Len( &direction );
    if ( dist != 0.0f ) { nw::math::VEC3Normalize( &direction, &direction ); }

    // Update values
    m_TargetPos = dirObj.GetTargetPos();
    m_Light->Transform().SetTranslate( dirObj.GetTranslate() );
    this->SetDirection( direction );
}

/*---------------------------------------------------------------------------*
  Name:         Light::RotatePos
  Description:  Rotates light position around the focal point
  Arguments:    directionalObj Directional object
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
Light::Rotate( vw::g3d::DirectionalObj& directionalObj )
{
    directionalObj.Rotate( true );
}

/*---------------------------------------------------------------------------*
  Name:         Light::Dolly
  Description:  Camera movement routine
  Arguments:    directionalObj
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
Light::Dolly( vw::g3d::DirectionalObj& directionalObj )
{
    directionalObj.Dolly( true );
}

/*---------------------------------------------------------------------------*
  Name:         Light::Track
  Description:  Light movement routine
  Arguments:    directionalObj
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
Light::Track( vw::g3d::DirectionalObj& directionalObj )
{
    directionalObj.Track( true );
}

/*---------------------------------------------------------------------------*
  Name:         Light::Slide
  Description:  Light movement routine
  Arguments:    directionalObj
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
Light::Slide( vw::g3d::DirectionalObj& directionalObj )
{
    directionalObj.Slide( true );
}

/*---------------------------------------------------------------------------*
  Name:         Light::Tilt
  Description:  Tilt lights.
  Arguments:    directionalObj
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
Light::Tilt( vw::g3d::DirectionalObj &directionalObj )
{
    directionalObj.Tilt( true );
}

/*---------------------------------------------------------------------------*
  Name:         Light::GetPosition
  Description:  Gets light position
  Arguments:    None.
  Returns:      Light position
 *---------------------------------------------------------------------------


*/
nw::math::VEC3
Light::GetPosition() const
{
    nw::math::VEC3 lightPos;
    m_Light->Transform().GetTranslate( &lightPos );
    return lightPos;
}

/*---------------------------------------------------------------------------*
  Name:         Light::GetDirection
  Description:  Light directional vector set
  Arguments:    None.
  Returns:      Gets direction
 *---------------------------------------------------------------------------


*/
nw::math::VEC3
Light::GetDirection() const
{
    nw::math::VEC3 direction;
    if( m_CanRotate )
    {
        nw::gfx::res::ResSceneObject resObject = m_Light->GetResSceneObject();
        switch ( resObject.GetTypeInfo() )
        {
        case nw::gfx::res::ResFragmentLight::TYPE_INFO:
            direction = nw::ut::ResDynamicCast<nw::gfx::res::ResFragmentLight>(resObject).GetDirection();
            break;
        case nw::gfx::res::ResVertexLight::TYPE_INFO:
            direction = nw::ut::ResDynamicCast<nw::gfx::res::ResVertexLight>(resObject).GetDirection();
            break;
        case nw::gfx::res::ResHemiSphereLight::TYPE_INFO:
            direction = nw::ut::ResDynamicCast<nw::gfx::res::ResHemiSphereLight>(resObject).GetDirection();
            break;
        }
    }
    return direction;
}

/*---------------------------------------------------------------------------*
  Name:         Light::SetDirection
  Description:  Light directional vector set
  Arguments:    direction Light direction
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
Light::SetDirection(const nw::math::VEC3& direction)
{
    nw::gfx::res::ResSceneObject resObject = m_Light->GetResSceneObject();
    switch ( resObject.GetTypeInfo() )
    {
    case nw::gfx::res::ResFragmentLight::TYPE_INFO:
        nw::ut::ResDynamicCast<nw::gfx::res::ResFragmentLight>(resObject).SetDirection( direction );
        break;
    case nw::gfx::res::ResVertexLight::TYPE_INFO:
        nw::ut::ResDynamicCast<nw::gfx::res::ResVertexLight>(resObject).SetDirection( direction );
        break;
    case nw::gfx::res::ResHemiSphereLight::TYPE_INFO:
         nw::ut::ResDynamicCast<nw::gfx::res::ResHemiSphereLight>(resObject).SetDirection( direction );
        break;
    }
}

}   /* namespace g3d */
}   /* namespace vw  */
