/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#include <viewer/system.h>
#include "../../../include/viewer/g3d/LightMgr.h"
#include "../include/SceneManager.h"

namespace vw  {
namespace g3d {

LightMgr* LightMgr::m_Instance = NULL;

/*static*/LightMgr*
LightMgr::GetInstance()
{
    NW_ASSERT( m_Instance != NULL );
    return m_Instance;
}

bool IsExist(
    nw::ut::MoveArray<Light>::iterator beginIter,
    nw::ut::MoveArray<Light>::iterator endIter,
    nw::gfx::Light *light
) {
    for (; beginIter!=endIter; ++beginIter)
    {
        if (beginIter->GetLight()==light)
        {
            return true;
        }
    }
    return false;
}

bool IsExist(
    SceneElemSelector::LightList::iterator beginIter,
    SceneElemSelector::LightList::iterator endIter,
    nw::gfx::Light *light
) {
    for (; beginIter!=endIter; ++beginIter)
    {
        if ((*beginIter)->GetSceneNode()==light)
        {
            return true;
        }
    }
    return false;
}

/*---------------------------------------------------------------------------*
  Name:         LightMgr::MoveLight3D

  Description:  Sets up the camera array.

  Arguments:    scene Scene class

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
LightMgr::Setup(SceneElemSelector::LightList::iterator lightsBegin,
                 SceneElemSelector::LightList::iterator lightsEnd)
{
    // Deletes excluding the ones that already exist.
    {
        nw::ut::MoveArray<Light>::iterator it = m_LightArray.begin();
        while( it != m_LightArray.end() )
        {
            if (!IsExist(lightsBegin, lightsEnd, it->GetLight()))
            {
                nw::ut::MoveArray<Light>::iterator next = it;
                ++next;
                it = m_LightArray.erase(it, next);
            }
            else
            {
                ++it;
            }
        }
    }

    // Adds new light.
    for(SceneElemSelector::LightList::iterator it = lightsBegin; it != lightsEnd; ++it)
    {
        if (!IsExist(m_LightArray.begin(), m_LightArray.end(), (*it)->GetSceneNode()))
        {
            m_LightArray.push_back( Light( (*it)->GetSceneNode(), (*it)->GetName() ) );
        }
    }

    m_TargetIdx = 0;

    // If there is one that matches a saved name, that one is made the TargetLight.
    for (int i=0; i<m_LightArray.size(); ++i)
    {
        if (!strcmp(m_LightArray[i].GetName(), m_PrevTargetLightName))
        {
            m_TargetIdx = i;
            break;
        }
    }
}


/*---------------------------------------------------------------------------*
  Name:         LightMgr::MoveLight3D

  Description:  Gets the current target camera.

  Arguments:    hidStatus HID input

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
LightMgr::MoveLight3D(vw::system::HidStatus& hidStatus)
{
    Light* target = &m_LightArray[m_TargetIdx];

    // Target camera operation
    target->MoveLight3D(hidStatus);
}

/* ctor */
LightMgr::LightMgr()
    : m_TargetIdx(0)
#if defined( NN_PCGX )
{
    ZeroMemory(m_PrevTargetLightName, 512);
#else
    , m_PrevTargetLightName("\0")
{
#endif
    // create stereo light
    nw::os::IAllocator* pAllocator = Memory_GetAllocator();
    m_LightArray = nw::ut::MoveArray<g3d::Light>(pAllocator);
}

/* dtor */
LightMgr::~LightMgr()
{
}

// Initialize
void
LightMgr::Initialize()
{
    if (m_Instance != NULL) return;

    void *buf = Memory_GetAllocator()->Alloc(sizeof(LightMgr));
    if (buf == NULL) return;

    m_Instance = new(buf) LightMgr();
}

// Finalize
void
LightMgr::Finalize()
{
    if (m_Instance == NULL) return;

    m_Instance->~LightMgr();
    Memory_GetAllocator()->Free(m_Instance);

    m_Instance = NULL;
}

// IsFirstTargetLight()
bool
LightMgr::IsFirstTargetLight()
{
    return ( m_TargetIdx == 0 );
}

// IsLastTargetLight()
bool
LightMgr::IsLastTargetLight()
{
    return ( m_TargetIdx == m_LightArray.size() - 1 );
}

// SwitchToFirstTargetLight()
void
LightMgr::SwitchToFirstTargetLight()
{
    m_TargetIdx = 0;
    strcpy( m_PrevTargetLightName, m_LightArray[m_TargetIdx].GetName() );
}

// SwitchToLastTargetLight()
void
LightMgr::SwitchToLastTargetLight()
{
    m_TargetIdx = m_LightArray.size() - 1;
    strcpy( m_PrevTargetLightName, m_LightArray[m_TargetIdx].GetName() );
}

// SwitchToPrevTargetLight()
void
LightMgr::SwitchToPrevTargetLight()
{
    m_TargetIdx = m_TargetIdx == 0 ? m_LightArray.size() - 1 : m_TargetIdx - 1;
    strcpy( m_PrevTargetLightName, m_LightArray[m_TargetIdx].GetName() );
}

// SwitchToNextTargetLight()
void
LightMgr::SwitchToNextTargetLight()
{
    m_TargetIdx = m_TargetIdx == m_LightArray.size() - 1 ? 0 : m_TargetIdx + 1;
    strcpy( m_PrevTargetLightName, m_LightArray[m_TargetIdx].GetName() );
}

// ReloadDefaultLight()
void
LightMgr::ReloadDefaultLight( nw::gfx::Light* light )
{
    for (int i=0; i<m_LightArray.size(); ++i)
    {
        if ( m_LightArray[i].GetLight() == light )
        {
            m_LightArray[i] = Light( light, m_LightArray[i].GetName() );
            return;
        }
    }
}

}   /* namespace g3d */
}   /* namespace vw  */
