/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#include <viewer/system.h>
#include "../../system/include/HidManager.h"
#include "../../system/include/LogSystem.h"
#include "../include/LogSubMode.h"

namespace vw  {
namespace g3d {

static const u32 sLogColumn = 10;

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
LogSubMode::LogSubMode()
{
    m_LogType = ERROR;
    m_Column = 0;
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
LogSubMode::~LogSubMode()
{
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
LogSubMode::Main(vw::system::HidStatus& hidStatus)
{
    // Get current pad status
    switch( system::GetPadEvent( hidStatus, system::GetLogSubModePadBindings() ) )
    {
    case system::LOG_EXPORT:
        ExportLog();
        break;
    case system::LOG_CLEAR:
        ClearLog();
        break;
    case system::LOG_SWITCH_PREV_LOG:
        switch( m_LogType ){
        case ERROR:   m_LogType=HISTORY; break;
        case WARNING: m_LogType=ERROR;   break;
        case SYSTEM:  m_LogType=WARNING; break;
        case HISTORY: m_LogType=SYSTEM;  break;
        }
        m_Column = 0;
        break;
    case system::LOG_SWITCH_NEXT_LOG:
        switch( m_LogType ){
        case ERROR:   m_LogType=WARNING; break;
        case WARNING: m_LogType=SYSTEM;  break;
        case SYSTEM:  m_LogType=HISTORY; break;
        case HISTORY: m_LogType=ERROR;   break;
        }
        m_Column = 0;
        break;
    case system::LOG_SWITCH_PREV_LOG_ITEM:
        m_Column--;
        if( m_Column < 0 ) m_Column = 0;
        break;
    case system::LOG_SWITCH_NEXT_LOG_ITEM:
        m_Column++;
        this->CalcColumn();
        break;
    }
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void LogSubMode::Draw()
{
    vw::system::Draw2DHelper& draw2D = vw::system::Draw2DHelper::GetInstance();
    draw2D.BeginDraw();

    draw2D.Disable(
        vw::system::Draw2DHelper::ALPHABLEND |
        vw::system::Draw2DHelper::CULLFACE |
        vw::system::Draw2DHelper::DEPTHTEST );

    s32 aw = 10;
    s32 ah = 40;
    s32 w = 320-aw*2;
    s32 h = 240-ah*2;
    draw2D.FillRectangle(aw,ah,w,h, nw::ut::Color8(64,64,96,220));
    draw2D.EndDraw();

    vw::system::LogSystem* pLogSystem = vw::system::LogSystem::GetInstance();
    vw::system::FontDrawer& font = vw::system::FontDrawer::GetInstance();
    font.SetColor( nw::ut::Color8::WHITE );
    font.BeginDraw();

    static const u32 sColumnStep = 14;

    switch ( m_LogType )
    {
    case ERROR:
        {
            font.DrawString(120, 40, "ERROR LOG");
            font.SetFontSize(8,12);
            font.SetFixedWidth(5.5f);
            for ( u32 i=m_Column, col=0; i<pLogSystem->GetNumError(); ++i )
            {
                vw::system::LogSystem::Log* pLog = pLogSystem->GetError(i);
                if ( pLog != 0 )
                {
#if defined( NN_PCGX )
                    font.DrawString(15, 60+col*sColumnStep, "%s", pLog->m_Message.Get(), NULL );
#else
                    font.DrawString(15, 60+col*sColumnStep, "%s", pLog->m_Message.Get() );
#endif
                    col++;
                }
                if ( col >= sLogColumn )
                {
                    break;
                }
            }
        }
        break;
    case WARNING:
        {
            font.DrawString(120, 40, "WARNING LOG");
            font.SetFontSize(8,12);
            font.SetFixedWidth(5.5f);
            for ( u32 i=m_Column, col=0; i<pLogSystem->GetNumWarning(); ++i )
            {
                vw::system::LogSystem::Log* pLog = pLogSystem->GetWarning(i);
                if ( pLog != 0 )
                {
#if defined( NN_PCGX )
                    font.DrawString(15, 60+col*sColumnStep, "%s", pLog->m_Message.Get(), NULL );
#else
                    font.DrawString(15, 60+col*sColumnStep, "%s", pLog->m_Message.Get() );
#endif
                    col++;
                }
                if ( col >= sLogColumn )
                {
                    break;
                }
            }
        }
        break;
    case SYSTEM:
        {
            font.DrawString(120, 40, "SYSTEM LOG");
            font.SetFontSize(8,12);
            font.SetFixedWidth(5.5f);
            for ( u32 i=m_Column, col=0; i<pLogSystem->GetNumSystemLog(); ++i )
            {
                vw::system::LogSystem::Log* pLog = pLogSystem->GetSystemLog(i);
                if ( pLog != 0 )
                {
#if defined( NN_PCGX )
                    font.DrawString(15, 60+col*sColumnStep, "%s", pLog->m_Message.Get(), NULL );
#else
                    font.DrawString(15, 60+col*sColumnStep, "%s", pLog->m_Message.Get() );
#endif
                    col++;
                }
                if ( col >= sLogColumn )
                {
                    break;
                }
            }
        }
        break;
    case HISTORY:
        {
            font.DrawString(120, 40, "LOG HISTORY");
            font.SetFontSize(8,12);
            font.SetFixedWidth(5.5f);
            for ( u32 i=m_Column, col=0; i<pLogSystem->GetNumLog(); ++i )
            {
                vw::system::LogSystem::Log* pLog = pLogSystem->GetLog(i);
                if ( pLog != 0 )
                {
                    switch (pLog->m_Type)
                    {
                    case vw::system::LogSystem::ERROR:
                        font.SetColor( nw::ut::Color8(255, 110, 110, 255) );
                        break;
                    case vw::system::LogSystem::WARNING:
                        font.SetColor( nw::ut::Color8::WHITE );
                        break;
                    case vw::system::LogSystem::SYSTEM_LOG:
                        font.SetColor( nw::ut::Color8(255, 200, 100, 255) );
                        break;
                    }
#if defined( NN_PCGX )
                    font.DrawString(15, 60+col*sColumnStep, "%s", pLog->m_Message.Get(), NULL );
#else
                    font.DrawString(15, 60+col*sColumnStep, "%s", pLog->m_Message.Get() );
#endif
                    col++;
                }

                if ( col >= sLogColumn )
                {
                    break;
                }
            }
        }
        break;
    }
    font.SetColor( nw::ut::Color8::WHITE );
    font.EndDraw();
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
const char*
LogSubMode::GetModeName() const
{
    return "Log";
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
SubModeType
LogSubMode::GetModeType() const
{
    return SubMode_Log;
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
LogSubMode::SetLogType(LOG_TYPE type)
{
    m_LogType = type;
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
LogSubMode::ExportLog()
{
    bool IsRunning = false;
    nn::Result ret = nn::hio::GetHioDaemonStatus( &IsRunning );

    if(ret.IsSuccess() && IsRunning == true)
    {
        vw::common::CharArray filepath( vw::system::Memory_GetSysAllocator() );

        nn::os::CTR::TargetHardware target = nn::os::CTR::GetRunningTargetHardware();

        if ( target == nn::os::CTR::TARGET_HARDWARE_PRODUCT )
        {
            char path[1024];
            nn::hio::CTR::GetEnvironmentVariable(path, 1024 , "NW4C_ROOT");
            filepath  = path;
            filepath += "\\tools\\Viewer\\log.txt";
        }
        else
        {
            filepath = "log.txt";
        }

        vw::system::LogSystem::TYPE logType = vw::system::LogSystem::ERROR;
        switch( m_LogType )
        {
        case ERROR:
            logType = vw::system::LogSystem::ERROR;
            break;
        case WARNING:
            logType = vw::system::LogSystem::WARNING;
            break;
        case SYSTEM:
            logType = vw::system::LogSystem::SYSTEM_LOG;
            break;
        case HISTORY:
            logType = vw::system::LogSystem::HISTORY;
            break;
        }
        vw::system::LogSystem::GetInstance()->OutputLog(filepath.Get(), logType);
    }
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
LogSubMode::ClearLog()
{
    switch( m_LogType )
    {
    case ERROR:
        vw::system::LogSystem::GetInstance()->Clear(vw::system::LogSystem::ERROR);
        break;
    case WARNING:
        vw::system::LogSystem::GetInstance()->Clear(vw::system::LogSystem::WARNING);
        break;
    case SYSTEM:
        vw::system::LogSystem::GetInstance()->Clear(vw::system::LogSystem::SYSTEM_LOG);
        break;
    case HISTORY:
        vw::system::LogSystem::GetInstance()->Clear(vw::system::LogSystem::HISTORY);
        break;
    }
    m_Column = 0;
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
LogSubMode::CalcColumn()
{
    vw::system::LogSystem* log = vw::system::LogSystem::GetInstance();

    u32 count = 0;
    switch( m_LogType )
    {
    case ERROR:
        count = log->GetNumError();
        break;
    case WARNING:
        count = log->GetNumWarning();
        break;
    case SYSTEM:
        count = log->GetNumSystemLog();
        break;
    case HISTORY:
        count = log->GetNumLog();
        break;
    }
    if( count < sLogColumn )
    {
        m_Column = 0;
    }
    if( m_Column > count - sLogColumn )
    {
        m_Column = count - sLogColumn;
    }
}


}   /* namespace g3d */
}   /* namespace vw  */
