/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#include <viewer/system.h>
#include "../../system/include/HidManager.h"
#include "../include/ProtocolExec.h"
#include "../include/ProtocolExecCS.h"
#include "../include/ModuleG3dImpl.h"

#include <viewer/g3d/CameraMgr.h>
#include <viewer/g3d/LightMgr.h>
#include "../include/Config3D.h"
#include "../include/ResourceManager.h"
#include "../include/SceneManager.h"
#include "../include/SceneElemSelector.h"
#include "../include/SceneRenderer.h"
#include "../include/AnimManager.h"
#include "../include/SubModeManager.h"
#include "../include/UserDataManager.h"
#include "../include/InfoDrawer.h"
#include "../include/ParamSetter.h"
#include "../include/SurvivalStreamTask.h"
#include "../../effect/include/AnimManager.h"


namespace vw  {
namespace g3d {



static ModuleG3dImpl sInstance;

// When changing the target mode, whether it is permitted to move the focus to SAMPLE mode is determined by the IsHidden() method.
// Variable to dynamically set with these conditions.
bool ModuleG3dImpl::sHiddenFlag = true;


// Function to get instance
ModuleG3d* ModuleG3d::GetInstance()
{
    return &sInstance;
}
ModuleG3dImpl* ModuleG3dImpl::GetInstance()
{
    return &sInstance;
}


/*---------------------------------------------------------------------------*
  Name:         ModuleG3dImpl::ModuleG3dImpl
  Description:  G3D Constructor
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------*/
ModuleG3dImpl::ModuleG3dImpl()
 : ModuleG3d(), m_pWorkMemory(NULL), m_CanDraw(true),
   m_3DEditorActiveChannel(Protocol4C::Constants::CHANNEL_ACTIVE),
   m_3DEditorPassiveChannel(Protocol4C::Constants::CHANNEL_PASSIVE),
   m_3DEditorSurvivalChannel(Protocol4C::Constants::CHANNEL_SURVIVAL),
   m_CSconnect(true),
   m_CSconnectionOpen(false)
{
}

/*---------------------------------------------------------------------------*
  Name:         ModuleG3dImpl::~ModuleG3dImpl
  Description:  G3D Destructor
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------*/
ModuleG3dImpl::~ModuleG3dImpl()
{
}

// Command for playback all is executed
void
ModuleG3dImpl::OnAnmCmd( u32   command    ,     // Command (RESET/PLAY/REV_PLAY/STOP/STEP/REV_STEP/LOOP/SET_FRAME)
                         u32   target     ,     // Target (ALL/3D_MDL/2D_LYT)
                         u32   targetID   ,     // Target ID (do not use when target is ANM_TGT_ALL)
                         void* arg        ,     // Option parameters
                         u32   trigModule  )    // Module ID that is the trigger
{
    switch (target)
    {
    case system::ANM_TGT_ALL:
    case system::ANM_TGT_3D_MDL:
        AnimManager::GetInstance()->OnAnmCmd( command, arg );
        break;
    default:
        break;
    }
}

/*---------------------------------------------------------------------------*
  Name:         ModuleG3dImpl::Init
  Description:  Sample module initialization function
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------*/
void
ModuleG3dImpl::Init()
{
    // Scene, resource initialization
    ResourceManager::Initialize();
    SceneManager::Initialize();
    SceneRenderer::Initialize();
    SceneElemSelector::Initialize();
    UserDataManager::Initialize();
    AnimManager::Initialize();
    CameraMgr::Initialize();
    LightMgr::Initialize();
    InfoDrawer::Initialize();
    SubModeManager::Initialize();
    ParamSetter::Initialize();

    // Scene resource initialization
    SceneManager::GetInstance()->InitializeResource();

    // Config settings
    Config3D_Execute();

    // Initialize stream for existence confirmation
    SurvivalStream_Init();

    // Establish serial communication connection with CS
    const u32 BODY_BUFFER_SIZE = 32 * 1024;
    if (m_CSconnect)
    {
        OpenCSconnection();
    }

    // Serial connection with 3DEditor
    vw::system::Stream_CreateChannel( m_3DEditorActiveChannel, BODY_BUFFER_SIZE );
    vw::system::Stream_CreateChannel( m_3DEditorPassiveChannel, BODY_BUFFER_SIZE );

    // Updates the scene context.
    SceneManager::GetInstance()->UpdateSceneContext();

    // Updates the selection object array.
    SceneElemSelector::GetInstance()->Setup();

    // Load default model
    vw::system::SystemSetting setting;
    vw::system::System_GetSetting(&setting);
    if (strcmp(setting.defaultModelPath.Get(),""))
    {
        ResourceManager::GetInstance()->LoadBinaryData(
            ResourceSet::Type_Packed,
            0,
            0,
            setting.defaultModelPath.Get(),
            false
        );
    }
}

/*---------------------------------------------------------------------------*
  Name:         ModuleG3dImpl::InitMenu
  Description:  Initialize menu
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------*/
void
ModuleG3dImpl::InitMenu( void )
{
    InfoDrawer::GetInstance()->InitializeMenu();
}

/*---------------------------------------------------------------------------*
  Name:         ModuleG3dImpl::InitGraphics
  Description:  Initializes graphics
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------*/
void
ModuleG3dImpl::InitGraphics( void )
{
    SceneRenderer::GetInstance()->InitializeGraphics();
    // Config settings
    Config3D_Execute();
}

/*---------------------------------------------------------------------------*
  Name:         ModuleG3dImpl::PollStream
  Description:  Sample module communication function
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------*/
void
ModuleG3dImpl::PollStream( void )
{
    if (vw::system::App::GetInstance().IsSoundOnlyMode())
    {
        return;
    }

    // Serial communications check
    {
        // Check serial port once every 20 frames
        static int frameCount = 0;
        if (m_CSconnect)
        {
            if ( frameCount == 20 )
            {
                if(!m_CSconnectionOpen)
                {
                    OpenCSconnection();
                }

                if(m_CSconnectionOpen)
                {
                    vw::g3d::ProtocolExecCS_Main(m_Serial);
                    frameCount = 0;
                }
            }
            ++frameCount;
        }
        ProtocolExec_Main();
    }
}

/*---------------------------------------------------------------------------*
  Name:         ModuleG3dImpl::Main
  Description:  The main function of the sample module
  Arguments:    pad  Gets the status of the pad assigned to the sample module.
  Returns:      None.
 *---------------------------------------------------------------------------*/
void
ModuleG3dImpl::Main(vw::system::HidStatus& hidStatus)
{
    // Config load process
    if (system::ConfigFile::IsReloaded()) Config3D_Execute();

    // All types of control with HID
    if (vw::system::G3D_ID == vw::system::App::GetInstance().GetFocusID())
    {
        // Sub-mode control
        SubModeManager::GetInstance()->Main(hidStatus);
        SubModeManager::GetInstance()->GetTouchPanelEvent(hidStatus);

        // Update of control managing the direct control of the module
        if(SubMode_Camera == SubModeManager::GetInstance()->GetCurrentModeType())
        {
            s32 padEvent = system::GetPadEvent(hidStatus, system::GetG3dPadBindings());
            padEvent = InfoDrawer::GetInstance()->GetSelectionCommand(padEvent);

            // Switch resources with the touch panel
            InfoDrawer::GetInstance()->GetTouchPanelEvent(hidStatus,padEvent);

            // Animation-related control update
            AnimManager::GetInstance()->HidControl(padEvent);

            switch (padEvent)
            {
            case system::G3D_SWITCH_PREV_SELECTION:
                InfoDrawer::GetInstance()->ChangeToPrevSelection();
                break;
            case system::G3D_SWITCH_NEXT_SELECTION:
                InfoDrawer::GetInstance()->ChangeToNextSelection();
                break;
            case system::G3D_SWITCH_PREV_VIEW_CAMERA:
                CameraMgr::GetInstance()->SwitchToPrevViewCamera();
                CameraMgr::GetInstance()->SynchronizeTargetCameraToViewCamera();
                InfoDrawer::GetInstance()->SetControlTargetType(InfoDrawer::CONTROL_CAMERA);
                break;
            case system::G3D_SWITCH_NEXT_VIEW_CAMERA:
                CameraMgr::GetInstance()->SwitchToNextViewCamera();
                CameraMgr::GetInstance()->SynchronizeTargetCameraToViewCamera();
                InfoDrawer::GetInstance()->SetControlTargetType(InfoDrawer::CONTROL_CAMERA);
                break;
            case system::G3D_SWITCH_PREV_CONTROL_TARGET_TYPE:
                InfoDrawer::GetInstance()->ChangeToPrevControlTargetType();
                break;
            case system::G3D_SWITCH_NEXT_CONTROL_TARGET_TYPE:
                InfoDrawer::GetInstance()->ChangeToNextControlTargetType();
                break;
            case system::G3D_SWITCH_PREV_CONTROL_TARGET:
                InfoDrawer::GetInstance()->ChangeToPrevControlTarget();
                break;
            case system::G3D_SWITCH_NEXT_CONTROL_TARGET:
                InfoDrawer::GetInstance()->ChangeToNextControlTarget();
                break;
            case system::G3D_SWITCH_PREV_CONTROL_TARGET_SEQUENTIALLY:
                if ( InfoDrawer::GetInstance()->IsFirstControlTarget() ) {
                    InfoDrawer::GetInstance()->ChangeToPrevControlTargetType();
                    InfoDrawer::GetInstance()->ChangeToLastControlTarget();
                } else {
                    InfoDrawer::GetInstance()->ChangeToPrevControlTarget();
                }
                break;
            case system::G3D_SWITCH_NEXT_CONTROL_TARGET_SEQUENTIALLY:
                if ( InfoDrawer::GetInstance()->IsLastControlTarget() ) {
                    InfoDrawer::GetInstance()->ChangeToNextControlTargetType();
                    InfoDrawer::GetInstance()->ChangeToFirstControlTarget();
                } else {
                    InfoDrawer::GetInstance()->ChangeToNextControlTarget();
                }
                break;
            case system::G3D_SWITCH_PREV_RESOURCE_TYPE:
                InfoDrawer::GetInstance()->ChangeToPrevResourceType();
                break;
            case system::G3D_SWITCH_NEXT_RESOURCE_TYPE:
                InfoDrawer::GetInstance()->ChangeToNextResourceType();
                break;
            case system::G3D_SWITCH_PREV_RESOURCE:
                SceneElemSelector::GetInstance()->ChangeToPrevElem(
                    (SceneElemSelector::SelectElemType)InfoDrawer::GetInstance()->GetInfoResourceType()
                );
                break;
            case system::G3D_SWITCH_NEXT_RESOURCE:
                SceneElemSelector::GetInstance()->ChangeToNextElem(
                    (SceneElemSelector::SelectElemType)InfoDrawer::GetInstance()->GetInfoResourceType()
                );
                break;
            case system::G3D_SWITCH_PREV_RESOURCE_SEQUENTIALLY:
                {
                    SceneElemSelector::SelectElemType type =
                        (SceneElemSelector::SelectElemType)InfoDrawer::GetInstance()->GetInfoResourceType();
                    if ( SceneElemSelector::GetInstance()->IsFirstElem( type ) ) {
                        InfoDrawer::GetInstance()->ChangeToPrevResourceType();
                        type = (SceneElemSelector::SelectElemType)InfoDrawer::GetInstance()->GetInfoResourceType();
                        SceneElemSelector::GetInstance()->ChangeToLastElem( type );
                    } else {
                        SceneElemSelector::GetInstance()->ChangeToPrevElem( type );
                    }
                }
                break;
            case system::G3D_SWITCH_NEXT_RESOURCE_SEQUENTIALLY:
                {
                    SceneElemSelector::SelectElemType type =
                        (SceneElemSelector::SelectElemType)InfoDrawer::GetInstance()->GetInfoResourceType();
                    if ( SceneElemSelector::GetInstance()->IsLastElem( type ) ) {
                        InfoDrawer::GetInstance()->ChangeToNextResourceType();
                        type = (SceneElemSelector::SelectElemType)InfoDrawer::GetInstance()->GetInfoResourceType();
                        SceneElemSelector::GetInstance()->ChangeToFirstElem( type );
                    } else {
                        SceneElemSelector::GetInstance()->ChangeToNextElem( type );
                    }
                }
                break;
            case system::G3D_SWITCH_PREV_ANIMATION:
                SceneElemSelector::GetInstance()->ChangeToPrevAnim(
                    (SceneElemSelector::SelectElemType)InfoDrawer::GetInstance()->GetInfoResourceType()
                );
                break;
            case system::G3D_SWITCH_NEXT_ANIMATION:
                SceneElemSelector::GetInstance()->ChangeToNextAnim(
                    (SceneElemSelector::SelectElemType)InfoDrawer::GetInstance()->GetInfoResourceType()
                );
                break;
            case system::G3D_SWITCH_PREV_SCENE:
                SceneElemSelector::GetInstance()->ChangeToPrevSceneEnvironmentSetting();
                break;
            case system::G3D_SWITCH_NEXT_SCENE:
                SceneElemSelector::GetInstance()->ChangeToNextSceneEnvironmentSetting();
                break;
            case system::G3D_CHANGE_INFO_RENDER_STATUS:
                InfoDrawer::GetInstance()->ChangeRenderInfoFlag();
                break;
            default:
                break;
            }
        }
    }

    // Animation bar slider input processing
    InfoDrawer::GetInstance()->Main( hidStatus );

    // Animation frame set
    f32 value = 0.0f;
    bool touch = InfoDrawer::GetInstance()->GetSeekBarInfo(value);
    if( touch )
    {
        SceneElemSelector::SelectElemType type =
            (SceneElemSelector::SelectElemType)InfoDrawer::GetInstance()->GetInfoResourceType();

        AnimController *pController = (AnimController*)SceneElemSelector::GetInstance()->GetTargetElemController( type );
        if( pController )
        {
            if(value == 1.0f)
            {
                value = 0.9999f;
            }
            if(pController->GetAnimSetLinker()->GetLongestAnimation())
            {
                f32 longest_frame = pController->GetAnimSetLinker()->GetLongestAnimation()->GetEndFrame();
                f32 frame = longest_frame * value;
                f32 frame_fate = nn::math::FAbs(pController->GetStepFrame());

                // Correction so that the frame is a multiple of the playback rate
                if(frame_fate > 0.0f)
                {
                    f32 mod;
                    nn::math::FModf(frame/frame_fate+0.5f, &mod);
                    frame = nn::math::Clamp(mod*frame_fate, 0.0f, longest_frame*0.9999f);
                }

                AnimManager::GetInstance()->PauseAll();
                AnimManager::GetInstance()->SetFrame(frame);
            }
        }
    }
}

/*---------------------------------------------------------------------------*
  Name:         ModuleG3dImpl::Calc
  Description:  The main function of the sample module
  Arguments:    pad  Gets the status of the pad assigned to the sample module.
  Returns:      None.
 *---------------------------------------------------------------------------*/
void
ModuleG3dImpl::Calc()
{
    // Update the scenes such as animation
    SceneRenderer::GetInstance()->UpdateScene();
}

/*---------------------------------------------------------------------------*
  Name:         ModuleG3dImpl::Draw
  Description:  Function to display to the screen
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------*/
void
ModuleG3dImpl::Draw()
{
    if ( !m_CanDraw ) return;

    // Scene rendering
    SceneRenderer::GetInstance()->RenderStereoScene();
}

/*---------------------------------------------------------------------------*
  Name:         ModuleG3dImpl::DrawInfo
  Description:  Function to display information on the screen
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------*/
void
ModuleG3dImpl::DrawInfo()
{
    // Check for skip processing
    bool IsUpdateScene = vw::system::App::GetInstance().IsUpdateScene();
    bool IsRenderScene = vw::system::App::GetInstance().IsRenderScene();
    if (!IsUpdateScene || !IsRenderScene) return;
    
    if (vw::system::App::GetInstance().IsSoundOnlyMode()) return;

    // Display graphics mode information
    if ( vw::g3d::InfoDrawer::GetInstance()->IsRender2DInfo() ||
         SubMode_Menu == SubModeManager::GetInstance()->GetCurrentModeType())
    {
        // Get resource information
        SceneElemSelector::SelectElemType type =
            (SceneElemSelector::SelectElemType)InfoDrawer::GetInstance()->GetInfoResourceType();
        nw::gfx::SceneObject* pObject = SceneElemSelector::GetInstance()->GetTargetElem( type );
        AnimSetLinker* pLinker = SceneElemSelector::GetInstance()->GetTargetElemLinker( type );
        IAnimController *pController = SceneElemSelector::GetInstance()->GetTargetElemController( type );
        // Resource name settings
        if (pObject) {
            InfoDrawer::GetInstance()->SetResouceName(SceneElemSelector::GetInstance()->GetTargetElemName( type ));
        } else {
            InfoDrawer::GetInstance()->SetResouceName(NULL);
        }

        // Set the animation information
        if (pLinker) {
            InfoDrawer::GetInstance()->SetAnimationGroup(
                pLinker->GetCurrentAnimSetName(),
                pLinker->GetNumCurrentAnim(AnimElem::Anim_Skeletal),
                pLinker->GetNumCurrentAnim(AnimElem::Anim_Material),
                pLinker->GetNumCurrentAnim(AnimElem::Anim_Visibility) );
            InfoDrawer::GetInstance()->SetMaxFrame(pLinker->GetEndFrame());
            InfoDrawer::GetInstance()->SetFrameCount(pLinker->GetFrame());
            InfoDrawer::GetInstance()->SetFrameRate(pController->GetStepFrame());
        } else {
            InfoDrawer::GetInstance()->SetAnimationGroup(NULL, false, false, false);
            InfoDrawer::GetInstance()->SetMaxFrame( 0.0f );
            InfoDrawer::GetInstance()->SetFrameCount(0.0f);
            InfoDrawer::GetInstance()->SetFrameRate( 0.0f );
        }

        // Sets sub-mode name.
        const char* SubModeName = SubModeManager::GetInstance()->GetCurrentModeName();
        InfoDrawer::GetInstance()->SetSubModeName(SubModeName);

        // Set view camera name
        InfoDrawer::GetInstance()->SetViewCameraName(CameraMgr::GetInstance()->GetViewCameraName());

        // Set operation target name
        switch (InfoDrawer::GetInstance()->GetControlTargetType())
        {
        case InfoDrawer::CONTROL_CAMERA:
            InfoDrawer::GetInstance()->SetControlTargetName(
                CameraMgr::GetInstance()->GetTargetCameraName()
            );
            break;
        case InfoDrawer::CONTROL_LIGHT:
            InfoDrawer::GetInstance()->SetControlTargetName(
                LightMgr::GetInstance()->GetTargetLightName()
            );
            break;
        default:
            break;
        }

        // Set scene setting name
        SceneEnvironmentSettingElem* scene = SceneElemSelector::GetInstance()->GetTargetSceneEnvironmentSetting();
        if (scene) {
            InfoDrawer::GetInstance()->SetSceneName(scene->GetSceneNode()->GetName());
        } else {
            InfoDrawer::GetInstance()->SetSceneName(NULL);
        }

        // Lower screen information display
        vw::system::Draw2DHelper::GetInstance().SetScreenSize(vw::system::Video_GetScreenWidth(), vw::system::Video_GetScreenHeight());
        vw::system::FontDrawer::GetInstance().SetScreenSize(vw::system::Video_GetScreenWidth(), vw::system::Video_GetScreenHeight());
        InfoDrawer::GetInstance()->DrawInfo();

        // Sub-mode information display
        SubModeManager::GetInstance()->Draw();
    }
}

/*---------------------------------------------------------------------------*
  Name:         ModuleG3dImpl::Destroy
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------*/
void
ModuleG3dImpl::Destroy()
{
    if (m_CSconnect)
    {
        m_Serial.Close();
        m_CSconnectionOpen = false;

        if( m_pWorkMemory )
        {
            Memory_Free(m_pWorkMemory);
            m_pWorkMemory = NULL;
        }
    }

    ParamSetter::Finalize();
    InfoDrawer::Finalize();
    SubModeManager::Finalize();
    AnimManager::Finalize();

    UserDataManager::Finalize();

    ResourceManager::Finalize();
    SceneManager::Finalize();

    CameraMgr::Finalize();
    LightMgr::Finalize();

    SceneElemSelector::Finalize();
    SceneRenderer::Finalize();
}

/*---------------------------------------------------------------------------*
  Name:         ModuleG3dImpl::OnModuleFocused
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------*/
void
ModuleG3dImpl::OnModuleFocused()
{
    AnimManager::GetInstance()->OnTargetModuleChanged( vw::system::G3D_ID );
    
    if (!vw::system::App::GetInstance().IsSoundOnlyMode())
    {
        vw::effect::AnimManager::GetInstance()->OnTargetModuleChanged( vw::system::G3D_ID );
    }
}



/*---------------------------------------------------------------------------*
  Name:         OpenCSconnection
  Description:  Open connection with CreativeStudio
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------*/
void
ModuleG3dImpl::OpenCSconnection( void )
{
    m_pWorkMemory = Memory_GetAllocatorDevice()->Alloc(nn::hio::CTR::WORKMEMORY_SIZE);
    nn::Result result = m_Serial.Open(0, m_pWorkMemory);
    if(result.IsSuccess())
    {
        m_CSconnectionOpen = true;
    }
    else
    {
        Memory_Free(m_pWorkMemory);
        m_pWorkMemory = NULL;
    }
}


}   /* namespace g3d */
}   /* namespace vw  */
