/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#include <viewer/system.h>
#include <nw/gfx/gfx_CalculatedTransform.h>
#include "../include/NodeConstraint.h"
#include "../include/SceneRenderer.h"

namespace vw {
namespace g3d {


/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void MultScale(nw::math::MTX34* dstMatrix, const nw::math::MTX34* srcMatrix, const nw::math::VEC3& scale)
{
    dstMatrix->f._00 = srcMatrix->f._00 * scale.x;
    dstMatrix->f._10 = srcMatrix->f._10 * scale.x;
    dstMatrix->f._20 = srcMatrix->f._20 * scale.x;

    dstMatrix->f._01 = srcMatrix->f._01 * scale.y;
    dstMatrix->f._11 = srcMatrix->f._11 * scale.y;
    dstMatrix->f._21 = srcMatrix->f._21 * scale.y;

    dstMatrix->f._02 = srcMatrix->f._02 * scale.z;
    dstMatrix->f._12 = srcMatrix->f._12 * scale.z;
    dstMatrix->f._22 = srcMatrix->f._22 * scale.z;
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void CopyRotate(nw::math::MTX34* dstMatrix, const nw::math::MTX34& srcMatrix)
{
    dstMatrix->f._00 = srcMatrix.f._00;
    dstMatrix->f._10 = srcMatrix.f._10;
    dstMatrix->f._20 = srcMatrix.f._20;

    dstMatrix->f._01 = srcMatrix.f._01;
    dstMatrix->f._11 = srcMatrix.f._11;
    dstMatrix->f._21 = srcMatrix.f._21;

    dstMatrix->f._02 = srcMatrix.f._02;
    dstMatrix->f._12 = srcMatrix.f._12;
    dstMatrix->f._22 = srcMatrix.f._22;
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void CopyTranslate(nw::math::MTX34* dstMatrix, const nw::math::MTX34& srcMatrix)
{
    dstMatrix->f._03 = srcMatrix.f._03;
    dstMatrix->f._13 = srcMatrix.f._13;
    dstMatrix->f._23 = srcMatrix.f._23;
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
CalculateWorldMatrix(
    nw::math::MTX34* transformMatrix,
    nw::math::VEC3* scale,
    const nw::gfx::CalculatedTransform& localTransform,
    const nw::gfx::CalculatedTransform& parentWorldTransform,
    const nw::gfx::CalculatedTransform& parentLocalTransform,
    const bool inheritTransform[ 3 ]
) {
    NW_NULL_ASSERT(transformMatrix);

    const nw::math::MTX34& parentMatrix = parentWorldTransform.TransformMatrix();
    nw::math::VEC3 localTranslate = localTransform.TransformMatrix().GetColumn(3);

    // TR(i) = TR(i-1) * S(i-1) * t(i) * r(i)
    MTX34Identity(transformMatrix);

    if (inheritTransform[ 1 ]) CopyRotate(transformMatrix, parentMatrix);
    if (inheritTransform[ 2 ]) CopyTranslate(transformMatrix, parentMatrix);
    if (inheritTransform[ 0 ]) {
        MultScale(transformMatrix, transformMatrix, parentLocalTransform.Scale());
        nw::math::VEC3Mult(scale, &parentWorldTransform.Scale(), &localTransform.Scale());
    } else {
        scale->Set(localTransform.Scale());
    }
    nw::math::MTX34MultTranslate(transformMatrix, transformMatrix, &localTranslate);
    nw::math::MTX33Mult(transformMatrix, transformMatrix, &localTransform.TransformMatrix());
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
NodeConstraint::NodeConstraint(
    nw::gfx::Model* model,
    nw::os::IAllocator* allocator)
: m_Constraints(allocator),
  m_Skeleton(NULL),
  m_Allocator(allocator)
{
    if (
        model &&
        model->GetResSceneObject().GetTypeInfo() == nw::gfx::res::ResSkeletalModel::TYPE_INFO
    ) {
        m_Skeleton = nw::ut::DynamicCast<nw::gfx::SkeletalModel*>(model)->GetSkeleton();
    }
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
NodeConstraint::~NodeConstraint()
{
    ClearConstraints();
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
NodeConstraint::AddConstraint(const Constraint& constraint, const char* boneName)
{
    if (!m_Skeleton) return;
    if ((nw::gfx::TransformNode*)m_Skeleton == constraint.m_ConstrainedObject) return;

    int boneIdx = m_Skeleton->GetResSkeleton().GetBonesIndex(boneName);
    if (boneIdx < 0) return;

    CallbackConstraint* obj = NULL;
    ConstraintArray::iterator it = m_Constraints.begin();
    for ( ; it != m_Constraints.end(); ) {
        if ( (*it)->GetTargetBoneIndex() == boneIdx ) {
            obj = *it;
        } else {
            (*it)->DeleteConstraint(constraint.m_ConstrainedObject);
            if ((*it)->GetConstraintCount() == 0) {
                m_Skeleton->PostCalculateMatrixSignal().Disconnect((*it));
                (*it)->~CallbackConstraint();
                m_Allocator->Free((*it));
                ConstraintArray::iterator next = it; ++next;
                it = m_Constraints.erase(it, next);
                continue;
            }
        }
        ++it;
    }

    if ( !obj ) {
        void* slotMemory = m_Allocator->Alloc(sizeof(CallbackConstraint));
        obj = new(slotMemory) CallbackConstraint(m_Skeleton, boneIdx, m_Allocator);
        m_Skeleton->PostCalculateMatrixSignal().Connect(obj);
        m_Constraints.push_back(obj);
    }

    obj->AddConstraint(constraint);
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
NodeConstraint::DeleteConstraint(nw::gfx::TransformNode* constrainedObject)
{
    if (!m_Skeleton) return;

    ConstraintArray::iterator it = m_Constraints.begin();
    for ( ; it != m_Constraints.end(); ) {
        (*it)->DeleteConstraint(constrainedObject);
        if ((*it)->GetConstraintCount() == 0) {
            m_Skeleton->PostCalculateMatrixSignal().Disconnect((*it));
            (*it)->~CallbackConstraint();
            m_Allocator->Free((*it));
            ConstraintArray::iterator next = it; ++next;
            it = m_Constraints.erase(it, next);
            continue;
        }
        ++it;
    }
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
NodeConstraint::DeleteConstraint(const char* boneName)
{
    if (!m_Skeleton) return;

    int boneIdx = m_Skeleton->GetResSkeleton().GetBonesIndex(boneName);
    if (boneIdx < 0) return;

    CallbackConstraint* obj = NULL;
    ConstraintArray::iterator it = m_Constraints.begin();
    ConstraintArray::iterator end = m_Constraints.end();
    for ( ; it != end; ++it ) {
        if ( (*it)->GetTargetBoneIndex() == boneIdx ) { obj = *it; break; }
    }

    if ( obj ) {
        obj->ClearConstraints();
        m_Skeleton->PostCalculateMatrixSignal().Disconnect(obj);
        obj->~CallbackConstraint();
        m_Allocator->Free(obj);
        ConstraintArray::iterator next = it; ++next;
        m_Constraints.erase(it, next);
    }
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
NodeConstraint::ClearConstraints()
{
    if (!m_Skeleton) return;

    ConstraintArray::iterator it = m_Constraints.begin();
    ConstraintArray::iterator end = m_Constraints.end();
    for ( ; it != end; ++it ) {
        (*it)->ClearConstraints();
        m_Skeleton->PostCalculateMatrixSignal().Disconnect((*it));
        (*it)->~CallbackConstraint();
        m_Allocator->Free((*it));
    }

    if(m_Constraints.size()!=0)
        m_Constraints.clear();
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
const nw::gfx::CalculatedTransform*
NodeConstraint::GetBoneTransform(nw::gfx::TransformNode* constrainedObject)
{
    if (!constrainedObject) return NULL;

    ConstraintArray::iterator it = m_Constraints.begin();
    ConstraintArray::iterator end = m_Constraints.end();
    for ( ; it != end; ++it ) {
        CallbackConstraint::ConstraintArray::iterator bone_it = (*it)->m_Constraints.begin();
        CallbackConstraint::ConstraintArray::iterator bone_end = (*it)->m_Constraints.end();
        for ( ; bone_it != bone_end; ++bone_it ) {
            if ( constrainedObject == bone_it->m_ConstrainedObject ) {
                return m_Skeleton->WorldTransformPose().GetTransform((*it)->m_BoneIdx);
            }
        }
    }

    return NULL;
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
NodeConstraint::CallbackConstraint::CallbackConstraint(
    nw::gfx::Skeleton* skeleton,
    int boneIdx,
    nw::os::IAllocator* allocator
) : nw::gfx::Skeleton::CalculateMatrixSlot(allocator),
    m_Constraints(allocator),
    m_Skeleton(skeleton),
    m_BoneIdx(boneIdx)
{}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
NodeConstraint::CallbackConstraint::~CallbackConstraint()
{
    ClearConstraints();
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
NodeConstraint::CallbackConstraint::AddConstraint(const Constraint& constraint)
{
    Constraint* obj = NULL;
    ConstraintArray::iterator it = m_Constraints.begin();
    ConstraintArray::iterator end = m_Constraints.end();
    for ( ; it != end; ++it ) {
        if ( constraint.m_ConstrainedObject == it->m_ConstrainedObject ) {
            obj = &(*it); break;
        }
    }

    if ( obj == NULL ) {
        m_Constraints.push_back(constraint);
        // Add to axis rendering target
        {
            SceneRenderer::DrawAxisInfo info;
            info.axisType = SceneRenderer::AxisConstraintTargetBone;
            info.axisTransform = m_Skeleton->WorldTransformPose().GetTransform(m_BoneIdx);
            info.key1 = m_Skeleton;
            info.key2 = constraint.m_ConstrainedObject;
            SceneRenderer::GetInstance()->AddDrawAxis(info);
        }
    } else {
        memcpy( obj->m_OffsetScale, constraint.m_OffsetScale, sizeof(f32) * 3 );
        memcpy( obj->m_OffsetRotate, constraint.m_OffsetRotate, sizeof(f32) * 3 );
        memcpy( obj->m_OffsetTranslate, constraint.m_OffsetTranslate, sizeof(f32) * 3 );
        memcpy( obj->m_UseOffset, constraint.m_UseOffset, sizeof(bool) * 3 );
        memcpy( obj->m_InheritTransform, constraint.m_InheritTransform, sizeof(bool) * 3 );
     }
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
NodeConstraint::CallbackConstraint::DeleteConstraint(nw::gfx::TransformNode* constrainedObject)
{
    if (!constrainedObject) return;

    ConstraintArray::iterator it = m_Constraints.begin();
    for ( ; it != m_Constraints.end(); ) {
        if ( constrainedObject == it->m_ConstrainedObject )
        {
            // Restore to the default value registered in ResTransformNode at the time of deletion
            if ( it->m_ConstrainedObject->GetResTransformNode().IsValid() ) {
                it->m_ConstrainedObject->Transform().SetTransform(
                    it->m_ConstrainedObject->GetResTransformNode().GetTransform()
                );
            // If there is no ResTransformNode, restore to unit conversion
            } else {
                MTX34Identity(&(it->m_ConstrainedObject->Transform().DirectTransformMatrix()));
                it->m_ConstrainedObject->Transform().SetScale(1.0f, 1.0f, 1.0f);
            }

            // Remove from axis rendering target
            SceneRenderer::GetInstance()->DeleteDrawAxis(
                m_Skeleton,
                it->m_ConstrainedObject
            );

            ConstraintArray::iterator next = it; ++next;
            it = m_Constraints.erase( it, next );
            continue;
        }
        it++;
    }
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
NodeConstraint::CallbackConstraint::ClearConstraints()
{
    ConstraintArray::iterator it = m_Constraints.begin();
    ConstraintArray::iterator end = m_Constraints.end();
    for ( ; it != end; ++it )
    {
        // Restore to the default value registered in ResTransformNode at the time of deletion
        if ( it->m_ConstrainedObject->GetResTransformNode().IsValid() ) {
            it->m_ConstrainedObject->Transform().SetTransform(
                it->m_ConstrainedObject->GetResTransformNode().GetTransform()
            );
        // If there is no ResTransformNode, restore to unit conversion
        } else {
            MTX34Identity(&(it->m_ConstrainedObject->Transform().DirectTransformMatrix()));
            it->m_ConstrainedObject->Transform().SetScale(1.0f, 1.0f, 1.0f);
        }

        // Remove from axis rendering target
        SceneRenderer::GetInstance()->DeleteDrawAxis(
            m_Skeleton,
            it->m_ConstrainedObject
        );
    }

    if(m_Constraints.size()!=0)
        m_Constraints.clear();
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
int
NodeConstraint::CallbackConstraint::GetConstraintCount()
{
    return m_Constraints.size();
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
int
NodeConstraint::CallbackConstraint::GetTargetBoneIndex(void)
{
    return m_BoneIdx;
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
NodeConstraint::CallbackConstraint::Invoke(nw::gfx::Skeleton* skeleton, int boneIdx)
{
    if (skeleton != m_Skeleton || boneIdx != m_BoneIdx) return;

    nw::gfx::CalculatedTransform* localPose = skeleton->LocalTransformPose().GetTransform(boneIdx);
    nw::gfx::CalculatedTransform* worldPose = skeleton->WorldTransformPose().GetTransform(boneIdx);

    ConstraintArray::iterator end = m_Constraints.end();
    for (ConstraintArray::iterator iter = m_Constraints.begin() ; iter != end ; ++iter)
    {
        // Set local offset conversion information
        nw::math::Transform3 transForm(
            nw::math::VEC3( 1.0f, 1.0f, 1.0f ),
            nw::math::VEC3( 0.0f, 0.0f, 0.0f ),
            nw::math::VEC3( 0.0f, 0.0f, 0.0f )
        );
        if (iter->m_UseOffset[0]) transForm.scale = nw::math::VEC3(iter->m_OffsetScale);
        if (iter->m_UseOffset[1]) transForm.rotate = nw::math::VEC3(iter->m_OffsetRotate);
        if (iter->m_UseOffset[2]) transForm.translate = nw::math::VEC3(iter->m_OffsetTranslate);

        // Local offset conversion
        nw::gfx::CalculatedTransform localOffset;
        localOffset.SetTransform( transForm );

        // Calculation of world conversion matrix with offset
        CalculateWorldMatrix(
            &iter->m_ConstrainedObject->Transform().DirectTransformMatrix(),
            &iter->m_ConstrainedObject->Transform().DirectScale(),
            localOffset,
            *worldPose,
            *localPose,
            iter->m_InheritTransform
        );

        // Invalidate cache
        //iter->m_ConstrainedObject->Transform().EnableFlags(
        //    nw::gfx::CalculatedTransform::FLAG_IS_DIRTY
        //);

        // Updates the final world conversion matrix
        SceneRenderer::GetInstance()->GetWorldMatrixUpdater()->UpdateBasic(
            &iter->m_ConstrainedObject->WorldMatrix(),
            &iter->m_ConstrainedObject->WorldTransform(),
            iter->m_ConstrainedObject->Transform(),
            nw::gfx::CalculatedTransform::Identity(),
            nw::gfx::CalculatedTransform::Identity()
        );
        iter->m_ConstrainedObject->InvalidateInverseWorldMatrix();

        // Enables the child node update flag
        nw::gfx::SceneNodeChildren::iterator it = iter->m_ConstrainedObject->GetChildBegin();
        nw::gfx::SceneNodeChildren::iterator end_  = iter->m_ConstrainedObject->GetChildEnd();
        for ( ; it != end_; ++it )
        {
            nw::gfx::TransformNode* node = nw::ut::DynamicCast<nw::gfx::TransformNode*>((*it));
            if ( node ) {
                node->Transform().EnableFlags(nw::gfx::CalculatedTransform::FLAG_IS_DIRTY);
            } else {
                (*it)->EnableTraversalResults(nw::gfx::SceneNode::FLAG_IS_DIRTY);
            }
        }
    }
}

}
}
