/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#include <viewer/system.h>
#include "../include/ParamSetter.h"

using namespace Protocol4C;

namespace vw  {
namespace g3d {

const u32   BODY_BUFFER_SIZE    =   32 * 1024;

static void     ExecCommand          ( const Header& header, void* body );
static void     ExecCategoryFile     ( const Header& header, void* body );
static void     ExecCategoryAnimation( const Header& header, void* body );
static void     ExecCategoryMaterial ( const Header& header, void* body );
static void     ExecCategoryObject   ( const Header& header, void* body );
static void     ExecCategoryOperation( const Header& header, void* body );
static void     ExecCategoryModel   ( const Header& header, void* body );
static void     ExecCategoryMesh    ( const Header& header, void* body );
static void     ExecCategoryLUT     ( const Header& header, void* body );
static void     ExecCategoryBone    ( const Header& header, void* body );
static void     ExecCategoryFog     ( const Header& header, void* body );
static void     ExecCategoryLight   ( const Header& header, void* body );
static void     ExecCategoryCamera  ( const Header& header, void* body );
static void     ExecCategoryScene   ( const Header& header, void* body );
static void     ExecCategoryEdit    ( const Header& header, void* body );
static void     ExecCategoryAnimCtrl( const Header& header, void* body );
static void     ExecCategoryUserData( const Header& header, void* body );

static void     UnknownCommandError( u16 type );

// Buffer for reading the body portion
static u32       sBodyBuffer[ BODY_BUFFER_SIZE / sizeof(u32) ];
static bool     sProtocolClearFlg = true;
static bool     sHeaderRead = false;

/*---------------------------------------------------------------------------*
  Name:         ProtocolExec_Main
  Description:  3D protocol processing main routine
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
ProtocolExec_Main()
{
    static Header header;

    const u32 active  = ModuleG3dImpl::GetInstance()->Get3DEditorActiveChannel();
    const u32 passive = ModuleG3dImpl::GetInstance()->Get3DEditorPassiveChannel();

    if (vw::system::Stream_Polling( active ))
    {
        system::App::GetInstance().SetFocusID( system::G3D_ID );
    }
    vw::system::Stream_Polling( passive );

    while( vw::system::Stream_IsConnected( active ) )
    {
        // Checks the header read complete flag
        if ( !sHeaderRead )
        {
            u32 ReadableBytes = vw::system::Stream_GetReadableBytes( active );
            if( ReadableBytes < sizeof(Header)) return;

            vw::system::Stream_Read( active, sBodyBuffer, sizeof(Header) );

            if( !vw::system::Stream_IsConnected( active ) ){
                NW_VIEWER_LOG("Connection lost.\n");
                sHeaderRead = false;
                return;
            }

            // Copy header information
            header.type = ((Header*)sBodyBuffer)->type;
            header.size = ((Header*)sBodyBuffer)->size;
            header.arraySize = ((Header*)sBodyBuffer)->arraySize;

            // Enables the header read complete flag
            sHeaderRead = true;
        }

        if ( sHeaderRead )
        {
            u32 ReadableBytes = vw::system::Stream_GetReadableBytes( active );
            if ( ReadableBytes < header.size + header.arraySize ) return;

            // Load the data itself
            u32 rSize = vw::system::Stream_Read( active, sBodyBuffer, header.size + header.arraySize );
            if ( rSize != header.size + header.arraySize ) return;

            ExecCommand( header, sBodyBuffer );

            // Reset the header read complete flag
            sHeaderRead = false;

            // Sends response notification to the host side
            int reply = 1;
            vw::system::Stream_Write(active, &reply, sizeof(int) );
        }
    }
}

/*---------------------------------------------------------------------------*
  Name:         ExecCommand
  Description:  Command analysis
  Arguments:    header  Reference to the header portion
                body    Pointer to the body portion. NULL if the message size is 0
  Returns:      None.
 *---------------------------------------------------------------------------


*/
static void
ExecCommand( const Header& header, void* body )
{
    switch ( header.type & Constants::MESSAGE_CATEGORY_MASK )
    {
    //---------------------------------
    // File-related category
    case MessageCategory_File:
        ExecCategoryFile( header, body );
        break;

    //---------------------------------
    // Material Management-related Category
    case MessageCategory_Material:
        ExecCategoryMaterial( header, body );
        break;

    //---------------------------------
    // Model State Change Category
    case MessageCategory_Object:
        ExecCategoryObject( header, body );
        break;

    //---------------------------------
    // Rendering Change Category for Models
    case MessageCategory_Model:
        ExecCategoryModel( header, body );
        break;

    //---------------------------------
    // Rendering Change Category for Mesh
    case MessageCategory_Mesh:
        ExecCategoryMesh( header, body );
        break;

    //---------------------------------
    // Rendering Change Category for Bones
    case MessageCategory_Bone:
        ExecCategoryBone( header, body );
        break;

    //---------------------------------
    // Rendering Change Category for LUT
    case MessageCategory_LUT:
        ExecCategoryLUT( header, body );
        break;

    //---------------------------------
    // Rendering Change Category for Fog
    case MessageCategory_Fog:
        ExecCategoryFog( header, body );
        break;

    //---------------------------------
    // Rendering Change Category for Light
    case MessageCategory_Light:
        ExecCategoryLight( header, body );
        break;

    //---------------------------------
    // Rendering Change Category for Cameras
    case MessageCategory_Camera:
        ExecCategoryCamera( header, body );
        break;

    //---------------------------------
    // Rendering Change Category for Scenes
    case MessageCategory_Scene:
        ExecCategoryScene( header, body );
        break;

    //---------------------------------
    // Animation change category
    case MessageCategory_Animation:
        ExecCategoryAnimation( header, body );
        break;

    //---------------------------------
    // User data change category
    case MessageCategory_UserData:
        ExecCategoryUserData( header, body );
        break;

    //---------------------------------
    // Operation change category
    case MessageCategory_Operation:
        ExecCategoryOperation( header, body );
        break;

    //---------------------------------
    default:
        UnknownCommandError( header.type );
    }
}



/*---------------------------------------------------------------------------*
  Name:         ExecCategoryFile
  Description:  Execute system-related category commands
  Arguments:    header  Reference to the header portion
                body    Pointer to the body portion. NULL if the message size is 0
  Returns:      None.
 *---------------------------------------------------------------------------


*/
static void
ExecCategoryFile( const Header& header, void* body )
{
    const u32 active = ModuleG3dImpl::GetInstance()->Get3DEditorActiveChannel();

    switch ( header.type )
    {
    //---------------------------------
    //  Load command
    case MESSAGEPACKET_File_LoadBinaryData::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_File_LoadBinaryData----\n");
        {
            MESSAGEPACKET_File_LoadBinaryData* loadBinaryData = reinterpret_cast<MESSAGEPACKET_File_LoadBinaryData*>(body);
            ResourceManager::GetInstance()->LoadBinaryData(
                (vw::g3d::ResourceSet::BinaryFileType)loadBinaryData->Type,
                loadBinaryData->Id,
                active,
                loadBinaryData->FilePath.Value
            );
        }
        break;
    case MESSAGEPACKET_File_Close::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_File_Close----\n");
        {
            MESSAGEPACKET_File_Close* closeBinaryData = reinterpret_cast<MESSAGEPACKET_File_Close*>(body);
            ResourceManager::GetInstance()->DeleteResource(
                (vw::g3d::ResourceSet::BinaryFileType)closeBinaryData->Type,
                closeBinaryData->Id
            );
        }
        break;
    case MESSAGEPACKET_File_CloseAll::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_File_CloseAll----\n");
        ResourceManager::GetInstance()->ClearResources(active);
        break;
    //---------------------------------
    default:
        UnknownCommandError( header.type );
    }
}

/*---------------------------------------------------------------------------*
  Name:         ExecCategoryFile
  Description:  Execute system-related category commands
  Arguments:    header  Reference to the header portion
                body    Pointer to the body portion. NULL if the message size is 0
  Returns:      None.
 *---------------------------------------------------------------------------


*/
static void
ExecCategoryMaterial( const Header& header, void* body )
{
    vw::g3d::ParamSetter &paramSetter = vw::g3d::ParamSetter::GetInstance();
    switch ( header.type )
    {
    //---------------------------------
    case MESSAGEPACKET_Material_SetCullingMode::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Material_SetCullingMode----\n");
        paramSetter.SetCullingMode(((MESSAGEPACKET_Material_SetCullingMode*)body)->Mode);
        break;
    case MESSAGEPACKET_Material_SetTranslucencyKind::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Material_SetTranslucencyKind----\n");
        paramSetter.SetTranslucencyKind(((MESSAGEPACKET_Material_SetTranslucencyKind*)body)->Kind);
        break;
    //  Color change command
    case MESSAGEPACKET_Material_SetColor::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Material_SetColor----\n");

        // Determination of parameter to change
        if(((MESSAGEPACKET_Material_SetColor*)body)->ModifiedProperties & MESSAGEPACKET_Material_SetColor::PropertyFlag_Ambient)
        {
            NW_VIEWER_LOG("----Exec SetAmbient()----\n");
            paramSetter.SetAmbient(&((MESSAGEPACKET_Material_SetColor*)body)->Ambient.R, ((MESSAGEPACKET_Material_SetColor*)body)->ModifiedPropertyElements);
        }

        if(((MESSAGEPACKET_Material_SetColor*)body)->ModifiedProperties & MESSAGEPACKET_Material_SetColor::PropertyFlag_Diffuse)
        {
            NW_VIEWER_LOG("----Exec SetDiffuse()----\n");
            paramSetter.SetDiffuse(&((MESSAGEPACKET_Material_SetColor*)body)->Diffuse.R, ((MESSAGEPACKET_Material_SetColor*)body)->ModifiedPropertyElements);
        }

        if(((MESSAGEPACKET_Material_SetColor*)body)->ModifiedProperties & MESSAGEPACKET_Material_SetColor::PropertyFlag_Emission)
        {
            NW_VIEWER_LOG("----Exec SetEmission()----\n");
            paramSetter.SetEmission(&((MESSAGEPACKET_Material_SetColor*)body)->Emission.R, ((MESSAGEPACKET_Material_SetColor*)body)->ModifiedPropertyElements);
        }

        if(((MESSAGEPACKET_Material_SetColor*)body)->ModifiedProperties & MESSAGEPACKET_Material_SetColor::PropertyFlag_Specular0)
        {
            NW_VIEWER_LOG("----Exec SetSpecular0()----\n");
            paramSetter.SetSpecular0(&((MESSAGEPACKET_Material_SetColor*)body)->Specular0.R, ((MESSAGEPACKET_Material_SetColor*)body)->ModifiedPropertyElements);
        }

        if(((MESSAGEPACKET_Material_SetColor*)body)->ModifiedProperties & MESSAGEPACKET_Material_SetColor::PropertyFlag_Specular1)
        {
            NW_VIEWER_LOG("----Exec SetSpecular1()----\n");
            paramSetter.SetSpecular1(&((MESSAGEPACKET_Material_SetColor*)body)->Specular1.R, ((MESSAGEPACKET_Material_SetColor*)body)->ModifiedPropertyElements);
        }

        if(((MESSAGEPACKET_Material_SetColor*)body)->ModifiedProperties & MESSAGEPACKET_Material_SetColor::PropertyFlag_VertexColorScale)
        {
            NW_VIEWER_LOG("----Exec SetVertexColorScale()----\n");
            // In SetAmbient, the vertex color scale is overwritten, so set the vertex color scale after SetAmbient.
            // 
            paramSetter.SetVertexColorScale(((MESSAGEPACKET_Material_SetColor*)body)->VertexColorScale);
        }

        if(((MESSAGEPACKET_Material_SetColor*)body)->ModifiedProperties & MESSAGEPACKET_Material_SetColor::PropertyFlag_Constant0)
        {
            NW_VIEWER_LOG("----Exec SetConstant0()----\n");
            paramSetter.SetConstant0(&((MESSAGEPACKET_Material_SetColor*)body)->Constant0.R, ((MESSAGEPACKET_Material_SetColor*)body)->ModifiedPropertyElements);
        }
        if(((MESSAGEPACKET_Material_SetColor*)body)->ModifiedProperties & MESSAGEPACKET_Material_SetColor::PropertyFlag_Constant1)
        {
            NW_VIEWER_LOG("----Exec SetConstant1()----\n");
            paramSetter.SetConstant1(&((MESSAGEPACKET_Material_SetColor*)body)->Constant1.R, ((MESSAGEPACKET_Material_SetColor*)body)->ModifiedPropertyElements);
        }
        if(((MESSAGEPACKET_Material_SetColor*)body)->ModifiedProperties & MESSAGEPACKET_Material_SetColor::PropertyFlag_Constant2)
        {
            NW_VIEWER_LOG("----Exec SetConstant2()----\n");
            paramSetter.SetConstant2(&((MESSAGEPACKET_Material_SetColor*)body)->Constant2.R, ((MESSAGEPACKET_Material_SetColor*)body)->ModifiedPropertyElements);
        }
        if(((MESSAGEPACKET_Material_SetColor*)body)->ModifiedProperties & MESSAGEPACKET_Material_SetColor::PropertyFlag_Constant3)
        {
            NW_VIEWER_LOG("----Exec SetConstant3()----\n");
            paramSetter.SetConstant3(&((MESSAGEPACKET_Material_SetColor*)body)->Constant3.R, ((MESSAGEPACKET_Material_SetColor*)body)->ModifiedPropertyElements);
        }
        if(((MESSAGEPACKET_Material_SetColor*)body)->ModifiedProperties & MESSAGEPACKET_Material_SetColor::PropertyFlag_Constant4)
        {
            NW_VIEWER_LOG("----Exec SetConstant4()----\n");
            paramSetter.SetConstant4(&((MESSAGEPACKET_Material_SetColor*)body)->Constant4.R, ((MESSAGEPACKET_Material_SetColor*)body)->ModifiedPropertyElements);
        }
        if(((MESSAGEPACKET_Material_SetColor*)body)->ModifiedProperties & MESSAGEPACKET_Material_SetColor::PropertyFlag_Constant5)
        {
            NW_VIEWER_LOG("----Exec SetConstant5()----\n");
            paramSetter.SetConstant5(&((MESSAGEPACKET_Material_SetColor*)body)->Constant5.R, ((MESSAGEPACKET_Material_SetColor*)body)->ModifiedPropertyElements);
        }
        break;
    case MESSAGEPACKET_Material_SetTexture::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Material_SetTexture----\n");
        paramSetter.SetTexture(((MESSAGEPACKET_Material_SetTexture*)body));
        break;
    case MESSAGEPACKET_Material_SetFog::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Material_SetFog----\n");
        paramSetter.SetFog((MESSAGEPACKET_Material_SetFog*)body);
        break;
    case MESSAGEPACKET_Material_SetLight::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Material_SetLight----\n");
        paramSetter.SetLight((MESSAGEPACKET_Material_SetLight*)body);
        break;
    case MESSAGEPACKET_Material_SetFShaderConfig::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Material_SetFShaderConfig----\n");
        paramSetter.SetFShaderConfig(((MESSAGEPACKET_Material_SetFShaderConfig*)body)->Config);
        break;
    case MESSAGEPACKET_Material_SetFragmentLighting::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Material_SetFragmentLighting----\n");
        paramSetter.SetFragmentLighting(((MESSAGEPACKET_Material_SetFragmentLighting*)body));
        break;
    case MESSAGEPACKET_Material_SetAlphaTest::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Material_SetAlphaTest----\n");
        paramSetter.SetAlphaTest(((MESSAGEPACKET_Material_SetAlphaTest*)body));
        break;
    case MESSAGEPACKET_Material_SetDepthTest::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Material_SetDepthTest----\n");
        paramSetter.SetDepthTest(((MESSAGEPACKET_Material_SetDepthTest*)body));
        break;
    case MESSAGEPACKET_Material_SetStencilTest::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Material_SetStencilTest----\n");
        paramSetter.SetStencilTest(((MESSAGEPACKET_Material_SetStencilTest*)body));
        break;
    case MESSAGEPACKET_Material_SetBlend::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Material_SetBlend----\n");
        paramSetter.SetBlend(((MESSAGEPACKET_Material_SetBlend*)body));
        break;
    case MESSAGEPACKET_Material_SetTextureCombiner::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Material_SetTextureCombiner----\n");
        paramSetter.SetTextureCombiner(((MESSAGEPACKET_Material_SetTextureCombiner*)body));
        break;
    case MESSAGEPACKET_Material_SetFShaderBufferColor::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Material_SetFShaderBufferColor----\n");
        paramSetter.SetBufferColor(&((MESSAGEPACKET_Material_SetFShaderBufferColor*)body)->BufferColor.R);
        break;
    case MESSAGEPACKET_Material_SetTextureCoordinateConfig::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Material_SetTextureCoordinateConfig----\n");
        paramSetter.SetTexCoordConfig(((MESSAGEPACKET_Material_SetTextureCoordinateConfig*)body)->TextureCoordinateConfig);
        break;
    case MESSAGEPACKET_Material_SetTextureMinFilter::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Material_SetTextureMinFilter----\n");
        paramSetter.SetTextureMinFilter(((MESSAGEPACKET_Material_SetTextureMinFilter*)body)->Index, ((MESSAGEPACKET_Material_SetTextureMinFilter*)body)->MinFilter);
        break;
    case MESSAGEPACKET_Material_SetTextureMagFilter::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Material_SetTextureMagFilter----\n");
        paramSetter.SetTextureMagFilter(((MESSAGEPACKET_Material_SetTextureMagFilter*)body)->Index, ((MESSAGEPACKET_Material_SetTextureMagFilter*)body)->MagFilter);
        break;
    case MESSAGEPACKET_Material_SetTextureWrapS::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Material_SetTextureWrapS----\n");
        paramSetter.SetTextureWrap(ParamSetter::Wrap_S, ((MESSAGEPACKET_Material_SetTextureWrapS*)body)->Index, ((MESSAGEPACKET_Material_SetTextureWrapS*)body)->WrapS);
        break;
    case MESSAGEPACKET_Material_SetTextureWrapT::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Material_SetTextureWrapT----\n");
        paramSetter.SetTextureWrap(ParamSetter::Wrap_T, ((MESSAGEPACKET_Material_SetTextureWrapT*)body)->Index, ((MESSAGEPACKET_Material_SetTextureWrapT*)body)->WrapT);
        break;
    case MESSAGEPACKET_Material_SetTextureMinLod::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Material_SetTextureMinLod----\n");
        paramSetter.SetTextureMinLod(((MESSAGEPACKET_Material_SetTextureMinLod*)body)->Index, (unsigned int)(((MESSAGEPACKET_Material_SetTextureMinLod*)body)->MinLod));
        break;
    case MESSAGEPACKET_Material_SetTextureLodBias::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Material_SetTextureLodBias----\n");
        paramSetter.SetTextureLodBias(((MESSAGEPACKET_Material_SetTextureLodBias*)body)->Index, (((MESSAGEPACKET_Material_SetTextureLodBias*)body)->LodBias));
        break;
    case MESSAGEPACKET_Material_SetTextureBorderColor::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Material_SetTextureBorderColor----\n");
        paramSetter.SetTextureBorderColor(((MESSAGEPACKET_Material_SetTextureBorderColor*)body)->Index, &((MESSAGEPACKET_Material_SetTextureBorderColor*)body)->BorderColor.R);
        break;
    case MESSAGEPACKET_Material_SetTexCoordSourceCoordinate::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Material_SetTexCoordSourceCoordinate----\n");
        paramSetter.SetTexcoordSourceCoord(((MESSAGEPACKET_Material_SetTexCoordSourceCoordinate*)body)->Index, ((MESSAGEPACKET_Material_SetTexCoordSourceCoordinate*)body)->SourceCoordinate);
        break;
    case MESSAGEPACKET_Material_SetTexCoordMappingMethod::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Material_SetTexCoordMappingMethod----\n");
        paramSetter.SetTexcoordMappingMethod(((MESSAGEPACKET_Material_SetTexCoordMappingMethod*)body)->Index, ((MESSAGEPACKET_Material_SetTexCoordMappingMethod*)body)->MappingMethod);
        break;
    case MESSAGEPACKET_Material_SetTexCoordReferenceCamera::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Material_SetTexCoordReferenceCamera----\n");
        paramSetter.SetTexcoordReferenceCamera(((MESSAGEPACKET_Material_SetTexCoordReferenceCamera*)body)->Index, ((MESSAGEPACKET_Material_SetTexCoordReferenceCamera*)body)->ReferenceCamera);
        break;
    case MESSAGEPACKET_Material_SetTexCoordMatrixMode::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Material_SetTexCoordMatrixMode----\n");
        paramSetter.SetTexcoordMatrixMode(((MESSAGEPACKET_Material_SetTexCoordMatrixMode*)body)->Index, ((MESSAGEPACKET_Material_SetTexCoordMatrixMode*)body)->MatrixMode);
        break;
    case MESSAGEPACKET_Material_SetTexCoordScaleS::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Material_SetTexCoordScaleS----\n");
        paramSetter.SetTexcoordScale(ParamSetter::TexS, ((MESSAGEPACKET_Material_SetTexCoordScaleS*)body)->Index, ((MESSAGEPACKET_Material_SetTexCoordScaleS*)body)->ScaleS,0);
        break;
    case MESSAGEPACKET_Material_SetTexCoordScaleT::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Material_SetTexCoordScaleS----\n");
        paramSetter.SetTexcoordScale(ParamSetter::TexT, ((MESSAGEPACKET_Material_SetTexCoordScaleT*)body)->Index, 0, ((MESSAGEPACKET_Material_SetTexCoordScaleT*)body)->ScaleT);
        break;
    case MESSAGEPACKET_Material_SetTexCoordScaleST::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Material_SetTexCoordScaleST----\n");
        paramSetter.SetTexcoordScale(ParamSetter::TexS|ParamSetter::TexT, ((MESSAGEPACKET_Material_SetTexCoordScaleST*)body)->Index, ((MESSAGEPACKET_Material_SetTexCoordScaleST*)body)->ScaleS, ((MESSAGEPACKET_Material_SetTexCoordScaleST*)body)->ScaleT);
        break;
    case MESSAGEPACKET_Material_SetTexCoordRotate::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Material_SetTexCoordRotate----\n");
        paramSetter.SetTexcoordRotate(((MESSAGEPACKET_Material_SetTexCoordRotate*)body)->Index, ((MESSAGEPACKET_Material_SetTexCoordRotate*)body)->Rotate);
        break;
    case MESSAGEPACKET_Material_SetTexCoordTranslateS::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Material_SetTexCoordTranslateS----\n");
        paramSetter.SetTexcoordTranslate(ParamSetter::TexS, ((MESSAGEPACKET_Material_SetTexCoordTranslateS*)body)->Index, ((MESSAGEPACKET_Material_SetTexCoordTranslateS*)body)->TranslateS);
        break;
    case MESSAGEPACKET_Material_SetTexCoordTranslateT::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Material_SetTexCoordTranslateT----\n");
        paramSetter.SetTexcoordTranslate(ParamSetter::TexT, ((MESSAGEPACKET_Material_SetTexCoordTranslateT*)body)->Index, ((MESSAGEPACKET_Material_SetTexCoordTranslateT*)body)->TranslateT);
        break;
    case MESSAGEPACKET_Material_SetBumpTextureIndex::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Material_SetTexCoordTranslateT----\n");
        paramSetter.SetBumpTexIndex(((MESSAGEPACKET_Material_SetBumpTextureIndex*)body)->TextureIndex);
        break;
    case MESSAGEPACKET_Material_SetBumpMode::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Material_SetBumpMode----\n");
        paramSetter.SetBumpMode(((MESSAGEPACKET_Material_SetBumpMode*)body)->Mode);
        break;
    case MESSAGEPACKET_Material_SetIsBumpRenormalize::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Material_SetIsBumpRenormalize----\n");
        paramSetter.SetRenormalize(((MESSAGEPACKET_Material_SetIsBumpRenormalize*)body)->IsBumpRenormalize);
        break;
    case MESSAGEPACKET_Material_SetTextureCombinerList::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Material_SetTextureCombinerList----\n");
        for( int n = 0; n < ((MESSAGEPACKET_Material_SetTextureCombinerList*)body)->ListCount; n++ ){
            TextureCombiner *combiner = (TextureCombiner *)((unsigned char*)body+MESSAGEPACKET_Material_SetTextureCombinerList::StructSize+TextureCombiner::StructSize*n);
            paramSetter.SetTextureCombiner(n,combiner);
        }
        break;
    case MESSAGEPACKET_Material_SetIsPolygonOffsetEnabled::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Material_SetIsPolygonOffsetEnabled----\n");
        paramSetter.SetPolygonOffset(((MESSAGEPACKET_Material_SetIsPolygonOffsetEnabled*)body)->IsPolygonOffsetEnabled);
        break;
    case MESSAGEPACKET_Material_SetPolygonOffsetUnit::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Material_SetPolygonOffsetUnit----\n");
        paramSetter.SetPolygonOffsetUnit(((MESSAGEPACKET_Material_SetPolygonOffsetUnit*)body)->PolygonOffsetUnit);
        break;
    case MESSAGEPACKET_Material_SetFragmentLightingTable::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Material_SetFragmentLightingTable----\n");
        paramSetter.SetFragmentLightingTable((MESSAGEPACKET_Material_SetFragmentLightingTable*)body);
        break;
    case MESSAGEPACKET_Material_SetShaderProgramDescriptionIndex::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Material_SetShaderProgramDescriptionIndex----\n");
        paramSetter.SetShaderProgramDescriptionIndex( (MESSAGEPACKET_Material_SetShaderProgramDescriptionIndex*)body );
        break;
    //---------------------------------
    default:
        UnknownCommandError( header.type );
    }

    // callback call
    paramSetter.CallMaterialCallBack();
}

static void
ExecCategoryModel( const Header& header, void* body )
{
    vw::g3d::ParamSetter &paramSetter = vw::g3d::ParamSetter::GetInstance();
    unsigned int nID;

    switch ( header.type )
    {
    case MESSAGEPACKET_Model_SetModelVisibility::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Model_SetModelVisibility----\n");
        nID = ((MESSAGEPACKET_Model_SetModelVisibility*)body)->ModelId;
        paramSetter.SetModelVisibility(nID, ((MESSAGEPACKET_Model_SetModelVisibility*)body)->IsVisible);
        break;
    case MESSAGEPACKET_Model_SetModelTransformScaleX::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Model_SetModelTransformScaleX----\n");
        nID = ((MESSAGEPACKET_Model_SetModelTransformScaleX*)body)->ModelId;
        paramSetter.SetModelScale( ParamSetter::AXIS_X, nID, ((MESSAGEPACKET_Model_SetModelTransformScaleX*)body)->Value);
        break;
    case MESSAGEPACKET_Model_SetModelTransformScaleY::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Model_SetModelTransformScaleY----\n");
        nID = ((MESSAGEPACKET_Model_SetModelTransformScaleY*)body)->ModelId;
        paramSetter.SetModelScale( ParamSetter::AXIS_Y, nID, ((MESSAGEPACKET_Model_SetModelTransformScaleY*)body)->Value);
        break;
    case MESSAGEPACKET_Model_SetModelTransformScaleZ::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Model_SetModelTransformScaleZ----\n");
        nID = ((MESSAGEPACKET_Model_SetModelTransformScaleZ*)body)->ModelId;
        paramSetter.SetModelScale( ParamSetter::AXIS_Z, nID, ((MESSAGEPACKET_Model_SetModelTransformScaleZ*)body)->Value);
        break;
    case MESSAGEPACKET_Model_SetModelTransformRotateX::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Model_SetModelTransformRotateX----\n");
        nID = ((MESSAGEPACKET_Model_SetModelTransformRotateX*)body)->ModelId;
        paramSetter.SetModelRotate( ParamSetter::AXIS_X, nID, ((MESSAGEPACKET_Model_SetModelTransformRotateX*)body)->Value);
        break;
    case MESSAGEPACKET_Model_SetModelTransformRotateY::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Model_SetModelTransformRotateY----\n");
        nID = ((MESSAGEPACKET_Model_SetModelTransformRotateY*)body)->ModelId;
        paramSetter.SetModelRotate( ParamSetter::AXIS_Y, nID, ((MESSAGEPACKET_Model_SetModelTransformRotateY*)body)->Value);
        break;
    case MESSAGEPACKET_Model_SetModelTransformRotateZ::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Model_SetModelTransformRotateZ----\n");
        nID = ((MESSAGEPACKET_Model_SetModelTransformRotateZ*)body)->ModelId;
        paramSetter.SetModelRotate( ParamSetter::AXIS_Z, nID, ((MESSAGEPACKET_Model_SetModelTransformRotateZ*)body)->Value);
        break;
    case MESSAGEPACKET_Model_SetModelTransformTranslateX::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Model_SetModelTransformTranslateX----\n");
        nID = ((MESSAGEPACKET_Model_SetModelTransformTranslateX*)body)->ModelId;
        paramSetter.SetModelTranslate( ParamSetter::AXIS_X, nID, ((MESSAGEPACKET_Model_SetModelTransformTranslateX*)body)->Value);
        break;
    case MESSAGEPACKET_Model_SetModelTransformTranslateY::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Model_SetModelTransformTranslateY----\n");
        nID = ((MESSAGEPACKET_Model_SetModelTransformTranslateY*)body)->ModelId;
        paramSetter.SetModelTranslate( ParamSetter::AXIS_Y, nID, ((MESSAGEPACKET_Model_SetModelTransformTranslateY*)body)->Value);
        break;
    case MESSAGEPACKET_Model_SetModelTransformTranslateZ::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Model_SetModelTransformTranslateZ----\n");
        nID = ((MESSAGEPACKET_Model_SetModelTransformTranslateZ*)body)->ModelId;
        paramSetter.SetModelTranslate( ParamSetter::AXIS_Z, nID, ((MESSAGEPACKET_Model_SetModelTransformTranslateZ*)body)->Value);
        break;

    case MESSAGEPACKET_Model_SetModelPreviewVisibility::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Model_SetModelPreviewVisibility----\n");
        nID = ((MESSAGEPACKET_Model_SetModelPreviewVisibility*)body)->ModelId;
        paramSetter.SetModelPreviewVisibility(nID, ((MESSAGEPACKET_Model_SetModelPreviewVisibility*)body)->IsVisible);
        break;
    case MESSAGEPACKET_Model_SetModelPreviewTransformScaleX::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Model_SetModelPreviewTransformScaleX----\n");
        nID = ((MESSAGEPACKET_Model_SetModelPreviewTransformScaleX*)body)->ModelId;
        paramSetter.SetModelPreviewScale( ParamSetter::AXIS_X, nID, ((MESSAGEPACKET_Model_SetModelPreviewTransformScaleX*)body)->Value);
        break;
    case MESSAGEPACKET_Model_SetModelPreviewTransformScaleY::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Model_SetModelPreviewTransformScaleY----\n");
        nID = ((MESSAGEPACKET_Model_SetModelPreviewTransformScaleY*)body)->ModelId;
        paramSetter.SetModelPreviewScale( ParamSetter::AXIS_Y, nID, ((MESSAGEPACKET_Model_SetModelPreviewTransformScaleY*)body)->Value);
        break;
    case MESSAGEPACKET_Model_SetModelPreviewTransformScaleZ::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Model_SetModelPreviewTransformScaleZ----\n");
        nID = ((MESSAGEPACKET_Model_SetModelPreviewTransformScaleZ*)body)->ModelId;
        paramSetter.SetModelPreviewScale( ParamSetter::AXIS_Z, nID, ((MESSAGEPACKET_Model_SetModelPreviewTransformScaleZ*)body)->Value);
        break;
    case MESSAGEPACKET_Model_SetModelPreviewTransformRotateX::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Model_SetModelPreviewTransformRotateX----\n");
        nID = ((MESSAGEPACKET_Model_SetModelPreviewTransformRotateX*)body)->ModelId;
        paramSetter.SetModelPreviewRotate( ParamSetter::AXIS_X, nID, ((MESSAGEPACKET_Model_SetModelPreviewTransformRotateX*)body)->Value);
        break;
    case MESSAGEPACKET_Model_SetModelPreviewTransformRotateY::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Model_SetModelPreviewTransformRotateY----\n");
        nID = ((MESSAGEPACKET_Model_SetModelPreviewTransformRotateY*)body)->ModelId;
        paramSetter.SetModelPreviewRotate( ParamSetter::AXIS_Y, nID, ((MESSAGEPACKET_Model_SetModelPreviewTransformRotateY*)body)->Value);
        break;
    case MESSAGEPACKET_Model_SetModelPreviewTransformRotateZ::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Model_SetModelPreviewTransformRotateZ----\n");
        nID = ((MESSAGEPACKET_Model_SetModelPreviewTransformRotateZ*)body)->ModelId;
        paramSetter.SetModelPreviewRotate( ParamSetter::AXIS_Z, nID, ((MESSAGEPACKET_Model_SetModelPreviewTransformRotateZ*)body)->Value);
        break;
    case MESSAGEPACKET_Model_SetModelPreviewTransformTranslateX::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Model_SetModelPreviewTransformTranslateX----\n");
        nID = ((MESSAGEPACKET_Model_SetModelPreviewTransformTranslateX*)body)->ModelId;
        paramSetter.SetModelPreviewTranslate( ParamSetter::AXIS_X, nID, ((MESSAGEPACKET_Model_SetModelPreviewTransformTranslateX*)body)->Value);
        break;
    case MESSAGEPACKET_Model_SetModelPreviewTransformTranslateY::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Model_SetModelPreviewTransformTranslateY----\n");
        nID = ((MESSAGEPACKET_Model_SetModelPreviewTransformTranslateY*)body)->ModelId;
        paramSetter.SetModelPreviewTranslate( ParamSetter::AXIS_Y, nID, ((MESSAGEPACKET_Model_SetModelPreviewTransformTranslateY*)body)->Value);
        break;
    case MESSAGEPACKET_Model_SetModelPreviewTransformTranslateZ::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Model_SetModelPreviewTransformTranslateZ----\n");
        nID = ((MESSAGEPACKET_Model_SetModelPreviewTransformTranslateZ*)body)->ModelId;
        paramSetter.SetModelPreviewTranslate( ParamSetter::AXIS_Z, nID, ((MESSAGEPACKET_Model_SetModelPreviewTransformTranslateZ*)body)->Value);
        break;

    case MESSAGEPACKET_Model_SetModelTransformValue::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Model_SetModelTransformValue----\n");
        nID = ((MESSAGEPACKET_Model_SetModelTransformValue*)body)->ModelId;
        paramSetter.SetModelTransform((MESSAGEPACKET_Model_SetModelTransformValue*)body);
        break;
    case MESSAGEPACKET_Model_SetModelPreviewTransformValue::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Model_SetModelPreviewTransformValue----\n");
        nID = ((MESSAGEPACKET_Model_SetModelPreviewTransformValue*)body)->ModelId;
        paramSetter.SetModelPreviewTransform((MESSAGEPACKET_Model_SetModelPreviewTransformValue*)body);
        break;
    case MESSAGEPACKET_Model_SetModelPreviewBoneLink::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Model_SetModelPreviewBoneLink----\n");
        nID = ((MESSAGEPACKET_Model_SetModelPreviewBoneLink*)body)->ModelId;
        paramSetter.SetModelPreviewBoneLink((MESSAGEPACKET_Model_SetModelPreviewBoneLink*)body);
        break;
    case MESSAGEPACKET_Model_SetModelPreviewBoneUnLink::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Model_SetModelPreviewBoneUnLink----\n");
        nID = ((MESSAGEPACKET_Model_SetModelPreviewBoneUnLink*)body)->ModelId;
        paramSetter.SetModelPreviewBoneUnLink((MESSAGEPACKET_Model_SetModelPreviewBoneUnLink*)body);
        break;
    }

    // callback call
    paramSetter.CallModelCallBack( nID );
}

static void
ExecCategoryMesh( const Header& header, void* body )
{
    vw::g3d::ParamSetter &paramSetter = vw::g3d::ParamSetter::GetInstance();
    unsigned int nID;

    switch ( header.type )
    {
    case MESSAGEPACKET_Mesh_SetMeshVisibility::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Mesh_SetMeshVisibility----\n");
        nID = ((MESSAGEPACKET_Mesh_SetMeshVisibility*)body)->ModelId;
        paramSetter.SetMeshVisibility(nID,((MESSAGEPACKET_Mesh_SetMeshVisibility*)body)->MeshIndex,((MESSAGEPACKET_Mesh_SetMeshVisibility*)body)->IsVisible);
        break;
    case MESSAGEPACKET_Mesh_SetMeshNodeVisibility::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Mesh_SetMeshNodeVisibility----\n");
        nID = ((MESSAGEPACKET_Mesh_SetMeshNodeVisibility*)body)->ModelId;
        paramSetter.SetMeshNodeVisibility(nID,((MESSAGEPACKET_Mesh_SetMeshNodeVisibility*)body)->MeshIndex,((MESSAGEPACKET_Mesh_SetMeshNodeVisibility*)body)->IsVisible);
        break;
    case MESSAGEPACKET_Mesh_SetMeshRenderPriority::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Mesh_SetMeshRenderPriority----\n");
        nID = ((MESSAGEPACKET_Mesh_SetMeshRenderPriority*)body)->ModelId;
        paramSetter.SetMeshRenderPriority(nID,((MESSAGEPACKET_Mesh_SetMeshRenderPriority*)body)->MeshIndex,((MESSAGEPACKET_Mesh_SetMeshRenderPriority*)body)->RenderPriority);
        break;
    }

    // callback call
    paramSetter.CallModelCallBack( nID );
}

static void
ExecCategoryBone( const Header& header, void* body )
{
    vw::g3d::ParamSetter &paramSetter = vw::g3d::ParamSetter::GetInstance();
    switch ( header.type )
    {
    case MESSAGEPACKET_Bone_SetBoneBillboard::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Bone_SetBoneBillboard----\n");
        paramSetter.SetBoneBillboard(((MESSAGEPACKET_Bone_SetBoneBillboard*)body)->ModelId,((MESSAGEPACKET_Bone_SetBoneBillboard*)body)->BoneName,((MESSAGEPACKET_Bone_SetBoneBillboard*)body)->BillboardMode);
        break;
    }
}

static void
ExecCategoryLUT( const Header& header, void* body )
{
    vw::g3d::ParamSetter &paramSetter = vw::g3d::ParamSetter::GetInstance();
    switch ( header.type )
    {
    case MESSAGEPACKET_LUT_ModifyLookUpTable::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_LUT_ModifyLookUpTable----\n");
        paramSetter.SetLUT(((MESSAGEPACKET_LUT_ModifyLookUpTable*)body));
        break;
    }
}

static void
ExecCategoryObject( const Header& header, void* body )
{
    vw::g3d::ParamSetter &paramSetter = vw::g3d::ParamSetter::GetInstance();
    nw::ut::MoveArray<EditSelectedObject> selectedObject( Memory_GetAllocator() );
    switch( header.type )
    {
    //---------------------------------
    //  Select command
    case MESSAGEPACKET_Object_Select::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Object_Select----\n");
        NW_VIEWER_LOG("%d objects seleced. (Id, Index) = ", header.arraySize/EditSelectedObject::StructSize);
        for (int i=0; i<header.arraySize/EditSelectedObject::StructSize; ++i)
        {
            selectedObject.push_back(*(EditSelectedObject*)((unsigned char*)body+MESSAGEPACKET_Object_Select::StructSize+EditSelectedObject::StructSize*i));
            NW_VIEWER_LOG("(%u, %u), ", selectedObject.back().Id, selectedObject.back().Index);
        }
        NW_VIEWER_LOG("\n");
        paramSetter.SelectObject(((MESSAGEPACKET_Object_Select*)body)->ObjectType, selectedObject);
        break;
    case MESSAGEPACKET_Object_Forcus::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Object_Forcus----\n");
        paramSetter.Focus();
        break;
    default:
        UnknownCommandError( header.type );
    }
}

static void
ExecCategoryOperation( const Header& header, void* body)
{
    vw::g3d::ParamSetter &paramSetter = vw::g3d::ParamSetter::GetInstance();
    char szName[260];
    switch( header.type )
    {
    case MESSAGEPACKET_Operation_EnableDrawing::MessagePacketType:
        paramSetter.EnableDrawing();
        break;
    case MESSAGEPACKET_Operation_DisableDrawing::MessagePacketType:
        paramSetter.DisableDrawing();
        break;
    case MESSAGEPACKET_Operation_CaptureScreen::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Operation_CaptureScreen----\n");
        vw::system::ConvertStrWtoUTF8(
            ((MESSAGEPACKET_Operation_CaptureScreen*)body)->CapturePath.ToString(),
            szName,
            260
        );
        vw::system::GetCapture().Capture(szName);
        break;
    }
}

static void
ExecCategoryFog( const Header& header, void* body)
{
    vw::g3d::ParamSetter &paramSetter = vw::g3d::ParamSetter::GetInstance();
    switch( header.type )
    {
    case MESSAGEPACKET_Fog_SetFog::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Fog_SetFog----\n");
        paramSetter.SetFog((MESSAGEPACKET_Fog_SetFog*)body);
        break;
    case MESSAGEPACKET_Fog_SetFogIsZFlip::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Fog_SetFogIsZFlip----\n");
        paramSetter.SetFogIsZFlip((MESSAGEPACKET_Fog_SetFogIsZFlip*)body);
        break;
    case MESSAGEPACKET_Fog_SetFogColor::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Fog_SetFogColor----\n");
        paramSetter.SetFogColor((MESSAGEPACKET_Fog_SetFogColor*)body);
        break;
    case MESSAGEPACKET_Fog_SetMinFogDepth::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Fog_SetMinFogDepth----\n");
        paramSetter.SetMinFogDepth((MESSAGEPACKET_Fog_SetMinFogDepth*)body);
        break;
    case MESSAGEPACKET_Fog_SetMaxFogDepth::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Fog_SetMaxFogDepth----\n");
        paramSetter.SetMaxFogDepth((MESSAGEPACKET_Fog_SetMaxFogDepth*)body);
        break;
    case MESSAGEPACKET_Fog_SetFogDensity::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Fog_SetFogDensity----\n");
        paramSetter.SetFogDensity((MESSAGEPACKET_Fog_SetFogDensity*)body);
        break;
    }

    // callback call
    paramSetter.CallFogCallBack();
}

static void
ExecCategoryLight( const Header& header, void* body )
{
    vw::g3d::ParamSetter &paramSetter = vw::g3d::ParamSetter::GetInstance();
    const u32 passive = ModuleG3dImpl::GetInstance()->Get3DEditorPassiveChannel();

    switch( header.type )
    {
    case MESSAGEPACKET_Light_SetLightIsLightEnabled::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Light_SetLightIsLightEnabled----\n");
        paramSetter.SetLightEnable((MESSAGEPACKET_Light_SetLightIsLightEnabled*)body);
        break;
    case MESSAGEPACKET_Light_SetLightTranslate::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Light_SetLightTranslate----\n");
        paramSetter.SetLightTranslate((MESSAGEPACKET_Light_SetLightTranslate*)body);
        break;
    case MESSAGEPACKET_Light_SetAmbientLightColor::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Light_SetAmbientLightColor----\n");
        paramSetter.SetAmbientLightColor((MESSAGEPACKET_Light_SetAmbientLightColor*)body);
        break;
    case MESSAGEPACKET_Light_SetHemiSphereGroundColor::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Light_SetHemiSphereGroundColor----\n");
        paramSetter.SetHemiSphereGroundColor((MESSAGEPACKET_Light_SetHemiSphereGroundColor*)body);
        break;
    case MESSAGEPACKET_Light_SetHemiSphereSkyColor::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Light_SetHemiSphereSkyColor----\n");
        paramSetter.SetHemiSphereSkyColor((MESSAGEPACKET_Light_SetHemiSphereSkyColor*)body);
        break;
    case MESSAGEPACKET_Light_SetHemiSphereDirection::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Light_SetHemiSphereDirection----\n");
        paramSetter.SetHemiSphereDirection((MESSAGEPACKET_Light_SetHemiSphereDirection*)body);
        break;
    case MESSAGEPACKET_Light_SetHemiSphereLerpFactor::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Light_SetHemiSphereLerpFactor----\n");
        paramSetter.SetHemiSphereLerpFactor((MESSAGEPACKET_Light_SetHemiSphereLerpFactor*)body);
        break;
    case MESSAGEPACKET_Light_SetHemiSphereIsInheritingDirectionRotate::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Light_SetHemiSphereIsInheritingDirectionRotate----\n");
        paramSetter.SetHemiSphereIsInheritingDirectionRorate((MESSAGEPACKET_Light_SetHemiSphereIsInheritingDirectionRotate*)body);
        break;
    case MESSAGEPACKET_Light_SetVLLightKind::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Light_SetVLLightKind----\n");
        paramSetter.SetVLLightKind((MESSAGEPACKET_Light_SetVLLightKind*)body);
        break;
    case MESSAGEPACKET_Light_SetVLAmbient::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Light_SetVLAmbient----\n");
        paramSetter.SetVLAmbient((MESSAGEPACKET_Light_SetVLAmbient*)body);
        break;
    case MESSAGEPACKET_Light_SetVLDiffuse::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Light_SetVLDiffuse----\n");
        paramSetter.SetVLDiffuse((MESSAGEPACKET_Light_SetVLDiffuse*)body);
        break;
    case MESSAGEPACKET_Light_SetVLDirection::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Light_SetVLDirection----\n");
        paramSetter.SetVLDirection((MESSAGEPACKET_Light_SetVLDirection*)body);
        break;
    case MESSAGEPACKET_Light_SetVLIsInheritingDirectionRotate::MessagePacketType:
        NW_VIEWER_LOG("----Revc MESSAGEPACKET_Light_SetVLIsInheritingDirectionRotate----\n");
        paramSetter.SetVLInheritingDirectionRotate((MESSAGEPACKET_Light_SetVLIsInheritingDirectionRotate*)body);
        break;
    case MESSAGEPACKET_Light_SetVLDistanceAttenuationConstant::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Light_SetVLDistanceAttenuationConstant----\n");
        paramSetter.SetVLDistanceAttenuationConstant((MESSAGEPACKET_Light_SetVLDistanceAttenuationConstant*)body);
        break;
    case MESSAGEPACKET_Light_SetVLDistanceAttenuationLinear::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Light_SetVLDistanceAttenuationLinear----\n");
        paramSetter.SetVLDistanceAttenuationLinear((MESSAGEPACKET_Light_SetVLDistanceAttenuationLinear*)body);
        break;
    case MESSAGEPACKET_Light_SetVLDistanceAttenuationQuadratic::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Light_SetVLDistanceAttenuationQuadratic----\n");
        paramSetter.SetVLDistanceAttenuationQuadratic((MESSAGEPACKET_Light_SetVLDistanceAttenuationQuadratic*)body);
        break;
    case MESSAGEPACKET_Light_SetVLIsDistanceAttenuationEnabled::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Light_SetVLIsDistanceAttenuationEnabled----\n");
        paramSetter.SetVLDistanceAttenuationEnable((MESSAGEPACKET_Light_SetVLIsDistanceAttenuationEnabled*)body);
        break;
    case MESSAGEPACKET_Light_SetVLSpotExponent::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Light_SetVLSpotExponent----\n");
        paramSetter.SetVLSpotExponet((MESSAGEPACKET_Light_SetVLSpotExponent*)body);
        break;
    case MESSAGEPACKET_Light_SetVLSpotCutoffAngle::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Light_SetVLSpotCutoffAngle----\n");
        paramSetter.SetVLSpotCutoffAngle((MESSAGEPACKET_Light_SetVLSpotCutoffAngle*)body);
        break;
    case MESSAGEPACKET_Light_SetFLKind::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Light_SetFLKind----\n");
        paramSetter.SetFLKind((MESSAGEPACKET_Light_SetFLKind*)body);
        break;
    case MESSAGEPACKET_Light_SetFLAmbient::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Light_SetFLAmbient----\n");
        paramSetter.SetFLAmbient((MESSAGEPACKET_Light_SetFLAmbient*)body);
        break;
    case MESSAGEPACKET_Light_SetFLDiffuse::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Light_SetFLDiffuse----\n");
        paramSetter.SetFLDiffuse((MESSAGEPACKET_Light_SetFLDiffuse*)body);
        break;
    case MESSAGEPACKET_Light_SetFLIsTwoSideDiffuse::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Light_SetFLIsTwoSideDiffuse----\n");
        paramSetter.SetFLIsTowSideDiffuse((MESSAGEPACKET_Light_SetFLIsTwoSideDiffuse*)body);
        break;
    case MESSAGEPACKET_Light_SetFLSpecular0::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Light_SetFLSpecular0----\n");
        paramSetter.SetFLSpecular(0,((MESSAGEPACKET_Light_SetFLSpecular0*)body)->Specular0);
        break;
    case MESSAGEPACKET_Light_SetFLSpecular1::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Light_SetFLSpecular1----\n");
        paramSetter.SetFLSpecular(1,((MESSAGEPACKET_Light_SetFLSpecular1*)body)->Specular1);
        break;
    case MESSAGEPACKET_Light_SetFLIsInheritingDirectionRotate::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Light_SetFLIsInheritingDirectionRotate----\n");
        paramSetter.SetFLIsInheritingDirectionRotate((MESSAGEPACKET_Light_SetFLIsInheritingDirectionRotate*)body);
        break;
    case MESSAGEPACKET_Light_SetFLDirection::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Light_SetFLDirection----\n");
        paramSetter.SetFLDirection((MESSAGEPACKET_Light_SetFLDirection*)body);
        break;
    case MESSAGEPACKET_Light_SetFLSamplerLUTS::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Light_SetFLSamplerLUTS----\n");
        paramSetter.SetFLSamplerLUTS((MESSAGEPACKET_Light_SetFLSamplerLUTS*)body);
        break;
    case MESSAGEPACKET_Light_SetFLIsDistanceAttenuationEnabled::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Light_SetFLIsDistanceAttenuationEnabled----\n");
        paramSetter.SetFLDistanceAttenuationEnable((MESSAGEPACKET_Light_SetFLIsDistanceAttenuationEnabled*)body);
        break;
    case MESSAGEPACKET_Light_SetFLIsDistanceAttenuationStart::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Light_SetFLIsDistanceAttenuationStart----\n");
        paramSetter.SetFLDistanceAttenuationStart((MESSAGEPACKET_Light_SetFLIsDistanceAttenuationStart*)body);
        break;
    case MESSAGEPACKET_Light_SetFLIsDistanceAttenuationEnd::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Light_SetFLIsDistanceAttenuationEnd----\n");
        paramSetter.SetFLDistanceAttenuationEnd((MESSAGEPACKET_Light_SetFLIsDistanceAttenuationEnd*)body);
        break;
    case MESSAGEPACKET_Light_SetFLDistanceSamplerLUT::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Light_SetFLDistanceSamplerLUT----\n");
        paramSetter.SetFLDistanceSamplerLUT((MESSAGEPACKET_Light_SetFLDistanceSamplerLUT*)body);
        break;
    case MESSAGEPACKET_Light_SetFLAngleSamplerLUT::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Light_SetFLAngleSamplerLUT----\n");
        paramSetter.SetFLAngleSamplerLUT((MESSAGEPACKET_Light_SetFLAngleSamplerLUT*)body);
        break;
    case MESSAGEPACKET_Light_SetFLAngleSamplerInput::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Light_SetFLAngleSamplerInput----\n");
        paramSetter.SetFLAngleSamplerInput((MESSAGEPACKET_Light_SetFLAngleSamplerInput*)body);
        break;
    case MESSAGEPACKET_Light_SetFLAngleSamplerScale::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Light_SetFLAngleSamplerScale----\n");
        paramSetter.SetFLAngleSamplerScale((MESSAGEPACKET_Light_SetFLAngleSamplerScale*)body);
        break;
    case MESSAGEPACKET_Light_GetFragmentLightInfo::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Light_GetFragmentLightInfo----\n");
        paramSetter.GetFragmentLightInfo((MESSAGEPACKET_Light_GetFragmentLightInfo*)body);
        vw::system::Stream_Write( passive, &header, header.StructSize );
        vw::system::Stream_Write( passive, body, MESSAGEPACKET_Light_GetFragmentLightInfo::StructSize );
        break;
    }

    // callback call
    paramSetter.CallLightCallBack();
}

static void
ExecCategoryCamera( const Header& header, void* body )
{
    vw::g3d::ParamSetter &paramSetter = vw::g3d::ParamSetter::GetInstance();
    const u32 passive = ModuleG3dImpl::GetInstance()->Get3DEditorPassiveChannel();

    switch( header.type )
    {
    case MESSAGEPACKET_Camera_SetTargetViewAim::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Camera_SetTargetViewAim----\n");
        paramSetter.SetTargetViewAim((MESSAGEPACKET_Camera_SetTargetViewAim*)body);
        break;
    case MESSAGEPACKET_Camera_SetTargetViewLookAt::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Camera_SetTargetViewLookAt----\n");
        paramSetter.SetTargetViewLookAt((MESSAGEPACKET_Camera_SetTargetViewLookAt*)body);
        break;
    case MESSAGEPACKET_Camera_SetTargetViewRotate::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Camera_SetTargetViewRotate----\n");
        paramSetter.SetTargetViewRotate((MESSAGEPACKET_Camera_SetTargetViewRotate*)body);
        break;
    case MESSAGEPACKET_Camera_SetOrthoProjection::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Camera_SetOrthoProjection---\n");
        paramSetter.SetOrthoProjection((MESSAGEPACKET_Camera_SetOrthoProjection*)body);
        break;
    case MESSAGEPACKET_Camera_SetPerspectiveProjection::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Camera_SetPerspectiveProjection----\n");
        paramSetter.SetPerspectiveProjection((MESSAGEPACKET_Camera_SetPerspectiveProjection*)body);
        break;
    case MESSAGEPACKET_Camera_SetCameraTransform::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Camera_SetCameraTransform----\n");
        paramSetter.SetCameraTransform((MESSAGEPACKET_Camera_SetCameraTransform*)body);
        break;
    case MESSAGEPACKET_Camera_SetCameraIsWBuffer::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Camera_SetCameraIsWBuffer----\n");
        paramSetter.SetCameraIsWBuffer((MESSAGEPACKET_Camera_SetCameraIsWBuffer*)body);
        break;
    case MESSAGEPACKET_Camera_GetCameraInfo::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Camera_GetCameraInfo----\n");
        paramSetter.GetCameraInfo((MESSAGEPACKET_Camera_GetCameraInfo*)body);
        vw::system::Stream_Write( passive, &header, header.StructSize );
        vw::system::Stream_Write( passive, body, MESSAGEPACKET_Camera_GetCameraInfo::StructSize );
        break;
    }

    // callback call
    paramSetter.CallCameraCallBack();
}

static void
ExecCategoryScene( const Header& header, void* body)
{
    vw::g3d::ParamSetter &paramSetter = vw::g3d::ParamSetter::GetInstance();
    switch( header.type )
    {
    case MESSAGEPACKET_Scene_SetSceneActive::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Scene_SetSceneActive----\n");
        paramSetter.SetSceneActive((MESSAGEPACKET_Scene_SetSceneActive*)body);
        break;
    }
}

static void
ExecCategoryAnimation( const Header& header, void* body )
{
    vw::g3d::ParamSetter &paramSetter = vw::g3d::ParamSetter::GetInstance();
    switch( header.type)
    {
    case MESSAGEPACKET_Animation_AddAnimationSet::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Animation_AddAnimationSet----\n");
        paramSetter.AddAnimationSet(header, (MESSAGEPACKET_Animation_AddAnimationSet*)body);
        break;
    case MESSAGEPACKET_Animation_AddStrayAnimations::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Animation_AddStrayAnimations----\n");
        paramSetter.AddStrayAnimations(header, (MESSAGEPACKET_Animation_AddStrayAnimations*)body);
        break;
    case MESSAGEPACKET_Animation_ClearAnimationSet::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Animation_ClearAnimationSet----\n");
        paramSetter.ClearAnimationSet((MESSAGEPACKET_Animation_ClearAnimationSet*)body);
    case MESSAGEPACKET_Animation_ActivateAnimationSet::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Animation_ActivateAnimationSet----\n");
        paramSetter.ActivateAnimationSet((MESSAGEPACKET_Animation_ActivateAnimationSet*)body);
        break;
    case MESSAGEPACKET_Animation_DeActivateAnimationSet::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Animation_DeActivateAnimationSet----\n");
        paramSetter.DeactivateAnimationSet((MESSAGEPACKET_Animation_DeActivateAnimationSet*)body);
        break;
    case MESSAGEPACKET_Animation_AddLightAnimation::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Animation_AddLightAnimation----\n");
        paramSetter.AddLightAnimation((MESSAGEPACKET_Animation_AddLightAnimation*)body);
        break;
    case MESSAGEPACKET_Animation_AddCameraAnimation::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Animation_AddCameraAnimation----\n");
        paramSetter.AddCameraAnimation((MESSAGEPACKET_Animation_AddCameraAnimation*)body);
        break;
    case MESSAGEPACKET_Animation_AddFogAnimation::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Animation_AddFogAnimation----\n");
        paramSetter.AddFogAnimation((MESSAGEPACKET_Animation_AddFogAnimation*)body);
        break;
    case MESSAGEPACKET_Animation_RenameAnimationSet::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Animation_RenameAnimationSet----\n");
        paramSetter.RenameAnimationSet((MESSAGEPACKET_Animation_RenameAnimationSet*)body);
        break;
    case MESSAGEPACKET_Animation_BindSceneAnimation::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Animation_BindSceneAnimation----\n");
        paramSetter.BindSceneAnimation((MESSAGEPACKET_Animation_BindSceneAnimation*)body);
        break;
    case MESSAGEPACKET_Animation_StoreCurrentFrame::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Animation_StoreCurrentFrame----\n");
        paramSetter.StoreCurrentFrame((MESSAGEPACKET_Animation_StoreCurrentFrame*)body);
        break;
    case MESSAGEPACKET_Animation_RestoreCurrentFrame::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_Animation_RestoreCurrentFrame----\n");
        paramSetter.RestoreCurrentFrame((MESSAGEPACKET_Animation_RestoreCurrentFrame*)body);
        break;
    }
}

static void
ExecCategoryUserData( const Header& header, void* body )
{
    vw::g3d::ParamSetter &paramSetter = vw::g3d::ParamSetter::GetInstance();
    switch( header.type )
    {
    case MESSAGEPACKET_UserData_SendUserData::MessagePacketType:
        NW_VIEWER_LOG("----Recv MESSAGEPACKET_UserData_SendUserData----\n");
        paramSetter.SetUserData((MESSAGEPACKET_UserData_SendUserData*)body);
        break;
    }
}

/*---------------------------------------------------------------------------*
  Name:         UnknownCommandError
  Description:  Sends undefined command error
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
static void
UnknownCommandError( u16 type )
{
    // NW4R_WARNING( false, RESTXT_ON_UNKNOWN_COMMAND_x, type );

    //if ( type & HEADER_REPLY_FLAG )
    // Returns error message to those with enabled reply flags
    //{
    //    ErrSystemUnknownCommand reply;
    //    reply.command = type;
    //
    //    Stream3d_Reply( ERR_SYSTEM_UNKNOWN_COMMAND, reply );
    //}
}

}   /* namespace g3d */
}   /* namespace vw  */
