/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#include "../include/ResourceHelper.h"
#include "../include/ResourceManager.h"

namespace vw {
namespace g3d {

nw::ut::ResTypeInfo
GetActualTypeInfo(nw::gfx::res::ResTexture texture)
{
    if (nw::gfx::res::ResReferenceTexture_TYPE_INFO == texture.ref().typeInfo)
        return GetActualTypeInfo(texture);

    return texture.ref().typeInfo;
}

nw::gfx::res::ResTexture
GetDereferenceTexture(nw::gfx::res::ResTexture texture)
{
    if (nw::gfx::res::ResReferenceTexture_TYPE_INFO == texture.ref().typeInfo)
        return GetDereferenceTexture(
            nw::ut::ResDynamicCast<nw::gfx::res::ResReferenceTexture>( texture ).GetTargetTexture() );
    return texture;
}

nw::gfx::Result
SetupResTexture(
    nw::gfx::res::ResTexture resTex,
    nw::os::IAllocator* allocator,
    nw::gfx::res::ResGraphicsFile graphicsFile)
{
    nw::gfx::Result result = resTex.Setup( allocator, graphicsFile );

    resTex = GetDereferenceTexture( resTex );

    if (nw::gfx::res::ResCubeTexture::TYPE_INFO == resTex.ref().typeInfo)
    {
        nw::gfx::res::ResCubeTexture resCubeTex
            = nw::gfx::ResDynamicCast<nw::gfx::res::ResCubeTexture>( resTex );

        u32 loadFlag = resCubeTex.GetLocationFlag();
        GLenum transtype = loadFlag & 0xFFFF0000;

        if ( (NN_GX_MEM_VRAMA | GL_NO_COPY_FCRAM_DMP) == transtype ||
             (NN_GX_MEM_VRAMB | GL_NO_COPY_FCRAM_DMP) == transtype )
        {
            // Gets the texture address for each face.
            u32 texAddresses[nw::gfx::res::ResCubeTexture::MAX_CUBE_FACE];
            for ( int face = 0; face < nw::gfx::res::ResCubeTexture::MAX_CUBE_FACE; ++face )
            {
                nw::gfx::res::ResPixelBasedImage resImage
                    = resCubeTex.GetImage( static_cast<nw::gfx::res::ResCubeTexture::CubeFace>( face ) );

                texAddresses[ face ] = resImage.GetLocationAddress();
            }

            // Finds the minimum address.
            int minFace = 0;
            u32 minAddress = texAddresses[ 0 ];
            for (int face = 1; face < nw::gfx::res::ResCubeTexture::MAX_CUBE_FACE; ++face )
            {
                if (minAddress > texAddresses[ face ])
                {
                    minAddress = texAddresses[ face ];
                    minFace = face;
                }
            }

            // Flips if the minimum address is not POS_X.
            if ( minFace != nw::gfx::res::ResCubeTexture::CUBE_FACE_POSITIVE_X )
            {
                u32 texID = reinterpret_cast<u32>(resTex.ptr());

                // Transfers POS_X to minAdress.
                {
                    nw::gfx::res::ResPixelBasedImage resImage
                        = resCubeTex.GetImage( nw::gfx::res::ResCubeTexture::CUBE_FACE_POSITIVE_X );

                    u32 size = resImage.GetImageDataCount();
                    if (size == 0)
                    {
                        size = resImage.GetWidth() * resImage.GetHeight() * resImage.GetBitsPerPixel() / 8;
                    }

                    if ( resImage.GetImageData() )
                    {
                        nngxAddVramDmaCommand( resImage.GetImageData(), (void*)minAddress, size );
                    }

                    resImage.SetLocationAddress( minAddress );
                }

                // Transfers minFace to the POS_X address.
                {
                    nw::gfx::res::ResPixelBasedImage resImage
                        = resCubeTex.GetImage( static_cast<nw::gfx::res::ResCubeTexture::CubeFace>( minFace ) );

                    u32 size = resImage.GetImageDataCount();
                    if (size == 0)
                    {
                        size = resImage.GetWidth() * resImage.GetHeight() * resImage.GetBitsPerPixel() / 8;
                    }

                    u32 address = texAddresses[nw::gfx::res::ResCubeTexture::CUBE_FACE_POSITIVE_X];

                    if ( resImage.GetImageData() )
                    {
                        nngxAddVramDmaCommand( resImage.GetImageData(), (void*)address, size );
                    }

                    resImage.SetLocationAddress( address );
                }
            }
        }
    }

    return result;
}

nw::gfx::Result
SetupResGraphicsFile(nw::os::IAllocator* allocator, nw::gfx::res::ResGraphicsFile graphicsFile)
{
    nw::gfx::Result result = nw::gfx::RESOURCE_RESULT_OK;
    NW_ASSERT( graphicsFile.IsValid() );
    NW_ASSERT( nw::gfx::internal::ResCheckRevision( graphicsFile ) );

    s32 lutSetsNum = graphicsFile.GetLutSetsCount();
    for ( int i = 0; i < lutSetsNum; ++i )
    {
        result |= graphicsFile.GetLutSets( i ).Setup(allocator, graphicsFile);
    }

    s32 lightNum = graphicsFile.GetLightsCount();
    for ( int i = 0; i < lightNum; ++i )
    {
        result |= graphicsFile.GetLights( i ).Setup(allocator, graphicsFile);
    }

    s32 fogNum = graphicsFile.GetFogsCount();
    for ( int i = 0; i < fogNum; ++i )
    {
        result |= graphicsFile.GetFogs( i ).Setup(allocator, graphicsFile);
    }

    s32 shaderNum = graphicsFile.GetShadersCount();
    for ( int i = 0; i < shaderNum; ++i )
    {
        result |= graphicsFile.GetShaders( i ).Setup(allocator, graphicsFile);
    }

    s32 textureNum = graphicsFile.GetTexturesCount();
    for ( int i = 0; i < textureNum; ++i )
    {
        //result |= graphicsFile.GetTextures( i ).Setup(allocator, graphicsFile);
        result |= SetupResTexture( graphicsFile.GetTextures( i ), allocator, graphicsFile);
    }

    s32 skeletalAnimNum = graphicsFile.GetSkeletalAnimsCount();
    for ( int i = 0; i < skeletalAnimNum; ++i )
    {
        result |= graphicsFile.GetSkeletalAnims( i ).Setup(allocator, graphicsFile);
    }

    s32 materialAnimNum = graphicsFile.GetMaterialAnimsCount();
    for ( int i = 0; i < materialAnimNum; ++i )
    {
        result |= graphicsFile.GetMaterialAnims( i ).Setup(allocator, graphicsFile);

        // Checks broken references.
        nw::anim::res::ResAnim materialAnim = graphicsFile.GetMaterialAnims( i );
        for(int j=0; j<materialAnim.GetMemberAnimSetCount(); j++)
        {
            nw::anim::res::ResMemberAnim resMemberAnim = materialAnim.GetMemberAnimSet(j);
            if(resMemberAnim.GetPrimitiveType() == nw::anim::res::ResMemberAnim::PRIMITIVETYPE_TEXTURE)
            {
                nw::anim::res::ResTextureAnim& resTextureAnim = static_cast<nw::anim::res::ResTextureAnim&>(resMemberAnim);
                for (int k = 0; k < resTextureAnim.GetTexturesCount(); ++k)
                {
                    nw::gfx::ResReferenceTexture resReferenceTexture = resTextureAnim.GetTextures(k);
                    if (!resReferenceTexture.GetTargetTexture().IsValid())
                    {
                        // Configures the default texture.
                        void* pDefaultTexture = ResourceManager::GetInstance()->GetDefaultTexture().ptr();
                        resReferenceTexture.ref().toTargetTexture.set_ptr(pDefaultTexture);
                    }
                }
            }
        }
    }

    s32 visibilityAnimNum = graphicsFile.GetVisibilityAnimsCount();
    for ( int i = 0; i < visibilityAnimNum; ++i )
    {
        result |= graphicsFile.GetVisibilityAnims( i ).Setup(allocator, graphicsFile);
    }

    s32 cameraAnimNum = graphicsFile.GetCameraAnimsCount();
    for ( int i = 0; i < cameraAnimNum; ++i )
    {
        result |= graphicsFile.GetCameraAnims( i ).Setup(allocator, graphicsFile);
    }

    s32 lightAnimNum = graphicsFile.GetLightAnimsCount();
    for ( int i = 0; i < lightAnimNum; ++i )
    {
        result |= graphicsFile.GetLightAnims( i ).Setup(allocator, graphicsFile);
    }

    s32 modelNum = graphicsFile.GetModelsCount();
    for ( int i = 0; i < modelNum; ++i )
    {
        result |= graphicsFile.GetModels( i ).Setup(allocator, graphicsFile);

        // Checks broken references.
        s32 cnt = graphicsFile.GetModels( i ).GetMaterialsCount();
        for( int j = 0; j < cnt; ++j )
        {
            nw::gfx::ResShader res_shader = graphicsFile.GetModels( i ).GetMaterials( j ).GetShader();
            if ( res_shader.IsValid() )
            {
                nw::gfx::ResReferenceShader ref_shader =
                    nw::ut::ResDynamicCast< nw::gfx::ResReferenceShader >( res_shader );

                // Cast check
                NW_ASSERT( ref_shader.IsValid() );

                // Broken reference detection
                if ( !ref_shader.GetTargetShader().IsValid() )
                {
                    // Determines whether it is ParticleMode.
                    nw::gfx::ResParticleModel particle_model = nw::ut::ResDynamicCast< nw::gfx::ResParticleModel >( graphicsFile.GetModels( i ) );
                    if ( particle_model.IsValid() )
                    {
                        // Configures the default particle shader.
                        nw::gfx::ResShader def_shader = vw::g3d::ResourceManager::GetInstance()->GetDefaultParticleShader();
                        graphicsFile.GetModels( i ).GetMaterials( j ).SetShader( def_shader );
                    }
                    else
                    {
                        // Configures the default shader.
                        nw::gfx::ResShader def_shader = vw::g3d::ResourceManager::GetInstance()->GetDefaultShader();
                        graphicsFile.GetModels( i ).GetMaterials( j ).SetShader( def_shader );
                    }
                }
            }
        }
    }

    return result;
}

}
}
