/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#include <viewer/system.h>
#include <nw/viewer/sys/Application.h>
#include "../../system/include/ConvertCharEncoding.h"
#include "../../system/include/LogUtil.h"
#include "../../system/include/LogSystem.h"
#include "../../system/include/RenderSystem.h"
#include "../include/ResourceHelper.h"
#include "../include/ResourceManager.h"
#include "../include/SceneManager.h"
#include "../include/SceneElemSelector.h"
#include "../include/AnimManager.h"
#include "../include/Setting3Dmisc.h"
#include "../include/ErrorCheck.h"

#include <viewer/system/Memory.h>

extern "C" {

#if defined( NN_PCGX )
    extern u8* DEFAULTLUT_BEGIN;
    extern u8* DEFAULTLUT_END;
    extern u8* DEFAULTTEXTURE_BEGIN;
    extern u8* DEFAULTTEXTURE_END;
    extern u8* DEFAULTSHADER_BEGIN;
    extern u8* DEFAULTSHADER_END;
    extern u8* DEFAULTPARTICLESHADER_BEGIN;
    extern u8* DEFAULTPARTICLESHADER_END;
#else
    extern u8* DEFAULTLUT_BEGIN[];
    extern u8* DEFAULTLUT_END[];
    extern u8* DEFAULTTEXTURE_BEGIN[];
    extern u8* DEFAULTTEXTURE_END[];
    extern u8* DEFAULTSHADER_BEGIN[];
    extern u8* DEFAULTSHADER_END[];
    extern u8* DEFAULTPARTICLESHADER_BEGIN[];
    extern u8* DEFAULTPARTICLESHADER_END[];
#endif

}

namespace vw {
namespace g3d {

#define RESOLVE_REF( src, dst, cond ) \
    it = m_Resources.begin(); \
    end = m_Resources.end(); \
    if ( !(src)->m_isResolved ) \
    { \
        for ( ; it != end; ++it ) { \
            if ( cond ) { \
                result = (src)->GetResource().Setup( m_pAllocator, (dst)->GetResource() ); \
                (src)->m_isResolved = ( result.GetCode() == nw::gfx::RESOURCE_RESULT_OK ); \
                if ( (src)->m_isResolved ) break; \
            } \
        } \
    }

#define CHECK_CONTAINDED( type ) \
    if ( resourceSet.IsContained( res_ref.ref().to##type.to_ptr() ) ) { \
        res_ref.ref().to##type.set_ptr( NULL ); \
    }

#define CHECK_ARRAY_CONTAINDED( obj, type ) \
    for ( int count = 0; count < obj.Get##type##sCount(); ++count ) { \
        nw::gfx::res::ResReferenceSceneObject res_ref = obj.Get##type##s( count ); \
        CHECK_CONTAINDED( Target ) \
    }



/*---------------------------------------------------------------------------*
  Name:         GetResourceSetTypeName

  Description:  (For debugging) Converts the resource type in a character string.

  Arguments:    ResourceSet::BinaryFileType type

  Returns:      Resource type character string.
 *---------------------------------------------------------------------------*/
const char* GetResourceSetTypeName(ResourceSet::BinaryFileType type)
{
#define RETURN_RESOURCESETTYPE_IF_MATCH(enumType) if(type==ResourceSet::enumType) return #enumType;

RETURN_RESOURCESETTYPE_IF_MATCH(Type_Model)
RETURN_RESOURCESETTYPE_IF_MATCH(Type_Texture)
RETURN_RESOURCESETTYPE_IF_MATCH(Type_LookupTableSet)
RETURN_RESOURCESETTYPE_IF_MATCH(Type_SkeletalAnim)
RETURN_RESOURCESETTYPE_IF_MATCH(Type_MaterialAnim)
RETURN_RESOURCESETTYPE_IF_MATCH(Type_SceneEnvironmentSetting)
RETURN_RESOURCESETTYPE_IF_MATCH(Type_Camera)
RETURN_RESOURCESETTYPE_IF_MATCH(Type_Light)
RETURN_RESOURCESETTYPE_IF_MATCH(Type_Fog)
RETURN_RESOURCESETTYPE_IF_MATCH(Type_VisibilityAnim)
RETURN_RESOURCESETTYPE_IF_MATCH(Type_Particle)
RETURN_RESOURCESETTYPE_IF_MATCH(Type_AnimSnd)
RETURN_RESOURCESETTYPE_IF_MATCH(Type_UserData)
RETURN_RESOURCESETTYPE_IF_MATCH(Type_Packed)
return "type not found";
}

ResourceManager* ResourceManager::m_pInstance = NULL;

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
bool
IsContained(nw::gfx::ResGraphicsFile file, void* res_ptr)
{
    if (file.ptr() == NULL) return false;

    const u32 ptr  = (u32)res_ptr;
    const u32 head = (u32)file.ptr();
    const u32 tail = head + file.GetFileSize();

    return ( ptr >= head && ptr <= tail );
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
ResourceSet::ResourceSet(
    nw::ut::MoveArray<u8>& buffer,
    BinaryFileType type,
    unsigned int hashID,
    int channelID
) {
    m_buffer = buffer;
    m_type = type;
    m_hashID = hashID;
    m_channelID = channelID;
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
ResourceSet::ResourceSet(const ResourceSet& resourceSet)
{
    m_buffer = resourceSet.m_buffer;
    m_type = resourceSet.m_type;
    m_hashID = resourceSet.m_hashID;
    m_channelID = resourceSet.m_channelID;
    memcpy(
        m_userData,
        resourceSet.m_userData,
        sizeof( float ) * RES_USE_DATA_SIZE
    );
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
ResourceSet::~ResourceSet()
{
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
*---------------------------------------------------------------------------*/
nw::gfx::ResGraphicsFile
ResourceSet::GetResource()
{
    return nw::gfx::ResGraphicsFile(&(m_buffer.front()));
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
bool
ResourceSet::IsContained(void* res_ptr)
{
    if (m_buffer.size()== 0) return false;
    return vw::g3d::IsContained(GetResource(), res_ptr);
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
ResourceSet::ResourceSet()
{
    memset(
        m_userData,
        0,
        sizeof( float ) * RES_USE_DATA_SIZE
    );
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
ResourceManager*
ResourceManager::GetInstance()
{
    return m_pInstance;
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
ResourceManager::AddResource(ResourceSet& resourceSet)
{
    // Resource setup
    SetupResGraphicsFile( m_pAllocator, resourceSet.GetResource() );

    // Copy of resource set
    void *buf = m_pAllocator->Alloc(sizeof(ResourceSet));
    ResourceSet* res = new(buf) ResourceSet(resourceSet);
    m_Resources.push_back(res);

    // Adds scene node
    vw::g3d::SceneManager::GetInstance()->AddElem(
        res,
        resourceSet.m_type == ResourceSet::Type_Packed || resourceSet.m_type == ResourceSet::Type_Particle
    );

    // Updates resource
    SetupScene();
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
ResourceManager::DeleteResource(const ResourceSet::BinaryFileType type, const unsigned int hashID)
{
    ResourceList::iterator it = Find(type, hashID);
    if (it != m_Resources.end())
    {
        DeleteResource(it);
    }


    // Updates resource
    SetupScene();


}


/*---------------------------------------------------------------------------*
  Name:         DeleteResource

  Description:  Deletes resource

  Arguments:    it : A valid m_Resources iterator (an iterator that is not m_Resources.end)

  Returns:      The next iterator after iterator "it," of the passed m_Resources.
 *---------------------------------------------------------------------------*/
ResourceManager::ResourceList::iterator
ResourceManager::DeleteResource(ResourceManager::ResourceList::iterator it)
{
    // Gets the resource set pointer.
    ResourceSet* resourceSet = &(*it);

    // Deletes it from the log.
    vw::system::LogSystem::GetInstance()->Clear(resourceSet->GetHashID());

    // Deletes from link list.
    ResourceList::iterator next = it; ++next;
    ResourceList::iterator result = m_Resources.erase(it, next);

    // Deletes the scene node.
    vw::g3d::SceneManager::GetInstance()->DeleteElem(resourceSet);

    // Deletes resource.
    resourceSet->GetResource().Cleanup();

    // Deletes object.
    resourceSet->~ResourceSet();
    m_pAllocator->Free(resourceSet);

    return result;
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
ResourceManager::ClearResources(const int channelID)
{


    ResourceList::iterator it = GetFirstResourceByChannel(channelID);
    while( it != m_Resources.end() )
    {
        DeleteResource(it);

        // Updates iterator.
        it = GetFirstResourceByChannel(channelID);
    }

    // Updates resource
    SetupScene();


}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
ResourceManager::ClearResources()
{


    ResourceList::iterator it = m_Resources.begin();
    while( it != m_Resources.end() )
    {
        it = DeleteResource(it);
    }

    // Updates resource
    SetupScene();
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
ResourceManager::LoadBinaryData(
    const ResourceSet::BinaryFileType type,
    const unsigned int hashID,
    const int channelID,
    const void *filename,
    const bool isWChar
) {
    // Loads resource.
    nw::ut::MoveArray<u8> file;
    bool correctlyLoaded = false;

    while (!correctlyLoaded)
    {
        // Checks loading and the 32 MB boundary.
        // Also destroys the buffer loaded at the previous loop.
        bool isOnThe32MBLine = true;
        if (isWChar) {
            file = vw::system::Utility::LoadFileFromHost(m_pDeviceAllocator, (wchar_t*)filename, &isOnThe32MBLine);
        } else {
            file = vw::system::Utility::LoadFileFromHost(m_pDeviceAllocator, (char*)filename, &isOnThe32MBLine);
        }

        // Checks to determine if the cube map has been maintained.
        bool containsCubemaps = false;
        if (file.size() != 0)
        {
            nw::gfx::ResGraphicsFile resGraphicsFile(&(file.front()));
            for (int i=0; i<resGraphicsFile.GetTexturesCount(); ++i)
            {
                if (nw::gfx::res::ResCubeTexture::TYPE_INFO == GetActualTypeInfo(resGraphicsFile.GetTextures(i)))
                {
                    containsCubemaps = true;
                    break;
                }
            }
        }

        // Ends if it is not "maintain cube map and on the 32 MB boundary."
        if (!isOnThe32MBLine || !containsCubemaps)
        {
            correctlyLoaded = true;
        }
    }

    u32 size = (u32)file.size();
    if ( size != 0 )
    {
        // Creates resource set.
        ResourceSet resourceSet(file, type, hashID, channelID);

        // Places resource in VRAM.
        if( vw::system::Video_CanAllocateA(size) )
        {
            if( Setting::GetInstance().GetTexAllocType() == Setting::ALLOCTYPE_VRAM )
            {
                resourceSet.GetResource().ForeachTexture(nw::gfx::TextureLocationFlagSetter(NN_GX_MEM_VRAMA | GL_NO_COPY_FCRAM_DMP));
            }

            if( Setting::GetInstance().GetModelAllocType() == Setting::ALLOCTYPE_VRAM )
            {
                resourceSet.GetResource().ForeachIndexStream(nw::gfx::IndexStreamLocationFlagSetter(NN_GX_MEM_VRAMA | GL_NO_COPY_FCRAM_DMP));
                resourceSet.GetResource().ForeachVertexStream(nw::gfx::VertexStreamLocationFlagSetter(NN_GX_MEM_VRAMA | GL_NO_COPY_FCRAM_DMP));
            }
        }
        else if( vw::system::Video_CanAllocateB(size) )
        {
            if( Setting::GetInstance().GetTexAllocType() == Setting::ALLOCTYPE_VRAM )
            {
                resourceSet.GetResource().ForeachTexture(nw::gfx::TextureLocationFlagSetter(NN_GX_MEM_VRAMB | GL_NO_COPY_FCRAM_DMP));
            }

            if( Setting::GetInstance().GetModelAllocType() == Setting::ALLOCTYPE_VRAM )
            {
                resourceSet.GetResource().ForeachIndexStream(nw::gfx::IndexStreamLocationFlagSetter(NN_GX_MEM_VRAMB | GL_NO_COPY_FCRAM_DMP));
                resourceSet.GetResource().ForeachVertexStream(nw::gfx::VertexStreamLocationFlagSetter(NN_GX_MEM_VRAMB | GL_NO_COPY_FCRAM_DMP));
            }
        }
        else
        {
            if (vw::system::RenderSystem::GetInstance()->IsGraphicsMemoryAllocatorInitialized())
            {
                ErrorCheck::Error_VRAM_Alloc();
            }
        }

        // Adds the explicit flush flag.
        resourceSet.GetResource().ForeachModel(nw::gfx::ParticleSetIsBufferFlushEnabledSetter(true));

        // Gets the callback class.
        nw::viewer::g3d::IResourceListener* g3dListener =
            nw::viewer::sys::App::GetInstance().GetModule3D().GetResourceListener();
        nw::viewer::ef::IResourceListener* efListener =
            nw::viewer::sys::App::GetInstance().GetModuleEF().GetResourceListener();

        // Invoke callbacks
        if (g3dListener) {
            g3dListener->OnLoadResource(resourceSet.GetResource());
        }

        // Invoke callbacks
        if (efListener) {
            efListener->OnLoadResource(resourceSet.GetResource());
        }

        // Deallocates resource with the same type and the same hash name just in case.
        DeleteResource(type, hashID);

        // Registers in ResourceManager.
        AddResource(resourceSet);
    }

    if (isWChar) {
        const int MAX_PATH_SIZE = 260;
        char utf8Str[MAX_PATH_SIZE];
        vw::system::ConvertStrWtoUTF8((wchar_t*)filename, utf8Str, MAX_PATH_SIZE);
        NW_VIEWER_DEV_LOG("Filename : %s\n", utf8Str);
    } else {
        NW_VIEWER_DEV_LOG("Filename : %s\n", (char*)filename);
    }
    NW_VIEWER_LOG("ObjectID : %u\n", hashID);
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
ResourceManager::ResourceList::iterator
ResourceManager::Find(const ResourceSet::BinaryFileType type, const unsigned int hashID)
{
    ResourceList::iterator it = m_Resources.begin();
    for (; it != m_Resources.end(); ++it) {
        if (it->m_type == type && it->m_hashID == hashID) return it;
    }
    NW_VIEWER_LOG("Resource type[%u], id[%u] not found.\n", type, hashID);
    return m_Resources.end();
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
ResourceManager::ResourceList::iterator
ResourceManager::Find(void* ptr)
{
    ResourceList::iterator it = m_Resources.begin();
    for (; it != m_Resources.end(); ++it) {
        if (it->IsContained(ptr)) return it;
    }
    NW_VIEWER_LOG("Resource ptr[%u] not found.\n", ptr);
    return m_Resources.end();
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
ResourceManager::ResourceList::iterator
ResourceManager::GetResourcesBegin()
{
    return m_Resources.begin();
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
ResourceManager::ResourceList::iterator
ResourceManager::GetResourcesEnd()
{
    return m_Resources.end();
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
nw::gfx::res::ResImageLookupTable
ResourceManager::GetDefaultLookupTable()
{
    nw::gfx::res::ResImageLookupTable lut;

    nw::gfx::res::ResGraphicsFile file(m_DefaultLookupTableSet.GetResource());
    if (file.IsValid()) {
        nw::gfx::res::ResLookupTableSet lutSet = file.GetLutSets( 0 );
        if (lutSet.IsValid()) lut = lutSet.GetSamplers( 0 );
    }

    return lut;
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
nw::gfx::res::ResTexture
ResourceManager::GetDefaultTexture()
{
    nw::gfx::res::ResTexture texture;

    nw::gfx::res::ResGraphicsFile file(m_DefaultTexture.GetResource());
    if (file.IsValid()) texture = file.GetTextures( 0 );

    return texture;
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
nw::gfx::res::ResShader
ResourceManager::GetDefaultShader()
{
    nw::gfx::res::ResShader shader;

    nw::gfx::res::ResGraphicsFile file(m_DefaultShader.GetResource());
    if (file.IsValid()) shader = file.GetShaders( 0 );

    return shader;
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
nw::gfx::res::ResShader
ResourceManager::GetDefaultParticleShader()
{
    nw::gfx::res::ResShader particleShader;

    nw::gfx::res::ResGraphicsFile file(m_DefaultParticleShader.GetResource());
    if (file.IsValid()) particleShader = file.GetShaders( 0 );

    return particleShader;
}


/*---------------------------------------------------------------------------*
  Name:         DumpMemory

  Description:  (For debugging) Dumps the allocator used by the ResourceManager.

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
ResourceManager::DumpMemory()
{
    vw::system::Memory::Allocator* pAllocator = static_cast<vw::system::Memory::Allocator*>(m_pAllocator);

    if(!pAllocator)
    {
        NW_VIEWER_DEV_LOG("ResourceManager::DumpMemory - cast fail \n");
    }
    else
    {
        pAllocator->Dump();
    }
}


/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
ResourceManager::ResourceManager()
{
    m_pAllocator = Memory_GetAllocator();
    m_pDeviceAllocator = Memory_GetAllocatorDevice();

    LoadDefaultLookupTableSet();
    LoadDefaultTexture();
    LoadDefaultShader();
    LoadDefaultParticleShader();
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
ResourceManager::~ResourceManager()
{
    m_DefaultLookupTableSet.GetResource().Cleanup();
    m_DefaultTexture.GetResource().Cleanup();
    m_DefaultShader.GetResource().Cleanup();
    m_DefaultParticleShader.GetResource().Cleanup();
    ClearResources();
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
ResourceManager::Initialize()
{
    if (m_pInstance != NULL) return;

    void *buf = Memory_GetAllocator()->Alloc(sizeof(ResourceManager));
    if (buf == NULL) return;

    m_pInstance = new(buf) ResourceManager();
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
ResourceManager::Finalize()
{
    if (m_pInstance == NULL) return;

    m_pInstance->~ResourceManager();
    Memory_GetAllocator()->Free(m_pInstance);

    m_pInstance = NULL;
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
ResourceManager::LoadDefaultLookupTableSet()
{
    // Loads the default reference table.
    u32 size = static_cast<u32>( reinterpret_cast<int>(DEFAULTLUT_END) - reinterpret_cast<int>(DEFAULTLUT_BEGIN) );
    NW_ASSERT( size != 0 );

     // Allocates the buffer.
    void* memory = m_pDeviceAllocator->Alloc(size, static_cast<u8>(128));
    NW_NULL_ASSERT( memory );

    nw::ut::MoveArray<u8> buffer(memory, size, m_pDeviceAllocator);
    if ( !buffer.resize(size) )
    {
        NW_VIEWER_DEV_LOG("Failed to resize buffer.\n");
    }

    memcpy(&buffer.front(), DEFAULTLUT_BEGIN, size);

    // Sets resource.
    m_DefaultLookupTableSet.m_buffer = buffer;
    m_DefaultLookupTableSet.m_type = ResourceSet::Type_Packed;
    m_DefaultLookupTableSet.m_hashID = 0;
    m_DefaultLookupTableSet.m_channelID = -1;

    // Initializes resources.
    nw::gfx::Result result = m_DefaultLookupTableSet.GetResource().Setup( m_pAllocator );
    NW_ASSERT( result.GetCode() == nw::gfx::RESOURCE_RESULT_OK );
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
ResourceManager::LoadDefaultTexture()
{
    // Loads the default texture
    u32 size = static_cast<u32>( reinterpret_cast<int>(DEFAULTTEXTURE_END) - reinterpret_cast<int>(DEFAULTTEXTURE_BEGIN) );
    NW_ASSERT( size != 0 );

    // Allocates the buffer.
    void* memory = m_pDeviceAllocator->Alloc(size, static_cast<u8>(128));
    NW_NULL_ASSERT( memory );

    nw::ut::MoveArray<u8> buffer(memory, size, m_pDeviceAllocator);
    if ( !buffer.resize(size) )
    {
        NW_VIEWER_DEV_LOG("Failed to resize buffer.\n");
    }

    // Load from file
    memcpy(&buffer.front(), DEFAULTTEXTURE_BEGIN, size);

    // Sets resource.
    m_DefaultTexture.m_buffer = buffer;
    m_DefaultTexture.m_type = ResourceSet::Type_Packed;
    m_DefaultTexture.m_hashID = 0;
    m_DefaultTexture.m_channelID = -1;

    // Initializes resources.
    if(false == vw::system::App::IsEmbeddedViewer())
    {
        m_DefaultTexture.GetResource().ForeachTexture(
            nw::gfx::TextureLocationFlagSetter(NN_GX_MEM_VRAMA | GL_NO_COPY_FCRAM_DMP)
        );
    }

    nw::gfx::Result result = m_DefaultTexture.GetResource().Setup( m_pAllocator );
    NW_ASSERT( result.GetCode() == nw::gfx::RESOURCE_RESULT_OK );
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
ResourceManager::LoadDefaultShader()
{
    // Loads the default shader
    u32 size = static_cast<u32>( reinterpret_cast<int>(DEFAULTSHADER_END) - reinterpret_cast<int>(DEFAULTSHADER_BEGIN) );
    NW_ASSERT( size != 0 );

    // Allocates the buffer.
    void* memory = m_pDeviceAllocator->Alloc(size, static_cast<u8>(128));
    NW_NULL_ASSERT( memory );

    nw::ut::MoveArray<u8> buffer(memory, size, m_pDeviceAllocator);
    if ( !buffer.resize(size) )
    {
        NW_VIEWER_DEV_LOG("Failed to resize buffer.\n");
    }

    // Load from file
    memcpy(&buffer.front(), DEFAULTSHADER_BEGIN, size);

    // Sets resource.
    m_DefaultShader.m_buffer = buffer;
    m_DefaultShader.m_type = ResourceSet::Type_Packed;
    m_DefaultShader.m_hashID = 0;
    m_DefaultShader.m_channelID = -1;

    // Initializes resources.
    nw::gfx::Result result = m_DefaultShader.GetResource().Setup( m_pAllocator );
    NW_ASSERT( result.GetCode() == nw::gfx::RESOURCE_RESULT_OK );
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
ResourceManager::LoadDefaultParticleShader()
{
    // Loads the default particle shader
    u32 size = static_cast<u32>( reinterpret_cast<int>(DEFAULTPARTICLESHADER_END) - reinterpret_cast<int>(DEFAULTPARTICLESHADER_BEGIN) );
    NW_ASSERT( size != 0 );

    // Allocates the buffer.
    void* memory = m_pDeviceAllocator->Alloc(size, static_cast<u8>(128));
    NW_NULL_ASSERT( memory );

    nw::ut::MoveArray<u8> buffer(memory, size, m_pDeviceAllocator);
    if ( !buffer.resize(size) )
    {
        NW_VIEWER_DEV_LOG("Failed to resize buffer.\n");
    }

    // Load from file
    memcpy(&buffer.front(), DEFAULTPARTICLESHADER_BEGIN, size);

    // Sets resource.
    m_DefaultParticleShader.m_buffer = buffer;
    m_DefaultParticleShader.m_type = ResourceSet::Type_Packed;
    m_DefaultParticleShader.m_hashID = 0;
    m_DefaultParticleShader.m_channelID = -1;

    // Initializes resources.
    nw::gfx::Result result = m_DefaultParticleShader.GetResource().Setup( m_pAllocator );
    NW_ASSERT( result.GetCode() == nw::gfx::RESOURCE_RESULT_OK );
}


/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
ResourceManager::SetupScene()
{
    // Updates the scene context.
    SceneManager::GetInstance()->UpdateSceneContext();

    // Updates the selection object array.
    SceneElemSelector::GetInstance()->Setup();
}



/*---------------------------------------------------------------------------*
  Name:            GetFirstResourceByChannel

  Description:  Searches m_Resources elements by their m_channelID parameter
                (Searches m_Resources using m_channelID.)

  Arguments:    channel ID

  Returns:      Returns an iterator to the first element in m_Resources whose m_channelID matches the specified parameter channelID
                Returns m_channelID.end() if no such element was found
                (Returns the first resource iterator found that agrees. Returns m_channelID.end if not found.)
 *---------------------------------------------------------------------------*/
ResourceManager::ResourceList::iterator
ResourceManager::GetFirstResourceByChannel(const int channelID)
{
    ResourceList::iterator it;
    for( it = m_Resources.begin(); it != m_Resources.end(); ++it)
    {
        if( it->m_channelID == channelID )
            break;
    }
    return it;
}



}
}
