/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#include <viewer/system.h>
#include <nw/viewer/sys/Application.h>
#include <viewer/g3d/Camera.h>
#include <viewer/g3d/LightMgr.h>
#include "../include/SceneManager.h"
#include "../include/ResourceManager.h"
#include "../include/SceneElemSelector.h"
#include "../include/SceneRenderer.h"
#include "../include/UserDataManager.h"
#include "../include/ErrorCheck.h"
#include "../../system/include/LogUtil.h"

#if NN_CURRENT_VERSION_NUMBER >= NN_VERSION_NUMBER(4,0,0,0)
#pragma diag_suppress 2530 // padding added to end of struct.
#endif
namespace vw {
namespace g3d {

#define ARRAY_ERASE( type, it, name ) \
    type::iterator it##_next = it; \
    ++it##_next; \
    it = m_##name##Array.erase( it, it##_next ); \

#define BUILD_LIST_ELEM( type, res ) \
    type##Elem *type##_elem = \
        type##Elem::Builder() \
        .ResSource( resourceSet ) \
        .ResObject( res ) \
        .Create(m_pAllocator, m_pDeviceAllocator); \
    m_##type##Array.push_back( type##_elem );

#if 0
#define ADD_SCENE_ELEM( src, dst ) \
    for ( int i = 0; i < file.Get##src##sCount(); ++i ) \
    { \
        BUILD_LIST_ELEM( dst, file.Get##src##s( i ) ) \
        sceneNodeArray.push_back( dst##_elem->GetSceneNode() ); \
    }
#endif

#define ADD_ANIM_ELEM( type, moveArray ) \
    for ( int i = 0; i < file.Get##type##AnimsCount(); ++i ) \
    { \
        AnimElem *elem = AnimElem::Builder()\
                                  .ResSource( resourceSet )\
                                  .ResObject( file.Get##type##Anims(i) )\
                                  .AnimType( AnimElem::Anim_##type )\
                                  .Create(m_pAllocator, m_pDeviceAllocator); \
        if (g3dListener) { \
            g3dListener->OnAddAnimation( \
                elem->GetResObject(), \
                nw::viewer::g3d::Anim_##type \
            ); \
        } \
        NW_NULL_ASSERT( elem );\
        m_AnimArray.push_back( elem ); \
        if( solveAnimReferences ) moveArray.push_back( elem ); \
    }

#define ADD_ANIM_SCENE_ELEM( src, dst ) \
    { \
        nw::ut::MoveArray<AnimSetLinker*> linkers( m_pAllocator ); \
        for ( int i = 0; i < file.Get##src##sCount(); ++i ) \
        { \
            BUILD_LIST_ELEM( dst, file.Get##src##s( i ) ) \
            sceneNodeArray.push_back( dst##_elem->GetSceneNode() ); \
            linkers.push_back( dst##_elem->GetAnimLinker() ); \
        } \
        if ( solveAnimReferences ) \
            AnimSetLinker::SolveAnimBindingsByName( linkers, dst##AnimArray, m_pAllocator ); \
    }

#define DELETE_RES_ELEM( type ) { \
    type##List::iterator it  = m_##type##Array.begin(); \
    for ( ; it != m_##type##Array.end(); ) { \
        if ( resourceSet == it->GetResSource() ) { \
            type##Elem* elem = &(*it); \
            if (g3dListener) { \
                g3dListener->OnRemove##type( elem->GetSceneNode() ); \
            } \
            ARRAY_ERASE( type##List, it, type ) \
            elem->Destroy(); \
            continue; \
        } \
        it++; \
    } }

#define DELETE_SCENE_ELEM( type ) { \
    type##List::iterator it  = m_##type##Array.begin(); \
    for ( ; it != m_##type##Array.end(); ) { \
        if ( resourceSet == it->GetResSource() ) { \
            SceneRenderer::GetInstance()->DeleteDrawAxis( it->GetSceneNode(), NULL ); \
            NW_ASSERT( subRootNode == NULL || subRootNode == it->GetSceneNode()->GetParent() ); \
            subRootNode = it->GetSceneNode()->GetParent(); \
            type##Elem* elem = &(*it); \
            if (g3dListener) { \
                g3dListener->OnRemove##type( elem->GetSceneNode() ); \
            } \
            ARRAY_ERASE( type##List, it, type ) \
            elem->Destroy(); \
            continue; \
        } \
        it++; \
    } }

#define DELETE_MODEL_ELEM( type ) { \
    type##List::iterator it  = m_##type##Array.begin(); \
    for ( ; it != m_##type##Array.end(); ) { \
        if ( resourceSet == it->GetResSource() ) { \
            nw::gfx::SceneNode* previewNode = it->GetSceneNode()->GetParent(); \
            nw::gfx::TransformNode* subRoot = \
                nw::ut::DynamicCast<nw::gfx::TransformNode*>( previewNode->GetParent() ); \
            FOREACH_ELEM( ModelList, model_it, Model ) { \
                model_it->GetNodeConstraint().DeleteConstraint(subRoot); \
            } \
            SceneRenderer::GetInstance()->DeleteDrawAxis( it->GetSceneNode(), NULL ); \
            NW_ASSERT( subRootNode == NULL || subRootNode == subRoot ); \
            subRootNode = subRoot; \
            type##Elem* elem = &(*it); \
            if (g3dListener) { \
                g3dListener->OnRemove##type( elem->GetSceneNode() ); \
            } \
            ARRAY_ERASE( type##List, it, type ) \
            elem->Destroy(); \
            nw::gfx::SafeDestroy( previewNode ); \
            continue; \
        } \
        it++; \
    } }

#define DELETE_PARTICLE_ELEM( type ) { \
    type##List::iterator it  = m_##type##Array.begin(); \
    for ( ; it != m_##type##Array.end(); ) { \
        if ( resourceSet == it->GetResSource() ) { \
            FOREACH_ELEM( ModelList, model_it, Model ) { \
                model_it->GetNodeConstraint().DeleteConstraint(it->GetSceneNode()); \
            } \
            SceneRenderer::GetInstance()->DeleteDrawAxis( it->GetSceneNode(), NULL ); \
            subRootNode = it->GetSceneNode()->GetParent(); \
            type##Elem* elem = &(*it); \
            if (efListener) { \
                efListener->OnRemove##type( elem->GetSceneNode() ); \
            } \
            ARRAY_ERASE( type##List, it, type ) \
            elem->Destroy(); \
            continue; \
        } \
        it++; \
    } }

#define FOREACH_ELEM( type, iteratorName, name ) \
    for (type::iterator iteratorName = m_##name##Array.begin(), end_it_ = m_##name##Array.end(); \
        iteratorName != end_it_; ++iteratorName )

/*---------------------------------------------------------------------------*/
#define DECL_RES_FUNC( type, resType, name ) \
    SceneManager::name##List::iterator \
    SceneManager::Find##name( unsigned int hashID, unsigned int index ) \
    { \
        unsigned int found = 0; \
        name##List::iterator it = Get##name##sBegin(); \
        name##List::iterator end = Get##name##sEnd(); \
        for ( ; it != end; ++it ) { \
            if ( \
                it->GetResSource()->GetHashID() == hashID && \
                it->GetResSource()->GetBinaryType() != ResourceSet::Type_Packed \
            ) { if ( index == found++ ) return it; } \
        } \
        if (it == end) NW_VIEWER_LOG("File not found [ %u, %u ]\n", hashID, index);\
        return it; \
    } \
    SceneManager::name##List::iterator \
    SceneManager::Find##name( resType object ) \
    { \
        name##List::iterator it = Get##name##sBegin(); \
        name##List::iterator end = Get##name##sEnd(); \
        for ( ; it != end; ++it ) { \
            if ( it->GetResObject().ptr() == object.ptr() ) return it; \
        } \
        return it; \
    } \
    resType \
    SceneManager::FindRes##name( unsigned int hashID, unsigned int index ) \
    { \
        resType res; \
        name##List::iterator it = Find##name( hashID, index ); \
        if ( it != Get##name##sEnd() ) res = it->GetResObject(); \
        return res; \
    } \

#define DECL_SCENE_FUNC( type, name ) \
    DECL_RES_FUNC( type, nw::gfx::Res##name, name ) \
    SceneManager::name##List::iterator \
    SceneManager::Find##name( nw::gfx::name* object ) \
    { \
        name##List::iterator it = Get##name##sBegin(); \
        name##List::iterator end = Get##name##sEnd(); \
        for ( ; it != end; ++it ) { \
            if ( it->GetSceneNode() == object ) return it; \
        } \
        return it; \
    } \
    nw::gfx::name* \
    SceneManager::Find##name##Node( unsigned int hashID, unsigned int index ) \
    { \
        nw::gfx::name* node = NULL; \
        name##List::iterator it = Find##name( hashID, index ); \
        if ( it != Get##name##sEnd() ) node = it->GetSceneNode(); \
        return node; \
    } \

DECL_SCENE_FUNC( ModelElem, Model )
DECL_RES_FUNC( TextureElem, nw::gfx::ResTexture, Texture )
DECL_RES_FUNC( LookupTableSetElem, nw::gfx::ResLookupTableSet, LookupTableSet )
DECL_RES_FUNC( AnimElem, nw::anim::ResAnim, Anim )

DECL_SCENE_FUNC( CameraElem, Camera )
DECL_SCENE_FUNC( LightElem, Light )
DECL_SCENE_FUNC( FogElem, Fog )
DECL_SCENE_FUNC( SceneEnvironmentSettingElem, SceneEnvironmentSetting )

DECL_SCENE_FUNC( ParticleModelElem, ParticleModel )
DECL_SCENE_FUNC( ParticleEmitterElem, ParticleEmitter )

SceneManager* SceneManager::m_pInstance = NULL;

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
SceneManager::SetCameraWScale(nw::gfx::Camera* objCamera)
{
    f32 wScale = 0.0f;
    {
        nw::gfx::res::ResCamera resCamera = objCamera->GetResCamera();
        f32 nFar = 1.0f;
        if( resCamera.GetProjectionType() == nw::gfx::res::ResCamera::PROJTYPE_PERSPECTIVE ){
            nw::gfx::res::ResPerspectiveProjectionUpdater resPPUpdater =  nw::ut::ResDynamicCast<nw::gfx::res::ResPerspectiveProjectionUpdater>(resCamera.GetProjectionUpdater());
            nFar = resPPUpdater.GetFar();
        }else if( resCamera.GetProjectionType() == nw::gfx::res::ResCamera::PROJTYPE_ORTHO ){
            nw::gfx::res::ResOrthoProjectionUpdater resOPUpdater =  nw::ut::ResDynamicCast<nw::gfx::res::ResOrthoProjectionUpdater>(resCamera.GetProjectionUpdater());
            nFar = resOPUpdater.GetFar();
        }
        wScale = 1.0f / nFar;
    }
    objCamera->SetWScale( wScale );
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
SceneManager*
SceneManager::GetInstance()
{
    NW_ASSERT( m_pInstance != NULL );
    return m_pInstance;
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
SceneManager::AddElem(
    ResourceSet* resourceSet,
    bool solveAnimReferences
) {
    // Returned immediately for special data for user data
    if (resourceSet->GetBinaryType() == ResourceSet::Type_UserData) return;

    // Gets the callback class.
    nw::viewer::g3d::IResourceListener* g3dListener =
        nw::viewer::sys::App::GetInstance().GetModule3D().GetResourceListener();
    nw::viewer::ef::IResourceListener* efListener =
        nw::viewer::sys::App::GetInstance().GetModuleEF().GetResourceListener();

    // Getting ResGraphicsFile.
    nw::gfx::res::ResGraphicsFile file = resourceSet->GetResource();

    // Adds objects that have scene nodes
    nw::ut::MoveArray<nw::gfx::SceneNode*> sceneNodeArray(m_pAllocator);
    nw::ut::MoveArray<ModelElem*> modelArray(m_pAllocator);
    nw::ut::MoveArray<ParticleModelElem*> particleModels(m_pAllocator);
    nw::ut::MoveArray<nw::gfx::ParticleEmitter*> particleEmitters(m_pAllocator);
    nw::ut::MoveArray<AnimElem*> modelAnimArray(m_pAllocator);
    nw::ut::MoveArray<AnimElem*> CameraAnimArray(m_pAllocator);
    nw::ut::MoveArray<AnimElem*> LightAnimArray(m_pAllocator);
    nw::ut::MoveArray<AnimElem*> FogAnimArray(m_pAllocator);
    {
        // Add a particle emitter
        for ( int i = 0; i < file.GetEmittersCount(); ++i )
        {
            BUILD_LIST_ELEM( ParticleEmitter, file.GetEmitters( i ) )
            particleEmitters.push_back( ParticleEmitter_elem->GetSceneNode() );
            sceneNodeArray.push_back( ParticleEmitter_elem->GetSceneNode() );
        }

        // Add model animation
        {
            ADD_ANIM_ELEM( Skeletal, modelAnimArray )
            ADD_ANIM_ELEM( Material, modelAnimArray )
            ADD_ANIM_ELEM( Visibility, modelAnimArray )
        }

        // Add camera animation
        ADD_ANIM_ELEM( Camera, CameraAnimArray )

        // Add light animation
        ADD_ANIM_ELEM( Light, LightAnimArray )

        // Add fog animation
        ADD_ANIM_ELEM( Fog, FogAnimArray )

        // Model generation
        {
            nw::ut::MoveArray<AnimSetLinker*> linkers(m_pAllocator);
            for ( int i = 0; i < file.GetModelsCount(); ++i )
            {
                nw::gfx::res::ResModel model = file.GetModels( i );
                nw::gfx::res::ResParticleModel particle_model =
                    nw::ut::ResDynamicCast<nw::gfx::res::ResParticleModel>( model );

                // For ParticleModel
                if ( particle_model.ptr() != NULL ) {
                    BUILD_LIST_ELEM( ParticleModel, particle_model )
                    sceneNodeArray.push_back( ParticleModel_elem->GetSceneNode() );
                    particleModels.push_back( ParticleModel_elem );
                    linkers.push_back( ParticleModel_elem->GetAnimController()->GetAnimSetLinker() );
                // In all other cases
                } else {
                    BUILD_LIST_ELEM( Model, model )
                    sceneNodeArray.push_back( Model_elem->GetSceneNode() );
                    modelArray.push_back( Model_elem );
                    linkers.push_back( Model_elem->GetAnimLinker() );
                }
            }

            if ( solveAnimReferences )
                AnimSetLinker::SolveAnimBindingsByName( linkers, modelAnimArray, m_pAllocator );
        }

        // Camera generation
        ADD_ANIM_SCENE_ELEM( Camera, Camera )

        // Light generation
        {
            nw::ut::MoveArray<AnimSetLinker*> linkers( m_pAllocator );
            for ( int i = 0; i < file.GetLightsCount(); ++i )
            {
                nw::gfx::res::ResLight resLight = file.GetLights( i );
            #if !defined(NW_GFX_VERTEX_LIGHT_ENABLED)
                if ( resLight.GetTypeInfo() == nw::gfx::res::ResVertexLight::TYPE_INFO ) {
                    ErrorCheck::Error_VertexLight_Disalbe();
                    continue;
                }
            #endif
                BUILD_LIST_ELEM( Light, resLight )
                sceneNodeArray.push_back( Light_elem->GetSceneNode() );
                linkers.push_back( Light_elem->GetAnimLinker() );
            }
            if ( solveAnimReferences )
                AnimSetLinker::SolveAnimBindingsByName( linkers, LightAnimArray, m_pAllocator );
        }

        // Fog generation
        ADD_ANIM_SCENE_ELEM( Fog, Fog )
    }

    // Adds objects that do not have scene nodes
    {
        // Add texture resource
        for ( int i = 0; i < file.GetTexturesCount(); ++i )
        {
            nw::gfx::res::ResTexture texture = file.GetTextures( i );
            BUILD_LIST_ELEM( Texture, texture )
            FOREACH_ELEM( ModelList, it, Model ) { it->GetResourceLinker()->Link( texture ); }
        }

        // Add reference table resource
        for ( int i = 0; i < file.GetLutSetsCount(); ++i )
        {
            nw::gfx::res::ResLookupTableSet lutSet = file.GetLutSets( i );
            BUILD_LIST_ELEM( LookupTableSet, lutSet )
            FOREACH_ELEM( ModelList, it, Model ) { it->GetResourceLinker()->Link( lutSet ); }
            FOREACH_ELEM( LightList, it, Light ) { it->GetResourceLinker()->Link( lutSet ); }
        }

        // Add environment setting resource
        for ( int i = 0; i < file.GetSceneEnvironmentSettingsCount(); ++i )
        {
            nw::gfx::res::ResSceneEnvironmentSetting scene = file.GetSceneEnvironmentSettings( i );
            BUILD_LIST_ELEM( SceneEnvironmentSetting, scene )
            if (g3dListener) {
                g3dListener->OnAddSceneEnvironmentSetting( SceneEnvironmentSetting_elem->GetSceneNode() );
            }
            ErrorCheck::CheckSceneEnvironmentSetting( SceneEnvironmentSetting_elem->GetSceneNode() );
        }
    }

    // Resource setup
    SetupResource( sceneNodeArray );

    // Error check
    CheckError( sceneNodeArray );

    // callback call
    for (int i = 0; i < sceneNodeArray.size(); ++i)
    {
        nw::gfx::SceneNode* node = sceneNodeArray[ i ];
        switch( node->GetResSceneObject().GetTypeInfo() )
        {
        case nw::gfx::res::ResModel_TYPE_INFO:
        case nw::gfx::res::ResSkeletalModel_TYPE_INFO:
            if (g3dListener) {
                g3dListener->OnAddModel( nw::ut::DynamicCast<nw::gfx::Model*>( node ) );
            }
            break;
        case nw::gfx::res::ResCamera_TYPE_INFO:
            if (g3dListener) {
                g3dListener->OnAddCamera( nw::ut::DynamicCast<nw::gfx::Camera*>( node ) );
            }
            break;
        case nw::gfx::res::ResLight_TYPE_INFO:
            if (g3dListener) {
                g3dListener->OnAddLight( nw::ut::DynamicCast<nw::gfx::Light*>( node ) );
            }
            break;
        case nw::gfx::res::ResFog_TYPE_INFO:
            if (g3dListener) {
                g3dListener->OnAddFog( nw::ut::DynamicCast<nw::gfx::Fog*>( node ) );
            }
            break;
        case nw::gfx::res::ResParticleModel_TYPE_INFO:
            if (efListener) {
                efListener->OnAddParticleModel( nw::ut::DynamicCast<nw::gfx::ParticleModel*>( node ) );
            }
            break;
        case nw::gfx::res::ResParticleEmitter_TYPE_INFO:
            if (efListener) {
                efListener->OnAddParticleEmitter( nw::ut::DynamicCast<nw::gfx::ParticleEmitter*>( node ) );
            }
            break;
        }
    }

    // Insert preview node for model resource
    for (int i = 0; i < modelArray.size(); ++i)
    {
        // Gets object
        ModelElem* Model_elem = modelArray[ i ];

        // Gets the parent node
        nw::gfx::SceneNode* parent = Model_elem->GetSceneNode()->GetParent();

        // Deletes model node from the parent node
        parent->DetachChild( Model_elem->GetSceneNode() );

        // Generates node for Preview conversion
        nw::gfx::SceneNode *previewNode = nw::gfx::TransformNode::DynamicBuilder()
            .IsFixedSizeMemory(false)
            .Create(m_pDeviceAllocator);

        // Add Preview conversion node to the parent node
        parent->AttachChild( previewNode );

        // Add model node to Preview conversion node
        previewNode->AttachChild( Model_elem->GetSceneNode() );
    }

    // Bind the particle resources
    for (int i = 0; i < particleModels.size(); ++i)
    {
        nw::gfx::ParticleModel *pParticleModel = particleModels[i]->GetSceneNode();
        for (int j=0; j<pParticleModel->GetParticleSetsCount(); ++j)
        {
            for (int k=0; k<particleEmitters.size(); ++k)
            {
                if ( particleEmitters[k]->GetParticleSet() == pParticleModel->GetParticleSets(j))
                {
                    particleModels[i]->AddParticleEmitter( particleEmitters[k] );
                }
            }
        }
    }
}

/*---------------------------------------------------------------------------*
  Name:
  Description:  perform the following actions on resourceSet (if its type is not userdata) :
                - unlink from the UserDataManager
                - remove it from SceneManager&apos;s resource arrays (m_ModelArray, m_ParticlesArray, etc), including child node, related resources
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
SceneManager::DeleteElem(ResourceSet* resourceSet)
{
    // Returned immediately for special data for user data
    if (resourceSet->GetBinaryType() == ResourceSet::Type_UserData) return;

    // Gets the callback class.
    nw::viewer::g3d::IResourceListener* g3dListener =
        nw::viewer::sys::App::GetInstance().GetModule3D().GetResourceListener();
    nw::viewer::ef::IResourceListener* efListener =
        nw::viewer::sys::App::GetInstance().GetModuleEF().GetResourceListener();

    // Deletes the resource in UserDataManager
    if ( UserDataManager::GetInstance() ) {
        UserDataManager::GetInstance()->Unlink( resourceSet );
    }

    // Deletes objects that have scene nodes
    nw::gfx::SceneNode* subRootNode = NULL;
    {
        DELETE_MODEL_ELEM( Model )
        DELETE_SCENE_ELEM( Camera )
        DELETE_SCENE_ELEM( Light )
        DELETE_SCENE_ELEM( Fog )
        DELETE_PARTICLE_ELEM( ParticleEmitter )
        DELETE_PARTICLE_ELEM( ParticleModel )
    }

    // Deletes objects that do not have scene nodes
    {
        // Deletes texture resources
        {
            TextureList::iterator it = m_TextureArray.begin();
            for ( ; it != m_TextureArray.end(); ) {
                if ( resourceSet == it->GetResSource() ) {
                    FOREACH_ELEM( ModelList, model_it, Model ) {
                        model_it->GetResourceLinker()->Unlink( it->GetResObject() );
                    }
                    TextureElem* elem = &(*it);
                    ARRAY_ERASE( TextureList, it, Texture )
                    elem->Destroy();
                    continue;
                }
                it++;
            }
        }

        // Deletes reference table resource
        {
            LookupTableSetList::iterator it = m_LookupTableSetArray.begin();
            for ( ; it != m_LookupTableSetArray.end(); ) {
                if ( resourceSet == it->GetResSource() ) {
                    FOREACH_ELEM( ModelList, model_it, Model ) {
                        model_it->GetResourceLinker()->Unlink( it->GetResObject() );
                    }
                    FOREACH_ELEM( LightList, light_it, Light ) {
                        light_it->GetResourceLinker()->Unlink( it->GetResObject() );
                    }
                    LookupTableSetElem* elem = &(*it);
                    ARRAY_ERASE( LookupTableSetList, it, LookupTableSet )
                    elem->Destroy();
                    continue;
                }
                it++;
            }
        }

        // Deletes environment setting resource
        DELETE_RES_ELEM( SceneEnvironmentSetting )
    }

    // Deletes animation object
    {
        AnimList::iterator it  = m_AnimArray.begin();
        for ( ; it != m_AnimArray.end(); ) {
            if ( resourceSet == it->GetResSource() ) {
                AnimElem* elem = &(*it);
                if (g3dListener) {
                    nw::viewer::g3d::AnimType type;
                    switch( elem->GetAnimType() ) {
                    case AnimElem::Anim_Skeletal:
                        type = nw::viewer::g3d::Anim_Skeletal;
                        break;
                    case AnimElem::Anim_Material:
                        type = nw::viewer::g3d::Anim_Material;
                        break;
                    case AnimElem::Anim_Visibility:
                        type = nw::viewer::g3d::Anim_Visibility;
                        break;
                    case AnimElem::Anim_Camera:
                        type = nw::viewer::g3d::Anim_Camera;
                        break;
                    case AnimElem::Anim_Light:
                        type = nw::viewer::g3d::Anim_Light;
                        break;
                    case AnimElem::Anim_Fog:
                        type = nw::viewer::g3d::Anim_Fog;
                        break;
                    default:
                        break;
                    }
                    g3dListener->OnRemoveAnimation( elem->GetResAnim(), type );
                }
                ARRAY_ERASE( AnimList, it, Anim )
                elem->Destroy();
                continue;
            }
            it++;
        }
    }

    // Deallocates sub root
    if ( subRootNode ) {
        subRootNode->GetParent()->DetachChild( subRootNode );
        nw::gfx::SafeDestroy( subRootNode );
    }
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
SceneManager::InitContext(const Description& description)
{
    if (m_pSceneContext || m_pParticleContext) return;

    // Scene context initialization
    m_pSceneContext = nw::gfx::SceneContext::Builder()
        .MaxSceneNodes(description.maxSceneNodes)
        .IsFixedSizeMemory(false)
        .Create(m_pAllocator);

    // Particle context initialization
    m_pParticleContext = nw::gfx::ParticleContext::Builder()
        .MaxEmission(description.maxEmission)
        .Create(m_pAllocator);
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
SceneManager::UpdateSceneContext()
{
    // Get context
    nw::gfx::SceneContext* context = GetSceneContext();

    // Clears context
    context->Clear();

    // Initialize.
    m_pSceneRoot->Accept(m_pSceneInitializer);

    // Traverses context
    m_pSceneTraverser->Begin(context);
    m_pSceneRoot->Accept(m_pSceneTraverser);
    m_pSceneTraverser->End();

    // Set by performing environment reference resolutions.
    {
        SceneEnvironmentSettingList::iterator it = GetSceneEnvironmentSettingsBegin();
        SceneEnvironmentSettingList::iterator end = GetSceneEnvironmentSettingsEnd();
        for ( ; it != end; ++it ) {
            it->GetSceneNode()->Clear();
            it->GetSceneNode()->ResolveReference(*context);
        }
    }

    // Process to add emitter axis display
    {
        SceneRenderer::GetInstance()->DeleteDrawAxis(SceneRenderer::AxisParticleEmitter);
        nw::gfx::ParticleEmitterArray::iterator it = context->GetParticleEmitterBegin();
        nw::gfx::ParticleEmitterArray::iterator end = context->GetParticleEmitterEnd();
        for ( ; it != end; ++it ) {
            SceneRenderer::DrawAxisInfo info;
            info.axisType = SceneRenderer::AxisParticleEmitter;
            info.axisTransform = &((*it)->WorldTransform());
            info.key1 = *it;
            info.key2 = NULL;
            SceneRenderer::GetInstance()->AddDrawAxis(info);
        }
    }
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
nw::gfx::TransformNode*
SceneManager::GetSceneRoot()
{
    return m_pSceneRoot;
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
nw::gfx::SceneContext*
SceneManager::GetSceneContext()
{
    if (!m_pSceneContext) InitContext( Description() );
    return m_pSceneContext;
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
nw::gfx::ParticleContext*
SceneManager::GetParticleContext()
{
    if (!m_pParticleContext) InitContext( Description() );
    return m_pParticleContext;
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
SceneManager::AddExternalCamera(nw::gfx::Camera* camera)
{
    if (camera == NULL) return;
    CameraElem* elem = CameraElem::Builder()
                           .ResSource( NULL )
                           .ResObject( camera->GetResCamera() )
                           .SceneNode( camera )
                           .DestroyNode( false )
                           .Create( m_pAllocator, m_pDeviceAllocator );
    m_ExternalCameraArray.push_back(elem);

    // Updates the selection object array.
    SceneElemSelector::GetInstance()->Setup();
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
SceneManager::RemoveExternalCamera(nw::gfx::Camera* camera)
{
    if (camera == NULL) return;
    nw::ut::MoveArray<CameraElem*>::iterator it  = m_ExternalCameraArray.begin();
    for ( ; it != m_ExternalCameraArray.end(); ) {
        if ( (*it)->GetSceneNode() == camera ) {
            SceneRenderer::GetInstance()->DeleteDrawAxis( (*it)->GetSceneNode(), NULL );
            CameraElem* elem = (*it);
            ARRAY_ERASE( nw::ut::MoveArray<CameraElem*>, it, ExternalCamera )
            elem->Destroy();
            continue;
        }
        it++;
    }

    // Updates the selection object array.
    SceneElemSelector::GetInstance()->Setup();
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
SceneManager::AddExternalLight(nw::gfx::Light* light)
{
    if (light == NULL) return;
    nw::gfx::res::ResLight resLight =
        nw::ut::ResDynamicCast<nw::gfx::res::ResLight>( light->GetResSceneNode() );
    if (resLight.IsValid()) {
        LightElem* elem = LightElem::Builder()
                           .ResSource( NULL )
                           .ResObject( resLight )
                           .SceneNode( light )
                           .DestroyNode( false )
                           .Create( m_pAllocator, m_pDeviceAllocator );
        m_ExternalLightArray.push_back(elem);
    }

    // Updates the selection object array.
    SceneElemSelector::GetInstance()->Setup();
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
SceneManager::RemoveExternalLight(nw::gfx::Light* light)
{
    if (light == NULL) return;
    nw::ut::MoveArray<LightElem*>::iterator it  = m_ExternalLightArray.begin();
    for ( ; it != m_ExternalLightArray.end(); ) {
        if ( (*it)->GetSceneNode() == light ) {
            SceneRenderer::GetInstance()->DeleteDrawAxis( (*it)->GetSceneNode(), NULL );
            LightElem* elem = (*it);
            ARRAY_ERASE( nw::ut::MoveArray<LightElem*>, it, ExternalLight )
            elem->Destroy();
            continue;
        }
        it++;
    }

    // Updates the selection object array.
    SceneElemSelector::GetInstance()->Setup();
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
SceneManager::SetForceUseCameraWScale( const bool flag )
{
    CameraList::iterator it  = m_CameraArray.begin();
    CameraList::iterator end = m_CameraArray.end();
    if ( flag ) {
        for ( ; it != end; ++it ) {
            SetCameraWScale( it->GetSceneNode() );
        }
        if ( m_pDefaultCamera ) {
            SetCameraWScale( m_pDefaultCamera->GetSceneNode() );
        }
    } else {
        for ( ; it != end; ++it ) {
            it->GetSceneNode()->SetWScale( it->GetResObject().GetWScale() );
        }
        if ( m_pDefaultCamera ) {
            m_pDefaultCamera->GetSceneNode()->SetWScale(
                m_pDefaultCamera->GetResObject().GetWScale()
            );
        }
    }
    m_ForceUseCameraWScale = flag;
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
SceneManager::SceneManager()
 : m_pSceneRoot( NULL ), m_pSceneInitializer( NULL ), m_pSceneTraverser( NULL ),
   m_pSceneContext( NULL ), m_pParticleContext( NULL ),
   m_pDefaultAmbientLight( NULL ), m_pDefaultFragmentLight( NULL ),
   m_pDefaultLightSet( NULL ), m_pVoidLightSet( NULL ),
   m_pVoidFog( NULL ), m_pDefaultCamera( NULL ),
   m_ExternalCameraArray( Memory_GetAllocator() ),
   m_ExternalLightArray( Memory_GetAllocator() ),
   m_ForceUseCameraWScale( false )
{
    m_pAllocator = Memory_GetAllocator();
    m_pDeviceAllocator = Memory_GetAllocatorDevice();
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
SceneManager::~SceneManager()
{
    if( m_pDefaultCamera->GetResObject().ref().toName.to_ptr() != NULL )
    {
        m_pAllocator->Free((void*)m_pDefaultCamera->GetResObject().ref().toName.to_ptr());
    }

    nw::gfx::SafeDestroy(this->m_pSceneInitializer);
    nw::gfx::SafeDestroy(this->m_pSceneTraverser);
    nw::gfx::SafeDestroy(this->m_pSceneContext);
    nw::gfx::SafeDestroy(this->m_pParticleContext);
    nw::gfx::SafeDestroy(this->m_pDefaultCamera);
    nw::gfx::SafeDestroy(this->m_pDefaultAmbientLight);
    nw::gfx::SafeDestroy(this->m_pDefaultFragmentLight);
    nw::gfx::SafeDestroy(this->m_pDefaultLightSet);
    nw::gfx::SafeDestroy(this->m_pVoidLightSet);
    nw::gfx::SafeDestroy(this->m_pVoidFog);
    nw::gfx::SafeDestroy(this->m_pSceneRoot);

    // Delete external reference camera array
    {
        nw::ut::MoveArray<CameraElem*>::iterator it  = m_ExternalCameraArray.begin();
        for ( ; it != m_ExternalCameraArray.end(); ) {
            SceneRenderer::GetInstance()->DeleteDrawAxis( (*it)->GetSceneNode(), NULL );
            CameraElem* elem = (*it);
            ARRAY_ERASE( nw::ut::MoveArray<CameraElem*>, it, ExternalCamera )
            elem->Destroy();
        }
    }

    // Delete external reference light array
    {
        nw::ut::MoveArray<LightElem*>::iterator it  = m_ExternalLightArray.begin();
        for ( ; it != m_ExternalLightArray.end(); ) {
            SceneRenderer::GetInstance()->DeleteDrawAxis( (*it)->GetSceneNode(), NULL );
            LightElem* elem = (*it);
            ARRAY_ERASE( nw::ut::MoveArray<LightElem*>, it, ExternalLight )
            elem->Destroy();
        }
    }
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
SceneManager::Initialize()
{
    if (m_pInstance != NULL) return;

    void *buf = Memory_GetAllocator()->Alloc(sizeof(SceneManager));
    if (buf == NULL) return;

    m_pInstance = new(buf) SceneManager();
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
SceneManager::Finalize()
{
    if (m_pInstance == NULL) return;

    m_pInstance->~SceneManager();
    Memory_GetAllocator()->Free(m_pInstance);

    m_pInstance = NULL;
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
SceneManager::InitializeResource()
{
    // Scene transverser initialization
    m_pSceneTraverser = nw::gfx::SceneTraverser::Builder()
        .Create(m_pAllocator);

    // Initialize the material ID generator
    nw::gfx::IMaterialIdGenerator* materialIdGenerator =
        nw::gfx::SortingMaterialIdGenerator::Builder()
        .IsFixedSizeMemory(false)
        .Create(m_pAllocator);

    // Scene initializer initialization
    m_pSceneInitializer = nw::gfx::SceneInitializer::Builder()
        .MaterialIdGenerator(materialIdGenerator)
        .Create(m_pAllocator);

    // Generate scene root
    m_pSceneRoot = nw::gfx::TransformNode::DynamicBuilder()
        .IsFixedSizeMemory(false)
        .MaxChildren(128)
        .Create(m_pAllocator);

    // Default light generation
    AddDefaultLight();

    // Default camera generation
    AddDefaultCamera();

    // Default fog generation
    AddDefaultFog();
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
SceneManager::CheckError( nw::gfx::SceneNode* sceneNode )
{
    nw::gfx::Model* model = nw::ut::DynamicCast<nw::gfx::Model*>( sceneNode );
    if ( model ) {
        ErrorCheck::CheckModelTextureCombiner( model );
        ErrorCheck::CheckModelTextureReference( model );
        ErrorCheck::CheckModelLookupTableReference( model );
        ErrorCheck::CheckModelFogIndexBoundary( model );
        ErrorCheck::CheckModelFogIndexReference( model );
        ErrorCheck::CheckModelLightSetIndexBoundary( model );
        ErrorCheck::CheckModelLightSetIndexReference( model );
        ErrorCheck::CheckModelCameraReferenceIndexBoundary( model );
        ErrorCheck::CheckModelCameraReferenceIndexReference( model );
    }

    nw::gfx::Light* light = nw::ut::DynamicCast<nw::gfx::Light*>( sceneNode );
    if ( light ) {
        ErrorCheck::CheckLightLookupTableReference( light );
    }

    nw::gfx::Camera* camera = nw::ut::DynamicCast<nw::gfx::Camera*>( sceneNode );
    if ( camera ) {
        ErrorCheck::CheckCameraPosition( camera );
        ErrorCheck::CheckCameraNear( camera );
        ErrorCheck::CheckCameraNearFar( camera );
        ErrorCheck::CheckCameraUpVectorZero( camera );
        ErrorCheck::CheckCameraUpVectorParallel( camera );

        // Check WScale
        if ( m_ForceUseCameraWScale ) SetCameraWScale( camera );
    }
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
SceneManager::CheckError( nw::ut::MoveArray<nw::gfx::SceneNode*>& sceneNodeArray )
{
    nw::ut::MoveArray<nw::gfx::SceneNode*>::iterator it  = sceneNodeArray.begin();
    nw::ut::MoveArray<nw::gfx::SceneNode*>::iterator end = sceneNodeArray.end();
    for ( ; it != end; ++it ) { CheckError( *it ); }
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
SceneManager::SetupResource( nw::ut::MoveArray<nw::gfx::SceneNode*>& sceneNodeArray )
{
    // Returns immediately if the array is empty
    if (sceneNodeArray.size() == 0) return;

    // Reference resolution and initial settings for particle objects
    nw::gfx::ParticleUtil::SetupParticleObject( &sceneNodeArray, this->GetParticleContext() );

    // Resolve parent-child reference relationships
    nw::gfx::SceneHelper::ResolveReference( sceneNodeArray );

    // Generate sub root node
    nw::gfx::SceneNode *subRootNode = nw::gfx::TransformNode::DynamicBuilder()
        .IsFixedSizeMemory(false)
        .Create(m_pDeviceAllocator);

    // Add node to sub root node
    nw::gfx::SceneHelper::ForeachRootNodes(
        sceneNodeArray.Begin(),
        sceneNodeArray.End(),
        nw::gfx::AttachNode(subRootNode)
    );

    // Add sub root to scene root
    m_pSceneRoot->AttachChild(subRootNode);
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
SceneManager::AddDefaultLight()
{
    // Generate default ambient light
    {
        nw::gfx::AmbientLight *ambientLight
            = nw::gfx::AmbientLight::DynamicBuilder()
            .MaxCallbacks(0)
            .MaxChildren(0)
            .Create(m_pAllocator);

        nw::gfx::ResAmbientLight resAmbientLight( ambientLight->GetResAmbientLight() );
        resAmbientLight.SetLightEnabled(true);
        const char* name = "Default Ambient Light";
        resAmbientLight.ref().toName.set_ptr(AllocateAndCopyString(name, m_pAllocator, strlen(name)+1));
        m_pDefaultAmbientLight
            = LightElem::Builder()
            .ResSource( NULL )
            .ResObject( resAmbientLight )
            .SceneNode( ambientLight )
            .Create( m_pAllocator, m_pDeviceAllocator );

        // Attach to the scene root
        m_pSceneRoot->AttachChild(m_pDefaultAmbientLight->GetSceneNode());
    }

    // Generate default fragment light
    {
        nw::gfx::FragmentLight *fragment
            = nw::gfx::FragmentLight::DynamicBuilder()
            .MaxCallbacks(0)
            .MaxChildren(0)
            .Create(m_pAllocator);

        nw::gfx::ResFragmentLight resFragmentLight(fragment->GetResFragmentLight());
        resFragmentLight.SetLightEnabled(true);
        const char* name = "Default Fragment Light";
        resFragmentLight.ref().toName.set_ptr(AllocateAndCopyString(name, m_pAllocator, strlen(name)+1));

        m_pDefaultFragmentLight
            = LightElem::Builder()
            .ResSource( NULL )
            .ResObject( resFragmentLight )
            .SceneNode( fragment )
            .Create( m_pAllocator, m_pDeviceAllocator );

        // Attach to the scene root
        m_pSceneRoot->AttachChild(m_pDefaultFragmentLight->GetSceneNode());
    }

    // Generate default light set
    m_pDefaultLightSet = nw::gfx::LightSet::DynamicBuilder().Create(m_pAllocator);

    // Generate empty light set
    m_pVoidLightSet = nw::gfx::LightSet::DynamicBuilder().Create(m_pAllocator);
    m_pVoidLightSet->ClearAll();
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
SceneManager::AddDefaultCamera()
{
    nn::math::VEC3 cameraPos    = g3d::DirectionalObj::GetDirectionalInfo().Position;
    nn::math::VEC3 cameraTarget = g3d::DirectionalObj::GetDirectionalInfo().TargetPos;
    nn::math::VEC3 cameraUp     = g3d::DirectionalObj::GetDirectionalInfo().UpVector;

    // Generate view updater
    nw::gfx::LookAtTargetViewUpdater* viewUpdater;
    {
        viewUpdater = nw::gfx::LookAtTargetViewUpdater::Create(m_pAllocator);

        nw::gfx::ResLookAtTargetViewUpdater resViewUpdater =
            nw::gfx::ResStaticCast<nw::gfx::ResLookAtTargetViewUpdater>(viewUpdater->GetResource());
        resViewUpdater.SetTargetPosition(cameraTarget);
        resViewUpdater.SetUpwardVector(cameraUp);
    }

    // Generate projection updater
    nw::gfx::PerspectiveProjectionUpdater* projectionUpdater;
    {
        projectionUpdater = nw::gfx::PerspectiveProjectionUpdater::Create(m_pAllocator);

        nw::gfx::ResPerspectiveProjectionUpdater resProjectionUpdater =
            nw::gfx::ResStaticCast<nw::gfx::ResPerspectiveProjectionUpdater>(
            projectionUpdater->GetResource());
        resProjectionUpdater.SetNear(g3d::DirectionalObj::GetDirectionalInfo().Near);
        resProjectionUpdater.SetFar(g3d::DirectionalObj::GetDirectionalInfo().Far);
        resProjectionUpdater.SetFovy(g3d::DirectionalObj::GetDirectionalInfo().Fovy);

        f32 aspectRatio = 240.0f / 400.0f;
        resProjectionUpdater.SetAspectRatio(aspectRatio);
    }

    // Generate camera node
    {
        nw::gfx::Camera *camera =
            nw::gfx::Camera::DynamicBuilder()
            .MaxChildren(0)
            .MaxCallbacks(0)
            .ViewUpdater(viewUpdater)
            .ProjectionUpdater(projectionUpdater)
            .Create(m_pAllocator);

        camera->Transform().SetTranslate(cameraPos);

        const char* name = "Default Camera";
        camera->GetResCamera().ref().toName.set_ptr(AllocateAndCopyString(name, m_pAllocator, strlen(name)+1));

        m_pDefaultCamera = CameraElem::Builder()
                           .ResSource( NULL )
                           .ResObject( camera->GetResCamera() )
                           .SceneNode( camera )
                           .Create( m_pAllocator, m_pDeviceAllocator );
    }

    // Add to scene tree
    m_pSceneRoot->AttachChild( m_pDefaultCamera->GetSceneNode() );

    // Check WScale
    if ( m_ForceUseCameraWScale ) SetCameraWScale( m_pDefaultCamera->GetSceneNode() );
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
SceneManager::AddDefaultFog()
{
    // Empty fog generation
    m_pVoidFog = nw::gfx::Fog::DynamicBuilder().Create(m_pAllocator);
    m_pVoidFog->GetResFog().GetFogUpdater().SetFogUpdaterType(
        nw::gfx::res::ResFogUpdater::FOG_UPDATER_TYPE_LINEAR
    );
    m_pVoidFog->GetResFog().GetFogUpdater().SetMinFogDepth( 0.0f );
    m_pVoidFog->GetResFog().GetFogUpdater().SetMaxFogDepth( 100000.0f );
    m_pVoidFog->GetResFog().SetColor( 1.0f, 1.0f, 1.0f );

    // Add to scene tree
    m_pSceneRoot->AttachChild( m_pVoidFog );
}

}
}
#if NN_CURRENT_VERSION_NUMBER >= NN_VERSION_NUMBER(4,0,0,0)
#pragma diag_default 2530 // padding added to end of struct.
#endif
