/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/
#ifndef NW4CVIEWER_DISPLAYBUFFERSWAPPER_H_
#define NW4CVIEWER_DISPLAYBUFFERSWAPPER_H_

#include <viewer.h>

#if NN_CURRENT_VERSION_NUMBER >= NN_VERSION_NUMBER(4,0,0,0)
#pragma diag_suppress 1301 // padding inserted in struct.
#endif
namespace vw     {
namespace system {

/*
  
 */
enum BufferTransferMode
{
    TRANSFER_MODE_ANTIALIASE_NOT_USED,  //
    TRANSFER_MODE_ANTIALIASE_2x1,       //
    TRANSFER_MODE_ANTIALIASE_2x2,       //
    TRANSFER_MODE_ANTIALIASE_COUNT      //
};

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------
class DisplayBufferSwapper : public nw::gfx::GfxObject
{
private:
    NW_DISALLOW_COPY_AND_ASSIGN(DisplayBufferSwapper);

public:
    NW_UT_RUNTIME_TYPEINFO;

    //----------------------------------------
    //
    //

    //
    struct Description
    {
        ScreenKind screenKind;          //
        s32 width;                      //
        s32 height;                     //
        nw::gfx::RenderColorFormat format;  //
        BufferTransferMode transferMode;//
        nw::gfx::GraphicsMemoryArea memoryArea;  //
        bool isTransferFlipX;           //
        s32 transferOffsetX;            //
        s32 transferOffsetY;            //
        s32 displayOffsetX;             //
        s32 displayOffsetY;             //
        size_t bufferCount;             //

        //
        Description()
         : screenKind(UPPER_SCREEN),
           width(400),
           height(240),
           format(nw::gfx::RENDER_COLOR_FORMAT_RGB8),
           transferMode(TRANSFER_MODE_ANTIALIASE_NOT_USED),
           memoryArea(nw::gfx::MEMORY_AREA_FCRAM),
           isTransferFlipX(false),
           transferOffsetX(0),
           transferOffsetY(0),
           displayOffsetX(0),
           displayOffsetY(0),
           bufferCount(1)
        {}
    };

    //
    class Builder
    {
    public:
        //
        Builder& BufferDescription(const Description& description)
        {
            m_Description = description;
            return *this;
        }

        //
        Builder& ScreenKind(ScreenKind kind)
        {
            m_Description.screenKind = kind;
            return *this;
        }

        //
        Builder& BufferSize(s32 width, s32 height)
        {
            m_Description.width = width;
            m_Description.height = height;
            return *this;
        }

        //
        Builder& Width(s32 width) { m_Description.width = width; return *this; }

        //
        Builder& Height(s32 height) { m_Description.height = height; return *this; }

        //
        Builder& Format(nw::gfx::RenderColorFormat format)
        {
            m_Description.format = format;
            return *this;
        }

        //
        Builder& TransferMode(BufferTransferMode mode)
        {
            m_Description.transferMode = mode;
            return *this;
        }

        //
        Builder& MemoryArea(nw::gfx::GraphicsMemoryArea area)
        {
            m_Description.memoryArea = area;
            return *this;
        }

        //
        Builder& IsTransferFlipX(bool isTransferFlipX)
        {
            m_Description.isTransferFlipX = isTransferFlipX;
            return *this;
        }

        //
        Builder& TransferOffset(s32 x, s32 y)
        {
            m_Description.transferOffsetX = x;
            m_Description.transferOffsetY = y;
            return *this;
        }

        //
        Builder& DisplayOffset(s32 x, s32 y)
        {
            m_Description.displayOffsetX = x;
            m_Description.displayOffsetY = y;
            return *this;
        }

        //
        Builder& BufferCount(size_t count) { m_Description.bufferCount = count; return *this; }

        //---------------------------------------------------------------------------
        //
        //
        //
        //
        //
        //---------------------------------------------------------------------------
        DisplayBufferSwapper* Create(nw::os::IAllocator* allocator);

    private:
        Description m_Description;
    };

    //
    //
    //
    const Description& GetDescription() const { return m_Description; }

    //


    //---------------------------------------------------------------------------
    //
    //
    //
    //
    //
    //
    //
    //
    //
    //
    //
    //---------------------------------------------------------------------------
    void MakeTransferBufferCommand(nw::gfx::IRenderTarget* target, bool isMultiCommandList);

    //---------------------------------------------------------------------------
    //
    //---------------------------------------------------------------------------
    void ActivateBuffer();

    //----------------------------------------
    //
    //
private:
    //
    DisplayBufferSwapper(
        nw::os::IAllocator* allocator,
        const Description& description);

    DisplayBufferSwapper(nw::os::IAllocator* allocator) : GfxObject(allocator) {}

public:
    //
    virtual ~DisplayBufferSwapper()
    {
        if (m_DisplayBuffers)
        {
            nngxDeleteDisplaybuffers( m_Description.bufferCount, m_DisplayBuffers );
            nw::os::SafeFree(m_DisplayBuffers, &this->GetAllocator());
        }
    }

    //

private:
    Description m_Description;
    u32 m_CurrentDisplay;
    GLuint* m_DisplayBuffers;
};

}   /* namespace system */
}   /* namespace vw     */
#if NN_CURRENT_VERSION_NUMBER >= NN_VERSION_NUMBER(4,0,0,0)
#pragma diag_default 1301 // padding inserted in struct.
#endif

#endif // NW4CVIEWER_DISPLAYBUFFERSWAPPER_H_
