/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/
#ifndef NW4CVIEWER_RENDER_SYSTEM_H_
#define NW4CVIEWER_RENDER_SYSTEM_H_


#include <viewer.h>
#include "DisplayBufferSwapper.h"

#if NN_CURRENT_VERSION_NUMBER >= NN_VERSION_NUMBER(4,0,0,0)
#pragma diag_suppress 2530 // padding added to end of struct.
#endif
namespace vw     {
namespace system {

class SceneSystem;
class CommandListSwapper;

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------
class RenderSystem
{
public:
    //---------------------------------------------------------------------------
    //
    //---------------------------------------------------------------------------
    struct CommandDesc
    {
        size_t commandBufferSize;
        size_t commandRequestCount;
        size_t reusableCommandBufferSize;
        size_t reusableCommandRequestCount;
        size_t commandListCount;
        CommandDesc();
    };

    //---------------------------------------------------------------------------
    //
    //---------------------------------------------------------------------------
    struct ScreenDesc
    {
        DisplayBufferSwapper::Description upperScreenDescription;
        DisplayBufferSwapper::Description lowerScreenDescription;
        size_t displayBufferCount;
        ScreenDesc();
    };

    // Gets an instance.
    static RenderSystem* GetInstance();

    // Creates drawing-related classes.
    static RenderSystem* Create(nw::os::IAllocator* allocator,
                                const CommandDesc& description,
                                bool isEmbeddedViewer = false);

    // Destroys the render system.
    void Destroy();

    // Initializes the command cache.
    void InitCommandCache(const CommandDesc& description);

    // Deletes the buffer.
    void ClearBuffer(nw::ut::FloatColor clearColor, f32 clearZ);

    // Clears the Z-buffer.
    void ClearDepth( f32 clearZ );

    // Transfers the buffer.
    void TransferBuffer( s32 screenKind );

    // Reserves a buffer swap.
    void SwapBuffer( s32 screenKind );

    // Performs buffer swap and VSync wait
    void PresentBuffer(s32 screenKind);

    // Initiates a Vsync wait.
    void WaitVSync( s32 screenKind );

    // Runs a bound command list.
    void RunCommandList();

    // Swaps command list with buffer.
    void SwapCommandList();

    // Waits for the executing command list to end.
    void WaitCommandList();

    // Sets targets to be rendered.
    void SetRenderTarget(nw::gfx::IRenderTarget* renderTarget);

    // Gets the render context.
    nw::gfx::RenderContext* GetRenderContext() { return this->m_RenderContext; }

    // Gets the display buffer swapper
    DisplayBufferSwapper* GetDisplayBufferSwapper( s32 screenKind ) const;

    // Changes the display buffer
    void ChangeDisplayBuffer( const ScreenDesc& description );

    // Register a graphics memory allocator.
    void RegistGraphicsMemoryAllocator();

    // Return whether the graphics memory allocator is initialized.
    bool IsGraphicsMemoryAllocatorInitialized();

private:
    //
    RenderSystem( nw::os::IAllocator* allocator,
                  const CommandDesc& description,
                  bool isEmbeddedViewer);

    nw::os::IAllocator* m_Allocator;
    nw::gfx::RenderContext* m_RenderContext;

    DisplayBufferSwapper* m_UpperSwapper;
    DisplayBufferSwapper* m_LowerSwapper;
    DisplayBufferSwapper* m_ExtensionSwapper;

    CommandListSwapper* m_CommandListSwapper;

    bool m_IsEmbeddedViewer;
};



}   /* namespace system */
}   /* namespace vw     */
#if NN_CURRENT_VERSION_NUMBER >= NN_VERSION_NUMBER(4,0,0,0)
#pragma diag_default 2530 // padding added to end of struct.
#endif

#endif // NW4CVIEWER_RENDER_SYSTEM_H_
