/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

//#include <dolphin.h>
#include <viewer/system.h>


#if defined( USE_MESSAGE_QUEUE )

namespace vw     {
namespace system {

namespace SysAnmCtrl {

typedef struct TAnmCtrl
{
    u32       command;
    u32       target;
    u32       targetID;
    void*     arg;
    u32       trigModule;
    u32       arglen;
    TAnmCtrl* next;
} TAnmCtrl;

static void      Push( TAnmCtrl* pCtrl );
static TAnmCtrl* Pop( void );

static TAnmCtrl* sHead = NULL;
static TAnmCtrl* sTail = NULL;


/*---------------------------------------------------------------------------*
  Name:         Push

  Description:  Inserts animation control commands into the queue.

  Arguments:    pCtrl   Animation control command

  Returns:      None.
 *---------------------------------------------------------------------------*/
static void
Push( TAnmCtrl*  pCtrl )
{
    pCtrl->next = NULL;

    //BOOL enabled = OSDisableInterrupts();

    if ( sTail == NULL )
    {
        sHead = pCtrl;
        sTail = pCtrl;
    }
    else
    {
        sTail->next = pCtrl;
        sTail = pCtrl;
    }
    //(void)OSRestoreInterrupts( enabled );
}


/*---------------------------------------------------------------------------*
  Name:         Pop

  Description:  Gets animation control commands from the queue.

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
static TAnmCtrl*
Pop( void )
{
    //BOOL enabled = OSDisableInterrupts();

    TAnmCtrl* pRet = sHead;

    if ( pRet == NULL )
    {
        goto out;
    }

    sHead = pRet->next;
    if ( sHead == NULL )
    {
        sTail = NULL;
    }

out:
    //(void)OSRestoreInterrupts( enabled );
    return pRet;
}

} /* namespace SysAnmCtrl */


/*---------------------------------------------------------------------------*
  Name:         SysAnmCtrl_Send

  Description:  Issues animation control commands.

  Arguments:    command     Command
                target      Target
                targetID    Target ID
                arg         Callback argument
                trigModule  Issue module ID
                arglen      Size of callback argument

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
SysAnmCtrl_Send( u32 command,  u32 target, u32 targetID, void* arg, u32 trigModule, u32 arglen )
{
    u32 block_size = sizeof(SysAnmCtrl::TAnmCtrl) + arglen;

    SysAnmCtrl::TAnmCtrl *pCtrl = reinterpret_cast<SysAnmCtrl::TAnmCtrl*>( Memory_AllocSys( block_size ) );

    pCtrl->command    = command;
    pCtrl->target     = target;
    pCtrl->targetID   = targetID;
    pCtrl->arg        = arg;
    pCtrl->trigModule = trigModule;
    pCtrl->arglen     = arglen;

    if ( arglen > 0 )
    {
        pCtrl->arg = nw::ut::AddOffsetToPtr( pCtrl, sizeof(SysAnmCtrl::TAnmCtrl) );
        ::std::memcpy( pCtrl->arg, arg, arglen );
    }
    SysAnmCtrl::Push( pCtrl );
}


/*---------------------------------------------------------------------------*
  Name:         SysAnmCtrl_Exec

  Description:  Executes animation control commands.

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
SysAnmCtrl_Exec( void )
{
    SysAnmCtrl::TAnmCtrl *pCtrl;

    while ( true )
    {
        pCtrl = SysAnmCtrl::Pop();
        if ( pCtrl == NULL )
        {
            break;
        }

        if ( (pCtrl->trigModule == vw::system::SYSTEM_ID) ||
             (pCtrl->trigModule == vw::system::App::GetInstance().GetFocusID()) )
        {
            App::GetInstance().OnAnmCmd( pCtrl->command, pCtrl->target, pCtrl->targetID, pCtrl->arg, pCtrl->trigModule );
        }

        Memory_Free( pCtrl );
    }
}




}   /* namespace system */
}   /* namespace vw     */


#endif /* defined( USE_MESSAGE_QUEUE ) */

