/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#include <viewer/system.h>
#include <viewer/g3d/ModuleG3d.h>
#include <viewer/g2d/g2d_main.h>
#include <viewer/effect/ModuleEffect.h>
#include <viewer/sound/ModuleSound.h>
#include <viewer/template/ModuleTemplate.h>
#include "../include/system_private.h"
#include <nw/middlewareSymbol.h>



namespace vw     {
namespace system {

App*    App::m_pInstance = NULL;
bool    App::sIsEmbedded = false;
bool    App::sUseMenuFlag = true;
bool    App::sUseGraphicsFlag = true;
bool    App::sUseGraphicsMemoryAllocator = true;

void*   sHioWorkMemory = NULL;
bool    sInitSystemFlg = false;

/*---------------------------------------------------------------------------*
  Name:         App::GetModuleInstance
  Description:  Gets the instance for each module class
  Arguments:    id      Module ID to be obtained
  Returns:      Returns pointer to instance if the ID is valid
                NULL is returned for an invalid ID.
 *---------------------------------------------------------------------------*/
/* static */ Module*
App::GetModuleInstance( u32 id )
{
    switch ( id )
    {
    case G3D_ID:
        return vw::g3d::ModuleG3d::GetInstance();
    case G2D_ID:
        return vw::g2d::ModuleG2d::GetInstance();
    case EFFECT_ID:
        return vw::effect::ModuleEffect::GetInstance();
#if defined(VIEWER_SND_ENABLE)
    case SOUND_ID:
        return vw::snd::ModuleSound::GetInstance();
#endif
    case TEMPLATE_ID:
        return vw::tmpl::ModuleTemplate::GetInstance();
    }
    return NULL;
}

App& App::GetInstance()
{
    NW_ASSERT( m_pInstance != NULL );
    return *m_pInstance;
}

App::App() : mAllocator( 0 ),
             mFocusID( SYSTEM_ID ),
             mConnectFlg( 0 ),
             mChangeFocusEnable( true ),
             mIsVisibleInfo( true ),
             mAnimState( ANIM_STATE_STOP ),
             mSkipCount( 0 ),
             mIsRequestRenderScene( false ),
             mIsRenderScene( true ),
             mUpdateCount( 0 ),
             mIsUpdateScene( true ),
             mSkipTime( 0.0f )
{
}

App::~App()
{
    for ( Module* module = Module::GetFirstSetupModule(); module != NULL; )
    {
        Module* next = module->GetNextSetup();

        if ( module->Available() )
        {
            module->Destroy();
        }
        module = next;
    }
}

void
App::InitializeSystem( nw::os::IAllocator* allocator )
{
    nn::fs::Initialize();
    nn::hid::Initialize();

    // Initialize the heap
    if( allocator == NULL )
    {
        size_t mainMemorySize = 0;
        Memory_Init( mainMemorySize );
        allocator = Memory_GetAllocatorDevice( 0 );
    }

    // Mounts ROM.
    const int MAX_FILE      = 256;
    const int MAX_DIRECTORY = 16;
    s32 workingMemorySize = nn::fs::GetRomRequiredMemorySize( MAX_FILE, MAX_DIRECTORY );
    void*   workingMemory = allocator->Alloc( workingMemorySize );
    nn::fs::MountRom( MAX_FILE, MAX_DIRECTORY, workingMemory, workingMemorySize );

    // Initializes HostIO.
    sHioWorkMemory = allocator->Alloc( nn::hio::CTR::WORKMEMORY_SIZE );
    nn::hio::Initialize( sHioWorkMemory );

    sInitSystemFlg = true;
}

/*---------------------------------------------------------------------------*
  Name:         App::Init
  Description:  Initialization
  Arguments:    allocator Allocator
  Returns:      None.
 *---------------------------------------------------------------------------*/
void
App::Initialize( nw::os::IAllocator* allocator )
{
    NW_PUT_MIDDLEWARE_SYMBOL(vwr);

    if (m_pInstance != NULL) return;
    void *buf = allocator->Alloc(sizeof(App));
    if (buf == NULL) return;

    m_pInstance = new(buf) App();
    m_pInstance->mAllocator = allocator;

    system::Init();

    Video_Init( NULL );

    system::ConfigSys_Execute();

    // Frame rate setting
    Video_SetFrameRate( 60 );

    // Initialize in order of priority
    for ( Module* module = Module::GetFirstSetupModule(); module != NULL; )
    {
        Module* next = module->GetNextSetup();

        if ( module->Available() )
        {
            module->Init();
        }
        module = next;
    }

    system::Start();

    // Start process time measurement for entire system
    vw::system::PerformMgr::GetInstance()->GetAllTimer().Begin();
    vw::system::PerformMgr::GetInstance()->SetTimerStart();

    const char* default_mode = vw::system::GetConfigFile()->GetString("/system/config/default_mode");
    if ( default_mode )
    {
        if(0 == strcmp(default_mode, "3D"))
        {
            m_pInstance->SetFocusID( G3D_ID );
        }
        else if(0 == strcmp(default_mode, "EFF"))
        {
            m_pInstance->SetFocusID( EFFECT_ID );
        }
        else if(0 == strcmp(default_mode, "LYT"))
        {
            m_pInstance->SetFocusID( G2D_ID );
        }
        else if(0 == strcmp(default_mode, "SND"))
        {
            m_pInstance->SetFocusID( SOUND_ID );
        }
#if 0
        else if(0 == strcmp(default_mode, "TMP"))
        {
            this->SetFocusID( TEMPLATE_ID );
        }
#endif
        else
        {
            m_pInstance->SetFocusID( G3D_ID );
        }
    }
#if defined( NN_PCGX )
    else
    {
        m_pInstance->SetFocusID( G3D_ID );
    }
#endif


    if( !sUseMenuFlag )
    {
        // Delete unnecessary information to conserve memory
        //vw::system::ConfigFile::GetInstance()->Clear();
    }
}

/*---------------------------------------------------------------------------*
  Name:         App::Finalize
  Description:  Menu initialization routine
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------*/
void
App::InitializeMenu()
{
    Video_InitMenu();
    vw::g3d::ModuleG3d::GetInstance()->InitMenu();
}

/*---------------------------------------------------------------------------*
  Name:         App::Finalize
  Description:  Graphics initialization routine
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------*/
void
App::InitializeGraphics()
{
    Video_InitGraphics();
    vw::g3d::ModuleG3d::GetInstance()->InitGraphics();
}

/*---------------------------------------------------------------------------*
  Name:         App::Finalize
  Description:  Finalization routine
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------*/
void
App::Finalize()
{
    if( m_pInstance && m_pInstance->mAllocator )
    {
        m_pInstance->~App();
        m_pInstance->mAllocator->Free( m_pInstance );
        m_pInstance = NULL;
    }

    system::Finalize();

    if( sHioWorkMemory )
    {
        nn::hio::Finalize();
        Memory_Free( sHioWorkMemory );
    }

    if( sInitSystemFlg )
    {
        nn::hid::Finalize();
    }
}

/*---------------------------------------------------------------------------*
  Name:         App::PollStream
  Description:  Main routine pre-processing
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------*/
void
App::PollStream()
{
    system::PollStream();

    for ( Module* module = Module::GetTopModule(); module != NULL; )
    {
        Module* next = module->GetNext();
        if ( module->Available() )
        {
            module->PollStream();
        }
        module = next;
    }
}

/*---------------------------------------------------------------------------*
  Name:         App::PreMain
  Description:  Main routine pre-processing
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------*/
void
App::PreMain()
{
    // End overall process time measurement
    vw::system::PerformMgr::GetInstance()->GetAllTimer().Begin();

    // Frame rate control
    const SystemSetting* settings = System_GetSetting();
    mIsUpdateScene = (mUpdateCount % settings->updateRate) == 0;
    mUpdateCount++;
    
    mIsSoundOnlyMode = settings->isSoundOnlyMode;

    system::PreMain();
    HidRead();

    for ( Module* module = Module::GetTopModule(); module != NULL; )
    {
        Module* next = module->GetNext();
        u16     id   = module->GetId();

        if ( module->Available() )
        {
            module->PreMain( mHidStatus );
        }
        module = next;
    }

    // Start 2D render process
    vw::system::Draw2DHelper::GetInstance().Begin();
}


/*---------------------------------------------------------------------------*
  Name:         App::Main
  Description:  Main routine
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------*/
void
App::Main()
{
    system::Main();

    if (mIsUpdateScene && mIsRenderScene && !sIsEmbedded)
    {
        system::DrawBG();
    }

    for ( Module* module = Module::GetTopModule(); module != NULL; )
    {
        Module* next = module->GetNext();
        u16     id   = module->GetId();
        if ( module->Available() )
        {
            module->Main( mHidStatus );
        }
        module = next;
    }
}


/*---------------------------------------------------------------------------*
  Name:         App::UpdateFrame
  Description:  Frame update.
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
App::UpdateFrame()
{
    SysAnmCtrl_Exec();

    for ( Module* module = Module::GetTopModule(); module != NULL; )
    {
        Module* next = module->GetNext();

        if ( module->Available() )
        {
            module->UpdateFrame();
        }
        module = next;
    }
}


/*---------------------------------------------------------------------------*
  Name:         App::CalcAndDraw
  Description:  Execute Calc and Draw alternately for each module
  Arguments:    sync
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
App::CalcAndDraw( bool sync /* = true */ )
{
    for ( Module* module = Module::GetTopModule(); module != NULL; )
    {
        Module* next = module->GetNext();
        u16     id   = module->GetId();

        if ( module->Available() )
        {
            module->Calc();
            module->Draw();
            module->RisePostCalcCallback();
            module->RisePostDrawCallback();
        }
        module = next;
    }
}


/*---------------------------------------------------------------------------*
  Name:         App::Calc
  Description:  Frame update routine
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
App::Calc()
{
    for ( Module* module = Module::GetTopModule(); module != NULL; )
    {
        Module* next = module->GetNext();
        if ( module->Available() )
        {
            module->Calc();

            module->RisePostCalcCallback();
        }
        module = next;
    }
}


/*---------------------------------------------------------------------------*
  Name:         App::Draw
  Description:  Render all again.
  Arguments:    sync
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
App::Draw( bool sync /* = true */ )
{
    for ( Module* module = Module::GetTopModule(); module != NULL; )
    {
        Module* next = module->GetNext();
        if ( module->Available() )
        {
            module->Draw();

            module->RisePostDrawCallback();
        }
        module = next;
    }
}

/*---------------------------------------------------------------------------*
  Name:         App::DrawInfo
  Description:  information display
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------*/
void
App::DrawInfo()
{
    vw::system::PerformMgr::GetInstance()->EndFrame();

    bool lowerPreview = System_GetSetting()->lowerPreview;
    if( lowerPreview )
    {
        vw::system::Video_SetRenderTarget(vw::system::UPPER_SCREEN);
    }
    else
    {
        vw::system::Video_SetRenderTarget(vw::system::LOWER_SCREEN);
    }

    if ( this->IsVisibleInfo() )
    {
        // Wait for the model type to complete rendering before information display then output the GPU process time result
        for ( Module* module = Module::GetTopModule(); module != NULL; )
        {
            Module* next = module->GetNext();

            if ( HasFocus( module->GetId() ) )
            {
                module->DrawInfo();
            }
            module = next;
        }
    }

    vw::system::Draw2DHelper::GetInstance().End();

#if defined( NN_PCGX )
    nngxSplitDrawCmdlist();
#endif
}


/*---------------------------------------------------------------------------*
  Name:         App::WaitNextFrame
  Description:  Wait for the next frame.
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------*/
void
App::WaitNextFrame()
{
    Video_TransferBuffer();
    Video_WaitCommandList();

    // End process time measurement for entire system
    vw::system::PerformMgr::GetInstance()->SetTimerEnd();

#if defined( NN_PCGX )
    vw::system::PerformMgr::GetInstance()->SetTimerStart();
    mIsRenderScene = true;
#else
    // Frame skip process
    bool frameSkip = System_GetSetting()->frameSkip;
    if ( frameSkip )
    {
        const f32 frameDuration = 1000.0f / 60.0f;
        f32 processTime = vw::system::PerformMgr::GetInstance()->GetSystemTime();

        if( mIsRenderScene )
        {
            Video_SwapBuffer();

            // Do not wait for VSync if does not fit in one frame
            if( processTime < frameDuration )
            {
                Video_WaitVSync();
            }
        }

        if( mSkipTime > 0.0f )
        {
            mSkipTime -= frameDuration - processTime;

            mIsRenderScene = false;

            mSkipCount++;
            if( mSkipCount >= System_GetSetting()->maxFrameSkipCount )
            {
                mSkipCount = 0;
                mSkipTime = 0.0f;
            }
            if( mSkipTime <= 0.0f )
            {
                Video_WaitVSync();

                mSkipTime = 0.0f;
                mSkipCount = 0;
                mIsRenderScene = true;
            }
        }
        else
        {
            f32 over_time = processTime - frameDuration;
            if( over_time > 0.0f )
            {
                mIsRenderScene = false;
                mSkipTime += over_time;
            }
            else
            {
                mIsRenderScene = true;
            }
        }

        // 1 frame time measurement
        vw::system::PerformMgr::GetInstance()->SetTimerStart();
    }
    else
    {
        Video_PresentBuffer();
        // 1 frame time measurement
        vw::system::PerformMgr::GetInstance()->SetTimerStart();
        mIsRenderScene = true;
    }
#endif
    // capture
    vw::system::CaptureInterface::Main();
}


/*---------------------------------------------------------------------------*
  Name:         App::OnAnmCmd
  Description:  Calls enter playback command for all animations for each module
  Arguments:    command     Command (RESET/PLAY/REV_PLAY/STOP/STEP/REV_STEP/LOOP/SET_FRAME)
                target      Target (ALL/3D_MDL/2D_LYT)
                arg         Option parameter
                trigModule  Module that is the trigger
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
App::OnAnmCmd( u32 command, u32 target, u32 targetID, void*  arg, u32  trigModule )
{
    for ( Module* module = Module::GetTopModule(); module != NULL; )
    {
        Module* next = module->GetNext();
        if ( module->Available() )
        {
            module->OnAnmCmd( command, target, targetID, arg, trigModule );
        }
        module = next;
    }
}


/*---------------------------------------------------------------------------*
  Name:         App::SetFocusID
  Description:  Changes the focus.
  Arguments:    focus
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
App::SetFocusID( u16 focus )
{
#if defined( NN_PCGX )
    if(focus == G2D_ID)   { focus = G3D_ID; }
    if(focus == SOUND_ID) { focus = G3D_ID; }
#endif
    if ( mFocusID == focus ) return;

    Module* oldModule = GetModuleInstance( mFocusID );
    if( oldModule != NULL )
    {
        oldModule->OnModuleUnfocused();
    }

    u16 before = mFocusID;
    mFocusID = focus;

    // Process during focus
    // Temporary process until OnAnmCmd is implemented
    Module* module = GetModuleInstance( mFocusID );
    module->OnModuleFocused();

/*    if ( before != SYSTEM_ID )
    {
        this->OnAnmCmd( ANM_CMD_RESET, ANM_TGT_ALL, 0, NULL, SYSTEM_ID );
    }*/
}


namespace {
    //---------------------------------------------------------------------------
    //  Description:  Updates touch panel click information.
    //---------------------------------------------------------------------------
    void UpdateClickState( HidStatus* status, bool lastTouch )
    {
        if (lastTouch != status->touchPanel.touch)
        {
            if (status->touchPanel.touch)
            {
                // The moment the touch panel is touched.
                status->isClicking = true;
                status->clickPosX = status->touchPanel.x;
                status->clickPosY = status->touchPanel.y;
            }
            else
            {
                // The moment the touch panel is released.
                if (status->isClicking)
                {
                    status->isClicking = false;
                    status->isClicked  = true;
                }
            }
        }
        else
        {
            if (status->touchPanel.touch)
            {
                const u32 CLICK_MARGIN = 5;

                if ( nw::ut::Abs( status->touchPanel.x - status->clickPosX ) > CLICK_MARGIN ||
                     nw::ut::Abs( status->touchPanel.y - status->clickPosY ) > CLICK_MARGIN )
                {
                    status->isClicking = false;
                }
            }
            else
            {
                status->isClicked = false;
            }
        }
    }
}


void
App::HidRead()
{
    static nn::hid::CTR::PadReader           padReader;
    static nn::hid::CTR::AccelerometerReader accelerometerReader;
    static nn::hid::CTR::DebugPadReader      debugPadReader;
    static nn::hid::CTR::TouchPanelReader    touchPanelReader;

    bool lastTouch = mHidStatus.touchPanel.touch;

    padReader.SetStickClampMode(nn::hid::CTR::PadReader::STICK_CLAMP_MODE_CROSS);
    if(!padReader.ReadLatest(&mHidStatus.pad))                     memset(&mHidStatus.pad,           0, sizeof(mHidStatus.pad));
    if(!accelerometerReader.ReadLatest(&mHidStatus.accelerometer)) memset(&mHidStatus.accelerometer, 0, sizeof(mHidStatus.accelerometer));
    if(!debugPadReader.ReadLatest(&mHidStatus.debugPad))           memset(&mHidStatus.debugPad,      0, sizeof(mHidStatus.debugPad));
    if(!touchPanelReader.ReadLatest(&mHidStatus.touchPanel))       memset(&mHidStatus.touchPanel,    0, sizeof(mHidStatus.touchPanel));

    UpdateClickState( &mHidStatus, lastTouch );

    mHidStatus.padReader = &padReader;
}



}   /* namespace system   */
}   /* namespace vw */

