/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#include "../include/BmpReaderImpl.h"
#include "../include/BitmapData.h"
#include "../include/TexUtil.h"

namespace vw     {
namespace system {

/*---------------------------------------------------------------------------*
  Name:         BmpReaderImpl::BmpReaderImpl
  Description:  Constructor
  Arguments:    data
  Returns:      None.
 *---------------------------------------------------------------------------


*/
/* ctor */
BmpReaderImpl::BmpReaderImpl( void* data )
 : ImageReaderImpl(),
   mBuf( static_cast<u8*>( data ) ),
   mBody( NULL ),
   mPltt( NULL ),
   mWidth( 0 ),
   mHeight( 0 ),
   mIsValid( true ),
   mBpp( 0 ),
   mCompressType( COMP_TYPE_NONE ),
   mPlttBpp( 0 )
{
    u32 COMP_TYPE_TABLE[ 4 ] =
    {
        COMP_TYPE_NONE,
        COMP_TYPE_RL8,
        COMP_TYPE_RL4,
        COMP_TYPE_BITFIELDS,
    };

    NW_NULL_ASSERT( data );

    mIsValid = this->CheckValid_();
    if ( ! mIsValid )
    {
   //     NW_WARNING( false, RESTXT_ON_UNSUPPORTED_IMAGE );
        return;
    }

    BitmapData::Header *hd = reinterpret_cast<BitmapData::Header*>( mBuf );

    u32 offset    = LEndian::ReadU32( &hd->bfOffBits );

    mBody = static_cast<u8*>( nw::ut::AddOffsetToPtr( mBuf, offset ) );

    u32 headerSize = LEndian::ReadU32( &hd->bcSize    );

    if ( headerSize == 12 )
    // OS/2 type
    {
        mWidth  = LEndian::ReadU16( &hd->os2.bcWidth    );
        mHeight = LEndian::ReadU16( &hd->os2.bcHeight   );
        mBpp    = LEndian::ReadU16( &hd->os2.bcBitCount );
        mPlttBpp = 3;
    }
    else
    // Windows type
    {
        mWidth  = LEndian::ReadU32( &hd->bcWidth    );
        mHeight = LEndian::ReadU32( &hd->bcHeight   );
        mBpp    = LEndian::ReadU16( &hd->bcBitCount );
        u32 compType = LEndian::ReadU32( &hd->compMode );
        mCompressType = COMP_TYPE_TABLE[ compType ];
        mPlttBpp = 4;
    }

    if ( mWidth > 1024 || mHeight > 1024 )
    {
        mIsValid = false;
   //     NW_WARNING( false, RESTXT_ON_TOO_LARGE_IMAGE_dd, mWidth, mHeight );
    }

    if ( mBpp == 1 || mBpp == 4 || mBpp == 8 ||
         mCompressType == COMP_TYPE_BITFIELDS )
    {
        mPltt = static_cast<u8*>( nw::ut::AddOffsetToPtr( mBuf, 14 + headerSize ) );
    }

    if ( mCompressType == COMP_TYPE_RL8 )
    {
        if ( mBpp != 8 )
        {
            mIsValid = false;
        }
    }
    else if ( mCompressType == COMP_TYPE_RL4 )
    {
        if ( mBpp != 4 )
        {
            mIsValid = false;
        }
    }
    else if ( mCompressType == COMP_TYPE_BITFIELDS )
    {
        if ( mBpp != 24 && mBpp != 32 )
        {
            mIsValid = false;
        }
    }

    if ( mCompressType == COMP_TYPE_RL4 || mCompressType == COMP_TYPE_BITFIELDS )
    // Currently, no support for RL4 and BITFIELDS
    {
        mIsValid = false;
    }
}


/*---------------------------------------------------------------------------*
  Name:         BmpReaderImpl::~BmpReaderImpl
  Description:  Destructor
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
/* dtor */
BmpReaderImpl::~BmpReaderImpl()
{
    this->Dispose();
}


/*---------------------------------------------------------------------------*
  Name:         BmpReaderImpl::IsValid
  Description:  Determines validity of file
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
bool
BmpReaderImpl::IsValid()
{
    return mIsValid;
}


/*---------------------------------------------------------------------------*
  Name:         BmpReaderImpl::GetWidth
  Description:  Gets the horizontal size of a BMP file.
  Arguments:    None.
  Returns:      Horizontal size
 *---------------------------------------------------------------------------


*/
u32
BmpReaderImpl::GetWidth()
{
    return mWidth;
}


/*---------------------------------------------------------------------------*
  Name:         BmpReaderImpl::GetHeight
  Description:  Gets the vertical size of a BMP file.
  Arguments:    None.
  Returns:      Vertical size
 *---------------------------------------------------------------------------


*/
u32
BmpReaderImpl::GetHeight()
{
    return mHeight;
}

/*---------------------------------------------------------------------------*
  Name:         BmpReaderImpl::GetType
  Description:  Gets the image type.
  Arguments:    None.
  Returns:      Image type
 *---------------------------------------------------------------------------


*/
u32
BmpReaderImpl::GetType()
{
    return TYPE_BMP;
}


/*---------------------------------------------------------------------------*
  Name:         BmpReaderImpl::Expand
  Description:  Expands a BMP file to RGB8 texture format.
  Arguments:    expBuf      Buffer to expand to
                bufWidth    Buffer horizontal size
                bufHeight   Buffer vertical size
  Returns:      If expanded normally, TRUE;
                if cannot be expanded, FALSE
 *---------------------------------------------------------------------------


*/
bool
BmpReaderImpl::Expand( void* expBuf, u32 bufWidth, u32 bufHeight )
{
//#pragma unused( bufHeight )
    u8* body = mBody;
    u8* pOut = static_cast<u8*>( expBuf );
    u32 bpp = mBpp / 8;
    u32 cidx;
    u32 shift = 0;
    u32 mask  = 0xFF;
    u8* uncompBuf = NULL;

    if ( mCompressType != COMP_TYPE_NONE )
    {
        uncompBuf = static_cast<u8*>( Memory_AllocSys( nn::math::RoundUp(mWidth, 2) * nn::math::RoundUp(mHeight, 2) * mBpp / 8 ) );

        this->Uncomp_( uncompBuf );
        body = uncompBuf;
    }

    if ( mBpp < 8 )
    {
        mask = u32( (~(0x100 - (1 << mBpp))) & 0xFF );
    }

    for ( u16 y = 0; y < mHeight; y++ )
    {
        for ( u16 x = 0; x < mWidth; x++ )
        {
            u8 a, b, g, r;
            if ( mPltt )
            {
                cidx = ( *body >> (8 - mBpp - shift ) ) & mask;
                u8* pCol = &mPltt[ cidx * mPlttBpp ];
                a = 0xFF;  // With BMP, alpha is always opaque
                b = pCol[ 0 ];
                g = pCol[ 1 ];
                r = pCol[ 2 ];
                shift += mBpp;
                if ( shift >= 8 )
                {
                    shift -= 8;
                    ++body;
                }
            }
            else
            {
                a = 0xFF;
                b = *body++;
                g = *body++;
                r = *body++;
                if ( bpp == 4 )
                {
                    a = *body++;
                }
            }

            u16 line = u16(mHeight - 1 - y);

            nw::ut::Color8 color(r,g,b,a);
            UtilRGBA8::SetColor( color, pOut, x, line, bufWidth, bufHeight);
        }
        // Align the size of 1 line to a multiple of 4
        {
            u32 line_size = mBpp * mWidth / 8;
            if ( mBpp * mWidth % 8 )
            {
                shift = 0;
                ++line_size;
                ++body;
            }

            body += ((4 - (line_size & 3)) & 3);
        }
    }

    if ( uncompBuf )
    {
        Memory_Free( uncompBuf );
    }

    return true;
}


/*---------------------------------------------------------------------------*
  Name:         BmpReaderImpl::CheckValid_
  Description:  Validity check
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
bool
BmpReaderImpl::CheckValid_()
{
    BitmapData::Header *hd = reinterpret_cast<BitmapData::Header*>( mBuf );

    if ( hd->bfType[0] != 'B' || hd->bfType[1] != 'M' )
    {
        return false;
    }

    u32 offset = LEndian::ReadU32( &hd->bfOffBits  );

    u32 headerSize = LEndian::ReadU32( &hd->bcSize );
    if ( headerSize == 12 )
    // OS/2 type
    {
        if ( offset < 26 )
        {
            return false;
        }

        u32 width = LEndian::ReadU16( &hd->os2.bcWidth  );
        if ( width == 0 )
        {
            return false;
        }

        u32 height = LEndian::ReadU32( &hd->os2.bcHeight );
        if ( height == 0 )
        {
            return false;
        }

        if ( LEndian::ReadU16( &hd->os2.bcPlanes ) != 1 )
        {
            return false;
        }

        u16 bitCount = LEndian::ReadU16( &hd->os2.bcBitCount );
        if ( (bitCount != 1) && (bitCount != 4) && (bitCount != 8) &&
             (bitCount !=16) && (bitCount != 24) && (bitCount != 32) )
        {
            return false;
        }
    }
    else if ( headerSize == 40 )
    // Windows type
    {
        if ( offset < 54 )
        {
            return false;
        }

        u32 width = LEndian::ReadU32( &hd->bcWidth  );
        if ( width == 0 )
        {
            return false;
        }

        u32 height = LEndian::ReadU32( &hd->bcHeight );
        if ( height == 0 )
        {
            return false;
        }

        if ( LEndian::ReadU16( &hd->bcPlanes ) != 1 )
        {
            return false;
        }

        u16 bitCount = LEndian::ReadU16( &hd->bcBitCount );
        if ( (bitCount != 1) && (bitCount != 4) && (bitCount != 8) &&
             (bitCount != 24) && (bitCount != 32) )
        {
            return false;
        }

        if ( LEndian::ReadU32( &hd->compMode ) >= 4 )
        {
            return false;
        }
    }
    else
    // Header size is invalid
    {
        return false;
    }

    // Check was OK
    return true;
}


/*---------------------------------------------------------------------------*
  Name:         BmpReaderImpl::Uncomp_
  Description:  Expands BMP compression format
  Arguments:    pOut    Buffer to expand to
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
BmpReaderImpl::Uncomp_( u8* pOut )
{
    if ( mCompressType == COMP_TYPE_RL8 )
    {
        this->UncompRL8_( pOut );
    }
    else if ( mCompressType == COMP_TYPE_RL4 )
    {
        this->UncompRL4_( pOut );
    }
    else if ( mCompressType == COMP_TYPE_BITFIELDS )
    {
        this->UncompBF_( pOut );
    }
}


/*---------------------------------------------------------------------------*
  Name:         BmpReaderImpl::UncompRL8_
  Description:  Expands RL8 compressed BMP
  Arguments:    pOut    Buffer to expand to
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
BmpReaderImpl::UncompRL8_( u8* pOut )
{
    u8* pIn = mBody;
    u32 len;
    u32 size = mHeight * mWidth;
    u32 count = 0;

    while ( count < size )
    {
        len = *pIn++;
        if ( len > 0 )
        {
            while ( len > 0 )
            {
                *pOut++ = *pIn;
                --len;
                ++count;
            }
            ++pIn;
        }
        else
        {
            len = *pIn++;

            if ( len >= 3 )
            {
                u8 padding = u8( len & 1 );
                while ( len > 0 )
                {
                    *pOut++ = *pIn++;
                    --len;
                    ++count;
                }
                if ( padding )
                {
                    ++pIn;
                }
            }
            else if ( len == 2 )
            {
                s8 hOfs = s8( *pIn++ );
                s8 vOfs = s8( *pIn++ );

                pOut += (hOfs + vOfs * mWidth);
                count += (hOfs + vOfs * mWidth);
            }
            else if ( len == 1 )
            {
                while ( count < size )
                {
                    *pOut++ = 0x00;
                    ++count;
                }
                return;
            }
            else if ( len == 0 )
            {
                while ( (count % mWidth) != 0 )
                {
                    *pOut++ = 0x00;
                    ++count;
                }
            }
        }
    }
}



/*---------------------------------------------------------------------------*
  Name:         BmpReaderImpl::UncompRL4_
  Description:  Expands RL4 compressed BMP
  Arguments:    pOut    Buffer to expand to
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
BmpReaderImpl::UncompRL4_( u8* pOut )
{
//#pragma unused( pOut )
#if 0
    u8* pIn = mBody;
    u32 len;
    u32 size = mHeight * mWidth;
    u32 count = 0;
    u32 shift = 0;
    u32 outShift = 0;

    while ( count < size )
    {
        len = *pIn++;
        if ( len > 0 )
        {
            shift = 0;
            while ( len > 0 )
            {
                *pOut = (*pOut & (0x0F << shift)) | (((*pIn >> (8 - shift)) & 0x0F) << (8 - outShift));
                shift = (shift + 4) % 8;
                outShift += 4;

                if ( outShift == 8 )
                {
                    outShift = 0;
                    ++pOut;
                }
                --len;
                ++count;
            }
            ++pIn;
        }
        else
        {
            len = *pIn++;

            if ( len >= 3 )
            {
                u8 padding = u8( len & 4 );
                shift = 0;
                while ( len > 0 )
                {
                    *pOut = (*pOut & (0x0F << shift)) | (((*pIn >> (8 - shift)) & 0x0F) << (8 - outShift));
                    shift    += 4;
                    outShift += 4;
                    if ( outShift == 8 )
                    {
                        outShift = 0;
                        ++pOut;
                    }
                    if ( shift == 8 )
                    {
                        shift = 0;
                        ++pIn;
                    }
                    --len;
                    ++count;
                }
                if ( padding > 0 )
                {
                    pIn += ((padding + 1) / 2);
                }
            }
            else if ( len == 2 )
            {
                s8 hOfs = s8( *pIn++ );
                s8 vOfs = s8( *pIn++ );

                s32 offset = hOfs + vOfs * mWidth;
                pOut  += (offset / 2);
                count += offset;
            }
            else if ( len == 1 )
            {
                while ( count < size )
                {
                    *pOut = (*pOut & (0x0F << outShift)) | (((*pIn >> (8 - shift)) & 0x0F) << (8 - outShift));

                    ++count;
                }
                return;
            }
            else if ( len == 0 )
            {
                while ( (count % mWidth) != 0 )
                {
                    *pOut++ = 0x00;
                    ++count;
                }
            }
        }
    }
#endif
}


/*---------------------------------------------------------------------------*
  Name:         BmpReaderImpl::UncompBF_
  Description:  Expands Bitfield compressed BMP
  Arguments:    pOut    Buffer to expand to
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
BmpReaderImpl::UncompBF_( u8* pOut )
{
//#pragma unused( pOut )

}

/*---------------------------------------------------------------------------*
  Name:         BmpReaderImpl::Dispose()
  Description:  Free allocated memory for internal processing.
  Arguments:    data
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
BmpReaderImpl::Dispose()
{
    /* DO_NOTHING */
}



}   /* namespace system */
}   /* namespace vw     */


