/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#include <viewer/system.h>
#include "../include/LogSystem.h"
#include "../../g3d/include/ModuleG3dImpl.h"
#include <viewer/g3d/CameraMgr.h>


namespace vw     {
namespace system {



static bool sCapture = false;
static bool sExecuting = false;
static u32 sMagnify = 1;
static s32 sCaptureCount = 0;
static vw::common::CharArray sDirectory;
static vw::common::CharArray sImagePath;



static void storeRGB8(u8* dst, const u8* __restrict src, u32 width, u32 height)
{
    u32 i=0, j=0;
    for(u32 x=0; x<width; ++x)
    {
        dst[j  ] = src[i+0];
        dst[j+1] = src[i+1];
        dst[j+2] = src[i+2];
        j += 3;
        i += 3 * height;
    }
}

static void storeRGBA4(u8* dst, const u16* __restrict src, u32 width, u32 height)
{
    u32 i=0, j=0;
    for(u32 x=0; x<width; ++x)
    {
        dst[j  ] = ((src[i] & 0x00F0) >>  4) << 4;
        dst[j+1] = ((src[i] & 0x0F00) >>  8) << 4;
        dst[j+2] = ((src[i] & 0xF000) >> 12) << 4;
        j += 3;
        i += height;
    }
}

static void storeRGB5_A1(u8* dst, const u16* __restrict src, u32 width, u32 height)
{
    u32 i=0, j=0;
    for(u32 x=0; x<width; ++x)
    {
        dst[j  ] = ((src[i] & 0x003E) >>  1) << 3;
        dst[j+1] = ((src[i] & 0x07C0) >>  6) << 3;
        dst[j+2] = ((src[i] & 0xF800) >> 11) << 3;
        j += 3;
        i += height;
    }
}

static void storeRGB565(u8* dst, const u16* __restrict src, u32 width, u32 height)
{
    u32 i=0, j=0;
    for(u32 x=0; x<width; ++x)
    {
        dst[j  ] = ((src[i] & 0x001F) >>  0) << 3;
        dst[j+1] = ((src[i] & 0x07E0) >>  5) << 2;
        dst[j+2] = ((src[i] & 0xF800) >> 11) << 3;
        j += 3;
        i += height;
    }
}

static void storeRect(u8* dst,
                      const u8* __restrict src,
                      u32 x, u32 y,
                      u32 dst_width,
                      u32 dst_height,
                      u32 src_width,
                      u32 src_height,
                      GLint format)
{
    const u32 width = (dst_height * src_width);
    const u32 offset = (src_height * x + width * y) * 3;

    for(u32 i=0; i<src_width; ++i)
    {
        u8* work = dst + i*dst_height*3 + offset;

        switch(format)
        {
        case GL_RGB8_OES:
            storeRGB8(work, src+i*3, src_height, src_width);
            break;
        case GL_RGBA4:
            storeRGBA4(work, (u16*)(src+i*2), src_height, src_width);
            break;
        case GL_RGB5_A1:
            storeRGB5_A1(work, (u16*)(src+i*2), src_height, src_width);
            break;
        case GL_RGB565:
            storeRGB565(work, (u16*)(src+i*2), src_height, src_width);
            break;
        }
    }
}

static void GetProjectionParameters(nn::math::Matrix44* proj, f32* left, f32* right,
                                    f32* bottom, f32* top, f32* near, f32* far)
{
    // Reverse calculate, assuming that the perspective projection matrix that was input takes CTR clip coordinates into account.
    *near   = proj->m[2][3] / proj->m[2][2];
    *far    = proj->m[2][3] / (proj->m[2][2] - 1.0f);

    f32 tmp = proj->m[2][3] / (proj->m[0][0] * proj->m[2][2]);
    *left   = tmp * (proj->m[0][2] - 1.0f);
    *right  = tmp * (proj->m[0][2] + 1.0f);

    tmp     = proj->m[2][3] / (proj->m[1][1] * proj->m[2][2]);
    *top    = tmp * (proj->m[1][2] + 1.0f);
    *bottom = tmp * (proj->m[1][2] - 1.0f);
}

static void GetProjectionForPartialCapture(nn::math::MTX44* pOut,
                                           nn::math::Matrix44* proj,
                                           s32 div,
                                           s32 indexH,
                                           s32 indexV,
                                           nn::math::PivotDirection pivot)
{
    NN_NULL_ASSERT(pOut);
    NN_NULL_ASSERT(proj);
    NN_ASSERT( 0 <= indexV || 0 <= indexH || indexV < div || indexH < div);
    NN_ASSERT( nn::math::PIVOT_NONE <= pivot || pivot <= nn::math::PIVOT_NUM );

    const f32 divH = static_cast<f32>(div);
    const f32 divV = static_cast<f32>(div);

    f32 near, far, top, bottom, left, right;
    GetProjectionParameters( proj, &left, &right, &bottom, &top, &near, &far);

    f32 partLeft, partRight, partTop, partBottom;

    partLeft   = (static_cast<f32>(indexH)/divH * right + (divH - static_cast<f32>(indexH))/divH * left);
    partRight  = (static_cast<f32>(indexH + 1)/divH * right + (divH - static_cast<f32>(indexH + 1))/divH * left);
    partTop    = (static_cast<f32>(indexV)/divV * bottom + (divV - static_cast<f32>(indexV))/divV * top);
    partBottom = (static_cast<f32>(indexV + 1)/divV * bottom + (divV - static_cast<f32>(indexV + 1))/divV * top);

    nn::math::MTX44FrustumPivot(pOut, partLeft, partRight, partBottom, partTop, near, far, pivot);
}

static void GetOrthoForPartialCapture(nn::math::MTX44* pOut,
                                      nw::gfx::ResOrthoProjectionUpdater resOrthoUpdater,
                                      s32 div,
                                      s32 indexV,
                                      s32 indexH,
                                      nn::math::PivotDirection pivot)
{
    nw::gfx::ResProjectionRect rect = resOrthoUpdater.GetRect();

    f32 halfWidth = rect.GetWidth() / 2.0f;
    f32 halfHeight = rect.GetHeight() / 2.0f;

    f32 left = rect.GetCenter().x - halfWidth;
    f32 right = rect.GetCenter().x + halfWidth;
    f32 bottom = rect.GetCenter().y - halfHeight;
    f32 top = rect.GetCenter().y + halfHeight;

    f32 step_x = (right - left) / div;
    f32 step_y = (top - bottom) / div;

    s32 x = indexV;
    s32 y = div-indexH-1;
    nn::math::MTX44OrthoPivot(
        pOut,
        left + step_x* x,
        left + step_x*(x+1),
        bottom + step_y* y,
        bottom + step_y*(y+1),
        resOrthoUpdater.GetNear(),
        resOrthoUpdater.GetFar(),
        pivot);
}

static GLenum GetDisplayFormat()
{
    GLenum format;

    vw::system::VideoSetting setting;
    vw::system::Video_GetSetting( &setting );
    switch( setting.colorFormat )
    {
    case nw::gfx::RENDER_COLOR_FORMAT_RGBA8:
    case nw::gfx::RENDER_COLOR_FORMAT_RGB8:
        format = GL_RGB8_OES;
        break;
    case nw::gfx::RENDER_COLOR_FORMAT_RGBA4:
        format = GL_RGBA4;
        break;
    case nw::gfx::RENDER_COLOR_FORMAT_RGB5_A1:
        format = GL_RGB5_A1;
        break;
    case nw::gfx::RENDER_COLOR_FORMAT_RGB565:
        format = GL_RGB565;
        break;
    }

    return format;
}

static void DrawScene(GLuint disp, u32 x, u32 y, u32 div)
{
    nw::math::MTX44* pOutProj = &vw::g3d::CameraMgr::GetInstance()->GetLeftViewCamera()->ProjectionMatrix();
    nw::gfx::Camera* camera = vw::g3d::CameraMgr::GetInstance()->GetTargetCamera();
    nw::gfx::ResCameraProjectionUpdater resProjUpdater  = camera->GetProjectionUpdater()->GetResource();
    nw::gfx::ResOrthoProjectionUpdater  resOrthoUpdater = nw::gfx::ResDynamicCast<nw::gfx::ResOrthoProjectionUpdater>( resProjUpdater );

    if( resOrthoUpdater.IsValid() )
    {
        GetOrthoForPartialCapture(pOutProj, resOrthoUpdater, div, x, div-y-1, nn::math::PIVOT_UPSIDE_TO_TOP);
    }
    else
    {
        GetProjectionForPartialCapture(pOutProj, &camera->ProjectionMatrix(), div, x, div-y-1, nn::math::PIVOT_UPSIDE_TO_TOP);
    }

    sExecuting = true;
    vw::system::Video_SetRenderTarget(UPPER_SCREEN);
    vw::system::Video_ClearBuffer();
    vw::system::Draw2DHelper::GetInstance().Begin();
    vw::g3d::ModuleG3dImpl::GetInstance()->Draw();
    vw::system::Draw2DHelper::GetInstance().End();
    sExecuting = false;

    nngxTransferRenderImage(disp, NN_GX_ANTIALIASE_NOT_USED, GL_FALSE, 0, 0);

    nngxSplitDrawCmdlist();
    nngxRunCmdlist();
    nngxWaitCmdlistDone();
    nngxClearCmdlist();
    nngxStopCmdlist();

    vw::system::CommandListSwapper::GetInstance()->ResetGpuProfiling();
}

static bool writeBmpHeader(nn::hio::CTR::HostFile* file, u32 width, u32 height)
{
    static u8 bmp_header[54] =
    {
        0x42,0x4D,0x04,0x03,0x02,0x00,0x00,0x00,
        0x00,0x00,0x36,0x00,0x00,0x00,0x28,0x00,
        0x00,0x00,0x80,0x02,0x00,0x00,0xE0,0x01,
        0x00,0x00,0x01,0x00,0x18,0x00,0x00,0x00,
        0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
        0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
        0x00,0x00,0x00,0x00,0x00,0x00
    };

    u32 xpad = (4 - ((width * 3) & 3)) & 3;
    u32 i = 54 + width * height * 3 + xpad * height;
    bmp_header[ 2] = static_cast<u8>(i);
    bmp_header[ 3] = static_cast<u8>(i >> 8);
    bmp_header[ 4] = static_cast<u8>(i >> 16) & 255;
    bmp_header[18] = static_cast<u8>(height);
    bmp_header[19] = static_cast<u8>(height >> 8);
    bmp_header[22] = static_cast<u8>(width);
    bmp_header[23] = static_cast<u8>(width >> 8);

    size_t size = 0;
    nn::Result result = file->Write(&size,(char*)bmp_header,54);

    return result.IsSuccess();
}

static bool GetFilePath(vw::common::CharArray* filePath)
{
    if( filePath == 0 )
        return false;

    SystemSetting setting;
    System_GetSetting(&setting);
    if ( !setting.captureImagePath.Get() )
    {
        return false;
    }
    if( sImagePath != setting.captureImagePath )
    {
        sDirectory.Clear();
    }
    sImagePath = vw::common::CharArray( Memory_GetSysAllocator() );
    sImagePath = setting.captureImagePath;

    if( sDirectory.Get() == 0 )
    {
        sDirectory = vw::common::CharArray( Memory_GetSysAllocator() );

        // Get the folder name
        vw::common::CharArray directory( Memory_GetSysAllocator() );
        if( *( setting.captureImagePath.Get() ) == '\0')
        {
            nn::os::CTR::TargetHardware target = nn::os::CTR::GetRunningTargetHardware();
            if ( target == nn::os::CTR::TARGET_HARDWARE_PRODUCT )
            {
                char path[1024];
                nn::hio::CTR::GetEnvironmentVariable(path, 1024 , "NW4C_ROOT");
                directory  = path;
                directory += "\\tools\\Viewer";
            }
            else
            {
                directory  = ".";
            }
        }
        else
        {
            directory = setting.captureImagePath.Get();
        }

        // Create temp folder
        vw::common::CharArray temp_dir( Memory_GetSysAllocator() );
        temp_dir = directory;
        temp_dir += "\\temp";
        for(int i=0; i<128; ++i)
        {
            char temp_name[256];
            sprintf(temp_name, "%s%d", temp_dir.Get(), i);
            nn::Result result = nn::hio::CTR::CreateDirectory(temp_name);
            if ( result.IsSuccess() )
            {
                temp_dir = temp_name;
                break;
            }
        }

        // Get the creation data and time for the temp folder
        nn::hio::CTR::HostDirectoryA hostDirectory;
        nn::Result result = hostDirectory.Open( temp_dir.Get() );
        if ( result.IsFailure() )
            return false;

        nn::hio::CTR::DirectoryEntryA directoryEntry;
        result = hostDirectory.GetNextEntry(&directoryEntry);
        if ( result.IsFailure() )
            return false;

        result = hostDirectory.Close();
        if ( result.IsFailure() )
            return false;

        // Rename the folder based on the temp folder creation date and time
        char str[256];
        sprintf(str, "\\capture_%04d_%02d%02d_%02d%02d",
                    directoryEntry.creationTime.year,
                    directoryEntry.creationTime.month,
                    directoryEntry.creationTime.day,
                    directoryEntry.creationTime.hour,
                    directoryEntry.creationTime.minute);
        directory += str;
        nn::hio::CTR::Rename(directory.Get(), temp_dir.Get());
        sDirectory = directory;
    }

    // If the specified folder does not exists, create it and reset the capture count
    nn::hio::CTR::HostDirectoryA dirA;
    nn::Result result = dirA.Open( sDirectory.Get() );
    if ( result.IsFailure() )
    {
        nn::hio::CTR::CreateDirectory( sDirectory.Get() );
        sCaptureCount = 0;
    }

    (*filePath) = sDirectory;
    char str[256];
    sprintf(str, "\\%d.bmp", sCaptureCount++);
    (*filePath) += str;

    return true;
}

class DisplayBuffer
{
public:
    DisplayBuffer(GLenum format, u32 width, u32 height)
    {
        nngxGenDisplaybuffers(1, &m_Display);
        nngxBindDisplaybuffer(m_Display);
        nngxDisplaybufferStorage(format, width, height, NN_GX_MEM_FCRAM);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_ADDRESS, &m_Address);
    }
    ~DisplayBuffer()
    {
        nngxDeleteDisplaybuffers(1, &m_Display);
    }
    GLuint GetDisplay() const
    {
        return m_Display;
    }
    const u8* GetAddress() const
    {
        return (const u8*)m_Address;
    }

private:
    GLuint m_Display;
    GLint  m_Address;
};

// 1/2 capture
static bool
CaptureMin(const char* filePath)
{
    // Display buffer generation
    GLenum format = GetDisplayFormat();
    DisplayBuffer disp(format, nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT);

    // Allocates the buffer.
    u32 workBufSize = nn::gx::DISPLAY0_WIDTH * nn::gx::DISPLAY0_HEIGHT * 3;
    u8* workBuf = (u8*)Memory_AllocSys(workBufSize);
    if( workBuf == 0)
    {
        GetLogSystem()->AddSystemLog( "Out of memory" );
        return false;
    }

    // Opens files
    nn::hio::CTR::HostFile file;
    nn::Result res = file.Open(filePath,
                               nn::hio::CTR::HostFile::ACCESS_MODE_WRITE,
                               nn::hio::CTR::HostFile::OPEN_DISP_CREATE_ALWAYS);

    // Write header
    writeBmpHeader(&file, nn::gx::DISPLAY0_WIDTH / 2, nn::gx::DISPLAY0_HEIGHT / 2);

    // Scene rendering
    DrawScene(disp.GetDisplay(), 0, 0, 1);

    storeRect(workBuf, disp.GetAddress(), 0, 0, nn::gx::DISPLAY0_WIDTH,
                                                nn::gx::DISPLAY0_HEIGHT,
                                                nn::gx::DISPLAY0_WIDTH,
                                                nn::gx::DISPLAY0_HEIGHT, format);

    // Size reduction.
    const u32 src_width = nn::gx::DISPLAY0_HEIGHT;
    const u32 src_height = nn::gx::DISPLAY0_WIDTH;
    const u32 dst_width = src_width / 2;
    const u32 dst_height = src_height / 2;

    u32 dstSize = dst_width * dst_height * 3;
    u8* dstBuf = (u8*)Memory_AllocSys( dstSize );
    const u8* srcBuf = static_cast<const u8*>(workBuf);

    for(u32 y=0; y<dst_height; ++y)
    {
        for(u32 x=0; x<dst_width; ++x)
        {
            for(u32 c = 0; c < 3; ++c){
                u8 c0 = srcBuf[(x*2)  * 3 + c + 0];
                u8 c1 = srcBuf[(x*2+1)* 3 + c + 0];
                u8 c2 = srcBuf[(x*2)  * 3 + c + src_width * 3];
                u8 c3 = srcBuf[(x*2+1)* 3 + c + src_width * 3];
                u32 a = c0 + c1 + c2 + c3;
                dstBuf[(x + y*dst_width) * 3 + c] = u8(a / 4);
            }
        }
        srcBuf += src_width * 3 * 2;
    }

    // write out the image
    size_t size = 0;
    file.Write(&size, dstBuf, dstSize);
    file.Close();

    Memory_Free(workBuf);
    Memory_Free(dstBuf);

    return true;
}

// Divided capture
static bool
CaptureMag(const char* filePath)
{
    // Display buffer generation
    u32 magnify = static_cast<u32>(sMagnify);
    GLenum format = GetDisplayFormat();
    DisplayBuffer disp1(format, nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT);
    DisplayBuffer disp2(format, nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT);

    // Allocates the buffer.
    u32 workBufSize = nn::gx::DISPLAY0_WIDTH * nn::gx::DISPLAY0_HEIGHT * magnify * 3;
    u8* workBuf = (u8*)Memory_AllocSys(workBufSize);
    if( workBuf == 0)
    {
        GetLogSystem()->AddSystemLog( "Out of memory" );
        return false;
    }

    // Opens files
    nn::hio::CTR::HostFile file;
    nn::Result res = file.Open(filePath,
                               nn::hio::CTR::HostFile::ACCESS_MODE_WRITE,
                               nn::hio::CTR::HostFile::OPEN_DISP_CREATE_ALWAYS);

    // Write header
    writeBmpHeader(&file, nn::gx::DISPLAY0_WIDTH * magnify,
                          nn::gx::DISPLAY0_HEIGHT * magnify);

    // write out the image
    for(u32 i=0, y=0; y<magnify; ++y)
    {
        for(u32 x=0; x<magnify; ++x, ++i)
        {
            DisplayBuffer* disp = (i%2)? &disp1 : &disp2;
            DrawScene(disp->GetDisplay(), x, y, magnify);

            storeRect(workBuf, disp->GetAddress(), x, 0,
                      nn::gx::DISPLAY0_WIDTH,
                      nn::gx::DISPLAY0_HEIGHT * magnify,
                      nn::gx::DISPLAY0_WIDTH,
                      nn::gx::DISPLAY0_HEIGHT,
                      format);
        }
        size_t size = 0;
        file.Write(&size, workBuf, workBufSize);
    }
    file.Close();

    Memory_Free(workBuf);

    return true;
}

static bool
CaptureG2D( const char* filePath )
{
    // Allocates the buffer.
    u32 workBufSize = nn::gx::DISPLAY0_WIDTH * nn::gx::DISPLAY0_HEIGHT * 3;
    u8* workBuf = (u8*)Memory_AllocSys(workBufSize);
    if( workBuf == 0)
    {
        GetLogSystem()->AddSystemLog( "Out of memory" );
        return false;
    }

    // Opens files
    nn::hio::CTR::HostFile file;
    nn::Result res = file.Open(filePath,
                               nn::hio::CTR::HostFile::ACCESS_MODE_WRITE,
                               nn::hio::CTR::HostFile::OPEN_DISP_CREATE_ALWAYS);

    // Write header
    writeBmpHeader(&file, nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT);

    const GLenum display = NN_GX_DISPLAY0;
    nngxActiveDisplay(display);
    GLint format, width, height, address;
    nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_ADDRESS, &address);
    void* srcArea = reinterpret_cast<u8*>(address);
    nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_FORMAT, &format);
    nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_WIDTH,  &width);
    nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_HEIGHT, &height);

    // write out the image
    storeRect(workBuf, reinterpret_cast<u8*>(address), 0, 0,
              nn::gx::DISPLAY0_WIDTH,
              nn::gx::DISPLAY0_HEIGHT,
              nn::gx::DISPLAY0_WIDTH,
              nn::gx::DISPLAY0_HEIGHT,
              format);
    size_t size = 0;
    file.Write(&size, workBuf, workBufSize);

    file.Close();

    Memory_Free(workBuf);
    return true;
}

//============================================
// Interface class for screen capture
//============================================
/* static */ CaptureInterface CaptureInterface::sCaptureInterface;

/*---------------------------------------------------------------------------*
  Name:         CaptureInterface::GetInstance

  Description:  Gets the interface

  Arguments:    None.

  Returns:      Returns the capture class interface.
 *---------------------------------------------------------------------------*/
/* static */ CaptureInterface&
CaptureInterface::GetInstance()
{
    return sCaptureInterface;
}


/*---------------------------------------------------------------------------*
  Name:         CaptureInterface::Finalize

  Description:  Finalize.

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
/* static */ void
CaptureInterface::Finalize()
{
    sDirectory = vw::common::CharArray();
    sImagePath = vw::common::CharArray();
}


/*---------------------------------------------------------------------------*
  Name:         CaptureInterface::Capture

  Description:  Executes capture.

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
/* static */ void
CaptureInterface::Capture(const char* filePath)
{
    bool IsRunning = false;
    nn::Result ret = nn::hio::GetHioDaemonStatus( &IsRunning );
    if(ret.IsSuccess() && IsRunning == true)
    {
        sCapture = true;
    }
    else
    {
        NN_LOG("Screen Capture Failure\n");
        GetLogSystem()->AddSystemLog( "Screen Capture Failure" );
    }
}

/*---------------------------------------------------------------------------*
  Name:         CaptureInterface::Main

  Description:  Executes capture.

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
CaptureInterface::Main()
{
    if( sCapture )
    {
        sCapture = false;
    }
    else
    {
        return;
    }

    // Get the folder name
    vw::common::CharArray filePath( Memory_GetSysAllocator() );
    if( !GetFilePath( &filePath ) )
    {
        return;
    }

    // Executes capture.
    const char* path = filePath.Get();

    bool result = false;
    if(G2D_ID == vw::system::App::GetInstance().GetFocusID())
    {
        result = CaptureG2D( path );
    }
    else
    {
        result = (sMagnify < 1.0f)? CaptureMin( path ):
                                    CaptureMag( path );
    }

    // Output log.
    if( result )
    {
        NN_LOG("Screen Capture Success\n");
        GetLogSystem()->AddSystemLog("Screen Capture Success");
    }
    else
    {
        NN_LOG("Screen Capture Failure\n");
        GetLogSystem()->AddSystemLog( "Screen Capture Failure" );
    }
}

/*---------------------------------------------------------------------------*
  Name:         CaptureInterface::IsExecuting

  Description:  Gets the capture executing flag

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
bool
CaptureInterface::IsExecuting()
{
    return sExecuting;
}

/*---------------------------------------------------------------------------*
  Name:         CaptureInterface::GetMagnify

  Description:  Gets the number of divisions

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
f32
CaptureInterface::GetMagnify()
{
    return sMagnify;
}

/*---------------------------------------------------------------------------*
  Name:         CaptureInterface::SetMagnify

  Description:  Sets number of divisions

  Arguments:    magnify Number of divisions

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
CaptureInterface::SetMagnify( f32 magnify )
{
    sMagnify = magnify;
}

}   /* namespace system */
}   /* namespace vw     */
