/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#include <cstring>
#include <cstdlib>
#include <viewer/common.h>
#include <viewer/system.h>
#include "../include/LogUtil.h"



extern "C" {
#if defined( NN_PCGX )
    extern u8* VIEWER_BEGIN;
    extern u8* VIEWER_END;
#else
    extern u8* VIEWER_BEGIN[];
    extern u8* VIEWER_END[];
#endif
}



namespace vw     {
namespace system {

/*---------------------------------------------------------------------------*
  Name:         ParseToken_
  Description:  Parses token, and gets the name and value string for 'name = value'.

  Arguments:    namePtr     Pointer to the name string
                valuePtr    Pointer to the value string
                line        Pointer to the string to parse
  Returns:      None.
 *---------------------------------------------------------------------------


*/
static bool
ParseToken_( char** namePtr, char** valuePtr, char* line )
{
    char *p;

    p = std::strchr( line, '#' );
    if ( p )
    {
        *p = '\0';
    }

    p = std::strchr( line, '=' );
    if ( ! p )
    {
        return false;
    }

    *p = '\0';

    char* pName = std::strtok( line, " \t\n\r" );

    if ( ! pName )
    {
        *p = '=';
        return false;
    }
    *namePtr = pName;

    char* pValue = std::strtok( p + 1, " \t\n\r" );
#if 0
    if ( ! pValue )
    {
        *p = '=';
        return false;
    }
#endif
    *valuePtr = pValue;
    return true;
}

/*---------------------------------------------------------------------------*
  Name:         ParseComment_
  Description:  Gets the start position of the comment text from the first line of the string.
  Arguments:    commentPtr  Argument to return the comment start position when a comment has been detected
                line        Pointer to the string to parse
  Returns:      Returns TRUE when a comment is discovered in the specified line and returns FALSE when no comment exists.

 *---------------------------------------------------------------------------


*/
static bool
ParseComment_( char** commentPtr, const char* line )
{
    *commentPtr = std::strchr( const_cast<char*>( line ), '#' );

    if ( *commentPtr )
    {
        return true;
    }
    else
    {
        return false;
    }
}


/*---------------------------------------------------------------------------*
  Name:         ConfigNode::ConfigNode
  Description:  The Config node constructor.
  Arguments:    name    name string
                value   value string
  Returns:      None.
 *---------------------------------------------------------------------------


*/
/* ctor */
ConfigNode::ConfigNode( const char* name, const char* value /* = NULL */ )
: mName ( name,  Memory_GetAllocator( SYSTEM_ID ) ),
  mValue( value, Memory_GetAllocator( SYSTEM_ID ) )
{
//    NW4R_DB_NULL_ASSERT( name );
}



/*---------------------------------------------------------------------------*
  Name:         ConfigNode::CompareName
  Description:  Gets indication of whether the settings <CODE>name</CODE> is the same as the specified value.
  Arguments:    name    String to use for comparing for equivalency
  Returns:      Returns <CODE>0</CODE> if they are equal.
                Returns a non-zero value if they are not the same.
 *---------------------------------------------------------------------------


*/
s32
ConfigNode::CompareName( const char* name )
{
    if ( mName.Get() == NULL )
    {
        return -1;
    }

    return std::strcmp( name, mName.Get() );
}


/*---------------------------------------------------------------------------*
  Name:         ConfigNode::Dump
  Description:  Dump output.
  Arguments:    value
  Returns:      None.
 *---------------------------------------------------------------------------


*/
s32
ConfigNode::Dump( char* buf /* = NULL */ )
{
    if ( mName.Get() )
    {
        if ( buf )
        {
            if ( mValue.Get() )
            {
                return std::sprintf( buf, "%s = %s", mName.Get(), mValue.Get() );
            }
            else
            {
                return std::sprintf( buf, "%s = ", mName.Get() );
            }
        }
        else
        {
            NW_VIEWER_LOG("%s = %s\n", mName.Get(), mValue.Get());
            return 0;
        }
    }
    else
    {
        if ( mValue.Get() )
        {
            if ( buf )
            {
                return std::sprintf( buf, "%s", mValue.Get() );
            }
            else
            {
                NW_VIEWER_LOG("%s\n", mValue.Get());
                return 0;
            }
        }
    }
    return 0;
}


ConfigFile*     ConfigFile::sInstance = NULL;
bool            ConfigFile::sReloadedTrig = false;
bool            ConfigFile::sReloadedNext = false;
CharArray       ConfigFile::sHostFilePath = CharArray();


/*---------------------------------------------------------------------------*
  Name:         ConfigFile::ConfigFile
  Description:  Constructor
  Arguments:    filePath
  Returns:      None.
 *---------------------------------------------------------------------------


*/
/* ctor */
ConfigFile::ConfigFile()
 : mExistsFile( false )
{
    // Initializes child list
    //ut::List_Init( , offsetof( ConfigNode, mLink ) );
}

/*---------------------------------------------------------------------------*
  Name:         ConfigFile::~ConfigFile
  Description:  Destructor
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
/* dtor */
ConfigFile::~ConfigFile( void )
{
    ConfigNode *node;

    while ( ( node = static_cast<ConfigNode*>( mChildList.GetNext( NULL ) ) ) != NULL )
    {
        mChildList.erase( node );
        node->~ConfigNode();
        vw::system::Memory_Free( node );
    }

    mDefaultFilePath.Clear();
}

/*---------------------------------------------------------------------------*
  Name:         ConfigFile::Create
  Description:  Creates an instance of the Config settings.
  Arguments:    filePath    File path
                mcsOnly     Flag for whether to read via MCS and not from the DVD
  Returns:      None.
 *---------------------------------------------------------------------------


*/
ConfigFile*
ConfigFile::Create()
{
    if ( sInstance != NULL )
    {
        return sInstance;
    }

    vw::common::CharArray tmpPath( vw::system::Memory_GetSysAllocator() );
    const char* host_file_path = NULL;
    const char* rom_file_path = "rom:/Viewer.ini";

#if defined( NN_PCGX )
    tmpPath = "../Viewer/Viewer.ini";
    host_file_path = tmpPath.Get();
#else

    bool IsRunning = false;
    nn::Result ret = nn::hio::GetHioDaemonStatus( &IsRunning );
    if ( ret.IsSuccess() && IsRunning == true )
    {
        char path[1024];
        nn::os::CTR::TargetHardware target = nn::os::CTR::GetRunningTargetHardware();

        if ( sHostFilePath.Get() == NULL )
        {
            if ( target == nn::os::CTR::TARGET_HARDWARE_PRODUCT )
            {
                nn::hio::CTR::GetEnvironmentVariable(path, 1024 , "NW4C_ROOT");
                tmpPath = path;
    #ifdef NW_BUILD_SCONS
                tmpPath += "/tools/Viewer/Viewer.ini";
    #else
                tmpPath += "/sources/tools/Viewer/Viewer.ini";
    #endif
            }
            else
            {
                if(vw::system::App::GetInstance().IsEmbeddedViewer())
                {
                    tmpPath = "./Viewer.ini";
                }
                else
                {
                    tmpPath = "../Viewer.ini";
                }
            }
        }
        else
        {
            tmpPath = sHostFilePath;
        }

        host_file_path = tmpPath.Get();
    }
#endif

    void* buf = vw::system::Memory_AllocSys( sizeof(ConfigFile) );
    sInstance = new(buf) ConfigFile();

    sInstance->mDefaultFilePath = vw::common::CharArray(tmpPath, vw::system::Memory_GetSysAllocator());
    sHostFilePath = sInstance->mDefaultFilePath;

    bool success = false;

    if ( ! success && host_file_path )
    {
        success = sInstance->ReadFromHost( host_file_path );
    }

    if ( ! success )
    {
        u32 file_size = static_cast<u32>( reinterpret_cast<int>(VIEWER_END) - reinterpret_cast<int>(VIEWER_BEGIN) );
        void* data = vw::system::Memory_AllocSys( file_size );
        memcpy(data, VIEWER_BEGIN, file_size);
        success = sInstance->ReadFromMEM( data, file_size );
        vw::system::Memory_Free( data );

        // For ViewerLibrary, communication with CS is disabled by default
        if (vw::system::App::IsEmbeddedViewer())
        {
            sInstance->SetInt("/g3d/config/cs_connection", 0 );
        }
    }

    //NW4R_WARNING( success, RESTXT_ON_CREATE_INI_ERROR );

    return sInstance;
}

/*---------------------------------------------------------------------------*
  Name:         ConfigFile::Destroy
  Description:  Destroys an instance of the Config settings.
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
ConfigFile::Destroy( void )
{
    if ( ! sInstance )
    {
        return;
    }

    sInstance->~ConfigFile();
    vw::system::Memory_Free( sInstance );

    sInstance = NULL;
    sReloadedTrig = false;
    sReloadedNext = false;
    sHostFilePath.Clear();
}

/*---------------------------------------------------------------------------*
  Name:         ConfigFile::ReadFromDVD
  Description:  Loads and parses INI file from DVD.
  Arguments:    filePath    File path
  Returns:      Returns true if loading the file succeeded, returns false if it fails

 *---------------------------------------------------------------------------


*/
bool
ConfigFile::ReadFromRom( const char* filePath )
{
    // Exit as-is if cannot access DVD
    //if ( ! system::App::GetInstance().IsEnableDVD() )
    //{
    //    return false;
    //}

    nn::fs::FileReader fileReader;

    nn::Result result = fileReader.TryInitialize(filePath);
    if ( result.IsFailure() )
    {
        return false;
    }

    u32 size = fileReader.GetSize();
    char* fileBuf = static_cast<char*>( Memory_AllocSysDevice( size, 32 ) );
    fileReader.Read( fileBuf, s32( size ) );
    fileReader.Finalize();

    this->Parse_( fileBuf, s32( size ) );

    Memory_Free( fileBuf );

    mExistsFile = true;

    return true;
}

/*---------------------------------------------------------------------------*
  Name:         ConfigFile::ReadFromMcs
  Description:  Loads and parses INI file from MCS.
  Arguments:    filePath    File path
  Returns:      Returns true if loading the file succeeded, returns false if it fails

 *---------------------------------------------------------------------------


*/
bool
ConfigFile::ReadFromHost( const char* filePath )
{
    // Opens files
    nn::hio::CTR::HostFile hostFile;
    nn::Result result = hostFile.Open(
        filePath,
        nn::hio::CTR::HostFile::ACCESS_MODE_READ,
        nn::hio::CTR::HostFile::OPEN_DISP_OPEN_EXISTING
    );
    if ( result.IsFailure() )
    {
        return false;
    }

    // Get the file size
    s64 size;
    hostFile.Seek(&size,0,nn::hio::CTR::HostFile::SEEK_TYPE_END);
    hostFile.Seek(0,nn::hio::CTR::HostFile::SEEK_TYPE_SET);

    // Allocates the buffer.
    char* fileBuf = static_cast<char*>( Memory_AllocSysDevice( size, 128 ) );
    hostFile.Read( fileBuf, s32( size ) );
    hostFile.Close();

    this->Parse_( fileBuf, s32( size ) );
    Memory_Free( fileBuf );

    mExistsFile = true;

    return true;
}


/*---------------------------------------------------------------------------*
  Name:         ConfigFile::ReadFromMEM
  Description:  Loads data in memory as an <CODE>INI</CODE> file and analyzes it.
  Arguments:    data     Data pointer
                size        Size
  Returns:      Returns true if loading the file succeeded, returns false if it fails

 *---------------------------------------------------------------------------


*/
bool
ConfigFile::ReadFromMEM( void* data, u32 size )
{
    if ( ! data )
    {
        return false;
    }

    this->Parse_( static_cast<char*>(data), s32(size) );
    mExistsFile = true;
    return true;
}


/*---------------------------------------------------------------------------*
  Name:         ConfigFile::Clear
  Description:  Clears the settings.
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
ConfigFile::Clear()
{
    ConfigNode *node;

    while ( ( node = static_cast<ConfigNode*>( mChildList.GetNext( NULL ) ) ) != NULL )
    {
        node->~ConfigNode();
        mChildList.erase( node );
        vw::system::Memory_Free( node );
    }
}


/*---------------------------------------------------------------------------*
  Name:         ConfigFile::Reload
  Description:  Reloads configuration file from MCS
  Arguments:    filePath
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
ConfigFile::Reload( const char* filePath )
{
    vw::common::CharArray tmpPath( vw::system::Memory_GetSysAllocator() );

    if ( filePath == NULL && sHostFilePath.Get() == NULL )
    {
        char path[1024];
        nn::os::CTR::TargetHardware target = nn::os::CTR::GetRunningTargetHardware();

        if ( target == nn::os::CTR::TARGET_HARDWARE_PRODUCT )
        {
            nn::hio::CTR::GetEnvironmentVariable(path, 1024 , "NW4C_ROOT");
            tmpPath = path;
#ifdef NW_BUILD_SCONS
            tmpPath += "/tools/Viewer/Viewer.ini";
#else
            tmpPath += "/sources/tools/Viewer/Viewer.ini";
#endif
        }
        else
        {
            if(vw::system::App::GetInstance().IsEmbeddedViewer())
            {
                tmpPath = "./Viewer.ini";
            }
            else
            {
                tmpPath = "../Viewer.ini";
            }
        }

        filePath = tmpPath.Get();
    }
    else if ( filePath == NULL && sHostFilePath.Get() != NULL )
    {
        filePath = sHostFilePath.Get();
    }

    this->Clear();
    if ( this->ReadFromHost( filePath ) )
    {
        sReloadedNext = true;
    }
}


/*---------------------------------------------------------------------------*
  Name:         ConfigFile::Reload
  Description:  Reloads the Settings file.
  Arguments:    data
                size
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
ConfigFile::Reload( void* data, u32 size )
{
    this->Clear();
    if ( this->ReadFromMEM( data, size ) )
    {
        sReloadedNext = true;
    }
}



/*---------------------------------------------------------------------------*
  Name:         ConfigFile::Parse_
  Description:  Parses <CODE>INI</CODE> file and converts to list of <CODE>name</CODE> and <CODE>value</CODE>.
  Arguments:    fileBuf Pointer to INI file raw data
                size    Size of raw data
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
ConfigFile::Parse_( char* fileBuf, s32 size )
{
    NW_NULL_ASSERT( fileBuf );

    fileBuf[ size - 1 ] = '\0';

    char* pLine = fileBuf;
    char* pchr;
    char* name;
    char* val;
    char* comment;

    while ( (pchr = std::strchr( pLine, '\n' )) != NULL )
    {
        if (*(pchr - 1) == '\r')
        {
            *(pchr -1) = '\0';
        }
        *pchr++ = '\0';
        (void)ParseComment_( &comment, pLine );

        if ( ParseToken_( &name, &val, pLine ) )
        {
            this->SetString( name, val );
            if ( comment )
            {
                this->SetComment("\t");
            }
        }
        else
        {
            this->SetComment( pLine );
        }
        if ( comment )
        {
            *comment = '#';
            this->SetComment( comment );
        }
        this->SetComment("\r\n");
        pLine = pchr;
    }
}



/*---------------------------------------------------------------------------*
  Name:         ConfigFile::Write
  Description:  Exports the INI file via MCS.
  Arguments:    filePath    Output file path
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
ConfigFile::Write( const char* filePath )
{
    // Opens files
    nn::hio::CTR::HostFile hostFile;
    nn::Result result = hostFile.Open(
        filePath,
        nn::hio::CTR::HostFile::ACCESS_MODE_WRITE,
        nn::hio::CTR::HostFile::OPEN_DISP_CREATE_ALWAYS
    );
    if ( result.IsFailure() )
    {
        return;
    }

    #define CONFIG_FILE_SIZE_MAX 0x10000

    // Allocates the buffer.
    char *str_buf = static_cast<char*>( Memory_AllocSysDevice( CONFIG_FILE_SIZE_MAX, 128 ) );
    char *p       = str_buf;
    u32  str_len  = 0;

    ConfigNode* node = static_cast<ConfigNode*>( mChildList.GetNext( NULL ) );

    bool  has_name_before = false;
    while ( node != NULL )
    {
        if ( has_name_before )
        {
            if ( node->HasName() )
            {
                *p++ = '\r';
                *p++ = '\n';
                str_len += 2;
            }
        }
        has_name_before = node->HasName();

        s32 len = node->Dump( p );
        str_len += len;
        p += len;
        node = static_cast<ConfigNode*>( mChildList.GetNext( node ) );
    }
    if ( has_name_before )
    {
        *p++ = '\r';
        *p++ = '\n';
        str_len += 2;
    }

    u32 size;
    hostFile.Write( &size, str_buf, str_len );
    hostFile.Close();

    Memory_Free( str_buf );
}

/*---------------------------------------------------------------------------*
  Name:         ConfigFile::SeekNode
  Description:  Searches for the settings value with the specified name.
  Arguments:    path    name value of value to search for
  Returns:      None.
 *---------------------------------------------------------------------------


*/
ConfigNode*
ConfigFile::SeekNode( const char* path )
{
    ConfigNode *node = static_cast<ConfigNode*>( mChildList.GetNext( NULL ) );
    while ( node != NULL )
    {
        if ( node->CompareName( path ) == 0 )
        {
            return node;
        }
        node = static_cast<ConfigNode*>( mChildList.GetNext( node ) );
    }
    return NULL;
}


/*---------------------------------------------------------------------------*
  Name:         ConfigFile::DumpNode_
  Description:  Outputs the settings parameters as character strings.
  Arguments:    path    Parameter name to output
                buf     Pointer to the destination buffer
  Returns:      None.
 *---------------------------------------------------------------------------


*/
s32
ConfigFile::DumpNode_( const char* path, char* buf )
{
    ConfigNode* node = this->SeekNode( path );
    if ( node == NULL )
    {
        return 0;
    }
    else
    {
        s32 len = node->Dump( buf );

        if ( ! mExistsFile )
        {
            len += std::sprintf( buf + len, "\r\n" );
        }
        return len;
    }
}


/*---------------------------------------------------------------------------*
  Name:         ConfigFile::AppendNode
  Description:  Adds a new value to the list of settings values.
  Arguments:    child   Configuration node to add
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
ConfigFile::AppendNode( ConfigNode* child )
{
#if 0
    ConfigNode* node = static_cast<ConfigNode*>( ut::List_GetNext( &mChildList, NULL ) );

    while ( node != NULL )
    {
        if ( child->CompareName( node->GetName() ) >= 0 )
        {
            ut::List_Insert( &mChildList, node, child );
            return;
        }
        node = static_cast<ConfigNode*>( ut::List_GetNext( &mChildList, node ) );
    }
#endif
    mChildList.push_back( child );
}


/*---------------------------------------------------------------------------*
  Name:         ConfigFile::CreateNode
  Description:  Creates a new settings node instance.
  Arguments:    path    Setting name
  Returns:      Pointer to configuration node
 *---------------------------------------------------------------------------


*/
ConfigNode*
ConfigFile::CreateNode( const char* path )
{
    ConfigNode* node = SeekNode( path );
    if ( ! node )
    {
        void* buf = vw::system::Memory_AllocSys( sizeof(ConfigNode) );
        node = new(buf) ConfigNode(path);
        AppendNode( node );
    }
    return node;
}

/*---------------------------------------------------------------------------*
  Name:         ConfigFile::GetString
  Description:  Gets the value specified by <CODE>name</CODE> as a character string.
                If no value exists, the function returns <CODE>NULL</CODE>.
  Arguments:    path    Setting name
  Returns:      Returns a value as a character string.
                If the settings name specified for the argument cannot be found, the function returns <CODE>NULL</CODE>.
 *---------------------------------------------------------------------------


*/
const char*
ConfigFile::GetString( const char* path )
{
    ConfigNode *node = SeekNode( path );
    if ( node == NULL )
    {
        return NULL;
    }
    return node->GetValue();
}


/*---------------------------------------------------------------------------*
  Name:         ConfigFile::SetString
  Description:  Sets the Config value in character-string format.
  Arguments:    path    name string
                value   value string
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
ConfigFile::SetString( const char* path, const char* value )
{
    NW_NULL_ASSERT( path );

    ConfigNode* node;

    if ( path != NULL )
    {
        node = CreateNode( path );
        NW_NULL_ASSERT( node );
    }
    node->SetValue( value );
}

/*---------------------------------------------------------------------------*
  Name:         ConfigFile::SetComment
  Description:  Adds a comment.
  Arguments:    comment    Comment string
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
ConfigFile::SetComment( const char* comment )
{
    void* buf = vw::system::Memory_AllocSys( sizeof(ConfigNode) );
    ConfigNode* node = new(buf) ConfigNode( NULL, comment);

    AppendNode( node );
}


/*---------------------------------------------------------------------------*
  Name:         ConfigFile::GetInt
  Description:  Gets the Config value in Int format.
  Arguments:    path    Setting name string to get
                pValue   Buffer for getting the value
  Returns:      Returns <CODE>true</CODE> if the specified <SPAN class="argument">path</SPAN> setting exists.
                Returns <CODE>false</CODE> if it does not exist.
 *---------------------------------------------------------------------------


*/
bool
ConfigFile::GetInt  ( const char* path, s32* pValue )
{
    const char* val = GetString( path );

    if ( val == NULL )
    {
        return false;
    }

    if ( val[ 0 ] == '0' && val[ 1 ] == 'x' )
    {
        *pValue = std::strtol( val, NULL, 16 );
    }
    else
    {
        *pValue = std::strtol( val, NULL, 10 );
    }
    return true;
}


/*---------------------------------------------------------------------------*
  Name:         ConfigFile::GetUInt
  Description:  Gets the Config value in UInt format.
  Arguments:    path    Setting name string to get
                pValue   Buffer for getting the value
  Returns:      Returns <CODE>true</CODE> if the specified <SPAN class="argument">path</SPAN> setting exists.
                Returns <CODE>false</CODE> if it does not exist.
 *---------------------------------------------------------------------------


*/
bool
ConfigFile::GetUInt  ( const char* path, u32* pValue )
{
    const char* val = GetString( path );

    if ( val == NULL )
    {
        return false;
    }

    if ( val[ 0 ] == '0' && val[ 1 ] == 'x' )
    {
        *pValue = static_cast<unsigned int>( std::strtoul( val, NULL, 16 ) );
    }
    else
    {
        *pValue = static_cast<unsigned int>( std::strtoul( val, NULL, 10 ) );
    }
    return true;
}


/*---------------------------------------------------------------------------*
  Name:         ConfigFile::SetInt
  Description:  Sets the Config value in Int format.
                The value is changed to a string internally.
  Arguments:    path    name string
                value : Value
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
ConfigFile::SetInt  ( const char* path, s32 value )
{
    char valueBuf[ 32 ];

    (void)std::sprintf( valueBuf, "%d", value );

    SetString( path, valueBuf );
}


/*---------------------------------------------------------------------------*
  Name:         ConfigFile::SetUInt
  Description:  Sets the Config value in UInt format.
                The value is changed to a string internally.
  Arguments:    path    name string
                value : Value
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
ConfigFile::SetUInt  ( const char* path, u32 value )
{
    char valueBuf[ 32 ];

    (void)std::sprintf( valueBuf, "%u", value );

    SetString( path, valueBuf );
}


/*---------------------------------------------------------------------------*
  Name:         ConfigFile::SetInt16
  Description:  Sets the Config value in hexadecimal integer format.
                The value is changed to a string internally.
  Arguments:    path    name string
                value : Value
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
ConfigFile::SetInt16( const char* path, s32 value )
{
    char valueBuf[ 32 ];

    (void)std::sprintf( valueBuf, "0x%08X", value );

    SetString( path, valueBuf );
}


/*---------------------------------------------------------------------------*
  Name:         ConfigFile::SetUInt16
  Description:  Sets the Config value in hexadecimal integer format.
                The value is changed to a string internally.
  Arguments:    path    name string
                value : Value
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
ConfigFile::SetUInt16 ( const char* path, u32 value )
{
    char valueBuf[ 32 ];

    (void)std::sprintf( valueBuf, "0x%08X", value );

    SetString( path, valueBuf );
}


/*---------------------------------------------------------------------------*
  Name:         ConfigFile::GetFloat
  Description:  Gets the Config value in float format.
  Arguments:    path    Setting name string to get
                pValue   Buffer for getting the value
  Returns:      Returns <CODE>true</CODE> if the specified <SPAN class="argument">path</SPAN> setting exists.
                Returns <CODE>false</CODE> if it does not exist.
 *---------------------------------------------------------------------------


*/
bool
ConfigFile::GetFloat( const char* path, f32* pValue )
{
    const char* val = GetString( path );

    if ( val == NULL )
    {
        return false;
    }

    *pValue = static_cast<f32>(std::strtod( val, NULL ));
    return true;
}


/*---------------------------------------------------------------------------*
  Name:         ConfigFile::SetFloat
  Description:  Sets the Config value in floating-point format.
                The value is changed to a string internally.
  Arguments:    path    name string
                value : Value
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
ConfigFile::SetFloat( const char* path, f32 value )
{
    char valueBuf[ 32 ];

    (void)std::sprintf( valueBuf, "%f", value );

    SetString( path, valueBuf );
}

/*---------------------------------------------------------------------------*
  Name:         ConfigFile::SetConfigFilePath
  Description:  Sets the path for loading the config file.
  Arguments:    filePath    Specifies the file path for loading the config file.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
ConfigFile::SetConfigFilePath( const char *filePath )
{
    if ( sHostFilePath.Get() != NULL )
    { sHostFilePath.Clear(); }
    sHostFilePath = CharArray( filePath, vw::system::Memory_GetSysAllocator() );
}

/*---------------------------------------------------------------------------*
  Name:         ConfigFile::GetConfigFilePath
  Description:  Gets the path for loading the config file.
  Arguments:    None.
  Returns:      Returns the path for loading the config file.
 *---------------------------------------------------------------------------


*/
const char*
ConfigFile::GetConfigFilePath()
{
    return sHostFilePath.Get();
}

#if 0
/*---------------------------------------------------------------------------*
  Name:         ConfigFileExporter::StartImport
  Description:  Starts INI file import from MCS.
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
ConfigFileExporter::StartImport()
{
    mFilePath[ 0 ] = '\0';

    BOOL result = vw::system::FileIOTask_RequestSelectFileForLoad( &mFilter, ImportFileSelectClbk_, this );
    if ( result )
    {
        mState = STATE_SELECTING;
        this->CreateMsgDialog_("Import File Select", "Please specify the input path on PC" );
    }
    else
    {
        this->Cancel();
    }
}


/*---------------------------------------------------------------------------*
  Name:         ConfigFileExporter::StartExport
  Description:  Starts INI file export to MCS.
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
ConfigFileExporter::StartExport()
{
    mFilePath[ 0 ] = '\0';

    BOOL result = vw::system::FileIOTask_RequestSelectFileForSave( &mFilter, ExportFileSelectClbk_, this );
    if ( result )
    {
        // Writes back the configuration content to the ConfigFile structure.
        App::GetInstance().WriteBackConfigFile();

        mState = STATE_SELECTING;
        this->CreateMsgDialog_("Export File Select", "Please specify the output path on PC" );
    }
    else
    {
        this->Cancel();
    }
}


/*---------------------------------------------------------------------------*
  Name:         ConfigFileExporter::CreateMsgDialog_
  Description:  Generates message boxes in the file export.
  Arguments:    title   Title string for the message box
                message Message string for message box
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
ConfigFileExporter::CreateMsgDialog_( const char* title, const char* message )
{
    mDialogWindow = new(SYSTEM_ID)MessageDialog( title, message );
    const dw::Size size = mDialogWindow->GetSize();
    mDialogWindow->Create( (CANVAS_WIDTH - size.width) / 2, (CANVAS_HEIGHT - size.height) / 2 );
}


/*---------------------------------------------------------------------------*
  Name:         ConfigFileExporter::Cancel
  Description:  Cancels file export.
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
ConfigFileExporter::Cancel()
{
    if ( mDialogWindow )
    {
        delete mDialogWindow;
        mDialogWindow = NULL;
    }
    mState = STATE_IDLE;
}


/*---------------------------------------------------------------------------*
  Name:         ConfigFileExporter::ImportFileSelectClbk_
  Description:  Callback called when the INI file to import is selected with FILEIO.

  Arguments:    result   INI file selection result on the PC
                filename: Filename
                arg      Pointer to ConfigFileExporter instance
  Returns:      None.
 *---------------------------------------------------------------------------


*/
/* static */ void
ConfigFileExporter::ImportFileSelectClbk_( u32 result, const char* filename, void* arg )
{
    ConfigFileExporter* exporter = static_cast<ConfigFileExporter*>( arg );

    if ( result ==  nw4r::mcs::FILEIO_ERROR_SUCCESS )
    {
        exporter->StartLoad( filename );
    }
    else
    {
        // Cancel.
        exporter->Cancel();
    }
}


/*---------------------------------------------------------------------------*
  Name:         ConfigFileExporter::ExportFileSelectClbk_
  Description:  Callback called when the INI file to export is selected with FILEIO.

  Arguments:    result   output file selection result on the PC
                filename: Filename
                arg      Pointer to ConfigFileExporter instance
  Returns:      None.
 *---------------------------------------------------------------------------


*/
/* static */ void
ConfigFileExporter::ExportFileSelectClbk_( u32 result, const char* filename, void* arg )
{
    ConfigFileExporter* exporter = static_cast<ConfigFileExporter*>( arg );

    if ( result ==  nw4r::mcs::FILEIO_ERROR_SUCCESS )
    {
        exporter->StartSave( filename );
    }
    else
    {
        // Cancel.
        exporter->Cancel();
    }
}


/*---------------------------------------------------------------------------*
  Name:         ConfigFileExporter::Update
  Description:  Flashes the message
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
ConfigFileExporter::Update()
{
    if ( mState == STATE_EXPORTING )
    {
        GetConfigFile()->Write( mFilePath );
        this->Cancel();
    }
    else if ( mState == STATE_IMPORTING )
    {
        GetConfigFile()->Reload( mFilePath );
        this->Cancel();
    }
}


/*---------------------------------------------------------------------------*
  Name:         ConfigFileExporter::StartLoad
  Description:  Transitions to the INI file load start state.
  Arguments:    filename
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
ConfigFileExporter::StartLoad( const char* filename )
{
    mState = STATE_IMPORTING;
    vw::common::strlcpy( mFilePath, filename, MAX_PATH );
}


/*---------------------------------------------------------------------------*
  Name:         ConfigFileExporter::StartSave
  Description:  Transitions to the INI file save start state.
  Arguments:    filename
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
ConfigFileExporter::StartSave( const char* filename )
{
    mState = STATE_EXPORTING;
    vw::common::strlcpy( mFilePath, filename, MAX_PATH );
}
#endif


}   /* namespace system */
}   /* namespace viewer */




