/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/
#include <viewer/system.h>



extern "C" {
#if defined( NN_PCGX )
    extern u8* COMMON_3D_BEGIN;
    extern u8* COMMON_3D_END;
#else
    extern u8* COMMON_3D_BEGIN[];
    extern u8* COMMON_3D_END[];
#endif
}

namespace vw     {
namespace system {



/*
  
 */
CommandCache::CommandCache() : m_Buffer( 0 ), m_Size( 0 )
{
}
CommandCache::~CommandCache()
{
    Release();
}
// Starts the command cache.
u32*
CommandCache::Begin()
{
    if( m_Data == 0 )
    {
        m_Data = static_cast<u32*>( vw::system::Memory_AllocSysDevice( 1024 * 32 ) );
    }
   // Release();
    return m_Data;
}
// Ends the command cache.
void
CommandCache::End( u32* command )
{
    u32* start=m_Data, *end=command;
    if ( start && end )
    {
        ResizeBuffer( nw::ut::GetOffsetFromPtr(start, end) );
        if( m_Buffer )
        {
            std::memcpy(m_Buffer, start, m_Size);
        }
    }
}
// Resizes the buffer.
void
CommandCache::ResizeBuffer(u32 size)
{
    if(m_Size < size)
    {
        Release();
        m_Buffer = (u32*)vw::system::Memory_AllocSysDevice( size );
        m_Size = size;
    }
}
// Copies the command.
u32*
CommandCache::MakeCommand( u32* command ) const
{
    if( m_Size == 0 )
    {
        return command;
    }
    std::memcpy( command, m_Buffer, m_Size );
    u8* cmd8 = (u8*)command;
    return (u32*)( cmd8 + m_Size );
}
void
CommandCache::Release()
{
    if( m_Buffer )
    {
        vw::system::Memory_Free( m_Buffer );
    }
    m_Buffer = 0;
    m_Size = 0;
}
void
CommandCache::Finalize()
{
    if( m_Data ) vw::system::Memory_Free( m_Data );
    m_Data = NULL;
}
u32* CommandCache::m_Data = 0;



/*
  
 */
XYZTC::XYZTC() : m_Size(0), m_Data(0), m_VertexData(0), m_IndexData(0)
{
    m_VertexSize = 0;
    m_IndexSize  = 0;
}

XYZTC::~XYZTC()
{
    ReleaseBuffer();
}

bool
XYZTC::CreateBuffer( u32 vertexNum, u32 indexNum )
{
    NN_ASSERT( vertexNum > 0 );
    NN_ASSERT( indexNum > 0 );

    ReleaseBuffer();

    // create buffer
    m_VertexSize = sizeof(Vertex) * vertexNum;
    m_IndexSize  = sizeof(u16)    * indexNum;
    m_Size = m_VertexSize + m_IndexSize;
    m_Data = reinterpret_cast<u8*>( vw::system::Memory_AllocSysDevice(m_Size, 128) );
    m_VertexData = (Vertex*)( m_Data );
    m_IndexData  =    (u16*)( m_Data + m_VertexSize );

    // interleave info
    nn::gr::CTR::Vertex::InterleaveInfo interleave;
    interleave.dataNum = 3;
    interleave.dataType[0] = PICA_DATA_SIZE_3_FLOAT;
    interleave.dataType[1] = PICA_DATA_SIZE_2_FLOAT;
    interleave.dataType[2] = PICA_DATA_SIZE_4_UNSIGNED_BYTE;
    interleave.symbol[0]   = Draw3DHelper::GetInstance().GetSymbolPos();
    interleave.symbol[1]   = Draw3DHelper::GetInstance().GetSymbolTex();
    interleave.symbol[2]   = Draw3DHelper::GetInstance().GetSymbolClr();

    // Sets up vertex information.
    uptr vtx_physical_addr = nngxGetPhysicalAddr( reinterpret_cast< uptr >( m_VertexData ) );
    m_VertexStream.EnableInterleavedArray( interleave, vtx_physical_addr );

    // Sets up index.
    m_IndexStream.physicalAddr = nngxGetPhysicalAddr( reinterpret_cast< uptr >( m_IndexData ) );
    m_IndexStream.drawVtxNum = indexNum;
    m_IndexStream.isUnsignedByte = false;

    // Generate the command cache.
    CreateCommand();
    return true;
}

// Deallocate buffer
void
XYZTC::ReleaseBuffer()
{
    if( m_Data )
    {
        vw::system::Memory_Free( (void*)m_Data );
    }
    m_VertexStream = nn::gr::Vertex();
    m_IndexStream = nn::gr::Vertex::IndexStream();
    m_Size = 0;
    m_Data = 0;
    m_VertexData = 0;
    m_IndexData  = 0;
    m_VertexSize = 0;
    m_IndexSize  = 0;
    m_Command.Release();
}

void
XYZTC::WriteBuffer(const Vertex* vertex, const u16* index)
{
    NN_ASSERT( vertex !=  0 );
    NN_ASSERT( index != 0 );

    std::memcpy( m_VertexData, vertex, m_VertexSize );
    std::memcpy( m_IndexData, index, m_IndexSize );
    nngxUpdateBuffer( reinterpret_cast<void*>(m_Data), m_Size );
}

void
XYZTC::UpdateVertex(u32 start, u32 size)
{
    if( start + size < m_VertexSize / sizeof(Vertex) )
    {
        nngxUpdateBuffer( reinterpret_cast<void*>(m_VertexData + start), sizeof(Vertex)*size );
    }
}

void
XYZTC::UpdateIndex(u32 start, u32 size)
{
    if( start + size < m_IndexSize / sizeof(u16) )
    {
        nngxUpdateBuffer( reinterpret_cast<void*>(m_IndexData + start), sizeof(u16)*size );
    }
}

void
XYZTC::UpdateData()
{
    if( m_Data != 0 )
    {
        nngxUpdateBuffer( reinterpret_cast<void*>(m_Data), m_Size );
    }
}

void
XYZTC::SetColor( float r, float g, float b, float a )
{
    nn::gr::BindSymbolVSInput m_SymbolClr = *Draw3DHelper::GetInstance().GetSymbolClr();
    float color[4] = { r*255.0f, g*255.0f, b*255.0f, a*255.0f };
    m_VertexStream.EnableAttrAsConst( m_SymbolClr, 4, color );

    CreateCommand();
}

void
XYZTC::Draw() const
{
    u32*& command = Draw3DHelper::GetInstance().GetCommand();
    command = m_Command.MakeCommand( command );
}

void
XYZTC::Draw(u32 drawVtxNum)
{
    if( drawVtxNum == 0) return;
    if( drawVtxNum >= m_IndexSize)
        drawVtxNum  = m_IndexSize;

    m_IndexStream.drawVtxNum = drawVtxNum;

    u32*& command = Draw3DHelper::GetInstance().GetCommand();
    command = m_VertexStream.MakeDisableAttrCommand( command );
    command = m_VertexStream.MakeEnableAttrCommand( command );
    command = m_VertexStream.MakeDrawCommand( command, m_IndexStream );
    command = m_VertexStream.MakeDisableAttrCommand( command );
    command = nn::gr::CTR::FrameBuffer::MakeClearCacheCommand( command );
}

// Generates the command cache.
void
XYZTC::CreateCommand()
{
    u32* command = m_Command.Begin();
    command = m_VertexStream.MakeDisableAttrCommand( command );
    command = m_VertexStream.MakeEnableAttrCommand( command );
    command = m_VertexStream.MakeDrawCommand( command, m_IndexStream );
    command = m_VertexStream.MakeDisableAttrCommand( command );
    command = nn::gr::CTR::FrameBuffer::MakeClearCacheCommand( command );
    m_Command.End( command );
}

Draw3DHelper* Draw3DHelper::m_pInstance = NULL;

/*

*/
Draw3DHelper& Draw3DHelper::GetInstance()
{
    NW_ASSERT( m_pInstance != NULL );
    return *m_pInstance;
}

// Constructor
Draw3DHelper::Draw3DHelper()
{
    m_Viewport.Set( 0, 0, 240, 400 );
    m_Scissor.Set( 0, 0, 240, 400 );

    // Load shader binary
    u32 shader_file_size = static_cast<u32>( reinterpret_cast<int>(COMMON_3D_END) - reinterpret_cast<int>(COMMON_3D_BEGIN) );
    m_ShaderBinary = static_cast<u8*>( vw::system::Memory_AllocSysDevice(shader_file_size, 128) );
    memcpy( m_ShaderBinary, COMMON_3D_BEGIN, shader_file_size );

    // Specify shader binary and initialize
    // Enable the executable image at position 0 as the vertex shader
    m_Shader.SetupBinary( m_ShaderBinary, 0, -1 );

    // Search for "Location."
    m_Shader.SearchBindSymbol( &m_SymbolPos, "Position" );
    m_Shader.SearchBindSymbol( &m_SymbolTex, "TexCoord0" );
    m_Shader.SearchBindSymbol( &m_SymbolClr, "Color" );

    // Uniform search
    m_Shader.SearchBindSymbol( &m_SymbolWorldMtx, "user" );
    m_Shader.SearchBindSymbol( &m_SymbolViewMtx,  "view" );
    m_Shader.SearchBindSymbol( &m_SymbolProjMtx,  "proj" );
}

// Destructor
Draw3DHelper::~Draw3DHelper()
{
    if( m_ShaderBinary != 0 )
    {
        vw::system::Memory_Free( m_ShaderBinary );
        m_ShaderBinary = 0;
    }
}

// Initialization
void
Draw3DHelper::Initialize()
{
    if (m_pInstance != NULL) return;
    void *buf = Memory_AllocSys(sizeof(Draw3DHelper));
    if (buf == NULL) return;

    m_pInstance = new(buf) Draw3DHelper();
}

// End processing
void
Draw3DHelper::Finalize()
{
    if (m_pInstance == NULL) return;

    m_pInstance->~Draw3DHelper();
    Memory_Free(m_pInstance);

    Draw3DUtil::Finalize();

    m_pInstance = NULL;
}

// Configures the screen size.
void
Draw3DHelper::SetScreenSize(u32 width, u32 height)
{
    m_Viewport.Set( 0, 0, height, width);
    m_Scissor.Set( 0, 0, height, width);
    m_Scissor.SetBufferSize(height, width);
}
void
Draw3DHelper::Begin()
{
    // Get a pointer to the render command buffer
    nngxGetCmdlistParameteri( NN_GX_CMDLIST_CURRENT_BUFADDR, reinterpret_cast<GLint*>( &m_Command ) );

    // Get the start address (so that the pointer can be advanced at the end)
    m_StartAddr = m_Command;

    // Viewport settings
    m_Command = m_Viewport.MakeCommand( m_Command );

    // Scissor settings
    m_Command = m_Scissor.MakeCommand( m_Command );

    // Shader binary settings ( glUseProgram() )
    m_Command = m_Shader.MakeFullCommand( m_Command );

    // Blend mode settings
    nn::gr::CTR::RenderState rs;
    rs.blend.isEnable = false;
    rs.cullingTest.isEnable = false;
    m_Command = rs.MakeCommand( m_Command );

    // Texture combiner settings
    nn::gr::Combiner combiner;
    combiner.stage[ 0 ].rgb.combine   = PICA_DATA_TEX_ENV_COMBINE_REPLACE;
    combiner.stage[ 0 ].rgb.source[0] = PICA_DATA_TEX_ENV_SRC_RGBA_PRIMARY_COLOR;
    combiner.stage[ 0 ].rgb.source[1] = PICA_DATA_TEX_ENV_SRC_RGBA_FRAGMENT_SECONDARY_COLOR_DMP;
    combiner.stage[ 0 ].alpha.combine = PICA_DATA_TEX_ENV_COMBINE_REPLACE;
    combiner.stage[ 0 ].alpha.source[0] = PICA_DATA_TEX_ENV_SRC_RGBA_PRIMARY_COLOR;
    combiner.stage[ 0 ].alpha.source[1] = PICA_DATA_TEX_ENV_SRC_RGBA_FRAGMENT_SECONDARY_COLOR_DMP;
    m_Command = combiner.MakeCommand( m_Command );

    // Fog configuration
    nn::gr::CTR::Fog fog;
    fog.isEnable = false;
    m_Command = fog.MakeConfigCommand( m_Command );
}
void
Draw3DHelper::SetWorld( const nn::math::MTX34& world )
{
    m_Command = m_SymbolWorldMtx.MakeUniformCommand( m_Command, world );
}
void
Draw3DHelper::SetView( const nn::math::MTX34& view )
{
    m_Command = m_SymbolViewMtx.MakeUniformCommand( m_Command, view );
}
void
Draw3DHelper::SetProj( const nn::math::MTX44& proj )
{
    m_Command = m_SymbolProjMtx.MakeUniformCommand( m_Command, proj );
}
void
Draw3DHelper::End()
{
    // Advance the pointer to the render command buffer
    if( m_Command - m_StartAddr )
    {
        nngxMoveCommandbufferPointer( ( m_Command - m_StartAddr ) * sizeof(u32) );
    }
}

void
Draw3DHelper::Enable(STATE state)
{
    if( state & ALPHATEST )
    {
        m_RenderState.alphaTest.isEnable = true;
        m_RenderState.alphaTest.func = PICA_DATA_ALPHA_TEST_GREATER;
        m_RenderState.alphaTest.refValue = 90;    // For now, the parameter will be the same as that for Draw2DHelper::Enable(ALPHATEST).
        m_Command = m_RenderState.alphaTest.MakeCommand( m_Command );
    }
    if( state & ALPHABLEND )
    {
        m_RenderState.blend.isEnable = true;
        m_RenderState.blend.eqRgb  = PICA_DATA_BLEND_EQUATION_ADD;
        m_RenderState.blend.srcRgb = PICA_DATA_BLEND_FUNC_SRC_ALPHA;
        m_RenderState.blend.dstRgb = PICA_DATA_BLEND_FUNC_ONE_MINUS_SRC_ALPHA;
        m_Command = m_RenderState.blend.MakeCommand( m_Command );
    }
    if( state & CULLFACE )
    {
        m_RenderState.cullingTest.isEnable = true;
        m_Command = m_RenderState.cullingTest.MakeCommand( m_Command );
    }
    if( state & DEPTHWRITE )
    {
        m_RenderState.depthTest.isEnableWrite = true;
        m_Command = m_RenderState.depthTest.MakeCommand( m_Command );
    }
    if ( state & DEPTHTEST )
    {
        m_RenderState.depthTest.isEnable = true;
        m_Command = m_RenderState.depthTest.MakeCommand ( m_Command );
    }
}
void
Draw3DHelper::Disable(STATE state)
{
    if( state & ALPHATEST )
    {
        m_RenderState.alphaTest.isEnable = false;
        m_Command = m_RenderState.alphaTest.MakeCommand( m_Command );
    }
    if ( state & ALPHABLEND )
    {
        m_RenderState.blend.isEnable = false;
        m_Command = m_RenderState.blend.MakeCommand( m_Command );
    }
    if ( state & CULLFACE )
    {
        m_RenderState.cullingTest.isEnable = false;
        m_Command = m_RenderState.cullingTest.MakeCommand( m_Command );
    }
    if ( state & DEPTHWRITE )
    {
        m_RenderState.depthTest.isEnableWrite = false;
        m_Command = m_RenderState.depthTest.MakeCommand( m_Command );
    }
    if ( state & DEPTHTEST )
    {
        m_RenderState.depthTest.isEnable = false;
        m_Command = m_RenderState.depthTest.MakeCommand ( m_Command );
    }
}

vw::system::Draw3DHelper::STATE
Draw3DHelper::GetEnable()
{
    STATE state = 0;
    if ( m_RenderState.alphaTest.isEnable )
    { state |= ALPHATEST; }
    if ( m_RenderState.blend.isEnable )
    { state |= ALPHABLEND; }
    if ( m_RenderState.cullingTest.isEnable )
    { state |= CULLFACE; }
    if ( m_RenderState.depthTest.isEnableWrite )
    { state |= DEPTHWRITE; }
    if ( m_RenderState.depthTest.isEnable )
    { state |= DEPTHTEST; }
    return state;
}

vw::system::Draw3DHelper::STATE
Draw3DHelper::GetDisable()
{
    STATE state = 0;
    if ( !m_RenderState.alphaTest.isEnable )
    { state |= ALPHATEST; }
    if ( !m_RenderState.blend.isEnable )
    { state |= ALPHABLEND; }
    if ( !m_RenderState.cullingTest.isEnable )
    { state |= CULLFACE; }
    if ( !m_RenderState.depthTest.isEnableWrite )
    { state |= DEPTHWRITE; }
    if ( !m_RenderState.depthTest.isEnable )
    { state |= DEPTHTEST; }
    return state;
}


namespace Draw3DUtil
{



// Generates spherical shape.
bool
CreateSphere( XYZTC* model, const nw::math::VEC3& radius, nw::ut::Color8 color, u32 COL_NUM, u32 ROW_NUM )
{
    NN_ASSERT( COL_NUM > 0 );
    NN_ASSERT( ROW_NUM > 0 );

    u32 vertexNum = ROW_NUM * COL_NUM;
    u32 indexNum  = COL_NUM * ( ROW_NUM - 1 ) * 6;
    if( !model->CreateBuffer( vertexNum, indexNum ) )
        return false;

    const f32 delta_row =     nn::math::F_PI / (ROW_NUM - 1);
    const f32 delta_col = 2 * nn::math::F_PI / (COL_NUM - 1);

    XYZTC::Vertex* vertex = model->GetVertexData();
    u16*           index  = model->GetIndexData();

    f32 sin_row, cos_row, sin_col, cos_col;
    for(int row = 0; row < ROW_NUM; row++)
    {
        for(int col = 0; col < COL_NUM; col++)
        {
            XYZTC::Vertex* p_stream = &vertex[ row * COL_NUM + col ];

            nn::math::SinCosRad( &sin_row, &cos_row, delta_row * row );
            nn::math::SinCosRad( &sin_col, &cos_col, delta_col * col );

            p_stream->setPos( sin_row * cos_col,
                              cos_row,
                              sin_row * sin_col );
            p_stream->setColor(color);
            p_stream->setUV(0.0f,0.0f);
        }
    }

    for(int i = 0, row = 0; row < ROW_NUM - 1; row++)
    {
        for(int col = 0; col < COL_NUM; col++)
        {
            index[i++] = COL_NUM*(row)     + (col     );
            index[i++] = COL_NUM*(row)     + (col + 1 );
            index[i++] = COL_NUM*(row + 1) + (col     );
            index[i++] = COL_NUM*(row + 1) + (col     );
            index[i++] = COL_NUM*(row)     + (col + 1 );
            index[i++] = COL_NUM*(row + 1) + (col + 1 );
        }
    }

    for( u32 i=0; i<vertexNum; ++i )
    {
        vertex[i].x *= radius.x;
        vertex[i].y *= radius.y;
        vertex[i].z *= radius.z;
    }

    model->UpdateData();
    return true;
}

// Generates cubical shape.
bool
CreateCube( XYZTC* model, const nw::math::VEC3& radius, nw::ut::Color8 color )
{
    const f32 s_BoxVtxStreamPos[] =
    {
        -1.0f,  1.0f,  1.0f,   -1.0f,  1.0f, -1.0f,   -1.0f, -1.0f, -1.0f,
        -1.0f, -1.0f,  1.0f,   -1.0f,  1.0f,  1.0f,   -1.0f, -1.0f, -1.0f,
         1.0f,  1.0f,  1.0f,   -1.0f,  1.0f,  1.0f,   -1.0f, -1.0f,  1.0f,
         1.0f, -1.0f,  1.0f,    1.0f,  1.0f,  1.0f,   -1.0f, -1.0f,  1.0f,
         1.0f,  1.0f, -1.0f,    1.0f,  1.0f,  1.0f,    1.0f, -1.0f, -1.0f,
         1.0f, -1.0f, -1.0f,    1.0f,  1.0f,  1.0f,    1.0f, -1.0f,  1.0f,
        -1.0f,  1.0f, -1.0f,    1.0f,  1.0f, -1.0f,   -1.0f, -1.0f, -1.0f,
         1.0f,  1.0f, -1.0f,    1.0f, -1.0f, -1.0f,   -1.0f, -1.0f, -1.0f,
        -1.0f, -1.0f, -1.0f,    1.0f, -1.0f, -1.0f,   -1.0f, -1.0f,  1.0f,
        -1.0f, -1.0f,  1.0f,    1.0f, -1.0f, -1.0f,    1.0f, -1.0f,  1.0f,
        -1.0f,  1.0f, -1.0f,   -1.0f,  1.0f,  1.0f,    1.0f,  1.0f, -1.0f,
        -1.0f,  1.0f,  1.0f,    1.0f,  1.0f,  1.0f,    1.0f,  1.0f, -1.0f,
     };
    const u8 s_BoxVtxStreamIdx[] =
    {
         0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11,
        12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23,
        24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35
    };

    u32 vertexNum = 36;
    u32 indexNum  = 36;
    if( !model->CreateBuffer( vertexNum, indexNum ) )
        return false;

    XYZTC::Vertex* vertex = model->GetVertexData();
    u16*           index  = model->GetIndexData();
    for(u32 i=0; i<36; ++i)
    {
        u32 j=i*3;
        vertex[i].setPos( s_BoxVtxStreamPos[j+0],
                          s_BoxVtxStreamPos[j+1],
                          s_BoxVtxStreamPos[j+2] );
        vertex[i].setColor( color );
        index[i] = s_BoxVtxStreamIdx[i];
    }

    for( u32 i=0; i<vertexNum; ++i )
    {
        vertex[i].x *= radius.x;
        vertex[i].y *= radius.y;
        vertex[i].z *= radius.z;
    }

    model->UpdateData();

    return true;
}

// Generates disk shape.
bool
CreateDisc( XYZTC* model, const nw::math::VEC3& radius, nw::ut::Color8 color, u32 div )
{
    NN_ASSERT( div > 0 );

    u32 vertexNum = div + 1;
    u32 indexNum  = div * 3;
    if( !model->CreateBuffer( vertexNum, indexNum ) )
        return false;

    XYZTC::Vertex* vertex = model->GetVertexData();
    u16*           index  = model->GetIndexData();

    vertex[div].setPos(0.0f,0.0f,0.0f).setColor(0xFFFFFFFF);

    f32 delta = 2 * nn::math::F_PI / (div);
    f32 x, y;
    for( u32 i=0; i<div; ++i )
    {
        nn::math::SinCosRad( &y, &x, delta * i );
        vertex[i].setPos(x,0.0f,y).setColor(color);

        index[i*3+0] = div;
        index[i*3+1] =(i+1)%div;
        index[i*3+2] = i;
    }

    for( u32 i=0; i<vertexNum; ++i )
    {
        vertex[i].x *= radius.x;
        vertex[i].y *= radius.y;
        vertex[i].z *= radius.z;
    }

    model->UpdateData();

    return true;
}

// Generates cylindrical shape.
bool
CreateCylinder( XYZTC* model, const nw::math::VEC3& radius, nw::ut::Color8 color, u32 div )
{
    NN_ASSERT( div > 0 );

    u32 vertexNum = div*2 + 2;
    u32 indexNum  = div*3 + div*3 + div*6;
    if( !model->CreateBuffer( vertexNum, indexNum ) )
        return false;

    XYZTC::Vertex* vertex = model->GetVertexData();
    u16*           index  = model->GetIndexData();

    vertex[div].    setPos(0.0f, 1.0f,0.0f).setColor(color);
    vertex[div*2+1].setPos(0.0f,-1.0f,0.0f).setColor(color);

    f32 delta = 2.0f * nn::math::F_PI / div;
    f32 x, y;
    for(u32 i=0; i<div; ++i)
    {
        nn::math::SinCosRad( &y, &x, delta * i );

        u32 j=i;
        vertex[j].      setPos(x, 1.0f,y).setColor(color);
        vertex[div+1+j].setPos(x,-1.0f,y).setColor(color);

        j = i*3;
        index[j+0] = div;
        index[j+1] =(i+1)%div;
        index[j+2] = i;

        j = div*3 + i*3;
        index[j+0] = div*2+1;
        index[j+1] = div+1+i;
        index[j+2] = div+1+(i+1)%div;

        j = (div+i)*6;
        index[j+0] = i;
        index[j+4] = index[j+1] =(i+1)%div;
        index[j+3] = index[j+2] = div+1+i;
        index[j+5] = div+1+(i+1)%div;
    }

    for( u32 i=0; i<vertexNum; ++i )
    {
        vertex[i].x *= radius.x;
        vertex[i].y *= radius.y;
        vertex[i].z *= radius.z;
    }

    model->UpdateData();

    return true;
}

// Generates XZ rectangular shape.
bool
CreateRectangle( XYZTC* model, const nw::math::VEC3& radius, nw::ut::Color8 color )
{
    if( radius.x < 0.0f || radius.y < 0.0f || radius.z < 0.0f )
        return false;

    const f32 s_BoxVtxStreamPos[] =
    {
        -1.0f, 0.0f, -1.0f,    1.0f,  0.0f, -1.0f,   -1.0f, 0.0f,  1.0f,
        -1.0f, 0.0f,  1.0f,    1.0f,  0.0f, -1.0f,    1.0f, 0.0f,  1.0f
    };
    if( !model->CreateBuffer( 6, 6 ) )
        return false;

    XYZTC::Vertex* vertex = model->GetVertexData();
    u16*           index  = model->GetIndexData();
    for(u32 i=0; i<6; ++i)
    {
        u32 j=i*3;
        vertex[i].x = s_BoxVtxStreamPos[j+0] * radius.x;
        vertex[i].y = s_BoxVtxStreamPos[j+1] * radius.y;
        vertex[i].z = s_BoxVtxStreamPos[j+2] * radius.z;
        vertex[i].color = color;
        index[i] = i;
    }
    model->UpdateData();

    return true;
}

bool
CreateCone( XYZTC* model, const nw::math::VEC3& radius, nw::ut::Color8 color, u32 div )
{
    NN_ASSERT( div > 0 );

    u32 vertexNum = div + 2;
    u32 indexNum  = div * 6;
    if( !model->CreateBuffer( vertexNum, indexNum ) )
        return false;

    XYZTC::Vertex* vertex = model->GetVertexData();
    u16*           index  = model->GetIndexData();

    vertex[div].setPos(0.0f,-1.0f,0.0f).setColor(0xFFFFFFFF);
    vertex[div+1].setPos(0.0f,1.0f,0.0f).setColor(0xFFFFFFFF);

    f32 delta = 2 * nn::math::F_PI / (div);
    f32 x, y;
    for( u32 i=0; i<div; ++i )
    {
        nn::math::SinCosRad( &y, &x, delta * i );
        vertex[i].setPos(x,-1.0f,y).setColor(color);

        index[i*6+0] = div;
        index[i*6+1] =(i+1)%div;
        index[i*6+2] = i;

        index[i*6+3] = div+1;
        index[i*6+4] = i;
        index[i*6+5] =(i+1)%div;
    }

    for( u32 i=0; i<vertexNum; ++i )
    {
        vertex[i].x *= radius.x;
        vertex[i].y *= radius.y;
        vertex[i].z *= radius.z;
    }

    model->UpdateData();

    return true;
}


static XYZTC* sSphere = NULL;
static XYZTC* sCube = NULL;
static XYZTC* sDisc = NULL;
static XYZTC* sCylinder = NULL;
static XYZTC* sRectangle = NULL;
static XYZTC* sCone = NULL;

// Sets the default of shape rendering primitive.
void SetupSphere( u32 COL_NUM, u32 ROW_NUM )
{
    if( sSphere == NULL )
    {
        void *buf = vw::system::Memory_AllocSys( sizeof(XYZTC) );
        sSphere = new(buf) XYZTC();
        CreateSphere( sSphere, nw::math::VEC3(1.0f,1.0f,1.0f), 0, COL_NUM, ROW_NUM );
    }
}
void SetupDisc( u32 div )
{
    if( sDisc == NULL )
    {
        void *buf = vw::system::Memory_AllocSys( sizeof(XYZTC) );
        sDisc = new(buf) XYZTC();
        CreateDisc( sDisc, nw::math::VEC3(1.0f,1.0f,1.0f), 0, div );
    }
}
void SetupCylinder( u32 div )
{
    if( sCylinder == NULL )
    {
        void *buf = vw::system::Memory_AllocSys( sizeof(XYZTC) );
        sCylinder = new(buf) XYZTC();
        CreateCylinder( sCylinder, nw::math::VEC3(1.0f,1.0f,1.0f), 0, div );
    }
}

void SetWorldMatrix( const nw::math::MTX34& world, const nw::math::VEC3& radius )
{
    nw::math::MTX34 scale;
    nw::math::MTX34Scale( &scale, radius );
    nw::math::MTX34Mult( &scale, &world, &scale );
    Draw3DHelper::GetInstance().SetWorld( scale );
}

void Finalize()
{
    if( sSphere )
    {
        sSphere->~XYZTC();
        Memory_Free( sSphere );
        sSphere = NULL;
    }
    if( sCube )
    {
        sCube->~XYZTC();
        Memory_Free( sCube );
        sCube = NULL;
    }
    if( sDisc )
    {
        sDisc->~XYZTC();
        Memory_Free( sDisc );
        sDisc = NULL;
    }
    if( sCylinder )
    {
        sCylinder->~XYZTC();
        Memory_Free( sCylinder );
        sCylinder = NULL;
    }
    if( sRectangle )
    {
        sRectangle->~XYZTC();
        Memory_Free( sRectangle );
        sRectangle = NULL;
    }
    if( sCone )
    {
        sCone->~XYZTC();
        Memory_Free( sCone );
        sCone = NULL;
    }
}

// Shape rendering function
void
DrawSphere( const nw::math::MTX34& world, const nw::math::VEC3& radius, const nw::ut::FloatColor& color )
{
    if( sSphere == NULL )
    {
        void *buf = vw::system::Memory_AllocSys( sizeof(XYZTC) );
        sSphere = new(buf) XYZTC();
        CreateSphere( sSphere, nw::math::VEC3(1.0f,1.0f,1.0f), 0, 20, 20 );
    }

    if( sSphere )
    {
        SetWorldMatrix( world, radius );
        sSphere->SetColor( color.r, color.g, color.b, color.a );
        sSphere->Draw();
    }
}
void
DrawCube( const nw::math::MTX34& world, const nw::math::VEC3& radius, const nw::ut::FloatColor& color )
{
    if( sCube == NULL )
    {
        void *buf = vw::system::Memory_AllocSys( sizeof(XYZTC) );
        sCube = new(buf) XYZTC();
        CreateCube( sCube, nw::math::VEC3(1.0f,1.0f,1.0f), 0 );
    }

    if( sCube )
    {
        SetWorldMatrix( world, radius );
        sCube->SetColor( color.r, color.g, color.b, color.a );
        sCube->Draw();
    }
}

void
DrawDisc( const nw::math::MTX34& world, const nw::math::VEC3& radius, const nw::ut::FloatColor& color )
{
    if( sDisc == NULL )
    {
        void *buf = vw::system::Memory_AllocSys( sizeof(XYZTC) );
        sDisc = new(buf) XYZTC();
        CreateDisc( sDisc, nw::math::VEC3(1.0f,1.0f,1.0f), 0, 20 );
    }

    if( sDisc )
    {
        SetWorldMatrix( world, radius );
        sDisc->SetColor( color.r, color.g, color.b, color.a );
        sDisc->Draw();
    }
}

void
DrawCylinder( const nw::math::MTX34& world, const nw::math::VEC3& radius, const nw::ut::FloatColor& color )
{
    if( sCylinder == NULL )
    {
        void *buf = vw::system::Memory_AllocSys( sizeof(XYZTC) );
        sCylinder = new(buf) XYZTC();
        CreateCylinder( sCylinder, nw::math::VEC3(1.0f,1.0f,1.0f), 0, 20 );
    }

    if( sCylinder )
    {
        SetWorldMatrix( world, radius );
        sCylinder->SetColor( color.r, color.g, color.b, color.a );
        sCylinder->Draw();
    }
}

void
DrawRectangle( const nw::math::MTX34& world, const nw::math::VEC3& radius, const nw::ut::FloatColor& color )
{
    if( sRectangle == NULL )
    {
        void *buf = vw::system::Memory_AllocSys( sizeof(XYZTC) );
        sRectangle = new(buf) XYZTC();
        CreateRectangle( sRectangle, nw::math::VEC3(1.0f,1.0f,1.0f), 0 );
    }

    if( sRectangle )
    {
        SetWorldMatrix( world, radius );
        sRectangle->SetColor( color.r, color.g, color.b, color.a );
        sRectangle->Draw();
    }

}

void
DrawCone( const nw::math::MTX34& world, const nw::math::VEC3& radius, const nw::ut::FloatColor& color )
{
    if( sCone == NULL )
    {
        void *buf = vw::system::Memory_AllocSys( sizeof(XYZTC) );
        sCone = new(buf) XYZTC();
        CreateCone( sCone, nw::math::VEC3(1.0f,1.0f,1.0f), 0 );
    }

    if( sCone )
    {
        SetWorldMatrix( world, radius );
        sCone->SetColor( color.r, color.g, color.b, color.a );
        sCone->Draw();
    }
}


}

} /* namespace NW4CViewer */
} /* namespace nw   */
