/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/
#include <viewer/system.h>
#include <string.h>
#include <cstdio>       // for vsnprintf
#include "../include/LogUtil.h"

namespace vw     {
namespace system {
namespace {

bool s_IsDrawing = false;

f32	 s_X=0.0f, s_Y=0.0f;
nw::ut::Color8 s_FontColor8 = nw::ut::Color8::WHITE;
f32  s_FontSize;

static const int TEXT_BUF_SIZE = 4096;
char s_TextBuf[ TEXT_BUF_SIZE ];

f32 s_LineWidth = 2.0f;

GLuint  s_ProgramId;
GLuint  s_TextureObj;
GLuint  s_ArrayBufferObj;
GLuint  s_ElementArrayBufferObj;

void RunCmdlistDone(void);

void RunCmdlistDone(void)
{
    // Separate the 3D commands, and begin execution.
    nngxSplitDrawCmdlist();
    nngxRunCmdlist();
    while (1)
    {
        GLint param;
        nngxGetCmdlistParameteri(NN_GX_CMDLIST_IS_RUNNING, &param);
        if ( param == 0)
            break;
    }
}

//======================================================================//
//						Alloc / Free
//======================================================================//
//Alloc file
//		Allocate a region of the size of the file and load the file
// 		After using, deallocate with the Free function.
void * AllocFile( const char * path, size_t * pReadSize=NULL )
{
    nn::fs::FileInputStream fileReader;
    if ( ! fileReader.TryInitialize( path ).IsSuccess() )
    {
		NW_VIEWER_LOG("Error:Can't open file.(%s)\n", path );
		NW_ASSERT(0);
		return NULL;
	}

	const size_t fileSize = (size_t)fileReader.GetSize();
	void * const buf = Memory_Alloc( fileSize, 128 );
	const s32 readSize = fileReader.Read( buf, fileSize );
	NW_ASSERT( readSize == fileSize );
	if ( pReadSize ) (*pReadSize) = fileSize;
	return buf;
}

//=======================================================================/
//						Uniform location
//=======================================================================/
//There is a small cost for getting the location because of the string comparison.
//  Since the location does not change for each program, the location is saved to increase speed
int s_loc_FragmentLightingEnabled			= 0;
int s_loc_FogMode							= 0;
int s_Loc_FragOperation_EnableAlphaTest 	= 0;
int s_Loc_Texure0_SamplerType 				= 0;

int s_Loc_TexEnv2_combineRgb 				= 0;
int s_Loc_TexEnv2_combineAlpha 				= 0;
int s_Loc_TexEnv2_SrcRgb 					= 0;
int s_Loc_TexEnv2_SrcAlpha 					= 0;
int s_Loc_TexEnv2_OperandRgb 				= 0;
int s_Loc_TexEnv2_OperandAlpha 				= 0;
int s_Loc_TexEnv2_scaleRgb					= 0;
int s_Loc_TexEnv2_scaleAlpha				= 0;
int s_Loc_TexEnv2_ConstRgba 				= 0;

void getLoc( int * loc, const char *name ) //Save one location
{
	(*loc) = glGetUniformLocation( s_ProgramId, name );
	NW_ASSERT( (*loc) > 1 ); //Fails to get if the name is mistaken
}

void SaveLocations() //Save location (only once after creating the program)
{
	getLoc( & s_loc_FragmentLightingEnabled
										, "dmp_FragmentLighting.enabled" );
	getLoc( & s_loc_FogMode				, "dmp_Fog.mode" );
	getLoc( & s_Loc_FragOperation_EnableAlphaTest
										, "dmp_FragOperation.enableAlphaTest" );
	getLoc( & s_Loc_Texure0_SamplerType	, "dmp_Texture[0].samplerType" );

	getLoc( & s_Loc_TexEnv2_combineRgb	, "dmp_TexEnv[2].combineRgb"	);
	getLoc( & s_Loc_TexEnv2_combineAlpha, "dmp_TexEnv[2].combineAlpha"	);
	getLoc( & s_Loc_TexEnv2_SrcRgb		, "dmp_TexEnv[2].srcRgb" 		);
	getLoc( & s_Loc_TexEnv2_SrcAlpha	, "dmp_TexEnv[2].srcAlpha" 		);
	getLoc( & s_Loc_TexEnv2_OperandRgb	, "dmp_TexEnv[2].operandRgb"	);
	getLoc( & s_Loc_TexEnv2_OperandAlpha, "dmp_TexEnv[2].operandAlpha" 	);
	getLoc( & s_Loc_TexEnv2_scaleRgb	, "dmp_TexEnv[2].scaleRgb" 		);
	getLoc( & s_Loc_TexEnv2_scaleAlpha	, "dmp_TexEnv[2].scaleAlpha" 	);
	getLoc( & s_Loc_TexEnv2_ConstRgba	, "dmp_TexEnv[2].constRgba" 	);
}

void SetupLocations() //Write setting values to each location (for each Begin)
{
	//Fragment lighting invalidation
	glUniform1i(  s_loc_FragmentLightingEnabled			, GL_FALSE);
	//Invalidate fog
    glUniform1i(  s_loc_FogMode							, GL_FALSE);
	//Invalidate alpha test
	glUniform1i(  s_Loc_FragOperation_EnableAlphaTest	, GL_FALSE);
	//Set texture unit 0 to a 2D texture
	glUniform1i(  s_Loc_Texure0_SamplerType				, GL_TEXTURE_2D);
	//Texture combiner settings
	{
		{ //Input
								  //Input0: Polygon; Input1: Texture; Input2: Unused
			static const GLint src[] = { GL_PRIMARY_COLOR, GL_TEXTURE0, GL_PREVIOUS };
		    glUniform3iv( s_Loc_TexEnv2_SrcRgb    , 1, src );
		    glUniform3iv( s_Loc_TexEnv2_SrcAlpha  , 1, src );
		}

		{ //Combining function
			//RGB is polygon color only
		    glUniform1i(  s_Loc_TexEnv2_combineRgb , GL_REPLACE );
		    //Alpha is polygon x texture
		    glUniform1i( s_Loc_TexEnv2_combineAlpha, GL_MODULATE);
		}
	}
}


//======================================================================//
//						Render request stream
//======================================================================//
/*
	The affect on the processing time is greater for how may times the glDrawElements function is called than for how many triangles are rendered.
	
	It is faster with larger stream sizes, but the effect decreases as the number increases.
	
*/
const int MAX_TRIANGLE_NUM = 4000; //46000 bytes with 1000 items
const int MAX_IDX_STREAM_N = 3 * MAX_TRIANGLE_NUM;
const int MAX_VTX_STREAM_N = 2 * MAX_TRIANGLE_NUM;

struct Vtx { //Vertex information
	nw::math::VEC2 pos;
	nw::ut::Color8 color;
	nw::math::VEC2 texcoord;
};

Vtx s_VtxStream[MAX_VTX_STREAM_N]; //Vertex stream
u16 s_IdxStream[MAX_IDX_STREAM_N]; //Vertex index stream
int s_VtxStreamN = 0; //Number of vertices accumulated in the vertex stream
int s_IdxStreamN = 0; //Number of vertex indices accumulated in the vertex index stream

void InitStream() //Initialize vertex stream
{
	NN_LOG("stream size = %d [byte]\n", sizeof( s_VtxStream ) + sizeof( s_IdxStream ) );
	s_VtxStreamN = 0;
	s_IdxStreamN = 0;
}

void FlushStream() //Render accumulated render requests all at once
{
	if ( s_IdxStreamN > 0 ) {

		//Wait if the previous graphics command has not executed
		RunCmdlistDone();

		// Load vertex information to the bound VBO.
		glBufferData(GL_ARRAY_BUFFER,
				sizeof(Vtx) * s_VtxStreamN,
				s_VtxStream,
				GL_STATIC_DRAW );

		// Load vertex indices to the bound VBO.
		glBufferData(GL_ELEMENT_ARRAY_BUFFER,
	    		sizeof(u16) * s_IdxStreamN ,
	    		s_IdxStream,
	    		GL_STATIC_DRAW);

		{ //Data set from fixed values or VBO to vertex attributes

		    glEnableVertexAttribArray(0); //pos.xy
		    glVertexAttribPointer(0, 2, GL_FLOAT, GL_FALSE, sizeof(Vtx), 0);

		    glDisableVertexAttribArray(1); //pos.z
			glVertexAttrib1f(1, 0.0f);

		    glEnableVertexAttribArray(2); //color
		    glVertexAttribPointer(2, 4, GL_UNSIGNED_BYTE, GL_FALSE, sizeof(Vtx),
		    	reinterpret_cast<const void *>( offsetof(Vtx, color) ) );

		    glEnableVertexAttribArray(3); //tex
		    glVertexAttribPointer(3, 2, GL_FLOAT, GL_FALSE, sizeof(Vtx),
		    	reinterpret_cast<const void *>( offsetof(Vtx, texcoord) ) );
		}

		//Render
		glDrawElements(GL_TRIANGLES, s_IdxStreamN, GL_UNSIGNED_SHORT, 0);
		glGetError(); //Ignore even if an error occurs.


	    //RunCmdlistDone(); // Efficiency is bad if wait here
		//nngxSplitDrawCmdlist();

	}
	s_VtxStreamN = 0;
	s_IdxStreamN = 0;
}

void EntryStream( //Set render request in stream (if full, automatic rendering)
		const Vtx * 	verteces,
		int 			vtx_n,
		const u16 *		indeces,
		int 			idx_n
	)
{
	NW_ASSERT( vtx_n <= idx_n );
	NW_ASSERT( idx_n % 3 == 0 );

	static const int SAMETIME_SET_IDX_N_MAX = 6; //Maximum number of vertices that can be set at the same time
	NW_ASSERT( idx_n <= SAMETIME_SET_IDX_N_MAX );

	const int idxBase = s_VtxStreamN;
	//Vertex information set
	for( int i=0; i<vtx_n; ++i ) {
		s_VtxStream[ s_VtxStreamN ++ ] = verteces[i];
	}
	//Vertex index set
	for( int i=0; i<idx_n; ++i ) {
		s_IdxStream[ s_IdxStreamN ++ ] = static_cast<u16>( indeces[i] + idxBase );
	}

	//Render all at once if the stream is full
	if ( 	s_IdxStreamN >= MAX_IDX_STREAM_N - SAMETIME_SET_IDX_N_MAX
	   ||   s_VtxStreamN >= MAX_VTX_STREAM_N - SAMETIME_SET_IDX_N_MAX
	) {
		FlushStream();
	}
}

//======================================================================//
//							Texture
//======================================================================//
static const f32 FONT_BASE_SIZE = 12.0f;
static const s32 BMP_WIDTH  = 256;
static const s32 BMP_HEIGHT = 256;
static const s32 BMP_PIX_BIT = 24;

inline f32 GetPolyWidth()  { return s_FontSize * ( 9.0f / FONT_BASE_SIZE); }
inline f32 GetPolyHeight() { return s_FontSize * (14.0f / FONT_BASE_SIZE); }
inline f32 GetCursorDx()   { return s_FontSize * (9.5f  / FONT_BASE_SIZE); }
inline f32 GetCursorDy()   { return s_FontSize * (14.0f / FONT_BASE_SIZE); }

inline nw::ut::Color8 getPixRgb( const u8 *bmp, int x, int y ) //Get BMP pixel color
{
	static const int OFS = 0x36; //Offset from the start to the image data
	const int revY = BMP_HEIGHT - 1 - y; //BMP is usually input with the top/bottom reversed.
	const u8 * const pix = bmp + OFS +
				(x + revY * BMP_WIDTH) * (BMP_PIX_BIT / 8);
	return nw::ut::Color8( pix[2], pix[1], pix[0], 255 );
}

//Addressing conversion (OpenGL->Native)
static int OpenGlUVToNativeAdrs( int u, int v )
{
	const int blockNo = u / 8 + (v / 8) * (BMP_WIDTH / 8);
	const int bu = u % 8;
	const int bv = v % 8;
	const u8 BlockUvToOfsAdrs[8][8] = {
		{  0, 1, 4, 5,16,17,20,21 },
		{  2, 3, 6, 7,18,19,22,23 },
		{  8, 9,12,13,24,25,28,29 },
		{ 10,11,14,15,26,27,30,31 },
		{ 32,33,36,37,48,49,52,53 },
		{ 34,35,38,39,50,51,54,55 },
		{ 40,41,44,45,56,57,60,61 },
		{ 42,43,46,47,58,59,62,63 }
	};
	return blockNo * 64 + BlockUvToOfsAdrs[bv][bu];
}

void LoadTexture() //Load fonts (24-bit BMP) from file and perform up to the glTexImage2D function
{
	//Load BMP files
	u8 * const bmp = reinterpret_cast<u8 *>(
				AllocFile( "rom:/sampleFont.bmp" ) );
	//Allocate texture image temporary creation region
	size_t texSize = BMP_WIDTH * BMP_HEIGHT;
	u8 * const tex = reinterpret_cast<u8 *>(
				Memory_Alloc( texSize, 4 ) );
	const bool onVram = true;
	if ( onVram )
	{
		//Format conversion: 24bit BMP -> Native A8 (Alpha 8bit)
		for( int y=0; y<BMP_HEIGHT; ++y ) {
			for( int x=0; x<BMP_WIDTH; ++x ) {
				nw::ut::Color8 pixRgb = getPixRgb( bmp, x, y );
				const u8 alpha = static_cast<u8>( 255 - pixRgb.r );
				tex[ OpenGlUVToNativeAdrs( x, BMP_HEIGHT - 1 - y ) ] = alpha;
			}
		}
		//Transfer created texture to the graphics library
		const GLenum FORMAT = GL_ALPHA_NATIVE_DMP;
		glTexImage2D(GL_TEXTURE_2D | NN_GX_MEM_VRAMA | GL_NO_COPY_FCRAM_DMP,
			0, FORMAT, BMP_WIDTH, BMP_HEIGHT, 0, FORMAT, GL_UNSIGNED_BYTE, tex );
	}
	else
	{
		//Format conversion: 24bit BMP -> A8 (Alpha 8bit)
		for( int y=0; y<BMP_HEIGHT; ++y ) {
			for( int x=0; x<BMP_WIDTH; ++x ) {
				nw::ut::Color8 pixRgb = getPixRgb( bmp, x, y );
				const u8 alpha = static_cast<u8>( 255 - pixRgb.r );
				tex[ x + y * BMP_WIDTH ] = alpha; //openGL
			}
		}
		//Transfer created texture to the graphics library
		const GLenum FORMAT = GL_ALPHA;
		glTexImage2D(GL_TEXTURE_2D,
			0, FORMAT, BMP_WIDTH, BMP_HEIGHT, 0, FORMAT, GL_UNSIGNED_BYTE, tex );
	}

	NW_GL_ASSERT();

	//Release BMP file and texture creation temporary region
	Memory_Free( tex );
	Memory_Free( bmp );
}

//ASCII code -> texture coordinates
void CharToTexPos( int c, nw::math::VEC2 *posLT, nw::math::VEC2 *posRT,
                          nw::math::VEC2 *posLB, nw::math::VEC2 *posRB )
{
	using namespace nw::math;
	static const f32 CELL_X0 =  2.0f / BMP_WIDTH  ; //Cell horizontal offset
	static const f32 CELL_Y0 =  2.0f / BMP_HEIGHT ; //Cell vertical offset
	static const f32 CELL_W  =  9.0f / BMP_WIDTH  ; //Font width
	static const f32 CELL_H  = 14.0f / BMP_HEIGHT ; //Font height
	static const f32 BLOCK_W = 13.0f / BMP_WIDTH  ; //Block width
	static const f32 BLOCK_H = 20.0f / BMP_HEIGHT ; //Block height
	static const VEC2 ofsLT = VEC2( CELL_X0         , CELL_Y0 		   );
	static const VEC2 ofsRT = VEC2( CELL_X0 + CELL_W, CELL_Y0 		   );
	static const VEC2 ofsLB = VEC2( CELL_X0         , CELL_Y0 + CELL_H );
	static const VEC2 ofsRB = VEC2( CELL_X0 + CELL_W, CELL_Y0 + CELL_H );

	c -= 0x20;
	const f32 xn = c % 16;
	const s32 yn = c / 16;
	VEC2 blockPos = VEC2( xn * BLOCK_W, yn * BLOCK_H );

	(*posLT) = blockPos + ofsLT;
	(*posRT) = blockPos + ofsRT;
	(*posLB) = blockPos + ofsLB;
	(*posRB) = blockPos + ofsRB;
}

//======================================================================//
//							Initialization
//======================================================================//
void Initialize()
{
	//Do nothing if already initialized
	static bool isInitialized = false;
	if ( isInitialized ) return;
	isInitialized = true;

	//Initialize font size
    s_FontSize = FONT_BASE_SIZE;

	//Initialize triangle rendering request stream
	InitStream();

	//Program appropriates text writer one
	GLint programId;
	glGetIntegerv(  GL_CURRENT_PROGRAM, & programId );
	s_ProgramId = programId;

    { // Load font texture
	    glGenTextures(1, &s_TextureObj);
	    glActiveTexture(GL_TEXTURE0);
	    glBindTexture(GL_TEXTURE_2D, s_TextureObj);
		LoadTexture(); //Load fonts from file and perform up to the glTexImage2D function
	}

    // Create vertex and vertex index VBO (vertex buffer object)
    glGenBuffers(1, &s_ArrayBufferObj);
    glGenBuffers(1, &s_ElementArrayBufferObj);

	//Save matrix and uniform location for shader
	SaveLocations();

	//Restore to unused state
    glBindTexture(GL_TEXTURE_2D, 0); 			//Set texture to unused state
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0 ); 	//Remove association with vertex index buffer.

	NW_GL_ASSERT();
}

//======================================================================//
//							Render one character
//======================================================================//
//Set render request for one character in the stream
void DrawChar(f32 x, f32 y, char c)
{
	//Vertex information set
	Vtx verteces[4]; //0: Upper left; 1: Upper right; 2: Lower left, 3: Lower right
	{
		//Display coordinates
		const f32 x2 = x + GetPolyWidth();
		const f32 y2 = y + GetPolyHeight();
		verteces[0].pos.Set( x , y  );
		verteces[1].pos.Set( x2, y  );
		verteces[2].pos.Set( x , y2 );
		verteces[3].pos.Set( x2, y2 );

		//Color
		for( int i=0; i<4; ++i ) verteces[i].color = s_FontColor8;

		//Texture coordinates
		CharToTexPos( c, 	& verteces[0].texcoord,
						 	& verteces[1].texcoord,
						 	& verteces[2].texcoord,
						 	& verteces[3].texcoord
					);
	}
	//Vertex index
	static const u16 indeces[6] = { 0, 1, 2,	2, 1, 3 };
	//Render request stream set
	EntryStream( verteces, 4, indeces, 6 );
}

} /* namespace anonymous */

///////////////////////////////////////////////////////////////////////////
//							Global functions
///////////////////////////////////////////////////////////////////////////


//======================================================================//
//						Start/stop rendering
//======================================================================//
//Begin rendering
void Begin()
{
 	NW_ASSERT( ! s_IsDrawing );
 	s_IsDrawing = true;

	NW_GL_ASSERT();

	//Initialize only if not initialized
	Initialize();

	//Text parameter reset
	SetTextColor( nw::ut::Color8::WHITE );
	SetFontSize( 12.0f );
	SetTextPos( 0.0f, 0.0f );

	//Note: Programs and matrices appropriate the text writer one.

    // Turn off features not used by GL (depth test, culling, logical operations)
	glDisable(GL_DEPTH_TEST);
	glDisable(GL_CULL_FACE);
	glDisable(GL_COLOR_LOGIC_OP);

	//Associate vertex buffer
    glBindBuffer(GL_ARRAY_BUFFER, s_ArrayBufferObj);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, s_ElementArrayBufferObj);

    //Texture settings
    glActiveTexture(GL_TEXTURE0); //Enable texture0
    glBindTexture(GL_TEXTURE_2D, s_TextureObj); //Bind texture (data is read)
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR); //Linear interpolation is on during enlarging
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR); //Linear interpolation is on during reduction

	//Uniform settings
	SetupLocations();

	//Color blend settings
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glBlendEquation(GL_FUNC_ADD);

	NW_GL_ASSERT();
}

//Rendering ended
void End()
{
	//Render if render requests remain
	FlushStream();

 	NW_ASSERT( s_IsDrawing );
 	if ( ! s_IsDrawing ) return; //So warnings are not output with resource mode
 	s_IsDrawing = false;

	//Reset buffer type associations
    glBindBuffer(GL_ARRAY_BUFFER, 0);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
    glBindTexture(GL_TEXTURE_2D, 0);
    glDisableVertexAttribArray(0);
    glDisableVertexAttribArray(1);
    glDisableVertexAttribArray(2);
    glDisableVertexAttribArray(3);

	NW_GL_ASSERT();

}
//======================================================================//
//							Text display
//======================================================================//
//Color
void SetTextColor( nw::ut::Color8 color )
{
	s_FontColor8 = color;
}
nw::ut::Color8 GetTextColor()
{
	return s_FontColor8;
}

//Font size
void SetFontSize( f32 fontSize )
{
	s_FontSize = fontSize;
}
f32 GetFontSize()
{
	return s_FontSize;
}

//Cursor position
void SetTextPos( f32 x, f32 y )
{
	s_X = x;
	s_Y = y;
}
nw::math::VEC2 GetTextPos()
{
	return nw::math::VEC2( s_X, s_Y );
}

//Print series
void Printf( const char *format, ... )
{
    va_list vlist;

    va_start(vlist, format );
 	VPrintf( format, vlist );
    va_end(vlist);
}

void VPrintf( const char *format, std::va_list vlist )
{
	s32 stringSize = vsnprintf(s_TextBuf, TEXT_BUF_SIZE, format, vlist);

	if (stringSize < 0)
    {
        NW_VIEWER_LOG("[WARN] input string is too long.\n");
        return;
    }

    Print( s_TextBuf );
}

void Print( const char *str )
{
	NW_POINTER_ASSERT( str );
	NW_ASSERT( s_IsDrawing );

	const f32 x0 = s_X; //X-coordinate restore position after carriage return

	for(;;) {
		const char c = *str++;

		if ( 0x20 <= c && c <= 0x7F ) { //1 character output

			if ( c != 0x20 ) { //Half space only moves the cursor
				DrawChar( s_X, s_Y, c ); //Register render request for one character in the stream
			}
			s_X += GetCursorDx();

		} else if ( c == '\n' ) { //Carriage return?

			s_X = x0;
			s_Y += GetCursorDy();

		} else if ( c == '\0' ) {

			break; //Quit

		} else { //If an unusable character is present, display a warning.
			NW_VIEWER_LOG("sdemofasttxt::Print(): Character-code 0x%02X cannot be used.\n", (u8)c );

		}
	}
}

nw::math::VEC2 StrToCharSize( const char* str ) //String -> display character size
{
	NW_POINTER_ASSERT( str );

	int lineNum = 0;
	int maxCharNumInLine = 0;
	int charNumInLine = 0;

	for(;;) {
		const char c = *str++;

		if ( 0x20 <= c && c <= 0x7F ) { //1 character output

			++ charNumInLine;

		} else if ( c == '\n' ) { // Carriage return?

			++ lineNum;
			if ( charNumInLine > maxCharNumInLine ) {
				maxCharNumInLine = charNumInLine;
			}
			charNumInLine = 0;

		} else if ( c == '\0' ) {

			if ( charNumInLine > 0 )
			{
				++ lineNum;
			}
			if ( charNumInLine > maxCharNumInLine )
			{
				maxCharNumInLine = charNumInLine;
			}
			break; //Quit

		} else { //If an unusable character is present, display a warning.

			NW_VIEWER_LOG("sdemofasttxt::Print(): Character-code 0x%02X cannot be used.\n", (u8)c );
		}
	}

	return nw::math::VEC2( maxCharNumInLine, lineNum );
}

//======================================================================//
//							  Shapes display
//======================================================================//
//Set the texture coordinates to be colored and the specified color in the vertex array.
static void SetNoTexAndColToVtxs( nw::ut::Color8 color, Vtx *vtxs, int vtxNum )
{
	static const f32 NO_TEX_S = 0.99; //The lower right corner of the texture is colored.
	static const f32 NO_TEX_T = 0.99;
	for( int i=0; i<vtxNum; ++i ) {
		vtxs[i].color = color;
		vtxs[i].texcoord.Set( NO_TEX_S, NO_TEX_T );
	}
}
//◇ Property (Can be called even outside Begin() - End().)
//Line width
void SetLineWidth( f32 lineWidth )
{
	s_LineWidth = lineWidth;
}
f32 GetLineWidth()
{
	return s_LineWidth;
}

//◇ Render Function (If not between Begin() - End(), an assert failure occurs.)

//Straight line
void DrawLine( f32 x1, f32 y1, f32 x2, f32 y2, const nw::ut::Color8 & color )
{
    DrawLine( nw::math::VEC2( x1, y1 ), nw::math::VEC2( x2, y2 ), color );
}
void DrawLine( const nw::math::VEC2 & p1, const nw::math::VEC2 & p2, const nw::ut::Color8 & color )
{
    NW_ASSERT( s_IsDrawing );

    nw::math::VEC2 direction  = p2 - p1;
    if ( nw::ut::Abs<float>( direction.x ) < 0.1f && nw::ut::Abs<float>( direction.y ) < 0.1f ) {
		direction = nw::math::VEC2( 1, 0 );
	} else {
		nw::math::VEC2Normalize( & direction, & direction );
	}
    direction *= s_LineWidth / 2;

    nw::math::VEC2 ortho = nw::math::VEC2(direction.y, -direction.x);

    Vtx verteces[4];

    verteces[0].pos = p1 - ortho - direction ;		//[1] (above) ortho  	    [3]
    verteces[1].pos = p1 + ortho - direction ;		//   p1  direction -> p2
    verteces[2].pos = p2 - ortho + direction ;		//[0]     			  	[2]
    verteces[3].pos = p2 + ortho + direction ;

	SetNoTexAndColToVtxs( color, verteces, 4 );

	static const u16 indeces[6] = { 0, 1, 3, 	0, 3, 2 };

	EntryStream( verteces, 4, indeces, 6 );
}
//Square line
void DrawRectangle( f32 posh, f32 posv, f32 sizeh, f32 sizev, const nw::ut::Color8 & color )
{
    DrawRectangle( nw::math::VEC2(  posh,  posv ),
                   nw::math::VEC2( sizeh, sizev ), color );
}
void DrawRectangle( const nw::math::VEC2& pos, const nw::math::VEC2& size, const nw::ut::Color8 & color )
{
    NW_ASSERT( s_IsDrawing );

	const nw::math::VEC2 p1 = pos;
    const nw::math::VEC2 p2 = pos + nw::math::VEC2( size.x, 0 );
    const nw::math::VEC2 p3 = pos + size;
    const nw::math::VEC2 p4 = pos + nw::math::VEC2( 0, size.y );

    DrawLine( p1, p2, color );
    DrawLine( p2, p3, color );
    DrawLine( p3, p4, color );
    DrawLine( p4, p1, color );
}
//Square
void FillRectangle( f32 posh, f32 posv, f32 sizeh, f32 sizev, const nw::ut::Color8 & color )
{
	FillRectangle( nw::math::VEC2( posh, posv ), nw::math::VEC2( sizeh, sizev ), color );
}
void FillRectangle( const nw::math::VEC2& pos, const nw::math::VEC2& size, const nw::ut::Color8 & color )
{
    NW_ASSERT( s_IsDrawing );

    Vtx verteces[4];

	verteces[0].pos = pos;										// 0 1
	verteces[1].pos = nw::math::VEC2( pos.x + size.x, pos.y );  // 2 3
	verteces[2].pos = nw::math::VEC2( pos.x , pos.y + size.y );
	verteces[3].pos = pos + size;

	SetNoTexAndColToVtxs( color, verteces, 4 );

	static const u16 indeces[6] = { 0, 1, 2, 	2, 1, 3 };

	EntryStream( verteces, 4, indeces, 6 );
}

//Triangle
void FillTriangle( const nw::math::VEC2& pt1, const nw::math::VEC2& pt2, const nw::math::VEC2& pt3, const nw::ut::Color8 & color )
{
    NW_ASSERT( s_IsDrawing );

    Vtx verteces[3];

	verteces[0].pos = pt1;
	verteces[1].pos = pt2;
	verteces[2].pos = pt3;

	SetNoTexAndColToVtxs( color, verteces, 3 );

	static const u16 indeces[3] = { 0, 1, 2};

	EntryStream( verteces, 3, indeces, 3 );
}

} /* namespace system */
} /* namespace vw   */

