/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/
#include <viewer/system.h>
#include "../include/system_private.h"

#if NN_CURRENT_VERSION_NUMBER >= NN_VERSION_NUMBER(4,0,0,0)
#pragma diag_suppress 2530 // padding added to end of struct.
#endif
namespace vw     {
namespace system {
namespace Memory {

struct Descriptor
{
    unsigned int size_;
    bool    allocated_;
    Descriptor(unsigned int size, bool allocated) : size_(size), allocated_(allocated)
    {
    }
};

class Descriptors
{
public:
    Descriptors() {}
    void setup()
    {
        m_Array = nw::ut::MoveArray<Descriptor>( 0, vw::system::Memory_GetSysAllocator(), nw::ut::ARRAY_VARIABILITY );
    }
    u32 getCount() const
    {
        return (u32)m_Array.size();
    }
    const Descriptor& get(int index) const
    {
        return m_Array[index];
    }
    Descriptor& get(int index)
    {
        return m_Array[index];
    }
    void set(int index, const Descriptor& x)
    {
        m_Array[index] = x;
    }
    void insert(int index, const Descriptor& x)
    {
        m_Array.push_back( x );
        for(s32 i = getCount() - 1; i > index; --i)
        {
            m_Array[i] = m_Array[i - 1];
        }
        m_Array[index] = x;
    }
    void remove(int index)
    {
        if( 0 == getCount() )
        {
            return;
        }

        for(u32 i = index, size = getCount() - 1; i < size; ++i)
        {
            m_Array[ i ] = m_Array[i + 1];
        }
        m_Array.pop_back();
    }

private:
    nw::ut::MoveArray< Descriptor > m_Array;
};

class VRAMAllocator
{
    Descriptors descriptors_;
    void* rawMemory_;
    unsigned alignment_;

public:
    VRAMAllocator() : rawMemory_( 0 ), alignment_( 0 )
    {
    }
    void setup(void* rawMemory, unsigned size, unsigned alignment)
    {
        rawMemory_ = rawMemory;
        alignment_ = alignment;
        descriptors_.setup();
        Descriptor descriptor( size, false );
        descriptors_.insert( 0, descriptor );
    }
    void* mallocAligned(unsigned int size, unsigned int alignment)
    {
        return reallocAligned( size, alignment );
    }
    void* reallocAligned(unsigned int size, unsigned int alignment)
    {
        if( size <= 0 )
        {
            return NULL;
        }

        if( alignment < alignment_)
        {
            alignment = alignment_;
        }

        char*        pointer      = reinterpret_cast<char*>( rawMemory_ );
        char*        bestPointer  = NULL;
        int          bestFitIndex = -1;
        unsigned int bestCost     = 0xffffffff;
        unsigned int bestLeftPad  = 0;

        const int count = descriptors_.getCount();
        for (int i = 0; i < count; ++i)
        {
            const Descriptor& descriptor = descriptors_.get(i);
            const unsigned int blockSize = descriptor.size_;
            if ( descriptor.allocated_ )
            {
                pointer += blockSize;
                continue;
            }

            unsigned int leftPad  = alignment  - 1 - ((reinterpret_cast<unsigned int>(pointer) - 1) & (alignment - 1));
            if ( leftPad+size > blockSize )
            {
                pointer += blockSize;
                continue;
            }

            unsigned int cost = computeCost( blockSize, size, leftPad );
            if ( cost < bestCost )
            {
                bestFitIndex = i;
                bestCost = cost;
                bestLeftPad = leftPad;
                bestPointer = pointer;
            }

            unsigned int rightPad = (blockSize - leftPad - size) & (alignment - 1);
            cost = computeCost(blockSize, size, rightPad);
            if ( cost < bestCost )
            {
                bestFitIndex = i;
                bestCost = cost;
                bestLeftPad = blockSize - size - rightPad;
                bestPointer = pointer;
            }

            pointer += blockSize;
        }

        if ( bestFitIndex < 0 )
        {
            return NULL;
        }

        allocateBlock(bestFitIndex, size, bestLeftPad);
        return reinterpret_cast<void*>(bestPointer + bestLeftPad);
    }
    void free(void* mem)
    {
        if ( !mem )
        {
            return;
        }

        int newBlock;
        unsigned int pad;
        int block = getDescriptorIndex( mem );

        freeBlock(block, newBlock, pad);
    }
    bool canAllocate(u32 size, unsigned int alignment)
    {
        if(size <= 0)
        {
            return NULL;
        }

        if( alignment < alignment_)
        {
            alignment = alignment_;
        }

        char*        pointer      = reinterpret_cast<char*>( rawMemory_ );
        char*        bestPointer  = NULL;
        int          bestFitIndex = -1;
        unsigned int bestCost     = 0xffffffff;
        unsigned int bestLeftPad  = 0;
        const int count = descriptors_.getCount();

        for (int i = 0; i < count; ++i)
        {
            const Descriptor& descriptor = descriptors_.get( i );
            const unsigned int blockSize = descriptor.size_;

            if ( descriptor.allocated_ )
            {
                pointer += blockSize;
                continue;
            }

            unsigned int leftPad  = alignment - 1 - ((reinterpret_cast<unsigned int>(pointer) - 1) & (alignment - 1));
            if (leftPad + size > blockSize)
            {
                pointer += blockSize;
                continue;
            }

            if ( computeCost(blockSize, size, leftPad) < bestCost )
            {
                return true;
            }

            unsigned int rightPad = (blockSize - leftPad - size) & (alignment - 1);
            if ( computeCost(blockSize, size, rightPad) < bestCost )
            {
                return true;
            }

            pointer += blockSize;
        }
        return false;
    }

private:
    int getDescriptorIndex(void* ptr) const
    {
        char* pointer = reinterpret_cast<char*>( rawMemory_ );
        const int count = descriptors_.getCount();

        for (int i = 0; i < count; ++i)
        {
            if ( pointer == ptr )
            {
                return i;
            }

            if ( pointer > ptr )
            {
                return-1;
            }

            const Descriptor& descriptor = descriptors_.get( i );
            pointer += descriptor.size_;
        }

        return -1;
    }
    void allocateBlock(int index, unsigned int size, unsigned int pad)
    {
        int newIndex = index;

        Descriptor descriptor = descriptors_.get( newIndex );

        unsigned int rightPad = descriptor.size_ - size - pad;

        if (pad > 0)
        {
            Descriptor leftDescriptor( pad, false );
            descriptors_.insert( newIndex, leftDescriptor );
            ++newIndex;
        }

        descriptor.size_ = size;
        descriptor.allocated_ = true;
        descriptors_.set( newIndex, descriptor );
        ++newIndex;

        if (rightPad > 0)
        {
            Descriptor rightDescriptor( rightPad, false );
            descriptors_.insert( newIndex, rightDescriptor );
        }
    }
    void freeBlock(int index, int& newBlock, unsigned int& pad)
    {
        descriptors_.get( index ).allocated_ = false;

        int newIndex = index;
        if (newIndex > 0)
        {
            const Descriptor& descriptor = descriptors_.get( newIndex - 1 );
            if( !descriptor.allocated_ )
            {
                pad = descriptor.size_;
                --newIndex;
            }
            else
            {
                pad = 0;
            }
        }
        newBlock = newIndex;
        Descriptor newDescriptor = descriptors_.get( newIndex );

        while( newIndex + 1 < descriptors_.getCount() )
        {
            const Descriptor &descriptor = descriptors_.get( newIndex + 1 );
            if( descriptor.allocated_ )
            {
                break;
            }

            newDescriptor.size_ += descriptor.size_;
            descriptors_.remove( newIndex + 1 );
        }
        descriptors_.set( newIndex,newDescriptor );
    }
    static unsigned int computeCost(unsigned int freeBlockSize, unsigned int size, unsigned int pad)
    {
        const unsigned int otherPad = freeBlockSize - size - pad;
        return freeBlockSize * freeBlockSize - pad * pad - otherPad * otherPad;
    }
};

#if defined( NN_PCGX )
static VRAMAllocator *sVRAM_A = NULL;
static VRAMAllocator *sVRAM_B = NULL;
#else
static VRAMAllocator sVRAM_A;
static VRAMAllocator sVRAM_B;
#endif
static GraphicsMemoryAllocator* sGraphicsMemoryAllocator = 0;

    //
GraphicsMemoryAllocator::GraphicsMemoryAllocator()
     : m_Initialized( false )
{
    sGraphicsMemoryAllocator = this;
}

const int ALIGNMENT_SYSTEM_BUFFER     = 4;
const int ALIGNMENT_VERTEX            = 4;
const int ALIGNMENT_TEXTURE           = 128;
const int ALIGNMENT_RENDER_BUFFER     = 64;    // 24-bit format (RGB8, D24) buffer must be aligned to 96 bytes
const int ALIGNMENT_DISPLAY_BUFFER    = 16;
const int ALIGNMENT_3D_COMMAND_BUFFER = 16;

//--------------------------------------------------------------------------
void GraphicsMemoryAllocator::Initialize()
{
    if (!m_Initialized)
    {
        u32 sizeA = u32(nn::gx::GetVramEndAddr(nn::gx::MEM_VRAMA) - nn::gx::GetVramStartAddr(nn::gx::MEM_VRAMA));
        u32 sizeB = u32(nn::gx::GetVramEndAddr(nn::gx::MEM_VRAMB) - nn::gx::GetVramStartAddr(nn::gx::MEM_VRAMB));

#if defined( NN_PCGX )
        PCGX_DELETE(sVRAM_A);
        PCGX_DELETE(sVRAM_B);
        sVRAM_A = new VRAMAllocator;
        sVRAM_B = new VRAMAllocator;
        sVRAM_A->setup((void*)nn::gx::GetVramStartAddr(nn::gx::MEM_VRAMA), sizeA, ALIGNMENT_SYSTEM_BUFFER);
        sVRAM_B->setup((void*)nn::gx::GetVramStartAddr(nn::gx::MEM_VRAMB), sizeB, ALIGNMENT_SYSTEM_BUFFER);
#else
        sVRAM_A.setup((void*)nn::gx::GetVramStartAddr(nn::gx::MEM_VRAMA), sizeA, ALIGNMENT_SYSTEM_BUFFER);
        sVRAM_B.setup((void*)nn::gx::GetVramStartAddr(nn::gx::MEM_VRAMB), sizeB, ALIGNMENT_SYSTEM_BUFFER);
#endif

        m_Initialized = true;
    }
}

//--------------------------------------------------------------------------
void GraphicsMemoryAllocator::Finalize(void)
{
#if defined( NN_PCGX )
    //  Free resources here because an error results on PCs during finalization due to differences in when global variables and allocator variables are freed.
    //  
    PCGX_DELETE(sVRAM_A);
    PCGX_DELETE(sVRAM_B);
#endif

    m_Initialized = false;
}

//--------------------------------------------------------------------------
bool GraphicsMemoryAllocator::CanAllocateA(u32 size)
{
    // VRAM management is not performed unless initialized.
    if( !m_Initialized ) return false;

#if defined( NN_PCGX )
    PCGX_ASSERT(sVRAM_A);
    return sVRAM_A->canAllocate(size,ALIGNMENT_TEXTURE);
#else
    return sVRAM_A.canAllocate(size,ALIGNMENT_TEXTURE);
#endif
}
//--------------------------------------------------------------------------
bool GraphicsMemoryAllocator::CanAllocateB(u32 size)
{
    // VRAM management is not performed unless initialized.
    if( !m_Initialized ) return false;

#if defined( NN_PCGX )
    PCGX_ASSERT(sVRAM_B);
    return sVRAM_B->canAllocate(size,ALIGNMENT_TEXTURE);
#else
    return sVRAM_B.canAllocate(size,ALIGNMENT_TEXTURE);
#endif
}
//--------------------------------------------------------------------------
bool GraphicsMemoryAllocator::IsInitialized()
{
    return m_Initialized;
}
//--------------------------------------------------------------------------
void* GraphicsMemoryAllocator::Allocate(GLenum area, GLenum aim, GLuint id, GLsizei size)
{
    NW_UNUSED_VARIABLE(id);

    NW_ASSERTMSG (m_Initialized, "[DEMO:GraphicsMemoryAllocator] not initialized.\n");

    if (size == 0) { return 0; }

    int addrAlign = 8;
    void* resultAddr = NULL;

    // Take into account the alignment restrictions on each data set's position
    switch (aim)
    {
    case NN_GX_MEM_SYSTEM:
        addrAlign = ALIGNMENT_SYSTEM_BUFFER;
        break;
    case NN_GX_MEM_TEXTURE:
        addrAlign = ALIGNMENT_TEXTURE;
        break;
    case NN_GX_MEM_VERTEXBUFFER:
        addrAlign = ALIGNMENT_VERTEX;
        break;
    case NN_GX_MEM_RENDERBUFFER:
        addrAlign = ALIGNMENT_RENDER_BUFFER;
        break;
    case NN_GX_MEM_DISPLAYBUFFER:
        addrAlign = ALIGNMENT_DISPLAY_BUFFER;
        break;
    case NN_GX_MEM_COMMANDBUFFER:
        addrAlign = ALIGNMENT_3D_COMMAND_BUFFER;
        break;
    default:
        NW_FATAL_ERROR("[DEMO:GraphicsMemoryAllocator] invalid parameter.(0x%X)\n", aim);
        break;
    }

    switch (area)
    {
    case NN_GX_MEM_FCRAM:
        // When using FCRAM, no need to manage this because it is allocated from the expanded heap that is supported by the SDK
        resultAddr = Memory_AllocSysDevice(size, addrAlign);
        break;
#if defined( NN_PCGX )
    case NN_GX_MEM_VRAMA:
        PCGX_ASSERT(sVRAM_A);
        resultAddr = sVRAM_A->mallocAligned(size, addrAlign);
        break;
    case NN_GX_MEM_VRAMB:
        PCGX_ASSERT(sVRAM_B);
        resultAddr = sVRAM_B->mallocAligned(size, addrAlign);
        break;
#else
    case NN_GX_MEM_VRAMA:
        resultAddr = sVRAM_A.mallocAligned(size, addrAlign);
        break;
    case NN_GX_MEM_VRAMB:
        resultAddr = sVRAM_B.mallocAligned(size, addrAlign);
        break;
#endif
    default:
        // This is not likely in normal use; FATAL_ERROR
        NW_FATAL_ERROR("[DEMO:GraphicsMemoryAllocator] invalid parameter.(0x%X)\n", area);
        break;
    }

    NW_NULL_ASSERT(resultAddr);

    return resultAddr;
}

//--------------------------------------------------------------------------
void GraphicsMemoryAllocator::Deallocate(GLenum area, GLenum aim, GLuint id, void* addr)
{
    NW_ASSERTMSG (m_Initialized, "[DEMO:GraphicsMemoryAllocator] not initialized.\n");
    NW_UNUSED_VARIABLE(aim);
    NW_UNUSED_VARIABLE(id);

    switch (area)
    {
    case NN_GX_MEM_FCRAM:
        Memory_Free(addr);
        break;
#if defined( NN_PCGX )
    case NN_GX_MEM_VRAMA:
        PCGX_ASSERT(sVRAM_A);
        sVRAM_A->free(addr);
        break;
    case NN_GX_MEM_VRAMB:
        PCGX_ASSERT(sVRAM_B);
        sVRAM_B->free(addr);
        break;
#else
    case NN_GX_MEM_VRAMA:
        sVRAM_A.free(addr);
        break;
    case NN_GX_MEM_VRAMB:
        sVRAM_B.free(addr);
        break;
#endif
    }
}

GraphicsMemoryAllocator* GraphicsMemoryAllocator::GetInstance(){
    return sGraphicsMemoryAllocator;
}


}   /* namespace Memory */
}   /* namespace system */
}   /* namespace vw     */
#if NN_CURRENT_VERSION_NUMBER >= NN_VERSION_NUMBER(4,0,0,0)
#pragma diag_default 2530 // padding added to end of struct.
#endif
