/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#include <dolphin.h>
#include <viewer/common.h>
#include <viewer/system.h>
#include <cstring>
#include "TextResourcesForWarning.h"
#include "system_private.h"

using namespace nw4r;

namespace vw     {
namespace system {


/*---------------------------------------------------------------------------*
  Name:         JpegData::JpegData
  Description:  Constructor
  Arguments:    width   Image horizontal size
                height  Image vertical size
                bpp     Image bpp size
  Returns:      None.
 *---------------------------------------------------------------------------


*/
/* ctor */
JpegData::JpegData( u32 width, u32 height, u32 quality )
 : ImageFmt( width, height, 4 ),
   mBodySize( 0 ),
   mQuality( quality )
{
    u32 size = ut::RoundUp(mWidth, 4) * ut::RoundUp(mHeight, 4) * sizeof(u32);

    mBody = Memory_TryAllocMEM2( system::SYSTEM_ID, size, 128 );
}


/*---------------------------------------------------------------------------*
  Name:         JpegData::~JpegData
  Description:  Destructor
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
/* dtor */
JpegData::~JpegData( void )
{
}


/*---------------------------------------------------------------------------*
  Name:         JpegData::ConvertData
  Description:  Converts EFB data to JPEG format.
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
JpegData::ConvertData( void )
{
    this->ConvertDataImpl_();
    this->CompressDataImpl_();
}


/*---------------------------------------------------------------------------*
  Name:         JpegData::ConvertData
  Description:  Converts texture format RBGA data to JPEG data.
  Arguments:    src_image   The pointer to the texture data
                width       Texture horizontal size
                height      Texture vertical size
                dst_ofs_x   X position in the overall JPEG image
                dst_ofs_y   Y position in the overall JPEG image
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
JpegData::ConvertData( void* src_image, u32 width, u32 height, u32 dst_ofs_x, u32 dst_ofs_y )
{
    this->ConvertDataImpl_( src_image, width, height, dst_ofs_x, dst_ofs_y, 0, 0, width, height );
}


/*---------------------------------------------------------------------------*
  Name:         JpegData::ConvertData
  Description:  Converts texture format RBGA data to JPEG data.
  Arguments:    src_image   The pointer to the texture data
                src_width   Texture horizontal size
                src_height  Texture vertical size
                dst_ofs_x   X position in the overall JPEG image
                dst_ofs_y   Y position in the overall JPEG image
                src_ofs_x   X-offset of the rectangle used in the texture
                src_ofs_y   Y-offset of the rectangle used in the texture
                width       Horizontal size of the rectangle used in the texture
                height      Vertical size of the rectangle used in the texture
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
JpegData::ConvertData( void* src_image,
                      u32 src_width, u32 src_height,
                      u32 dst_ofs_x, u32 dst_ofs_y,
                      u32 src_ofs_x, u32 src_ofs_y,
                      u32 width,     u32 height )
{
    this->ConvertDataImpl_( src_image, src_width, src_height, dst_ofs_x, dst_ofs_y, src_ofs_x, src_ofs_y, width, height );
}



/*---------------------------------------------------------------------------*
  Name:         JpegData::ConvertDataImpl_
  Description:  Convert RGBA data in buffer to match format.
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
JpegData::ConvertDataImpl_( void )
{
    const u32 SRC_BPP = 4;

    u8* pTmp   = static_cast<u8*>( Memory_Alloc2( system::SYSTEM_ID, ut::RoundUp(mWidth, 4) * ut::RoundUp(mHeight, 4) * SRC_BPP ) );
    std::memcpy( pTmp, mBody, mWidth * mHeight * SRC_BPP );

    this->ConvertDataImpl_( pTmp, mWidth, mHeight, 0, 0, 0, 0, mWidth, mHeight );

    Memory_Free( pTmp );
}


/*---------------------------------------------------------------------------*
  Name:         JpegData::ConvertDataImpl_
  Description:  Copies texture format RBGA data.
  Arguments:    src_image   The pointer to the texture data
                src_width   Texture horizontal size
                src_height  Texture vertical size
                dst_ofs_x   X position in the overall BMP image
                dst_ofs_y   Y position in the overall BMP image
                src_ofs_x   X-offset of the rectangle used in the texture
                src_ofs_y   Y-offset of the rectangle used in the texture
                width       Horizontal size of the rectangle used in the texture
                height      Vertical size of the rectangle used in the texture
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
JpegData::ConvertDataImpl_( void* src_image,
                      u32 src_width, u32 src_height,
                      u32 dst_ofs_x, u32 dst_ofs_y,
                      u32 src_ofs_x, u32 src_ofs_y,
                      u32 width, u32 height )
{
#pragma unused( src_height )
    u8* pRead = reinterpret_cast<u8*>( src_image );
    u8* pWrite = reinterpret_cast<u8*>( mBody );

    for ( u32 line = 0; line < height; ++line )
    {
        u32 y = src_height - line - 1 - (src_height - height - src_ofs_y);
        u32 dst_y = mHeight - (line + dst_ofs_y) - 1;
        for ( u32 i = 0; i < width; ++i )
        {
            u32 x = i + src_ofs_x;
            u32 dst_x = i + dst_ofs_x;

            nw::ut::Color8 col = UtilRGBA8::GetColor( pRead, x, y, src_width );
            UtilRGBA8::SetColor( col, pWrite, dst_x, dst_y, mWidth );
        }
    }
}

/*---------------------------------------------------------------------------*
  Name:         JpegData::CompressDataImpl_
  Description:  Data compression
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
JpegData::CompressDataImpl_()
{
    s32   result;
    u32   predictSize;
    u32   outputSize;
    void* buf;

    void* work = Memory_AllocSys2( TMCC_JPEG_WORKAREA_SIZE );

    result = TMCCJPEGEncGetOutputSize( work,
                                       mBody,
                                       s32(mWidth), s32(mHeight),
                                       GX_TF_RGBA8,
                                       &predictSize,
                                       mQuality );

    if ( result < 0 )
    {
        goto out;
    }

    buf = Memory_AllocSys2( predictSize );

    result = TMCCJPEGEncEncode( work,
                                mBody,
                                s32(mWidth), s32(mHeight),
                                GX_TF_RGBA8,
                                mQuality,
                                buf,
                                predictSize,
                                &outputSize );
    if ( result >= 0 )
    {
        Memory_Free( mBody );
        mBody = buf;
        mBodySize = outputSize;
    }

out:
    Memory_Free( work );
}


/*---------------------------------------------------------------------------*
  Name:         JpegData::Finalize
  Description:  Performs data finalization such as compression and puts into a state that can be output to the PC as a file.
                
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
JpegData::Finalize( void )
{
    this->CompressDataImpl_();
}


}   // namespace system
}   // namespace vw
