/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/


#include "../include/JpegReaderImpl.h"
#include "../include/TexUtil.h"

#include <nn/jpeg.h>



namespace vw     {
namespace system {


/*---------------------------------------------------------------------------*
  Name:         JpegReaderImpl::JpegReaderImpl
  Description:  Constructor
  Arguments:    data   Specifies JPEG file data.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
/* ctor */
JpegReaderImpl::JpegReaderImpl( void* data, u32 dataSize )
 : ImageReaderImpl(),
   mValid( false ),
   mBuf( data ),
   mHeight( 0 ),
   mWidth( 0 ),
   mDecBuf( NULL )
{
    nn::jpeg::JpegMpDecoder decoder;
    void* decoderWork = Memory_AllocSys(decoder.GetWorkBufferSize(), 4);
    decoder.Initialize(decoderWork, decoder.GetWorkBufferSize());

    const nn::jpeg::PixelFormat dstPixelFormat = nn::jpeg::PIXEL_FORMAT_RGBA8;

    decoder.ExtractExif((const u8*)data, dataSize, false);
    mWidth = decoder.GetLastWidth();
    mHeight = decoder.GetLastHeight();
    if ( mWidth > 1024 || mHeight > 1024 )
    {
        Memory_Free( decoderWork );
        return;
    }

    size_t dstBufferSize = decoder.GetDstBufferSize(mWidth, mHeight, dstPixelFormat);
    mDecBuf = Memory_AllocSys(dstBufferSize, 4);
    size_t decodeResult = decoder.StartJpegDecoder(mDecBuf,
                                                   dstBufferSize,
                                                   (const u8*)data,
                                                   dataSize,
                                                   mWidth,
                                                   mHeight,
                                                   dstPixelFormat,
                                                   false);
    decoder.Finalize();
    Memory_Free( decoderWork );

    // decodeResult is the number of bytes of the decoded image and is non-zero if successful.
    // The result varies depending on the size of the decoded image.
    if(decodeResult != dstBufferSize)
    {
        return;
    }
    mValid = true;
}


/*---------------------------------------------------------------------------*
  Name:         TgaReaderImpl::~TgaReaderImpl
  Description:  Destructor
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
/* dtor */
JpegReaderImpl::~JpegReaderImpl()
{
    this->Dispose();
}


/*---------------------------------------------------------------------------*
  Name:         JpegReaderImpl::Expand
  Description:  Expands JPEG data to RGBA8 texture
  Arguments:    expBuf      Buffer to expand to
                bufWidth    Buffer horizontal size
                bufHeight   Buffer vertical size
  Returns:      None.
 *---------------------------------------------------------------------------


*/
bool
JpegReaderImpl::Expand( void* expBuf, u32 bufWidth, u32 bufHeight )
{
    if ( ! this->IsValid() ) { return false; }

    u8* pOut = static_cast<u8*>( expBuf );
    u8* pIn  = static_cast<u8*>( mDecBuf );

    for ( u16 y = 0; y < mHeight; ++y )
    {
        for ( u16 x = 0; x < mWidth; ++x )
        {
            nw::ut::Color8 col = UtilRGBA8::GetRGBA( pIn, x, y, mWidth );
            UtilRGBA8::SetColor( col, pOut, x, y, bufWidth, bufHeight );
        }
    }
    return true;
}

/*---------------------------------------------------------------------------*
  Name:         JpegReaderImpl::Dispose()
  Description:  Free allocated memory for internal processing.
  Arguments:    data
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
JpegReaderImpl::Dispose()
{
    if ( mDecBuf )
    {
        Memory_Free( mDecBuf );
        mDecBuf = NULL;
    }
}

}   /* namespace system */
}   /* namespace vw     */


