/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/
#include <viewer/system.h>
#include "../include/system_private.h"



namespace vw     {
namespace system {


RenderSystem::CommandDesc::CommandDesc()
: commandBufferSize(0x100000),
  commandRequestCount(512),
  reusableCommandBufferSize(0),
  reusableCommandRequestCount(0),
  commandListCount(2){}

RenderSystem::ScreenDesc::ScreenDesc()
: displayBufferCount(2)
{
    upperScreenDescription.screenKind = UPPER_SCREEN;
    upperScreenDescription.width      = NN_GX_DISPLAY0_HEIGHT;
    upperScreenDescription.height     = NN_GX_DISPLAY0_WIDTH;
    lowerScreenDescription.screenKind = LOWER_SCREEN;
    lowerScreenDescription.width      = NN_GX_DISPLAY1_HEIGHT;
    lowerScreenDescription.height     = NN_GX_DISPLAY1_WIDTH;
}


namespace
{
    Memory::GraphicsMemoryAllocator s_GraphicsMemoryAllocator;
}

//----------------------------------------
void* AllocateGraphicsMemory(GLenum area, GLenum aim, GLuint id, GLsizei size)
{
    return s_GraphicsMemoryAllocator.Allocate(area, aim, id, size);
}

//----------------------------------------
void DeallocateGraphicsMemory(GLenum area, GLenum aim, GLuint id, void* addr)
{
    s_GraphicsMemoryAllocator.Deallocate(area, aim, id, addr);
}

static RenderSystem* sInstance = NULL;

RenderSystem*
RenderSystem::GetInstance()
{
    return sInstance;
}

//----------------------------------------
RenderSystem*
RenderSystem::Create(
    nw::os::IAllocator* allocator,
    const CommandDesc& description,
    bool isEmbeddedViewer
)
{
    void* memory = allocator->Alloc(sizeof(RenderSystem));
    sInstance = new(memory) RenderSystem( allocator, description, isEmbeddedViewer );

    NW_GL_ASSERT();

    return sInstance;
}


//----------------------------------------
void
RenderSystem::Destroy()
{
    nw::gfx::SafeDestroy(this->m_ExtensionSwapper);
    nw::gfx::SafeDestroy(this->m_LowerSwapper);
    nw::gfx::SafeDestroy(this->m_UpperSwapper);
    nw::gfx::SafeDestroy(this->m_RenderContext);
    nw::gfx::SafeDestroy(this->m_CommandListSwapper);

    if( s_GraphicsMemoryAllocator.IsInitialized() )
    {
        nngxFinalize();
        s_GraphicsMemoryAllocator.Finalize();
    }

    NW_GL_ASSERT();

    if( this->m_Allocator )
    {
        void* memory = static_cast<void*>(this);
        this->m_Allocator->Free( memory );
    }
}

//----------------------------------------
void
RenderSystem::InitCommandCache(const CommandDesc& description)
{
    // get current command list
    GLint command_id;
    nngxGetCmdlistParameteri(NN_GX_CMDLIST_BINDING, &command_id);

    vw::system::CommandListSwapper::Description swapperDescription;
    swapperDescription.commandListCount          = description.commandListCount;
    swapperDescription.bufferSize                = description.commandBufferSize;
    swapperDescription.requestCount              = description.commandRequestCount;
    swapperDescription.reusableBufferSize        = description.reusableCommandBufferSize;
    swapperDescription.reusableRequestCount      = description.reusableCommandRequestCount;
    swapperDescription.maxGpuProfilingEntryCount = 8;
    m_CommandListSwapper = vw::system::CommandListSwapper::Create(Memory_GetSysAllocatorDevice(), swapperDescription);
    NW_POINTER_ASSERT( m_CommandListSwapper );

    // reset command list
    nngxBindCmdlist( command_id );
}

//----------------------------------------
void
RenderSystem::ChangeDisplayBuffer( const ScreenDesc& description )
{
    nw::os::IAllocator* allocator = m_Allocator;

    nw::gfx::SafeDestroy(this->m_ExtensionSwapper);
    nw::gfx::SafeDestroy(this->m_LowerSwapper);
    nw::gfx::SafeDestroy(this->m_UpperSwapper);

    this->m_UpperSwapper = DisplayBufferSwapper::Builder()
        .BufferDescription( description.upperScreenDescription )
        .BufferCount (description.displayBufferCount )
        .Create( allocator );

    this->m_LowerSwapper = DisplayBufferSwapper::Builder()
        .BufferDescription( description.lowerScreenDescription )
        .BufferCount( description.displayBufferCount )
        .Create( allocator );

    nngxSetDisplayMode( NN_GX_DISPLAYMODE_STEREO );

    vw::system::DisplayBufferSwapper::Description extensionDesc = description.upperScreenDescription;
    extensionDesc.screenKind = EXTENSION_SCREEN;
    this->m_ExtensionSwapper = DisplayBufferSwapper::Builder()
        .BufferDescription( extensionDesc )
        .Create(allocator);
}


//----------------------------------------
void
RenderSystem::ClearBuffer( nw::ut::FloatColor clearColor , f32 clearZ )
{
    this->m_RenderContext->ClearBuffer(
        GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT,
        clearColor,
        clearZ);
    NW_GL_ASSERT();
}

//----------------------------------------
void
RenderSystem::ClearDepth( f32 clearZ )
{
    this->m_RenderContext->ClearBuffer(
        GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT,
        nw::ut::FloatColor(),
        clearZ);
    NW_GL_ASSERT();
}

//----------------------------------------
void
RenderSystem::TransferBuffer(
    s32 screenKind
)
{
    nw::gfx::RenderContext* renderContext = this->GetRenderContext();

    // When there are duplicates in the command list, you need to change the display buffer at the transfer target since the timing for the transfer to the display buffer will change.
    // 
    // 
    // For more information, see the comments on the PresentBuffer function.

    bool isMultiCommandList = m_CommandListSwapper->GetListCount() > 1;

    if (screenKind & UPPER_SCREEN)
    {
        this->m_UpperSwapper->MakeTransferBufferCommand(renderContext->GetRenderTarget(), isMultiCommandList);
    }

    if (screenKind & LOWER_SCREEN)
    {
        this->m_LowerSwapper->MakeTransferBufferCommand(renderContext->GetRenderTarget(), isMultiCommandList);
    }

    if (screenKind & EXTENSION_SCREEN)
    {
        this->m_ExtensionSwapper->MakeTransferBufferCommand(renderContext->GetRenderTarget(), isMultiCommandList);
    }
}

//----------------------------------------
void
RenderSystem::SwapBuffer( s32 screenKind )
{
    GLenum display;
    if (screenKind & (UPPER_SCREEN | LOWER_SCREEN))
    {
        this->m_UpperSwapper->ActivateBuffer();
        this->m_LowerSwapper->ActivateBuffer();

        if (screenKind & EXTENSION_SCREEN)
        {
            this->m_ExtensionSwapper->ActivateBuffer();
        }

        display = NN_GX_DISPLAY_BOTH;
    }
    else if (screenKind & UPPER_SCREEN)
    {
        this->m_UpperSwapper->ActivateBuffer();

        if (screenKind & EXTENSION_SCREEN)
        {
            this->m_ExtensionSwapper->ActivateBuffer();
        }

        display = NN_GX_DISPLAY0;
    }
    else if (screenKind & LOWER_SCREEN)
    {
        this->m_LowerSwapper->ActivateBuffer();

        display = NN_GX_DISPLAY1;
    }

    nngxSwapBuffers(display);
    glBindFramebuffer(GL_FRAMEBUFFER, 0);
}

//----------------------------------------
void
RenderSystem::PresentBuffer(s32 screenKind)
{
    if (m_CommandListSwapper->GetListCount() > 1)
    {
        SwapBuffer(screenKind);
        WaitVSync(screenKind);
        RunCommandList();
        SwapCommandList();
    }
    else
    {
        SwapBuffer(screenKind);
        WaitVSync(screenKind);
    }
}

//----------------------------------------
void
RenderSystem::WaitVSync(s32 screenKind)
{
    NW_MAX_ASSERT(screenKind, (int)BOTH_SCREENS);

    GLenum display;
    if (screenKind & (UPPER_SCREEN | LOWER_SCREEN)) { display = NN_GX_DISPLAY_BOTH; }
    else if (screenKind & UPPER_SCREEN) { display = NN_GX_DISPLAY0; }
    else if (screenKind & LOWER_SCREEN) { display = NN_GX_DISPLAY1; }

    nngxWaitVSync(display);
}

//----------------------------------------
void
RenderSystem::RunCommandList()
{
    m_CommandListSwapper->RunAsync();
}

//----------------------------------------
void
RenderSystem::SwapCommandList()
{
    m_CommandListSwapper->Swap();
    m_CommandListSwapper->Bind();
}

//----------------------------------------
void
RenderSystem::WaitCommandList()
{
    if (m_CommandListSwapper->GetListCount() > 1)
    {
        m_CommandListSwapper->WaitDone();
    }
    else
    {
        RunCommandList();
        m_CommandListSwapper->WaitDone();
    }
}

//----------------------------------------
void
RenderSystem::SetRenderTarget(nw::gfx::IRenderTarget* renderTarget)
{
    nw::gfx::RenderContext* renderContext = this->GetRenderContext();
    renderContext->SetRenderTarget(renderTarget);
}

//----------------------------------------
DisplayBufferSwapper*
RenderSystem::GetDisplayBufferSwapper( s32 screenKind ) const
{
    if(screenKind == UPPER_SCREEN) return m_UpperSwapper;
    if(screenKind == LOWER_SCREEN) return m_LowerSwapper;
    return m_ExtensionSwapper;
}

//----------------------------------------
void
RenderSystem::RegistGraphicsMemoryAllocator()
{
    if (!s_GraphicsMemoryAllocator.IsInitialized())
    {
        s_GraphicsMemoryAllocator.Initialize();
        if (nngxInitialize(AllocateGraphicsMemory, DeallocateGraphicsMemory) == GL_FALSE)
        {
            NW_FATAL_ERROR("nngxInitialize failed.\n");
        }
    }
}

//----------------------------------------
bool
RenderSystem::IsGraphicsMemoryAllocatorInitialized()
{
    return s_GraphicsMemoryAllocator.IsInitialized();
}

RenderSystem::RenderSystem(nw::os::IAllocator* allocator,
                           const CommandDesc& description,
                           bool isEmbeddedViewer) : m_Allocator( allocator ),
                                                    m_RenderContext( NULL ),
                                                    m_UpperSwapper( NULL ),
                                                    m_LowerSwapper( NULL ),
                                                    m_ExtensionSwapper( NULL ),
                                                    m_CommandListSwapper( NULL ),
                                                    m_IsEmbeddedViewer( isEmbeddedViewer )
{
    if( vw::system::App::GetInstance().GetUseGraphicsMemoryAllocator() )
    {
        RegistGraphicsMemoryAllocator();
    }

    if( !isEmbeddedViewer )
    {
        nngxStartLcdDisplay();
    }

    if( description.commandListCount > 0 )
    {
        // get current command list
        GLint command_id;
        nngxGetCmdlistParameteri(NN_GX_CMDLIST_BINDING, &command_id);

        vw::system::CommandListSwapper::Description swapperDescription;
        swapperDescription.commandListCount          = description.commandListCount;
        swapperDescription.bufferSize                = description.commandBufferSize;
        swapperDescription.requestCount              = description.commandRequestCount;
        swapperDescription.reusableBufferSize        = description.reusableCommandBufferSize;
        swapperDescription.reusableRequestCount      = description.reusableCommandRequestCount;
        swapperDescription.maxGpuProfilingEntryCount = 8;
        m_CommandListSwapper = vw::system::CommandListSwapper::Create(Memory_GetSysAllocatorDevice(), swapperDescription);
        NW_POINTER_ASSERT( m_CommandListSwapper );

        // reset command list
        if( isEmbeddedViewer )
        {
            nngxBindCmdlist( command_id );
        }
        else
        {
            m_CommandListSwapper->Bind();
            m_CommandListSwapper->RunAsync();
        }
    }

    m_RenderContext = nw::gfx::RenderContext::Builder()
        .Create( allocator );
}

}   /* namespace system */
}   /* namespace vw     */
