/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#include <viewer.h>
#include <viewer/system.h>
#include "../include/LogSystem.h"
#include <viewer/system/Stream.h>

#if NN_CURRENT_VERSION_NUMBER >= NN_VERSION_NUMBER(4,0,0,0)
#pragma diag_suppress 1301 // padding inserted in struct.
#endif
namespace vw     {
namespace system {



class SerialChannel
{
public:
    SerialChannel(u16 channel, u32 bufSize=0x20000) : m_ReadSize(0), m_BufSize(bufSize), m_Cursor(0), m_Buffer(0), m_IsConnect(false), m_IsRunning(false), m_pWorkMemory( 0 )
    {
        nn::Result ret = nn::hio::GetHioDaemonStatus( &m_IsRunning );

        if(ret.IsSuccess() && m_IsRunning == true)
        {
            m_pWorkMemory = vw::system::Memory_AllocSysDevice(nn::hio::WORKMEMORY_SIZE);
            nn::Result res = m_Serial.Open( channel, m_pWorkMemory );
        #if 0
            if ( res.IsSuccess() ) {
                NW_DEV_LOG("Serial[%d] is Opened\n", channel );
            } else {
                NW_DEV_LOG("Serial[%d] is Failed\n", channel );
            }
        #endif
            m_Buffer     = (u8*)vw::system::Memory_AllocSys(bufSize);
            m_TempBuffer = (u8*)vw::system::Memory_AllocSys(bufSize);
        }
    }
    ~SerialChannel(){
        if( m_Buffer )
        {
            vw::system::Memory_Free( m_Buffer );
        }
        if( m_TempBuffer )
        {
            vw::system::Memory_Free( m_TempBuffer );
        }
        if( m_IsConnect ){
            m_IsConnect = false;
            m_Serial.Disconnect();
        }

        if( m_IsRunning )
        {
            m_Serial.Close();
        }

        if( m_pWorkMemory )
        {
            Memory_Free( m_pWorkMemory );
        }
    }
    bool Polling(){
        if ( !m_IsConnect && m_IsRunning)
        {
            nn::Result result = m_Serial.WaitHost( nn::hio::CTR::SerialChannel::ATTRIBUTE_NO_WAIT );
            if ( result == nn::hio::CTR::ResultConnected() )
            {
                m_IsConnect = true;
                return true;
            }
        }
        return false;
    }
    u32 GetReadableBytes(){
        read_buffer();
        return m_ReadSize;
    }
    u32 Read(void* data, u32 size){
        if ( GetReadableBytes() < size || size==0)
        {
            return 0;
        }
        read(data,size);
        return size;
    }
    u32 Peek(void* data, u32 size){
        if ( GetReadableBytes() < size )
        {
            return 0;
        }
        peek(data,size);
        return size;
    }
    u32 Skip(u32 size){
        if ( GetReadableBytes() < size )
        {
            return 0;
        }
        skip(size);
        return size;
    }
    u32 GetWritableBytes( u32* pWritableBytes ){
        if( m_IsConnect ){
            nn::Result result = m_Serial.GetWritableSize((size_t*)&pWritableBytes, nn::hio::CTR::SerialChannel::ATTRIBUTE_IN_PC );
            if( !process_result( result ))
                return 0;
            if(result.IsSuccess())
                return 1;
        }
        return 0;
    }
    u32 Write(const void* data, u32 size){
        if( !m_IsConnect )
            return 0;
        u32 writesize = 0;
        nn::Result result = m_Serial.Write(&writesize,
                                            (size_t*)data,
                                            size,
                                            nn::hio::CTR::SerialChannel::ATTRIBUTE_NONE);
        process_result( result );
        // process_result( m_Serial.Flush() );
        return size;
    }
    u8* GetBuffer() const
    {
        return m_Buffer;
    }
    bool IsConnect() const
    {
        return m_IsConnect;
    }

private:
    void read(void* data, u32 size){
        peek(data,size);
        skip(size);
    }
    void peek(void* data, u32 size){
        u8* dst = (u8*)data;
        size_t read_size = nw::math::Min(size, m_BufSize-m_Cursor);
        std::memcpy( dst, &m_Buffer[m_Cursor], read_size );
        if(size-read_size > 0){
            std::memcpy( &dst[read_size], m_Buffer,  size-read_size);
        }
    }
    void skip(u32 size){
        m_Cursor = (m_Cursor+size) % m_BufSize;
        m_ReadSize -= size;
    }
    void read_buffer(){
        if ( !m_IsConnect )
            return;

        if( m_ReadSize >= m_BufSize )
            return;

        // get readable size
        size_t ReadableSize=0;
        nn::Result result = m_Serial.Read(&ReadableSize, m_TempBuffer, m_BufSize-m_ReadSize, nn::hio::CTR::SerialChannel::ATTRIBUTE_NO_WAIT);
        if ( !process_result(result) ){
            return;
        }
        if( ReadableSize == 0 ){
            return;
        }
        u32 write_start = (m_Cursor+m_ReadSize) % m_BufSize;
        u32 write_end   = (write_start+ReadableSize) % m_BufSize;
        if(write_start < write_end){
            memcpy(m_Buffer+write_start, m_TempBuffer, write_end-write_start);
        }else{
            u32 size = m_BufSize-write_start;
            memcpy(m_Buffer+write_start, m_TempBuffer, size);
            u32 rest = ReadableSize - size;
            memcpy(m_Buffer+write_end-rest, m_TempBuffer+size, rest);
        }
        m_ReadSize += ReadableSize;
    }
    bool process_result(nn::Result result){
        if (result == nn::Result(nn::Result::LEVEL_PERMANENT, nn::Result::SUMMARY_STATUS_CHANGED, nn::Result::MODULE_NN_HIO, 8))
        {
            m_Serial.Disconnect();
            m_IsConnect = false;
            return false;
        }
        if ( result.IsFailure() )
        {
            m_Serial.Disconnect();
            m_IsConnect = false;
            return false;
        }
        return true;
    }

private:
    nn::hio::CTR::SerialChannel m_Serial;
    u32                         m_ReadSize;
    u32                         m_BufSize;
    u32                         m_Cursor;
    u8*                         m_Buffer;
    u8*                         m_TempBuffer;
    bool                        m_IsConnect;
    bool                        m_IsRunning;
    void*                       m_pWorkMemory;
};


static nw::ut::MoveArray<SerialChannel*> s_Serial;


/*---------------------------------------------------------------------------*
  Name:         OpenDevice_
  Description:  Opens device for HostI/O
  Arguments:    None.
  Returns:      true        When opening succeeded
                false       When it failed

 *---------------------------------------------------------------------------*/
static bool
OpenDevice_( void )
{
    return true;
}


/*---------------------------------------------------------------------------*
  Name:         Stream_Init
  Description:  HostI/O initialization
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------*/
void
Stream_Init( void )
{
    bool isRunning;
    nn::Result ret = nn::hio::GetHioDaemonStatus( &isRunning );
    if(ret.IsSuccess() && isRunning == false)
    {
        GetLogSystem()->AddSystemLog( "HIODaemon is Disconnected" );
    }

    const u32 NumSerial = 32;
    s_Serial = nw::ut::MoveArray<SerialChannel*>( NumSerial, vw::system::Memory_GetSysAllocator() );
    s_Serial.Resize(NumSerial);
    for(u32 i=0; i<NumSerial; ++i)
    {
        s_Serial[i] = 0;
    }
}


/*---------------------------------------------------------------------------*
  Name:         Stream_UnInit
  Description:  HostI/O finalization
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------*/
void Stream_UnInit( void )
{
    for(u32 i=0; i<s_Serial.size(); ++i)
    {
        if( s_Serial[i] != NULL )
        {
            s_Serial[i]->~SerialChannel();
            vw::system::Memory_Free(s_Serial[i]);
        }
    }
    s_Serial = nw::ut::MoveArray<SerialChannel*>();
}

/*---------------------------------------------------------------------------*
  Name:         Stream_IsConnected
  Description:  Flag to confirm whether connected to the MCS server
  Arguments:    None.
  Returns:      TRUE        Connected
                FALSE       Not connected
 *---------------------------------------------------------------------------*/
bool
Stream_IsConnected( u16 channel )
{
    SerialChannel* serial = s_Serial[channel];
    return serial->IsConnect();
}


/*---------------------------------------------------------------------------*
  Name:         Stream_Polling
  Description:  HostI/O polling process
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------*/
bool
Stream_Polling( u16 channel )
{
    SerialChannel* serial = s_Serial[channel];
    return serial->Polling();
}


/*---------------------------------------------------------------------------*
  Name:         Stream_GetFileSize
  Description:  Gets file size of specified file on the PC
  Arguments:    fileName: Filename
  Returns:      Returns the file size.
                0 if file cannot be found
 *---------------------------------------------------------------------------*/
u32
Stream_GetFileSize( const char *fileName )
{
    const int FILE_PATH_MAX = 260;
    if ( std::strlen( fileName ) > FILE_PATH_MAX )
    // Cannot load when the filename exceeds the max
    {
        return 0;
    }

    bool IsRunning = false;
    nn::Result ret = nn::hio::GetHioDaemonStatus( &IsRunning );
    if(ret.IsSuccess() || IsRunning == false)
    {
        return 0;
    }

    // Opens files
    nn::hio::CTR::HostFile hostFile;
    nn::Result result = hostFile.Open(
        fileName,
        nn::hio::CTR::HostFile::ACCESS_MODE_READ,
        nn::hio::CTR::HostFile::OPEN_DISP_OPEN_EXISTING);
    if (result.IsFailure())
    {
        return 0;
    }
    // Get the file size
    s64 filesize;
    hostFile.Seek(&filesize,0,nn::hio::CTR::HostFile::SEEK_TYPE_END);
    hostFile.Seek(0,nn::hio::CTR::HostFile::SEEK_TYPE_SET);
    if ( filesize == 0 )
    {
        return 0;
    }

    return u32(filesize);
}


/*---------------------------------------------------------------------------*
  Name:         Stream_ReadFile
  Description:  Loads file on the PC to the specified buffer
  Arguments:    fileName: Filename
                buf             Buffer where the file is loaded
  Returns:      Returns the file size.
                0 if file cannot be found
 *---------------------------------------------------------------------------*/
u32
Stream_ReadFile( const char *fileName, void* buf )
{
    const int FILE_PATH_MAX = 260;
    if ( std::strlen( fileName ) > FILE_PATH_MAX )
    // Cannot load when the filename exceeds the max
    {
        return 0;
    }

    bool IsRunning = false;
    nn::Result ret = nn::hio::GetHioDaemonStatus( &IsRunning );
    if(ret.IsSuccess() || IsRunning == false)
    {
        return 0;
    }

    // Opens files
    nn::hio::CTR::HostFile hostFile;
    nn::Result result = hostFile.Open(
        fileName,
        nn::hio::CTR::HostFile::ACCESS_MODE_READ,
        nn::hio::CTR::HostFile::OPEN_DISP_OPEN_EXISTING);
    if (result.IsFailure())
    {
        return 0;
    }
    // Get the file size
    s64 filesize;
    hostFile.Seek(&filesize,0,nn::hio::CTR::HostFile::SEEK_TYPE_END);
    hostFile.Seek(0,nn::hio::CTR::HostFile::SEEK_TYPE_SET);
    if ( filesize == 0 )
    {
        return 0;
    }

    if ( ! buf )
    {
        return 0;
    }

    // Load from file
    size_t rSize = 0;
    result = hostFile.Read(&rSize,buf,(s32)filesize);
    if( result.IsFailure() )
    {
        return 0;
    }
    hostFile.Close();

    return rSize;
}


/*---------------------------------------------------------------------------*
  Name:         Stream_WriteFile
  Description:  Writes the data in memory to the PC file
  Arguments:    fileName: Filename
                buf             Pointer to the data to write
                size            Data size
  Returns:      The actual data size that was written
                0 if an error occurred midway
 *---------------------------------------------------------------------------*/
u32
Stream_WriteFile( const char *fileName, const void* buf, u32 size )
{
    const int FILE_PATH_MAX = 260;
    if ( std::strlen( fileName ) > FILE_PATH_MAX )
    // Cannot load when the filename exceeds the max
    {
        return 0;
    }

    bool IsRunning = false;
    nn::Result ret = nn::hio::GetHioDaemonStatus( &IsRunning );
    if(ret.IsSuccess() || IsRunning == false)
    {
        return 0;
    }

    // Opens files
    nn::hio::CTR::HostFile hostFile;
    nn::Result result = hostFile.Open(
        fileName,
        nn::hio::CTR::HostFile::ACCESS_MODE_WRITE,
        nn::hio::CTR::HostFile::OPEN_DISP_CREATE_ALWAYS);
    if (result.IsFailure())
    {
        return 0;
    }

    if ( ! buf )
    {
        return 0;
    }

    // Write to the file
    size_t wSize = 0;
    result = hostFile.Write(&wSize, buf, size);
    if( result.IsFailure() )
    {
        return 0;
    }
    hostFile.Close();

    return wSize;
}
/*---------------------------------------------------------------------------*
  Name:         Stream_CreateChannel
  Description:  Creates channel for host IO
  Arguments:    channel         Channel number (0 - 0x7FFF)
                bufSize         Buffer size to allocate
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
Stream_CreateChannel( u16 channel, u32 bufSize /* = STREAM_DEFAULT_BUF_SIZE */ )
{
    // Open SerialChannel
    if( s_Serial[channel] == NULL )
    {
        void* buf = vw::system::Memory_AllocSys(sizeof(SerialChannel));
        s_Serial[channel] = new(buf) SerialChannel(channel,bufSize);
    }
}


/*---------------------------------------------------------------------------*
  Name:         Stream_TryCreateChannel
  Description:  Creates channel for host IO
  Arguments:    channel         Channel number (0 - 0x7FFF)
                bufSize         Buffer size to allocate
  Returns:      Returns pointer to buffer
 *---------------------------------------------------------------------------


*/
void*
Stream_TryCreateChannel( u16 channel, u32 bufSize /* = STREAM_DEFAULT_BUF_SIZE */ )
{
    // Open SerialChannel
    if( s_Serial[channel] == NULL )
    {
        void* buf = vw::system::Memory_AllocSys(sizeof(SerialChannel));
        s_Serial[channel] = new(buf) SerialChannel(channel,bufSize);
    }
    return (void*)s_Serial[channel]->GetBuffer();
}



/*---------------------------------------------------------------------------*
  Name:         Stream_GetReadableBytes
  Description:  Gets the data size that can be loaded from MCS.
                Returns 0 if not connected normally.
  Arguments:    channel         Channel number (0 - 0x7FFF)
  Returns:      Size of data that can be loaded.
 *---------------------------------------------------------------------------


*/
u32
Stream_GetReadableBytes( u16 channel )
{
    SerialChannel* serial = s_Serial[channel];
    if ( serial == NULL || !serial->IsConnect() )
    {
        return 0;
    }
    return serial->GetReadableBytes();
}


/*---------------------------------------------------------------------------*
  Name:         Stream_Read
  Description:  Loads data of the size specified by MCS and returns the size of data loaded.
                Does nothing and returns 0 if data of the amount specified can be loaded but does not exist.
  Arguments:    channel         Channel number (0 - 0x7FFF)
                data            Pointer to the read start region
                size            Size of data to read
  Returns:      Size of the data actually read
 *---------------------------------------------------------------------------


*/
u32
Stream_Read( u16 channel, void* data, u32 size )
{
    SerialChannel* serial = s_Serial[channel];
    if ( serial == NULL || !serial->IsConnect() )
    {
        return 0;
    }
    return serial->Read(data,size);
}


/*---------------------------------------------------------------------------*
  Name:         Stream_Peek
  Description:  Loads data of the size specified by MCS and returns the size of data loaded.
                Note that obtained data is not cleared from the receive buffer.
                Does nothing and returns 0 if data of the amount specified can be loaded but does not exist.
  Arguments:    channel         Channel number (0 - 0x7FFF)
                data            Pointer to the read start region
                size            Size of data to read
  Returns:      Size of the data actually read
 *---------------------------------------------------------------------------


*/
u32
Stream_Peek( u16 channel, void* data, u32 size )
{
    SerialChannel* serial = s_Serial[channel];
    if ( serial == NULL || !serial->IsConnect() )
    {
        return 0;
    }
    return serial->Peek(data,size);
}


/*---------------------------------------------------------------------------*
  Name:         Stream_Skip
  Description:  Skips reading data of the specified size from MCS.
                Does nothing and returns 0 if data of the amount specified can be loaded but does not exist.
  Arguments:    channel         Channel number (0 - 0x7FFF)
                size            Size of data to skip reading
  Returns:      Size of the data actually skipped for reading
 *---------------------------------------------------------------------------


*/
u32
Stream_Skip( u16 channel, u32 size )
{
    SerialChannel* serial = s_Serial[channel];
    if ( serial == NULL || !serial->IsConnect() )
    {
        return 0;
    }
    return serial->Skip(size);
}


/*---------------------------------------------------------------------------*
  Name:         Stream_GetWritableBytes
  Description:  Obtains the size (number of bytes) that can be sent at once.
                When writing data that exceeds the size that can be sent once with Mcs_Write, function calls are blocked on the PC side until the written data can be read, and control is not returned.
                
                
  Arguments:    channel         Channel number (0 - 0x7FFF)
                pWritableBytes  Pointer to the variable that has the size (number of bytes) that can be sent at once.
  Returns:      0 if the function succeeds; a non-zero value otherwise.
 *---------------------------------------------------------------------------


*/
u32
Stream_GetWritableBytes( u16 channel, u32* pWritableBytes )
{
    SerialChannel* serial = s_Serial[channel];
    if ( serial == NULL || !serial->IsConnect() )
    {
        return 0;
    }
    return serial->GetWritableBytes(pWritableBytes);
}

/*---------------------------------------------------------------------------*
  Name:         Stream_Write
  Description:  Sends data.
  Arguments:    channel         Channel number (0 - 0x7FFF)
                data            Pointer to the data to be sent
                size            Size of the data to send
  Returns:      0 if the function succeeds; a non-zero value otherwise.
 *---------------------------------------------------------------------------


*/
u32
Stream_Write( u16 channel, const void* data, u32 size )
{
    SerialChannel* serial = s_Serial[channel];
    if ( serial == NULL || !serial->IsConnect() )
    {
        return 0;
    }
    return serial->Write( data, size );
}


}   /* namespace system */
}   /* namespace vw     */
#if NN_CURRENT_VERSION_NUMBER >= NN_VERSION_NUMBER(4,0,0,0)
#pragma diag_default 1301 // padding inserted in struct.
#endif


