/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#include <viewer/system.h>
#include "../include/TgaData.h"
#include "../include/TexUtil.h"

namespace vw     {
namespace system {


//===============================================================
//
// Temporary buffer class for reading and writing data
//
//===============================================================
class ReadWriteBuffer
{
public:
    /* ctor */  ReadWriteBuffer( void* src, u32 cacheSize );
    /* dtor */  ~ReadWriteBuffer() { if ( mCache ) { Memory_Free( mCache ); } }

    u8          Peek() { return *mReadPtr; }
    void        Write( u8 data );
    void        SeekReadPtr( u32 size );

    u32         PeekU24() { return PeekU24( mReadPtr ); }
    static u32  PeekU24( const void* addr );
    void        WriteU24( u32 data );

    u32         PeekU32() { return PeekU32( mReadPtr ); }
    static u32  PeekU32( const void* addr );
    void        WriteU32( u32 data );

    bool        IsValid() const { return (mCache != NULL)? true : false; }

    static u32  PeekUBytes( const void* addr, u8 bytes )
        {
            NW_ASSERT( bytes == 3 || bytes == 4 );
            if ( bytes == 3 )
            {
                return PeekU24( addr );
            }
            return PeekU32( addr );
        }

    void        WriteUBytes( u32 data, u8 bytes )
        {
            NW_ASSERT( bytes == 3 || bytes == 4 );
            if ( bytes == 3 )
            {
                this->WriteU24( data );
            }
            else
            {
                this->WriteU32( data );
            }
        }

private:
    u8*     mBuf;
    u8*     mReadPtr;
    u8*     mWritePtr;
    u8*     mCache;
    u8*     mCacheHead;
    u8*     mCacheTail;
    u32     mCacheSize;

    bool    CacheIsEmpty_() const { return (mCacheHead == mCacheTail); }

    void    PushToCache_( u8 data )
    {
        *mCacheHead++ = data;
        if ( mCacheHead == (mCache + mCacheSize) )
        {
            mCacheHead = mCache;
        }
        NW_ASSERT( mCacheHead != mCacheTail );
    }

    u8      PopFromCache_()
    {
        u8 data = *mCacheTail++;
        if ( mCacheTail == (mCache + mCacheSize) )
        {
            mCacheTail = mCache;
        }
        return data;
    }
};


/*---------------------------------------------------------------------------*
  Name:         ReadWriteBuffer::ReadWriteBuffer
  Description:  Constructor
  Arguments:    src
                cacheSize
  Returns:      None.
 *---------------------------------------------------------------------------


*/
/* ctor */
ReadWriteBuffer::ReadWriteBuffer( void* src, u32 cacheSize )
 : mBuf( reinterpret_cast<u8*>(src) ),
   mReadPtr( mBuf ),
   mWritePtr( mBuf ),
   mCacheSize( cacheSize )
{
    mCache = reinterpret_cast<u8*>( Memory_AllocSys( cacheSize ) );
    mCacheHead = mCache;
    mCacheTail = mCache;
}


/*---------------------------------------------------------------------------*
  Name:         ReadWriteBuffer::Seek
  Description:  Moves the read position by the specified size amount
  Arguments:    size    Move size
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
ReadWriteBuffer::SeekReadPtr( u32 size )
{
    mReadPtr += size;

    while ( size > 0 )
    {
        if ( this->CacheIsEmpty_() )
        {
            break;
        }
        *mWritePtr++ = this->PopFromCache_();
        --size;
    }
}


/*---------------------------------------------------------------------------*
  Name:         ReadWriteBuffer::Write
  Description:  Writes data
  Arguments:    data
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
ReadWriteBuffer::Write( u8 data )
{
    if ( mReadPtr > mWritePtr )
    {
        *mWritePtr++ = data;
    }
    else
    {
        this->PushToCache_( data );
    }
}


/*---------------------------------------------------------------------------*
  Name:         ReadWriteBuffer::PeekU24
  Description:  Reads 24-bit data from the specified address.
  Arguments:    addr    Address
  Returns:      24-bit data
 *---------------------------------------------------------------------------


*/
/* static */ u32
ReadWriteBuffer::PeekU24( const void* addr )
{
    const u8* p = reinterpret_cast<const u8*>( addr );
    u32 val = static_cast<u32>( (*(p + 2) << 16) + (*(p + 1) << 8) + *p );
    return val;
}

/*---------------------------------------------------------------------------*
  Name:         ReadWriteBuffer::WriteU24
  Description:  Writes 24-bit data.
  Arguments:    data    Data
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
ReadWriteBuffer::WriteU24( u32 data )
{
    this->Write( static_cast<u8>( data & 0xFF ) );
    this->Write( static_cast<u8>( (data >> 8)  & 0xFF ) );
    this->Write( static_cast<u8>( (data >> 16) & 0xFF ) );
}


/*---------------------------------------------------------------------------*
  Name:         ReadWriteBuffer::PeekU32
  Description:  Reads 32-bit data from the specified address.
  Arguments:    Address
  Returns:      32-bit data
 *---------------------------------------------------------------------------


*/
/* static */ u32
ReadWriteBuffer::PeekU32( const void* addr )
{
    const u8* p = reinterpret_cast<const u8*>( addr );
    u32 val = static_cast<u32>( (*(p + 3) << 24) + (*(p + 2) << 16) + (*(p + 1) << 8) + *p );
    return val;
}

/*---------------------------------------------------------------------------*
  Name:         ReadWriteBuffer::WriteU32
  Description:  Writes 32-bit data.
  Arguments:    data    32-bit data to write
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
ReadWriteBuffer::WriteU32( u32 data )
{
    this->Write( static_cast<u8>( data & 0xFF ) );
    this->Write( static_cast<u8>( (data >> 8)  & 0xFF ) );
    this->Write( static_cast<u8>( (data >> 16) & 0xFF ) );
    this->Write( static_cast<u8>( (data >> 24) & 0xFF ) );
}



/*---------------------------------------------------------------------------*
  Name:         TgaData::TgaData
  Description:  Constructor
  Arguments:    width   Image horizontal size
                height  Image vertical size
                bpp     Image bpp size
  Returns:      None.
 *---------------------------------------------------------------------------


*/
/* ctor */
TgaData::TgaData( u32 width, u32 height, u8 bpp /* = 3 */ )
 : ImageFmt( width, height, bpp )
{
    u32 size = mWidth * mHeight * bpp;
    u32 allocSize = nn::math::RoundUp(mWidth, 4) * nn::math::RoundUp(mHeight, 4) * bpp;
    mBody = Memory_AllocSys( allocSize, 128 );

    mBodySize = width * height * bpp;

    mHeader.bfIdLength = 0x00;
    mHeader.bfType     = 0;
    mHeader.bfFormat   = FMT_RGB;
    LEndian::WriteU16( mHeader.bfPlttIdx,    0 );
    LEndian::WriteU16( mHeader.bfPlttLength, 0 );
    mHeader.bfPlttSize = 0;
    LEndian::WriteU16( mHeader.bfOriginX,    0 );
    LEndian::WriteU16( mHeader.bfOriginY,    0 );
    LEndian::WriteU16( mHeader.bfWidth,      static_cast<u16>( mWidth ) );
    LEndian::WriteU16( mHeader.bfHeight,     static_cast<u16>( mHeight ) );
    mHeader.bfBpp      = static_cast<u8>( mBpp * 8 );
    mHeader.bfDescript = 0x08;

    LEndian::WriteU32( mFooter.bfOffset,     0 );
    LEndian::WriteU32( mFooter.bfDevOffset,  0 );
    std::strcpy( reinterpret_cast<char*>( mFooter.bfSignature ), TGA_SIGNATURE );
    mFooter.bfPadding_ = 0;
}


/*---------------------------------------------------------------------------*
  Name:         TgaData::~TgaData
  Description:  Destructor
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
/* dtor */
TgaData::~TgaData( void )
{
}


/*---------------------------------------------------------------------------*
  Name:         TgaData::ConvertData
  Description:  Converts EFB data to Tga format.
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
TgaData::ConvertData( void )
{
    this->ConvertDataImpl_();
    this->CompressDataImpl_();
}


/*---------------------------------------------------------------------------*
  Name:         TgaData::ConvertData
  Description:  Converts texture format RBGA data to Tga data.
  Arguments:    src_image   The pointer to the texture data
                width       Texture horizontal size
                height      Texture vertical size
                dst_ofs_x   X position in the overall TGA image
                dst_ofs_y   Y position in the overall TGA image
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
TgaData::ConvertData( void* src_image, u32 width, u32 height, u32 dst_ofs_x, u32 dst_ofs_y )
{
    this->ConvertDataImpl_( src_image, width, height, dst_ofs_x, dst_ofs_y, 0, 0, width, height );
}


/*---------------------------------------------------------------------------*
  Name:         TgaData::ConvertData
  Description:  Converts texture format RBGA data to Tga data.
  Arguments:    src_image   The pointer to the texture data
                src_width   Texture horizontal size
                src_height  Texture vertical size
                dst_ofs_x   X position in the overall TGA image
                dst_ofs_y   Y position in the overall TGA image
                src_ofs_x   X-offset of the rectangle used in the texture
                src_ofs_y   Y-offset of the rectangle used in the texture
                width       Horizontal size of the rectangle used in the texture
                height      Vertical size of the rectangle used in the texture
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
TgaData::ConvertData( void* src_image,
                      u32 src_width, u32 src_height,
                      u32 dst_ofs_x, u32 dst_ofs_y,
                      u32 src_ofs_x, u32 src_ofs_y,
                      u32 width,     u32 height )
{
    this->ConvertDataImpl_( src_image, src_width, src_height, dst_ofs_x, dst_ofs_y, src_ofs_x, src_ofs_y, width, height );
}


/*---------------------------------------------------------------------------*
  Name:         TgaData::ConvertAlphaData
  Description:  Converts EFB data to Tga format.
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
TgaData::ConvertAlphaData( void )
{
    this->ConvertAlphaDataImpl_();
    this->CompressDataImpl_();
}

/*---------------------------------------------------------------------------*
  Name:         TgaData::ConvertAlphaData
  Description:  Performs substitution processing from the RGB8 texture data to the alpha channel
  Arguments:    src_image   The pointer to the texture data
                width       Texture horizontal size
                height      Texture vertical size
                dst_ofs_x   X position in the overall image
                dst_ofs_y   Y position in the overall image
                op          Calculation specifications when substituting
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
TgaData::ConvertAlphaData( void* src_image, u32 width, u32 height, u32 dst_ofs_x, u32 dst_ofs_y, Operation op )
{
    this->ConvertAlphaDataImpl_( src_image, width, height, dst_ofs_x, dst_ofs_y, 0, 0, width, height, op );
}


/*---------------------------------------------------------------------------*
  Name:         TgaData::ConvertAlphaData
  Description:  Performs substitution processing from the RGB8 texture data to the alpha channel
  Arguments:    src_image   The pointer to the texture data
                src_width   Texture horizontal size
                src_height  Texture vertical size
                dst_ofs_x   X position in the overall BMP image
                dst_ofs_y   Y position in the overall BMP image
                src_ofs_x   X-offset of the rectangle used in the texture
                src_ofs_y   Y-offset of the rectangle used in the texture
                width       Horizontal size of the rectangle used in the texture
                height      Vertical size of the rectangle used in the texture
                op          Calculation specifications when substituting
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
TgaData::ConvertAlphaData( void* src_image,
                      u32 src_width, u32 src_height,
                      u32 dst_ofs_x, u32 dst_ofs_y,
                      u32 src_ofs_x, u32 src_ofs_y,
                      u32 width,     u32 height,
                      Operation op )
{
    this->ConvertAlphaDataImpl_( src_image, src_width, src_height, dst_ofs_x, dst_ofs_y, src_ofs_x, src_ofs_y, width, height, op );
}


/*---------------------------------------------------------------------------*
  Name:         TgaData::Finalize
  Description:  Performs data finalization such as compression and puts into a state that can be output to the PC as a file.
                
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
TgaData::Finalize( void )
{
#if 0
    this->CompressDataImpl_();
#else
    if ( this->GetCompressedDataSize_() < this->GetBodySize() )
    {
        this->CompressDataImplThin_();
    }
#endif
}


/*---------------------------------------------------------------------------*
  Name:         TgaData::ConvertDataImpl_
  Description:  Performs data sorting and bit field conversion.
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
TgaData::ConvertDataImpl_( void )
{
    const u32 SRC_BPP = 4;

    u32 allocSize = nn::math::RoundUp(mWidth, 4) * nn::math::RoundUp(mHeight, 4) * SRC_BPP;
    u8* pTmp   = static_cast<u8*>( Memory_AllocSys( allocSize ) );
    std::memcpy( pTmp, mBody, mWidth * mHeight * SRC_BPP );

    this->ConvertDataImpl_( pTmp, mWidth, mHeight, 0, 0, 0, 0, mWidth, mHeight );

    Memory_Free( pTmp );
}

/*---------------------------------------------------------------------------*
  Name:         TgaData::ConvertAlphaDataImpl_
  Description:  Performs data sorting for alpha only and bit field conversion.
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
TgaData::ConvertAlphaDataImpl_( void )
{
    const u32 SRC_BPP = 4;

    u32 allocSize = nn::math::RoundUp(mWidth, 4) * nn::math::RoundUp(mHeight, 4) * SRC_BPP;
    u8* pTmp   = static_cast<u8*>( Memory_AllocSys( system::SYSTEM_ID, allocSize ) );
    std::memcpy( pTmp, mBody, mWidth * mHeight * SRC_BPP );

    this->ConvertAlphaDataImpl_( pTmp, mWidth, mHeight, 0, 0, 0, 0, mWidth, mHeight );

    Memory_Free( pTmp );
}

/*---------------------------------------------------------------------------*
  Name:         TgaData::ConvertDataImpl_
  Description:  Converts texture format RBGA data to Tga data.
  Arguments:    src_image   The pointer to the texture data
                src_width   Texture horizontal size
                src_height  Texture vertical size
                dst_ofs_x   X position in the overall TGA image
                dst_ofs_y   Y position in the overall TGA image
                src_ofs_x   X-offset of the rectangle used in the texture
                src_ofs_y   Y-offset of the rectangle used in the texture
                width       Horizontal size of the rectangle used in the texture
                height      Vertical size of the rectangle used in the texture
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
TgaData::ConvertDataImpl_(
    void*   src_image,
    u32     src_width, u32 src_height,
    u32     dst_ofs_x, u32 dst_ofs_y,
    u32     src_ofs_x, u32 src_ofs_y,
    u32     width,     u32 height )
{
    const u32 SRC_BPP = 4;

    u8* pRead  = reinterpret_cast<u8*>( src_image );
    u8* pWrite = reinterpret_cast<u8*>( mBody );
    u32 i;

    u32 line = 0;

    for ( line = 0; line < height; line++ )
    {
        u32 y = src_height - line - 1 - (src_height - height - src_ofs_y);
        u32 dst_addr = ( (dst_ofs_y + line) * mWidth + dst_ofs_x ) * mBpp;
        pWrite = reinterpret_cast<u8*>( mBody ) + dst_addr;

        for ( i = 0; i < width; i ++ )
        {
            u32 x = i + src_ofs_x;
            nw::ut::Color8 col = UtilRGBA8::GetColor( pRead, x, y, src_width );

            switch ( mBpp )
            {
            //---------------
            // Output 24-bit color BMP
            case 3:
                {
                    *pWrite++   = col.b;
                    *pWrite++   = col.g;
                    *pWrite++   = col.r;
                }
                break;
            //---------------
            // Output 32-bit color BMP
            case 4:
                {
                    *pWrite++   = col.b;
                    *pWrite++   = col.g;
                    *pWrite++   = col.r;
                    *pWrite++   = col.a;
                }
            }
        }
    }
}

/*---------------------------------------------------------------------------*
  Name:         TgaData::ConvertDataImpl_
  Description:  Performs substitution processing from the RGB8 texture data to the alpha channel
  Arguments:    src_image   The pointer to the texture data
                src_width   Texture horizontal size
                src_height  Texture vertical size
                dst_ofs_x   X position in the overall BMP image
                dst_ofs_y   Y position in the overall BMP image
                src_ofs_x   X-offset of the rectangle used in the texture
                src_ofs_y   Y-offset of the rectangle used in the texture
                width       Horizontal size of the rectangle used in the texture
                height      Vertical size of the rectangle used in the texture
                op          Calculation specifications when substituting
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
TgaData::ConvertAlphaDataImpl_(
    void*   src_image,
    u32     src_width, u32 src_height,
    u32     dst_ofs_x, u32 dst_ofs_y,
    u32     src_ofs_x, u32 src_ofs_y,
    u32     width,     u32 height,
    Operation operarion )
{
    const u32 SRC_BPP = 4;

    u8* pRead  = reinterpret_cast<u8*>( src_image );
    u8* pWrite = reinterpret_cast<u8*>( mBody );
    u32 i;

    u32 line = 0;

    for ( line = 0; line < height; line++ )
    {
        u32 y = src_height - line - 1 - (src_height - height - src_ofs_y);
        u32 dst_addr = ( (dst_ofs_y + line) * mWidth + dst_ofs_x ) * mBpp;
        pWrite = reinterpret_cast<u8*>( mBody ) + dst_addr;

        for ( i = 0; i < width; i ++ )
        {
            u32 x = i + src_ofs_x;
            nw::ut::Color8 col = UtilRGBA8::GetColor( pRead, x, y, src_width );

            switch ( mBpp )
            {
            //---------------
            // Output 24-bit color BMP
            case 3:
                {
                    pWrite += 3;
                }
                break;
            //---------------
            // Output 32-bit color BMP
            case 4:
                {
                    pWrite += 3;
                    switch ( operarion )
                    {
                    case OPE_SETVALUE:
                        *pWrite++ = col.r;
                        break;
                    case OPE_ADD:
                        *pWrite++ += col.r;
                        break;
                    case OPE_DEC:
                        *pWrite++ -= col.r;
                        break;
                    case OPE_NOTDEC:
                        *pWrite = static_cast<u8>(~(*pWrite - col.r));
                        pWrite++;
                        break;
                    }
                }
            }
        }
    }
}


/*---------------------------------------------------------------------------*
  Name:         TgaData::GetCompressedDataSize_
  Description:  Calculates data size when data is compressed with RLE.
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
u32
TgaData::GetCompressedDataSize_( void )
{
    if ( mHeader.bfFormat == FMT_CMP_RGB24 )
    {
        return mBodySize;
    }

    u32 srcSize = mWidth * mHeight * mBpp;
    u32 dstSize = 0;
    u8* pSrc = reinterpret_cast<u8*>( mBody );
    u8* pSrcEnd  = reinterpret_cast<u8*>( mBody ) + srcSize;
    u8* pLineEnd = pSrc + mWidth * mBpp;
    u32 lineNum = 0;

    while ( pSrc < pSrcEnd )
    {
        u8* p = pSrc;
        u32 state = LEndian::ReadUBytes( p, mBpp );
        p += mBpp;

        if ( ( p < pLineEnd ) && LEndian::ReadUBytes( p, mBpp ) == state )
        // RUN data
        {
            u32 runLen = 1;
            do
            {
                p += mBpp;
                if ( ++runLen == 128 || ( p >= pLineEnd ) )
                // Maximum length 128
                {
                    break;
                }
            } while ( LEndian::ReadUBytes( p, mBpp ) == state );

            // Write data to output buffer
            dstSize += 4;
            pSrc = p;
        }
        else
        // RAW data
        {
            u32 runLen = 0;
            do
            {
                if ( p >= pLineEnd )
                {
                    break;
                }

                state = LEndian::ReadUBytes( p, mBpp );
                p += mBpp;
                if ( ++runLen == 128 )
                // Maximum length 128
                {
                    break;
                }
            } while ( LEndian::ReadUBytes( p, mBpp ) != state );

            if ( p >= pLineEnd && runLen < 128  )
            {
                ++runLen;
            }

            // Write data to output buffer
            dstSize += (runLen * mBpp + 1);
            pSrc += runLen * mBpp;
        }
        if ( pSrc >= pLineEnd )
        {
            pLineEnd += mWidth * mBpp;
            ++lineNum;
        }
    }
    return dstSize;
}

/*---------------------------------------------------------------------------*
  Name:         TgaData::CompressDataImpl_
  Description:  RLE compression function
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
TgaData::CompressDataImpl_( void )
{
    if ( mHeader.bfFormat == FMT_CMP_RGB24 )
    {
        return;
    }

    u32 srcSize = mWidth * mHeight * mBpp;
    u32 dstSize = 0;
    u8* pBuf = reinterpret_cast<u8*>( Memory_AllocSys( srcSize, 128 ) );
    u8* pSrc = reinterpret_cast<u8*>( mBody );
    u8* pDst = pBuf;
    u8* pSrcEnd = reinterpret_cast<u8*>( mBody ) + srcSize;
    u8* pLineEnd = pSrc + mWidth * mBpp;
    u32 lineNum = 0;

    while ( pSrc < pSrcEnd )
    {
        u8* p = pSrc;
        u32 state = LEndian::ReadU24( p );
        p += 3;

        if ( ( p < pLineEnd ) && LEndian::ReadU24( p ) == state )
        // RUN data
        {
            u32 runLen = 1;
            do
            {
                p += 3;
                if ( ++runLen == 128 || ( p >= pLineEnd ) )
                // Maximum length 128
                {
                    break;
                }
            } while ( LEndian::ReadU24( p ) == state );

            // Write data to output buffer
            dstSize += 4;
            if ( dstSize > srcSize )
            {
                Memory_Free( pBuf );
                return;
            }
            *pDst++ = static_cast<u8>( 0x80 | (runLen - 1) );
            LEndian::WriteU24( pDst, state );
            pDst += 3;
            pSrc = p;
        }
        else
        // RAW data
        {
            u32 runLen = 0;
            do
            {
                if ( p >= pLineEnd )
                {
                    break;
                }

                state = LEndian::ReadU24( p );
                p += 3;
                if ( ++runLen == 128 )
                // Maximum length 128
                {
                    break;
                }
            } while ( LEndian::ReadU24( p ) != state );

            if ( p >= pLineEnd && runLen < 128 )
            {
                ++runLen;
            }

            // Write data to output buffer
            dstSize += (runLen * 3 + 1);
            if ( dstSize > srcSize )
            {
                Memory_Free( pBuf );
                return;
            }
            *pDst++ = static_cast<u8>( runLen - 1 );
            for ( u32 i = 0; i < runLen * 3; i++ )
            {
                *pDst++ = *pSrc++;
            }
        }
        if ( pSrc >= pLineEnd )
        {
            pLineEnd += mWidth * mBpp;
            ++lineNum;
        }
    }

    Memory_Free( mBody );
    mBody = pBuf;
    mBodySize = dstSize;
    mHeader.bfFormat = FMT_CMP_RGB24;
}



/*---------------------------------------------------------------------------*
  Name:         TgaData::CompressDataImplThin_
  Description:  Must check if memory conservation compression and compression in advance with the GetCompressedDataSize_ function are permitted.
                
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
TgaData::CompressDataImplThin_( void )
{
    if ( mHeader.bfFormat == FMT_CMP_RGB24 )
    {
        return;
    }

    u32 srcSize = mWidth * mHeight * mBpp;
    u32 dstSize = 0;
    u32 bufSize = srcSize / 128 * 3;
    ReadWriteBuffer buffer( mBody, bufSize );
    u8* pSrc = reinterpret_cast<u8*>( mBody );
    u8* pSrcEnd = reinterpret_cast<u8*>( mBody ) + srcSize;
    u8* pLineEnd = pSrc + mWidth * mBpp;
    u32 lineNum = 0;

    if ( ! buffer.IsValid() )
    {
        return;
    }

    while ( pSrc < pSrcEnd )
    {
        u8* p = pSrc;
        u32 state = buffer.PeekUBytes( p, mBpp );
        p += mBpp;

        if ( ( p < pLineEnd ) && buffer.PeekUBytes( p, mBpp ) == state )
        // RUN data
        {
            buffer.SeekReadPtr( mBpp );
            u32 runLen = 1;
            do
            {
                p += mBpp;
                buffer.SeekReadPtr( mBpp );
                if ( ++runLen == 128 || ( p >= pLineEnd ) )
                // Maximum length 128
                {
                    break;
                }
            } while ( buffer.PeekUBytes( p, mBpp ) == state );

            // Write data to output buffer
            dstSize += (mBpp + 1);
            if ( dstSize > srcSize )
            {
                return;
            }
            buffer.Write( u8( 0x80 | (runLen - 1) ) );
            buffer.WriteUBytes( state, mBpp );
            pSrc = p;
        }
        else
        // RAW data
        {
            u32 runLen = 0;
            do
            {
                if ( p >= pLineEnd )
                {
                    break;
                }

                state = buffer.PeekUBytes( p, mBpp );
                p += mBpp;
                if ( ++runLen == 128 )
                // Maximum length 128
                {
                    break;
                }
            } while ( buffer.PeekUBytes( p, mBpp ) != state );

            if ( p >= pLineEnd && runLen < 128  )
            {
                ++runLen;
            }

            // Write data to output buffer
            dstSize += (runLen * mBpp + 1);
            if ( dstSize > srcSize )
            {
                return;
            }
            buffer.Write( u8( runLen - 1 ) );
            for ( u32 i = 0; i < runLen; i++ )
            {
                u32 data = buffer.PeekUBytes( pSrc, mBpp );
                buffer.SeekReadPtr( mBpp );
                buffer.WriteUBytes( data, mBpp );
                pSrc += mBpp;
            }
        }
        if ( pSrc >= pLineEnd )
        {
            pLineEnd += mWidth * mBpp;
            ++lineNum;
        }
    }

    mBodySize = dstSize;
    mHeader.bfFormat = FMT_CMP_RGB24;
}


}   // namespace system
}   // namespace vw
