/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/
#include <viewer/system.h>
#include "../include/LogUtil.h"

namespace vw     {
namespace system {
namespace Utility{

/*
  
 */
const char* ErrorToString( ERROR errCode )
{
    switch (errCode)
    {
        case SUCCESS :
            return "Success";

        case OUT_OF_MEMORY :
            return "Out of memory";

        case INVALID_PARAMS :
            return "Invalid params";
    }

    return "Unknown error";
}

//----------------------------------------
nw::ut::MoveArray<u8>
LoadFile(
    nw::os::IAllocator* allocator,
    const wchar_t* filePath,
    u32 align /* = 32 */
)
{
    nn::fs::FileReader fileReader;

    nn::Result result = fileReader.TryInitialize(filePath);
    if (result.IsFailure())
    {
        return nw::ut::MoveArray<u8>();
    }

    s32 fileSize = static_cast<s32>(fileReader.GetSize());

    void* memory = allocator->Alloc(fileSize, static_cast<u8>(align));
    if(memory == 0)
    {
        return nw::ut::MoveArray<u8>();
    }

    nw::ut::MoveArray<u8> buffer(memory, fileSize, allocator);
    buffer.resize(fileSize);

    fileReader.Read(&buffer.front(), fileSize);

    fileReader.Finalize();

    // The container is returned, but the buffer is not copied.
    // Move the ownership.
    return buffer;
}


/*
  

  

  
 */
nw::ut::MoveArray<u8>
LoadFileFromHost(
    nw::os::IAllocator* allocator,
    const wchar_t* resourcePath,
    bool *isOnThe32MBLine)
{
    static const int RESOURCE_ALIGNMENT = 128;

    bool IsRunning = false;
    nn::Result ret = nn::hio::GetHioDaemonStatus( &IsRunning );
    if(ret.IsSuccess() && IsRunning == false)
    {
        return nw::ut::MoveArray<u8>();
    }

    // Opens files
    nn::hio::CTR::HostFile hostFile;
    nn::Result result = hostFile.Open(
        resourcePath,
        nn::hio::CTR::HostFile::ACCESS_MODE_READ,
        nn::hio::CTR::HostFile::OPEN_DISP_OPEN_EXISTING);

    if (result.IsFailure())
    {
        NW_VIEWER_DEV_LOG("File Open failed.\n");
    }

    // Get the file size
    s64 filesize;
    hostFile.Seek(&filesize,0,nn::hio::CTR::HostFile::SEEK_TYPE_END);
    hostFile.Seek(0,nn::hio::CTR::HostFile::SEEK_TYPE_SET);

    // Allocates the buffer.
    void* memory = allocator->Alloc(filesize, static_cast<u8>(RESOURCE_ALIGNMENT));
    if(memory == 0)
    {
        return nw::ut::MoveArray<u8>();
    }

    // 32 MB boundary check.
    if (NULL != isOnThe32MBLine)
    {
        char* startAddress = (char*)memory;
        char* endAddress = startAddress + static_cast<size_t>(filesize);
        *isOnThe32MBLine = ((u32)startAddress & ~0x1FFFFFF) != ((u32)endAddress & ~0x1FFFFFF);
    }

    nw::ut::MoveArray<u8> buffer(memory, static_cast<size_t>(filesize), allocator);
    if (!buffer.resize(filesize))
    {
        NW_VIEWER_DEV_LOG("Failed to resize buffer.\n");
    }

    // Load from file
    u8* pFront = &buffer.front();
    s32 rSize = hostFile.Read(pFront,filesize);
    const int MAX_PATH_SIZE = 260;
    char utf8Str[MAX_PATH_SIZE];
    if (!buffer)
    {
        //ConvertStrWtoUTF8(resourcePath, utf8Str, MAX_PATH_SIZE);
        //NW_VIEWER_DEV_LOG("Failed to read \"%s\".\n", utf8Str);
        //NW_VIEWER_DEV_LOG("Failed to read.\n");
    }
    else
    {
        //ConvertStrWtoUTF8(resourcePath, utf8Str, MAX_PATH_SIZE);
        //NW_VIEWER_DEV_LOG("Resource file \"%s\" loaded. [%d bytes]\n", utf8Str, rSize);
        //NW_VIEWER_DEV_LOG("Resource file loaded.\n");
    }

    hostFile.Close();

    return buffer;
}

/*
  

  

  
 */
nw::ut::MoveArray<u8>
LoadFileFromHost(
    nw::os::IAllocator* allocator,
    const char* resourcePath,
    bool *isOnThe32MBLine)
{
    NW_VIEWER_DEV_LOG("Now Loading....\n");

    static const int RESOURCE_ALIGNMENT = 128;

    bool IsRunning = false;
    nn::Result ret = nn::hio::GetHioDaemonStatus( &IsRunning );
    if(ret.IsSuccess() && IsRunning == false)
    {
        return nw::ut::MoveArray<u8>();
    }

    // Opens files
    nn::hio::CTR::HostFile hostFile;
    nn::Result result = hostFile.Open(
        resourcePath,
        nn::hio::CTR::HostFile::ACCESS_MODE_READ,
        nn::hio::CTR::HostFile::OPEN_DISP_OPEN_EXISTING);

    if (result.IsFailure())
    {
        NW_VIEWER_DEV_LOG("File Open failed.\n");
        nw::ut::MoveArray<u8> buffer(0, allocator);
        return buffer;
    }

    // Get the file size
    s64 filesize;
    hostFile.Seek(&filesize,0,nn::hio::CTR::HostFile::SEEK_TYPE_END);
    hostFile.Seek(0,nn::hio::CTR::HostFile::SEEK_TYPE_SET);

    // Allocates the buffer.
    void* memory = allocator->Alloc(filesize, static_cast<u8>(RESOURCE_ALIGNMENT));
    if(memory == 0)
    {
        return nw::ut::MoveArray<u8>();
    }

    // 32 MB boundary check.
    if (NULL != isOnThe32MBLine)
    {
        char* startAddress = (char*)memory;
        char* endAddress = startAddress + static_cast<size_t>(filesize);
        *isOnThe32MBLine = ((u32)startAddress & ~0x1FFFFFF) != ((u32)endAddress & ~0x1FFFFFF);
    }

    nw::ut::MoveArray<u8> buffer(memory, static_cast<size_t>(filesize), allocator);
    if (!buffer.resize(filesize))
    {
        NW_VIEWER_DEV_LOG("Failed to resize buffer.\n");
    }

    // Load from file
    u8* pFront = &buffer.front();
    s32 rSize = hostFile.Read(pFront,filesize);
    const int MAX_PATH_SIZE = 260;
    char utf8Str[MAX_PATH_SIZE];
    if (!buffer)
    {
        //ConvertStrWtoUTF8(resourcePath, utf8Str, MAX_PATH_SIZE);
        //NW_VIEWER_DEV_LOG("Failed to read \"%s\".\n", utf8Str);
        NW_VIEWER_DEV_LOG("Failed to read.\n");
    }
    else
    {
        //ConvertStrWtoUTF8(resourcePath, utf8Str, MAX_PATH_SIZE);
        //NW_VIEWER_DEV_LOG("Resource file \"%s\" loaded. [%d bytes]\n", utf8Str, rSize);
        NW_VIEWER_DEV_LOG("Resource file loaded.\n");
    }

    hostFile.Close();

    return buffer;
}

/*
  
 */
void LOG_ERROR( int errCode )
{
    // To be implemented
    if (errCode==SUCCESS)
        return;
}

} // Utility
} // system
} // vw
