/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#include <viewer/system.h>
#include "../include/system_private.h"

#if NN_CURRENT_VERSION_NUMBER >= NN_VERSION_NUMBER(4,0,0,0)
#pragma diag_suppress 1301 // padding inserted in struct.
#pragma diag_suppress 2530 // padding added to end of struct.
#endif
#ifdef __cplusplus
extern "C" {
#endif

extern bool VIEnableDimming( bool enable );

#ifdef __cplusplus
}/* extern "C" */
#endif



namespace vw     {
namespace system {


static const VideoSetting
DEFAULT_VIDEO_SETTING =
{
    nw::ut::FloatColor(0.25f,0.25f,0.25f,0.25f),// Screen clear color
    1.0f,                                       // Screen clear depth value
    60,                                         // Number of frames to update in a second
    NN_GX_DISPLAY0_WIDTH,                       // Upper screen horizontal size
    NN_GX_DISPLAY0_HEIGHT,                      // Upper screen vertical size
    NN_GX_DISPLAY1_WIDTH,                       // Lower screen horizontal size
    NN_GX_DISPLAY1_HEIGHT,                      // Lower screen vertical size
    nw::gfx::RENDER_COLOR_FORMAT_RGBA8,         // Screen pixel format
    nw::gfx::RENDER_DEPTH_FORMAT_24_STENCIL8,   // Screen Z format
};



// Video settings
static VideoSetting sVideoSetting;

static f32 sFPS = 0.0f;

static RenderSystem* s_pRenderSystem;
static u32 sScreenWidth = 400;
static u32 sScreenHeight = 240;


const int RENDER_TARGET_COUNT = 3;
typedef nw::ut::FixedSizeArray<nw::gfx::IRenderTarget*, RENDER_TARGET_COUNT> RenderTargetArray;
static RenderTargetArray s_RenderTargets;



/*---------------------------------------------------------------------------*
    Name:           Video_ChangeSetting

    Description:    Changes video settings

    Arguments:      setting     Video settings

    Returns:        None.
 *---------------------------------------------------------------------------*/
void
Video_ChangeSetting( const VideoSetting* setting )
{
    sVideoSetting = *setting;
    Video_ChangeFormat();
}

void
Video_GetSetting( VideoSetting* setting )
{
    *setting = sVideoSetting;
}

/*---------------------------------------------------------------------------*
    Name:           Video_Init

    Description:    Initializes video

    Arguments:      setting     Video settings

    Returns:        None.
 *---------------------------------------------------------------------------*/
void
Video_Init( const VideoSetting* setting /* = NULL */ )
{
    if ( setting == NULL )
    {
        setting = &DEFAULT_VIDEO_SETTING;
    }

    vw::system::RenderSystem::CommandDesc commandDescription;
    bool isEmbeddedViewer = vw::system::App::GetInstance().IsEmbeddedViewer();
    if( isEmbeddedViewer )
    {
        commandDescription.commandBufferSize            = 0x40000;
        commandDescription.reusableCommandBufferSize    = 0x100000;
    }
    else
    {
        commandDescription.commandBufferSize            = 0x400000;
        commandDescription.reusableCommandBufferSize    = 0x400000;
    }
    commandDescription.commandRequestCount              = 2048;
    commandDescription.reusableCommandRequestCount      = 2048;
    commandDescription.commandListCount                 = 1;    // Do not duplicate command list.

    // Process when graphics are not used by the built-in viewer
    if( !vw::system::App::GetUseGraphicsFlg() )
        commandDescription.commandListCount = 0;

    nw::os::IAllocator* pAllocator = Memory_GetSysAllocator();
    nw::os::IAllocator* pDeviceAllocator = Memory_GetSysAllocatorDevice();

    s_pRenderSystem = vw::system::RenderSystem::Create(pAllocator, commandDescription, isEmbeddedViewer);

    if( vw::system::App::GetUseMenuFlg() )
    {
        FontDrawer::Initialize();
        Draw2DHelper::Initialize(4096);
    }
    if( vw::system::App::GetUseGraphicsFlg() )
    {
        Draw3DHelper::Initialize();
    }
}

void
Video_InitMenu()
{
    FontDrawer::Initialize();
    Draw2DHelper::Initialize(4096);
}

void
Video_InitGraphics()
{
    vw::system::RenderSystem::CommandDesc commandDescription;
    commandDescription.commandBufferSize            = 0x40000;
    commandDescription.reusableCommandBufferSize    = 0x100000;
    commandDescription.commandRequestCount          = 2048;
    commandDescription.reusableCommandRequestCount  = 2048;
    commandDescription.commandListCount             = 1;    // Do not duplicate command list.
    s_pRenderSystem->InitCommandCache( commandDescription );

    Draw3DHelper::Initialize();
}

void
Video_Finalize()
{
    s_pRenderSystem->Destroy();

    if( vw::system::App::GetUseMenuFlg() )
    {
        FontDrawer::Finalize();
        Draw2DHelper::Finalize();
    }
    if( vw::system::App::GetUseGraphicsFlg() )
    {
        Draw3DHelper::Finalize();
    }
}

void
Video_SetFrameRate( u32 frameRate )
{
    sVideoSetting.frameRate = frameRate;
}

/*---------------------------------------------------------------------------*
  Name:         Video_GetFrameRate

  Description:  Gets frame rate settings.

  Arguments:    None.

  Returns:      Returns frame rate settings.
 *---------------------------------------------------------------------------*/
u32
Video_GetFrameRate( void )
{
    return sVideoSetting.frameRate;
}

void
Video_TransferBuffer()
{
    Video_SetRenderTarget(UPPER_SCREEN);
    s_pRenderSystem->TransferBuffer(UPPER_SCREEN);

#if !defined( NN_PCGX )
    Video_SetRenderTarget(EXTENSION_SCREEN);
    s_pRenderSystem->TransferBuffer(EXTENSION_SCREEN);
#endif

    Video_SetRenderTarget(LOWER_SCREEN);
    s_pRenderSystem->TransferBuffer(LOWER_SCREEN);
}

void
Video_WaitCommandList()
{
    s_pRenderSystem->WaitCommandList();
}

void
Video_PresentBuffer()
{
    s_pRenderSystem->PresentBuffer(UPPER_SCREEN | LOWER_SCREEN | EXTENSION_SCREEN);
}

void
Video_SwapBuffer()
{
    s_pRenderSystem->SwapBuffer( UPPER_SCREEN | LOWER_SCREEN | EXTENSION_SCREEN );
}

void
Video_WaitVSync()
{
    s_pRenderSystem->WaitVSync( UPPER_SCREEN | LOWER_SCREEN | EXTENSION_SCREEN );
}

nw::gfx::RenderContext*
Video_GetRenderContext()
{
    return s_pRenderSystem->GetRenderContext();
}

nw::gfx::IRenderTarget*
Video_GetRenderTarget(ScreenKind screenKind)
{
    u32 index = 0;
    switch(screenKind){
        case UPPER_SCREEN:
            index = 0;
            break;
        case EXTENSION_SCREEN:
            index = 1;
            break;
        case LOWER_SCREEN:
            index = 2;
            break;
    }

    return (index < s_RenderTargets.size())? s_RenderTargets[index] : NULL;
}

void
Video_SetRenderTarget(ScreenKind screenKind)
{
#if defined( NN_PCGX )
    nngxSplitDrawCmdlist();

    if (!gpGX->drawSkipMode)
    {
        switch (screenKind)
        {
        case UPPER_SCREEN:
        case EXTENSION_SCREEN:
            gpGX->GetGL()->pDeviceContext->OMSetRenderTargets(1, PCGXGetRenderViews(0), PCGXGetDepthStencilView(0));
            gpGX->SetViewport(0, 0, PCGX_DISP0_WID, PCGX_DISP0_HGT);
            break;
        case LOWER_SCREEN:
            gpGX->GetGL()->pDeviceContext->OMSetRenderTargets(1, PCGXGetRenderViews(1), PCGXGetDepthStencilView(1));
            gpGX->SetViewport(0, 0, PCGX_DISP1_WID, PCGX_DISP1_HGT);
            break;
        }
    }
#endif

    nw::gfx::IRenderTarget* renderTarget = Video_GetRenderTarget(screenKind);
    if( renderTarget != 0 )
    {
        s_pRenderSystem->SetRenderTarget(renderTarget);

        glBindFramebuffer( GL_FRAMEBUFFER, renderTarget->GetBufferObject() );

        // set screen size
        switch(screenKind){
        case UPPER_SCREEN:
        case EXTENSION_SCREEN:
            sScreenWidth  = 400;
            sScreenHeight = 240;
            break;
        case LOWER_SCREEN:
            sScreenWidth  = 320;
            sScreenHeight = 240;
            break;
        }
    }
}

void
Video_ClearBuffer()
{
    s_pRenderSystem->ClearBuffer( sVideoSetting.clearColor, sVideoSetting.clearZ );
}

void
Video_ClearDepth()
{
    s_pRenderSystem->ClearDepth( sVideoSetting.clearZ );
}

void
Video_ClearBuffers()
{
    Video_SetRenderTarget(UPPER_SCREEN);
    s_pRenderSystem->ClearBuffer(sVideoSetting.clearColor,sVideoSetting.clearZ);

    Video_SetRenderTarget(LOWER_SCREEN);
    s_pRenderSystem->ClearBuffer(sVideoSetting.clearColor,sVideoSetting.clearZ);

#if !defined( NN_PCGX )
    Video_SetRenderTarget(EXTENSION_SCREEN);
    s_pRenderSystem->ClearBuffer(sVideoSetting.clearColor,sVideoSetting.clearZ);
#endif
}

f32
Video_GetFPS()
{
    return sFPS;
}

void
Video_CalcFPS()
{
    f64 time = f64( s64(nn::os::Tick::GetSystemCurrent() )) / nn::os::Tick::TICKS_PER_SECOND;
    static f64 lastTime = time;
    static u32 FPS = 0;
    FPS++;

    if( time-lastTime >= 1.0 )
    {
        sFPS = FPS / (time-lastTime);
        FPS = 0;
        lastTime = time;
    }
}

u32
Video_GetScreenWidth()
{
    return sScreenWidth;
}

u32
Video_GetScreenHeight()
{
    return sScreenHeight;
}

void
Video_SetScreenSize(u32 width, u32 height)
{
    sScreenWidth  = width;
    sScreenHeight = height;
}

static nw::gfx::RenderColorFormat sColorFormat = nw::gfx::RENDER_COLOR_FORMAT_NONE;
static nw::gfx::RenderDepthFormat sDepthFormat = nw::gfx::RENDER_DEPTH_FORMAT_NONE;

void
Video_ChangeFormat()
{
    if( vw::system::App::GetInstance().IsEmbeddedViewer() )
        return;

    if(sVideoSetting.colorFormat == sColorFormat &&
       sVideoSetting.depthFormat == sDepthFormat)
       return;

    sColorFormat = sVideoSetting.colorFormat;
    sDepthFormat = sVideoSetting.depthFormat;

    nw::os::IAllocator* pAllocator = Memory_GetSysAllocator();
    nw::os::IAllocator* pDeviceAllocator = Memory_GetSysAllocatorDevice();

    vw::system::RenderSystem::ScreenDesc screenDescription;
    vw::system::DisplayBufferSwapper::Description& upperScreenDescription =
        screenDescription.upperScreenDescription;

    upperScreenDescription.screenKind = vw::system::UPPER_SCREEN;
    upperScreenDescription.width       = 240;
    upperScreenDescription.height      = 400;
    upperScreenDescription.bufferCount = 2;
    upperScreenDescription.format = (nw::gfx::RENDER_COLOR_FORMAT_RGBA8 == sColorFormat) ?
                                     nw::gfx::RENDER_COLOR_FORMAT_RGB8 : sColorFormat;

    vw::system::DisplayBufferSwapper::Description& lowerScreenDescription =
        screenDescription.lowerScreenDescription;

    lowerScreenDescription.screenKind = vw::system::LOWER_SCREEN;
    lowerScreenDescription.width       = 240;
    lowerScreenDescription.height      = 320;
    lowerScreenDescription.bufferCount = 2;
    lowerScreenDescription.format = (nw::gfx::RENDER_COLOR_FORMAT_RGBA8 == sColorFormat) ?
                                     nw::gfx::RENDER_COLOR_FORMAT_RGB8 : sColorFormat;

    for(u32 i=0; i<s_RenderTargets.Size(); ++i)
    {
        nw::gfx::SafeDestroy( s_RenderTargets[i] );
    }

    s_pRenderSystem->ChangeDisplayBuffer( screenDescription );

    // create render targets
    s_RenderTargets.Clear();

    s_RenderTargets.push_back(
        nw::gfx::IRenderTarget::Builder()
        .BufferSize(upperScreenDescription.width, upperScreenDescription.height)
        .ColorFormat( sColorFormat )
        .DepthFormat( sDepthFormat )
        .Create( pAllocator ));

    s_RenderTargets.push_back(
        nw::gfx::IRenderTarget::Builder()
        .BufferSize(upperScreenDescription.width, upperScreenDescription.height)
        .ColorFormat( sColorFormat )
        .DepthFormat( sDepthFormat )
        .Create( pAllocator ));

    s_RenderTargets.push_back(
        nw::gfx::IRenderTarget::Builder()
        .BufferSize(lowerScreenDescription.width, lowerScreenDescription.height)
        .ColorFormat( sColorFormat )
        .DepthFormat( sDepthFormat )
        .Create( pAllocator ));
    NW_ASSERT( !s_RenderTargets.empty() );
}

bool Video_CanAllocateA(u32 size)
{
    return vw::system::Memory::GraphicsMemoryAllocator::GetInstance()->CanAllocateA( size );
}

bool Video_CanAllocateB(u32 size)
{
    return vw::system::Memory::GraphicsMemoryAllocator::GetInstance()->CanAllocateB( size );
}


}   /* namespace system */
}   /* namespace vw     */
#if NN_CURRENT_VERSION_NUMBER >= NN_VERSION_NUMBER(4,0,0,0)
#pragma diag_default 2530 // padding added to end of struct.
#pragma diag_default 1301 // padding inserted in struct.
#endif
